{ Sheffield University Editor }
{ Copyright (c) University of Sheffield 1990 }

{******************************************}
{ GLOBAL VARIABLES FOR EDITOR AND VDUSETUP }
{******************************************}

const
   version_string = '8.6.7';            { version of editor and VDUSETUP }
   copyright_string = 'Copyright (c) University of Sheffield 1990';
   info_version = 2;                    { version of vdu_info record }
   term_gv_name   = '.SHEFF_TERMINAL';  { global variable name for vdu_type }
   report_gv_name = '.SHEFF_FILE_NAME';  { global variable name for -REPORT  }

%insert syscom>keys.ins.pas
%insert syscom>errd.ins.pas

{********************}
{ EDITOR FILE SYSTEM }
{********************}

const
   sam_type = 0;                        { values returned by open routines }
   dam_type = 1;
   cam_type = 7;

   { Check the following consts for consistency }
   block_max_words = 1200;              { block size in words }
   block_max_wordsp1 = 1201;            { ditto + 1 }
   block_max_chars = 2400;              { ditto in characters }
   block_tide_words = 1150;             { initially filled up to here }
   { block_tide_words should be around 95% of block_max_words }
   { and also greater than line_max_words so that a maximum length line }
   { will fit into a block }

   index_max = 50;                      { entries per index block }
   dynamic_segments = 2;                { heap segments to reserve }
   minimum_free_segments = 8;           { ditto to avoid using }
   index_block_max = 200;               { max index blocks }
   buffer_max = 1000;                   { number of files that can be edited }
                                        { simultaneously }

   { Check the following consts for consistency }
   line_max_words = 1024;               { maximum line size in words }
   line_max_chars = 2048;               { ditto in characters }
   line_max_charsm1 = 2047;             { ditto - 1}
   line_max_charsp1 = 2049;             { and + 1}

   tree_max_chars = 256;                { maximum tree length }
   locate_list_length_max = 100;        { maximum number of separate elements }
                                        { in a search string }
   word_max_chars = 8;                  { max chars in command names etc. }
   num_string_buffers = 10;
   tabs_max = 30;                       { maximum tab stop positions }
   count_width_max = 80;                { maximum width for mode count }
   large_int = 200000000;
   fairly_large_int = 100000000;
   first_print_char = ' ';              { first and last printable characters }
   last_print_char = '~';
   { codes for various ASCII characters; must be >= 128 because of ECS }
   nul_code = 200b;
   bs_code  = 210b;
   lf_code  = 212b;
   ff_code  = 214b;
   cr_code  = 215b;
   dle_code = 220b;
   dc1_code = 221b;
   dc3_code = 223b;
   can_code = 230b;
   esc_code = 233b;
   del_code = 377b;
   double_newline = -72566b;            { two newline's in a shortint }

type
   smallint = 0..255;
   data_chars_range = 1..block_max_chars;
   data_chars_type = packed array[data_chars_range] of char;
   data_words_range = 1..block_max_words;
   data_words0_range = 0..block_max_words;
   data_words0p1_range = 0..block_max_wordsp1;
   data_words_type = array[data_words_range] of shortint;

   tree_range = 1..tree_max_chars;
   tree0_range = 0..tree_max_chars;
   tree_chars = packed array[tree_range] of char;
   tree_string = string[tree_max_chars];
   tree_type = record                   { treat as string or len/chars }
      case boolean of
         false: (strng: tree_string);
         true:  (len: shortint; chars: tree_chars)
      end;
   ptr_tree_type = ^ tree_type;

   line_words_range = 1..line_max_words;
   line_words_type = array[line_words_range] of shortint;
   line_chars_range = 1..line_max_chars;
   line_chars_plus_range = 1..line_max_charsp1;
   line_chars0_range = 0..line_max_chars;
   line_chars_type = array[line_chars_range] of char;
   long_chars = packed array[1..line_max_chars] of char;
   long_string = string[line_max_chars];

   word_range = 1..word_max_chars;
   word0_range = 0..word_max_chars;
   word = packed array[word_range] of char;
   word_string = string[word_max_chars];

   tabs_range = 1..tabs_max;
   tabs0_range = 0..tabs_max;
   tabs_array = array[tabs_range] of 0..line_max_chars;

   three_state = (is_false, is_true, is_unknown);
   char3 = packed array[1..3] of char;
   char4 = packed array[1..4] of char;
   char6 = packed array[1..6] of char;
   char10 = packed array[1..10] of char;
   char12 = packed array[1..12] of char;
   char16 = packed array[1..16] of char;
   char32 = packed array[1..32] of char;
   string8 = string[8];
   string12 = string[12];
   string32 = string[32];
   string80 = string[80];
   string128 = string[128];
   string_buffers0_range = 0..num_string_buffers;
   string_buffers2_range = -2..num_string_buffers;
   search_type = (searching, found, not_found);
   direction_type = (go_forwards, go_backwards);
   set_of_char = set of char;
   word_bits = (bit1, bit2, bit3, bit4, bit5, bit6, bit7, bit8,
                bit9, bit10, bit11, bit12, bit13, bit14, bit15, bit16);
   word_as_bits = set of word_bits;     { for tty$rs etc. }

   case_patterns = (mixed_case, all_upper, all_lower, first_upper);

   file_block = record
      case boolean of
         true:  (data_words: data_words_type);
         false: (data_chars: data_chars_type)
      end;

   ptr_file_block = ^file_block;

   index_entry = record                 { block index entry }
      block_lines: integer;             { lines in the block }
      block_ad: ptr_file_block;         { pointer to block or "nil" if block }
                                        { is offline in paging file }
      file_key: integer;                { record number in paging file }
      paged_out_copy_ok: boolean;       { true if copy in paging file is }
                                        { up to date }
      split_start: integer;             { word subscript of split in block }
      split_end: integer                {   "      "     of part after split }
   end;

   index_range = 1..index_max;
   index0_range = 0..index_max;
   ptr_index_block = ^index_block;
   index_block = record                 { a block of index entries for a buffer}
      entries: array[index_range] of index_entry;
      top_entry: index0_range;          { number of entries present }
      previous, next: ptr_index_block   { pointers to rest of index if any }
   end;

   buffer_range = 1..buffer_max;
   buffer0_range = 0..buffer_max;
   line_status_type = (top, bottom, normal);
   file_places = record
      saved_line: integer;
      saved_state: line_status_type;
      saved_position: line_chars_range
   end;

   load_use_type = (use_primos, use_buffer);
   primos_file = record                 { Primos file information }
      treename: tree_type;
      file_is_open: boolean;            { true if open }
      file_type: integer;               { 0=SAM, 1=DAM, 7=CAM }
      original_pathname: boolean;       { still pathname used in ED comm etc. }
      rename_done: boolean;             { rename to .BAKn already done }
      case load_use: load_use_type of
         use_primos: (                  { using a real Primos file }
            truncate_later: boolean;    { true when file is to be truncated }
                                        { after closing }
            unit: shortint;             { Primos file unit number }
            prepos: integer             { position for use with k$prer in }
                                        { next call to PRWF$$ }
            );
         use_buffer: (                  { using a memory buffer }
            load_buffer_number: buffer_range;  { the buffer number }
            load_index: ptr_index_block;  { index block pointer }
            load_index_sub: index_range  { entry subscript in index block }
            )
   end;


   {************************}
   { THE EDITOR FILE BUFFER }
   {************************}

   ptr_buffer_info = ^buffer_info;
   buffer_info = record                 { all the information about one file }
                                        { being edited }
      first_index,                      { -> first index block }
      current_index: ptr_index_block;   { -> current index block }
      index_sub: index_range;           { subscript in current index block }
      block_number: integer;            { overall block number in file }
      blocks: integer;                  { total blocks in file }
      memory_blocks: integer;           { total ditto actually in memory }
      lines_above,                      { total lines in blocks above current }
      last_line_in_block: integer;      { last line in current block - not }
                                        { always kept up to date }

      { data for current line }

      line_text: line_chars_type;       { text of line - uncompressed }
      line_number: integer;             { line number in file }
      line_length: line_chars0_range;   { length in chars excluding newline }
      line_position: line_chars_range;  { position within current line }
      line_status: line_status_type;
      line_exists: boolean;             { true if line not deleted }
      line_altered: boolean;            { true if line differs from file }
      line_start,                       { start word index in file block }
      line_end: data_words0p1_range;    { end     "    "    "   "    "   }

      marked_line: integer;             { when set to the number of a line in }
                                        { the workfile, the filesystem will }
                                        { keep it up to date, allowing for }
                                        { insertions and deletions. }
      iw_marked_line: integer;          { ditto used by IW command }
      point_line_pos: line_chars0_range;  { line position marked by character }
                                        { marker in window editor }
      point_line_no: integer;           { corresponding line number }
      file_altered: boolean;            { true when any alterations to file. }
                                        { this is kept up to date by the low }
                                        { level routines; high level routines }
                                        { just need to set "line_altered" when }
                                        { necessary. }
      file_truncated: boolean;          { true when file closed prematurely }
      main_file: primos_file;           { the details of the file being edited }
      must_redraw_window: boolean;      { set by commands which drastically }
                                        { change the file to inform the window }
                                        { editor that the window no longer }
                                        { corresponds to the workfile }
      some_lines_changed: boolean       { one or more lines in this buffer }
                                        { changed and must be redrawn }
   end { buffer_info } ;

   buffer_pointer_rec = record
      bptr: ptr_buffer_info;            { pointer to buffer or nil }
      pred_buff,                        { predecessor buffer - mode_chain }
      succ_buff: buffer0_range          { successor   buffer              }
   end { buffer_pointer_rec } ;

var
   new_block_max: integer;              { max file blocks in memory at once }
   current_buffer_number: buffer_range;  { the number of buffer being edited }
   target_buffer_number: buffer_range;  { the number of the last buffer robbed }
                                        { of a memory block, kept to ensure }
                                        { they are robbed in turn }
   current_buffer: buffer_info;         { its details are kept in globals to }
                                        { speed access }
   multiple_buffers_used: boolean;      { true if user has made use of more }
                                        { than one buffer }
   lowest_empty_buffer: integer;        { next buffer for NEW to try }
   buffer_pointers:                     { pointers to buffers }
      array[buffer_range] of buffer_pointer_rec;
   first_buffer_in_chain,               { info for mode_chain }
   last_buffer_in_chain,
   first_chain_setting: buffer0_range;
   index_block_count: integer;          { count of index blocks in memory }
   free_blocks,                         { chain of index entries containing }
                                        { unused file memory blocks }
   free_keys: ptr_index_block;          { ditto free paging file keys }
   paging_file: file of file_block;     { repository for paged out file blocks }
   paging_file_open: boolean;           { true if open }
   paging_file_name: tree_string;       { name when open }
   blocks_paged_out: integer;
   top_paging_key: integer;             { records currently in paging file }
   new_block_count: integer;            { count of blocks created by "new" - }
                                        { must not exceed "new_block_max" to }
                                        { avoid filling the heap }
   avoid_index: ptr_index_block;        { index block not to be attacked when }
                                        { seeking file blocks to be paged out }

{*************************}
{ EDITOR COMMAND ANALYSIS }
{*************************}

const
   init_wild = '!';
   init_blanks = '#';
   init_tab = '\';
   init_escape = '^';
   init_semicolon = ';';
   init_counter = '@';

type
   edline_type = record                 { edit line or string buffer }
      ed_text: line_chars_type;         { the text of the line }
      text_length: line_chars0_range;   { length of text excluding newline }
      being_obeyed: boolean             { true if currently being executed }
   end;
   ptr_edline = ^edline_type;

   commands = (                         { editor commands - must be in same }
                                        { order as literal command names }
      blankc   , findc    , locatec  , nextc    ,
      changec  , retypec  , overlayc , deletec  ,
      printc   , pointc   , pprintc  , bottomc  ,
      topc     , nfindc   , nlocatec , appendc  ,
      oopsc    , modifyc  , wherec   , windowc  ,
      insertc  , iwc      , ibc      , xeqc     ,
      gmodifyc , helpc    , verifyc  , briefc   ,
      symbolc  , filec    , savec    , quitc    ,
      qfc      , loadc    , dunloadc , unloadc  ,
      tabsetc  , modec    , erasec   , killc    ,
      fnamec   , movec    , psymbolc , pausec   ,
      inputc   , lineszc  , outputc  , ptabsetc ,
      punchc   , markc    , editc    , terminalc,
      macclearc, macloadc , macprintc, macsavec ,
      stylec   , textc    , bufferc  , saveallc ,
      fileallc , delbufc  , versionc , punctc   ,
      alphabetc, mergec   , extractc , replacec ,
      statsc   , newc     , splitc   , centrec  ,
      reloadc  , expandc  , sizec    , obeyc    ,
      spellc   , savestatc, loadstatc, cominputc,
      nextbuffc, prevbuffc);

   actions =                            { used in "many_commands" }
      ( print_act, delete_act, unload_act, backwards_act, case_act,
        ident_act, locate_act, find_act, not_act, mark_act, start_act);
   set_of_actions = set of actions;

   symbols =                            { editor special symbols }
      ( skill     , serase    , swild     , sblanks   ,
        stab      , sescape   , scprompt  , sdprompt  ,
        ssemicolon, scounter  , svisible  , swpc      ,
        sverbatim );

   modes =                              { mode commands }
      ( mckpar  , mcolumn , mcount  , minfo   ,
        mnumber , mprompt , msemi   , mindent ,
        mquiet  , mmargin , mdirect , msafety ,
        mcomo   , minsert , mdisp   , mchain  ,
        mbackup , mmsg    , mlon    , mtest   );

   elements =                           { the different elements of a string }
      (el_string, blanks, wild, nothing);
   element_entry = record
      element: elements;
      arg1, arg2: integer
   end;

   style_type = record                  { text processing style information }
      left_margin,                      { left and }
      right_margin,                     { right margin columns }
      para_indent,                      { indentation (+ve or -ve) for paras }
      para_blanks: integer;             { blank lines between paras }
      adjust: boolean                   { adjust or fill }
   end;

   { record to contain all command line args except for pathnames }

   command_line_args_type = record
      init_sw: boolean; init_path: string128;
      help_sw: boolean; help_key: string80;
      terminal_sw: boolean; terminal_type: integer;
      tidy_sw: boolean; tidy_wildcard: string32;
      sort_dtm_sw,
      reverse_sw,
      report_sw,
      window_sw,
      no_line_sw,
      screen_sw,
      no_init_sw,
      no_primos_sw,
      testing_sw: boolean
   end;

   sel_type = record                    { selection criteria for DIR$SE }
      version_no: shortint;
      wild_ptr: integer;
      wild_count: shortint;
      desired_types: packed record
         dirs, seg_dirs, files, access_cats, rbf: boolean;
         filler: 0..2047
      end;
      mdb, mda, svb, sva: integer
   end;

   dtm_type = integer;
   short4 = array[1..4] of shortint;

   dir_entry = record
      ecw: shortint;
      entryname: string32;
      protection: shortint;
      file_info: shortint;
      dtm: dtm_type;
      non_default_acl: shortint;
      logical_type: shortint;
      trunc: shortint;
      dts: integer
   end;

var
   command_line_args: command_line_args_type;  { args except for pathnames }
   newline: char;
   load_file: primos_file;              { a file being loaded }
   edit_mode,                           { true if in EDIT mode }
   edit_mode_after_init: boolean;       { value for edit_mode when init file }
                                        { exhausted }
   ld_buffer: buffer_range;             { buffer that contains current ld file }
                                        { for -TIDY }
   ld_pathname: long_string;            { pathname to directory for -TIDY }
   original_length: integer;            { length of ld_pathname in orig dir }
   ld_cpl_filename: long_string;        { name of CPL file to obey or '' }
   main_state: (running, command_error,  { state of obeying commands }
      top_or_bottom, quit_signal, quit_comm);
   obey_commands: boolean;              { false if command line error or -HELP }
   return_code: integer;                { main program return code }
   reading_command_line: boolean;       { used in quit on-unit }

   cominput_on: boolean;                { editor simulated cominput in action }
   cominput_buffer: 0..buffer_max;      { cominput buffer; 0 => file }
   cominput_lineno: integer;            { line number of cominput }
   cominput_file: text;                 { Primos cominput file }
   obeying_loadstate: boolean;          { true if LOADSTATE not COMINPUT }

   must_redraw_screen: boolean;         { set by commands which need the whole }
                                        { screen to be refreshed }
   first_changed_line,                  { set by CHANGE etc to indicate range }
   last_changed_line: integer;          { of lines changed }
   split_command_used: boolean;         { true after SPLIT command used }
   switch_buffer_used: boolean;         { true after buffer changed }
   display_buffer_name: boolean;        { display new buffer if true }
   suppress_output: boolean;            { need to redraw screen when the user }
                                        { ceases to type ahead }
   in_window_code: boolean;             { true when a command is being obeyed }
                                        { from within the window code }
   in_help_system: boolean;             { true when in help system }
   message: long_string;                { place where messages are built }
   message_lines_to_clear: integer;     { number of message lines on screen }
   continue_screen_messages: boolean;   { carry on printing }
   copy_suppress: boolean;              { true after an XEQ with no parameter }
   replace_command_given: boolean;      { true if REPLACE given }
   replace_ptr_1,                       { replacement strings on heap }
   replace_ptr_2: ptr_tree_type;
   replace_options: alfa;               { list of options }
   spell_command_given: boolean;        { true if SPELL given }
   wi_oops_count: integer;
   last_match_position: line_chars0_range;  { last column matched by Locate }
   input_state: (                       { source of command input }
      not_open,
      default_file,
      init_file,
      user_input);
   test_line_size: integer;             { set by LINESZ command }

   mode_verify: boolean;                { verify as opposed to brief }
   mode_ckpar,                          { mode settings }
   mode_column,
   mode_count,
   mode_info,
   mode_number,
   mode_prompt,
   mode_semi,
   mode_indent,
   mode_quiet,
   mode_margin,
   mode_direct,
   mode_safety,
   mode_como,
   mode_insert,
   mode_disp,
   mode_chain,
   mode_backup,
   mode_msg,
   mode_lon,
   mode_test: boolean;                  { set true by -TESTING on command line }
                                        { or MODE TEST - switches on test code }
   count_value, count_increment,        { mode count information }
   count_width: integer;
   count_type: (zblank, zprint, zsuppress);
   margin_column: line_chars_range;
   max_backup: 1..9;
   number_of_tabs: tabs0_range;         { TAB information }
   htabs: tabs_array;
   screen_tabs_set: boolean;            { true when hardware tabs set on VDU }
   number_of_ptabs: tabs0_range;
   ptabs: tabs_array;

   { information about the search string for FIND, LOCATE etc. }

   locate_string_exists: boolean;
   locate_element_list: array[1..locate_list_length_max] of element_entry;
   locate_list_length: 0..locate_list_length_max;  { number of entries }
   locate_zero: boolean;
   locate_string,                       { contains string(s) to be tested }
   locate_upper,                        { upper and lower case versions }
   locate_lower:  line_chars_type;
   locate_use_ql: boolean;              { able to use "quick_locate" }
   locate_ql_start, locate_ql_end: integer;  { location of quick_locate string }

   { information about the search string for CHANGE and MODIFY }

   change_string_exists: boolean;
   change_zero: boolean;                { zero length }
   change_string: line_chars_type;      { contains old and new strings }
   change_use_ql: boolean;              { able to use "quick_locate" }
   change_ql_start,                     { location of "quick_locate" string }
   change_ql_end: integer;
   change_length1,                      { length of old string }
   change_length2: line_chars0_range;   { length of new string }
   change_same_length: boolean;         { line shifting required? }
   spaces_to_add: line_chars0_range;    { add this many spaces (M only) }
   change_general,                      { 'G' argument given }
   change_ident,                        { 'I' argument given }
   change_to_mark,                      { 'M' argument given }
   change_star: boolean;                { '*' argument given }
   change_lines: integer;               { number of lines to work on }
   change_backwards: boolean;           { backwards? }
   style_rec: style_type;               { text processing style information }

   special: array[symbols] of char;     { special symbols }
   illegal_chars,                       { these not to be used as symbols }
   symbol_chars,                        { current set of symbols }
   string_terminators,                  { terminate LOCATE strings etc. }
   string_specials,                     { treat specially in ditto }
   input_specials,                      { treat specially in INPUT etc. }
   overlay_specials,                    { ditto OVERLAY }
   gmod_input_specials,                 { ditto for use in GMODIFY }
   gmod_overlay_specials,
   print_specials: set_of_char;         { specials treatment for printing }
   punctuation: set_of_char;            { used by IL command etc. }
   upper_alphabet,                      { used by DEL A/a in screen editor }
   lower_alphabet: set_of_char;
   ascii_printable: set_of_char;        { ASCII codes 240b .. 376b }

   edit_line: edline_type;              { main edit line }
   string_buffers:                      { editor string buffers }
      array[string_buffers0_range] of ptr_edline;
   str_index: string_buffers0_range;    { [0] is used as edit line copy }
   ed_com_line: edline_type;            { edit line used by window code }
   ed_com_line_valid: (nothing_defined, ed_command);  { is it valid? }

   blank_word: word;
   command_names:                       { literal command names }
      array[commands] of word;
   action_names:
      array[actions] of char;
   symbol_names:                        { literal symbol names }
      array[symbols] of word;
   mode_names:                          { literal mode names }
      array[modes] of word;
   control_names:                       { literal control character names }
      array[200b..237b] of char3;

value
   test_line_size = line_max_charsm1;
   mode_verify = true;
   mode_ckpar = false;
   mode_column = false;
   mode_count = false;
   mode_info = false;
   mode_number = false;
   mode_prompt = true;
   mode_semi = true;
   mode_indent = false;
   mode_quiet = false;
   mode_margin = false;
   mode_direct = false;
   mode_safety = false;
   mode_como = false;
   mode_insert = false;
   mode_disp = false;
   mode_chain = false;
   mode_backup = true;
   mode_msg = false;
   mode_lon = false;
   mode_test = false;
   count_value = 1;                     { mode count defaults }
   count_increment = 1;
   count_width = 5;
   count_type = zprint;
   margin_column = 78;
   max_backup = 1;
   number_of_tabs = 4;                  { default tab stops }
   htabs = (6, 12, 30, 80,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0);
   number_of_ptabs = 0;
   style_rec = (0, 79, 0, 1, true);

   special = (' ', ' ', init_wild, init_blanks, init_tab, init_escape,
              '$', '&', init_semicolon, init_counter, ' ', '.', '~');

   blank_word = '        ';

   command_names = (
{     '        ', 'FIND    ', 'LOCATE  ', 'NEXT    ',  {***std***}
{     'CHANGE  ', 'RETYPE  ', 'OVERLAY ', 'DELETE  ',  {***std***}
{     'PRINT   ', 'POINT   ', 'PPRINT  ', 'BOTTOM  ',  {***std***}
{     'TOP     ', 'NFIND   ', 'NLOCATE ', 'APPEND  ',  {***std***}
{     'OOPS    ', 'MODIFY  ', 'WHERE   ', 'WINDOW  ',  {***std***}
{     'INSERT  ', 'IW      ', 'IB      ', 'XEQ     ',  {***std***}
{     'GMODIFY ', 'HELP    ', 'VERIFY  ', 'BRIEF   ',  {***std***}
{     'SYMBOL  ', 'FILE    ', 'SAVE    ', 'QUIT    ',  {***std***}
{     'QF      ', 'LOAD    ', 'DUNLOAD ', 'UNLOAD  ',  {***std***}
{     'TABSET  ', 'MODE    ', 'ERASE   ', 'KILL    ',  {***std***}
{     'FNAME   ', 'MOVE    ', 'PSYMBOL ', 'PAUSE   ',  {***std***}
{     'INPUT   ', 'LINESZ  ', 'OUTPUT  ', 'PTABSET ',  {***std***}
      '||||||||', '||||||||', '||||||||', '||||||||',  {***ds***}
      '||||||||', '||||||||', '||||||||', '||||||||',  {***ds***}
      '||||||||', '||||||||', '||||||||', '||||||||',  {***ds***}
      '||||||||', '||||||||', '||||||||', '||||||||',  {***ds***}
      '||||||||', '||||||||', '||||||||', '||||||||',  {***ds***}
      '||||||||', '||||||||', '||||||||', '||||||||',  {***ds***}
      '||||||||', '||||||||', '||||||||', '||||||||',  {***ds***}
      '||||||||', '||||||||', '||||||||', '||||||||',  {***ds***}
      '||||||||', '||||||||', '||||||||', '||||||||',  {***ds***}
      '||||||||', '||||||||', '||||||||', '||||||||',  {***ds***}
      '||||||||', '||||||||', '||||||||', '||||||||',  {***ds***}
      '||||||||', '||||||||', '||||||||', '||||||||',  {***ds***}
      'PUNCH   ', 'MARK    ', 'EDIT    ', 'TERMINAL',
      'MACCLEAR', 'MACLOAD ', 'MACPRINT', 'MACSAVE ',
      'STYLE   ', 'TEXT    ', 'BUFFER  ', 'SAVEALL ',
      'FILEALL ', 'DELBUF  ', 'VERSION ', 'PUNCT   ',
      'ALPHABET', 'MERGE   ', 'EXTRACT ', 'REPLACE ',
      'STATS   ', 'NEW     ', 'SPLIT   ', 'CENTRE  ',
      'RELOAD  ', 'EXPAND  ', 'SIZE    ', 'OBEY    ',
      'SPELL   ', 'SAVESTAT', 'LOADSTAT', 'COMINPUT',
      'NEXTBUFF', 'PREVBUFF');

   action_names =
      ( 'P', 'D', 'U', 'B', 'C', 'I', 'L', 'F', 'N', 'M', 'S');

   symbol_names = (
      'KILL    ', 'ERASE   ', 'WILD    ', 'BLANKS  ',
      'TAB     ', 'ESCAPE  ', 'CPROMPT ', 'DPROMPT ',
      'SEMICOLO', 'COUNTER ', 'VISIBLE ', 'WPC     ',
      'VERBATIM');

   mode_names = (
      'CKPAR   ', 'COLUMN  ', 'COUNT   ', 'INFO    ',
      'NUMBER  ', 'PROMPT  ', 'SEMI    ', 'INDENT  ',
      'QUIET   ', 'MARGIN  ', 'DIRECT  ', 'SAFETY  ',
      'COMO    ', 'INSERT  ', 'DISP    ', 'CHAIN   ',
      'BACKUP  ', 'MSG     ', 'LON     ', 'TEST    ');

   control_names = (
      'NUL', 'SOH', 'STX', 'ETX', 'EOT', 'ENQ', 'ACK', 'BEL',
      'BS ', 'HT ', 'LF ', 'VT ', 'FF ', 'CR ', 'SO ', 'SI ',
      'DLE', 'DC1', 'DC2', 'DC3', 'DC4', 'NAK', 'SYN', 'ETB',
      'CAN', 'EM ', 'SUB', 'ESC', 'FS ', 'GS ', 'RS ', 'US ');


{******************************}
{ SPELLING CHECKER INFORMATION }
{******************************}

const
   spell_word_max = 32;
   first_suffix = 91;                   { first ASCII character after 'Z' }
   n_dict_blocks = 5;
   block_bytes_max = 130000;            { almost one segment }

type
   spell_alphabet = 'A' .. 'Z';
   spell_word_string = string[spell_word_max];
   char_type = (punct, upper, lower, extra);
   suffix_range = first_suffix..255;
   suffix_pairs = record
      suffix_start, suffix_count: 0..maxint
   end;

   dict_block = packed array[1..block_bytes_max] of smallint;
   dict_ptr = ^ dict_block;

   offset_entry = record
      any_words: boolean;
      dict_block_no: 1..n_dict_blocks;
      dict_block_offset: integer
   end;

var
   spell_char_array: array[128..255] of char_type;
   suffixes: array[suffix_range] of spell_word_string;
   suffix_map: array[1..8, spell_alphabet] of suffix_pairs;
   null_suffix: suffix_range;
   dict_blocks: array[1..n_dict_blocks] of dict_ptr;
   dictionary_online: boolean;
   spell_stop_line: integer;
   offset_table: array[spell_alphabet, spell_alphabet] of offset_entry;

value
   spell_char_array = (
      punct, punct, punct, punct, punct, punct, punct, punct,
      punct, punct, punct, punct, punct, punct, punct, punct,
      punct, punct, punct, punct, punct, punct, punct, punct,
      punct, punct, punct, punct, punct, punct, punct, punct,
      punct, punct, punct, punct, punct, punct, punct, punct,
      punct, punct, punct, punct, punct, punct, punct, punct,
      extra, extra, extra, extra, extra, extra, extra, extra,
      extra, extra, punct, punct, punct, punct, punct, punct,
      punct, upper, upper, upper, upper, upper, upper, upper,
      upper, upper, upper, upper, upper, upper, upper, upper,
      upper, upper, upper, upper, upper, upper, upper, upper,
      upper, upper, upper, punct, punct, punct, punct, extra,
      punct, lower, lower, lower, lower, lower, lower, lower,
      lower, lower, lower, lower, lower, lower, lower, lower,
      lower, lower, lower, lower, lower, lower, lower, lower,
      lower, lower, lower, punct, punct, punct, punct, punct);


{***********************************}
{ VDU AND WINDOW EDITOR INFORMATION }
{***********************************}

const
   cols_max = 1023;                     { vdu screen size max }
   cols_maxp1 = 1024;
   rows_max = 100;
   rows_maxp1 = 101;

   window_max = 8;                      { maximum number of windows }

   screen_margin = 4;                   { try to keep cursor away from top and }
                                        { bottom this many lines of screen }
   vdu_chars_max = 2400;                { available area in vdu_info record is }
   vdu_entry_max = 300;                 { used for characters (one per word) or}
                                        { "char_plus_entry"s. Current length }
                                        { of "char_plus_entry" is 8 words so }
                                        { vdu_chars_max = vdu_entry_max * 8 }
   macro_codes_max = 100;               { codes per macro block }
   macro_total_max = 500;               { overall limit per macro }
   seq_max = 10;                        { chars/block in sequence }

type
   cont_type = (rekord, half_duplex, xonrec, reinstate);

   controls = (
      backtab     , scrolloff   , charinsert  , chardelete  ,
      clearatab   , cleartabs   , clearscreen , scrollon    ,
      ctrlcodeoff , ctrlcodeon  , cursorup    , cursordown  ,
      cursorleft  , cursorright , foldoff     , foldon      ,
      horiztab    , arrowoff    , arrowon     , lineerase   ,
      lineinsert  , linedelete  , pageerase   , reverself   ,
      settab      , messoff     , messon      );

   col_range = 1..cols_max;
   row_range = 1..rows_max;
   col0_range = 0..cols_max;
   row0_range = 0..rows_max;
   col0p1_range = 0..cols_maxp1;
   row0p1_range = 0..rows_maxp1;

   window_range = 1..window_max;
   window0_range = 0..window_max;

   { all information about a window on the screen }

   window_rec = record
      { size of window is recorded as inclusive row and col ranges }
      window_top, window_bottom, window_length: row_range;
      window_left, window_right, window_width: col_range;
      window_at_right,                  { this window is a right hand edge }
      window_at_bottom,                 { ditto at whole bottom edge }
      window_full_width: boolean;       { ditto full width of screen }
      window_buffer: buffer_range;      { file buffer in window }
      window_mode: boolean;             { true for WI, false for IW }
      window_status: array[row_range] of
         record
            file_line: integer;         { editor workfile line number }
            offset: 0..maxint;          { start offset within line }
            row_status: (row_normal, row_null);
            row_drawn: boolean          { drawn on screen yet? }
         end;
      cursor_row: row0p1_range;         { position of cursor relative to }
      cursor_col: col0p1_range          { window }
   end;

   { The order of the following sumbolic type must not be changed because }
   { users have its ord stored in their macro files.  Add any new ones to }
   { the end }

   action_proc = (
      emptyact     , sequence     , breakline    , carret       ,  {**order**}
      deleteonech  , deleteonel   , dofile       , doquit       ,  {**order**}
      dosave       , downmany     , dunlfile     , duplicate    ,  {**order**}
      edcom        , emacro       , eraseline    , erroract     ,  {**order**}
      exitwi       , formtext     , godown       , godownx      ,  {**order**}
      gohome       , goleft       , goleftx      , goright      ,  {**order**}
      gorightx     , goup         , goupx        , helpact      ,  {**order**}
      insertmode   , insertonech  , insertonel   , joinline     ,  {**order**}
      lfchar       , lfcharx      , lfile        , lowcase      ,  {**order**}
      markon       , markoff      , negmacro     , nullact      ,  {**order**}
      openit       , closeit      , pointchon    , pointchoff   ,  {**order**}
      posmacro     , redrawl      , redraws      , repeatit     ,  {**order**}
      revlfchar    , revlfcharx   , smacro       , dmacro       ,  {**order**}
      ispaces      , dspaces      , storeonech   , tableft      ,  {**order**}
      tabright     , tabstopon    , tabstopoff   , unlfile      ,  {**order**}
      upcase       , upmany       , wiswitch     , xcomm        ,  {**order**}
      zap          , downscreen   , upscreen     , litchar      ,  {**order**}
      sinthost     , ssuspend     , srestart     , sleftupd     ,  {**order**}
      sfirstns     , salastns     , seraprev     , seraline     ,  {**order**}
      seraright    , shomefld     , snewline     , nextwind     ,  {**order**}
      prevwind    );  {**order**}

   set_of_action_proc = set of action_proc;

   repair_type = (ok, rest1, rest2, rest3, rest4, func);

   ware_type = (hardware , software);

   vdu_char_range = 1..vdu_chars_max;
   vdu_char0_range = 0..vdu_chars_max;

   vdu_char_pointer = record
      start_index: vdu_char_range;      { start location and length of chars }
      string_length: vdu_char0_range    { in char_array field of vdu_info }
   end;

   vdu_entry_range = 1..vdu_entry_max;
   vdu_entry0_range = 0..vdu_entry_max;

   vdu_entry_pointer = record
      start_index: vdu_entry_range;
      number_of_entries: vdu_entry0_range
   end;

   { Description of the input action for a character }

   char_entry = record
      char_action: action_proc;
      case boolean of
         false: (char_pointer: vdu_entry_pointer);
         true:  (char_repair: repair_type;
                 char_ware: ware_type;
                 char_reply: vdu_char_pointer)
   end;

   { Ditto but including the character in question }

   char_plus_entry = record
      entry_char: char;
      entry_rest: char_entry
   end;

   ptr_char_table = ^char_table;

   seq_range = 1..seq_max;

   { expanded form of character sequence used in VDUSETUP }

   ptr_seq_block = ^seq_block;
   seq_block = record
      next_seq_block: ptr_seq_block;
      seq_chs: array[seq_range] of integer
   end;

   sequences = record
      seq_len: 0..maxint;
      first_seq_block: ptr_seq_block
   end;

   { Description of input character used in VDUSETUP }

   full_entry = record
      full_action: action_proc;
      full_pointer: vdu_entry_pointer;
      full_repair: repair_type;
      full_ware: ware_type;
      full_reply: sequences;
      table_ptr: ptr_char_table
   end;

   char_table = array[char] of full_entry;

   {***************************}
   { THE VDU DESCRIPTION TABLE }
   {***************************}

   ptr_vdu_info = ^vdu_info;
   vdu_info = record                    { all the information about the vdu }
      stored_version: integer;          { version number of vdu_info structure }
      vdu_identification: string[80];
      control_array: array[controls] of vdu_char_pointer;
      reset_string, exit_string, cursor_string: vdu_char_pointer;
      two_way: record
         case boolean of
            false: (char_array: array[vdu_char_range] of shortint);
            true:  (entry_array: array[vdu_entry_range] of char_plus_entry)
         end { two_way } ;
      mess_type1, arrow_type1: boolean;
      func_expect_return: boolean;
      func_repair: repair_type;
      cols_per_row: col_range;
      rows_per_page: row_range;
      home_location: record
         home_row: row0p1_range;
         home_col: col0p1_range
      end;
      right_pos_ok,
      right_prt_ok,
      left_pos_ok,
      up_pos_ok,
      down_pos_ok,
      data_protect: boolean;
      posmacro_repair, negmacro_repair: repair_type;
      posmacro_reply, negmacro_reply: vdu_char_pointer;
      action_array: array[128..255] of char_entry
   end;

   mb_range = 1..macro_codes_max;
   mb0_range = 0..macro_codes_max;

   looprec = record                     { information about repeat command }
      end_condition: (once_only, ntimes,
         stop_after_mark, stop_before_mark,
         stop_eq_char, stop_ne_char);
      loop_counter: integer;
      loop_test_char: char
   end;

   ptr_macro_buffer = ^macro_buffer;
   macro_buffer = record                { one block of macro text }
      next: ptr_macro_buffer;
      code_count: mb0_range;
      macro_text: array[mb_range] of shortint
   end;

   ptr_macro_desc = ^macro_desc;
   macro_desc = record                  { macro descriptor }
      code_start,                       { start of macro code }
      current_buf: ptr_macro_buffer;    { current place in macro execution }
      codes_used: mb0_range;            { code spaces used in current buffer }
      total_count: integer;             { total codes in macro }
      macro_being_obeyed: boolean;      { true while being obeyed }
      loop_spec: looprec;               { repetition information }
      previous: ptr_macro_desc          { -> calling macro if any }
   end;

   macro_type = (posm, negm);
   macro_pair = record
      pointers: array[macro_type] of ptr_macro_desc
   end;

var
   duplex_word: shortint;               { current duplex state of terminal }
                                        { saved between switching off echo }
                                        { and switching it on again }
   echoing_on: boolean;                 { true when in full duplex }
   reading_action: boolean;             { true when reading action in WI }
   vdu_type: integer;                   { current vdu terminal type }
   vdu_table: ptr_vdu_info;             { -> vdu control information record }
   vdu_treename: tree_type;             { file name for vdu tables }

   macro_pointers: array[first_print_char..last_print_char]
                    of macro_pair;      { macro defs }
   current_macro_desc: ptr_macro_desc;  { -> macro currently being obeyed }
   macro_head,                          { -> top of macro being stored }
   store_macro_buffer: ptr_macro_buffer;  { -> current buffer for storing }
   store_macro_count: integer;          { total characters stored so far }
   macro_id_char: char;                 { identifying character of macro being }
                                        { stored }
   macro_id_type: macro_type;           { its type - posm or negm }
   mac_treename: tree_type;             { name of macro file }

   savestat_treename: tree_type;        { name of SAVESTAT file }

   ch_ins_del_avail,                    { true if vdu has char insert/delete }
   line_ins_del_avail,                  { true if vdu has line insert/delete }
   lineerase_avail,                     { true if vdu has line erase }
   scrolloff_avail,                     { ditto scrolloff (scroll suppress) }
   tabs_avail,                          { ditto hardware tabs of either type }
   reverself_avail,                     { ditto reverse linefeed }
   horiztab_avail,                      { ditto horizontal tab }
   backtab_avail,                       { ditto backwards tab }
   ctrlcode_avail,                      { ditto control code display }
   fold_avail,                          { ditto half intensity }
   pageerase_avail: boolean;            { ditto page erase }

   { information about all windows }
   window_count: window0_range;         { number of windows; 0 => no split }
   window_array: array[window0_range] of window_rec;
   windows_across, windows_down: window_range;  { details of split }
   cols_across, rows_down: integer;     { size+1 in rows or cols }

   action_proc_print:                   { mnemonic action names for MACPRINT }
      array[action_proc] of word;

value
   action_proc_print = (
      'empty   ','seq     ','brkline ','cr      ',
      'chdel   ','linedel ','file    ','quit    ',
      'save    ','v-many  ','delunl  ','dupl    ',
      'edcom   ','emacro  ','trunc   ','error   ',
      'exit    ','text    ','V       ','bottom  ',
      'home    ','<       ','<start  ','>       ',
      '>end    ','^       ','top     ','help    ',
      'insmode ','chins   ','lineins ','join    ',
      'V       ','bottom  ','load    ','lower   ',
      'mark    ','markoff ','negmacro','null    ',
      'open    ','close   ','point   ','pointoff',
      'macro   ','redrawl ','redraws ','repeat  ',
      '^       ','top     ','smacro  ','dmacro  ',
      'spaces  ','delsps  ','storech ','backtab ',
      'tab     ','settab  ','unsettab','unload  ',
      'upper   ','^many   ','wisw    ','xcomm   ',
      'zap     ','downscrn','upscrn  ','litchar ',
      'kinthost','ksuspend','krestart','kleftupd',
      'kfirstns','kalastns','keraprev','keraline',
      'kerarght','khomefld','knewline','nextwind',
      'prevwind');

{******************}
{ SSMP INFORMATION }
{******************}

const
   frame_start = 213b;                  { output frame start character }
   max_pars = 4;                        { max parameters for primitive }
   zmaxrow = 59;                        { allows a maximum of 60 rows }
   zmaxcol = 131;                       { allows a maximum of 132 columns }
   zfieldlimit = 239;                   { allows a maximum of 240 fields }

   { Subscripts of SSMP mode array elements }

   tlevel    = 0;     tmaxrow   = 1;     tmaxcol   = 2;
   dsinvalid = 3;     notify    = 4;     selectgr  = 5;
   reqshift  = 6;     cursor    = 7;     icharmode = 8;
   ilinerow  = 9;     intsignal = 10;    kinthost  = 11;
   ksuspend  = 12;    krestart  = 13;    kenter    = 14;
   kcsrup    = 15;    kcsrdown  = 16;    kcsrleft  = 17;
   kcsrright = 18;    knexttab  = 19;    kprevtab  = 20;
   kleftupd  = 21;    kfirstns  = 22;    kalastns  = 23;
   kinsmode  = 24;    keraright = 25;    kinsspac  = 26;
   kdelchar  = 27;    keraprev  = 28;    kinsline  = 29;
   kdelline  = 30;    keraline  = 31;    kappline  = 32;
   ksplline  = 33;    knextfld  = 34;    kprevfld  = 35;
   khomefld  = 36;    knewline  = 37;
   klast = 37;                          { last entry used }

   { Values for SELECTGR mode array element }

   defaultrend   = 0;                   { default rendition }
   bold          = 1;                   { bold, or increased intensity }
   faint         = 2;                   { faint, or reduced intensity }
   italicised    = 3;                   { italicised, or slanted }
   underlined    = 4;                   { underlined }
   slowblink     = 5;                   { slowly blinking (less than 150/min) }
   rapidblink    = 6;                   { rapidly blinking (150/min or more) }
   negativeimage = 7;                   { negative image (reverse video) }
   concealed     = 8;                   { concealed (displayed as SPACEs) }
   crossedout    = 9;                   { crossed out (marked for deletion) }

   frame_size = 60;                     { characters per SSMP frame output }
   frames_per_packet = 2;               { frames per packet }
   packet_max = 128;                    { network buffer size for one packet }

type
   packet_range = 1..packet_max;
   ssmp_row_range = 0..zmaxrow;         { in use, 0..maxrow }
   ssmp_col_range = 0..zmaxcol;         { in use, 0..maxcol }
   field_index = 0..zfieldlimit;        { in use, 0..((maxrow+1)*4-1) }
   mode_range = 0..63;                  { 37 is highest defined index }
   mode_array = array [mode_range]
                       of smallint;     { terminal emulation status }

   table = array [128..255] of smallint;
   pars_range = 1..max_pars;
   pars0_range = 0..max_pars;
   primpar = array [pars_range] of smallint;
   pend_type = (nopend, csrpend, fldpend, chinspend, chdelpend);

var
   term_is_ssmp: boolean;               { true if terminal is SSMP type }
   term_is_direct: boolean;             { true if Primos direct line }
   token: (withhost, withterm);         { access control token }
   saved_row: ssmp_row_range;           { cursor row }
   saved_col: ssmp_col_range;           { cursor column }
   curfield, maxfield: field_index;
   box_top, box_bottom: ssmp_row_range;  { update limits - rows }
   box_left, box_right: ssmp_col_range;  { update limits - columns }
   ssmp_mode: mode_array;               { terminal emulation status }

   pend_state: pend_type;               { setcursor/setfield optimisation }
   saved_ssmp_count: integer;           { number of ch ins/del stacked }
   filter: table;                       { filter input and strip parity }
   encode, decode: table;               { iso to/from ssmp conversions }
   input_buffer, output_buffer:         { input and output frame buffers }
      array[packet_range] of char;
   input_buffer_chars,                  { characters in input buffer }
   input_buffer_pointer,                { buffer pointers }
   output_buffer_pointer,
   frame_number,                        { index of output frame in packet }
   out_remainder: smallint;             { characters available in frame }

   usual_ssmp_mode,                     { initial mode array values }
   reply_mode: mode_array;              { modes used in answers }
   stacked_ssmp_codes: record           { SSMP input codes stacked }
      code1, code2: integer;
      multiple_count: smallint;
      mult_code1, mult_code2: smallint;
      row_diff, col_diff: integer;      { cursor position report }
      tree_report: boolean;
      tree_cr_given: boolean;
      stacked_tree: long_string
   end;
   token_here_after_request: boolean;   { true when special token value  }
                                        { received after H-REQTOKEN      }
   token_requested: boolean;            { true after H-REQTOKEN }
   special_limits_set: boolean;         { true when cursor is limited to }
                                        { a small box on the screen:     }
                                        { inhibits normal limit setting  }
   special_top, special_bottom: ssmp_row_range;  { special update limits }
   special_left, special_right: ssmp_col_range;

value

{  The following tables are used for encoding and decoding characters for
   SSMP transmission; all are subscripted by a character ord (128..255).

   ENCODE takes a character to be sent for display on the screen and delivers
      either 0, meaning send it as is, or the internal code of the character
      to be sent after ':' as the encoding.
}

   encode = (
       48, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79,
       80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 49, 50, 51, 52, 53,
        0, 97, 98, 99,100,101,102,103,  0,  0,  0,  0,  0,  0,  0,  0,
        0,  0,  0,  0,  0,  0,  0,  0,  0,  0,104,105,106,107,108,109,
      110,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
        0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,111,112,113,114,115,
      116,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
        0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,117,118,119,120,121
            );

{  FILTER takes incoming characters and converts them to either themselves
   or zero if they are to be ignored.
}

   filter = (
        0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
        0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
       32,  0,  0,  0,  0,  0,  0,  0, 40, 41, 42, 43, 44, 45, 46, 47,
       48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59,  0, 61,  0,  0,
        0, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79,
       80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90,  0,  0,  0,  0,  0,
        0, 97, 98, 99,100,101,102,103,104,105,106,107,108,109,110,111,
      112,113,114,115,116,117,118,119,120,121,122,  0,  0,  0,  0,  0
            );

{  DECODE takes the character after an incoming ':' and converts it to
   the appropriate target character.
}

   decode = (
       32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32,
       32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32,
       32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32,
        0, 27, 28, 29, 30, 31, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32,
       32,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14, 15,
       16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 32, 32, 32, 32, 32,
       32, 33, 34, 35, 36, 37, 38, 39, 58, 59, 60, 61, 62, 63, 64, 91,
       92, 93, 94, 95, 96,123,124,125,126,127, 32, 32, 32, 32, 32, 32
            );

   usual_ssmp_mode = (
      0,       { tlevel      }
      0,       { tmaxrow     }
      0,       { tmaxcol     }
      0,       { dsinvalid   }
      55,      { notify      }
      0,       { selectgr    }
      1,       { reqshift    }
      0,       { cursor      }
      0,       { icharmode   }
      0,       { ilinerow    }
      0,       { intsignal   }
      33,      { kinthost    }
      62,      { ksuspend    }
      63,      { krestart    }
      0,       { kenter      }
      0,       { kcsrup      }
      0,       { kcsrdown    }
      0,       { kcsrleft    }
      0,       { kcsrright   }
      0,       { knexttab    }
      0,       { kprevtab    }
      0,       { kleftupd    }
      0,       { kfirstns    }
      0,       { kalastns    }
      0,       { kinsmode    }
      0,       { keraright   }
      0,       { kinsspac    }
      0,       { kdelchar    }
      0,       { keraprev    }
      0,       { kinsline    }
      0,       { kdelline    }
      0,       { keraline    }
      0,       { kappline    }
      0,       { ksplline    }
      0,       { knextfld    }
      0,       { kprevfld    }
      0,       { khomefld    }
      0,       { knewline    }
      0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
      0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
      0, 0, 0, 0, 0, 0
      );

   reply_mode = (
      0,       { tlevel      }
      0,       { tmaxrow     }
      0,       { tmaxcol     }
      0,       { dsinvalid   }
      0,       { notify      }
      0,       { selectgr    }
      1,       { reqshift    }
      0,       { cursor      }
      0,       { icharmode   }
      0,       { ilinerow    }
      0,       { intsignal   }
      33,      { kinthost    }
      62,      { ksuspend    }
      63,      { krestart    }
      13,      { kenter      }
      0,       { kcsrup      }
      0,       { kcsrdown    }
      0,       { kcsrleft    }
      0,       { kcsrright   }
      0,       { knexttab    }
      0,       { kprevtab    }
      0,       { kleftupd    }
      0,       { kfirstns    }
      0,       { kalastns    }
      0,       { kinsmode    }
      0,       { keraright   }
      0,       { kinsspac    }
      0,       { kdelchar    }
      0,       { keraprev    }
      0,       { kinsline    }
      0,       { kdelline    }
      0,       { keraline    }
      0,       { kappline    }
      0,       { ksplline    }
      0,       { knextfld    }
      0,       { kprevfld    }
      0,       { khomefld    }
      0,       { knewline    }
      0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
      0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
      0, 0, 0, 0, 0, 0
      );


{**************************************************}
{ INFORMATION USED ONLY WHEN IN THE WINDOW EDITOR. }
{ BEWARE: THESE MAY BE UNSET IF ACCIDENTALLY       }
{ ACCESSED WHEN NOT IN WINDOW EDITOR.              }
{**************************************************}

const
   max_ins_del = 6;                     { we do not do more than this many }
                                        { line insert/deletes at once to }
                                        { avoid overtaxing the terminal }
   wi_oops_max = 10;                    { size of OOPS stack }

type
   insertonel_type = (do_blank, do_duplicate, do_oops);
   wi_oops_range = 1..wi_oops_max;
   wi_oops0_range = 0..wi_oops_max;

var
   current_window_number: window0_range;
   current_window: window_rec;
   screen_is_split: boolean;            { true when screen is split }
   full_screen: boolean;                { true when all screen is drawn }
                                        { false after e.g. WI -3 3 }
   window_mode: boolean;                { true if WI mode }
   entry_line_position: line_chars_range;  { line_position on entry }
   macro_ptr: ptr_macro_desc;
   action_code: action_proc;
   interrupt_deferred,                  { true when quit occurred part way }
                                        { through executing a command }
   need_to_redraw,                      { need to redraw current file line }
                                        { when we next change screen line }
   fill_in_undrawn_rows,                { set true when SSMP line delete done }
                                        { and we need to fill in at bottom of }
                                        { window later }
   need_to_output,                      { true when the editor needs to output }
                                        { the code received, e.g. when echo }
                                        { off or code is 'software' }
   character_omitted,                   { true when a dumb terminal fails to }
                                        { fill in last col of last row }
   always_output,                       { true when echoing off }
   como_switched_off,
   exit_window,
   getting_single_char: boolean;
   first_update_row,                    { range of rows set as update limits }
   last_update_row: integer;            { for SSMP terminal }
   input_char: char;                    { character read from keyboard }
   waiting_for_input: boolean;          { true when in c1in }

   macro_dropout: boolean;              { set while dropping out of macro(s) }

   msgst: shortint;                     { current message status }
   half_state: boolean;                 { true when vdu in half-intensity }

   { stack of line images before alteration for use by OOPS command }
   wi_oops_stack: array[wi_oops_range] of ptr_edline;
   wi_oops_top: wi_oops0_range;

   { Primos routines used all over the editor }

   procedure srsfx$(key: shortint;
                    pathname: long_string;
                    var unit, tipe: shortint;
                    n_suffixes: shortint;
                    suffix_list: string32;
                    var base_name: string32;
                    var suffix_used: shortint;
                    var code: shortint);
      extern;

   procedure prwf$$(key: shortint;      { read or write file block }
                    funit: shortint;
                    buffad: integer;
                    nw: shortint;
                    var position: integer;
                    var rnw, code: shortint);
      extern;

   procedure dir$se(dir_unit, dir_type: shortint;
                    start_at_top: shortint;
                    sel_ptr, return_ptr: integer;
                    max_entries, entry_size: shortint;
                    var ents_returned: shortint;
                    var type_counts: short4;
                    max_type: shortint;
                    var code: shortint);
      extern;

   procedure lon$cn(key: shortint);     { switch logout notification on/off }
      extern;

   procedure at$or(key: shortint; var code: shortint);
      extern;

   procedure fil$dl(filename: long_string; var code: shortint);
      extern;

   procedure cnsig$(var code: shortint); { set "continue to signal" flag }
      extern;

   procedure tty$rs(flags: word_as_bits; var code: shortint);
      extern;

   procedure cp$(command: long_string;
                 var code1, code2: shortint);
      extern;

   procedure com$ab(var command: long_string;
                    max_length: shortint;
                    var code: shortint);
      extern;

   procedure comlv$; extern;            { starts a new command level }

   procedure gpath$(key, funit: shortint;
                    var buffer: tree_chars;
                    bufflen: shortint;
                    var pathlen, code: shortint);
      extern;

   procedure p$break(on_or_off: boolean);
      extern;                           { control system handling of breaks }

   procedure p$hpsize(segments: integer);
      extern;                           { reserve segments of heap for EPF }

   procedure como$$(key, filnam, namlen, xx: shortint;
                    var code: shortint);
      extern;

   function tty$in: shortint;
      extern;                           { test whether user has typed ahead }

   procedure gv$get(name: string32;
                    var valu: long_string;
                    val_len: shortint;
                    var code: shortint);
      extern;

   procedure gv$set(name: string;
                    valu: string;
                    var code: shortint);
      extern;

   procedure t1ou(ch: char);            { print a character }
      extern;

   procedure tonl;                      { print a newline }
      extern;

   procedure c1in( var ch: char);       { read a character }
      extern;

   procedure at$rel(key: shortint;
                    dir_name: long_string;
                    var code: shortint);
      extern;

   procedure at$(key: shortint;
                 pathname: long_string;
                 var code: shortint);
      extern;

   procedure cnam$$(old_name: char32; old_len: shortint;
                    new_name: char32; new_len: shortint;
                    var code: shortint);
      extern;

