%separate = true
%ecs = true
%ttybuffered = false
%epfargs = true
{ Sheffield University Editor }
{ Copyright (c) University of Sheffield 1990 }

{***************************************}
{ MAIN PROGRAM AND LINE EDITOR COMMANDS }
{***************************************}

program main(input, output);

%listing off
%insert pmed_globals.ins.pas
%insert pmed_mprocs.ins.pas
%insert pmed_procs.ins.pas
%insert pmed_ssmp.ins.pas
%insert pmed_vdu.ins.pas
%insert pmed_efs.ins.pas
%listing on

   { move "wd_count" words from "from_string[fromsub]" to "to_string[tosub]".
     Preserves the words moved by copying either forwards or backwards }

   procedure movewds(from_string: line_chars_type;
                     fromsub: integer;
                     var to_string: line_chars_type;
                     tosub: integer;
                     wd_count: integer);
      extern;                           { written in PMA }

{ On-unit for QUIT$ }

procedure quit_onunit(not_used: integer);                                       { QUIT_ONUNIT }

var
   code: shortint;

begin { quit_onunit }
   if cominput_open then
   begin { editing from cominput file - do not }
      cnsig$(code);                     { attempt to handle quits }
      if code <> 0
      then error_print(k$irtn, code, 'In QUIT$ onunit', 'CNSIG$')
   end { comi on }
   else
   begin { no cominput file open - deal with quits here }
      tty$rs([bit1,bit2],code);         { clear terminal input and output }
      if code <> 0
      then error_print(k$irtn, code, 'Clear both', 'TTY$RS');
      screen_message('', false);
      if reading_command_line then
      begin { waiting for user input }
         if edit_mode
         then message := 'T'
         else message := 'Press RETURN and t';
         append_message('ype QUIT to leave the Editor.')
      end { waiting for user input }
      else message := 'Quit';
      screen_message(message, true);
      lowest_empty_buffer := 1;
      main_state := quit_signal         { quit at next convenient place }
   end { comi not on }
end { quit_onunit } ;

{ Onunit for condition LOGOUT$ raised when user is about to be force }
{ logged-out }

procedure logout_onunit(not_used: integer);                                     { LOGOUT_ONUNIT }

var
   code: shortint;

begin { logout_onunit }
   lon$cn(1);                           { switch logout notification on }
   screen_message('', false);
   screen_message('You are being logged out.', true);
   screen_message('Attempting to save the file(s) you are editing.', true);
   main_state := running;
   saveall_proc(true);
   if main_state = running then
   begin { saves worked }
      if paging_file_open
      then close_paging_file(false);    { do not leave a T$ file }
      cnsig$(code);                     { allow standard on-unit to log the }
                                        { user out }
      if code <> 0
      then error_print(k$irtn, code, 'In LOGOUT$ onunit', 'CNSIG$')
   end { saves worked }
   else                                 { let the user do something }
end { logout_onunit } ;

%page
{********************************}
{ READ A LINE OF EDITOR COMMANDS }
{********************************}

{ Read a line of editor commands and process unescaped kill and erase }

procedure read_command_line(var command_line: edline_type;                      { READ_COMMAND_LINE }
                            function read_a_char: char);

var
   length: line_chars0_range;
   escaped: boolean;
   prompt_char, ch, nextch: char;
   lindex: line_chars_range;
   count: 0..3;
   ch_code: 0..777b;

   procedure store(ch: char);
   begin { store }
      if ch = chr(dc1_code)
      then { disallow storing a DC1 in the file as it is used for space }
           { suppression by the Primos file system }
      else
      with command_line do
      begin
         if length < line_max_charsm1 then
         begin { there is space }
            length := length + 1;
            ed_text[length] := ch
         end { there is space }
      end { with }
   end { store } ;

begin { read_command_line }
   with command_line do
   begin
      test_init_eof(false);             { test for end of init file }
      length := 0;
      if input_state in [default_file, init_file] then
      begin { read from file with no erase or kill processing }
         while not eoln(input) do
         begin
            read(ch);
            store(ch)
         end { while loop } ;
         readln
      end { input from file }
      else if command_line_args.window_sw and  { set up a WINDOW or IW command }
              (not in_window_code)      { unless called from window_code }
      then
      begin { set up command }
         edit_mode := true;             { ensure it is obeyed, not stored }
         if edit_mode_after_init
         then begin store('W'); store('I') end
         else begin store('I'); store('W') end;
      end { enter window }
      else
      begin { read from terminal and process erase and kill }
         if mode_prompt and             { MODE PROMPT - output prompt }
            (not in_window_code)
         then
         begin { give prompt }
            if edit_mode
            then prompt_char := special[scprompt]
            else prompt_char := special[sdprompt];
            if prompt_char <> chr(nul_code)  { no prompt if NUL }
            then write(prompt_char,' ')
         end { give prompt } ;
         escaped := false;
         repeat
            if length < line_max_charsm1
            then ch := read_a_char      { get the next input char }
            else ch := newline;         { assume newline when buffer full }
            if main_state = quit_signal then
            begin { clear command read so far on quit }
               length := 0;
               escaped := false;
               main_state := running
            end { clear command read so far on quit } ;
            if ch <> newline then
            begin { not newline }
               if escaped then
               begin { remove escape for kill and erase }
                  if (ch <> special[skill]) and (ch <> special[serase])
                  then store(special[sescape]);
                  store(ch);
                  escaped := false
               end { remove escape for kill and erase }
               else if ch = special[skill]  { kill resets input line }
               then length := 0
               else if ch = special[serase]  { erase cancels 1 character }
               then if length > 0
                    then length := length - 1
                    else
               else if ch = special[sescape]  { escape means 'treat as itself' }
               then escaped := true
               else store(ch)
            end { not newline }
         until (ch = newline);
         if escaped then
         begin { escape before newline }
            store(special[sescape]);
            escaped := false
         end { escape before newline }
      end { read from terminal and process erase and kill } ;
      ed_text[length + 1] := newline;   { add newline to end }

      { process escape characters in the text - currently only ^nnn }

      lindex := 1;
      length := 0;
      escaped := false;
      repeat
         ch := ed_text[lindex];
         if (ch = special[sescape]) and (not escaped) then
         begin { user escape char found }
            nextch := ed_text[lindex+1];
            if nextch in ['0'..'7'] then
            begin { octal digit }
               lindex := lindex + 1;
               ch_code := 0;            { octal code being built up }
               count := 3;
               while (count <> 0) and (nextch in ['0'..'7']) do
               begin
                  count := count - 1;
                  ch_code := ch_code*8 + ord(nextch) - ord('0');
                  lindex := lindex + 1;
                  nextch := ed_text[lindex]
               end { while loop } ;
               ch_code := ch_code mod 256;
               store(chr(ch_code));     { store char with that code }
               lindex := lindex - 1     { back to last octal digit }
            end { octal digit }
            else
            begin { other escaped char - leave alone }
               store(ch);
               escaped := true
            end { other escaped char - leave alone }
         end { user escape char found }
         else
         begin { not escape char }
            store(ch);                { unescaped char }
            escaped := false
         end { not escape char } ;
         if ch <> newline
         then lindex := lindex + 1
      until ch = newline;
      text_length := length - 1         { processed length ( less newline ) }
   end { with }
end { read_command_line } ;

   procedure obey_command_line(command_line: line_chars_type;
                               times: integer;
                               upto: line_chars_plus_range);
      forward;

   procedure window_code(number_of_args, arg1, arg2: integer;
                         init_wi: boolean);
      extern;

%page
{*********************************}
{ OBEY A LINE OF EDITOR COMMANDS. }
{*********************************}

{ This procedure is called
      1) after reading a command line
      2) the first time *nnn is detected in a line of commands
      3) in obeying the XEQ command.
  'times' is the number of times the line is to obeyed.
  'upto'  is the position on the line at which obedience is to cease.
  'upto'  is set to the terminator after the number (if any) for
          the repetition call, or newline position plus one for
          other calls. }

procedure obey_command_line;                                                    { OBEY_COMMAND_LINE }
                         {command_line: line_chars_type;
                          times: integer;
                          upto: line_chars_plus_range}

var
   sub: line_chars_plus_range;          { position on line }
   first_command_on_line: boolean;
   ch: char;
   rep_count: integer;
   command_name: word;
   command: commands;
   command_state: search_type;
   switch, result: boolean;

%page
{******************************************************}
{ UTILITY PROCEDURES USED BY COMMAND ANALYSIS ROUTINES }
{          ( INTERNAL TO OBEY_COMMAND_LINE )           }
{******************************************************}

{ Skip any spaces at current command line point }

procedure skip_spaces;                                                          { SKIP_SPACES }
begin { skip_spaces }
   while command_line[sub] = ' '
   do sub := sub + 1
end { skip_spaces } ;

{ Check that there is no extra information after the command }

function check_termination: boolean;                                            { CHECK_TERMINATION }

var
   correct_term: boolean;

begin { check_termination }
   skip_spaces;                         { do not mind a few spaces }
   correct_term := command_line[sub] in string_terminators;
   if not correct_term
   then comm_error('Extra information after command.');
   check_termination := correct_term
end { check_termination } ;

{ Get the next word from the command line - terminate on non-alphabetic char }

procedure get_word(var result: word);                                           { GET_WORD }

var
   finished, overflowed: boolean;
   ch: char;
   index: word_range;

begin { get_word }
   skip_spaces;
   result := blank_word;
   index := 1;
   finished := false;
   overflowed := false;
   repeat
      ch := command_line[sub];
      if ch in ['A'..'Z', 'a'..'z', '_'] then
      begin { alphabetic etc. }
         if ch in ['a'..'z']            { convert to upper case }
         then ch := chr(ord('A') - ord('a') + ord(ch));
         if not overflowed
         then result[index] := ch;
         sub := sub + 1;
         if index < word_max_chars
         then index := index + 1
         else overflowed := true
      end { alphabetic etc. }
      else finished := true
   until finished
end { get_word } ;

{ Read a decimal number from the command line; delivers true if number present }
{ returns 0 if no number present. }

function get_number(var result: integer): boolean;                              { GET_NUMBER }

var
   finished, negative: boolean;
   ch: char;
   old_sub: line_chars_plus_range;

begin { get_number }
   skip_spaces;
   result := 0;
   finished := false;
   old_sub := sub;                      { sub value at start of number }
   negative := command_line[sub] = '-';
   if negative then sub := sub + 1;
   repeat
      ch := command_line[sub];
      if ch in ['0'..'9'] then
      begin { numeric }
         if result < large_int          { avoid overflow }
         then result := result*10 + ord(ch) - ord('0');
         sub := sub + 1
      end { numeric }
      else finished := true
   until finished;
   if negative then result := - result;
   get_number := sub <> old_sub         { was a number actually present? }
end { get_number } ;

{ Get .charnumber for POINT and NEXT commands, if present }

procedure get_line_pos;                                                         { GET_LINE_POS }

var
   lpos: integer;

begin { get_line_pos }
   skip_spaces;
   with current_buffer do
   if command_line[sub] = '.' then
   begin { dot found }
      sub := sub + 1;
      if get_number(lpos)
      then if line_status = normal
           then if (lpos > 0) and (lpos < line_max_chars)
                then line_position := lpos
                else comm_error('Error in character number.')
           else                         { do not set for .top. or .bottom. }
      else comm_error('Error in character number.')
   end { . found }
end { get_line_pos } ;

{ Get the next (positive) number from the command line, allowing for ranges }
{ given as 3-5 etc.  Uses existing value of the_number when getting range.  }

function get_next_number(var the_number: integer): boolean;                     { GET_NEXT_NUMBER }

var
   old_sub: line_chars_plus_range;
   top_of_range: integer;

begin { get_next_number }
   get_next_number := false;
   skip_spaces;
   old_sub := sub;
   if (the_number > 0) and (command_line[sub] = '-') then
   begin { range found }
      sub := sub + 1;
      if get_number(top_of_range) then
      begin { top found }
         if the_number < top_of_range then
         begin { deliver next in range }
            the_number := the_number + 1;
            sub := old_sub;             { set back so that '-' found again }
            get_next_number := true
         end { deliver next in range }
         else if the_number = top_of_range then
         begin { range exhausted }
            the_number := 0;
            get_next_number := get_next_number(the_number)
         end { range exhausted }
         else comm_error('Error in number range.')
      end { top found }
      else comm_error('Error in number range.')
   end { range found }
   else
   begin { not range }
      get_next_number := get_number(the_number)
   end { not range }
end { get_next_number } ;

{ Read a string buffer name from the command line and return its index. }
{  Index values are:                                                    }
{     -2   ALL        (if allowed by command)                           }
{     -1   INLIN                                                        }
{     0    EDLIN                                                        }
{     1    STRA or STR.1                                                }
{     ..   ..                                                           }
{     10   STRJ or STR.10                                               }

procedure get_buffer_index(var bindex: string_buffers2_range;                   { GET_BUFFER_INDEX }
                           allow_all: boolean);

var
   number: integer;
   buffer_name: word;
   valid, result: boolean;
   ch: char;

begin { get_buffer_index }
   get_word(buffer_name);
   valid := buffer_name[1] <> ' ';
   if compare(buffer_name, 'INLIN   ')  { INLIN is current line }
   then number := -1
   else if allow_all and compare(buffer_name, 'ALL     ')
   then number := -2
   else if compare(buffer_name, 'EDLIN   ')  { EDLIN is last edit line }
   then number := 0
   else  if (buffer_name[1] = 'S') and  { STRa.. or STR1.. are string buffers }
            (buffer_name[2] = 'T') and
            (buffer_name[3] = 'R')
   then
   begin { STR found }
      if buffer_name[4] = ' ' then
      begin { '.' not picked up by "get_word" }
         if command_line[sub] = '.' then
         begin { get buffer number }
            sub := sub + 1;
            ch := command_line[sub];
            if (ch >= '0') and (ch <= '9')
            then  result := get_number(number)
            else  valid := false
         end { get buffer number }
         else  valid := false
      end { '.' not picked up by "get_word" }
      else if buffer_name[5] = ' '
      then number := ord(buffer_name[4]) - ord('A') + 1
      else valid := false;              { something after letter }
      if valid
      then valid := (number >= 0) and (number <= num_string_buffers)
   end { STR found }
   else  valid := false;                { not STR }
   if valid
   then bindex := number
   else comm_error('Error in format of string buffer name.')
end { get_buffer_index } ;

{ Copy a tree name or Primos command from command_line }

procedure copy_treename(var the_tree: tree_type;                                { COPY_TREENAME }
                        upcaseit, allow_spaces, copy_quotes,
                        do_abbrevs: boolean);

var
   finished, in_quotes, escaped, add_the_char: boolean;
   ch: char;
   tindex: tree_range;

begin { copy_treename }
   with the_tree do
   begin
      skip_spaces;
      len := 0;
      finished := false;
      in_quotes := false;
      escaped := false;
      repeat
         ch := command_line[sub];
         if not escaped and
            ((ch in string_terminators) or
            ( (ch = ' ') and (not allow_spaces) and (not in_quotes) ) )
         then  finished := true
         else
         begin { not terminator }
            sub := sub + 1;
            if (not escaped) and (ch = special[sescape])
            then escaped := true
            else
            begin { not escape char }
               escaped := false;
               if ch = '''' then
               begin { quote character }
                  in_quotes := not in_quotes;
                  add_the_char := copy_quotes
               end { quote character }
               else add_the_char := true;
               if add_the_char then
               begin { ok to add in the char }
                  if len < tree_max_chars then
                  begin { space available }
                     len := len + 1;
                     if upcaseit and    { upcase if requested }
                        (ch in ['a'..'z'])
                     then ch := chr(ord('A') - ord('a') + ord(ch));
                     chars[len] := ch
                  end { space available }
                  else finished := true
               end { ok to add in the char }
            end { not escape char }
         end { not terminator }
      until finished;
      for tindex := len + 1 to tree_max_chars  { pad with blanks }
      do chars[tindex] := ' ';
      if do_abbrevs
      then expand_abbrevs(the_tree)
   end { with }
end { copy_treename } ;

{ Called by commands that are not implemented }

procedure not_implemented;                                                      { NOT_IMPLEMENTED }

begin { not_implemented }
   word_screen_message('Command ~ is not implemented.', true,
                       command_names[command], true);
   main_state := command_error
end { not_implemented } ;

{ Print a command error message }

procedure bad_command;                                                          { BAD_COMMAND }

begin { bad_command }
   { if the command was a normal one, rather than one of the "many_commands", }
   { then replace the name the user typed by the full name from the list }
   if command_state = found
   then command_name := command_names[command];
   word_screen_message('Error in format of ~ command.', true, command_name, true);
   main_state := command_error
end { bad_command } ;

{ Command text begins with non-alpha char, giving blank command word }

procedure invalid_comm;                                                         { INVALID_COMM }

begin { invalid_comm }
   comm_error('Unrecognised command. Type HELP if you want assistance.')
end { invalid_comm } ;

{ Process the counter symbol (usually @) in input text }

procedure store_count_value(procedure store(ch: char) );                        { STORE_COUNT_VALUE }

var
   num_string: packed array[1..count_width_max] of char;
   start_at, cindex: 1..count_width_max;
   digit: 0..9;
   val: integer;
   negative, sign_done: boolean;
   ch: char;

begin { store_count_value }
   val := count_value;
   negative := val < 0;
   start_at := count_width_max - count_width + 1;
   if negative then
   begin { number < 0 }
      val := -val;
      start_at := start_at + 1;         { allow one place for sign }
      if count_type = zprint then       { for zero print insert sign now }
      begin { print zeroes }
         store('-');
         sign_done := true
      end { print zeroes }
      else sign_done := false           { sign not inserted yet }
   end { negative }
   else sign_done := true;              { no sign for positive value }

   { produce decimal count with leading spaces }

   for cindex := count_width_max downto start_at do
   begin
      if val > 0 then
      begin { still positive }
         digit := val mod 10;
         val := val div 10;
         num_string[cindex] := chr(ord('0') + digit)
      end { still positive }
      else num_string[cindex] := ' '
   end { for loop } ;
   if count_value = 0                   { zero must have > 0 digits }
   then num_string[count_width_max] := '0';

   { output count with correct treatment of leading zeroes }

   for cindex := start_at to count_width_max do
   begin
      ch := num_string[cindex];
      if ch = ' ' then
         case count_type of
            zblank: store(' ');
            zprint: store('0');
            zsuppress:
         end
      else
      begin { non-blank }
         if not sign_done then
         begin { output the sign now }
            store('-');
            sign_done := true
         end { output the sign now } ;
         store(ch)
      end { non-blank }
   end { for loop } ;
   count_value := count_value + count_increment
end { store_count_value } ;

{ Copy a delimited string from the command line }

function copy_string(procedure store(ch: char);                                 { COPY_STRING }
                     process_escape: boolean): boolean;

var
   terminator, ch: char;
   escaped: boolean;
   state: search_type;

begin { copy_string }
   terminator := command_line[sub];
   if terminator = newline
   then copy_string := false            { no opening delimiter }
   else
   begin
      if terminator in [' ','a'..'z','A'..'Z'] then
      begin { Is terminator correct? }
         ch_screen_message(
            '***WARNING*** Character ''~'' being used as string delimiter.',
            true, terminator)
      end { Is terminator correct? } ;
      escaped := false;
      state := searching;
      repeat
         sub := sub + 1;
         ch := command_line[sub];
         if ch = newline
         then  state := not_found       { terminator not found }
         else  if escaped then
         begin
            store(ch);
            escaped := false
         end
         else  if ch = terminator
         then  state := found
         else  if (ch = special[sescape]) and process_escape
         then  escaped := true
         else  store(ch)
      until state <> searching;
      copy_string := state = found
   end { string found }
end { copy_string } ;

{ Copy text from the current command_line into the current file line. }
{ Performs special processing on counter symbols etc. }

procedure copy_input_string(overlay: boolean;                                   { COPY_INPUT_STRING }
                            start_col: line_chars_range);

var
   copy_col, tabno: integer;
   specials: set_of_char;
   finished, escaped: boolean;
   ch: char;
   lindex: line_chars_range;
   state: search_type;

   procedure store(ch: char);
   begin { store }
      if copy_col < line_max_chars
      then current_buffer.line_text[copy_col] := ch;
      copy_col := copy_col + 1
   end { store } ;

begin { copy_input_string }
   with current_buffer do
   begin
      if overlay
      then specials := overlay_specials
      else specials :=   input_specials;
      copy_col := start_col;
      escaped := false;
      finished := false;
      repeat
         ch := command_line[sub];
         if ch = newline
         then  finished := true         { newline always terminates }
         else  if escaped then
         begin
            store(ch);                  { escaped char stored as is }
            escaped := false
         end
         else if (ch < chr(200b)) or    { parity off or }
                 (not (ch in specials))  { not a special }
         then store(ch)                 { ordinary character }
         else if ch = special[stab] then
         begin { process tabs }
            tabno := 1;
            state := searching;
            repeat
               if tabno > number_of_tabs
               then state := not_found
               else if htabs[tabno] > copy_col
               then state := found
               else tabno := tabno + 1
            until state <> searching;
            if state = not_found
            then store(' ')             { past last tab position }
            else
            begin { tab position found }
               for lindex := copy_col to htabs[tabno] - 1
               do if (not overlay) or (copy_col > line_length)
                  then store(' ')
                  else copy_col := copy_col + 1  { overlay situation }
            end { tab position found }
         end { tab character }
         else if ch = special[scounter]
         then store_count_value(store)  { deal with @ }
         else if ch = special[ssemicolon]
         then finished := true          { ';' terminates line (MODE SEMI) }
         else if ch = special[sescape]
         then escaped := true           { next char is escaped }
         else if ch = ' '
         then  if copy_col > line_length
               then  store(' ')         { beyond end of line }
               else copy_col := copy_col + 1  { overlay - just skip }
         else if ch = special[swild]
         then store(' ')                { overlay - wild becomes space }
         else abort('unrecognised special char');
         if not finished
         then sub := sub + 1
      until finished;

      if (not overlay) or
         (copy_col - 1 > line_length) then
      begin { new end of line position }
         if copy_col > line_max_chars then
         begin
            line_length := line_max_charsm1;
            comm_error('Line overflowed - extra character(s) lost.')
         end
         else line_length := copy_col - 1;
         line_text[line_length + 1] := newline
      end { new end of line position } ;
      line_altered := true
   end
end { copy_input_string } ;

{ Perform operations on lines of the file a given number of times or until top }
{ or bottom reached. Used by PRINT, DELETE, UNLOAD and DUNLOAD. }

procedure loop_by_number(what_to_do: set_of_actions;                            { LOOP_BY_NUMBER }
                         number_of_lines: integer);

var
   print, deletec, unload: boolean;
   direction: direction_type;
   start_line, stop_line, nlines: integer;

begin { loop_by_number }
   with current_buffer do
   begin
      print      := print_act  in what_to_do;
      deletec    := delete_act in what_to_do;
      unload     := unload_act in what_to_do;
      if backwards_act in what_to_do
      then direction := go_backwards
      else direction := go_forwards;
      if deletec
      then must_redraw_window := true;
      if print or ((direction = go_backwards) and unload) then
      begin { have to do it the slow way }
         if print then print_banner;
         repeat
            if print then print_line;   { print the line }
            if unload
            then if not unload_line     { copy the line out to file }
                 then main_state := command_error;
            if deletec then delete_line;  { delete the line }
            number_of_lines := number_of_lines - 1;
            if number_of_lines > 0 then
            begin { move to next line }
               case direction of
                  go_forwards:
                     if line_exists
                     then load_a_line(line_number + 1, true)
                     else load_a_line(line_number    , true);
                  go_backwards:
                     load_a_line(line_number - 1, true)
               end { case } ;
               test_line_landed_on      { test for top or bottom }
            end { move to next line }
         until (number_of_lines <= 0) or (main_state <> running)
      end { have to do it the slow way }
      else
      begin { use quick procedures }
         start_line := line_number;
         if number_of_lines = 0
         then nlines := 1               { D0 => D1 etc. }
         else nlines := number_of_lines;
         if direction = go_backwards then
         begin { backwards }
            stop_line := line_number - nlines + 1;
            if stop_line < 0
            then stop_line := 0;        { stop at top of file }
            if not line_exists
            then start_line := start_line -1;  { allow for line already deleted }
            if deletec and (stop_line <= start_line) then
            begin
               quick_delete(stop_line + 1, start_line);
               load_a_line(stop_line, true);
               delete_line              { allows later OOPS if necessary }
            end
         end { backwards }
         else
         begin { forwards }
            stop_line := line_number + nlines - 1;
            if not line_exists
            then stop_line := stop_line - 1;  { count one for line deleted }
            if stop_line >= start_line then
            begin
               if unload
               then  if not quick_unload(start_line, stop_line)
                     then main_state := command_error;
               if deletec and (main_state = running) then
               begin
                  quick_delete(start_line, stop_line - 1);
                  delete_line
               end;
               if line_status = bottom
               then screen_message('.bottom.', true);
               test_line_landed_on
            end { stop_line >= start_line }
         end { forwards }
      end { use quick procedures }
   end { with }
end { loop_by_number } ;

{ Structure the search string }

function structure_string: boolean;                                             { STRUCTURE_STRING }

var
   current_element: elements;
   escaped, finished, overflowed: boolean;
   ch: char;
   {---teststart---}
   lindex, sindex: integer;
   {---teststop---}
   locate_string_length: 0..line_max_chars;

   { Store part of a search string element in the structured string array.
     If it is different from the current, then a new entry is created. }

   procedure store(el: elements);

   begin { store }
      if el <> current_element then     { create a new entry }
      begin
         if locate_list_length < locate_list_length_max then
         begin
            locate_list_length := locate_list_length + 1;
            with locate_element_list[locate_list_length] do
            begin { initialise the new entry }
               element := el;
               arg1 := locate_string_length + 1;
               arg2 := 0
            end { initialise the new entry }
         end
         else overflowed := true;       { entry list is full }
         current_element := el
      end;

      with locate_element_list[locate_list_length] do
      case el of
         el_string:
            begin { there must be sufficient space }
               locate_string_length := locate_string_length + 1;
               locate_string[locate_string_length] := ch;
               if ch in lower_alphabet
               then ch := chr(ord('A') - ord('a') + ord(ch));  { upcase }
               locate_upper[locate_string_length] := ch;
               if ch in upper_alphabet
               then ch := chr(ord('a') - ord('A') + ord(ch));  { lowcase }
               locate_lower[locate_string_length] := ch;
               arg2 := locate_string_length
            end { there must be sufficient space } ;
         blanks: ;                      { no further information necessary }
         wild:
            arg2 := arg2 + 1;           { count of wild symbols }
         nothing:
            abort('structure_string: trying to store "nothing"')
      end { case }
   end { store } ;

begin { structure_string }
   locate_list_length := 0;
   locate_string_length := 0;
   current_element := nothing;
   escaped := false;
   finished := false;
   overflowed := false;
   repeat
      ch := command_line[sub];
      if ch in string_specials then
      begin { in string_specials }
         if ch = newline                { always terminates }
         then finished := true
         else if escaped then
         begin { escaped - store it as is }
            store(el_string);
            escaped := false
         end { escaped - store it as is }
         else if ch in string_terminators
         then finished := true
         else if ch = special[sescape]
         then escaped := true
         else if ch = special[sblanks]  { test for # }
         then store(blanks)
         else if ch = special[swild]    {      and ! }
         then store(wild)
         else abort('structure_string: special character not recognised')
      end { ch in string_specials }
      else
      begin { normal character }
         store(el_string);
         escaped := false               { in case user escaped it }
      end { normal character } ;
      if not finished
      then sub := sub + 1               { on to next character }
   until finished;
   locate_zero := locate_list_length = 0;
   structure_string := not overflowed;
   locate_string_exists := not overflowed;
   {---teststart---}
   if mode_test then
   begin { diagnostics }
      writeln('structure_string: locate_zero = ',locate_zero);
      for lindex := 1 to locate_list_length do
      with locate_element_list[lindex] do
      case element of
         el_string:
            begin
               write('    string: "');
               for sindex := arg1 to arg2 do
                  write(locate_string[sindex]);
               writeln('"');
               write('upp_string: "');
               for sindex := arg1 to arg2 do
                  write(locate_upper[sindex]);
               writeln('"');
               write('low_string: "');
               for sindex := arg1 to arg2 do
                  write(locate_lower[sindex]);
               writeln('"')
            end;
         blanks: writeln('   blanks');
         wild:   writeln('   wild: ', arg2:1);
         nothing: writeln('***nothing***')
      end { case }
   end { diagnostics }
   {---teststop---}
end { structure_string } ;

{ Select a string of characters to be used for "quick_locate".             }
{ Characters selected may include single blanks, but no blanks are allowed }
{ on beginning or end.                                                     }

function select_ql_string: boolean;                                             { SELECT_QL_STRING }

var
   index: 1..locate_list_length_max;
   state: search_type;
   spaces_run: integer;
   done: boolean;
   {---teststart---}
   qlindex: integer;
   {---teststop---}

begin { select_ql_string }
   index := 1;
   state := searching;
   repeat
      with locate_element_list[index] do
      begin
         if element = el_string then    { we are only looking for strings }
         begin
            locate_ql_start := arg1;
            while (locate_string[locate_ql_start] = ' ') and  { skip spaces }
                  (locate_ql_start <= arg2) do
                  locate_ql_start := locate_ql_start + 1;
            if locate_ql_start <= arg2 then  { anything left? }
            begin
               locate_ql_end := locate_ql_start;
               spaces_run := 0;         { count run of spaces }
               done := false;
               repeat
                  if locate_string[locate_ql_end] = ' ' then
                  begin
                     spaces_run := spaces_run + 1;
                     if spaces_run > 1  { stop if more than one }
                     then done := true
                  end
                  else spaces_run := 0;
                  if not done
                  then  if locate_ql_end < arg2
                        then locate_ql_end := locate_ql_end + 1
                        else done := true  { end of string }
               until done;
               { remove trailing spaces }
               locate_ql_end := locate_ql_end - spaces_run;
               state := found
            end
         end;
         if state = searching           { try next element }
         then  if index < locate_list_length
               then index := index + 1
               else state := not_found  { list exhausted }
      end { with }
   until state <> searching;
   select_ql_string := state = found;
   {---teststart---}
   if mode_test
   then  if state = found then
         begin
            write('quick_locate string is "');
            for qlindex := locate_ql_start to locate_ql_end
            do write(locate_string[qlindex]);
            writeln('"')
         end
         else writeln('No quick_locate string found.')
   {---teststop---}
end { select_ql_string } ;

{ Test the string specified by "locate_element_list" against the current line }
{ at the current column position. }

function string_matches(column: integer;                                        { STRING_MATCHES }
                        case_signif, ident_match: boolean): boolean;

var
   list_index: 1..locate_list_length_max;
   match_col, sindex: integer;
   state, inner_state: search_type;
   ch: char;
   left_ok, right_ok: boolean;

begin { string_matches }
   with current_buffer do
   begin
      match_col := column;              { current position in line }
      list_index := 1;
      state := searching;
      repeat
         with locate_element_list[list_index] do
         case element of
            el_string:
               begin { straightforward string }
                  sindex := arg1;
                  inner_state := searching;
                  if case_signif then
                  begin { test with case significant }
                     repeat
                        if match_col > line_max_charsm1
                        then inner_state := not_found  { max line length }
                        else
                        begin
                           if match_col > line_length  { consider blank padded }
                           then ch := ' '
                           else ch := line_text[match_col];
                           if ch = locate_string[sindex]
                           then  if sindex >= arg2
                                 then inner_state := found
                                 else
                                 begin
                                    sindex := sindex + 1;
                                    match_col := match_col + 1
                                 end
                           else  inner_state := not_found
                        end
                     until inner_state <> searching
                  end { test with case significant }
                  else
                  begin { test with case not significant }
                     repeat
                        if match_col > line_max_charsm1
                        then inner_state := not_found  { max line length }
                        else
                        begin
                           if match_col > line_length  { consider blank padded }
                           then ch := ' '
                           else ch := line_text[match_col];
                           if (ch = locate_upper[sindex]) or
                              (ch = locate_lower[sindex]) then
                           begin
                              if sindex >= arg2
                              then inner_state := found
                              else
                              begin
                                 sindex := sindex + 1;
                                 match_col := match_col + 1
                              end
                           end
                           else inner_state := not_found
                        end
                     until inner_state <> searching
                  end { test with case not significant } ;
                  match_col := match_col + 1;  { skip last character of string }
                  if inner_state = not_found
                  then state := not_found
               end { straightforward string } ;
            blanks:                     { any number of blanks }
               while (match_col <= line_length) and (line_text[match_col] = ' ')
               do match_col := match_col + 1;
            wild:                       { arg2 wild symbols match any char }
               if match_col + arg2 > line_length + 1
               then state := not_found  { line too short }
               else match_col := match_col + arg2  { skip the chars }
         end { case } ;
         if state = searching           { on to next element }
         then  if list_index = locate_list_length
               then state := found      { list exhausted }
               else list_index := list_index + 1
      until state <> searching;
      if ident_match and (state = found) then
      begin { identifier type match - check end }
                                        { points found }
         if column = 1
         then left_ok := true
         else left_ok := line_text[column-1] in punctuation;
         if match_col > line_length
         then right_ok := true
         else right_ok := line_text[match_col] in punctuation;
         if left_ok and right_ok
         then                           { both ends delimited by punctuation }
         else state := not_found        { part of longer word }
      end { identifier type match - check end } ;
      if state = found then
      begin { all matches }
         string_matches := true;
         last_match_position := match_col - 1
      end { all matches }
      else string_matches := false
   end { with current_buffer }
end { string_matches } ;

{ Perform command actions up to a terminating line specified by string. }

procedure loop_to_string(what_to_do: set_of_actions;                            { LOOP_TO_STRING }
                         start_column: line_chars_range);

var
   print, deletec, unload, backwards, locate_type, notflag,
   case_signif, ident_match, to_mark, scurrent: boolean;
   state: search_type;
   terminate, check_rest_of_line, col_matches: boolean;
   column: integer;
   chain_allowed, unload_later, delete_later: boolean;
   start_line, stop_line: integer;

begin { loop_to_string }
   with current_buffer do
   begin
      print       := print_act      in what_to_do;
      deletec     := delete_act     in what_to_do;
      unload      := unload_act     in what_to_do;
      backwards   := backwards_act  in what_to_do;
      case_signif := not (case_act  in what_to_do);
      ident_match := ident_act      in what_to_do;
      locate_type := locate_act     in what_to_do;
      notflag     := not_act        in what_to_do;
      to_mark     := mark_act       in what_to_do;
      scurrent    := start_act      in what_to_do;
      if deletec
      then must_redraw_window := true;
      chain_allowed := mode_chain and not (deletec or unload or to_mark);

      { decide whether we can use the quick procedures }

      unload_later := false;
      delete_later := false;
      if print or (backwards and unload)
      then { no }
      else
      begin { can use quick procedures }
         start_line := line_number;
         if backwards and (not line_exists)
         then start_line := start_line -1;  { allow for line deleted }
         if unload then
         begin
            unload := false;            { do not do it now }
            unload_later := true        { do it later }
         end;
         if deletec then
         begin
            deletec := false;
            delete_later := true
         end
      end { can use quick procedures } ;

      { examine the search string and structure it }

      if command_line[sub] in string_terminators
      then if locate_string_exists      { no string: use last one }
           then                         { ok - it exists }
           else bad_command
      else if command_line[sub] <> ' '
      then bad_command
      else
      begin { valid string found }
         sub := sub + 1;
         if structure_string
         then { string structured alright }
         else comm_error('Search string is too complicated.')
      end { valid string found } ;

      { can we use "quick_locate"? }

      if main_state = running then
      begin { still alright }
         if locate_zero or print or deletec or unload or notflag
         then locate_use_ql := false
         else locate_use_ql := select_ql_string
      end { still alright } ;

      { Decide where to stop }

      if to_mark then
      begin { stop at mark }
         if marked_line = 0
         then comm_error('No line is marked.')
         else
         begin { stop one line after marked line }
            if backwards
            then stop_line := marked_line - 1
            else stop_line := marked_line + 1
         end { stop one line after marked line }
      end { stop at mark }
      else stop_line := 0;

      { main loop }

      if print then print_banner;
      { if called from within the window editor LOCATEs etc. start at the }
      { character after the current line position rather than on the next }
      { line }
      if in_window_code
      then if notflag or                { NL etc. apply to whole line }
              (line_status <> normal)   { always move off null line }
           then check_rest_of_line := false
           else if locate_type          { always for LOCATE etc. }
           then check_rest_of_line := true
           else if backwards            { only if within range for FIND }
                then check_rest_of_line := start_column < line_position
                else check_rest_of_line := start_column > line_position
      else check_rest_of_line := false;

      terminate := false;
      while (not terminate) and (main_state = running) do
      begin
         if print then print_line;
         if unload
         then if not unload_line
              then main_state := command_error;
         if deletec then delete_line;

         { get to next candidate line for testing }

         if not check_rest_of_line then
         begin { do not check rest of line }
            if locate_use_ql
            then if case_signif
                 then quick_locate(locate_string,locate_string,locate_ql_start,
                                   locate_ql_end, backwards, case_signif, false,
                                   stop_line)
                 else quick_locate(locate_upper, locate_lower, locate_ql_start,
                                   locate_ql_end, backwards, case_signif, false,
                                   stop_line)
            else if backwards
            then load_a_line(line_number - 1, false)
            else if line_exists
            then load_a_line(line_number + 1, false)
            else load_a_line(line_number    , false)
         end { do not check rest of line } ;

         { test the line landed on }

         if to_mark and (line_number = stop_line)
         then main_state := top_or_bottom
         else test_line_landed_on;      { hit top or bottom? }
         if main_state = running then
         begin { stopped on normal line }
            if locate_zero
            then state := found         { it must match! }
            else
            begin { try string against line }
               { LOCATEs from within the window editor start after the current }
               { column rather than on the next line }
               if locate_type
               then if check_rest_of_line
                    then if scurrent
                         then column := line_position
                         else if backwards
                              then column := line_position - 1
                              else column := line_position + 1
                    else if backwards
                         then column := line_length
                         else column := start_column
               else column := start_column;
               if column < 1
               then state := not_found
               else
               begin { column >= 1 }
                  state := searching;
                  repeat
                     { test this start column only if within valid range }
                     if (column = start_column) or
                       ((column > start_column) and (column <= line_length))
                     then col_matches :=
                        string_matches(column, case_signif, ident_match)
                     else col_matches := false;
                     if col_matches then
                     begin { match at this column }
                        state := found;
                        line_position := column  { column number for match }
                     end { match at this column }
                     else if locate_type  { go on if "locate" type }
                          then if backwards
                               then if column <= start_column
                                    then state := not_found
                                    else column := column - 1
                               else if column >= line_length
                                    then state := not_found
                                    else column := column + 1
                          else state := not_found
                  until state <> searching
               end { column >= 1 }
            end { try string against line } ;
            if notflag
            then terminate := state = not_found
            else terminate := state = found;
            terminate := terminate or (line_number = stop_line);
            {---teststart---}
            if mode_test and locate_use_ql and (not terminate)
            then writeln('Failed ql candidate ',line_number:1)
            {---teststop---}
         end { stopped on normal line }
         else if main_state = top_or_bottom then
         begin { hit top or bottom of buffer }
            if chain_allowed then
            begin { chain to next buffer }
               { temporarily reset main_state to avoid frightening load_a_line }
               { into stopping short }
               main_state := running;
               if find_next_chain_buffer(backwards, true)
               then main_state := running  { alright to continue now }
               else main_state := top_or_bottom  { failed to chain along }
            end { chain to next buffer } ;
            if main_state = top_or_bottom then
            begin { print .top. or .bottom. }
               if line_status = top
               then screen_message('.top.', true)
               else screen_message('.bottom.', true)
            end { print .top. or .bottom. }
         end { hit top or bottom of buffer } ;
         check_rest_of_line := false
      end { while (not terminate) etc. } ;

      if display_buffer_name then
      begin { buffer has been changed }
         if in_window_code
         then { buffer name will be printed later }
         else if main_state = running
         then announce_buffer_name
         else display_buffer_name := false  { do not announce last one }
      end { buffer has been changed } ;

      { do "later" unload and delete }

      if (main_state = running) or (main_state = top_or_bottom) then
      begin { do "later" unload and delete }
         if backwards then
         begin { backwards }
            if line_status = top
            then stop_line := 1
            else stop_line := line_number + 1;
            if delete_later and (stop_line <= start_line) then
            begin { do deletes }
               main_state := running;
               quick_delete(stop_line, start_line);
               load_a_line(stop_line - 1, false);
               test_line_landed_on
            end { do deletes }
         end { backwards }
         else
         begin { forwards }
            stop_line := line_number - 1;  { last affected by delete/unload }
            if stop_line >= start_line then
            begin { stop_line >= start_line }
               main_state := running;   { otherwise drops out in quick procs }
               if unload_later
               then if quick_unload(start_line, stop_line)
                    then load_a_line(line_number + 1, false)  { leave on next }
                    else main_state := command_error;
               if main_state = running
               then if delete_later
                    then quick_delete(start_line, stop_line);
               test_line_landed_on      { again }
            end { stop_line >= start_line }
         end { forwards }
      end { do "later" unload and delete }
   end { with }
end { loop_to_string } ;

{ Get start column number for FIND type commands. }
{ It appears in (..) after the command name. }

function get_start_column(var start_col: integer): boolean;                     { GET_START_COLUMN }

var
   result: boolean;

begin { get_start_column }
   sub := sub + 1;
   result := get_number(start_col);
   if (start_col > 0) and
      (start_col < line_max_chars) and
      (command_line[sub] = ')')
   then get_start_column := true
   else
   begin
      get_start_column := false;
      screen_message('Column number error.', true)
   end;
   sub := sub + 1
end { get_start_column } ;

%page
{***********************************}
{        COMMAND PROCEDURES         }
{ ( INTERNAL TO OBEY_COMMAND_LINE ) }
{***********************************}

{ ALPHABET command }

procedure alphabet_comm;                                                        { ALPHABET_COMM }

var
   ch: char;
   escaped, done: boolean;
   new_upper: set_of_char;
   count: integer;

   procedure add_in(ch: char);

   begin { add_in }
      { first convert to upper case }
      if ch >= '`'
      then ch := chr(ord('A') - ord('a') + ord(ch));
      if ch < '@' then
      begin { out of range }
         ch_screen_message('Character ~ is not in the permissible alphabetic range @ to _',
                           true, ch);
         main_state := command_error
      end { out of range }
      else new_upper := new_upper + [ch]
   end { add_in } ;

begin { alphabet_comm }
   skip_spaces;
   new_upper := [];
   if command_line[sub] in string_terminators then
   begin { no args - interactive method }
      screen_message('For each possible alphabet character, indicate whether you want to include it in the set. RETURN means yes.',
                     false);

      message := '~ : ';
      for ch := '@' to '_' do
      begin
         message[1] := ch;
         if question_and_answer(message, false, true)
         then new_upper := new_upper + [ch]
      end { for loop }
   end { no args - interactive method }
   else
   begin { args given - copy into set }
      escaped := false;
      done := false;
      repeat
         ch := command_line[sub];
         if escaped then
         begin
            escaped := false;
            add_in(ch)
         end
         else if ch = special[sescape]
         then escaped := true
         else if ch in string_terminators
         then done := true
         else add_in(ch);
         if not done
         then sub := sub + 1
      until done
   end { args given } ;
   if main_state = running then
   begin { set up alphabets }
      upper_alphabet := new_upper;
      lower_alphabet := [];
      count := 0;
      for ch := '@' to '_'
      do if ch in upper_alphabet then
         begin
            lower_alphabet := lower_alphabet +
                              [chr(ord('a') - ord('A') + ord(ch))];
            count := count + 1
         end;
      int_screen_message('~ characters in alphabet.', false, count, 1)
   end { set up alphabets }
end { alphabet_comm } ;

{ APPEND and OVERLAY commands }

procedure append_and_overlay_proc(overlay: boolean);                            { APPEND_AND_OVERLAY_PROC }

var
   true_length: 0..line_max_chars;
   start_col: 1..line_max_chars;
   finished: boolean;

begin { append_and_overlay_proc }
   with current_buffer do
   if command_line[sub] <> ' '
   then bad_command
   else if not line_exists then
   begin { no line }
      print_line;
      main_state := command_error
   end { no line }
   else
   begin { line exists }
      update_changed_lines(line_number);
      if overlay
      then start_col := 1               { overlay - start at col 1 }
      else
      begin { append - start after last non-space }
         true_length := line_length;
         finished := false;
         repeat
            if true_length <= 0
            then finished := true
            else if line_text[true_length] <> ' '
            then finished := true
            else true_length := true_length - 1
         until finished;
         start_col := true_length + 1
      end { append - start after last non-space } ;
      sub := sub + 1;
      copy_input_string(overlay, start_col);
      verify_print
   end { line exists }
end { append_and_overlay_proc } ;

{ BUFFER command - switch to another editor buffer to edit a different file }

procedure buffer_comm;                                                          { BUFFER_COMM }

var
   new_buffer_number, saved_buffer_number: integer;
   pattern: tree_type;
   state: search_type;
   try_it: boolean;

begin { buffer_comm }
   if get_number(new_buffer_number) then
   begin { BUFFER n - switch buffers }
      if new_buffer_number = current_buffer_number
      then { no need to switch }
      else if (new_buffer_number >= 1) and (new_buffer_number <= buffer_max)
      then
      begin { in range }
         switch_buffer(new_buffer_number);
         if mode_disp and in_window_code
         then display_buffer_name := true
      end { in range }
      else
      begin { out of range }
         int_screen_message('Buffer number must be 1..~', true, buffer_max, 1);
         main_state := command_error
      end { out of range }
   end { BUFFER n }
   else if command_line[sub] in string_terminators
   then list_buffers(false)             { no args - just list buffers }
   else with current_buffer, main_file do
   begin { BUFFER name - look for filename }
      copy_treename(pattern, true, false, false, true);
      saved_buffer_number := current_buffer_number;
      new_buffer_number := 1;
      state := searching;
      repeat
         try_it := true;
         if new_buffer_number = current_buffer_number
         then                           { already there }
         else if buffer_pointers[new_buffer_number].bptr = nil
         then try_it := false           { null buffer }
         else switch_buffer(new_buffer_number);
         if try_it
         then if pos(pattern.strng, treename.strng) > 0
              then state := found;
         if state = searching
         then if new_buffer_number < buffer_max
              then new_buffer_number := new_buffer_number + 1
              else state := not_found
      until state <> searching;
      if state = found then
      begin { matching buffer found }
         if in_window_code
         then display_buffer_name := true
         else announce_buffer_name
      end { matching buffer found }
      else
      begin { not found }
         switch_buffer(saved_buffer_number);
         comm_error('No matching buffer found.')
      end { not found }
   end { BUFFER name }
end { buffer_comm } ;

{ CENTRE command - centre current line between STYLE left and right margins }

procedure centre_comm;                                                          { CENTRE_COMM }

var
   first_non_space, chindex: line_chars_range;
   style_width, new_spaces, new_length: integer;

begin { centre_comm }
   with current_buffer, style_rec do
   begin
      if not line_exists then
      begin { null line }
         print_line;
         if line_status = normal
         then main_state := command_error  { .null. line }
         else main_state := top_or_bottom  { .top. or .bottom. }
      end { null line }
      else
      begin { normal line }
         update_changed_lines(line_number);
         strip_trailing_spaces;
         line_altered := true;

         { remove any leading spaces }

         first_non_space := 1;
         while line_text[first_non_space] = ' '
         do first_non_space := first_non_space + 1;
         if first_non_space > 1 then
         begin { remove spaces }
            for chindex := first_non_space to line_length+1
            do line_text[chindex - first_non_space + 1] := line_text[chindex];
            line_length := line_length - first_non_space + 1
         end { remove spaces } ;

         { insert spaces to centralise }

         if left_margin = 0
         then style_width := right_margin  { auto processing }
         else style_width := right_margin - left_margin + 1;
         new_spaces := (style_width - line_length) div 2;
         if new_spaces < 0
         then new_spaces := 0;
         if left_margin > 0
         then new_spaces := new_spaces + left_margin - 1;
         new_length := line_length + new_spaces;
         if new_length <= line_max_charsm1 then
         begin { new size alright }
            for chindex := line_length+1 downto 1
            do line_text[chindex+new_spaces] := line_text[chindex];
            for chindex := 1 to new_spaces
            do line_text[chindex] := ' ';
            line_length := new_length;
            verify_print
         end { new size alright }
         else comm_error('File line is full.')
      end { normal line }
   end { with }
end { centre_comm } ;

{ CHANGE and MODIFY }

procedure change_comm(modify_command: boolean);                                 { CHANGE_COMM }

var
   ch: char;
   chain_allowed, finished, loop_finished, changes, result,
   change_lines_given: boolean;
   state: search_type;
   change_string_length: line_chars0_range;
   change_col: line_chars_range;
   saved_line_number, markln, currln, stop_line: integer;
   saved_line_position: line_chars_range;
   saved_buffer_number: buffer_range;

   procedure store(ch: char);

   begin { store }
      change_string_length := change_string_length + 1;
      change_string[change_string_length] := ch
   end { store } ;

   function match: boolean;             { check string against line }

   var
      state: search_type;
      pos1, pos2, lindex: line_chars_range;
      linech: char;
      left_ok, right_ok: boolean;

   begin { match }
      with current_buffer do
      begin
         pos1 := 1;                     { -> change string }
         pos2 := change_col;            { -> current line }
         linech := line_text[pos2];
         state := searching;
         repeat
            if change_string[pos1] <> linech
            then state := not_found
            else if pos1 = change_length1
            then state := found
            else
            begin
               pos1 := pos1 + 1;
               if pos2 < line_length then
               begin
                  pos2 := pos2 + 1;     { still within line }
                  linech := line_text[pos2]
               end
               else if pos2 < line_max_charsm1 then
               begin { assume padded with spaces }
                  pos2 := pos2 + 1;
                  linech := ' '
               end { assume padded with spaces }
               else state := not_found  { end of max line }
            end
         until state <> searching;
         { if ident type change check that string is delimited by punctuation }
         if change_ident and (state = found) then
         begin { check delimiters }
            if change_col = 1
            then left_ok := true
            else left_ok := line_text[change_col-1] in punctuation;
            if pos2 >= line_length
            then right_ok := true
            else right_ok := line_text[pos2+1] in punctuation;
            if left_ok and right_ok
            then                        { both ends ok }
            else state := not_found
         end { check delimiters } ;
         if state = found then
         begin { it matches }
            match := true;
            if pos2 > line_length then
            begin { it was necessary to pad with spaces }
                                        { to get a match }
               for lindex := (line_length + 1) to pos2 do
                  line_text[lindex] := ' ';
               line_length := pos2;
               line_text[line_length] := newline
            end { it was necessary to pad with spaces }
            else { no need to pad }
         end { it matches }
         else match := false
      end { with }
   end { match } ;

   function test_new_length(nl: integer): boolean;
   begin { test_new_length }
      if nl > line_max_charsm1 then
      begin { out of range }
         comm_error('File line is full.');
         test_new_length := false
      end { out of range }
      else test_new_length := true
   end { test_new_length } ;

   procedure perform_change;

   var
      index: line_chars_range;
      new_length: integer;

   begin { perform_change }
      with current_buffer do
      begin

         { move rest of current line up or down if necessary }

         if change_same_length then
         begin { MODIFY or strings same length }
            new_length := change_col + change_length2 - 1;  { minimum required }
            if new_length > line_length
            then                        { can happen for MODIFY command }
               if test_new_length(new_length)
               then line_text[new_length + 1] := newline
               else                     { line full }
            else new_length := line_length  { do not shorten line }
         end { MODIFY or strings same length }
         else
         begin { lengths are different }
            new_length := line_length - change_length1 + change_length2;
            {copy rest of line up or down }
            if test_new_length(new_length)
            then movewds(line_text, change_col + change_length1,
                         line_text, change_col + change_length2,
                         line_length + 2 - change_col - change_length1)
            else                        { line is full }
         end { lengths are different } ;

         if main_state = running then
         begin { changed alright }
            line_length := new_length;

            { copy new text over old }
            for index := 1 to change_length2
            do line_text[change_col + index - 1] :=
               change_string[change_length1 + index];
            change_col := change_col + change_length2;

            { add spaces if necessary (MODIFY command) }
            for index := 1 to spaces_to_add do
            begin
               line_text[change_col] := ' ';
               change_col := change_col + 1
            end;
            line_altered := true
         end { changed alright }
      end { with }
   end { perform_change } ;

begin { change_comm }
   with current_buffer do
   begin
      saved_buffer_number := current_buffer_number;
      if line_status = top              { treat .top. as line 0 }
      then currln := 0
      else currln := line_number;
      saved_line_number := currln;
      saved_line_position := line_position
   end { with } ;

   { analyse command strings and options if we have to }

   if command_line[sub] in string_terminators
   then if change_string_exists         { just C alone: use last strings }
        then                            { ok - they exist }
        else bad_command
   else with current_buffer do
   begin { analyse strings }
      change_general := false;
      change_ident := false;
      change_to_mark := false;
      change_star := false;
      change_lines_given := false;
      change_lines := 1;
      if command_line[sub] = ' '
      then sub := sub + 1;
      change_string_length := 0;
      if not copy_string(store, true)   { copy first string }
      then bad_command                  { terminator not found }
      else
      begin
         change_length1 := change_string_length;
         change_zero := change_length1 = 0;
         if copy_string(store, true)    { copy second string }
         then sub := sub + 1;           { skip terminator }
         change_length2 := change_string_length - change_length1;
         if modify_command then
         begin
            change_same_length := true;
            if change_length1 > change_length2  { MODIFY may add spaces for lineup }
            then spaces_to_add := change_length1 - change_length2
            else spaces_to_add := 0
         end
         else
         begin
            change_same_length := change_length1 = change_length2;
            spaces_to_add := 0
         end;

         { look for 'g' etc. after command }

         finished := false;
         repeat
            ch := command_line[sub];
            if ch in string_terminators
            then finished := true
            else if ch = ' '
            then sub := sub + 1         { skip spaces }
            else if (ch = 'g') or (ch = 'G') then
            begin
               change_general := true;
               sub := sub + 1
            end
            else if (ch = 'i') or (ch = 'I') then
            begin
               change_ident := true;
               sub := sub + 1
            end
            else if (ch = '*') then     { change throughout file }
            begin
               change_star := true;
               change_lines := large_int;  { and change a lot of lines }
               sub := sub + 1
            end
            else if (ch = 'm') or (ch = 'M') then  { change from here to mark }
            begin
               change_to_mark := true;
               sub := sub + 1
            end
            else if ch in ['-', '0'..'9'] then
            begin
               change_lines_given := true;
               result := get_number(change_lines);
               if change_lines = 0
               then change_lines := 1
            end
            else bad_command
         until finished or (main_state <> running)
      end;
      if ord(change_to_mark) +          { only 0 or 1 of these line specs }
         ord(change_star) +             { is allowed }
         ord(change_lines_given)
         > 1
      then bad_command;

      { select "quick_locate" string etc. }

      if main_state = running then
      begin
         change_use_ql := false;
         {---teststart---}
         if mode_test then
         begin
            write('strings "');
            for change_col := 1 to change_length1
            do write(change_string[change_col]);
            write('" "');
            for change_col := (change_length1 + 1) to
                              (change_length1 + change_length2)
            do write(change_string[change_col]);
            writeln('"');
            writeln('change_general = ',change_general,
                  '  change_ident = ',change_ident);
            writeln('change_to_mark = ', change_to_mark,
                  '  change_lines = ',change_lines:1);
            writeln('change_zero = ',change_zero,
                  '  change_same_length = ',change_same_length,
                  '  spaces_to_add = ',spaces_to_add:1)
         end;
         {---teststop---}
         if ((change_lines <> 1) or     { if multi-line change and }
             change_to_mark) and
            (not change_zero) then      { non-zero string }
         begin { select "quick_locate" string }
            change_ql_start := 1;
            while (change_string[change_ql_start] = ' ') and
                  (change_ql_start <= change_length1) do
               change_ql_start := change_ql_start + 1;  { skip spaces }
            if change_ql_start <= change_length1 then
            begin
               change_ql_end := change_ql_start;
               while (change_string[change_ql_end] <> ' ') and
                     (change_ql_end <= change_length1)
               do change_ql_end := change_ql_end + 1;
               change_ql_end := change_ql_end - 1;  { last non-space }
               change_use_ql := true
               {---teststart---}
               ;if mode_test then
               begin
                  write('quick_locate string "');
                  for change_col := change_ql_start to change_ql_end
                  do write(change_string[change_col]);
                  writeln('"')
               end
               {---teststop---}
            end
         end { select ql string }
      end { main_state = running }
   end { analyse strings } ;

   { main loop }

   change_string_exists := main_state = running;
   if main_state = running then
   with current_buffer do
   begin { analysed OK }
      chain_allowed := mode_chain and change_star;
      if change_star
      then top_or_bottom_proc(true);    { go to .top. }
      if change_to_mark then
      begin { find line count for change to mark }
         if marked_line = 0             { if no marked line, use current }
         then markln := currln
         else markln := marked_line;
         update_changed_lines(markln);
         marked_line := 0;              { delete the mark }
         if markln < currln             { calculate number of lines }
         then change_lines := markln - currln - 1  { backwards }
         else change_lines := markln - currln + 1  { forwards }
         {---teststart---}
         ;if mode_test
         then writeln('change_lines (to mark) = ', change_lines:1)
         {---teststop---}
      end { find line count for change to mark } ;
      change_backwards := change_lines < 0;
      {---teststart---}
      if mode_test
      then writeln('change_backwards = ', change_backwards);
      {---teststop---}
      if change_backwards
      then stop_line := currln + change_lines + 1
      else stop_line := currln + change_lines - 1;
      loop_finished := false;
      repeat

         { perform any changes on current line if it exists }

         changes := false;
         if line_exists then
         begin { change current line }
            change_col := 1;
            finished := false;
            repeat
               if change_zero
               then state := found
               else
               begin
                  state := searching;
                  repeat
                     if change_col > line_length
                     then state := not_found
                     else if match
                     then state := found
                     else change_col := change_col + 1
                  until state <> searching
               end;
               if state = found then
               begin
                  perform_change;
                  changes := true
               end
               else finished := true
            until finished or (not change_general) or (main_state <> running)
         end { change current line } ;

         if line_status = top           { treat .top. as line 0 }
         then currln := 0
         else currln := line_number;

         if changes or (currln = stop_line) then
         begin { display changed line }
            if display_buffer_name and (main_state = running) then
            begin { buffer has been changed }
               if in_window_code
               then { buffer name will be printed later }
               else announce_buffer_name
            end { buffer has been changed } ;
            verify_print
         end { display changed line } ;

         if changes
         then update_changed_lines(line_number);

         { move on to the next candidate line }

         if currln = stop_line
         then loop_finished := true
         else
         begin { not at stop line }
            if change_use_ql
            then quick_locate(change_string, change_string,
                              change_ql_start, change_ql_end,
                              change_backwards, true, false, stop_line)
            else if change_backwards
            then load_a_line(line_number - 1, false)
            else if line_exists
            then load_a_line(line_number + 1, false)
            else load_a_line(line_number    , false);
            test_line_landed_on;
            if main_state = top_or_bottom then
            begin { hit top or bottom of buffer }
               if chain_allowed then
               begin { chain to next buffer }
                  { temporarily reset main_state to avoid frightening }
                  { load_a_line into stopping short }
                  main_state := running;
                  if find_next_chain_buffer(change_backwards, true)
                  then main_state := running  { alright to continue now }
                  else main_state := top_or_bottom  { failed to chain along }
               end { chain to next buffer } ;
               if (main_state = top_or_bottom) and (not change_star) then
               begin { print .top. or .bottom. }
                  if line_status = top
                  then screen_message('.top.', true)
                  else screen_message('.bottom.', true)
               end { print .top. or .bottom. }
            end { hit top or bottom of buffer }
         end { not at stop line }
      until loop_finished or (main_state <> running);

      if change_star then               { if '*' given by user }
      begin { '*' given }
         if main_state = top_or_bottom
         then main_state := running;    { OK to hit top or bottom }
         if current_buffer_number <> saved_buffer_number then
         begin { switch back to original buffer }
            switch_buffer(saved_buffer_number);
            if in_window_code
            then display_buffer_name := true
            else announce_buffer_name
         end { switch back to original buffer } ;
         { Change * with mode chain means we must redraw the screen, because }
         { any assortment of buffers including current could be affected. }
         if in_window_code and display_buffer_name
         then must_redraw_screen := true;
         load_a_line(saved_line_number, false);  { back to start posn }
         line_position := saved_line_position
      end { '*' given }
   end { analysed OK }
end { change_comm } ;

{ COMINPUT or LOADSTATE command }

procedure cominput_comm(ls: boolean);                                           { COMINPUT_COMM }

label 8720;

var
   old_savestat_treename, cominput_treename: tree_type;
   temp_int: integer;

   procedure error_trap_onunit(cf_ptr: integer);

   begin { error_trap_onunit }
      main_state := command_error;
      goto 8720
   end { error_trap_onunit } ;

begin { cominput_comm }
   obeying_loadstate := ls;
   old_savestat_treename := savestat_treename;
   skip_spaces;
   if command_line[sub] in string_terminators
   then cominput_treename.strng := ''
   else copy_treename(cominput_treename, true, false, false, true);
   if check_termination then
   begin { no rubbish after treename }
      if ls then
      begin { LOADSTATE command }
         if (cominput_treename.len = 0)
         then cominput_treename := savestat_treename
         else savestat_treename := cominput_treename
      end { LOADSTATE command } ;
      if cominput_treename.len = 0
      then comm_error('Filename must be specified.')
      else
      begin { filename not null }
         on('ERROR_TRAP', error_trap_onunit);
         cominput_lineno := 1;
         if cominput_treename.strng[1] in ['0'..'9'] then
         begin { read from buffer }
            if string_to_integer(cominput_treename.strng, temp_int) then
            begin { integer conversion alright }
               if (temp_int >= 1) and (temp_int <= buffer_max) then
               begin { buffer number in range }
                  cominput_buffer := temp_int
               end { buffer number in range }
            end { integer conversion alright }
            else comm_error('Error in buffer number.')
         end { read from buffer }
         else
         begin { read from file }
            cominput_buffer := 0;       { => read from file }
            reset(cominput_file, cominput_treename.strng)
         end { read from file }
      end { filename not null }
   end { no rubbish after treename } ;
8720:
   if main_state = command_error
   then savestat_treename := old_savestat_treename
   else cominput_on := true
end { cominput_comm } ;

{ Delete one or more file buffers }

procedure delbuf_comm;                                                          { DELBUF_COMM }

var
   numbers_given, buffer_no: integer;
   delete_current: boolean;

   { Delete the given file buffer; if it is current then move to suitable }
   { alternative buffer.                                                  }

   procedure delete_buffer(buffer_no: buffer_range);

   var
      saved_buffer_number, new_buffer_number: buffer_range;
      ok_to_delete: boolean;

   begin { delete_buffer }
      with current_buffer do
      begin
         saved_buffer_number := current_buffer_number;
         switch_buffer(buffer_no);
         if line_altered or file_altered then
         begin
            display_buffer_modified;
            ok_to_delete := question_and_answer('OK to delete? ', true, false)
         end
         else ok_to_delete := true;
         if ok_to_delete then
         begin
            empty_buffer;
            if saved_buffer_number = current_buffer_number then
            begin { deleting current buffer }
               find_next_used_buffer(true, new_buffer_number);
               if new_buffer_number = current_buffer_number
               then init_buffer         { no other buffers }
               else switch_buffer(new_buffer_number)
            end { deleting current buffer }
            else switch_buffer(saved_buffer_number);
            if buffer_pointers[buffer_no].bptr <> nil
            then dispose(buffer_pointers[buffer_no].bptr);
            buffer_pointers[buffer_no].bptr := nil;
            if lowest_empty_buffer > buffer_no
            then lowest_empty_buffer := buffer_no
         end { ok_to_delete }
         else switch_buffer(saved_buffer_number)
      end { with }
   end { delete_buffer } ;

begin { delbuf_comm }
   buffer_no := 0;                      { preset get_next_number }
   numbers_given := 0;                  { count }
   delete_current := false;
   while get_next_number(buffer_no) and (main_state = running) do
   begin
      if (buffer_no < 1) or (buffer_no > buffer_max) then
      begin
         int_screen_message('Buffer number must be 1..~', true, buffer_max,
                            1);
         main_state := command_error
      end
      else if buffer_no = current_buffer_number
      then delete_current := true
      else if buffer_pointers[buffer_no].bptr = nil
      then { buffer is empty }
      else delete_buffer(buffer_no);
      numbers_given := numbers_given + 1
   end { while get buffer number } ;
   { delete current buffer last if necessary; this simplifies the }
   { 'which buffer to change to' question }
   if (numbers_given = 0) or            { no numbers => delete current }
      delete_current                    { delete explicitly }
   then
   begin { delete current buffer }
      delete_buffer(current_buffer_number);
      if in_window_code
      then display_buffer_name := true
      else announce_buffer_name
   end { delete current buffer }
end { delbuf_comm } ;

{ DELETE command }

procedure delete_comm;                                                          { DELETE_COMM }

var
   ch: char;
   toword: word;
   lines_to_delete: integer;
   result: boolean;

begin { delete_comm }
   skip_spaces;
   ch := command_line[sub];
   if (ch = 't') or (ch = 'T') then
   begin { "DELETE TO string" form }
      get_word(toword);
      if toword = 'TO      ' then
      begin { format alright }
         loop_to_string([delete_act,locate_act], 1);
         verify_print
      end { format alright }
      else bad_command
   end { "DELETE TO string" form }
   else
   begin { "DELETE number" form }
      result := get_number(lines_to_delete);
      if lines_to_delete < 0
      then loop_by_number([delete_act,backwards_act], -lines_to_delete)
      else loop_by_number([delete_act]             ,  lines_to_delete)
   end { "DELETE number" form }
end { delete_comm } ;

{ EDIT, NEW and RELOAD commands }
{ If find_empty then switch to empty buffer first }
{ If use_current_pathname then reload same file }

procedure edit_comm(find_empty, use_current_pathname: boolean);                 { EDIT_COMM }

var
   new_treename: tree_type;
   new_count: integer;
   all_buffers_full, any_expansion, new_edit_mode: boolean;

   procedure do_edit_comm(the_treename: tree_type);

   var
      ok_to_lose: boolean;
      saved_buffer_number, new_buffer_number: integer;
      state: (searching, possible, found, not_found);
      saved_file_place: file_places;

   begin { do_edit_comm }
      with current_buffer do
      begin
         if find_empty then
         begin { switch to an empty buffer }
            saved_buffer_number := current_buffer_number;
            if (lowest_empty_buffer >= 1) and
               (lowest_empty_buffer <= buffer_max)
            then new_buffer_number := lowest_empty_buffer
            else new_buffer_number := 1;
            state := searching;
            repeat
               if new_buffer_number = current_buffer_number
               then if (main_file.treename.len > 0) or
                       (line_number > 1) or
                       (line_status = normal) or
                       line_altered or file_altered
                    then { already in use }
                    else state := possible
               else if buffer_pointers[new_buffer_number].bptr = nil
               then state := found      { definitely empty }
               else with buffer_pointers[new_buffer_number].bptr^ do
                    if (main_file.treename.len > 0) or
                       (line_number > 1) or
                       (line_status = normal) or
                       line_altered or file_altered
                    then { already in use }
                    else state := possible;
               if state = possible then
               begin { look at this one more closely }
                  switch_buffer(new_buffer_number);
                  { if line 1 is bottom then file is empty }
                  save_place(saved_file_place);
                  load_a_line(1, false);
                  if line_status = bottom
                  then state := found
                  else state := searching;
                  restore_place(saved_file_place)
               end { look at this one more closely } ;
               if state = searching then
               begin { still searching }
                  if new_buffer_number < buffer_max
                  then new_buffer_number := new_buffer_number + 1
                  else state := not_found
               end { still searching }
            until state <> searching;
            if state = found then
            begin { empty found }
               switch_buffer(new_buffer_number);
               lowest_empty_buffer := new_buffer_number + 1
            end { empty found }
            else
            begin { switch failed }
               screen_message('There are no empty buffers.', true);
               switch_buffer(saved_buffer_number);
               main_state := command_error;
               all_buffers_full := true
            end { switch failed }
         end { switch to an empty buffer }
         else
         begin { stick with current buffer }
            if use_current_pathname
            then { RELOAD does not change anything }
            else lowest_empty_buffer := 1
         end { stick with current buffer } ;
         if main_state = running then
         begin { buffer switch alright }
            if file_altered or line_altered then
            begin { check if ok to overwrite }
               if multiple_buffers_used
               then display_buffer_modified;
               ok_to_lose := question_and_answer(
                  'Is it alright to lose the edits you have done? ',
                  true, false)
            end { check if ok to overwrite }
            else ok_to_lose := true;
            if ok_to_lose then
            with main_file do
            begin { ok to lose edits }
               must_redraw_window := true;
               empty_buffer;            { throw away current contents }
               init_buffer;             { set up buffer again }
               treename := the_treename;
               if open_edit_file        { set up the new file }
               then new_edit_mode := treename.len > 0
               else
               begin
                  treename.strng := '';  { clear erroneous treename }
                  main_state := command_error
               end;
               if find_empty or use_current_pathname then
               begin { display pathname }
                  if in_window_code
                  then display_buffer_name := true
                  else announce_buffer_name
               end { display pathname }
            end { ok to lose edits }
            else screen_message('Please SAVE first.', true)
         end { buffer switch alright }
      end { with }
   end { do_edit_comm } ;

   { Deal with an expanded treename generated by wildcarding etc. }

   procedure deliver(expanded_pathname: tree_string;
                     var continue_expanding: boolean);

   var
      temp_treename: tree_type;

   begin { deliver }
      if main_state = running then
      begin
         temp_treename.strng := expanded_pathname;
         do_edit_comm(temp_treename);
      end;
      { do not stop on most errors, only on user interrupt }
      if main_state = command_error
      then main_state := running
      else new_count := new_count + 1;
      continue_expanding := (main_state = running) and (not all_buffers_full)
   end { deliver } ;

begin { edit_comm }
   with current_buffer do
   begin
      new_edit_mode := true;
      all_buffers_full := false;
      if use_current_pathname then
      begin { check RELOAD }
         new_treename := main_file.treename;
         if length(new_treename.strng) = 0
         then comm_error('Cannot reload because buffer has no filename.')
      end { check RELOAD }
      else copy_treename(new_treename, true, find_empty, false, true);
      if (main_state = running) and check_termination then
      begin { pathname alright }
         { The NEW command allows iteration etc. in pathname }
         if find_empty then
         begin { NEW command }
            if new_treename.len = 0
            then do_edit_comm(new_treename)  { no treename given }
            else
            begin { treename given }
               new_count := 0;
               expand_pathname(new_treename.strng, any_expansion, deliver);
               if any_expansion then
               begin { report files loaded }
                  if new_count = 0
                  then screen_message('No files match pathname given.', false)
                  else int_screen_message('~ files loaded into buffers.',
                                          false, new_count, 1)
               end { report files loaded }
            end { treename given }
         end { NEW command }
         else do_edit_comm(new_treename);
         { switched to input mode? }
         if not new_edit_mode
         then set_edit_mode(new_edit_mode)
      end { pathname alright }
   end { with }
end { edit_comm } ;

{ Expand any tab characters in the current line according to htabs array }

procedure expand_comm;                                                          { EXPAND_COMM }

var
   to_line: line_chars_type;
   from_col, to_col: line_chars_range;
   overflowed, tab_found: boolean;
   ch: char;
   state: search_type;
   tabno: tabs_range;

   procedure store(ch: char);
   begin { store }
      if to_col < line_max_chars then
      begin
         to_line[to_col] := ch;
         to_col := to_col + 1
      end
      else overflowed := true
   end { store } ;

begin { expand_comm }
   with current_buffer do
   if not line_exists then
   begin
      print_line;
      if line_status = normal
      then main_state := command_error  { .null. line }
      else main_state := top_or_bottom  { .top. or .bottom. }
   end
   else
   begin { line exists }
      to_col := 1;
      overflowed := false;
      tab_found := false;
      for from_col := 1 to line_length do
      begin
         ch := line_text[from_col];
         if ch = special[stab] then
         begin { tab character }
            tab_found := true;
            tabno := 1;                 { find next screen tab }
            state := searching;
            repeat
               if tabno > number_of_tabs
               then state := not_found
               else if htabs[tabno] > to_col
               then state := found
               else if tabno >= tabs_max
               then state := not_found
               else tabno := tabno + 1
            until state <> searching;
            if state = found then
            begin { replace tab with spaces }
               while to_col < htabs[tabno]
               do store(' ')
            end { replace tab with spaces }
            else store(' ')             { no tabs left }
         end { tab character }
         else store(ch)
      end { for loop } ;
      if tab_found then
      begin { copy changed line }
         if overflowed
         then comm_error('File line is full.')
         else
         begin { new version fits }
            line_text := to_line;       { copy the new version }
            line_length := to_col - 1;
            line_text[to_col] := newline;
            line_altered := true;
            verify_print;
            update_changed_lines(line_number)
         end { new version fits }
      end { copy changed line }
   end { line exists }
end { expand_comm } ;

{ Extract a vertical stripe of text from current buffer and copy to another }

procedure extract_comm;                                                         { EXTRACT_COMM }

var
   main_buffer_number: buffer_range;
   merge_column, merge_column_2, last_column, new_mark, option_no: integer;
   option_word: word;
   lindex: line_chars_range;
   copy_line: line_chars_type;
   copy_length: line_chars0_range;
   blank_flag, delete_flag: boolean;
   saved_file_place: file_places;

begin { extract_comm }
   with current_buffer, load_file do
   begin
      save_place(saved_file_place);
      file_type := sam_type;
      copy_treename(treename, true, false, false, true);
      if not get_number(merge_column)
      then comm_error('Column number missing.')
      else if not get_number(merge_column_2)
      then comm_error('Column number missing.')
      else if (merge_column < 1) or (merge_column > line_max_charsm1)
      then comm_error('Column number out of range.')
      else if (merge_column_2 < 1) or (merge_column_2 > line_max_charsm1) or
              (merge_column_2 < merge_column)
      then comm_error('Column number out of range.')
      else
      begin
         blank_flag := false;           { analyse any options present }
         delete_flag := false;
         for option_no := 1 to 2 do
         begin
            get_word(option_word);
            if option_word = '        '
            then
            else if compare(option_word, 'BLANK   ')
            then blank_flag := true
            else if compare(option_word, 'DELETE  ')
            then delete_flag := true
            else bad_command
         end { for loop } ;
         if not line_exists
         then load_a_line(line_number, false);
         if line_status = bottom
         then load_a_line(line_number - 1, false);
         if line_status = normal then
         begin { not null file }
            original_pathname := false;
            rename_done := false;
            open_file(k$writ, load_file, 0, false, true);
            if (main_state = running) and file_is_open then
            begin
               main_buffer_number := current_buffer_number;
               if marked_line = 0 then  { if no mark then }
               begin { process whole file }
                  load_a_line(maxint, false);
                  marked_line := line_number - 1;
                  load_a_line(1, false)
               end { process whole file } ;
               if marked_line < line_number then
               begin { swap so that mark is after current }
                  new_mark := line_number;
                  load_a_line(marked_line, false);
                  marked_line := new_mark
               end { swap so that mark is after current } ;
               repeat
                  if merge_column_2 > line_length
                  then last_column := line_length
                  else last_column := merge_column_2;
                  if last_column < merge_column
                  then copy_length := 0
                  else copy_length := last_column - merge_column + 1;
                  movewds(line_text, merge_column, copy_line, 1, copy_length);
                  copy_line[copy_length+1] := newline;
                  if blank_flag or delete_flag then
                  begin { change 'from' line }
                     if blank_flag then
                     begin { blank out extracted part }
                        for lindex := merge_column to last_column
                        do line_text[lindex] := ' '
                     end { blank out extracted part }
                     else if delete_flag then
                     begin { delete extracted part }
                        movewds(line_text, merge_column+copy_length,
                                line_text, merge_column,
                                line_length+2-copy_length-merge_column);
                        line_length := line_length - copy_length
                     end { delete extracted part } ;
                     line_altered := true;
                     must_redraw_window := true
                  end { change 'from' line } ;
                  switch_buffer(load_buffer_number);
                  replace_line;
                  line_text := copy_line;
                  line_length := copy_length;
                  insert_a_line;
                  switch_buffer(main_buffer_number);
                  load_a_line(line_number+1, false)
               until (line_number > marked_line) or (main_state <> running);
               marked_line := 0;
               switch_buffer(load_buffer_number);
               load_a_line(0, false);
               file_altered := false;   { leave unaltered }
               switch_buffer(main_buffer_number)
            end { opened OK }
            else main_state := command_error
         end { not null file }
         else screen_message('Cannot extract from null file.', true)
      end { no obvious syntax errors } ;
      if main_state = top_or_bottom
      then main_state := running;       { OK to hit top or bottom }
      restore_place(saved_file_place)
   end { with }
end { extract_comm } ;

{ FIND and LOCATE and similar commands }

procedure find_or_locate_proc(what_to_do: set_of_actions);                      { FIND_OR_LOCATE_PROC }

var
   start_column: integer;

begin { find_or_locate_proc }
   start_column := 1;
   if command_line[sub] = '('           { look for start column in (..) }
   then if not get_start_column(start_column)  { for FIND and LOCATE }
        then bad_command;
   if main_state = running then
   begin { start column alright }
      loop_to_string(what_to_do, start_column);
      verify_print
   end { start column alright }
end { find_or_locate_proc } ;

{ FNAME command }

procedure fname_comm;                                                           { FNAME_COMM }

var
   new_treename: tree_type;

begin { fname_comm }
   skip_spaces;
   ch := command_line[sub];
   with current_buffer, main_file do
   begin
      if ch in string_terminators
      then write_tree(main_file)
      else
      begin { new filename present }
         copy_treename(new_treename, true, false, false, true);
         if check_termination then
         begin { new treename alright }
            treename := new_treename;
            original_pathname := false;
            rename_done := false
         end { new treename alright }
      end { new filename present }
   end { with }
end { fname_comm } ;

{ GMODIFY command }

procedure gmodify_comm;                                                         { GMODIFY_COMM }

var
   from_line, to_line: line_chars_type;
   from_col, to_col: line_chars_range;
   from_length: line_chars0_range;
   cols: integer;
   notflag, hit_bottom, finished, overflowed, result: boolean;
   errors: boolean;
   commch: char;

   procedure gmoderr;
   begin { gmoderr }
      bad_command;
      errors := true                    { test this for loop termination to }
                                        { avoid dropping out on quit }
   end { gmoderr } ;

   procedure store(ch: char);
   begin { store }
      if to_col < line_max_chars then
      begin
         to_line[to_col] := ch;
         to_col := to_col + 1
      end
      else overflowed := true
   end { store } ;

   procedure copy_to_line;

   begin { copy_to_line }
      with current_buffer do
      begin
         if overflowed
         then comm_error('File line is full.');
         line_text := to_line;          { copy the new version }
         line_length := to_col - 1;
         line_text[to_col] := newline;
         line_altered := true;
         verify_print
      end
   end { copy_to_line } ;

   procedure copy_enclosed_string(overlay, update_from: boolean);

   var
      specials: set_of_char;
      terminator, ch: char;
      escaped: boolean;
      state: search_type;

      procedure store_and_move(ch: char);
      begin { store_and_move }
         store(ch);
         if update_from and (from_col <= from_length)
         then from_col := from_col + 1
      end { store_and_move } ;

   begin { copy_enclosed_string }
      sub := sub + 1;
      terminator := command_line[sub];
      if terminator = newline
      then gmoderr
      else
      begin
         if overlay
         then specials := gmod_overlay_specials
         else specials := gmod_input_specials;
         specials := specials + [terminator];
         escaped := false;
         state := searching;
         repeat
            sub := sub + 1;
            ch := command_line[sub];
            if ch = newline
            then  state := not_found    { terminator not found }
            else  if escaped then
            begin
               store_and_move(ch);
               escaped := false
            end
            else if (not (ch in specials) )  { not a special }
            then store_and_move(ch)
            else if ch = terminator
            then state := found         { terminator found }
            else if ch = special[scounter]
            then store_count_value(store_and_move)
            else if ch = special[sescape]
            then escaped := true
            else if ch = ' '
            then if from_col > from_length
                 then store(' ')
                 else store_and_move(from_line[from_col])
            else if ch = special[swild]
            then store_and_move(' ')
            else abort('special character not recognised')
         until state <> searching;
         if state = not_found           { if no closing terminator }
         then sub := sub - 1            { backspace one col }
      end
   end { copy_enclosed_string } ;

   procedure copy_upto_ch(copy: boolean);

   var
      uptoch, ch: char;
      finished: boolean;

   begin { copy_upto_ch }
      sub := sub + 1;
      uptoch := command_line[sub];
      if uptoch = newline
      then gmoderr
      else
      begin
         finished := false;
         repeat
            ch := from_line[from_col];
            if ch = newline
            then finished := true       { always stop at newline }
            else if (not notflag and (ch =  uptoch)) or
                    (    notflag and (ch <> uptoch))
            then finished := true
            else
            begin
               if copy
               then store(ch);
               from_col := from_col + 1
            end
         until finished
      end;
      notflag := false
   end { copy_upto_ch } ;

   procedure copy_n_chars(copy: boolean);

   var
      cols_to_copy: integer;
      lindex: line_chars_range;
      result: boolean;

   begin { copy_n_chars }
      sub := sub + 1;
      if command_line[sub] in ['0'..'9'] then
      begin
         result := get_number(cols_to_copy);
         sub := sub - 1;                { otherwise miss next command }
         if (cols_to_copy + from_col) > (from_length + 1)
         then cols_to_copy := from_length + 1 - from_col;  { rest of line }
         if copy
         then for lindex := from_col to (from_col + cols_to_copy - 1)
              do  store(from_line[lindex]);
         from_col := from_col + cols_to_copy
      end
      else gmoderr
   end { copy_n_chars } ;

   procedure copy_rest;

   var
      cols_to_copy: line_chars0_range;
      lindex: line_chars_range;

   begin { copy_rest }
      cols_to_copy := from_length + 1 - from_col;
      for lindex := from_col to (from_col + cols_to_copy - 1)
      do store(from_line[lindex]);
      from_col := from_col + cols_to_copy
   end { copy_rest } ;

   procedure joinline;

   begin { joinline }
      with current_buffer do
      begin
         must_redraw_window := true;
         delete_line;                   { delete old first part }
         load_a_line(line_number, true);
         if line_status = bottom then
         begin { simulate zero length line }
            hit_bottom := true;         { remember because of insert later }
            line_length := 0;
            line_text[1] := newline;
            insert_a_line               { create a new line at bottom }
         end { simulate zero length line } ;
         from_line := line_text;        { copy text }
         from_length := line_length;
         from_col := 1
      end { with }
   end { joinline } ;

   procedure breakline;

   begin { breakline }
      with current_buffer do
      begin
         must_redraw_window := true;
         copy_to_line;                  { copy new text }
         replace_line;
         insert_a_line;                 { start off new line }
         to_col := 1
      end { with }
   end { breakline } ;

begin { gmodify_comm }
   with current_buffer do
   if not line_exists then
   begin
      print_line;
      if line_status = normal
      then main_state := command_error  { .null. line }
      else main_state := top_or_bottom  { .top. or .bottom. }
   end
   else
   begin { line exists }
      update_changed_lines(line_number);
      from_line := line_text;           { copy old line }
      from_length := line_length;
      from_col := 1;
      to_col := 1;
      notflag := false;
      overflowed := false;
      errors := false;
      hit_bottom := false;
      finished := false;
      repeat
         sub := sub + 1;
         skip_spaces;
         commch := command_line[sub];
         if commch in string_terminators
         then finished := true
         else
         begin
            if commch in ['a'..'z']     { convert to upper case }
            then commch := chr(ord('A') - ord('a') + ord(commch));
            case commch of
            'A':                        { Append }
               begin
                  copy_rest;
                  copy_enclosed_string(false, false)
               end;
            'B':                        { Back n }
               begin
                  sub := sub + 1;
                  if command_line[sub] in ['0'..'9'] then
                  begin
                     result := get_number(cols);
                     sub := sub - 1;
                     if cols >= from_col
                     then from_col := 1
                     else from_col := from_col - cols
                  end
                  else gmoderr
               end;
            'C':                        { Copy up to ch }
               copy_upto_ch(true);
            'D':                        { Delete up to ch }
               copy_upto_ch(false);
            'E':                        { Erase n }
               copy_n_chars(false);
            'F':                        { Finish }
               copy_rest;
            'I':                        { Insert }
               copy_enclosed_string(false, false);
            'J':                        { Join to next line }
               joinline;
            'K':                        { breaK line into two }
               breakline;
            'M':                        { Move n }
               copy_n_chars(true);
            'N':                        { Not }
               notflag := true;
            'O':                        { Overlay }
               copy_enclosed_string(true, true);
            'R':                        { Retype }
               copy_enclosed_string(false, true);
            'S':                        { Start }
               from_col := 1
            otherwise
               gmoderr                  { command not recognised }
            end { case }
         end
      until finished or errors;
      if not errors
      then copy_to_line;
      if hit_bottom
      then main_state := top_or_bottom  { Join hit bottom }
      else test_line_landed_on          { in case of join(s) }
   end { line exists }
end { gmodify_comm } ;

{ Call the HELP system }

procedure help_comm;                                                            { HELP_COMM }

var
   help_key: tree_type;

begin { help_comm }
   { get the parameters if any }
   copy_treename(help_key, true, true, false, false);
   if in_window_code
   then do_clear_screen;
   help_system(help_key.strng);
   must_redraw_screen := true           { ensure screen is redrawn }
end { help_comm } ;

{ INPUT mode }

procedure input_proc;                                                           { INPUT_PROC }

begin { input_proc }
   with current_buffer do
   begin
      replace_line;                     { ensure file copy ok }
      copy_input_string(false, 1);
      insert_a_line;
      must_redraw_window := true
   end { with }
end { input_proc } ;

{ INSERT and RETYPE commands }

procedure insert_and_retype_proc(insert: boolean;                               { INSERT_AND_RETYPE_PROC }
                                 before: boolean);

var
   must_create: boolean;

begin { insert_and_retype_proc }
   with current_buffer do
   begin
      if before                         { IB command }
      then load_a_line(line_number - 1, false);
      if command_line[sub] <> ' ' then
      begin
         must_redraw_window := true;
         if insert
         then if command_line[sub] = newline then
              begin
                 edit_mode := false;    { INSERT alone -> input mode }
                 change_to_input_mode
              end
              else bad_command
         else bad_command
      end
      else
      begin
         must_create :=                 { (re)create line if ... }
            insert or                   { INSERT or ... }
            (not line_exists);          { RETYPE .null. line }
         sub := sub + 1;
         if must_create then
         begin
            replace_line;
            must_redraw_window := true
         end
         else update_changed_lines(line_number);
         copy_input_string(false, 1);
         if must_create then insert_a_line
      end
   end
end { insert_and_retype_proc } ;

{ LINESZ command }

procedure linesz_comm;                                                          { LINESZ_COMM }

begin { linesz_comm }
   if get_number(test_line_size)
   then                                 { number actually given }
   else test_line_size := line_max_charsm1  { back to default }
end { linesz_comm } ;

{ LOAD command }

procedure load_comm;                                                            { LOAD_COMM }

begin { load_comm }
   with current_buffer, load_file do
   begin
      file_type := sam_type;
      copy_treename(treename, true, false, false, true);
      if treename.chars[1] = ' '
      then bad_command
      else if check_termination then
      begin { treename given }
         if load_a_file then
         begin
            screen_message('Edit', false);
            must_redraw_window := true
         end
         else main_state := command_error
      end { treename given }
   end { with }
end { load_comm } ;

{ Open the macro file for reading or writing }

function open_macro_file(var macfile: text;                                     { OPEN_MACRO_FILE }
                         for_writing: boolean): boolean;

var
   opened_ok: boolean;
   old_mac_treename: tree_type;

begin { open_macro_file }
   old_mac_treename := mac_treename;
   opened_ok := false;
   skip_spaces;
   if command_line[sub] in string_terminators
   then screen_message(mac_treename.strng, false)
   else copy_treename(mac_treename, true, false, false, true);
   if check_termination then
   begin { no rubbish after treename }
      if mac_treename.len = 0
      then comm_error('Filename must be specified.')
      else
      begin { filename not null }
         if for_writing
         then rewrite(macfile, mac_treename.strng)
         else reset  (macfile, mac_treename.strng);
         opened_ok := true
      end { filename not null }
   end { no rubbish after treename } ;
   if not opened_ok
   then main_state := command_error;
   { restore old filename if any error }
   if main_state = command_error
   then mac_treename := old_mac_treename;
   open_macro_file := opened_ok
end { open_macro_file } ;

{ Clear out all macros defined }

procedure macclear_comm;                                                        { MACCLEAR_COMM }

var
   ch: char;

begin { macclear_comm }
   for ch := first_print_char to last_print_char do
   with macro_pointers[ch] do
   begin
      free_macro(pointers[posm]);
      free_macro(pointers[negm])
   end
end { macclear_comm } ;

{ Load macros from a file }

procedure macload_comm;                                                         { MACLOAD_COMM }

label 121;

var
   macfile: text;
   visible_format, errors: boolean;
   line_number: integer;

   { Error encountered when reading macro file }

   procedure macl_error_message;

   begin { macl_error_message }
      int_screen_message('Error in macro file on line ~.', true,
                         line_number, 1);
      errors := true
   end { macl_error_message } ;

   { Read a macro file in numeric format.  The file format is decimal numbers }
   { 0..127 ASCII char, 128..  action ord + 128                               }
   { Converts to new ECS internal format:                                     }
   { 0..255 ASCII char, 256..  action ord + 256                               }

   procedure read_numeric_format;

   var
      mindex, macro_code, total: integer;
      id_char: char;
      mac_type: macro_type;

   begin { read_numeric_format }
      errors := false;
      { skip any null lines }
      while (not eof(macfile)) and (eoln(macfile))
      do readln(macfile);
      while not (eof(macfile) or errors) do
      begin
         if macfile^ = '^' then
         begin { ^code means neg macro }
            get(macfile);
            mac_type := negm
         end { ^code means neg macro }
         else mac_type := posm;         { no ^ means positive macro }
         read(macfile, macro_code);
         id_char := chr(macro_code+128);
         if id_char in ascii_printable then
         begin { code in file OK }
            readln(macfile, total);     { first read chars and store in temp }
            line_number := line_number + 1;
            new_macro_buffer(macro_head);  { macro buffer }
            store_macro_count := 0;
            if total > 0 then
            begin { read text }
               for mindex := 1 to total do
               begin
                  read(macfile, macro_code);
                  macro_code := macro_code + 128;
                  store_macro_code(macro_code);
                  if eoln(macfile) then
                  begin { deal with end of line }
                     readln(macfile);
                     line_number := line_number + 1
                  end { deal with end of line }
               end { for loop }
            end { read text } ;
            with macro_pointers[id_char] do
            begin
               free_macro(pointers[mac_type]);  { remove old version of macro }
               new(pointers[mac_type]);  { set up the new version }
               with pointers[mac_type]^ do
               begin
                  code_start := macro_head;
                  macro_head := nil;
                  total_count := store_macro_count;
                  macro_being_obeyed := false
               end
            end { with }
         end { code in file OK }
         else macl_error_message
      end { while loop }
   end { read_numeric_format } ;

   { Read a macro file in visible format }

   procedure read_visible_format;

   var
      macro_code: integer;
      ch, id_char: char;
      mac_type: macro_type;
      name_text: string8;
      done: boolean;

      { get the symbol name found }

      procedure get_symbol(var name_text: string8);

      var
         state: search_type;

      begin { get_symbol }
         name_text := '';
         state := searching;
         repeat
            if eoln(macfile)
            then state := not_found
            else
            begin { not eoln }
               read(macfile, ch);
               if ch = '}'
               then state := found
               else if length(name_text) >= 8
               then state := not_found
               else
               begin { add to end }
                  insert(' ',name_text,length(name_text)+1);
                  name_text[length(name_text)] := ch
               end { add to end } ;
               if name_text = '{'
               then state := found      { { coded as {{ }
            end { not eoln }
         until state <> searching;
         if state = not_found
         then macl_error_message
      end { get_symbol } ;

   begin { read_visible_format }
      errors := false;
      { skip any null lines }
      while (not eof(macfile)) and (eoln(macfile))
      do readln(macfile);
      while not (eof(macfile) or errors) do
      begin
         { first char is ' ' or '^' }
         read(macfile, ch);
         if ch = ' '
         then mac_type := posm
         else if ch = '^'
         then mac_type := negm
         else macl_error_message;
         read(macfile, id_char);
         if id_char in ascii_printable then
         begin { code in file OK }
            { skip spaces }
            while (not eof(macfile)) and (macfile^ = ' ')
            do get(macfile);
            read(macfile,ch);
            if ch = '"' then
            begin { first char alright }
               new_macro_buffer(macro_head);  { macro buffer }
               store_macro_count := 0;
               done := false;
               repeat
                  if eoln(macfile)
                  then done := true
                  else
                  begin { not eoln }
                     read(macfile, ch);
                     if (ch = '"') and eoln(macfile)
                     then done := true  { last  " on line }
                     else
                     begin { not end of macro }
                        { try to recognise the character }
                        if ch = '{' then
                        begin { starts with { }
                           get_symbol(name_text);
                           if name_text = '{'
                           then store_macro_code(ord('{'))
                           else if visible_to_macro_code(name_text, macro_code)
                           then store_macro_code(macro_code)
                           else macl_error_message
                        end { starts with { }
                        else store_macro_code(ord(ch))  { simple character }
                     end { not end of macro }
                  end { not eoln }
               until done;
               readln(macfile);
               line_number := line_number + 1;
               with macro_pointers[id_char] do
               begin
                  free_macro(pointers[mac_type]);  { remove old macro }
                  new(pointers[mac_type]);  { set up the new version }
                  with pointers[mac_type]^ do
                  begin
                     code_start := macro_head;
                     macro_head := nil;
                     total_count := store_macro_count;
                     macro_being_obeyed := false
                  end { with }
               end { with }
            end { first char alright }
            else macl_error_message
         end { code in file OK }
         else macl_error_message
      end { while loop }
   end { read_visible_format } ;

   procedure onunit(not_used: integer);
   begin { onunit }
      errors := true;
      goto 121
   end { onunit } ;

begin { macload_comm }
   if current_macro_desc <> nil then
   begin { called while obeying a macro - avoid nasty overwrite }
      screen_message('Cannot load macros while obeying a macro.', true);
      errors := true
   end { obeying macro }
   else
   begin { not obeying macro }
      free_mbufs(macro_head);           { free macro being recorded if any }
      on('ERROR_TRAP', onunit);
      if open_macro_file(macfile, false) then
      begin

         { Decide whether the file is in numeric or visible format }
         { by looking for first '"' }

         { skip any null lines }
         while (not eof(macfile)) and (eoln(macfile))
         do readln(macfile);
         if eof(macfile)
         then visible_format := false
         else
         begin { test for " on line }
            visible_format := false;
            while not eoln(macfile) do
            begin
               if macfile^ = '"'
               then visible_format := true;
               get(macfile)
            end { while loop }
         end { test for " on line } ;
         {---teststart---}
         if mode_test
         then writeln('MACLOAD: visible_format = ',visible_format);
         {---teststop---}
         reset(macfile);                { back to beginning }
         line_number := 1;

         if visible_format
         then read_visible_format
         else read_numeric_format

      end { file opened OK }
      else errors := true               { failed to open file }
   end { not obeying macro } ;
121:
   if errors then
   begin
      free_mbufs(macro_head);
      main_state := command_error
   end
end { macload_comm } ;

{ Print one or all macros in user-friendly form }

procedure macprint_comm;                                                        { MACPRINT_COMM }

var
   macro_char: char;
   no_printed, error_count, spaces_left: integer;

   procedure ensure_space(needed: integer);

   begin { ensure_space }
      if needed > spaces_left then
      begin
         screen_message(message, false);
         message := '     ';
         spaces_left := 74
      end;
      spaces_left := spaces_left - needed
   end { ensure_space } ;

   procedure print_macro(ch: char);

      procedure print_macro_text(not_ch: char; the_ptr: ptr_macro_desc);

      var
         cindex: mb_range;
         ptr: ptr_macro_buffer;
         total_chs, write_chs: integer;
         visible_rep: string12;

      begin { print_macro_text }
         with the_ptr^ do
         begin
            message := 'xx   "';
            message[1] := not_ch;
            message[2] := ch;
            no_printed := no_printed + 1;
            total_chs := total_count;
            ptr := code_start;
            spaces_left := 73;
            while total_chs > 0 do
            with ptr^ do
            begin
               if total_chs > code_count
               then write_chs := code_count
               else write_chs := total_chs;
               total_chs := total_chs - write_chs;
               for cindex := 1 to write_chs do
               begin
                  macro_code_to_visible(macro_text[cindex], visible_rep);
                  ensure_space(length(visible_rep));
                  append_message(visible_rep)
               end { for loop } ;
               ptr := next
            end { while loop, with ptr^ } ;
            ensure_space(1);
            append_char('"');
            screen_message(message, false)
         end { with }
      end { print_macro_text } ;

   begin { print_macro }
      if ch in ascii_printable then
      begin { valid macro char }
         with macro_pointers[ch] do
         if (pointers[posm] = nil) and (pointers[negm] = nil) then
         begin { no macros defined }
            ch_screen_message(' ~   No macro(s) defined.', false, ch);
            error_count := error_count + 1
         end { no macros defined }
         else
         begin { macro(s) defined }
            if pointers[posm] <> nil
            then print_macro_text(' ', pointers[posm]);
            if pointers[negm] <> nil
            then print_macro_text('^', pointers[negm])
         end { macro(s) defined }
      end { valid macro char }
      else
      begin { not a valid macro char }
         ch_screen_message('~ is not a valid macro char.', true, ch);
         main_state := command_error
      end { not a valid macro char }
   end { print_macro } ;

begin { macprint_comm }
   if command_line[sub] = ' '
   then sub := sub + 1;                 { skip one space if present }
   no_printed := 0;
   error_count := 0;
   if command_line[sub] = chr(lf_code) then
   begin { no argument(s) - print all }
      for macro_char := first_print_char to last_print_char do
      with macro_pointers[macro_char] do
      begin
         if (pointers[posm] <> nil) or (pointers[negm] <> nil) then
         begin
            if main_state = running     { no print if quit }
            then print_macro(macro_char)
         end
      end { for loop } ;
      if (no_printed = 0) and (main_state = running)
      then screen_message('No macros defined.', false)
   end { print all macros }
   else
   begin { print individual macro(s) }
      while not (command_line[sub] = chr(lf_code)) do
      begin
         macro_char := command_line[sub];
         sub := sub + 1;
         if main_state = running
         then print_macro(macro_char)
      end { while loop }
   end { print individual macros } ;
   if error_count > 0
   then main_state := command_error
end { macprint_comm } ;

{ Save all macros into a file }

procedure macsave_comm;                                                         { MACSAVE_COMM }

label 122;

const
   codes_per_line = 20;

var
   macfile: text;
   id_char: char;

   procedure write_macro_text(neg_ind, id_char: char; the_ptr: ptr_macro_desc);

   var
      ptr: ptr_macro_buffer;
      line_count, total_chs, write_chs: integer;
      cindex: mb_range;
      visible_rep: string12;

   begin { write_macro_text }
      with the_ptr^ do
      begin
         write(macfile, neg_ind);       { space or ^ }
         line_count := 1;               { governs newlines in file }
         write(macfile,id_char,'   "');
         total_chs := total_count;      { total characters to be written }
         ptr := code_start;
         while total_chs > 0 do
         with ptr^ do
         begin
            if total_chs > code_count
            then write_chs := code_count  { write whole of block }
            else write_chs := total_chs;  { write last few characters }
            total_chs := total_chs - write_chs;
            for cindex := 1 to write_chs do
            begin
               macro_code_to_visible(macro_text[cindex], visible_rep);
               { replace single { by {{ to avoid confusion when reading }
               if visible_rep = '{'
               then visible_rep := '{{';
               write(macfile, visible_rep)
            end { for loop } ;
            ptr := next
         end { while loop, with } ;
         writeln(macfile,'"')
      end { with }
   end { write_macro_text } ;

   procedure onunit(not_used: integer);
   begin { onunit }
      main_state := command_error;
      goto 122
   end { onunit } ;

begin { macsave_comm }
   on('ERROR_TRAP', onunit);
   if open_macro_file(macfile, true) then
   begin { file opened }
      for id_char := first_print_char to last_print_char do
      with macro_pointers[id_char] do
      begin
         if pointers[posm] <> nil
         then write_macro_text(' ', id_char, pointers[posm]);
         if pointers[negm] <> nil
         then write_macro_text('^', id_char, pointers[negm])
      end { for loop }
   end { file opened } ;
122:
end { macsave_comm } ;

{ MARK command }

procedure mark_comm;                                                            { MARK_COMM }

var
   markln: integer;

begin { mark_comm }
   with current_buffer do
   begin
      if marked_line <> 0
      then update_changed_lines(marked_line);
      if get_number(markln)
      then                              { number actually given }
         if markln < 0
         then bad_command
         else marked_line := markln
      else                              { no number given - mark current line }
         if line_status = top
         then comm_error('You cannot mark the .top. line.')
         else marked_line := line_number;
      if marked_line <> 0
      then update_changed_lines(marked_line)
   end
end { mark_comm } ;

{ Merge a buffer side-by-side with the current buffer }

procedure merge_comm;                                                           { MERGE_COMM }

var
   main_buffer_number: buffer_range;
   merge_column, merge_column_2, merge_width,
   merge_buffer_line, last_column: integer;
   option_word: word;
   lindex: line_chars_range;
   copy_line: line_chars_type;
   copy_length: line_chars0_range;
   insert_flag, done: boolean;
   saved_file_place: file_places;

begin { merge_comm }
   with current_buffer, load_file do
   begin
      save_place(saved_file_place);
      file_type := sam_type;
      copy_treename(treename, true, false, false, true);
      if not get_number(merge_column)
      then comm_error('Column number missing.')
      else if not get_number(merge_column_2)
      then comm_error('Column number missing.')
      else if (merge_column < 1) or (merge_column > line_max_charsm1)
      then comm_error('Column number out of range.')
      else if (merge_column_2 < 1) or (merge_column_2 > line_max_charsm1) or
              (merge_column_2 < merge_column)
      then comm_error('Column number out of range.')
      else
      begin
         insert_flag := false;          { analyse any options present }
         get_word(option_word);
         if option_word = '        '
         then
         else if compare(option_word, 'INSERT  ')
         then insert_flag := true
         else bad_command;
         original_pathname := false;
         rename_done := false;
         open_file(k$read, load_file, 0, false, true);
         if (main_state = running) and file_is_open then
         begin
            merge_width := merge_column_2 - merge_column + 1;
            if not line_exists
            then load_a_line(line_number, false);
            main_buffer_number := current_buffer_number;
            merge_buffer_line := 1;
            done := false;
            repeat
               switch_buffer(load_buffer_number);
               load_a_line(merge_buffer_line, false);
               if line_status = bottom
               then done := true
               else
               begin { not end of merge buffer }
                  merge_buffer_line := merge_buffer_line + 1;
                  copy_line := line_text;
                  if line_length <= merge_width
                  then copy_length := line_length
                  else copy_length := merge_width;
                  switch_buffer(main_buffer_number);
                  if line_status = bottom then
                  begin { append blank line to main buffer }
                     line_text[1] := newline;
                     line_length := 0;
                     insert_a_line
                  end { append blank line to main buffer } ;
                  last_column := merge_column + copy_length - 1;
                  if insert_flag then
                  begin { shift cols right for INSERT flag }
                     if (line_length + merge_width) > line_max_charsm1
                     then comm_error('Merged line too long.')
                     else if line_length < merge_column - 1 then
                     begin { need to expand }
                        for lindex := line_length+1 to last_column
                        do line_text[lindex] := ' ';
                        line_length := last_column;
                        line_text[line_length+1] := newline
                     end { need to expand }
                     else
                     begin
                        movewds(line_text, merge_column,
                                line_text, merge_column+merge_width,
                                line_length+2-merge_column);
                        line_length := line_length + merge_width
                     end
                  end { shift cols right for INSERT flag } ;
                  if main_state = running then
                  begin
                     if last_column > line_length then
                     begin { expand to make space }
                        for lindex := line_length+1 to merge_column-1
                        do line_text[lindex] := ' ';
                        line_length := last_column;
                        line_text[line_length+1] := newline
                     end { expand to make space } ;
                     movewds(copy_line, 1,
                             line_text, merge_column,
                             copy_length);
                     for lindex := copy_length+1 to merge_width
                     do line_text[lindex+merge_column-1] := ' ';
                     line_altered := true;
                     update_changed_lines(line_number);
                     load_a_line(line_number+1, false)
                  end { new line length in range }
               end { not end of merge buffer }
            until done or (main_state <> running);
            switch_buffer(load_buffer_number);
            load_a_line(0, false);
            file_altered := false;      { clear this now text loaded elsewhere }
            switch_buffer(main_buffer_number);
            must_redraw_window := true
         end { opened OK }
         else main_state := command_error
      end { no obvious syntax errors } ;
      if main_state = top_or_bottom
      then main_state := running;       { OK to hit top or bottom }
      if main_state = running
      then restore_place(saved_file_place)  { back to start place if OK }
   end { with }
end { merge_comm } ;

{ MODE command }

procedure mode_comm;                                                            { MODE_COMM }

var
   mode_name: word;
   windex: word_range;
   value_flag: boolean;
   remove: 1..2;
   mode_index: modes;

   procedure bad_mode(unrec: boolean);

   begin { bad_mode }
      if unrec
      then message := 'Command MODE ~ not recognised.'
      else message := 'Error in format of MODE ~ command.';
      ins_word(mode_name, true);
      screen_message(message, true);
      main_state := command_error
   end { bad_mode } ;

   function find_mode: boolean;         { search list of mode names }

   var
      state: search_type;

   begin { find_mode }
      if mode_name[1] = ' '
      then find_mode := false
      else
      begin { not blank }
         mode_index := modes.min;
         state := searching;
         repeat
            if compare(mode_name, mode_names[mode_index])
            then state := found
            else if mode_index = modes.max
            then state := not_found
            else mode_index := succ(mode_index)
         until state <> searching;
         find_mode := state = found
      end { not blank }
   end { find_mode } ;

   procedure mode_count_proc;           { MODE COUNT }

   var
      nindex: 1..4;
      ch: char;
      valu: integer;
      zeros_word: word;
      finished, result: boolean;

   begin { mode_count_proc }
      mode_count := value_flag;
      set_up_sets;
      if value_flag then
      begin { mode now set }
         nindex := 1;                   { indicates which numeric expected }
         finished := false;
         repeat
            skip_spaces;
            ch := command_line[sub];
            if ch in string_terminators
            then  finished := true
            else  if ch in ['-', '0'..'9'] then
            begin { numeric argument }
               result := get_number(valu);
               case nindex of
                  1: count_value := valu;
                  2: count_increment := valu;
                  3: if (valu >= 1) and (valu <= count_width_max)
                     then  count_width := valu
                     else  bad_mode(false)
                  otherwise bad_mode(false)
               end;
               if nindex < 4
               then nindex := nindex + 1
            end { numeric argument }
            else
            begin { non-numeric argument }
               get_word(zeros_word);
               if zeros_word = blank_word
               then bad_mode(false)
               else if compare(zeros_word, 'BLANK   ')
               then count_type := zblank
               else if compare(zeros_word, 'PRINT   ')
               then count_type := zprint
               else if compare(zeros_word, 'SUPPRESS')
               then count_type := zsuppress
               else bad_mode(false)
            end { non-numeric argument }
         until finished or (main_state <> running)
      end { mode now set }
   end { mode_count_proc } ;

   procedure mode_margin_proc;          { MODE MARGIN }

   var
      valu: integer;

   begin { mode_margin_proc }
      mode_margin := value_flag;
      if value_flag then
      begin
         if get_number(valu) then
         begin { new value for margin_column }
            if (valu < 1) or (valu >= line_max_chars) then
            begin { margin value out of range }
               int_screen_message(
                  'Margin column must be in the range 1..~', true,
                  line_max_charsm1, 1);
               mode_margin := false;
               main_state := command_error
            end { margin value out of range }
            else margin_column := valu
         end { number present }
      end { switch on }
   end { mode_margin_proc } ;

   procedure mode_backup_proc;          { MODE BACKUP }

   var
      valu: integer;

   begin { mode_backup_proc }
      mode_backup := value_flag;
      if value_flag then
      begin { switch on }
         if get_number(valu) then
         begin { new value for max_backup }
            if (valu < 1) or (valu > 9) then
            begin { backup value out of range }
               screen_message(
                  'Backup column must be in the range 1..9', true);
               mode_backup := false;
               main_state := command_error
            end { backup value out of range }
            else max_backup := valu
         end { number present }
      end { switch on }
   end { mode_backup_proc } ;

   procedure mode_chain_proc;          { MODE CHAIN }

   var
      buffer_no, buffer_count: integer;
      last_buffer_no: buffer0_range;
      temp_str: string32;

      procedure clear_buffer_chain;

      var
         bindex: buffer_range;

      begin { clear_buffer_chain }
         for bindex := 1 to buffer_max do
         with buffer_pointers[bindex] do
         begin
            pred_buff := 0;
            succ_buff := 0
         end { for loop } ;
         first_buffer_in_chain := 0;
         last_buffer_in_chain := 0
      end { clear_buffer_chain } ;

   begin { mode_chain_proc }
      mode_chain := value_flag;
      skip_spaces;
      if value_flag then
      begin { switch on }
         if not (command_line[sub] in string_terminators) then
         begin { number(s) present }
            clear_buffer_chain;
            first_chain_setting := 0;
            last_buffer_no := 0;
            buffer_count := 0;
            buffer_no := 0;             { preset for get_next_number }
            while get_next_number(buffer_no) and (main_state = running) do
            begin { while loop }
               if (buffer_no >= 1) and (buffer_no <= buffer_max) then
               with buffer_pointers[buffer_no] do
               begin { number in range }
                  buffer_count := buffer_count + 1;
                  if first_buffer_in_chain = 0 then
                  begin { this is the first }
                     first_buffer_in_chain := buffer_no;
                     pred_buff := 0;
                     succ_buff := 1     { just to show in use }
                  end { this is the first }
                  else
                  begin { this is not the first }
                     if (pred_buff = 0) and (succ_buff = 0) then
                     begin { buffer not already chained }
                        pred_buff := last_buffer_no;
                        buffer_pointers[last_buffer_no].succ_buff := buffer_no
                     end { buffer not already chained }
                     else comm_error('Buffer chain loops.')
                  end { this is not the first } ;
                  last_buffer_no := buffer_no
               end { number in range }
               else comm_error('Buffer number out of range.')
            end { while loop } ;
            if (main_state = running) and (buffer_count < 2)
            then comm_error('Only one buffer in chain.');
            if main_state = running then
            begin { chain alright }
               last_buffer_in_chain := last_buffer_no;
               buffer_pointers[last_buffer_no].succ_buff := 0
            end { chain alright }
            else
            begin { error }
               mode_chain := false;
               clear_buffer_chain
            end { error }
         end { number(s) present }
         else if first_chain_setting > 1 then
         begin { chain multiple buffers initially loaded }
            int_screen_message('Chaining buffers 1-~', false,
                               first_chain_setting, 1);
            str(first_chain_setting, temp_str);
            first_chain_setting := 0;
            temp_str := concat('MODE CHAIN 1-',temp_str);
            obey_command_string(temp_str)
         end { chain multiple buffers initially loaded }
         else if first_buffer_in_chain = 0 then
         begin { no buffers chained }
            comm_error('No buffers are chained.');
            mode_chain := false
         end { no buffers chained }
      end { switch on }
   end { mode_chain_proc } ;

   { Print out all current mode values }

   procedure print_modes;

   var
      tindex: tabs_range;
      ch: char;
      count: integer;

      procedure pmode(valu: boolean; word: word_string);
      begin { pmode }
         if valu
         then append_message('  ')
         else append_message('no');
         append_message(word)
      end { pmode } ;

      procedure print_mode_chain;

      var
         buffer_no, last_buffer_no: integer;
         printing_range: boolean;

      begin { print_mode_chain }
         pmode(mode_chain , 'chain   ');
         if first_buffer_in_chain > 0 then
         begin { print the chain }
            buffer_no := first_buffer_in_chain;
            printing_range := false;
            last_buffer_no := -1;       { avoid accidental range }
            repeat
               if printing_range then
               begin { accumulating range of buffers }
                  if buffer_no = last_buffer_no + 1
                  then { still part of range }
                  else
                  begin { found end of range }
                     append_message('-~ ~');
                     ins_int(last_buffer_no, 1);
                     ins_int(buffer_no, 1);
                     printing_range := false
                  end { found end of range }
               end { accumulating range of buffers }
               else
               begin { individual number }
                  if buffer_no = last_buffer_no + 1
                  then printing_range := true
                  else
                  begin { print single number }
                     append_message(' ~');
                     ins_int(buffer_no, 1)
                  end { print single number }
               end { individual number } ;
               last_buffer_no := buffer_no;
               buffer_no := buffer_pointers[buffer_no].succ_buff
            until buffer_no = 0;
            if printing_range then
            begin { finish last range }
               append_message('-~');
               ins_int(last_buffer_no, 1)
            end { finish last range }
         end { print the chain }
      end { print_mode_chain } ;

   begin { print_modes }
      screen_message('Current mode settings are:', false);
      if mode_verify
      then screen_message('verify', false)
      else screen_message('brief', false);
      message := 'mode ';
      pmode(mode_ckpar , 'ckpar   ');
      pmode(mode_column, 'column  ');
      pmode(mode_info  , 'info    ');
      pmode(mode_number, 'number  ');
      pmode(mode_prompt, 'prompt  ');
      pmode(mode_semi  , 'semi    ');
      pmode(mode_indent, 'indent  ');
      screen_message(message, false);
      message := 'mode ';
      pmode(mode_quiet , 'quiet   ');
      pmode(mode_direct, 'direct  ');
      pmode(mode_safety, 'safety  ');
      pmode(mode_como  , 'como    ');
      pmode(mode_insert, 'insert  ');
      pmode(mode_disp  , 'disp    ');
      screen_message(message, false);
      message := 'mode ';
      pmode(mode_msg   , 'msg     ');
      pmode(mode_lon   , 'lon     ');
      {---teststart---}
      pmode(mode_test  , 'test    ');
      {---teststop---}
      screen_message(message, false);
      message := 'mode ';
      pmode(mode_count , 'count   ');
      append_message('~ ~ ~ ~      ');
      ins_int(count_value, 1);
      ins_int(count_increment, 1);
      ins_int(count_width, 1);
      case count_type of
         zblank:    ins_word('blank   ', true);
         zprint:    ins_word('print   ', true);
         zsuppress: ins_word('suppress', true)
      end { case } ;
      pmode(mode_margin, 'margin  ');
      append_message('~  ');
      ins_int(margin_column, 1);
      pmode(mode_backup, 'backup  ');
      append_message('~');
      ins_int(max_backup, 1);
      screen_message(message, false);
      message := 'mode ';
      print_mode_chain;
      screen_message(message, false);
      if number_of_tabs = 0
      then screen_message('No tabs set.', false)
      else
      begin
         message := 'Tabs are';
         for tindex := 1 to number_of_tabs do
         begin
            append_message(' ~');
            ins_int(htabs[tindex], 1)
         end;
         screen_message(message, false)
      end;
      message := 'STYLE ~ ~ ~ ~ ~   LINESZ ~';
      with style_rec do
      begin
         ins_int(left_margin , 1);
         ins_int(right_margin, 1);
         ins_int(para_indent , 1);
         ins_int(para_blanks , 1);
         case adjust of
            false: ins_word('fill    ', true);
            true:  ins_word('adjust  ', true)
         end
      end;
      ins_int(test_line_size, 1);
      screen_message(message, false);
      message := 'Punctuation characters are space and ';
      for ch := '!' to '~'
      do if ch in punctuation then
         begin
            append_char(ch)
         end;
      screen_message(message, false);
      message := 'Alphabetic characters (~) are: ';
      count := 0;
      for ch := '@' to '_'
      do if ch in upper_alphabet then
         begin
            append_char(ch);
            count := count + 1
         end;
      ins_int(count, 1);
      screen_message(message, false);
      with current_buffer do
      begin
         screen_message('Current buffer information:', false);
         get_where_message;
         message := concat('   Position is ',message);
         screen_message(message, false);
         if marked_line = 0
         then message := '   No marked line.  '
         else
         begin
            message := '   Line ~ marked.  ';
            ins_int(marked_line, 1)
         end;
         if point_line_pos = 0
         then append_message('   No character pointer.')
         else
         begin
            append_message('   Character pointer is at ~.~');
            ins_int(point_line_no, 1);
            ins_int(point_line_pos, 1)
         end;
         screen_message(message, false);
         if in_window_code then
         begin
            if iw_marked_line = 0
            then message := '   WI mode'
            else
            begin
               message := '   IW mode (inserting before line ~)';
               ins_int(iw_marked_line, 1)
            end;
            screen_message(message, false)
         end
      end
   end { print_modes } ;

begin { mode_comm }
   skip_spaces;
   if command_line[sub] in string_terminators
   then print_modes                     { no argument - just print modes }
   else
   begin { argument present }
      get_word(mode_name);
      if find_mode
      then  value_flag := true          { switching on a mode }
      else  if mode_name[1] = 'N' then
      begin { switching off a mode }
         if mode_name[2] = 'O'          { allow for "N" and "NO" }
         then  remove := 2              { remove first part of mode_name }
         else  remove := 1;
         for windex := 1 to (word_max_chars - remove)
         do mode_name[windex] := mode_name[windex + remove];
         for windex := (word_max_chars - remove + 1) to word_max_chars
         do mode_name[windex] := ' ';
         if find_mode
         then value_flag := false
         else bad_mode(true)
      end { switching off a mode }
      else bad_mode(true);
      if main_state = running then
      case mode_index of
         mckpar:  mode_ckpar  := value_flag;
         mcolumn: mode_column := value_flag;
         mcount:  mode_count_proc;
         minfo:   mode_info   := value_flag;
         mnumber: mode_number := value_flag;
         mprompt: mode_prompt := value_flag;
         msemi:
            begin
               mode_semi := value_flag;
               set_up_sets
            end;
         mindent: mode_indent := value_flag;
         mquiet:  mode_quiet  := value_flag;
         mmargin: mode_margin_proc;
         mdirect: mode_direct := value_flag;
         msafety: mode_safety := value_flag;
         mcomo:   mode_como   := value_flag;
         minsert: mode_insert := value_flag;
         mdisp:   mode_disp   := value_flag;
         mchain:  mode_chain_proc;
         mbackup: mode_backup_proc;
         mmsg:    mode_msg    := value_flag;
         mlon:    mode_lon    := value_flag;
         mtest:   mode_test   := value_flag
      end { case }
   end { argument present }
end { mode_comm } ;

{ MOVE or OBEY command }

procedure move_or_obey_comm(obey_command: boolean);                             { MOVE_OR_OBEY_COMM }

var
   bindex1, bindex2: string_buffers2_range;
   ch: char;
   explicit, must_create, new_exists: boolean;
   strbuf_ptr: ptr_edline;

   procedure do_the_move(var target: line_chars_type;
                         var length: line_chars0_range;
                         var exists: boolean);

   var
      ptr: ptr_edline;

      procedure store(ch: char);
      begin { store }
         length := length + 1;
         target[length] := ch
      end { store } ;

   begin { do_the_move }
      with current_buffer do
      if explicit then
      begin { explicit text string given }
         length := 0;
         if copy_string(store, false)   { copy without processing escapes }
         then  sub := sub + 1;          { skip delimitor }
         target[length + 1] := newline;
         exists := true
      end { explicit text string given }
      else
      begin { not explicit string }
         if bindex2 < 0 then
         begin { copying from INLIN }
            target := line_text;
            length := line_length;
            exists := line_exists
         end { copying from INLIN }
         else
         begin { not INLIN }
            ptr := string_buffers[bindex2];
            if ptr = nil
            then exists := false        { old buffer does not exist }
            else with ptr^ do
            begin { copy buffer }
               target := ed_text;
               length := text_length;
               exists := true
            end { copy buffer }
         end { not INLIN }
      end { not explicit string }
   end { do_the_move } ;

begin { move_or_obey_comm }
   if obey_command
   then bindex1 := 0                    { not used }
   else get_buffer_index(bindex1, false);  { get first arg }
   if main_state = running then
   begin { buffer name alright }
      skip_spaces;
      ch := command_line[sub];
      if ch in string_terminators
      then bad_command
      else if ch in ['A'..'Z', 'a'..'z'] then
      begin { second buffer name given }
         get_buffer_index(bindex2, false);
         explicit := false
      end { second buffer name given }
      else explicit := true             { enclosed string given }
   end { buffer name alright } ;
   if main_state = running then
   with current_buffer do
   begin { args alright }
      if bindex1 < 0 then
      begin { copying to INLIN }
         must_redraw_window := true;
         must_create := not line_exists;
         if must_create
         then replace_line;
         do_the_move(line_text, line_length, new_exists);
         if new_exists then
         begin { not null line }
            line_altered := true;
            if must_create
            then  insert_a_line
         end { not null line }
         else delete_line
      end { copying to INLIN }
      else
      begin { copying to string buffer }
         if obey_command
         then strbuf_ptr := nil         { just use it as temp storage }
         else strbuf_ptr := string_buffers[bindex1];
         if strbuf_ptr = nil then
         begin { null buffer }
            new(strbuf_ptr);
            strbuf_ptr^.being_obeyed := false
         end { null buffer } ;
         with strbuf_ptr^ do
         do_the_move(ed_text, text_length, new_exists);
         if not new_exists then
         begin { buffer no longer exists }
            dispose(strbuf_ptr);
            strbuf_ptr := nil
         end { buffer no longer exists } ;
         if obey_command then
         begin { OBEY command }
            if strbuf_ptr = nil
            then { no-operation }
            else
            begin { command not null }
               substitute_variables(strbuf_ptr^);
               if main_state = running then
               with strbuf_ptr^ do
               begin { substitution alright }
                  { no need to check being_obeyed as line is temporary }
                  obey_command_line(ed_text, 1, text_length + 2)
               end { substitution alright } ;
               dispose(strbuf_ptr)
            end { command not null }
         end { OBEY command }
         else
         begin { MOVE command }
            string_buffers[bindex1] := strbuf_ptr;
            if bindex1 = 0              { if copying to EDLIN }
            then copy_suppress := true  { then do not copy on top of it later }
                                        { else the user has wasted his time }
         end { MOVE command }
      end { copying to string buffer }
   end { args alright }
end { move_or_obey_comm } ;

{ NEXT command }

procedure next_comm;                                                            { NEXT_COMM }

var
   increment: integer;
   result: boolean;

begin { next_comm }
   with current_buffer do
   begin
      result := get_number(increment);
      if increment = 0
      then increment := 1;              { default is NEXT 1 }
      if (increment > 0) and (not line_exists)
      then increment := increment - 1;  { allow for deleted line }
      load_a_line(line_number + increment, true);
      test_line_landed_on;              { top or bottom ? }
      verify_print;
      get_line_pos                      { process .charpos if present }
   end
end { next_comm } ;

{ NEXTBUFF and PREVBUFF commands }

procedure next_or_prev_buff(forwards: boolean);                                 { NEXT_OR_PREV_BUFF }

var
   new_buffer_number: buffer_range;

begin { next_or_prev_buff }
   find_next_used_buffer(forwards, new_buffer_number);
   if new_buffer_number = current_buffer_number
   then comm_error('No other buffers in use.')
   else
   begin { non-empty buffer found }
      switch_buffer(new_buffer_number);
      if in_window_code
      then display_buffer_name := true
      else announce_buffer_name
   end { non-empty buffer found }
end { next_or_prev_buff } ;

{ Obey a Primos command after expanding abbreviations if possible }

procedure obey_primos_proc;                                                     { OBEY_PRIMOS_PROC }

label 123;

var
   primos_command_line: tree_type;
   expanded_command: long_string;
   code, code1, code2: shortint;
   result: boolean;

   procedure quit_onunit(not_used: integer);
   begin { quit_onunit }
      main_state := quit_signal;
      screen_message('', false);
      if question_and_answer(
         'Do you want to terminate execution of the Primos command? ',
         true, true)
      then goto 123
   end { quit_onunit } ;

   procedure reenter_onunit(not_used: integer);
   begin { reenter_onunit }
      screen_message('', false);
      screen_message('You are now back in the editor.', true);
      screen_message('', false);
      goto 123
   end { reenter_onunit } ;

begin { obey_primos_proc }
   if command_line_args.no_primos_sw
   then comm_error('You are not allowed to use the ! command.')
   else
   begin { command is allowed }
      sub := sub + 1;
      skip_spaces;
      copy_treename(primos_command_line, false, true, true, false);
      expanded_command := primos_command_line.strng;
      com$ab(expanded_command, line_max_chars, code);
      on('QUIT$', quit_onunit);
      on('REENTER$', reenter_onunit);
      if in_window_code then
      begin
         do_clear_screen;
         if term_is_ssmp
         then stop_ssmp;
         lon$cn(1)                      { switch logout notification on }
      end;
      cp$(expanded_command, code1, code2);
      if in_window_code then
      begin
         if not mode_lon
         then lon$cn(0);                { switch logout notification off }
         write('Press RETURN to redraw the screen.');
         readln;
         must_redraw_screen := true;
         screen_tabs_set := false;      { in case Primos command did anything }
         if term_is_ssmp
         then result := init_ssmp
      end;
   123:
   end { command is allowed }
end { obey_primos_proc } ;

{ OOPS command }

procedure oops_comm;                                                            { OOPS_COMM }

var
   result: boolean;
   oops_count: integer;

begin { oops_comm }
   if in_window_code then
   begin { in window editor }
      result := get_number(oops_count);
      if oops_count <= 0
      then oops_count := 1;
      wi_oops_count := wi_oops_count + oops_count
   end { in window editor }
   else oops_proc
end { oops_comm } ;

{ PAUSE command }

procedure pause_comm;                                                           { PAUSE_COMM }

label 124;

   procedure reenter_onunit(cf_ptr: integer);
   begin { reenter_onunit }
      goto 124
   end { reenter_onunit } ;

begin { pause_comm }
   if command_line_args.no_primos_sw
   then comm_error('You are not allowed to use the PAUSE command.')
   else
   begin { command is allowed }
      if in_window_code then
      begin
         do_clear_screen;
         if term_is_ssmp
         then stop_ssmp
      end;
      writeln('Type START or REN to restart the Editor.');
      hook2(vdu_type);
      on('REENTER$', reenter_onunit);
      comlv$;
   124:
      hook1(vdu_type);
      get_erase_and_kill;
      writeln('You are now back in the Editor.');
      if in_window_code then
      begin
         must_redraw_screen := true;
         if term_is_ssmp
         then result := init_ssmp
      end
   end { command is allowed }
end { pause_comm } ;

{ POINT command }

procedure point_comm;                                                           { POINT_COMM }

var
   new_line_number: integer;

begin { point_comm }
   skip_spaces;
   with current_buffer do
   if command_line[sub] in ['m','M'] then
   begin
      sub := sub + 1;
      if marked_line = 0
      then comm_error('No line is marked.')
      else load_a_line(marked_line, false)
   end
   else if get_number(new_line_number) then
   begin
      load_a_line(new_line_number, true);
      test_line_landed_on;
      verify_print
   end;
   get_line_pos                         { process .charpos if present }
end { point_comm } ;

{ PP command }

procedure pprint_comm;                                                          { PPRINT_COMM }

var
   arg1, arg2, start_at, lines_to_print: integer;
   saved_file_place: file_places;
   result: boolean;

begin { pprint_comm }
   with current_buffer do
   begin
      save_place(saved_file_place);
      result := get_number(arg1);
      result := get_number(arg2);
      if arg2 = 0
      then  if arg1 = 0 then
            begin
               arg1 := -5;              { default settings }
               arg2 := 5
            end
            else  if arg1 > 0 then
                  begin
                     arg2 := arg1;
                     arg1 := 0
                  end;
      start_at := line_number + arg1;
      if start_at < 0                   { off top }
      then start_at := 0;
      lines_to_print := line_number + arg2 - start_at + 1;
      load_a_line(start_at, false);
      loop_by_number([print_act], lines_to_print);
      if main_state = top_or_bottom     { not an error to hit end of file }
      then main_state := running;
      restore_place(saved_file_place)
   end
end { pprint_comm } ;

{ PRINT command }

procedure print_comm;                                                           { PRINT_COMM }

var
   lines_to_print, start_at: integer;
   result: boolean;
   bindex: string_buffers2_range;

   procedure print_buffer(the_index: string_buffers0_range);

   var
      strbuf_ptr: ptr_edline;
      lindex: line_chars_range;

   begin { print_buffer }
      strbuf_ptr := string_buffers[bindex];
      if strbuf_ptr = nil
      then write('.null.')
      else with strbuf_ptr^ do
      begin
         write('"');
         for lindex := 1 to text_length
         do wrt_ch_mode(ed_text[lindex]);
         write('"')
      end;
      writeln
   end { print_buffer } ;

begin { print_comm }
   if in_window_code then
   begin
      screen_message('PRINT is not valid in the window editor.', true);
      main_state := command_error
   end
   else with current_buffer do
   begin
      skip_spaces;
      if command_line[sub] in ['a'..'z','A'..'Z'] then
      begin { PRINT STRA etc. }
         get_buffer_index(bindex, true);
         if main_state = running then
         begin
            if bindex = -2 then
            begin { print all }
               for bindex := 0 to num_string_buffers do
               begin
                  if string_buffers[bindex] <> nil then
                  begin { not null }
                     if bindex = 0
                     then write('EDLIN ')
                     else write('STR',chr(ord('A') - 1 + bindex),'  ');
                     print_buffer(bindex)
                  end { not null }
               end { for loop }
            end { print all }
            else if bindex = -1
            then print_line             { INLIN }
            else print_buffer(bindex);  { print numbered buffer }
            copy_suppress := true       { do not obliterate EDLIN }
         end { buffer name alright }
      end { print stra etc. }
      else
      begin { PRINT (n) }
         result := get_number(lines_to_print);
         if lines_to_print < 0 then
         begin
            start_at := line_number + lines_to_print + 1;
            if start_at < 0             { off top }
            then start_at := 0;
            lines_to_print := line_number - start_at + 1;
            load_a_line(start_at, false)
         end;
         loop_by_number([print_act], lines_to_print)
      end { PRINT (n) }
   end
end { print_comm } ;

{ PSYMBOL command }

procedure psymbol_comm;                                                         { PSYMBOL_COMM }

var
   symbol: symbols;

begin { psymbol_comm }
   for symbol := symbols.min to symbols.max do
   begin
      message := '~ ~';
      ins_word(symbol_names[symbol], false);
      ins_ch(special[symbol]);
      screen_message(message, false)
   end
end { psymbol_comm } ;

{ PUNCT command }

procedure punct_comm;                                                           { PUNCT_COMM }

var
   ch: char;
   escaped, done: boolean;
   message: string[4];

begin { punct_comm }
   skip_spaces;
   punctuation := [' '];
   if command_line[sub] in string_terminators then
   begin { no args - interactive method }
      screen_message('For each possible punctuation character, indicate whether you want to include it in the set. RETURN means yes.',
         false);
      message := '~ : ';
      for ch := '!' to '~' do
      begin
         if ch in ['!','"','#','$','%','&','''','(',')','*','+',',','-','.','/',
            ':',';','<','=','>','?','[','\',']','^','_','`','{','|','}','~']
         then
         begin
            message[1] := ch;
            if question_and_answer(message, false, true)
            then punctuation := punctuation + [ch]
         end
      end
   end { no args - interactive method }
   else
   begin { args given - copy into set }
      escaped := false;
      done := false;
      repeat
         ch := command_line[sub];
         if escaped then
         begin
            escaped := false;
            punctuation := punctuation + [ch]
         end
         else if ch = special[sescape]
         then escaped := true
         else if ch in string_terminators
         then done := true
         else punctuation := punctuation + [ch];
         if not done
         then sub := sub + 1
      until done
   end { args given }
end { punct_comm } ;

{ REPLACE command - only valid in screen editor }

procedure replace_comm;                                                         { REPLACE_COMM }

var
   oplen: 0..maxint;

   procedure store1(ch: char);
   begin { store1 }
      with replace_ptr_1^ do
      begin
         len := len + 1;
         strng[len] := ch
      end { with }
   end { store1 } ;

   procedure store2(ch: char);
   begin { store2 }
      with replace_ptr_2^ do
      begin
         len := len + 1;
         strng[len] := ch
      end { with }
   end { store2 } ;

begin { replace_comm }
   if replace_ptr_1 <> nil then
   begin { dispose first string }
      dispose(replace_ptr_1);
      replace_ptr_1 := nil
   end { dispose first string } ;
   if replace_ptr_2 <> nil then
   begin { dispose second string }
      dispose(replace_ptr_2);
      replace_ptr_2 := nil
   end { dispose second string } ;
   if in_window_code then
   begin { called from WI }
      if command_line[sub] = ' '
      then sub := sub + 1;
      replace_options := '        ';
      if command_line[sub] in string_terminators
      then { no options given - prompt for them in window code }
      else
      begin { get strings etc. from command line }
         new(replace_ptr_1);
         new(replace_ptr_2);
         replace_ptr_1^.strng := '';
         if not copy_string(store1, true)
         then bad_command               { terminator absent }
         else if replace_ptr_1^.len = 0
         then bad_command               { first string null }
         else
         begin { first string alright }
            replace_ptr_2^.strng := '';
            if copy_string(store2, true)
            then sub := sub + 1;        { skip terminator }
            oplen := 0;
            while not (command_line[sub] in string_terminators) do
            begin
               oplen := oplen + 1;
               if oplen <= 8
               then replace_options[oplen] := command_line[sub]
               else bad_command;
               sub := sub + 1
            end { while loop }
         end { first string alright } ;
         if main_state <> running then
         begin { error in format }
            dispose(replace_ptr_1);
            replace_ptr_1 := nil;
            dispose(replace_ptr_2);
            replace_ptr_2 := nil
         end { error in format }
      end { get strings from command line } ;
      if main_state = running
      then replace_command_given := true
      { REPLACE is obeyed later in the screen editor code }
   end { called from WI }
   else comm_error('REPLACE is only allowed when screen editing.')
end { replace_comm } ;

{ FILE or SAVE command }

procedure file_or_save_proc(fors: boolean);                                     { FILE_OR_SAVE_PROC }

var
   new_treename: tree_type;

begin { file_or_save_proc }
   skip_spaces;
   if command_line[sub] in string_terminators
   then new_treename.len := 0
   else copy_treename(new_treename, true, false, false, true);
   if check_termination then
   begin { no extra rubbish after command }
      if save_proc(new_treename, new_treename.len = 0) then
      begin { save worked }
         if fors
         then quit_proc(false)          { FILE - give up now }
         else                           { SAVE - completed ok }
      end { save worked }
   end { no extra rubbish after command }
end { file_or_save_proc } ;

{ FILEALL or SAVEALL command }

procedure fileall_or_saveall_proc(fors: boolean);                               { FILEALL_OR_SAVEALL_PROC }

begin { fileall_or_saveall_proc }
   saveall_proc(false);                 { save buffer(s) under current names }
   if fors
   then if main_state = running         { FILEALL - quit if all OK }
        then quit_proc(false)
   else                                 { SAVEALL - completed OK }
end { fileall_or_saveall_proc } ;

{ TABSET and PTABSET commands }

procedure settabs(var ntabs: tabs0_range;                                       { SETTABS }
                  var tab_settings: tabs_array);

var
   valu: integer;
   previous_tab: 0..line_max_chars;
   finished: boolean;

begin { settabs }
   screen_tabs_set := false;            { ensure terminal tabs reset next time }
   ntabs := 0;
   previous_tab := 0;
   finished := false;
   repeat
      if not get_number(valu)
      then finished := true             { no more numbers }
      else if (valu <= previous_tab) or (valu > line_max_charsm1)
      then bad_command
      else if ntabs = tabs_max
      then bad_command                  { too mant tabs }
      else
      begin
         ntabs := ntabs + 1;
         tab_settings[ntabs] := valu;
         previous_tab := valu
      end
   until finished or (main_state <> running)
end { settabs } ;

{ SAVESTAT command }

procedure savestat_comm;                                                        { SAVESTAT_COMM }

label 3729;

var
   old_savestat_treename: tree_type;
   savestat_file: text;
   saved_buffer_number, bindex: buffer_range;
   unas_count: integer;

   procedure write_buffer_info(buffer_number: buffer_range);

   begin { write_buffer_info }
      with current_buffer, main_file do
      begin
         if treename.len > 0 then
         begin { name assigned }
            writeln(savestat_file,'buffer ',buffer_number:1);
            writeln(savestat_file,'ed ',main_file.treename.strng);
            case line_status of
               top:
                  writeln(savestat_file,'top');
               normal:
                  writeln(savestat_file,'point ',line_number:1,'.',
                                                 line_position:1);
               bottom:
                  writeln(savestat_file,'bottom')
            end { case }
         end { name assigned }
         else unas_count := unas_count + 1
      end { with }
   end { write_buffer_info } ;

   procedure error_trap_onunit(cf_ptr: integer);

   begin { error_trap_onunit }
      main_state := command_error;
      goto 3729
   end { error_trap_onunit } ;

begin { savestat_comm }
   old_savestat_treename := savestat_treename;
   skip_spaces;
   if command_line[sub] in string_terminators
   then { no filename given }
   else copy_treename(savestat_treename, true, false, false, true);
   if check_termination then
   begin { no rubbish after treename }
      if savestat_treename.len = 0
      then comm_error('Filename must be specified.')
      else
      begin { filename not null }
         screen_message('Saving any modified files.', false);
         saveall_proc(false);           { save all buffers }
         { plough on even if not all saved }
         screen_message('Writing the state file.', false);
         on('ERROR_TRAP', error_trap_onunit);
         rewrite(savestat_file, savestat_treename.strng);
         saved_buffer_number := current_buffer_number;
         unas_count := 0;
         for bindex := 1 to buffer_max do
         begin
            if bindex = current_buffer_number  { first or only buffer }
            then write_buffer_info(bindex)
            else if buffer_pointers[bindex].bptr <> nil then
            begin { used buffer }
               switch_buffer(bindex);
               write_buffer_info(bindex)
            end { used buffer }
         end { for loop } ;
         switch_buffer(saved_buffer_number);
         writeln(savestat_file,'buffer ',current_buffer_number:1);
         str_screen_message('Current editing state saved to file ~', false,
                            savestat_treename.strng);
         if unas_count > 0
         then int_screen_message('Warning: ~ buffers have no filename assigned so their states not saved.',
                                 true, unas_count, 1)
      end { filename not null }
   end { no rubbish after treename } ;
3729:
   if main_state = command_error
   then savestat_treename := old_savestat_treename
end { savestat_comm } ;

{ SIZE command }

procedure size_comm;                                                            { SIZE_COMM }

begin { size_comm }
   list_buffers(true)
end { size_comm } ;

{ SPELL command - variants for line and screen editor working }

procedure spell_comm;                                                           { SPELL_COMM }

var
   lineno: integer;

   procedure show_line;

   begin { show_line }
      verify_print
   end { show_line } ;

   procedure ask_user(msg: string; var answer_char: char);

   var
      reply_text: long_string;
      cr_given: boolean;

   begin { ask_user }
      message_and_reply(msg, true, reply_text, cr_given);
      upcase_string(reply_text);
      if not cr_given
      then answer_char := 'Q'           { CTRL-X }
      else if length(reply_text) = 0
      then answer_char := 'A'
      else answer_char := reply_text[1]
   end { ask_user } ;

   procedure show_changed_line;

   begin { show_changed_line }
      verify_print
   end { show_changed_line } ;

begin { spell_comm }
   if check_termination then
   with current_buffer do
   begin { options alright }
      if free_dynamic_segments <= minimum_free_segments then
      begin { not enough memory }
         screen_message('There is not enough memory available to load the dictionary.', true);
         screen_message('Please free some dynamic segments and try again.', true);
         main_state := command_error
      end { not enough memory }
      else
      begin { there is enough memory }
         if marked_line = 0 then        { default is whole file }
         begin
            load_a_line(1, false);
            spell_stop_line := large_int
         end
         else
         begin { line marked - count mark to current }
            spell_stop_line := marked_line;
            update_changed_lines(marked_line);
            marked_line := 0;
            if spell_stop_line < line_number then
            begin { swap so that line_number <= spell_stop_line}
               lineno := line_number;
               load_a_line(spell_stop_line, false);
               spell_stop_line := lineno
            end { swap so that line_number <= spell_stop_line}
         end { line marked - count mark to current } ;
         if not line_exists
         then load_a_line(line_number, false);
         if in_window_code
         then spell_command_given := true
         { SPELL will be obeyed later in the screen editor code }
         else spell_proc(show_line, ask_user, show_changed_line)
      end { there is enough memory }
   end { options alright }
end { spell_comm } ;

{ SPLIT command }

procedure split_comm;                                                           { SPLIT_COMM }

var
   finished, current_buffer_found: boolean;
   vertical_split: boolean;
   valu, temp_count, way_param: integer;
   windex: window_range;
   buffer_array: array[window_range] of integer;
   ch: char;
   vertword: word;

   procedure print_split;

   var
      windex: window_range;

   begin { print_split }
      if window_count > 0 then
      begin
         int_screen_message('Screen is split into ~ windows.',
                            false, window_count, 1);
         screen_message('    buffer      left     right       top    bottom     width    length', false);
         for windex := 1 to window_count do
         with window_array[windex] do
         begin
            message := '~~~~~~~';
            ins_int(window_buffer, 10);
            ins_int(window_left  , 10);
            ins_int(window_right , 10);
            ins_int(window_top   , 10);
            ins_int(window_bottom, 10);
            ins_int(window_width , 10);
            ins_int(window_length, 10);
            screen_message(message, false)
         end
      end { screen is split }
      else screen_message('The screen is not split.', false)
   end { print_split } ;

begin { split_comm }
   skip_spaces;
   if command_line[sub] in string_terminators
   then print_split { no argument - just print split details }
   else
   begin
      { copy list of buffers into temp area - be sure not to alter permanent }
      { information until we know the command is alright }
      temp_count := 0;
      finished := false;
      repeat
         if not get_number(valu)
         then finished := true          { no more numbers }
         else if (valu < 1) or (valu > buffer_max)
         then comm_error('Buffer number out of range.')
         else if temp_count = window_max
         then comm_error('Too many windows specified.')
         else
         begin
            for windex := 1 to temp_count
            do if valu = buffer_array[windex]
               then comm_error('Buffer may not appear in two windows.');
            temp_count := temp_count + 1;
            buffer_array[temp_count] := valu
         end
      until finished or (main_state <> running);
      if main_state = running then
      begin
         skip_spaces;
         ch := command_line[sub];
         if ch in string_terminators
         then vertical_split := false   { default horizontal }
         else
         begin
            get_word(vertword);
            if compare(vertword, 'VERTICAL')
            then vertical_split := true
            else if compare(vertword, 'HORIZONT')
            then vertical_split := false
            else if compare(vertword, 'NO      ')
            then vertical_split := false
            else bad_command
         end { word present }
      end { buffer numbers alright } ;
      if main_state = running then
      begin
         skip_spaces;
         ch := command_line[sub];
         if ch in string_terminators
         then way_param := temp_count
         else
         begin { get split ways parameter }
            if get_number(way_param) then
            begin
               if (way_param < 1) or (way_param > temp_count)
               then comm_error('Split ways parameter out of range')
            end
            else bad_command
         end { get split ways parameter }
      end { alright so far } ;
      if main_state = running then
      begin { allocate windows }
         window_count := temp_count;
         if window_count = 0 then
         begin
            windows_across := 1;
            windows_down := 1
         end { window_count := 0 }
         else if vertical_split then
         begin
            windows_across := way_param;
            windows_down   := (window_count + way_param - 1) div way_param
         end { vertical }
         else
         begin
            windows_down   := way_param;
            windows_across := (window_count + way_param - 1) div way_param
         end { horizontal } ;
         { If user has given <= 1 buffers then screen is not split }
         if window_count = 0
         then window_array[0].window_buffer := current_buffer_number
         else if window_count = 1 then
         begin
            window_count := 0;
            window_array[0].window_buffer := buffer_array[1];
            switch_buffer(buffer_array[1])
         end { one buffer given }
         else
         begin { several buffers given }
            current_buffer_found := false;
            for windex := 1 to window_count do
            with window_array[windex] do
            begin
               window_buffer := buffer_array[windex];
               if window_buffer = current_buffer_number
               then current_buffer_found := true
            end { for loop } ;
            { switch to first buffer of split unless current buffer in split }
            if not current_buffer_found
            then switch_buffer(window_array[1].window_buffer)
         end { several buffers given } ;
         { ensure we know how big the screen is }
         if vdu_table = nil
         then read_vdu_table(false)     { calls calculate_window_sizes }
         else calculate_window_sizes;
         split_command_used := true
      end { allocate windows }
   end { args given }
end { split_comm } ;

{ STATS command }

procedure stats_comm;                                                           { STATS_COMM }

var
   saved_file_place: file_places;
   lineno, stop_line, comm_count, blank_count, text_count, word_count: integer;
   stats_col: line_chars_range;
   done, line_done, in_word: boolean;

begin { stats_comm }
   with current_buffer do
   begin
      save_place(saved_file_place);
      if marked_line = 0 then           { default is whole file }
      begin
         load_a_line(1, false);
         stop_line := large_int
      end
      else
      begin { line marked - count mark to current }
         stop_line := marked_line;
         update_changed_lines(marked_line);
         marked_line := 0;
         if stop_line < line_number then
         begin { swap so that line_number <= stop_line}
            lineno := line_number;
            load_a_line(stop_line, false);
            stop_line := lineno
         end { swap so that line_number <= stop_line}
      end { line marked - count mark to current } ;
      if line_status = top
      then load_a_line(1, false);       { ensure not on .top. }
      comm_count := 0;
      blank_count := 0;
      text_count := 0;
      word_count := 0;
      done := false;
      repeat
         if (line_number > stop_line) or  { all text counted or }
            (main_state <> running) or  { quit or }
            (line_status = bottom)      { hit bottom }
         then done := true
         else
         begin
            if line_length > 0 then
            begin
               if line_text[1] = special[swpc]
               then comm_count := comm_count + 1
               else
               begin
                  text_count := text_count + 1;
                  { count words in line }
                  stats_col := 1;
                  in_word := false;
                  line_done := false;
                  repeat
                     if stats_col > line_length
                     then line_done := true
                     else
                     begin
                        if line_text[stats_col] in punctuation
                        then in_word := false  { outside word }
                        else if not in_word then
                        begin { start of word }
                           in_word := true;
                           word_count := word_count + 1
                        end { start of word } ;
                        stats_col := stats_col + 1
                     end { stats_col within range }
                  until line_done;
               end { text line }
            end { non-blank line }
            else blank_count := blank_count + 1;
            load_a_line(line_number + 1, false)
         end
      until done;
      restore_place(saved_file_place);
      if main_state = quit_signal
      then screen_message('(Counts so far)', false);
      int_screen_message('~ lines made up of:', false,
                         comm_count+blank_count+text_count, 6);
      int_screen_message('~ text lines', false,
                         text_count, 12);
      int_screen_message('~ blank lines', false,
                         blank_count, 12);
      message := '~ lines beginning with ''~''';
      ins_int(comm_count, 12);
      ins_ch(special[swpc]);
      screen_message(message, false);
      int_screen_message('~ words in text lines', false,
                         word_count, 6)
   end { with }
end { stats_comm } ;

{ STYLE command }

procedure style_comm;                                                           { STYLE_COMM }

var
   saved_style_rec: style_type;
   arg_number, item: integer;
   ch: char;
   result, done: boolean;
   adjword: word;

begin { style_comm }
   saved_style_rec := style_rec;
   with style_rec do
   begin
      arg_number := 1;
      done := false;
      repeat
         skip_spaces;
         ch := command_line[sub];
         if ch in string_terminators
         then done := true
         else if ch in ['-', '0'..'9'] then
         begin { numeric }
            result := get_number(item);
            case arg_number of
               1: left_margin := item;
               2: right_margin := item;
               3: para_indent := item;
               4: para_blanks := item
               otherwise bad_command
            end;
            arg_number := arg_number + 1
         end { numeric }
         else
         begin { non-numeric }
            get_word(adjword);
            if compare(adjword, 'ADJUST  ')
            then adjust := true
            else if compare(adjword, 'FILL    ')
            then adjust := false
            else bad_command
         end { non-numeric }
      until done or (main_state <> running);
      if done then
      begin { check parameters given }
         if left_margin = 0
         then { auto processing - no checks }
         else if (left_margin < 0) or
                 (right_margin < left_margin) or
                 (right_margin > line_max_charsm1) or
                 ((left_margin + para_indent) <= 0) or
                 (para_blanks < 0)
         then bad_command
      end { check parameters given }
   end { with } ;
   { reinstate old values if any error }
   if main_state <> running
   then style_rec := saved_style_rec
end { style_comm } ;

{ Read and set a special symbol for SYMBOL, KILL, and ERASE commands }

procedure new_symbol(symbol: symbols);                                          { NEW_SYMBOL }

var
   new_value: char;

   procedure test_and_set(var special: char);

   var
      reduced: set_of_char;

   begin { test_and_set }
      reduced := symbol_chars - [special];
      if new_value in illegal_chars     { Is it illegal - e.g. comma? }
      then comm_error('New value is illegal.')
      else if new_value in reduced      { Is it already a special symbol? }
      then
      begin { duplicate }
         message := 'Symbol ~ already in use.';
         ins_ch(new_value);
         comm_error(message)
      end { duplicate }
      else
      begin { new value is alright }
         special := new_value;
         set_up_sets
      end { new value is alright }
   end { test_and_set } ;

begin { new_symbol }
   if command_line[sub] <> ' '
   then bad_command                     { no space after command }
   else
   begin { command format alright }
      sub := sub + 1;
      new_value := command_line[sub];
      if new_value = newline
      then bad_command                  { no new value given }
      else
      begin { new value present }
         sub := sub + 1;
         if symbol in [scprompt, sdprompt, svisible, swpc, sverbatim]
         then special[symbol] := new_value  { no checking for simple ones }
         else test_and_set(special[symbol])  { check for duplication }
      end { new value present }
   end { command format alright }
end { new_symbol } ;

{ SYMBOL command }

procedure symbol_comm;                                                          { SYMBOL_COMM }

var
   symbol_name: word;
   state: search_type;
   symbol: symbols;

begin { symbol_comm }
   get_word(symbol_name);
   if symbol_name[1] = ' '
   then state := not_found
   else
   begin { symbol name present }
      state := searching;
      symbol := symbols.min;
      repeat                            { look down list of symbol names }
         if compare(symbol_name, symbol_names[symbol])
         then state := found
         else if symbol = symbols.max
         then state := not_found
         else symbol := succ(symbol)
      until state <> searching
   end { symbol name present } ;
   if state = found
   then new_symbol(symbol)
   else bad_command
end { symbol_comm } ;

{ Display or change the terminal type }

procedure terminal_comm;                                                        { TERMINAL_COMM }

var
   type_read, saved_vdu_type: integer;
   old_treename: tree_type;

begin { terminal_comm }
   if get_number(type_read) then
   begin { set new terminal type }
      if type_read < 0 then
      begin
         writeln('Terminal number must be >= 0');
         main_state := command_error
      end
      else
      begin { new type >= 0 }
         saved_vdu_type := vdu_type;
         vdu_type := type_read;
         if vdu_type = 0 then
         begin
            if in_window_code then
            begin
               comm_error('Cannot switch to TTY when in window editor!');
               vdu_type := saved_vdu_type
            end
            else
            begin { switch to TTY }
               if vdu_table <> nil then
               begin
                  dispose(vdu_table);
                  vdu_table := nil
               end
            end { switch to TTY }
         end { vdu_type = 0 }
         else read_vdu_table(false);
         if main_state = running then
         begin { type is valid }
            hook5(vdu_type, type_read);
            get_erase_and_kill;
            set_term_gv;                { record it in globals }
            screen_tabs_set := false    { ensure tabs set up next time }
         end { type is valid }
         else
         begin { failed to read table }
            vdu_type := saved_vdu_type;  { restore original value }
            read_vdu_table(true);       { print possible types }
            main_state := command_error
         end { failed to read table }
      end { new type >= 0 }
   end { number present }
   else if command_line[sub] in string_terminators
   then read_vdu_table(true)            { print possible types }
   else if in_window_code
   then comm_error('Not allowed in window editor.')
   else
   begin { switch to different tables file }
      if vdu_type = 0                   { disable auto terminal recognition }
      then vdu_type := 1;
      old_treename := vdu_treename;     { preserve current tree name in case }
      copy_treename(vdu_treename, true, false, false, true);
      if check_termination then
      begin { no rubbish after treename }
         read_vdu_table(true);          { print possible types }
         if vdu_table <> nil then
         begin { tables not read now }
            dispose(vdu_table);
            vdu_table := nil
         end { tables not read now }
      end { no rubbish after treename } ;
      if main_state = command_error
      then vdu_treename := old_treename  { error - restore old treename }
   end { switch to different tables file }
end { terminal_comm } ;

{ TEXT command }

procedure text_comm;                                                            { TEXT_COMM }

var
   last_was_marked: boolean;
   old_left_margin, new_mark: integer;
   from_line, to_line: line_chars_type;

   function get_from_line(var from_line: line_chars_type;
                          var from_length: line_chars0_range): boolean;

   begin { get_from_line }
      with current_buffer do
      begin
         if last_was_marked or (main_state <> running)
         then get_from_line := false
         else
         begin { line available }
            if line_exists
            then load_a_line(line_number + 1, false)
            else load_a_line(line_number    , false);
            if line_status = normal then
            begin
               strip_trailing_spaces;
               from_line := line_text;
               from_length := line_length;
               get_from_line := true;
               last_was_marked := line_number = marked_line;
               delete_line
            end
            else get_from_line := false
         end { line available }
      end { with }
   end { get_from_line } ;

   procedure write_to_line(to_line: line_chars_type;
                           to_length: line_chars0_range);

   begin { write_to_line }
      with current_buffer do
      begin
         replace_line;
         line_text := to_line;
         line_length := to_length;
         insert_a_line;
         verify_print
      end { with }
   end { write_to_line } ;

begin { text_comm }
   result := get_number(old_left_margin);
   with current_buffer do
   if in_window_code and
      (point_line_no = line_number) and
      (point_line_pos <> 0)
   then
   begin { get from char pointer }
      old_left_margin := point_line_pos;
      point_line_no := 0;
      point_line_pos := 0
   end { get from char pointer } ;
   if main_state = running then
   with current_buffer do
   begin
      if marked_line = 0
      then marked_line := line_number
      else if marked_line < line_number then
      begin { swap so that mark is after current }
         new_mark := line_number;
         load_a_line(marked_line, false);
         marked_line := new_mark
      end { swap so that mark is after current } ;
      last_was_marked := false;
      load_a_line(line_number - 1, false);  { start on previous line to }
                                        { simulate usual state in loop }
      text_format(old_left_margin, get_from_line, write_to_line);
      if main_state = running
      then marked_line := 0;
      load_a_line(line_number + 1, true);  { move to line after formatted text }
      line_position := 1;
      must_redraw_window := true
   end { with }
end { text_comm } ;

{ UNLOAD and DUNLOAD commands }

procedure unload_proc(what_to_do: set_of_actions);                              { UNLOAD_PROC }

var
   ch: char;
   lines_to_unload, start_at: integer;
   result: boolean;
   start_column: integer;
   toword: word;

begin { unload_proc }
   start_column := 1;                   { default }
   if not (locate_act in what_to_do)
   then if command_line[sub] = '('      { get specified start column }
        then if not get_start_column(start_column)
             then bad_command;
   if main_state = running then
   with load_file do
   begin
      file_type := sam_type;
      copy_treename(treename, true, false, false, true);
      original_pathname := false;
      rename_done := false;
      skip_spaces;
      ch := command_line[sub];
      if (ch = 't') or (ch = 'T') then
      begin { to string form }
         get_word(toword);
         if toword = 'TO      ' then
         begin
            open_file(k$writ, load_file, 0, true, true);
            if file_is_open then
            begin { file opened }
               loop_to_string(what_to_do, start_column);
               verify_print;
               close_file(load_file)
            end { file opened }
            else main_state := command_error
         end
         else bad_command
      end { to string form }
      else with current_buffer do
      begin { number form }
         result := get_number(lines_to_unload);
         if check_termination then
         begin { no extra after command }
            open_file(k$writ, load_file, 0, true, true);
            if file_is_open then
            begin { file opened }
               if lines_to_unload < 0 then
               begin { negative number of lines }
                  start_at := line_number + lines_to_unload + 1;
                  if start_at < 1
                  then start_at := 1;
                  lines_to_unload := line_number - start_at + 1;
                  load_a_line(start_at, false)
               end { negative number of lines } ;
               loop_by_number(what_to_do, lines_to_unload);
               close_file(load_file)
            end { file opened }
            else main_state := command_error
         end { no extra after command }
      end { number form }
   end { command ok, with }
end { unload_proc } ;

{ VERSION command }

procedure version_comm;                                                         { VERSION_COMM }

begin { version_comm }
   str_screen_message('[Sheffield Editor version ~]', false, version_string);
   screen_message(copyright_string, false)
end { version_comm } ;

{ WHERE command }

procedure where_comm;                                                           { WHERE_COMM }

begin { where_comm }
   get_where_message;
   screen_message(message, false)
end { where_comm } ;

{ WINDOW and IW commands }

procedure window_comm(init_wi: boolean);                                        { WINDOW_COMM }

var
   number_of_args, arg1, arg2: integer;

begin { window_comm }
   if in_window_code
   then comm_error('You are already in the window editor!')
   else
   begin { not already in }
      number_of_args := 0;              { count number of actual arguments }
      if get_number(arg1)
      then number_of_args := number_of_args + 1;
      if get_number(arg2)
      then number_of_args := number_of_args + 1;
      if check_termination
      then window_code(number_of_args, arg1, arg2, init_wi)
   end { not already in }
end { window_comm } ;

{ XEQ command }

procedure xeq_comm;                                                             { XEQ_COMM }

var
   bindex: string_buffers2_range;
   strbuf_ptr: ptr_edline;
   line_text_copy: line_chars_type;

begin { xeq_comm }
   skip_spaces;
   if command_line[sub] in string_terminators
   then bindex := 0                     { XEQ alone => XEQ EDLIN }
   else get_buffer_index(bindex, false);
   if main_state = running then
   with current_buffer do
   begin
      if bindex = 0
      then copy_suppress := true;       { XEQ alone => do not copy edit line }
      if bindex < 0 then
      begin { XEQ INLIN }
         { copy line_text because the command could affect it }
         line_text_copy := line_text;
         obey_command_line(line_text_copy, 1, line_length + 2)
      end { XEQ INLIN }
      else
      begin { not INLIN }
         strbuf_ptr := string_buffers[bindex];
         if strbuf_ptr = nil
         then comm_error('Buffer is empty.')
         else with strbuf_ptr^ do
         begin
            if being_obeyed
            then comm_error('Buffer is already being executed.')
            else
            begin { not being executed }
               being_obeyed := true;
               obey_command_line(ed_text, 1, text_length + 2);
               being_obeyed := false
            end { not being executed }
         end { not empty }
      end { not INLIN }
   end { with }
end { xeq_comm } ;

{ Perform generalised FIND, LOCATE, PRINT, DELETE and UNLOAD.
  The command name is made up of a combination of single letters,
  each one indicating an element of the "what_to_do" set of actions,
  finishing with L or F for Locate or Find type.
   e.g. the command BDUL is a "backwards delete unload locate" }

procedure many_commands;                                                        { MANY_COMMANDS }

var
   action: actions;
   what_to_do: set_of_actions;
   windex: word_range;
   ch: char;
   state, inner_state: search_type;
   last_was_L_or_F: boolean;

begin { many_commands }
   what_to_do := [];                    { accumulated actions }
   windex := 1;
   state := searching;
   last_was_L_or_F := false;
   repeat
      ch := command_name[windex];
      if ch = ' '
      then state := found               { end of command_name }
      else if last_was_L_or_F
      then state := not_found           { L or F not at end }
      else
      begin { look in list of action names }
         inner_state := searching;
         action := actions.min;
         repeat
            if action_names[action] = ch
            then inner_state := found
            else if action = actions.max
            then inner_state := not_found
            else action := succ(action)
         until inner_state <> searching;
         if inner_state = not_found     { error - unrecognised letter }
         then state := not_found
         else
         begin { letter recognised }
            if action in what_to_do
            then state := not_found     { error - repeated letter }
            else
            begin { letter not repeated }
               what_to_do := what_to_do + [action];  { add in the action }
               last_was_L_or_F := ch in ['L','F'];
               if windex < word_max_chars
               then windex := windex + 1  { on to the next letter }
               else state := found      { end of command name }
            end { letter not repeated }
         end { letter recognised }
      end { look in list of action names }
   until state <> searching;

   if (state = found) and last_was_L_or_F then
   begin { command alright }
      if find_act in what_to_do         { 'F' => not locate_type }
      then what_to_do := what_to_do - [locate_act];
      if unload_act in what_to_do
      then unload_proc(what_to_do)
      else find_or_locate_proc(what_to_do)
   end { command alright }
   else
   begin { command format error }
      word_screen_message(
         'Unrecognised command "~". Type HELP if you need assistance.', true,
         command_name, true);
      main_state := command_error
   end { command format error }
end { many_commands } ;

%page
{***************************}
{ OBEY THE LINE OF COMMANDS }
{***************************}

begin { obey_command_line }
   repeat
      sub := 1;
      first_command_on_line := true;
      repeat
         ch := command_line[sub];
         if ch = newline                { decide whether to switch modes }
         then if edit_mode
              then switch := not mode_info
              else switch := true
         else switch := false;
         if switch then
         begin { switch modes }
            edit_mode := not edit_mode;
            if edit_mode
            then if in_window_code
                 then
                 else writeln('Edit')
            else change_to_input_mode
         end { switch modes }
         else if not edit_mode
         then input_proc                { input mode }
         else
         begin
            skip_spaces;
            ch := command_line[sub];
            if ch = '*' then            { repetition }
            begin
               if first_command_on_line
               then comm_error('No commands to repeat.')
               else
               begin
                  sub := sub + 1;
                  result := get_number(rep_count);
                  if sub < upto then
                  begin
                     if command_line[sub] in string_terminators then
                     begin
                        rep_count := rep_count - 1;
                        if rep_count <> 0
                        then obey_command_line(command_line, rep_count, sub)
                     end
                     else comm_error('Error in repeat specification.')
                  end { sub < upto }
               end { not first thing on line }
            end { * command repetition }
            else if ch = '!' then       { obey Primos command }
            begin { ! command }
               obey_primos_proc;
               first_command_on_line := false
            end { ! command }
            else
            begin { named command }
               get_word(command_name);
               if mode_info and         { in mode info NEXT is the default }
                  (command_name = '        ')
               then command_name := 'NEXT    ';
               command_state := searching;
               command := commands.min;
               repeat
                  if compare(command_name, command_names[command])
                  then  command_state := found
                  else  if command = commands.max
                        then  command_state := not_found
                        else  command := succ(command)
               until command_state <> searching;
               if command_state = found then
               begin
                  case command of
                     alphabetc: alphabet_comm;
                     appendc:   append_and_overlay_proc(false);
                     blankc:    invalid_comm;
                     bottomc:   top_or_bottom_proc(false);
                     briefc:    mode_verify := false;
                     bufferc:   buffer_comm;
                     centrec:   centre_comm;
                     changec:   change_comm(false);
                     cominputc: cominput_comm(false);
                     delbufc:   delbuf_comm;
                     deletec:   delete_comm;
                     dunloadc:  unload_proc([delete_act,unload_act,locate_act]);
                     editc:     edit_comm(false, false);
                     erasec:    new_symbol(serase);
                     expandc:   expand_comm;
                     extractc:  extract_comm;
                     fileallc:  fileall_or_saveall_proc(true);
                     filec:     file_or_save_proc(true);
                     findc:     find_or_locate_proc([]);
                     fnamec:    fname_comm;
                     gmodifyc:  gmodify_comm;
                     helpc:     help_comm;
                     ibc:       insert_and_retype_proc(true, true);
                     inputc:    not_implemented;
                     insertc:   insert_and_retype_proc(true, false);
                     iwc:       window_comm(false);
                     killc:     new_symbol(skill);
                     lineszc:   linesz_comm;
                     loadc:     load_comm;
                     loadstatc: cominput_comm(true);
                     locatec:   find_or_locate_proc([locate_act]);
                     macclearc: macclear_comm;
                     macloadc:  macload_comm;
                     macprintc: macprint_comm;
                     macsavec:  macsave_comm;
                     markc:     mark_comm;
                     mergec:    merge_comm;
                     modec:     mode_comm;
                     modifyc:   change_comm(true);
                     movec:     move_or_obey_comm(false);
                     newc:      edit_comm(true, false);
                     nextc:     next_comm;
                     nextbuffc: next_or_prev_buff(true);
                     nfindc:    find_or_locate_proc([not_act]);
                     nlocatec:  find_or_locate_proc([locate_act,not_act]);
                     obeyc:     move_or_obey_comm(true);
                     oopsc:     oops_comm;
                     outputc:   not_implemented;
                     overlayc:  append_and_overlay_proc(true);
                     pausec:    pause_comm;
                     pointc:    point_comm;
                     pprintc:   pprint_comm;
                     prevbuffc: next_or_prev_buff(false);
                     printc:    print_comm;
                     psymbolc:  psymbol_comm;
                     ptabsetc:  settabs(number_of_ptabs, ptabs);
                     punchc:    not_implemented;
                     punctc:    punct_comm;
                     qfc:       quit_proc(true);
                     quitc:     quit_proc(false);
                     reloadc:   edit_comm(false, true);
                     replacec:  replace_comm;
                     retypec:   insert_and_retype_proc(false, false);
                     savec:     file_or_save_proc(false);
                     saveallc:  fileall_or_saveall_proc(false);
                     savestatc: savestat_comm;
                     sizec:     size_comm;
                     spellc:    spell_comm;
                     splitc:    split_comm;
                     statsc:    stats_comm;
                     stylec:    style_comm;
                     symbolc:   symbol_comm;
                     tabsetc:   settabs(number_of_tabs, htabs);
                     terminalc: terminal_comm;
                     textc:     text_comm;
                     topc:      top_or_bottom_proc(true);
                     unloadc:   unload_proc([unload_act,locate_act]);
                     verifyc:   mode_verify := true;
                     versionc:  version_comm;
                     wherec:    where_comm;
                     windowc:   window_comm(true);
                     xeqc:      xeq_comm
                  end { case }
               end { command recognised }
               else many_commands;      { try for multi-function locate etc. }
               first_command_on_line := false
            end { named command }
         end { not switching to input mode } ;
         if main_state = running then
         begin { test whether command took all data }
            if check_termination
            then sub := sub + 1         { skip correct terminator }
         end { test whether command took all data }
      until (sub >= upto) or (main_state <> running);
      if times > 0                      { don't count for endless repetition }
      then times := times - 1
   until (times = 0) or (main_state <> running)
end { obey_command_line } ;

{ Read and obey a line of commands }
{ Separate procedure because of restrictions on non-local goto out of onunit }

procedure read_and_obey_command_line;                                           { READ_AND_OBEY_COMMAND_LINE }

label 111;

var
   error_trap_count: integer;

   { On-unit for condition ERROR_TRAP, presumably caused by an editor bug! }

   procedure errortrap_onunit(not_used: integer);

   {---teststart---}
   var
      handle_it: boolean;

   {---teststop---}
   begin { errortrap_onunit }
      lon$cn(1);                        { switch logout notification on }
      {---teststart---}
      if mode_test then
      begin
         screen_message('***Condition ERROR_TRAP raised.', true);
         handle_it := question_and_answer('Obey on-unit code? ', true, true)
      end
      else handle_it := true;
      if handle_it then
      {---teststop---}
      begin { handle the condition }
         error_trap_count := error_trap_count + 1;
         if error_trap_count = 1 then
         begin
            if in_window_code then
            begin
               if term_is_ssmp
               then stop_ssmp
               else vdu_string(vdu_table^.exit_string, 0, 0)
            end;
            in_window_code := false;
            writeln;
            writeln('An unexpected error has occurred.');
            writeln('Please report to analyst responsible.');
            writeln('Attempting to save the file(s) you are editing.');
            main_state := running;
            saveall_proc(true);
            if main_state = running then
            begin { saves worked }
               if paging_file_open
               then close_paging_file(false)  { do not leave a T$ file }
                                        { return to let normal error exit }
                                        { occur }
            end { saves worked }
            else goto 111               { let the user do something }
         end
         else { multiple error - just crash }
      end { handle the condition }
   end { errortrap_onunit } ;

begin { read_and_obey_command_line }
   on('ERROR_TRAP', errortrap_onunit);  { in case of catastrophe! }
   error_trap_count := 0;
   if cominput_on
   then read_cominput_line(edit_line);
   if cominput_on
   then { already read }
   else
   begin { read from terminal }
      reading_command_line := true;
      read_command_line(edit_line, get_c1in);
      reading_command_line := false
   end { read from terminal } ;
   main_state := running;
   copy_suppress := false;
   obey_command_line(edit_line.ed_text, 1, edit_line.text_length + 2);
   { if current line does not contain XEQ etc. then copy it to EDLIN }
   if not copy_suppress then
   begin { copy to EDLIN }
      if string_buffers[0] = nil
      then new(string_buffers[0]);      { create it first }
      string_buffers[0]^ := edit_line
   end { copy to EDLIN } ;
111:
   case main_state of
      running, top_or_bottom, quit_comm: ;
      command_error:
         begin
            writeln('?');
            return_code := -1           { warning return after any errors }
         end;
      quit_signal:
         with current_buffer do
         begin
            writeln('Instruction terminated at line ',line_number:1);
            if not edit_mode then
            begin { switch to edit mode }
               edit_mode := true;
               writeln('Edit')
            end { switch to edit mode } ;
            if cominput_on
            then stop_cominput;
            return_code := -1           { warning return after quit }
         end
   end { case main_state } ;
   if main_state <> running             { close init file after error }
   then test_init_eof(true)
end { read_and_obey_command_line } ;

{ Obey -TIDY on command line.  Write file containing LD-type information. }
{ Enter window editor.  Allow user to perform actions such as deleting    }
{ files.                                                                  }

procedure tidy_proc;                                                            { TIDY_PROC }

var
   ld_filename: long_string;
   temp_path: tree_type;
   code: shortint;

begin { tidy_proc }
   ld_buffer := 2;                      { buffer to use for ld file }
   with temp_path
   do gpath$(k$cura, 0, chars, tree_max_chars, len, code);
   if code <> 0 then
   begin { gpath$ failed }
      error_print(k$irtn, code, 'Getting current pathname', 'GPATH$');
      abort('Failed to get current pathname.')
   end { gpath$ failed } ;
   ld_pathname := temp_path.strng;      { pathname to current directory }
   original_length := length(ld_pathname);
   write_ld_file(ld_filename);          { read dir and write file }
   obey_command_string(concat('buffer 2; ed ',ld_filename,'; brief; point 9'));
   command_line_args.window_sw := true;
   command_line_args.no_line_sw := true;  { stop F1 working }
   edit_mode_after_init := true
end { tidy_proc } ;

   procedure cntblks(theword: word); extern;  {***ds***}

%page
{**************}
{ MAIN PROGRAM }
{**************}

begin { main }
   p$break(false);                      { no system break handling }
   cntblks(command_names[blankc]);      {***ds***}
   writeln('[Sheffield Editor version ',version_string,']');
   writeln(copyright_string);
   main_state := running;
   newline := chr(lf_code);
   punctuation := [' ', '!', '"', '#', '%', '&','''', '(', ')', '*',
                   '+', ',', '-', '.', '/', ':', ';', '<', '=', '>',
                   '?', '[', '\', ']', '^', '`', '{', '|', '}', '~'];
   upper_alphabet := ['A'..'Z'];
   lower_alphabet := ['a'..'z'];
   ascii_printable := [first_print_char .. last_print_char];
   return_code := 0;                    { assume OK }
   locate_string_exists := false;
   change_string_exists := false;
   obey_commands := true;
   suppress_output := false;
   message_lines_to_clear := 0;
   continue_screen_messages := true;
   must_redraw_screen := false;
   replace_command_given := false;
   replace_ptr_1 := nil;
   replace_ptr_2 := nil;
   wi_oops_count := 0;
   first_changed_line := maxint;
   last_changed_line := 0;
   split_command_used := false;
   switch_buffer_used := false;
   display_buffer_name := false;
   screen_tabs_set := false;
   in_window_code := false;
   in_help_system := false;
   current_macro_desc := nil;
   macro_head := nil;
   vdu_table := nil;                    { table not read yet }
   term_is_ssmp := false;
   window_count := 0;
   windows_across := 1;
   windows_down   := 1;
   vdu_treename.strng := 'SYSOVL>VDUTABLES';
   dictionary_online := false;

   if initialise then                   { set up defaults and open files }
   begin { args alright }
      get_erase_and_kill;               { get Primos erase and kill chars }
      if command_line_args.help_sw then
      begin { give command line help }
         reset(input, '@tty');          { just in case user had -init }
         input_state := user_input;
         help_system(command_line_args.help_key);
         obey_commands := false
      end { give command line help } ;
      if obey_commands then
      begin
         hook1(vdu_type);
         illegal_chars := [',', ' ', '*', chr(lf_code)];
         symbol_chars := [init_wild,   init_blanks,    init_tab,
                          init_escape, init_semicolon, init_counter];
         for str_index := 0 to num_string_buffers
         do string_buffers[str_index] := nil;
         edit_line.being_obeyed := false;  { other ents set by read_command_line }
         ed_com_line_valid := nothing_defined;
         ed_com_line.being_obeyed := false;
         reading_command_line := true;  { used by quit on-unit }
         on('QUIT$', quit_onunit);
         on('LOGOUT$', logout_onunit);

         if command_line_args.tidy_sw
         then tidy_proc;                { set up for -tidy operation }
         repeat read_and_obey_command_line
         until main_state = quit_comm;
         if command_line_args.tidy_sw then
         begin { run last CPL file if necessary }
            if length(ld_cpl_filename) > 0 then
            begin { it exists }
               obey_command_string(concat('! resume ',ld_cpl_filename));
               delete_primos_file(ld_cpl_filename)
            end { it exists }
         end { run last CPL file if necessary }
      end { obey commands }
   end { args alright }
   else return_code := 1;               { error in args }
   if input_state in [default_file, init_file] then
   begin { do not leave input file open }
      reset(input, '@tty');
      input_state := user_input
   end { do not leave input file open } ;
   epfresult(return_code);
   hook2(vdu_type)                      { leaving the editor }
end { main } .
