%separate = true
%ecs = true
%ttybuffered = false
%epfargs = true
{ Sheffield University Editor }
{ Copyright (c) University of Sheffield 1990 }

{****************************************}
{ PROCEDURES USED BY EDITOR MAIN PROGRAM }
{****************************************}

subprogram pmed_mprocs;

%listing off
%insert pmed_globals.ins.pas
%insert pmed_efs.ins.pas
%insert pmed_procs.ins.pas
%insert pmed_vdu.ins.pas
%listing on

{ Write a banner for print commands }

procedure print_banner;                                                         { PRINT_BANNER }

begin { print_banner }
   if mode_column
   then if mode_number
        then banner(6)                  { allow offset for line number }
        else banner(0)
end { print_banner } ;

{ Construct string for WHERE and MODE commands }

procedure get_where_message;                                                    { GET_WHERE_MESSAGE }

begin { get_where_message }
   with current_buffer do
   begin
      case line_status of
         top:    message := '(top) ';
         bottom: message := '(bottom) ';
         normal: message := ''
      end;
      append_message('line ~');
      ins_int(line_number, 1);
      if line_status = normal then
      begin
         append_message(' column ~');
         ins_int(line_position, 1)
      end;
      append_message(' in buffer ~');
      ins_int(current_buffer_number, 1)
   end { with }
end { get_where_message } ;

{ Read command line and set up editor buffers }

function initialise: boolean;                                                   { INITIALISE }

label 224;

const
   buffer_cl_max = 100;                 { maximum files on command line }

var
   null_buffer_pointer_rec: buffer_pointer_rec;
   cnt_str: string[10];
   command_line, picture, current_pathname: long_string;
   system_name: string32;
   pix_index, bad_index, code: shortint;
   edit_file_count, files_opened, pathnames_given, paren_count: integer;
   chindex: 1..maxint;
   this_item: string80;
   temp_pointer: ^ file_block;          { only used to get size }
   pathname_expansion, all_worked: boolean;
   bindex: buffer_range;
   ch: char;

   full_command_line_args: record
      buffer_names: array[1..buffer_cl_max] of string80;
      options: command_line_args_type
   end;

   procedure cl$pix(keys: shortint;
                    caller_name: string32;
                    picture_ptr: integer;
                    pixel_size: shortint;
                    com_args: long_string;
                    struct_ptr: integer;
                    var pix_index, bad_index, code: shortint;
                    local_vars_ptr: integer);
      extern;

   procedure heap_error_unit(cfaddr: integer);

   begin { heap_error_unit }
      writeln;
      writeln('Not enough memory to enter successfully.');
      writeln('Type RLS -ALL or ICE and try again.');
      writeln;
      initialise := false;
      goto 224
   end { heap_error_unit } ;

   { Feed current_pathname into the expansion mechanism }

   procedure expand_current_pathname;

   var
      any_expansion: boolean;

      { Deal with an expanded treename generated by wildcarding etc. }

      procedure deliver(expanded_pathname: tree_string;
                        var continue_expanding: boolean);

      var
         temp_treename: tree_type;

      begin { deliver }
{        writeln('expanded_pathname = "',expanded_pathname,'"');  {----}
         edit_file_count := edit_file_count + 1;
         if main_state = running then
         begin
            switch_buffer(edit_file_count);
            with current_buffer.main_file do
            begin
               treename.strng := expanded_pathname;
               if open_edit_file
               then files_opened := files_opened + 1
            end { with }
         end;
         continue_expanding := true
      end { deliver } ;

   begin { expand_current_pathname }
      pathnames_given := pathnames_given + 1;
{     writeln('pathnames_given = ',pathnames_given:1);  {----}
      if length(current_pathname) > tree_max_chars
      then delete(current_pathname, tree_max_chars+1,
                  length(current_pathname) - tree_max_chars);
      expand_pathname(current_pathname, any_expansion, deliver);
      pathname_expansion := pathname_expansion or any_expansion;
      current_pathname := ''
   end { expand_current_pathname } ;

begin { initialise }
   with null_buffer_pointer_rec do
   begin
      bptr := nil;
      pred_buff := 0;
      succ_buff := 0
   end { with } ;
   first_buffer_in_chain := 0;
   last_buffer_in_chain := 0;
   first_chain_setting := 0;
   for bindex := 1 to buffer_max
   do buffer_pointers[bindex] := null_buffer_pointer_rec;
   current_buffer_number := 1;          { start on buffer 1 }
   target_buffer_number := 1;           { rob buffer 2 first }
   paging_file_open := false;
   blocks_paged_out := 0;               { no memory blocks offline yet }
   top_paging_key := 0;
   free_blocks := nil;                  { no saved memory blocks }
   free_keys := nil;                    { ditto file keys }
   new_block_count := 0;                { no blocks created yet }
   index_block_count := 0;              { no index blocks yet }
   init_buffer;
   for ch := first_print_char to last_print_char do  { clear macro pointers }
   with macro_pointers[ch] do
   begin
      pointers[posm] := nil;
      pointers[negm] := nil
   end { for loop } ;
   cominput_on := false;
   mac_treename.strng := '';
   savestat_treename.strng := '';
   input_state := not_open;             { INPUT not RESET yet }
   ld_cpl_filename := '';
   if command_names[ptabsetc] <> 'PTABSET '  {***ds***}
   then halt('The Editor has timed out: contact your distributor.');  {***ds***}

   { read command line args }

   epfargs(command_line, system_name);
   { Use 'char' for pathnames because 'tree' objects to parentheses }
   str(buffer_cl_max, cnt_str);
   picture := concat(
      cnt_str,' * char; ',
      '-init, -i tree; -help, -h char = INDEX; -terminal, -t dec = -1;',
      '-tidy entry = @@; -sort_dtm, -sortm; -reverse, -rv;',
      '-report, -r; -window, -w, -wi; -no_line, -nl, -n; -screen, -s;',
      '-no_init, -ni; -no_primos, -np; -testing, -test;',
      ' end');
   cl$pix(0, system_name, addressof(picture), length(picture), command_line,
          addressof(full_command_line_args), pix_index, bad_index, code, 0);
   if code = 0 then
   with full_command_line_args, options do
   begin { args alright }
      vdu_type := terminal_type;
      if terminal_type >= 0
      then set_term_gv;                 { record it in globals }
      mode_test := testing_sw;
      if screen_sw then
      begin { -screen => -w -n }
         window_sw := true;
         no_line_sw := true
      end { -screen => -w -n } ;
      command_line_args := options;
      if command_line_args.help_sw
      then initialise := true           { No need to try opening files }
      else
      begin
         on('HEAP_ERROR', heap_error_unit);
         {---teststart---}
         if not mode_test then
         {---teststop---}
         p$hpsize(dynamic_segments);
         new_block_max := 64;
         edit_file_count := 0;
         files_opened := 0;
         pathnames_given := 0;
         paren_count := 0;
         current_pathname := '';
         pathname_expansion := false;
         { Any pathnames involving iteration come over as separate bits }
         for bindex := 1 to buffer_cl_max do
         begin { buffer loop }
            this_item := buffer_names[bindex];
            if length(this_item) > 0 then
            begin { buffer given }
{              writeln('bindex = ',bindex:1,'  this_item = "',this_item,'"');  {----}
               if length(current_pathname) > 0
               then current_pathname := concat(current_pathname,' ',this_item)
               else current_pathname := this_item;
{              writeln('current_pathname = "',current_pathname,'"');  {----}
               for chindex := 1 to length(this_item) do
               begin
                  ch := this_item[chindex];
                  if ch = '('
                  then paren_count := paren_count + 1
                  else if ch = ')'
                  then paren_count := paren_count - 1
               end { for loop } ;
{              writeln('paren_count = ',paren_count:1);  {----}
               if paren_count = 0
               then expand_current_pathname  { pathname is complete }
            end { buffer given }
         end { buffer loop } ;
         if length(current_pathname) > 0
         then expand_current_pathname;
         switch_buffer(1);              { start in buffer 1 }
         edit_mode_after_init := edit_file_count > 0;
         multiple_buffers_used := edit_file_count > 1;
         lowest_empty_buffer := edit_file_count + 1;
         first_chain_setting := edit_file_count;
         if main_state <> running
         then all_worked := false       { some error reported }
         else if (pathnames_given > 0) and (edit_file_count = 0) then
         begin { no expanded pathnames }
            writeln('No files match pathname(s) given.');
            all_worked := false
         end { no expanded pathnames }
         else all_worked := (edit_file_count = 0) or  { no files to open or }
                            (files_opened > 0);  { at least one opened ok }
         if all_worked then
         begin { all buffers opened }
            if pathname_expansion
            then writeln(files_opened:1,' files loaded into buffers 1 to ',
                         edit_file_count:1);
            set_edit_mode(edit_mode_after_init)  { set to EDIT or INPUT mode }
         end { all buffers opened }
         else
         begin { some file problem }
            edit_mode := true;          { just in case }
            quit_proc(true)             { close any open files }
         end { some file problem } ;
         if command_line_args.report_sw  { preset report gv to '' }
         then set_global(report_gv_name, '', code);
         initialise := all_worked
      end { not -HELP }
   end { args alright }
   else
   begin { error in args }
      writeln('Error in parameters.  Use -HELP if you need help.');
      initialise := false
   end { error in args } ;
224:
end { initialise } ;

{ Convert a macro code to the visible representation used by MACPRINT }

procedure macro_code_to_visible(macro_code: integer;                            { MACRO_CODE_TO_VISIBLE }
                                var visible_rep: string12);

var
   control_name: char3;
   fiddle: record
      case boolean of
         false: (action_part: action_proc);
         true:  (ord_part: shortint)
      end;
   name: record
      case boolean of
         false: (strng: string[word_max_chars]);
         true:  (len: shortint; chars: word)
      end;
   temp_str: string[5];
   chindex: 1..5;
   action_ord: integer;

   procedure do_printable;
   begin { do_printable }
      visible_rep := ' ';
      visible_rep[1] := chr(macro_code)
   end { do_printable } ;

begin { macro_code_to_visible }
   if macro_code < 200b
   then do_printable;
   if macro_code < 240b then
   begin { 200b .. 237b }
      if macro_code = lf_code
      then visible_rep := '{CR}'
      else
      begin { control char }
         control_name := control_names[macro_code];
         if control_name[3] = ' '
         then visible_rep := '{xx}'
         else visible_rep := '{xxx}';
         for chindex := 1 to length(visible_rep)-2
         do visible_rep[chindex+1] := control_name[chindex]
      end { control char }
   end { 200b .. 237b }
   else if macro_code < 377b
   then do_printable
   else if macro_code = 377b
   then visible_rep := '{DEL}'
   else
   begin { action proc }
      action_ord := macro_code - 256;
      if (action_ord < 0) or (action_ord > ord(action_proc.max)) then
      begin { out of range }
         name.chars := 'ACT     ';
         str(action_ord, temp_str);
         for chindex := 1 to length(temp_str)
         do name.chars[chindex+3] := temp_str[chindex]
      end { out of range }
      else with fiddle do
      begin { fiddle to get symbolic type from ord }
         ord_part := action_ord;
         name.chars := action_proc_print[action_part]
      end { fiddle to get symbolic type from ord } ;
      with name do
      begin
         len := word_max_chars;
         while (len > 1) and (chars[len] = ' ')
         do len := len - 1;
         visible_rep := concat('{',name.strng,'}')
      end { with }
   end { action proc }
end { macro_code_to_visible } ;

{ Convert a visible representation to a macro code }
{ The visible_rep has already had braces removed }

function visible_to_macro_code(visible_rep: string12;                           { VISIBLE_TO_MACRO_CODE }
                               var macro_code: integer): boolean;

type
   three_way = (one, two, three);

var
   result: boolean;
   the_action: action_proc;
   the_control: 200b..237b;
   state: search_type;
   name: record
      case three_way of
         one:   (strng: string[word_max_chars]);
         two:   (len1: shortint; chars: word);
         three: (len2: shortint; char_three: char3)
      end;

begin { visible_to_macro_code }
   result := true;
   if visible_rep = 'CR'
   then macro_code := lf_code
   else if visible_rep = 'DEL'
   then macro_code := del_code
   else
   begin { search tables }
      name.chars := '        ';
      name.strng := visible_rep;
      the_action := action_proc.min;
      state := searching;
      repeat
         if name.chars = action_proc_print[the_action]
         then state := found
         else if the_action = action_proc.max
         then state := not_found
         else the_action := succ(the_action)
      until state <> searching;
      if state = found
      then macro_code := ord(the_action) + 256
      else
      begin { not an action }
         the_control := 200b;
         state := searching;
         repeat
            if name.char_three = control_names[the_control]
            then state := found
            else if the_control = 237b
            then state := not_found
            else the_control := succ(the_control)
         until state <> searching;
         if state = found
         then macro_code := the_control
         else result := false
      end { not an action }
   end { search tables } ;
   visible_to_macro_code := result
end { visible_to_macro_code } ;

{ Return the value of the named variable if possible.    }
{ Calling code has already checked that name length > 0. }

procedure evaluate_variable(variable_name: string32;                            { EVALUATE_VARIABLE }
                            var variable_value: long_string);

var
   chindex: 1..32;
   ch4, ch: char;
   indent_col: line_chars_range;
   bindex: integer;
   recognised: boolean;
   code: shortint;

   procedure integer_value(the_value: integer);

   begin { integer_value }
      str(the_value, variable_value)
   end { integer_value } ;

   procedure copy_buffer(bindex: string_buffers2_range);

   var
      strbuf_ptr: ptr_edline;

      procedure copy_text(the_text: line_chars_type;
                          the_length: line_chars0_range);

      var
         chindex: line_chars0_range;

      begin { copy_text }
         { start with 80 for efficiency }
         variable_value := '                                                                                ';
         for chindex := 1 to the_length do
         begin
            if chindex > length(variable_value)
            then insert(' ', variable_value, length(variable_value)+1);
            variable_value[chindex] := the_text[chindex]
         end { for loop } ;
         if length(variable_value) > the_length
         then delete(variable_value, the_length+1,
                     length(variable_value) - the_length)
      end { copy_text } ;

   begin { copy_buffer }
      if bindex < 0 then
      with current_buffer do
      begin { INLIN }
         if line_exists
         then copy_text(line_text, line_length)
         else variable_value := ''
      end { INLIN }
      else
      begin { not INLIN }
         strbuf_ptr := string_buffers[bindex];
         if strbuf_ptr = nil
         then variable_value := ''
         else with strbuf_ptr^
         do copy_text(ed_text, text_length)
      end { not INLIN }
   end { copy_buffer } ;

   { Find the next decimal number on the current line }

   function next_number: integer;

   var
      col: line_chars_range;
      sign_mult: -1..1;
      valu: integer;
      ch: char;
      state: search_type;
      done: boolean;

   begin { next_number }
      with current_buffer do
      begin
         col := line_position;
         { skip to numeric }
         state := searching;
         repeat
            if col > line_length
            then state := not_found
            else if line_text[col] in ['0'..'9','+','-']
            then state := found
            else col := col + 1
         until state <> searching;
         { evaluate the number }
         if state = found then
         begin { digit found }
            ch := line_text[col];
            valu := 0;
            if ch = '+'
            then sign_mult := +1
            else if ch = '-'
            then sign_mult := -1
            else
            begin
               valu := ord(ch) - ord('0');
               sign_mult := +1
            end;
            done := false;
            repeat
               col := col + 1;
               ch := line_text[col];
               if ch in ['0'..'9'] then
               begin { digit }
                  if valu < 200000000
                  then valu := valu*10 + ord(ch) - ord('0')
               end { digit }
               else done := true
            until done;
            next_number := sign_mult * valu
         end { digit found }
         else next_number := 0
      end { with }
   end { next_number } ;

begin { evaluate_variable }
   with current_buffer do
   begin
      { convert name to upper case }
      for chindex := 1 to length(variable_name) do
      begin
         ch := variable_name[chindex];
         if ch in ['a'..'z']
         then variable_name[chindex] := chr(ord('A') - ord('a') + ord(ch))
      end { for loop } ;
      if variable_name[1] = '.' then
      begin { get global variable }
         gv$get(variable_name, variable_value, line_max_chars, code);
         if code = 0
         then { retrieved alright }
         else if code = e$unop
         then comm_error('Global variables are not active.')
         else comm_error(concat('Unset global variable ',variable_name))
      end { get global variable }
      else
      begin { not global variable }
         { test for standard names }
         if variable_name = 'BUFFNO'
         then integer_value(current_buffer_number)
         else if variable_name = 'LINENO'
         then integer_value(line_number)
         else if variable_name = 'COLNO'
         then integer_value(line_position)
         else if variable_name = 'MLINENO'
         then integer_value(marked_line)
         else if variable_name = 'PLINENO'
         then integer_value(point_line_no)
         else if variable_name = 'PCOLNO'
         then integer_value(point_line_pos)
         else if variable_name = 'FILENAME'
         then variable_value := main_file.treename.strng
         else if variable_name = 'INDENT' then
         begin { find indent col }
            find_indent(indent_col, maxint);
            integer_value(indent_col)
         end { find indent col }
         else if variable_name = 'NEXTNO'
         then integer_value(next_number)
         else if variable_name = 'INLIN'
         then copy_buffer(-1)
         else if variable_name = 'EDLIN'
         then copy_buffer(0)
         else
         begin { is it STR... }
            if length(variable_name) <= 3
            then recognised := false
            else if pos('STR', variable_name) = 1 then
            begin { begins with STR }
               ch4 := variable_name[4];
               if ch4 = '.' then
               begin { STR. form }
                  if length(variable_name) < 5
                  then recognised := false  { too short }
                  else
                  begin { decode number }
                     recognised := true;
                     bindex := 0;
                     for chindex := 5 to length(variable_name) do
                     begin
                        ch := variable_name[chindex];
                        if ch in ['0'..'9']
                        then bindex := bindex*10 + ord(ch) - ord('0')
                        else recognised := false
                     end { for loop }
                  end { decode number }
               end { STR. form }
               else if ch4 in ['A'..'Z'] then
               begin { STRA... }
                  bindex := ord(ch4) - ord('A') + 1;
                  recognised := true
               end { STRA... }
               else recognised := false;
               if recognised
               then recognised := (bindex >= 1) and (bindex <= num_string_buffers);
               if recognised
               then copy_buffer(bindex)
            end { begins with STR }
            else recognised := false;
            if not recognised
            then comm_error(concat('Unset variable ',variable_name))
         end { is it STR... }
      end { not global variable }
   end { with }
end { evaluate_variable } ;

{ Substitute values for any %variables% in the given command line }

procedure substitute_variables(var the_line: edline_type);                      { SUBSTITUTE_VARIABLES }

var
   all_substituted: boolean;
   variable_name: string32;
   variable_value: long_string;
   col, start_col, stop_col: line_chars_range;
   name_length: integer;
   chindex: 1..32;

   function find_next_percent: boolean;

   var
      state: search_type;

   begin { find_next_percent }
      with the_line do
      begin
         state := searching;
         repeat
            if col > text_length
            then state := not_found
            else if ed_text[col] = '%'
            then state := found
            else col := col + 1
         until state <> searching;
         find_next_percent := state = found
      end { with }
   end { find_next_percent } ;

   { Replace %variable% by the value found }

   procedure replace_variable;

   var
      length_increase, new_text_length: integer;
      chindex: line_chars_range;

      { Move "wd_count" words from "from_string[fromsub]" to "to_string[tosub]".
        Preserves the words moved by copying either forwards or backwards. }

      procedure movewds(from_string: line_chars_type;
                        fromsub: integer;
                        var to_string: line_chars_type;
                        tosub: integer;
                        wd_count: integer);
         extern;                        { written in PMA }

   begin { replace_variable }
      with the_line do
      begin
         length_increase := length(variable_value) - name_length - 2;
         new_text_length := text_length + length_increase;
         if new_text_length > line_max_charsm1
         then comm_error('Command too long after variable expansion.')
         else
         begin { length alright }
            if length_increase <> 0 then
            begin { length has changed }
               movewds(ed_text, stop_col+1,
                       ed_text, stop_col+1+length_increase,
                       text_length-stop_col+1);
               text_length := new_text_length;
               col := col + length_increase
            end { length has changed }
            else { same length - no move needed } ;
            { copy in new value }
            for chindex := 1 to length(variable_value)
            do ed_text[start_col + chindex - 1] := variable_value[chindex]
         end { length alright }
      end { with }
   end { replace_variable } ;

begin { substitute_variables }
   with the_line do
   begin
      col := 1;
      all_substituted := false;
      repeat
         if find_next_percent then
         begin { start of variable found }
            start_col := col;
            col := col + 1;
            if find_next_percent then
            begin { end of variable found }
               stop_col := col;
               col := col + 1;
               name_length := stop_col - start_col - 1;
               if name_length = 0
               then variable_value := '%'  { %% becomes % }
               else if name_length > 32 then
               begin { name too long }
                  comm_error('Variable name too long.');
                  all_substituted := true
               end { name too long }
               else
               begin { evaluate variable }
                  variable_name := '                                ';
                  for chindex := 1 to name_length
                  do variable_name[chindex] := ed_text[start_col + chindex];
                  variable_name := substr(variable_name, 1, name_length);
                  evaluate_variable(variable_name, variable_value)
               end { evaluate variable } ;
               if main_state = running
               then replace_variable
            end { end of variable found }
            else
            begin { no second % }
               comm_error('Unmatched percent symbols.');
               all_substituted := true
            end { no second % }
         end { start of variable found }
         else all_substituted := true
      until all_substituted
   end { with }
end { substitute_variables } ;

{ Obey a line of editor commands (given as a string) in line editor mode. }

procedure obey_command_string(the_commands: long_string);                       { OBEY_COMMAND_STRING }

var
   temp_line: edline_type;
   chindex: line_chars_range;

   procedure obey_command_line(var command_line: line_chars_type;
                               times: integer;
                               upto: line_chars_plus_range);
      extern;

begin { obey_command_string }
   with temp_line do
   begin
      if special[ssemicolon] <> ';' then
      begin { allow for non-standard semi symbol }
         for chindex := 1 to length(the_commands)
         do if the_commands[chindex] = ';'
            then the_commands[chindex] := special[ssemicolon]
      end { allow for non-standard semi symbol } ;
      for chindex := 1 to length(the_commands)
      do ed_text[chindex] := the_commands[chindex];
      ed_text[length(the_commands)+1] := newline;
      text_length := length(the_commands);
      being_obeyed := false;
      obey_command_line(ed_text, 1, text_length+2)
   end { with }
end { obey_command_string } ;

begin { pmed_mprocs }
end { pmed_mprocs } .
