%separate = true
%ecs = true
%ttybuffered = false
%epfargs = true
{ Sheffield University Editor }
{ Copyright (c) University of Sheffield 1990 }

{********************************************}
{ PROCEDURES USED BY ALL PARTS OF THE EDITOR }
{********************************************}

subprogram pmed_procs;

%listing off
%insert pmed_globals.ins.pas
%insert pmed_efs.ins.pas
%insert pmed_ssmp.ins.pas
%insert pmed_vdu.ins.pas
%listing on

{ Check whether a Primos file exists }

function file_exists(filename: tree_string): boolean;                           { FILE_EXISTS }

var
   unit, tipe, code: shortint;

begin { file_exists }
   call_srsfx$(k$exst+k$getu, filename, unit, tipe, code);
   file_exists := code = 0
end { file_exists } ;

{ Function to get an input character from terminal }

function get_c1in: char;                                                        { GET_C1IN }

var
   ch: char;

begin { get_c1in }
   c1in(ch);
   get_c1in := ch
end { get_c1in } ;

{ Test for end of (simulated) cominput file, and read next line if open. }
{ Assumes cominput was on at entry.                                      }

procedure read_cominput_line(var the_string_buffer: edline_type);               { READ_COMINPUT_LINE }

var
   saved_buffer_number: buffer_range;
   col: line_chars_range;

begin { read_cominput_line }
   with current_buffer, the_string_buffer do
   begin
      if cominput_buffer = 0 then
      begin { from file }
         if eof(cominput_file) then
         begin { at end }
            cominput_on := false;
            close(cominput_file)
         end { at end }
         else
         begin { not at end }
            col := 1;
            while (not eoln(cominput_file)) and (col <= line_max_charsm1) do
            begin
               read(cominput_file, ed_text[col]);
               col := col + 1
            end { while loop } ;
            readln(cominput_file);
            cominput_lineno := cominput_lineno + 1;
            ed_text[col] := newline;
            text_length := col - 1;
            being_obeyed := false
         end { not at end }
      end { from file }
      else
      begin { from buffer }
         saved_buffer_number := current_buffer_number;
         switch_buffer(cominput_buffer);
         load_a_line(cominput_lineno, false);
         cominput_lineno := cominput_lineno + 1;
         if line_status = normal then
         begin { get line from buffer }
            ed_text := line_text;
            text_length := line_length;
            being_obeyed := false
         end { get line from buffer }
         else cominput_on := false;
         switch_buffer(saved_buffer_number)
      end { from buffer } ;
      if not cominput_on then
      begin { message at end }
         if obeying_loadstate
         then str_screen_message('Editing state restored from file ~', false,
                                 savestat_treename.strng)
         else screen_message('End of cominput file: input from terminal.',
                             false);
         screen_message('', false)
      end { message at end }
   end { with }
end { read_cominput_line } ;

{ Stop cominput after quit. }
{ Assumes cominput was actually on. }

procedure stop_cominput;                                                        { STOP_COMINPUT }

begin { stop_cominput }
   cominput_on := false;
   if cominput_buffer = 0
   then close(cominput_file)
end { stop_cominput } ;

{ Print a Primos error message }

procedure error_print(key, code: shortint;                                      { ERROR_PRINT }
                      error_text: long_string;
                      name: string32);

type
   string1024 = string[1024];

var
   message_text: string1024;
   fiddle1: record
      case boolean of
         false: (strng: long_string);
         true:  (len: shortint; chars: long_chars)
      end;

   fiddle2: record
      case boolean of
         false: (strng: string32);
         true:  (len: shortint; chars: char32)
      end;

   procedure ertxt$(code: shortint; var errmsg: string1024);
      extern;

   procedure errpr$(key, code: shortint;
                    error_text: long_chars;
                    text_len: shortint;
                    name: char32;
                    namlen: shortint);
      extern;

begin { error_print }
   message_text := '';
   ertxt$(code, message_text);
   if length(message_text) > 0 then
   begin { message text found }
      screen_message(concat(message_text,' ',error_text,' (',name,')'), true)
   end { message text found }
   else
   begin { no message text found }
      fiddle1.strng := error_text;
      fiddle2.strng := name;
      errpr$(key, code, fiddle1.chars, fiddle1.len, fiddle2.chars, fiddle2.len)
   end { no message returned }
end { error_print } ;

{ Editor internal error - signal ERROR_TRAP to abort }

procedure abort(message: string);                                               { ABORT }

   { Signal an error }

   procedure signal_error(cond_name: string32; allow_return: boolean);

   const
      null_ptr = 1777600000b;           { PLP null pointer 7777/0 }

   type
      bit16 = (b1, b2, b3, b4, b5, b6, b7, b8, b9, b10,
               b11, b12, b13, b14, b15, b16);
      bit_word = set of bit16;

   var
      action: bit_word;

      procedure signl$(cond_name: string32;
                       ms_ptr: integer;
                       ms_len: shortint;
                       info_ptr: integer;
                       info_len: shortint;
                       action: bit_word);
         extern;

   begin { signal_error }
      if allow_return
      then action := [b1]
      else action := [];
      signl$(cond_name, null_ptr, 0, null_ptr, 0, action)
   end { signal_error } ;

begin { abort }
   screen_message('', false);
   str_screen_message('Internal error (~).', true, message);
   screen_message('', false);
   signal_error('ERROR_TRAP', true);
   signal_error('ERROR', false)
end { abort } ;

{ Test whether running from cominput file }

function cominput_open: boolean;                                                { COMINPUT_OPEN }

var
   switch, unit: shortint;

   procedure accom$(var switch, unit: shortint;
                    action: char4);
      extern;

begin { cominput_open }
   accom$(switch, unit, 'READ');
{  writeln('cominput_open: switch = ',switch:1,'  unit = ',unit:1);  {----}
   cominput_open := switch < 0
end { cominput_open } ;

{ Calculate how many dynamic segments are free (EPF version only) }

function free_dynamic_segments: integer;                                        { FREE_DYNAMIC_SEGMENTS }

var
   access, free_segments: integer;
   top_seg, segno: shortint;

   function dy$sgs: shortint;
      extern;

   function tl$sgs: shortint;
      extern;

   function rsegac$(segno: shortint; var access: integer): shortint;
      extern;

begin { free_dynamic_segments }
   free_segments := 0;
   top_seg := tl$sgs;
   for segno := (top_seg - dy$sgs + 1) to top_seg do
   begin
      if rsegac$(segno, access) = 0
      then free_segments := free_segments + 1
   end { for loop } ;
   free_dynamic_segments := free_segments
end { free_dynamic_segments } ;

{ Set or reset terminal characteristics }

procedure ctrl_term(cont: cont_type; switch_on: boolean);                       { CTRL_TERM }

var
   short_or_bits: record
      case boolean of
         true:  (int: shortint);
         false: (bits: packed array[1..16] of boolean)
      end;
   rubbish: shortint;

   function duplx$(newmode: shortint): shortint;
      extern;

begin { ctrl_term }
   with short_or_bits do
   case cont of
      rekord: duplex_word := duplx$(-1);  { record old state }
      reinstate: rubbish := duplx$(duplex_word);  { reinstate }
      half_duplex:
         begin
            int := duplx$(-1);          { get current state }
            bits[1] := switch_on;       { set half duplex }
            bits[2] := switch_on;       { and no lf echo }
            echoing_on := not switch_on;  { record current echo state }
            rubbish := duplx$(int)
         end;
      xonrec:
         begin
            int := duplx$(-1);          { get current state }
            bits[3] := switch_on;       { allow XON/XOFF }
            rubbish := duplx$(int)
         end
   end { case}
end { ctrl_term } ;

{ Same as ctrl_term except that SSMP terminals are not affected }

procedure ctrl_half_duplex(switch_on: boolean);                                 { CTRL_HALF_DUPLEX }

begin { ctrl_half_duplex }
   if term_is_ssmp
   then  { leave SSMP terminal duplex state alone }
   else ctrl_term(half_duplex, switch_on)
end { ctrl_half_duplex } ;

{ Trim leading and trailing spaces from a string }

procedure trim_string(var the_string: long_string);                             { TRIM_STRING }

var
   first_sig, last_sig: integer;
   done: boolean;

begin { trim_string }
   last_sig := length(the_string);
   done := false;
   repeat
      if last_sig <= 0
      then done := true
      else if the_string[last_sig] = ' '
      then last_sig := last_sig - 1
      else done := true
   until done;
   first_sig := 1;
   done := false;
   repeat
      if first_sig > last_sig
      then done := true
      else if the_string[first_sig] = ' '
      then first_sig := first_sig + 1
      else done := true
   until done;
   if (first_sig > 1) or (last_sig < length(the_string))
   then the_string := substr(the_string, first_sig, last_sig - first_sig + 1)
end { trim_string } ;

{ Convert a string to upper case }

procedure upcase_string(var the_string: long_string);                           { UPCASE_STRING }

var
   string_index: integer;
   ch:char;

begin { upcase_string }
   for string_index := 1 to length(the_string) do
   begin
      ch := the_string[string_index];
      if ch in ['a'..'z']
      then the_string[string_index] := chr(ord('A') - ord('a') + ord(ch))
   end
end { upcase_string } ;

{ Calculate the sizes of windows based on new screen size or split information }

procedure calculate_window_sizes;                                               { CALCULATE_WINDOW_SIZES }

var
   windex: integer;
   across_index, down_index: window_range;

begin { calculate_window_sizes }
   with vdu_table^ do
   begin
      { Initialise special [0] window used when screen not split }
      with window_array[0] do
      begin
         window_left := 1;
         window_right := cols_per_row;
         window_width := cols_per_row;
         window_top := 1;
         window_bottom := rows_per_page;
         window_length := rows_per_page;
         window_at_right := true;
         window_at_bottom := true;
         window_full_width := true;
         window_mode := true
      end { with } ;
      if window_count > 0 then
      begin { split screen }
         cols_across := (cols_per_row + 1) div windows_across;
         rows_down   := (rows_per_page+ 1) div windows_down;
         for across_index := 1 to windows_across do
         begin
            for down_index := 1 to windows_down do
            begin
               windex := across_index + (down_index-1)*windows_across;
               if windex <= window_count then
               with window_array[windex] do
               begin { window exists }
                  window_bottom := down_index * rows_down - 1;
                  window_top    := window_bottom - rows_down + 2;
                  window_length := rows_down - 1;
                  window_right  := across_index * cols_across - 1;
                  window_left   := window_right - cols_across + 2;
                  window_width  := cols_across - 1;
                  window_at_right := across_index = windows_across;
                  window_full_width := windows_across = 1;
                  window_at_bottom := window_full_width and
                                      (down_index = windows_down);
                  window_mode := true
               end { window exists }
            end { for loop }
         end { for loop }
      end { split screen }
   end { with }
end { calculate_window_sizes } ;

{ Print an error message and set main_state }

procedure comm_error(message: string);                                          { COMM_ERROR }

begin { comm_error }
   screen_message(message, true);
   main_state := command_error
end { comm_error } ;

{ Compare two words for equality.  Comparison ends at the first blank in the }
{ first word }

function compare(word1, word2: word): boolean;                                  { COMPARE }

var
   windex: word_range;
   state: search_type;
   ch: char;

begin { compare }
   windex := 1;
   state := searching;
   repeat
      ch := word1[windex];
      if ch = ' '
      then state := found
      else if ch = word2[windex]
      then if windex < word_max_chars
           then windex := windex + 1
           else state := found
      else state := not_found
   until state <> searching;
   compare := state = found
end { compare } ;

{ Write out a treename }

procedure write_tree(var pfile: primos_file);                                   { WRITE_TREE }

begin { write_tree }
   with pfile do
   begin
      if treename.len = 0
      then screen_message('Filename not specified.', true)
      else screen_message(treename.strng, true)
   end { with }
end { write_tree } ;

{ Convert string to decimal }

function tree_to_dec(in_tree: tree_type; var valu: integer): boolean;           { TREE_TO_DEC }

var
   tree_col: tree_range;
   state: search_type;
   ch: char;

begin { tree_to_dec }
   with in_tree do
   begin
      tree_col := 1;
      valu := 0;
      state := searching;
      repeat
         if tree_col > len
         then state := found
         else
         begin
            ch := chars[tree_col];
            if ch in ['0'..'9'] then
            begin
               if valu > large_int
               then state := not_found
               else
               begin
                  valu := valu*10 + ord(ch) - ord('0');
                  if tree_col < len
                  then tree_col := tree_col + 1
                  else state := found
               end
            end { valid numeric }
            else state := not_found
         end { still more characters }
      until state <> searching;
      tree_to_dec := state = found
   end { with }
end { tree_to_dec } ;

{ Write a character as its octal representation }

procedure wrt_ch_octal(ch: char);                                               { WRT_CH_OCTAL }

var
   valu: integer;
   digit_no: 1..3;
   digits: array[1..3] of char;

begin { wrt_ch_octal }
   t1ou(special[sescape]);
   valu := ord(ch);                     { ascii code for char }
   for digit_no := 3 downto 1 do
   begin
      digits[digit_no] := chr( (valu mod 8) + ord('0') );
      valu := valu div 8
   end { for loop } ;
   for digit_no := 1 to 3
   do t1ou(digits[digit_no])
end { wrt_ch_octal } ;

{ Write out a character, allowing for all the editor special treatments }

procedure wrt_ch_mode(ch: char);                                                { WRT_CH_MODE }

begin { wrt_ch_mode }
   if mode_ckpar and (ch < chr(200b))
   then wrt_ch_octal(ch)                { print as octal }
   else if ch in print_specials then
   begin { print special as ^ something }
      if ch = special[sescape] then
      begin { print ^ as ^^ }
         t1ou(ch); t1ou(ch)
      end { print ^ as ^^ }
      else wrt_ch_octal(ch)             { specials as ^nnn }
   end { print special as ^ something }
   else t1ou(ch)
end { wrt_ch_mode } ;

{ Print the current line }

procedure print_line;                                                           { PRINT_LINE }

var
   lindex: line_chars_range;
   ch: char;

begin { print_line }
   if in_window_code
   then { do nothing if command obeyed from within window code }
   else with current_buffer do
   case line_status of
      top:    writeln('.top.');
      normal:
         if line_exists then
         begin
            if mode_number
            then write(line_number:5, ':');
            for lindex := 1 to line_length do
            begin
               ch := line_text[lindex];
               if (ch < first_print_char) or (ch > last_print_char)
               then wrt_ch_mode(ch)     { tricky cases }
               else
               begin { printable }
                  if ch = special[sescape]
                  then t1ou(ch);        { ^ prints as ^^ }
                  t1ou(ch)
               end { printable }
            end;
            tonl
         end
         else writeln('.null.');
      bottom: writeln('.bottom.')
   end { not in window code }
end { print_line } ;

{ Do verification print }

procedure verify_print;                                                         { VERIFY_PRINT }

begin { verify_print }
   with current_buffer do
   if mode_verify and (line_status = normal) and (main_state = running)
   then print_line
end { verify_print } ;

{ Find the first non-blank column in the current line, provided it is within }
{ the screen width }

procedure find_indent(var indent_col: line_chars_range;                         { FIND_INDENT }
                      highest_allowed: integer);

begin { find_indent }
   with current_buffer do
   if line_status <> normal
   then indent_col := 1
   else
   begin { normal line }
      strip_trailing_spaces;
      indent_col := 1;
      while line_text[indent_col] = ' '
      do indent_col := indent_col + 1;
      if indent_col > highest_allowed
      then indent_col := 1
   end { normal line }
end { find_indent } ;

{ Write a banner heading for input mode or print (Mode Column) }

procedure banner(banner_indent: integer);                                       { BANNER }

var
   index: 1..79;

begin { banner }
   if in_window_code
   then { do not output it }
   else
   begin { not in window code }
      for index := 1 to banner_indent   { banner is offset to the right by }
      do write(' ');                    { "banner_indent" columns }
      for index := 1 to ((79 - banner_indent) div 10 )
      do write('         ', index:1);
      writeln;
      for index := 1 to banner_indent
      do write(' ');
      for index := 1 to (79 - banner_indent)
      do write(index mod 10:1);
      writeln
   end { not in window code }
end { banner } ;

{ Update the range of lines changed by commands such as CHANGE }

procedure update_changed_lines(ln: integer);                                    { UPDATE_CHANGED_LINES }

begin { update_changed_lines }
   with current_buffer do
   begin
      if ln < first_changed_line then
      begin { before first }
         first_changed_line := ln;
         {---teststart---}
         if mode_test
         then writeln('first_changed_line = ',first_changed_line:1)
         {---teststop---}
      end { before first } ;
      if ln > last_changed_line then
      begin { after last  }
         last_changed_line := ln;
         {---teststart---}
         if mode_test
         then writeln('last_changed_line = ',last_changed_line:1)
         {---teststop---}
      end { after last  } ;
      some_lines_changed := true
   end { with }
end { update_changed_lines } ;

{ Convert a string to integer }

function string_to_integer(the_string: string;                                  { STRING_TO_INTEGER }
                           var the_integer: integer): boolean;

var
   worked, negate: boolean;
   val: integer;
   sindex: 1..80;
   ch: char;

begin { string_to_integer }
   worked := true;
   negate := false;
   val := 0;
   for sindex := 1 to length(the_string) do
   begin
      ch := the_string[sindex];
      if ch in ['0'..'9']
      then if val < large_int
           then val := val*10 + ord(ch) - ord('0')
           else worked := false
      else if ch = ' '
      then                              { ignore spaces }
      else if ch = '-'
      then negate := not negate
      else worked := false              { not recognised }
   end { for loop } ;
   if negate
   then the_integer := -val
   else the_integer := val;
   string_to_integer := worked
end { string_to_integer } ;

{ Set up the sets of char used in command recognition }

procedure set_up_sets;                                                          { SET_UP_SETS }

var
   overlay_extras: set_of_char;

   {---teststart---}

   { Write out a set of characters }

   procedure write_set(the_set: set_of_char);

   var
      ch: char;
      first: boolean;

   begin { write_set }
      first := true;
      write('[');
      for ch := char.min to char.max do
      begin
         if ch in the_set then
         begin { char is in set }
            if first
            then first := false
            else write(' ');
            if ch in [chr(000b)..chr(237b), chr(377b)]
            then wrt_ch_octal(ch)
            else write(ch)
         end { char is in set } ;
      end { for loop } ;
      writeln(']')
   end { write_set } ;

   {---teststop---}

begin { set_up_sets }
   symbol_chars := [special[swild], special[sblanks], special[stab],
                    special[sescape], special[ssemicolon], special[scounter] ];
   { only include erase and kill if we are not processing -init file }
   if input_state = user_input then
   begin { reading from terminal }
      if (special[serase] in symbol_chars) or (special[skill] in symbol_chars)
      then screen_message(
           'Your erase or kill symbol conflicts with an editor symbol.', true);
      symbol_chars := symbol_chars + [special[skill], special[serase]]
   end { reading from terminal } ;
   string_terminators := [chr(lf_code), ',', special[ssemicolon]];
   string_specials := string_terminators +
                      [special[sescape], special[sblanks], special[swild]];
   gmod_input_specials := [chr(lf_code), special[sescape]];
   if mode_count
   then gmod_input_specials := gmod_input_specials + [special[scounter]];
   input_specials := gmod_input_specials + [special[stab]];
   if mode_semi
   then input_specials := input_specials + [special[ssemicolon]];
   overlay_extras := [' ', special[swild]];
   overlay_specials := input_specials + overlay_extras;
   gmod_overlay_specials := gmod_input_specials + overlay_extras;
   print_specials := [chr(000b)..chr(037b), chr(177b),
                      chr(200b)..chr(237b), chr(377b), special[sescape]];
   {---teststart---}
   if mode_test then
   begin
      write('symbol_chars = ');       write_set(symbol_chars);
      write('string_terminators = '); write_set(string_terminators);
      write('string_specials = ');    write_set(string_specials);
      write('input_specials = ');     write_set(input_specials);
      write('overlay_specials = ');   write_set(overlay_specials);
      write('gmodinput = ');          write_set(gmod_input_specials);
      write('gmodoverlay = ');        write_set(gmod_overlay_specials);
      write('print_specials = ');     write_set(print_specials)
   end
   {---teststop---}
end { set_up_sets } ;

{ Set the erase and kill symbols from the ones held by Primos }

procedure get_erase_and_kill;                                                   { GET_ERASE_AND_KILL }

type
   char2 = packed array[1..2] of char;

var
   code: shortint;
   temp_erase, temp_kill: char2;

   procedure erkl$$(key: shortint;      { read Primos erase and kill chars }
                    var erase, kill: char2;
                    var code: shortint);
      extern;

begin { get_erase_and_kill }
   erkl$$(k$read, temp_erase, temp_kill, code);
   if code = 0 then
   begin { read alright }
      { This code circumvents a Primos problem: the top eight bits }
      { may not be zero when returned }
      special[serase] := temp_erase[2];
      special[ skill] := temp_kill [2];
      set_up_sets
   end { read alright }
   else error_print(k$irtn, code, 'Reading erase and kill', 'ERKL$$')
end { get_erase_and_kill } ;

{ Perform a change to input mode, outputting column heading if requested. }

procedure change_to_input_mode;                                                 { CHANGE_TO_INPUT_MODE }

begin { change_to_input_mode }
   if not in_window_code
   then screen_message('Input', false);
   edit_mode := false;
   if mode_column
   then if mode_prompt
        then banner(2)                  { allow for prompt char and space }
        else banner(0)
end { change_to_input_mode } ;

{ Test for eof on the init file (if open), and reset to terminal if true }

procedure test_init_eof(force_eof: boolean);                                    { TEST_INIT_EOF }

   procedure try_next_file; forward;

   procedure open_input_file(tree: tree_string; print_fail: boolean);

   label 231;

   var
      reset_worked: boolean;

      procedure error_trap_unit(not_used: integer);
      begin { error_trap_unit }
         goto 231
      end { error_trap_unit } ;

      procedure io_error_unit(not_used: integer);
      begin { io_error_unit }
         if print_fail
         then                           { just return to let error message }
                                        { be printed }
         else goto 231                  { try next without message }
      end { io_error_unit } ;

   begin { open_input_file }
      reset_worked := false;
      if length(tree) = 0
      then { no point trying }
      else
      begin
         on('IO_ERROR', io_error_unit);
         on('ERROR_TRAP', error_trap_unit);
         reset(input, tree);
         reset_worked := true
      end;
   231:
      if not reset_worked
      then try_next_file
   end { open_input_file } ;

   { Open file PMED_INIT in origin directory if possible, but not if -NO_INIT }
   { given on command line }

   procedure open_default_init_file;

   label 232;

   var
      code: shortint;
      reset_worked: boolean;

      procedure onunit(not_used: integer);
      begin { onunit }
         goto 232
      end { onunit } ;

   begin { open_default_init_file }
      reset_worked := false;
      if not command_line_args.no_init_sw then
      begin { search for PMED_INIT allowed }
         at$or(k$setc, code);
         if code = 0 then
         begin
            on('IO_ERROR', onunit);
            on('ERROR_TRAP', onunit);
            reset(input, 'PMED_INIT');
            reset_worked := true
         end { attached to origin } ;
      232:
         attach_home
      end { search for PMED_INIT } ;
      if not reset_worked
      then try_next_file
   end { open_default_init_file } ;

   { Start reading commands from terminal }

   procedure open_user_input;

   begin { open_user_input }
      reset(input,'@tty');              { read from terminal }
      set_up_sets;                      { add in erase and kill symbols now }
      if in_window_code
      then edit_mode := true            { never switch to INPUT from window }
      else if edit_mode_after_init then
      begin { now edit mode }
         edit_mode := true;
         writeln('Edit')
      end { now edit mode }
      else change_to_input_mode;
      { avoid backing up T$ files }
      if command_line_args.tidy_sw
      then mode_backup := false
   end { open_user_input } ;

   procedure try_next_file;
   begin { try_next_file }
      input_state := succ(input_state);
      case input_state of
         default_file: open_input_file('SYSOVL>PMED_DEFAULTS', false);
         init_file:
            if length(command_line_args.init_path) > 0
            then open_input_file(command_line_args.init_path, true)
            else open_default_init_file;
         user_input:   open_user_input
      end { case }
   end { try_next_file } ;

begin { test_init_eof }
   case input_state of
      not_open: try_next_file;
      default_file, init_file:
         begin
            if force_eof or eof(input)
            then try_next_file
         end;
      user_input:
   end { case }
end { test_init_eof } ;

{ Set a global variable, allowing for a quirk in GV$SET; if the globals file }
{ is in use (e.g. the user is editing it, presumably by accident) then it }
{ waits for ever }

procedure set_global(name: string;                                              { SET_GLOBAL }
                     valu: string;
                     var code: shortint);

label 222;

   procedure quit_onunit(not_used: integer);
   begin { quit_onunit }
      screen_message('',false);
      screen_message('Abandoning setting of global variable.', true);
      screen_message('',false);
      goto 222
   end { quit_onunit } ;

begin { set_global }
   on('QUIT$', quit_onunit);
   gv$set(name, valu, code);
222:
end { set_global } ;

{ Get the default vdu_type from a global variable if possible }

procedure get_term_gv;                                                          { GET_TERM_GV }

var
   gv_val: long_string;
   code: shortint;

begin { get_term_gv }
   gv$get(term_gv_name, gv_val, line_max_chars, code);
   if code = 0 then
   begin { value obtained }
      if length(gv_val) > 80
      then { value too long }
      else if string_to_integer(gv_val, vdu_type)
      then if vdu_type < 0
           then vdu_type := 1
           else { valid type given }
      else { conversion error }
   end { value obtained }
end { get_term_gv } ;

{ Set the terminal global variable if possible }

procedure set_term_gv;                                                          { SET_TERM_GV }

var
   gv_val: string;
   code: shortint;

begin { set_term_gv }
   str(vdu_type, gv_val);
   set_global(term_gv_name, gv_val, code)
end { set_term_gv } ;

{ Decide on vdu_type from evidence in global variable or hook routine, but }
{ only if user has not specified it so far. }

procedure decide_vdu_type;                                                      { DECIDE_VDU_TYPE }

begin { decide_vdu_type }
   if vdu_type < 0 then
   begin { not yet decided }
      hook8(vdu_type);                  { automatic setting, or ask user }
      get_erase_and_kill;
      if vdu_type < 0                   { still not set }
      then get_term_gv                  { get from global or default 1 }
      else set_term_gv                  { record type chosen }
   end { not yet decided }
   else;                                { already decided by user }
   if vdu_type < 0
   then vdu_type := 1                   { set default type }
   else if vdu_type = 0 then
   begin
      writeln('Your terminal is recorded as being a dumb terminal and');
      writeln('therefore you cannot use the window editor system.');
      writeln('Type QUIT if you want to leave the editor.');
      main_state := command_error
   end
end { decide_vdu_type } ;

{ Print a message and wait for a reply }

procedure message_and_reply(message: long_string;                               { MESSAGE_AND_REPLY }
                            highlight_it: boolean;
                            var answer: long_string;
                            var cr_given: boolean);

var
   temp_tree: tree_type;
   done: boolean;
   ch: char;

begin { message_and_reply }
   continue_screen_messages := true;    { just in case }
   if in_window_code and term_is_ssmp
   then get_ssmp_reply(message, highlight_it, answer, cr_given)
   else with temp_tree do
   begin { not SSMP }
      no_lf_screen_message(message, highlight_it);
      test_init_eof(false);
      len := 0;
      done := false;
      repeat
         if eoln(input) then
         begin { eoln }
            readln;
            done := true;
            cr_given := true
         end { eoln }
         else
         begin { not eoln }
            read(ch);
            if ch = chr(can_code) then
            begin
               done := true;
               cr_given := false
            end
            else if ch = special[skill]
            then len := 0               { reset to start }
            else if ch = special[serase]
            then if len > 0             { erase one character }
                 then len := len - 1
                 else
            else
            begin
               if len < line_max_chars then
               begin
                  len := len + 1;
                  chars[len] := ch
               end
            end
         end { not eoln }
      until done;
      answer := strng;
      trim_string(answer)
   end { not SSMP } ;
   if in_window_code and highlight_it and vdu_table^.mess_type1
   then vdu_control(messoff)
end { message_and_reply } ;

{ Get a yes or no answer to a question }

function question_and_answer(message: long_string;                              { QUESTION_AND_ANSWER }
                             highlight_it: boolean;
                             default_value: boolean): boolean;

var
   answer: long_string;
   ch: char;
   cr_given, ok: boolean;

begin { question_and_answer }
   repeat
      message_and_reply(message, highlight_it, answer, cr_given);
      if length(answer) = 0
      then ch := ' '
      else ch := answer[1];
      ok := ch in ['Y','y','N','n',' ','Q','q'];
      if not ok
      then message := 'Please answer yes or no: '
   until ok;
   if cr_given
   then if ch = ' '
        then question_and_answer := default_value
        else question_and_answer := ch in ['Y','y']
   else question_and_answer := false    { CAN answer => false }
end { question_and_answer } ;

{ Format some text according to current STYLE.                                }
{ Gets next line of input by calling get_from_line and stores output lines by }
{ calling write_to_line.                                                      }

procedure text_format(old_left_margin: integer;                                 { TEXT_FORMAT }
                      function get_from_line(
                         var from_line: line_chars_type;
                         var from_length: line_chars0_range): boolean;
                      procedure write_to_line(
                         to_line: line_chars_type;
                         to_length: line_chars0_range) );

var
   temp_style_rec: style_type;
   visi_char, verbatim_char: char;
   from_line, to_line, next_line: line_chars_type;
   next_already_read, auto_process, start_at_right: boolean;
   to_words, blanks_skipped, lindex: integer;
   from_col, to_col, next_col, word_end: line_chars_range;
   from_length, next_length: line_chars0_range;

   { process visible spaces and call write procedure }

   procedure end_to_line(to_length: line_chars0_range);

   var
      lindex: line_chars_range;

   begin { end_to_line }
      if visi_char <> ' ' then
      begin { process visible spaces }
         for lindex := 1 to to_length
         do if to_line[lindex] = visi_char
            then to_line[lindex] := ' '
      end { process visible spaces } ;
      to_line[to_length + 1] := newline;
      write_to_line(to_line, to_length);
      to_col := 1;
      to_words := 0
   end { end_to_line } ;

   { store a character in the new line }

   procedure store_char(the_char: char);

   begin { store_char }
      if to_col < line_max_chars then
      begin
         to_line[to_col] := the_char;
         to_col := to_col + 1
      end
   end { store_char } ;

   { start off the output line at the correct margin }

   procedure empty_to_line(start_col: line_chars_range);

   var
      index: integer;

   begin { empty_to_line }
      to_col := 1;
      for index := 1 to (start_col - 1)
      do store_char(' ');
      to_words := 0
   end { empty_to_line } ;

   { Look at the first two lines of the new paragraph (if possible) }
   { to decide how it is laid out }

   procedure decide_para_type(var end_found: boolean);

   var
      all_skipped, hit_end, is_verbatim: boolean;
      lindex: integer;

      { Set new values for style record, but only in automatic case }

      procedure update_parameters(left, indent: integer);

      var
         cindex: line_chars_range;

      begin { update_parameters }
         if auto_process then
         with temp_style_rec do
         begin
            old_left_margin := left;
            left_margin := left;
            para_indent := indent;
            para_blanks := blanks_skipped;
            { replace any spaces in 'hanging' part of first line by }
            { visible space symbols }
            if para_indent < 0 then
            begin { hanging indent }
               for cindex := left_margin+para_indent to left_margin-1
               do if from_line[cindex] = ' '
                  then from_line[cindex] := visi_char
            end { hanging indent }
         end { with } ;
         blanks_skipped := 0
      end { update_parameters } ;

      { The next paragraph is only a single line:                  }
      { If the first line looks the same as the previous paragraph }
      { Then leave parameters alone                                }
      { Else assume non-indented paragraph                         }

      procedure single_line_paragraph;

      var
         cindex: line_chars_range;

      begin { single_line_paragraph }
         with temp_style_rec do
         begin
            if from_col = left_margin + para_indent
            then update_parameters(left_margin, para_indent)
            else update_parameters(from_col, 0);
            { If the single line already fits into the width then fill in }
            { with visible space characters so that it is copied verbatim }
            if from_length <= right_margin
            then for cindex := left_margin+para_indent to right_margin
                 do if from_line[cindex] = ' '
                    then from_line[cindex] := visi_char
         end { with }
      end { single_line_paragraph } ;

   begin { decide_para_type }
      end_found := false;               { true if all text consumed }
      { skip any blank lines }
      blanks_skipped := 0;
      all_skipped := false;
      hit_end := false;
      repeat
         if from_length = 0 then
         begin { blank line }
            blanks_skipped := blanks_skipped + 1;
            if get_from_line(from_line, from_length) then
            begin { read another }
               if (from_line[1] = verbatim_char) and (from_length >= 1) then
               begin { copy verbatim }
                  { Append any blanks skipped }
                  for lindex := 1 to blanks_skipped
                  do end_to_line(0);
                  blanks_skipped := 0;
                  repeat
                     write_to_line(from_line, from_length);
                     hit_end := not get_from_line(from_line, from_length);
                     is_verbatim := (from_line[1] = verbatim_char) and
                                    (from_length >= 1)
                  until (not is_verbatim) or hit_end
               end { copy verbatim } ;
               if not hit_end then
               begin { not all skipped }
                  from_col := 1;
                  while from_line[from_col] = ' '
                  do from_col := from_col + 1
               end { not all skipped }
            end { read another }
            else hit_end := true
         end { blank line }
         else all_skipped := true
      until all_skipped or hit_end;

      with temp_style_rec do
      if all_skipped then
      begin { first line found }
         if get_from_line(next_line, next_length) then
         begin { second line found }
            next_already_read := true;
            next_col := 1;
            while next_line[next_col] = ' '
            do next_col := next_col + 1;
            if (next_length = 0) or
               (next_line[1] = verbatim_char)
            then single_line_paragraph
            else
            begin { non-blank second line }
               update_parameters(next_col, from_col-next_col)
            end { non-blank second line }
         end { second line found }
         else single_line_paragraph
      end { first line found }
      else
      begin { hit end while skipping blanks }
         end_found := true
      end { hit end while skipping blanks }
   end { decide_para_type } ;

   { Get the next line of text to be processed }

   function get_next_line(first_time: boolean): boolean;

   var
      line_alright, end_found, is_verbatim: boolean;
      index: integer;

   begin { get_next_line }
      if next_already_read then
      begin { have looked ahead }
         from_line := next_line;
         from_length := next_length;
         next_already_read := false;
         line_alright := true
      end { have looked ahead }
      else line_alright := get_from_line(from_line, from_length);
      if line_alright then
      with temp_style_rec do
      begin
         if (from_line[1] = verbatim_char) and (from_length >= 1) then
         begin { copy line verbatim }
            if to_words > 0
            then end_to_line(to_col - 1);  { output end of last para }
            empty_to_line(temp_style_rec.left_margin);
            repeat
               write_to_line(from_line, from_length);
               line_alright := get_from_line(from_line, from_length);
               is_verbatim := (from_line[1] = verbatim_char) and
                              (from_length >= 1)
            until (not is_verbatim) or (not line_alright)
         end { copy line verbatim } ;
         if line_alright then
         begin { still got a line }
            from_col := 1;
            while from_line[from_col] = ' '
            do from_col := from_col + 1;
            end_found := false;
            if (from_length = 0) or     { blank line or }
               (from_col <> old_left_margin) or
               first_time
            then
            begin { start of new paragraph }
               { output end of last para }
               if to_words > 0
               then end_to_line(to_col - 1);
               { decide what type next para is }
               decide_para_type(end_found);
               if not end_found then
               begin { set up new paragraph }
                  for index := 1 to para_blanks
                  do end_to_line(0);
                  empty_to_line(left_margin + para_indent);
                  start_at_right := true
               end { set up new paragraph }
            end { start of new paragraph } ;
            get_next_line := not end_found
         end { still got a line }
         else get_next_line := false
      end { got line, with }
      else get_next_line := false       { hit .bottom. }
   end { get_next_line } ;

   { Get the boundaries of the next word on the line being processed }

   function get_next_word: boolean;

   var
      state: search_type;

   begin { get_next_word }
      while from_line[from_col] = ' '
      do from_col := from_col + 1;
      state := searching;
      repeat                            { skip to start of word }
         if from_line[from_col] = newline then
         begin { end of from_line }
            if get_next_line(false)
            then                        { moved on OK }
            else state := not_found     { hit .bottom. }
         end { end of from_line }
         else state := found
      until state <> searching;
      if state = found then
      begin
         word_end := from_col;
         while (from_line[word_end] <> ' ') and
               (from_line[word_end] <> newline)
         do word_end := word_end + 1
      end;
      get_next_word := state = found
   end { get_next_word } ;

   { adjust right margin }

   procedure adjust_margin(var to_line: line_chars_type);

   var
      spaces_to_add, gaps, spaces_each, bigger_gaps, bigger_start,
      bigger_stop, gap_index, spaces_index, spaces_to_insert: integer;
      incol, outcol: line_chars_range;

   begin { adjust_margin }
      with temp_style_rec do
      begin
         spaces_to_add := right_margin - to_col + 1;
         gaps := to_words - 1;
         if (spaces_to_add > 0) and     { any adjustment necessary }
            (gaps > 1)                  { and more than one gap }
                                        { padding a single gap looks ugly }
         then
         begin { expand gaps }
            spaces_each := spaces_to_add div gaps;
            bigger_gaps := spaces_to_add mod gaps;
            if start_at_right           { wider gaps at }
            then bigger_start := 1      { right or }
            else bigger_start := gaps - bigger_gaps + 1;  { left }
            bigger_stop := bigger_start + bigger_gaps - 1;
            incol := to_col - 1;
            outcol := right_margin;
            for gap_index := 1 to gaps do
            begin
               while to_line[incol] <> ' ' do
               begin { copy down to gap }
                  to_line[outcol] := to_line[incol];
                  incol := incol - 1;
                  outcol := outcol - 1
               end { copy down to gap } ;
               while to_line[incol] = ' ' do
               begin { copy space or spaces }
                  to_line[outcol] := to_line[incol];
                  incol := incol - 1;
                  outcol := outcol - 1
               end { copy space or spaces } ;
               if (gap_index >= bigger_start) and  { test for wider gap }
                  (gap_index <= bigger_stop)
               then spaces_to_insert := spaces_each + 1
               else spaces_to_insert := spaces_each;
               for spaces_index := 1 to spaces_to_insert do
               begin
                  to_line[outcol] := ' ';
                  outcol := outcol - 1
               end
            end { for loop } ;
            to_col := right_margin + 1
         end { expand gaps } ;
         start_at_right := not start_at_right
      end { with }
   end { adjust_margin } ;

   { add the word into the output line }

   procedure add_in_word;

   var
      spaces, index: integer;

   begin { add_in_word }
      with temp_style_rec do
      begin
         if to_words = 0                { calculate spaces before word }
         then spaces := 0
         else if to_line[to_col-1] in ['.', '!', '?'] then
         begin { two spaces at end of sentence }
            { It is only a sentence if it ends with a lower-case letter }
            if to_col >= 3
            then if to_line[to_col-2] in lower_alphabet
                 then spaces := 2
                 else spaces := 1
            else spaces := 1
         end { two spaces at end of sentence }
         else spaces := 1;
         if ( (right_margin - to_col + 1) >= (word_end - from_col + spaces) ) or
            (to_words = 0)              { long word would not fit }
         then
         begin { there is space in line }
            for index := 1 to spaces
            do store_char(' ');
            for index := from_col to word_end - 1
            do store_char(from_line[index]);
            to_words := to_words + 1;
            from_col := word_end
         end { there is space in line }
         else
         begin { output line is full }
            if adjust
            then adjust_margin(to_line);
            end_to_line(to_col - 1);
            empty_to_line(left_margin);
            add_in_word                 { put it in an empty line }
         end { output line is full }
      end { with }
   end { add_in_word } ;

begin { text_format }
   temp_style_rec := style_rec;
   visi_char := special[svisible];
   verbatim_char := special[sverbatim];
   with temp_style_rec do
   begin
      { Does the user want automatic recognition of paragraph parameters? }
      auto_process := left_margin = 0;
      if auto_process then
      begin { set start-up defaults }
         old_left_margin := 1;
         left_margin := 1;
         para_indent := 0;
         para_blanks := 1;
         { must have visible space character in case of hanging indents }
         if visi_char = ' '
         then visi_char := chr(nul_code)
      end { set start-up defaults }
      else if old_left_margin = 0
      then comm_error('You have not specified the old left-margin column.');
      if main_state = running then
      begin { left margin alright }
         next_already_read := false;
         start_at_right := true;
         blanks_skipped := 0;
         to_words := 0;
         to_col := 1;
         empty_to_line(left_margin);
         { Only assume para start for auto-processing because non-auto may }
         { start in middle of para }
         if get_next_line(auto_process) then
         begin { line got }
            while get_next_word
            do add_in_word;
            if to_words > 0
            then end_to_line(to_col - 1);
            { Append any blanks skipped }
            for lindex := 1 to blanks_skipped
            do end_to_line(0)
         end { line got }
      end { left margin alright }
   end { with }
end { text_format } ;

{ Clear out a chain of macro buffers }

procedure free_mbufs(var ptr: ptr_macro_buffer);                                { FREE_MBUFS }

begin { free_mbufs }
   if ptr <> nil then
   begin
      free_mbufs(ptr^.next);            { do from bottom up }
      dispose(ptr);
      ptr := nil
   end
end { free_mbufs } ;

{ Clear out a complete macro }

procedure free_macro(var ptr: ptr_macro_desc);                                  { FREE_MACRO }

begin { free_macro }
   if ptr <> nil then
   begin
      free_mbufs(ptr^.code_start);
      dispose(ptr);
      ptr := nil
   end
end { free_macro } ;

{ Start off a new macro buffer }

procedure new_macro_buffer(var ptr: ptr_macro_buffer);                          { NEW_MACRO_BUFFER }

begin { new_macro_buffer }
   new(ptr);
   store_macro_buffer := ptr;
   with ptr^ do
   begin
      next := nil;
      code_count := 0
   end
end { new_macro_buffer } ;

{ Store a character code in the temporary macro buffer }

procedure store_macro_code(the_code: integer);                                  { STORE_MACRO_CODE }

begin { store_macro_code }
   if the_code = ord(srestart) + 256
   then { never store "srestart" in macros }
   else with store_macro_buffer^ do
   begin
      if code_count < macro_codes_max then
      begin { there is space in buffer }
         code_count := code_count + 1;
         store_macro_count := store_macro_count + 1;
         macro_text[code_count] := the_code
      end { there is space in buffer }
      else if store_macro_count < macro_total_max then
      begin { start a new buffer }
         new_macro_buffer(next);
         store_macro_code(the_code)     { there should be space now }
      end { start a new buffer }
      else { too many characters! }
   end { with }
end { store_macro_code } ;

{ List a help file, pausing at lines beginning '--' or after 23 lines }

procedure list_help_file(help_filename, second_choice: string;                  { LIST_HELP_FILE }
                         var new_key: long_string);

label 225;

var
   help_file: text;
   help_line: long_string;
   lines_printed: integer;
   reset_worked, done: boolean;

   { Perform RESET on helpfile, allowing for errors }

   procedure perform_reset(the_filename: string; var reset_worked: boolean);

   label 228;

      procedure io_error_onunit(not_used: integer);
      begin { io_error_onunit }
         reset_worked := false;
         goto 228
      end { io_error_onunit } ;

   begin { perform_reset }
      reset_worked := true;
      on('IO_ERROR', io_error_onunit);
      reset(help_file, the_filename);
   228:
   end { perform_reset } ;

   procedure quit_onunit(not_used: integer);
   begin { quit_onunit }
      screen_message('', false);
      screen_message('', false);
      screen_message('Abandoning file listing.', false);
      screen_message('', false);
      goto 225
   end { quit_onunit } ;

   function ask_if_cont(prompt: long_string): boolean;

   var
      answer: long_string;
      cr_given: boolean;

   begin { ask_if_cont }
      message_and_reply(prompt, false, answer, cr_given);
      if cr_given then
      begin
         trim_string(answer);
         upcase_string(answer);
         if answer = ''
         then ask_if_cont := true
         else if length(answer) > 3 then
         begin
            ask_if_cont := false;
            new_key := answer
         end
         else if answer = substr('YES',1,length(answer))
         then ask_if_cont := true
         else if answer = substr('NO ',1,length(answer))
         then ask_if_cont := false
         else
         begin
            ask_if_cont := false;
            new_key := answer
         end
      end { cr_given }
      else
      begin
         ask_if_cont := false;
         new_key := 'QUIT'
      end
   end { ask_if_cont } ;

   procedure pad_to_screenful;

   var
      lindex: integer;

   begin { pad_to_screenful }
      if lines_printed < 22
      then for lindex := lines_printed+1 to 22
           do screen_message('', false)
   end { pad_to_screenful } ;

begin { list_help_file }
   on('QUIT$', quit_onunit);
   perform_reset(help_filename, reset_worked);
   if (not reset_worked) and (length(second_choice) > 0) then
   begin { try second choice }
      help_filename := second_choice;
      perform_reset(help_filename, reset_worked)
   end { try second choice } ;
   if reset_worked then
   begin { file opened alright }
      lines_printed := 0;
      new_key := '';
      done := false;
      repeat
         if eof(help_file)
         then done := true
         else
         begin { not eof }
            readln(help_file, help_line);
            if pos('--', help_line) = 1 then
            begin { pause line found }
               pad_to_screenful;
               done := not ask_if_cont(concat(help_line,' '));
               lines_printed := 0
            end { pause line found }
            else if lines_printed >= 23 then
            begin { screen full }
               done := not ask_if_cont('--more?-- ');
               if not done
               then screen_message(help_line, false);
               lines_printed := 1
            end { screen full }
            else
            begin { screen not full }
               screen_message(help_line, false);
               lines_printed := lines_printed + 1
            end { screen not full }
         end { not eof }
      until done;
      if eof(help_file)
      then pad_to_screenful
   end { file opened alright }
   else
   begin { open failed }
      screen_message('', false);
      str_screen_message('Sorry - Help file ~ is not available.', false,
                         help_filename)
   end { open failed } ;
225:
end { list_help_file } ;

{ Simple help system }

procedure help_system(first_key: long_string);                                  { HELP_SYSTEM }

label 226;

const
   key_file_name = 'HELP_SEARCH_LIST';

var
   key: long_string;
   key_len: integer;
   help_filename, second_choice: string32;
   done, cr_given: boolean;

   function attach_to_help: boolean;

   var
      code: shortint;

      procedure atch$$(ufdnam: char16;
                       namlen, ldisk: shortint;
                       passwd: char6;
                       key: shortint;
                       var code: shortint);
         extern;

   begin { attach_to_help }
      attach_to_help := false;
      atch$$('SYSOVL          ', 6, k$alld, '      ', k$imfd, code);
      if code <> 0
      then error_print(k$irtn, code, 'Attaching to SYSOVL', 'ATCH$$')
      else
      begin
         atch$$('PMED_HELP.UFD   ', 13, k$curr, '      ', k$icur, code);
         if code <> 0
         then error_print(k$irtn, code,
                          'Attaching to SYSOVL>PMED_HELP.UFD', 'ATCH$$')
         else attach_to_help := true
      end
   end { attach_to_help } ;

   function search_key_list(key: long_string;
                            var help_filename,
                                second_choice: string32): boolean;

   var
      key_file: text;
      key_line, alias: long_string;
      suffix: string[10];
      state: (searching, not_found, found);
      matches: boolean;
      comma_pos: integer;

      function check_key(test_key: long_string): boolean;

      begin { check_key }
         if key_len > length(test_key)
         then check_key := false
         else if key_len = length(test_key)
         then check_key := key = test_key
         else check_key := key = substr(test_key, 1, key_len)
      end { check_key } ;

   begin { search_key_list }
      reset(key_file, key_file_name);
      state := searching;
      repeat
         if eof(key_file)
         then state := not_found
         else
         begin
            readln(key_file, key_line);
            comma_pos := pos(',', key_line);
            if comma_pos = 0
            then if check_key(key_line)
                 then state := found
                 else
            else
            begin
               matches := false;
               repeat
                  alias := substr(key_line, 1, comma_pos-1);
                  key_line := substr(key_line, comma_pos+1,
                                     length(key_line)-comma_pos);
                  if check_key(alias)
                  then matches := true;
                  comma_pos := pos(',', key_line);
               until comma_pos = 0;
               if matches
               then state := found
               else if check_key(key_line)
                    then state := found
            end
         end
      until state <> searching;
      if key_line = 'SUMMARY' then
      begin { special case - screen code summary }
         decide_vdu_type;               { in case not known yet }
         str(vdu_type, suffix);
         help_filename := concat('WINDOW_HELP',suffix);
         { if no particuular help for this terminal give general help }
         second_choice := 'SUMMARY.HELP';
         { need to re-attach because GV$SET attaches home }
         search_key_list := attach_to_help
      end { special case - screen code summary }
      else
      begin { normal case }
         help_filename := concat(key_line,'.HELP');
         second_choice := '';
         search_key_list := state = found
      end { normal case }
   end { search_key_list } ;

   procedure onunit(not_used: integer);
   begin { onunit }
      screen_message('', false);
      screen_message('', false);
      screen_message('Leaving help system.', false);
      screen_message('', false);
      goto 226
   end { onunit } ;

begin { help_system }
   in_help_system := true;
   continue_screen_messages := true;
   on('ERROR_TRAP', onunit);
   on('QUIT$', onunit);
   if attach_to_help then
   begin
      key := first_key;
      trim_string(key);
      done := false;
      repeat
         if key = ''
         then key := 'INDEX';
         key_len := length(key);
         if search_key_list(key, help_filename, second_choice)
         then list_help_file(help_filename, second_choice, key)
         else
         begin
            str_screen_message('Help key ~ not found.', false, key);
            key := ''
         end;
         if key = '' then
         begin { the user just pressed RETURN }
            message_and_reply(
               'Type a keyword, QUIT to finish, or INDEX for a list: ',
               false, key, cr_given);
            if cr_given then
            begin
               trim_string(key);
               upcase_string(key)
            end
            else key := 'QUIT'
         end { the user just pressed RETURN } ;
         key_len := length(key);
         if (key_len >= 1) and (key_len <= 4)
         then if key = substr('QUIT',1,key_len)
              then done := true
      until done
   end;
226:
   attach_home;
   in_help_system := false
end { help_system } ;

{ SPELL command main procedure }

procedure spell_proc(procedure show_line;                                       { SPELL_PROC }
                     procedure ask_user(msg: string; var answer_char: char);
                     procedure show_changed_line);

const
   user_dict_name = 'EDITOR_USER_DICTIONARY';
   blank_spell_word = '                                ';

type
   spell_word_chars = packed array[1..spell_word_max] of char;
   spell_word_range = 1..spell_word_max;
   spell_word_both = record
      case boolean of
         false: (len: shortint; chars: spell_word_chars);
         true:  (strng: spell_word_string)
      end;

   ptr_glossary_entry = ^ glossary_entry;
   glossary_entry = record
      next, previous: ptr_glossary_entry;
      groot: spell_word_string;
      gsuffix: suffix_range
   end;

   ptr_word_entry = ^ word_entry;
   word_entry = record
      left, right: ptr_word_entry;
      eroot: spell_word_string;
      esuffix: suffix_range;
      estate: (new_word, in_dict, word_skipped)
   end;

var
   spell_state: (checking_words, all_words_checked, user_interrupt,
                 new_word_found);
   spell_word, upper_spell_word: spell_word_string;
   spell_word_start, spell_word_end: integer;  { cols within line }
   spell_word_count, spell_diff_count,
   comm_count, blank_count, text_count: integer;
   first_gloss, current_gloss: ptr_glossary_entry;
   spell_heads: array[spell_alphabet, spell_alphabet] of ptr_word_entry;
   block_sub: 0..block_bytes_max;
   dict_unit: shortint;
   offsets: array[spell_alphabet, spell_alphabet] of integer;
   input_text_file: text;
   low_word, high_word: spell_word_string;
   found_word_ptr: ptr_word_entry;
   new_glosses: integer;
   glossary_location: (file_glossary, local_glossary, origin_glossary,
                       no_glossary);
   gloss_pathname: tree_string;
   temp_dict_block: dict_ptr;

{ Decide what type a character is, allowing for parity possibly off }

function get_char_type(ch: char): char_type;                                    { GET_CHAR_TYPE }

begin { get_char_type }
   if ord(ch) < 200b
   then get_char_type := punct
   else get_char_type := spell_char_array[ord(ch)]
end { get_char_type } ;

{ Compare two root/suffix pairs, returning -1 0 1 as appropriate }

function compare_dict(root1: spell_word_string; suffix1: suffix_range;          { COMPARE_DICT }
                      root2: spell_word_string; suffix2: suffix_range): integer;

begin { compare_dict }
   if root1 < root2
   then compare_dict := -1
   else if root1 > root2
   then compare_dict := +1
   else if suffix1 < suffix2
   then compare_dict := -1
   else if suffix1 > suffix2
   then compare_dict := +1
   else compare_dict := 0
end { compare_dict } ;

{ Set up the glossary as empty }

procedure init_glossary;                                                        { INIT_GLOSSARY }

begin { init_glossary }
   new(first_gloss);
   with first_gloss^ do
   begin
      groot := low_word;
      gsuffix := suffix_range.min;
      previous := nil;
      new(next)
   end;
   with first_gloss^.next^ do
   begin
      groot := high_word;
      gsuffix := suffix_range.max;
      previous := first_gloss;
      next := nil
   end;
   current_gloss := first_gloss
end { init_glossary } ;

{ Insert a gloss into the glossary }

procedure insert_gloss(root: spell_word_string; suffix: suffix_range);          { INSERT_GLOSS }

var
   comp: integer;
   last, temp_ptr: ptr_glossary_entry;

begin { insert_gloss }
   with current_gloss^
   do comp := compare_dict(root, suffix, groot, gsuffix);
   if comp > 0 then
   begin { move forwards }
      repeat
         last := current_gloss;
         current_gloss := current_gloss^.next;
         with current_gloss^
         do comp := compare_dict(root, suffix, groot, gsuffix)
      until comp <= 0
   end { move forwards }
   else if comp = 0
   then
   else
   begin { move backwards }
      repeat
         last := current_gloss;
         current_gloss := current_gloss^.previous;
         with current_gloss^
         do comp := compare_dict(root, suffix, groot, gsuffix)
      until comp >= 0;
      temp_ptr := current_gloss;        { swap current_gloss before last }
      current_gloss := last;
      last := temp_ptr
   end { move backwards } ;
   if comp = 0
   then { ignore duplicate }
   else
   begin { add in new word }
      new(temp_ptr);
      temp_ptr^.next := current_gloss;
      temp_ptr^.previous := last;
      current_gloss^.previous := temp_ptr;
      last^.next := temp_ptr;
      current_gloss := temp_ptr;
      current_gloss^.groot := root;
      current_gloss^.gsuffix := suffix
   end { add in new word }
end { insert_gloss } ;

{ Scan the glossary for a word }

function scan_glossary(ch1, ch2: char;                                          { SCAN_GLOSSARY }
                       root: spell_word_string;
                       suffix: suffix_range): boolean;

var
   state: search_type;
   comp: integer;

begin { scan_glossary }
   root := concat('  ',root);
   root[1] := ch1;
   root[2] := ch2;
   state := searching;
   repeat
      with current_gloss^
      do comp := compare_dict(root, suffix, groot, gsuffix);
      if comp < 0
      then state := not_found
      else if comp = 0
      then state := found
      else current_gloss := current_gloss^.next
   until state <> searching;
   scan_glossary := state = found
end { scan_glossary } ;

{ Clear out the glossary }

procedure dispose_glossary;                                                     { DISPOSE_GLOSSARY }

var
   temp_ptr: ptr_glossary_entry;

begin { dispose_glossary }
   current_gloss := first_gloss;
   while current_gloss <> nil do
   with current_gloss^ do
   begin
      temp_ptr := current_gloss^.next;
      dispose(current_gloss);
      current_gloss := temp_ptr
   end { while loop } ;
   first_gloss := nil
end { dispose_glossary } ;

{ Close the dictionary file }

procedure close_dict;                                                           { CLOSE_DICT }

var
   code: shortint;

begin { close_dict }
   close_unit(dict_unit, code);
   dict_unit := 0;
   if code <> 0
   then error_print(k$irtn, code, 'Closing dictionary file', 'SRSFX$')
end { close_dict } ;

{ Open the dictionary file for reading or writing }

procedure open_dict(key: shortint; dict_pathname: string);                      { OPEN_DICT }

var
   tipe, rnw, code: shortint;
   position: integer;

begin { open_dict }
   call_srsfx$(key+k$ndam+k$getu, dict_pathname, dict_unit, tipe, code);
   if code = 0 then
   begin { opened alright }
      if key = k$writ then
      begin
         position := 0;
         prwf$$(k$trnc, dict_unit, 0, 0, position, rnw, code);
         if code <> 0 then
         begin
            error_print(k$irtn, code, 'Truncating', 'PRWF$$');
            close_dict
         end { truncate error }
      end { key = k$writ }
   end { opened alright }
   else error_print(k$irtn, code, dict_pathname, 'SRSFX$')
end { open_dict } ;

{ Set the dictionary file to a given position }

procedure set_position(new_position: integer);                                  { SET_POSITION }

var
   rnw, code: shortint;

begin { set_position }
   prwf$$(k$posn+k$prea, dict_unit, 0, 0, new_position, rnw, code);
   if code <> 0
   then error_print(k$irtn, code, 'Setting position', 'PRWF$$')
end { set_position } ;

{ Read words from the dictionary file }

procedure read_block(block_address, block_length: integer;                      { READ_BLOCK }
                     var word_count: integer);

var
   position: integer;
   nw, rnw, code: shortint;

begin { read_block }
   position := 0;
   { allow for the fact that nw is really 0..65535 }
   if block_length > 32767
   then nw := block_length - 65536
   else nw := block_length;
   prwf$$(k$read, dict_unit, block_address, nw, position, rnw, code);
   if code = e$eof
   then code := 0;
   if code = 0 then
   begin { read ok }
      { rnw is also 0..65535 }
      if rnw < 0
      then word_count := 65536 + rnw
      else word_count := rnw;
      if word_count = 0
      then abort('Reading empty dictionary block.')
   end { read ok }
   else error_print(k$irtn, code, 'Reading dictionary file', 'PRWF$$')
end { read_block } ;

{ Split a word into root and suffix }

procedure strip_suffix(the_word: spell_word_string;                             { STRIP_SUFFIX }
                       var ch1, ch2: char;
                       var the_root: spell_word_string;
                       var the_suffix: suffix_range);

var
   col, suff_len: integer;
   state, inner_state: search_type;
   sindex: suffix_range;
   test_suffix: spell_word_string;

begin { strip_suffix }
   ch1 := the_word[1];
   ch2 := the_word[2];
   if length(the_word) <= 3 then
   begin
      the_root := substr(the_word, 3, length(the_word)-2);
      the_suffix := null_suffix
   end
   else
   begin { at least 4 characters }
      col := 4;
      state := searching;
      repeat
         suff_len := length(the_word) - col + 1;
         if suff_len < 1
         then state := not_found
         else if suff_len > 8
         then                           { > max suffix length }
         else with suffix_map[suff_len, the_word[col] ] do
         begin { suffix length alright }
            if suffix_count > 0 then
            begin
               sindex := suffix_start;
               test_suffix := substr(the_word, col, suff_len);
               inner_state := searching;
               repeat
                  if test_suffix = suffixes[sindex]
                  then inner_state := found
                  else if test_suffix < suffixes[sindex]
                  then inner_state := not_found
                  else if sindex >= suffix_start + suffix_count
                  then inner_state := not_found
                  else sindex := sindex + 1
               until inner_state <> searching
            end { suffix_count > 0 }
            else inner_state := not_found;
            if inner_state = found
            then state := found
         end { suffix length alright } ;
         if state = searching
         then col := col + 1
      until state <> searching;
      if state = found then
      begin { suffix identified }
         the_root := substr(the_word, 3, col-3);
         the_suffix := sindex
      end { suffix identified }
      else
      begin { suffix not identified }
         the_root := substr(the_word, 3, length(the_word)-2);
         the_suffix := null_suffix
      end { suffix not identified }
   end { at least 4 characters }
end { strip_suffix } ;

{ Get the next byte from the dictionary }

procedure get_byte(var byte: smallint);                                         { GET_BYTE }

begin { get_byte }
   block_sub := block_sub + 1;
   byte := temp_dict_block^[block_sub]
end { get_byte } ;

{ Unpack a suffix from the dictionary }

procedure unpack_suffix(var the_suffix: spell_word_string);                     { UNPACK_SUFFIX }

var
   temp: spell_word_both;
   byte: smallint;

begin { unpack_suffix }
   with temp do
   begin
      get_byte(byte);
      len := 0;
      while byte < suffix_range.min do
      begin
         len := len + 1;
         chars[len] := chr(byte + 128);
         get_byte(byte)
      end;
      the_suffix := strng
   end { with }
end { unpack_suffix } ;

{ Check this file line for blank/command/normal status and count }

procedure check_this_line;                                                      { CHECK_THIS_LINE }

begin { check_this_line }
   with current_buffer do
   begin
      if (line_status = normal) and (line_number <= spell_stop_line) then
      begin { not eof etc. }
         line_position := 1;
         if line_length = 0
         then blank_count := blank_count + 1
         else if line_text[1] = special[swpc] then
         begin { begins with '.' - skip text }
            line_position := line_length + 1;
            comm_count := comm_count + 1
         end { begins with '.' - skip text }
         else text_count := text_count + 1
      end { not eof etc. }
   end { with }
end { check_this_line } ;

{ Split off the next word from the input file }

function next_word(read_from_file: boolean;                                     { NEXT_WORD }
                   var cont_extra: boolean): boolean;

var
   state, inner_state: search_type;
   done: boolean;
   ch: char;
   ch_type: char_type;
   temp, upper_temp: record
      case boolean of
         false: (len: shortint; chars: spell_word_chars);
         true:  (strng: spell_word_string)
      end;

   { Get next character from Primos file or editor buffer }

   function next_char(var the_char: char): boolean;

   begin { next_char }
      if read_from_file then
      begin { read from input_text_file }
         if eof(input_text_file)
         then next_char := false
         else
         begin { not eof }
            next_char := true;
            if eoln(input_text_file) then
            begin { end of line }
               the_char := ' ';
               readln(input_text_file)
            end { end of line }
            else read(input_text_file, the_char)
         end { not eof }
      end { read from input_text_file }
      else with current_buffer do
      begin { get from editor buffer }
         if (line_status <> normal) or
            (line_number > spell_stop_line)
         then next_char := false
         else
         begin { not eof }
            next_char := true;
            { Allow the linefeed to be returned as a normal punct char so that }
            { we stay on same line at end of word }
            if line_position > line_length+1 then
            begin { end of line }
               the_char := ' ';
               load_a_line(line_number+1, false);
               check_this_line
            end { end of line }
            else
            begin { normal character }
               the_char := line_text[line_position];
               line_position := line_position + 1
            end { normal character }
         end { not eof }
      end { get from editor buffer }
   end { next_char } ;

begin { next_word }
   temp.len := 0;
   upper_temp.len := 0;
   cont_extra := false;
   state := searching;
   repeat
      { skip punctuation }
      inner_state := searching;
      repeat
         if next_char(ch) then
         begin { got char }
            if get_char_type(ch) = punct
            then { carry on }
            else inner_state := found
         end { got char }
         else inner_state := not_found
      until inner_state <> searching;

      { split off the next word }

      if inner_state = not_found
      then state := not_found
      else
      begin { start of word found }
         if not read_from_file
         then spell_word_start := current_buffer.line_position - 1;
         done := false;
         repeat
            ch_type := get_char_type(ch);
            if ch_type = punct
            then done := true
            else
            begin { not punctuation }
               if ch_type = extra
               then cont_extra := true;
               if temp.len < spell_word_max then
               begin { not too long }
                  temp.len := temp.len + 1;
                  temp.chars[temp.len] := ch;
                  if ch_type = lower
                  then ch := chr(ord('A') - ord('a') + ord(ch));
                  upper_temp.len := upper_temp.len + 1;
                  upper_temp.chars[upper_temp.len] := ch
               end { not too long }
               else { skip end of long word } ;
               if next_char(ch)
               then { keep looking }
               else done := true
            end { not punctuation }
         until done;
         state := found;
         spell_word_count := spell_word_count + 1
      end { not eof }
   until state <> searching;
   spell_word := temp.strng;
   upper_spell_word := upper_temp.strng;
   next_word := state = found
end { next_word } ;

{ Get the next word that does not contain numbers etc. }

function next_proper_word(read_from_file: boolean): boolean;                    { NEXT_PROPER_WORD }

var
   state: search_type;
   cont_extra: boolean;

begin { next_proper_word }
   state := searching;
   repeat
      if next_word(read_from_file, cont_extra)
      then if cont_extra
           then { ignore because of special chars }
           else if length(spell_word) < 2
           then { ignore because short }
           else state := found
      else state := not_found
   until state <> searching;
   next_proper_word := state = found
end { next_proper_word } ;

{ Read the list of suffixes from either the compressed dictionary or the }
{ original source file }

procedure read_suffixes;                                                        { READ_SUFFIXES }

var
   done: boolean;
   suffix_no: suffix_range;
   inline: spell_word_string;
   len: 1..8;
   letter: spell_alphabet;

begin { read_suffixes }
   suffix_no := first_suffix;
   for len := 1 to 8
   do for letter := 'A' to 'Z'
      do with suffix_map[len, letter] do
      begin
         suffix_start := 0;
         suffix_count := 0
      end;
   done := false;
   repeat
      unpack_suffix(inline);
      if length(inline) > 8
      then writeln('Suffix "',inline,'" is too long.')
      else
      begin
         suffixes[suffix_no] := inline;
         if length(inline) > 0 then
         begin
            with suffix_map[length(inline), inline[1] ] do
            begin
               if suffix_start = 0
               then suffix_start := suffix_no;
               suffix_count := suffix_count + 1
            end { with } ;
            suffix_no := suffix_no + 1
         end
         else done := true
      end { valid suffix }
   until done;
   null_suffix := suffix_no
end { read_suffixes } ;

{ Read the text of the dictionary into memory }

procedure read_dictionary;                                                      { READ_DICTIONARY }

var
   ch1, ch2: char;
   dindex: 0..n_dict_blocks;
   all_block_processed, all_read: boolean;
   first_offset, last_offset, this_offset, word_count: integer;

begin { read_dictionary }
   ch1 := 'A';
   ch2 := 'A';
   dindex := 0;
   all_read  := false;
   repeat
      if dindex = n_dict_blocks then
      begin { out of blocks }
         screen_message('Dictionary is too big to fit into memory.', true);
         close_dict;
         all_read := true
      end { out of blocks }
      else
      begin { block no alright }
         dindex := dindex + 1;
         {---teststart---}
         if mode_test
         then writeln('Dictionary block ',dindex:1,' starts with ',ch1,ch2);
         {---teststop---}
         new(dict_blocks[dindex]);
         first_offset := offsets[ch1, ch2];
         set_position(first_offset);
         read_block(addressof(dict_blocks[dindex]^),
                    sizeof   (dict_blocks[dindex]^),
                    word_count);
         {---teststart---}
         if mode_test
         then writeln('Dictionary block ',dindex:1,' contains ',word_count:1,
                      ' words');
         {---teststop---}
         last_offset := first_offset + word_count;
         all_block_processed := false;
         repeat
            this_offset := offsets[ch1, ch2];
            if this_offset >= last_offset then
            begin { off end of block }
               if this_offset = last_offset
               then { fitted exactly }
               else
               begin { need to step backwards }
                  { the last dictionary section starts in the block but is }
                  { not complete: step back to last non-null entry and re-read }
                  { from there }
                  repeat
                     if ch2 > 'A'
                     then ch2 := pred(ch2)
                     else
                     begin
                        ch1 := pred(ch1);
                        ch2 := 'Z'
                     end
                  until offset_table[ch1, ch2].any_words
               end { need to step backwards } ;
               all_block_processed := true
            end { off end of block }
            else
            begin { within block }
               with offset_table[ch1, ch2] do
               begin
                  if this_offset < first_offset
                  then any_words := false  { no text for this section }
                  else
                  begin { text exists }
                     any_words := true;
                     dict_block_no := dindex;
                     dict_block_offset := this_offset - first_offset
                  end { text exists }
               end { with } ;
               if ch2 < 'Z'
               then ch2 := succ(ch2)
               else if ch1 < 'Z' then
               begin
                  ch1 := succ(ch1);
                  ch2 := 'A'
               end
               else
               begin { reached ZZ }
                  all_block_processed := true;
                  all_read := true
               end { reached ZZ }
            end { within block }
         until all_block_processed
      end { block no alright }
   until all_read
end { read_dictionary } ;

{ Read a glossary if possible.  Try in order: }
{    =.glossary                               }
{    editor_user_dictionary in local dir      }
{    editor_user_dictionary in origin dir     }

procedure read_glossary;                                                        { READ_GLOSSARY }

var
   reset_worked: boolean;
   code: shortint;
   ch1, ch2, ch: char;
   root: spell_word_string;
   suffix: suffix_range;
   paren_pos, last_dot: integer;
   chindex: tree_range;

   function try_to_reset(the_pathname: tree_string): boolean;

   label 1752;

      procedure io_error_onunit(cf_ptr: integer);

      begin { io_error_onunit }
         try_to_reset := false;
         goto 1752
      end { io_error_onunit } ;

   begin { try_to_reset }
      try_to_reset := true;
      on('IO_ERROR', io_error_onunit);
      reset(input_text_file, the_pathname);
   1752:
   end { try_to_reset } ;

begin { read_glossary }
   { construct glossary pathname as =.GLOSSARY }
   gloss_pathname := current_buffer.main_file.treename.strng;
   if length(gloss_pathname) > 0 then
   begin { filename not null }
      { trim off at '(' if present }
      paren_pos := pos('(', gloss_pathname);
      if paren_pos > 0
      then delete(gloss_pathname, paren_pos,
                  length(gloss_pathname) - paren_pos + 1);
      { trim off at last dot if present }
      last_dot := 0;
      for chindex := 1 to length(gloss_pathname) do
      begin
         ch := gloss_pathname[chindex];
         if ch = '.'
         then last_dot := chindex
         else if ch = '>'
         then last_dot := 0             { disallow dots in UFD names }
      end { for loop } ;
      if last_dot > 0
      then delete(gloss_pathname, last_dot,
                  length(gloss_pathname) - last_dot + 1);
      if length(gloss_pathname) + 9 <= tree_max_chars
      then insert('.GLOSSARY', gloss_pathname, length(gloss_pathname)+1)
      else gloss_pathname := ''         { resultant pathname too long }
   end { filename not null } ;
   {---teststart---}
   if mode_test
   then writeln('gloss_pathname = "',gloss_pathname,'"');
   {---teststop---}
   reset_worked := false;
   if length(gloss_pathname) > 0 then
   begin { try for file glossary }
      reset_worked := try_to_reset(gloss_pathname);
      if reset_worked then
      begin { file glossary found }
         glossary_location := file_glossary;
         str_screen_message('Reading document glossary from ~', true,
                            gloss_pathname)
      end { file glossary found }
   end { try for file glossary } ;
   if not reset_worked then
   begin { try local }
      reset_worked := try_to_reset(user_dict_name);
      if reset_worked then
      begin { local found }
         glossary_location := local_glossary;
         screen_message('Reading user dictionary in current directory.', true)
      end { local found }
   end { try local } ;
   if not reset_worked then
   begin { try in origin }
      at$or(k$setc, code);
      if code = 0
      then reset_worked := try_to_reset(user_dict_name);
      attach_home;
      if reset_worked then
      begin { found in origin }
         glossary_location := origin_glossary;
         screen_message('Reading user dictionary in origin directory.', true)
      end { found in origin }
      else glossary_location := no_glossary
   end { try in origin } ;
   if reset_worked then
   begin { managed to find file }
      while next_proper_word(true) do
      begin
         strip_suffix(upper_spell_word, ch1, ch2, root, suffix);
         root := concat('  ',root);
         root[1] := ch1;
         root[2] := ch2;
         insert_gloss(root, suffix)
      end;
      close(input_text_file)
   end { managed to find file } ;
   current_gloss := first_gloss
end { read_glossary } ;

{ Rewrite glossary either locally or in origin }

procedure write_glossary;                                                       { WRITE_GLOSSARY }

label 1989;

var
   glossfile: text;
   gloss_fn: tree_string;
   code: shortint;

   procedure error_trap_onunit(cf_ptr: integer);

   begin { error_trap_onunit }
      goto 1989
   end { error_trap_onunit } ;

begin { write_glossary }
   on('ERROR_TRAP', error_trap_onunit);
   gloss_fn := user_dict_name;
   case glossary_location of
      file_glossary:
         begin
            gloss_fn := gloss_pathname;
            message := concat('Alright to update glossary file ',gloss_pathname,
                              '? ')
         end;
      local_glossary:
         message := 'Alright to update user dictionary in current directory? ';
      origin_glossary:
         message := 'Alright to update user dictionary in origin directory? ';
      no_glossary:
         message := 'Alright to create user dictionary in origin directory? '
   end { case } ;
   if question_and_answer(message, true, true) then
   begin { alright to update }
      if glossary_location in [origin_glossary, no_glossary]
      then at$or(k$setc, code);
      rewrite(glossfile, gloss_fn);
      attach_home;
      current_gloss := first_gloss^.next;
      while current_gloss^.next <> nil do
      with current_gloss^ do
      begin
         writeln(glossfile,groot,suffixes[gsuffix]);
         current_gloss := next;
      end { while loop } ;
   1989:
   end { alright to update }
end { write_glossary } ;

{ Open dictionary for reading and read standard parts }

procedure online_dictionary;                                                    { ONLINE_DICTIONARY }

const
   suffix_block_size = 500;             { max words total for suffixes }

var
   block_words: integer;

begin { online_dictionary }
   open_dict(k$read, 'SYSOVL>NEW_DICT.DICT');
   if dict_unit <> 0 then
   begin { opened alright }
      read_block(addressof(offsets), sizeof(offsets), block_words);
      new(temp_dict_block);
      read_block(addressof(temp_dict_block^), suffix_block_size, block_words);
      block_sub := 0;
      read_suffixes;
      dispose(temp_dict_block);
      read_dictionary;
      if dict_unit <> 0 then
      begin { all alright }
         close_dict;
         dictionary_online := true
      end { all alright }
   end { opened alright }
end { online_dictionary } ;

   { Scan a section of the dictionary looking for a word }

   function scandict(root: spell_word_string;
                     suffix: suffix_range;
                     the_block: dict_block;
                     block_offset: integer): boolean;
      extern;

{ Check whether a single word is in the dictionary }

function check_word(ch1, ch2: char;                                             { CHECK_WORD }
                    root: spell_word_string;
                    suffix: suffix_range): boolean;

var
   in_dict: boolean;

begin { check_word }
   with offset_table[ch1, ch2] do
   begin
      if any_words
      then in_dict := scandict(root, suffix, dict_blocks[dict_block_no]^,
                               dict_block_offset)
      else in_dict := false
   end { with } ;
   if in_dict
   then check_word := true
   else
   begin { try glossary }
      current_gloss := first_gloss;
      check_word := scan_glossary(ch1, ch2, root, suffix)
   end { try glossary }
end { check_word } ;

{ Start of spell checking session; open dictionary etc. }

procedure start_spell;                                                          { START_SPELL }

var
   windex: spell_word_range;
   ch1, ch2: char;

begin { start_spell }
   dict_unit := 0;
   if dictionary_online
   then screen_message('Dictionary is already online.', true)
   else
   begin { need to read dictionary }
      screen_message('Reading dictionary.', true);
      online_dictionary
   end { need to read dictionary } ;
   if dictionary_online then
   begin { dictionary found }
      low_word := blank_spell_word;
      for windex := 1 to spell_word_max
      do low_word[windex] := char.min;
      high_word := blank_spell_word;
      for windex := 1 to spell_word_max
      do high_word[windex] := char.max;
      new_glosses := 0;
      init_glossary;
      read_glossary;
      for ch1 := 'A' to 'Z' do
      begin
         for ch2 := 'A' to 'Z'
         do spell_heads[ch1, ch2] := nil
      end { for loop } ;
      spell_word_count := 0;
      spell_diff_count := 0;
      comm_count := 0;
      blank_count := 0;
      text_count := 0;
      check_this_line
   end { dictionary found }
end { start_spell } ;

{ Run through file from current location forwards checking each word found }

procedure run_spell;                                                            { RUN_SPELL }

var
   word_is_alright: boolean;
   ch1, ch2: char;
   root: spell_word_string;
   suffix: suffix_range;

   { Add a word to the word tree }

   procedure add_word_to_tree(var ptr: ptr_word_entry);

   var
      comp: integer;

   begin { add_word_to_tree }
      if ptr = nil then
      begin { add it here }
         new(ptr);
         with ptr^ do
         begin
            left := nil; right := nil;
            eroot := root;
            esuffix := suffix;
            estate := new_word;
            found_word_ptr := ptr
         end { with } ;
         spell_diff_count := spell_diff_count + 1
      end { ptr = nil }
      else with ptr^ do
      begin { not nil }
         comp := compare_dict(root, suffix, eroot, esuffix);
         if comp < 0
         then add_word_to_tree(left)
         else if comp > 0
         then add_word_to_tree(right)
         else found_word_ptr := ptr     { word already in tree }
      end { ptr <> nil }
   end { add_word_to_tree } ;

begin { run_spell }
   spell_state := checking_words;
   repeat
      if main_state <> running
      then spell_state := user_interrupt
      else if next_proper_word(false) then
      begin { got word from text }
         strip_suffix(upper_spell_word, ch1, ch2, root, suffix);
         add_word_to_tree(spell_heads[ch1, ch2]);
         with found_word_ptr^ do
         begin
            if estate = new_word then
            begin { first time for this word }
               word_is_alright := check_word(ch1, ch2, root, suffix);
               if word_is_alright
               then estate := in_dict
            end { first time for this word } ;
            if estate = new_word then
            with current_buffer do
            begin
               spell_state := new_word_found;
               spell_word_end := line_position;
               line_position := spell_word_start
            end
         end { with }
      end { got word from text }
      else spell_state := all_words_checked
   until spell_state <> checking_words
end { run_spell } ;

{ Deal with the response given by the user }

procedure correct_spell;                                                        { CORRECT_SPELL }

label 1999;

var
   old_length, new_length, length_increase, new_line_length: integer;
   chindex: spell_word_range;
   ch1, ch2: char;
   root: spell_word_string;
   suffix: suffix_range;
   action_key: char;
   reply_text: long_string;
   cr_given: boolean;
   replacement_word: spell_word_string;

   { Move "wd_count" words from "from_string[fromsub]" to "to_string[tosub]".
     Preserves the words moved by copying either forwards or backwards. }

   procedure movewds(from_string: line_chars_type;
                     fromsub: integer;
                     var to_string: line_chars_type;
                     tosub: integer;
                     wd_count: integer);
      extern;                           { written in PMA }

   procedure quit_onunit(cf_ptr: integer);

   begin { quit_onunit }
      action_key := 'Q';
      spell_state := user_interrupt;
      goto 1999
   end { quit_onunit } ;

begin { correct_spell }
   on('QUIT$', quit_onunit);
   with current_buffer do
   begin
      int_screen_message(concat('Word found on line ~ is "',spell_word,'"'),
                         true, current_buffer.line_number, 1);
      repeat ask_user('Add, Skip, Retype or Quit? ', action_key)
      until action_key in ['A', 'S', 'R', 'Q'];
      if action_key = 'R' then
      begin { get replacement }
         message_and_reply('Please type correct spelling: ', true,
                           reply_text, cr_given);
         if cr_given then
         begin
            if length(reply_text) > spell_word_max
            then replacement_word := substr(reply_text, 1, spell_word_max)
            else replacement_word := reply_text
         end
         else action_key := 'Q'         { CTRL-X }
      end { get replacement } ;
   1999:
      case action_key of
         'A':
            begin { add to glossary }
               found_word_ptr^.estate := in_dict;
               strip_suffix(upper_spell_word, ch1, ch2, root, suffix);
               root := concat('  ',root);
               root[1] := ch1;
               root[2] := ch2;
               insert_gloss(root, suffix);
               new_glosses := new_glosses + 1;
               line_position := spell_word_end
            end { add to glossary } ;
         'S':
            begin { skip word }
               line_position := spell_word_end;
               found_word_ptr^.estate := word_skipped
            end { skip word } ;
         'R':
            with current_buffer do
            begin { replace with new word given by user }
               old_length := spell_word_end - 1 - spell_word_start;
               new_length := length(replacement_word);
               length_increase := new_length - old_length;
               new_line_length := line_length + length_increase;
               if new_line_length > line_max_charsm1
               then comm_error('File line too long after substitution.')
               else
               begin { length alright }
                  if length_increase <> 0 then
                  begin { length has changed }
                     movewds(line_text, spell_word_end-1,
                             line_text, spell_word_end-1+length_increase,
                             line_length-spell_word_end+3);
                     line_length := new_line_length
                  end { length has changed }
                  else { same length - no move needed } ;
                  { copy in new value }
                  for chindex := 1 to length(replacement_word)
                  do line_text[spell_word_start + chindex - 1] :=
                     replacement_word[chindex];
                  line_altered := true;
                  { leave line_position so that word is rescanned }
                  show_changed_line
               end { length alright }
            end { replace with new word given by user } ;
         'Q': spell_state := user_interrupt
      end { case }
   end { with }
end { correct_spell } ;

{ End of spelling session: close dictionary etc. }

procedure stop_spell;                                                           { STOP_SPELL }

var
   ch1, ch2: char;

   procedure dispose_tree(var ptr: ptr_word_entry);

   begin { dispose_tree }
      if ptr <> nil then
      with ptr^ do
      begin
         dispose_tree(left);
         dispose_tree(right);
         dispose(ptr);
         ptr := nil
      end { with }
   end { dispose_tree } ;

begin { stop_spell }
   if new_glosses > 0 then
   begin { words added to glossary }
      int_screen_message('~ words added to glossary.', true, new_glosses, 1);
      write_glossary
   end { words added to glossary } ;
   dispose_glossary;
   for ch1 := 'A' to 'Z' do
   begin
      for ch2 := 'A' to 'Z'
      do dispose_tree(spell_heads[ch1, ch2])
   end { for loop } ;
   message := 'End of Spelling check; ~ different words; ~ total words checked.';
   ins_int(spell_diff_count, 1);
   ins_int(spell_word_count, 1);
   screen_message(message, true);
   int_screen_message('~ total lines checked.', true,
                      comm_count+blank_count+text_count, 1);
   if comm_count > 0 then
   begin { wpc lines skipped }
      message := '~ lines beginning with ''~'' omitted.';
      ins_int(comm_count, 1);
      ins_ch(special[swpc]);
      screen_message(message, true)
   end { wpc lines skipped } ;
   if blank_count > 0
   then int_screen_message('~ blank lines.', true, blank_count, 1);
   screen_message('', false)
end { stop_spell } ;

   procedure spell_quit_onunit(cf_ptr: integer);

   begin { spell_quit_onunit }
      spell_state := user_interrupt
   end { spell_quit_onunit } ;

begin { spell_proc }
   on('QUIT$', spell_quit_onunit);
   start_spell;
   if dictionary_online then
   begin { dictionary alright }
      screen_message('Scanning text.', true);
      repeat
         run_spell;
         show_line;
         if spell_state = new_word_found
         then correct_spell
      until spell_state <> new_word_found;
      if spell_state =  user_interrupt
      then screen_message('Terminating spelling check.', true);
      stop_spell
   end { dictionary alright }
   else screen_message('Failed to read dictionary.', true)
end { spell_proc } ;

{ Write the LD-type file for ED -TIDY }

procedure write_ld_file(var ld_filename: long_string);                          { WRITE_LD_FILE }

const
   ld_file_prefix = 'T$PMED_LDFILE.';
   max_entries = 100;
   max_type = 4;

type
   ptr_sort_entry = ^ sort_entry;
   sort_entry = record
      left, right: ptr_sort_entry;
      sort_item: dir_entry
   end;

var
   wild_name, ld_entryname: string32;
   buffno_str: string[10];
   ldfile: text;
   dir_unit, dir_type, ret_code, ents_returned, start_at_top: shortint;
   error_printed, done: boolean;
   type_counts: short4;
   sel_struct: sel_type;
   return_struct: array[1..max_entries] of dir_entry;
   total_count, ent_index, lindex: integer;
   head: ptr_sort_entry;
   sindex: 1..4;

   { Add file to tree }

   procedure add_file_to_tree(the_entry: dir_entry);

      procedure add_to_tree(var ptr: ptr_sort_entry);

      var
         go_left: boolean;

      begin { add_to_tree }
         if ptr = nil then
         begin { end of branch }
            new(ptr);
            with ptr^ do
            begin
               left := nil;
               right := nil;
               sort_item := the_entry
            end { with }
         end { end of branch }
         else with ptr^ do
         begin { entry exists }
            if command_line_args.sort_dtm_sw
            then go_left := the_entry.dtm       > sort_item.dtm
            else go_left := the_entry.entryname < sort_item.entryname;
            if command_line_args.reverse_sw
            then go_left := not go_left;
            if go_left
            then add_to_tree(left)
            else add_to_tree(right)
         end { entry exists }
      end { add_to_tree } ;

   begin { add_file_to_tree }
      if the_entry.entryname = ld_entryname
      then type_counts[1] := type_counts[1] - 1  { omit ld file itself }
      else if (the_entry.entryname = 'MFD') and
              (pos('>MFD', ld_pathname) = length(ld_pathname) - 3) and
              (pos('>'   , ld_pathname) = length(ld_pathname) - 3)
      then type_counts[3] := type_counts[3] - 1  { omit MFD itself if in MFD }
      else add_to_tree(head)
   end { add_file_to_tree } ;

   { copy entries from tree to file }

   procedure copy_tree(var ptr: ptr_sort_entry);

   type
      string28 = string[28];

   var
      day_of_week: shortint;
      visual_date: string28;

      procedure cv$fdv(fs_date: dtm_type;
                       var day_of_week: shortint;
                       var visual_date: string28);
         extern;

   begin { copy_tree }
      with ptr^ do
      begin
         if left <> nil then
         begin { left branch exists }
            copy_tree(left);
            dispose(left)
         end { left branch exists } ;
         cv$fdv(sort_item.dtm, day_of_week, visual_date);
         writeln(ldfile,'  ',sort_item.entryname:-32,'  ',visual_date);
         if right <> nil then
         begin { right branch exists }
            copy_tree(right);
            dispose(right)
         end { right branch exists }
      end { with }
   end { copy_tree } ;

begin { write_ld_file }
   { construct ld_filename as pathname in original directory }
   ld_filename := substr(ld_pathname, 1, original_length);
   str(ld_buffer, buffno_str);
   ld_entryname := concat(ld_file_prefix,buffno_str);
   ld_filename := concat(ld_filename,'>',ld_entryname);
   rewrite(ldfile, ld_filename);
   writeln(ldfile,'L=LD display  I=Inspect  D=Mark for delete  F=Finish and do deletes  Q=Quit');
   writeln(ldfile,'---------------------------------------------------------------------------');
   writeln(ldfile);
   writeln(ldfile,'  ',ld_pathname);
   error_printed := false;
   total_count := 0;
   call_srsfx$(k$read+k$getu, '', dir_unit, dir_type, ret_code);
   if ret_code <> 0 then
   begin
      error_print(k$irtn, ret_code, 'Opening directory', 'SRSFX$');
      abort('Error opening directory')
   end;
   attach_home;
   with sel_struct do                   { fill in common items }
   begin
      version_no := 0;
      wild_ptr := addressof(wild_name);
      wild_name := command_line_args.tidy_wildcard;
      wild_count := 1;
      desired_types.rbf := false;
      desired_types.filler := 0;
      mdb := 0;
      mda := 0;
      svb := 0;
      sva := 0
   end { with } ;
   for sindex := 1 to 4 do
   begin
      with sel_struct.desired_types do
      begin
         dirs        := false;
         seg_dirs    := false;
         files       := false;
         access_cats := false;
         case sindex of
            1: files := true;
            2: seg_dirs := true;
            3: dirs := true;
            4: access_cats := true
         end { case }
      end { with } ;
      head := nil;
      start_at_top := -1;               { = true }
      done := false;
      repeat
         dir$se(dir_unit, dir_type, start_at_top, addressof(sel_struct),
                addressof(return_struct), max_entries,
                sizeof(return_struct[1]), ents_returned,
                type_counts, max_type, ret_code);
         if ret_code = e$eof then
         begin { hit eof }
            ret_code := 0;
            done := true
         end { hit eof } ;
         if ret_code <> 0 then
         begin { error reading dir }
            error_print(k$irtn, ret_code, 'Reading directory', 'DIR$SE');
            close_unit(dir_unit, ret_code);
            abort('Error reading directory')
         end { error reading dir } ;
         start_at_top := 0;             { = false }
         if ents_returned <= 0
         then done := true
         else
         begin { some entries read }
            for ent_index := 1 to ents_returned
            do add_file_to_tree(return_struct[ent_index])
         end { some entries read }
      until done;
      { unload tree of sorted entries }
      if head <> nil then
      begin { entries found }
         writeln(ldfile);
         total_count := total_count + type_counts[sindex];
         write(ldfile,'  ',type_counts[sindex]:1,' ');
         case sindex of
            1: writeln(ldfile,'Files.');
            2: writeln(ldfile,'Segment Directories.');
            3: writeln(ldfile,'Directories.');
            4: writeln(ldfile,'Access Categories.')
         end;
         writeln(ldfile,'  name                              date/time modified');
         writeln(ldfile,'  --------------------------------------------------------------');
         copy_tree(head)
      end { entries found }
   end { for loop } ;
   if total_count = 0 then
   begin { empty dir }
      writeln(ldfile);
      writeln(ldfile,'  No entries in directory.');
      { Write some extra lines because later we will position to the first }
      { file line.                                                         }
      for lindex := 1 to 5
      do writeln(ldfile)
   end { empty dir } ;
   writeln(ldfile);
   close_unit(dir_unit, ret_code)
end { write_ld_file } ;

{ Delete a Primos file }

procedure delete_primos_file(the_filename: long_string);                        { DELETE_PRIMOS_FILE }

var
   code: shortint;

begin { delete_primos_file }
   fil$dl(the_filename, code);
   if code <> 0
   then error_print(k$irtn, code, the_filename, 'FIL$DL')
end { delete_primos_file } ;

{ Expand any abbreviations in a pathname given by the user }

procedure expand_abbrevs(var the_treename: tree_type);                          { EXPAND_ABBREVS }

var
   the_command: long_string;
   code: shortint;

begin { expand_abbrevs }
   with the_treename do
   begin
      if len > 0 then
      begin { treename is not null }
         { precede the pathname with a long command name to ensure it is in    }
         { the argument position and also that the command itself does not get }
         { changed.                                                            }
         the_command := concat('long_command ',strng);
         com$ab(the_command, line_max_chars, code);
         if length(the_command) >= 13
         then delete(the_command, 1, 13)  { remove the command }
         else the_command := '';        { treename has disappeared }
         upcase_string(the_command);
         if length(the_command) <= tree_max_chars
         then strng := the_command
         else strng := substr(the_command, 1, tree_max_chars)
      end { treename is not null }
   end { with }
end { expand_abbrevs } ;

{ Trim spaces from both ends of a string }

procedure trim(var the_string: tree_string);                                    { TRIM }

var
   first_sig, last_sig: integer;
   done: boolean;

begin { trim }
   last_sig := length(the_string);
   done := false;
   repeat
      if last_sig <= 0
      then done := true
      else if the_string[last_sig] = ' '
      then last_sig := last_sig - 1
      else done := true
   until done;
   first_sig := 1;
   done := false;
   repeat
      if first_sig > last_sig
      then done := true
      else if the_string[first_sig] = ' '
      then first_sig := first_sig + 1
      else done := true
   until done;
   if (first_sig > 1) or (last_sig < length(the_string))
   then the_string := substr(the_string, first_sig, last_sig - first_sig + 1)
end { trim } ;

{ Split a string into two parts at the given sequence of characters }

procedure split_at_chars(source, splitter: tree_string;                         { SPLIT_AT_CHARS }
                         var before, after: tree_string);

var
   first_pos, after_start: integer;

begin { split_at_chars }
   first_pos := pos(splitter, source);
   if first_pos = 0 then
   begin { no splitter found }
      before := source;
      after := ''
   end { no splitter found }
   else
   begin { splitter found }
      before := substr(source, 1, first_pos - 1);
      after_start := first_pos + length(splitter);
      if after_start > length(source)
      then after := ''
      else after := substr(source, after_start, length(source)-after_start+1);
      trim(after)
   end { splitter found } ;
   trim(before)
end { split_at_chars } ;

{ Split a string into two parts at a space }

procedure split_at_space(source: tree_string; var before, after: tree_string);  { SPLIT_AT_SPACE }

begin { split_at_space }
   split_at_chars(source, ' ', before, after)
end { split_at_space } ;

{ Expand any iteration, treewalking and wildcards in a pathname. }
{ Expanded trees are returned via the given procedure. }

procedure expand_pathname(original_pathname: tree_string;                       { EXPAND_PATHNAME }
                          var any_expansion: boolean;
                          procedure deliver(expanded_pathname: tree_string;
                                            var continue_expanding: boolean));

var
   continue_expanding, contains_wildcards: boolean;
   left_par_count, right_par_count: integer;
   left_par_pos, right_par_pos: array[1..2] of integer;
   chindex: tree_range;
   ch: char;
   before_iter, iter1, iter2, after_iter,
   temp_iter1, temp_iter2, item1, item2: tree_string;

   { Expand any wildcards in the entry part of a pathname.             }
   { If no wildcards just deliver the unchanged path.                  }
   { We open the directory part by attaching and then opening current, }
   { because otherwise expanding such as CMDNC0>@@.CPL looks for       }
   { CMDNC0 in the current directory.                                  }

   procedure expand_wildcard(dir_part, entry_part: tree_string;
                             want_files: boolean;
                             procedure deliver(expanded_pathname: tree_string;
                                var continue_expanding: boolean));

   const
      max_entries = 100;
      max_type = 4;

   var
      dir_unit, dir_type, ret_code, ents_returned,
      start_at_top: shortint;
      type_counts: short4;
      sel_struct: sel_type;
      return_struct: array[1..max_entries] of dir_entry;
      sindex: 1..4;
      wild_found, done: boolean;
      chindex, ent_index: integer;

      procedure path_found(the_entry: tree_string);

      var
         temp_path: long_string;

      begin { path_found }
         temp_path := the_entry;
         if length(dir_part) > 0
         then temp_path := concat(dir_part,'>',temp_path);
         if length(temp_path) > tree_max_chars
         then temp_path := substr(temp_path, 1, tree_max_chars);
         deliver(temp_path, continue_expanding)
      end { path_found } ;

   begin { expand_wildcard }
{     writeln('expand_wildcard:');  {----}
{     writeln('   dir_part = "',dir_part,'"  entry_part = "',entry_part,'"');  {----}
{     writeln('   want_files = ',want_files);  {----}
      wild_found := false;
      for chindex := 1 to length(entry_part)
      do if entry_part[chindex] in ['@', '+', '^']
         then wild_found := true;
      if wild_found then
      begin { there are wildcards in entry part }
         if (dir_part = '*') or (dir_part = '')
         then ret_code := 0             { no need to attach to current dir }
         else at$(k$setc, dir_part, ret_code);
         if ret_code <> 0 then
         begin { attach failed }
            error_print(k$irtn, ret_code, dir_part, 'AT$');
            main_state := command_error
         end { attach failed }
         else
         begin { attach worked }
            call_srsfx$(k$read+k$getu, '', dir_unit, dir_type, ret_code);
            if ret_code <> 0 then
            begin { open failed }
               error_print(k$irtn, ret_code, dir_part, 'SRSFX$');
               main_state := command_error
            end { open failed }
            else
            begin { directory opened }
               with sel_struct do       { fill in common items }
               begin
                  version_no := 0;
                  wild_ptr := addressof(entry_part);
                  wild_count := 1;
                  desired_types.rbf := false;
                  desired_types.filler := 0;
                  mdb := 0;
                  mda := 0;
                  svb := 0;
                  sva := 0;
                  with desired_types do
                  begin
                     dirs        := not want_files;
                     seg_dirs    := false;
                     files       := want_files;
                     access_cats := false;
                  end { with }
               end { with } ;
               start_at_top := -1;      { = true }
               continue_expanding := true;
               done := false;
               repeat
                  dir$se(dir_unit, dir_type, start_at_top, addressof(sel_struct),
                         addressof(return_struct), max_entries,
                         sizeof(return_struct[1]), ents_returned,
                         type_counts, max_type, ret_code);
                  if ret_code = e$eof then
                  begin { hit eof }
                     ret_code := 0;
                     done := true
                  end { hit eof } ;
                  if ret_code <> 0 then
                  begin { error reading dir }
                     error_print(k$irtn, ret_code, 'Reading directory', 'DIR$SE');
                     main_state := command_error
                  end { error reading dir } ;
                  start_at_top := 0;    { = false }
                  if ents_returned <= 0
                  then done := true
                  else
                  begin { some entries read }
                     for ent_index := 1 to ents_returned do
                     with return_struct[ent_index] do
                     begin
                        if entryname = 'MFD'
                        then { do not descend into MFD! }
                        else if continue_expanding
                        then path_found(entryname)
                     end { for loop }
                  end { some entries read }
               until done or (not continue_expanding)
            end { directory opened } ;
            close_unit(dir_unit, ret_code)
         end { attach worked } ;
         attach_home
      end { there are wildcards in entry part }
      else path_found(entry_part)
   end { expand_wildcard } ;

   { Expand any wildcards in the pathname part of a tree (treewalking) }

   procedure expand_treewalk(the_pathname: tree_string);

   var
      last_gt, first_wild_pos, last_wild,
      last_gt_before_wild, first_gt_after_wild: integer;
      chindex: tree_range;
      ch: char;
      dir_part, entry_part, before, wild_dir, after: tree_string;
      wild_found: boolean;

      { Add >MFD if necessary to the directory part of a pathname }
      { e.g. A>B>C splits up as A>B    and C but                  }
      {      <*>C  splits up as <*>MFD and C                      }

      procedure add_mfd(var the_dir: tree_string);

      var
         chindex: 1..maxint;
         gt_count: integer;

      begin { add_mfd }
         if length(the_dir) = 0
         then { null }
         else if the_dir[1] = '<' then
         begin { begins with disc name }
            gt_count := 0;
            for chindex := 2 to length(the_dir)
            do if the_dir[chindex] = '>'
               then gt_count := gt_count + 1;
            if gt_count = 0
            then insert('>MFD', the_dir, length(the_dir)+1)
            else { no need }
         end { begins with disc name }
         else { no disc name }
      end { add_mfd } ;

      { Deal with a treewalk wildcard match found }

      procedure treewalk_deliver(the_pathname: tree_string;
                                 var continue_expanding: boolean);

      var
         temp_pathname: long_string;

      begin { treewalk_deliver }
{        writeln('treewalk_deliver: the_pathname = "',the_pathname,'"');  {----}
         temp_pathname := concat(the_pathname, after);
         if length(temp_pathname) >= tree_max_chars then
         begin { too long }
            comm_error('Pathname too long.');
            continue_expanding := false
         end { too long }
         else
         begin { length alright }
            expand_wildcard(temp_pathname, entry_part, true, deliver);
            temp_pathname :=
               concat(the_pathname,'>',wild_dir,after,'>',entry_part);
            if length(temp_pathname) < tree_max_chars
            then expand_treewalk(temp_pathname)
         end { length alright }
      end { treewalk_deliver } ;

   begin { expand_treewalk }
{     writeln('expand_treewalk: the_pathname = "',the_pathname,'"');  {----}
      last_gt := 0;
      first_wild_pos := 0;
      for chindex := 1 to length(the_pathname) do
      begin
         ch := the_pathname[chindex];
         if ch  = '>'
         then last_gt := chindex
         else if ch in ['@', '+', '^'] then
         begin { wildcard }
            if first_wild_pos = 0
            then first_wild_pos := chindex
         end { wildcard }
      end { for loop } ;
{     writeln('last_gt = ',last_gt:1);  {----}
      if last_gt = 0 then
      begin { not pathname }
         dir_part := '';
         entry_part := the_pathname;
         expand_wildcard(dir_part, entry_part, true, deliver)
      end { not pathname }
      else
      begin { is pathname }
         dir_part := substr(the_pathname, 1, last_gt-1);
         entry_part := substr(the_pathname, last_gt+1,
                              length(the_pathname)-last_gt);
{        writeln('dir_part = "',dir_part,'"');  {----}
         add_mfd(dir_part);             { add >MFD to end if necessary }
{        writeln('dir_part = "',dir_part,'"');  {----}
{        writeln('entry_part = "',entry_part,'"');  {----}
         { does the pathname contain wildcards? }
         if (first_wild_pos = 0) or (first_wild_pos > last_gt)
         then expand_wildcard(dir_part, entry_part, true, deliver)
         else
         begin { need to treewalk }
            { split dir_part into:                                        }
            { before: pathname up to but not inc wildcard dir             }
            { wild_dir: the directory name with wildcards in              }
            { after: rest of pathname                                     }
            { 'before' may not be null, 'wild_dir' must  not contain '>', }
            { 'after' may be null                                         }
            last_gt_before_wild := 0;
            wild_found := false;
            first_gt_after_wild := 0;
            last_wild := 0;
            for chindex := 1 to length(dir_part) do
            begin
               ch := dir_part[chindex];
               if ch = '>' then
               begin { '>' found }
                  if wild_found
                  then if first_gt_after_wild = 0
                       then first_gt_after_wild := chindex
                       else { already found first after wild }
                  else last_gt_before_wild := chindex
               end { '>' found }
               else if ch in ['@', '+', '^'] then
               begin { wildcard found }
                  wild_found := true;
                  last_wild := chindex
               end { wildcard found }
            end { for loop } ;
            if last_gt_before_wild = 0
            then comm_error('Invalid treewalk pathname.')
            else
            begin { 'before' is not null }
               before := substr(dir_part, 1, last_gt_before_wild-1);
{              writeln('before = "',before,'"');  {----}
               add_mfd(before);         { add >MFD if necessary }
{              writeln('before = "',before,'"');  {----}
               if first_gt_after_wild = 0 then
               begin { no '>' after wild_dir }
                  wild_dir := substr(dir_part, last_gt_before_wild+1,
                                     length(dir_part) - last_gt_before_wild);
                  after := ''
               end { no '>' after wild_dir }
               else
               begin { wild_dir is not last }
                  wild_dir := substr(dir_part, last_gt_before_wild+1,
                     first_gt_after_wild - last_gt_before_wild - 1);
                  after := substr(dir_part, first_gt_after_wild,
                     length(dir_part) - first_gt_after_wild);
                  if last_wild > first_gt_after_wild
                  then comm_error('Incorrect treewalking pathname.')
               end { wild_dir is not last } ;
{              writeln('wild_dir = "',wild_dir,'"');  {----}
{              writeln('after = "',after,'"');  {----}
               if main_state = running then
               begin { treewalk pathname alright }
                  expand_wildcard(before, wild_dir, false, treewalk_deliver)
               end { treewalk pathname alright }
            end { 'before' is not null }
         end { need to treewalk }
      end { is pathname }
   end { expand_treewalk } ;

begin { expand_pathname }
{  writeln('expand_pathname: original_pathname = "',original_pathname,'"');  {----}
   left_par_count := 0;
   right_par_count := 0;
   contains_wildcards := false;
   for chindex := 1 to length(original_pathname) do
   begin
      ch := original_pathname[chindex];
      if ch = '(' then
      begin { left paren }
         left_par_count := left_par_count + 1;
         if left_par_count <= 2
         then left_par_pos[left_par_count] := chindex
      end { left paren }
      else if ch = ')' then
      begin { right paren }
         right_par_count := right_par_count + 1;
         if right_par_count <= 2
         then right_par_pos[right_par_count] := chindex
      end { right paren }
      else if ch in ['@', '+', '^']
      then contains_wildcards := true
   end { for loop } ;
   any_expansion := contains_wildcards or (left_par_count > 0);
   if left_par_count <> right_par_count
   then comm_error('Mismatched parentheses in pathname.')
   else if left_par_count = 0
   then expand_treewalk(original_pathname)  { no iteration }
   else if left_par_count > 2
   then comm_error('Too many parentheses in pathname.')
   else if (right_par_pos[1] - left_par_pos[1] < 1) or
           ( (left_par_count = 2) and
             ( (right_par_pos[2] - left_par_pos[2] < 1) or
               (left_par_pos[2] <> right_par_pos[1]+1 ) ) )
   then comm_error('Incorrectly combined parentheses in pathname.')
   else
   begin { all parentheses correct }
      { split pathname up into parts }
      before_iter := substr(original_pathname, 1, left_par_pos[1]-1);
      iter1 := substr(original_pathname, left_par_pos[1]+1,
                      right_par_pos[1] - left_par_pos[1] - 1);
      trim(iter1);
      if left_par_count = 1 then
      begin { single iteration }
         iter2 := '';
         after_iter := substr(original_pathname, right_par_pos[1]+1,
                              length(original_pathname) - right_par_pos[1])
      end { single iteration }
      else
      begin { extract second iteration }
         iter2 := substr(original_pathname, left_par_pos[2]+1,
                         right_par_pos[2] - left_par_pos[2] - 1);
         trim(iter2);
         after_iter := substr(original_pathname, right_par_pos[2]+1,
                              length(original_pathname) - right_par_pos[2])
      end { extract second iteration } ;

      { perform the iterations }

      continue_expanding := true;
      temp_iter1 := iter1;
      repeat
         split_at_space(temp_iter1, item1, temp_iter1);
         temp_iter2 := iter2;
         repeat
            split_at_space(temp_iter2, item2, temp_iter2);
            expand_treewalk(concat(before_iter, item1, item2, after_iter))
         until (length(temp_iter2) = 0) or (not continue_expanding)
      until (length(temp_iter1) = 0) or (not continue_expanding)
   end { all parentheses correct }
end { expand_pathname } ;

begin { pmed_procs }
end { pmed_procs } .
