%separate = true
%ecs = true
%ttybuffered = false
%epfargs = true
{ Sheffield University Editor }
{ Copyright (c) University of Sheffield 1990 }

{***********************************************************}
{ SIMPLE SCREEN MANAGEMENT PROTOCOL (SSMP) HANDLING PACKAGE }
{***********************************************************}

subprogram pmed_ssmp;

%listing off
%insert pmed_globals.ins.pas
%insert pmed_procs.ins.pas
%insert pmed_vdu.ins.pas
%listing on

{ initialise our copy of the shared data structure }

procedure init_shared;                                                          { INIT_SHARED }

var
   fidx: field_index;
   trow: ssmp_row_range;
   tcol: ssmp_col_range;
   mindex: mode_range;

begin { init_shared }
   token := withhost;
   saved_row := 0;  saved_col := 0;

   curfield := 0;  maxfield := 0;

   { update limits }
   box_top := 0;   box_bottom := 0;
   box_left := 0;  box_right := 0;

   { ssmp_mode array }
   for mindex := 0 to 63 do ssmp_mode[mindex] := 0
end { init_shared } ;

{ Get next frame from network }

procedure net_read;                                                             { NET_READ }

var
    ch: char;

begin { net_read }
   input_buffer_chars := 0;
   while not eoln(input) do
   begin
      read(ch);
      input_buffer_chars := input_buffer_chars + 1;
      input_buffer[input_buffer_chars] := ch
   end;
   readln;
   input_buffer_pointer := 0
end { net_read } ;

{ Output assembled frames to the network }

procedure net_write;                                                            { NET_WRITE }

var
   pindex: packet_range;
   ch: char;

begin { net_write }
   for pindex := 1 to output_buffer_pointer do
   begin
      ch := output_buffer[pindex];
      if ch = newline
      then writeln
      else write(ch)
   end
end { net_write } ;

{ End current output frame }

procedure end_frame;                                                            { END_FRAME }

begin { end_frame }
   if output_buffer_pointer > 0 then
   begin
      if output_buffer[output_buffer_pointer] = ' ' then
      begin { foil trailing space trim }
         output_buffer[output_buffer_pointer+1] := '$';
         output_buffer_pointer := output_buffer_pointer + 1
      end { foil trailing space trim } ;
      output_buffer[output_buffer_pointer+1] := newline;
      output_buffer_pointer := output_buffer_pointer + 1;
      if frame_number < frames_per_packet
      then frame_number := frame_number + 1
      else
      begin { packet full }
         net_write;
         frame_number := 1;  output_buffer_pointer := 0
      end { packet full }
   end;
   out_remainder := 0
end { end_frame } ;

{ Push frame(s) to network }

procedure push_frame;                                                           { PUSH_FRAME }

begin { push_frame }
   end_frame;
   if output_buffer_pointer > 0 then
   begin
       net_write;
       frame_number := 1;  output_buffer_pointer := 0
   end
end { push_frame } ;

{ Add a character to the output frame being built }

procedure add_to_frame(ch: char);                                               { ADD_TO_FRAME }

begin { add_to_frame }
   if suppress_output
   then                                 { do not add anything }
   else
   begin
      if out_remainder = 0 then
      begin
         end_frame;
         output_buffer_pointer := output_buffer_pointer + 1;
         output_buffer[output_buffer_pointer] := chr(frame_start);
         out_remainder := frame_size
      end;
      output_buffer_pointer := output_buffer_pointer + 1;
      out_remainder := out_remainder - 1;
      output_buffer[output_buffer_pointer] := ch
   end { output not suppressed }
end { add_to_frame } ;

{ Get the next character from the current input frame }

procedure get_from_frame(var ch: char);                                         { GET_FROM_FRAME }

var
   code: integer;

begin { get_from_frame }
   repeat
      while input_buffer_pointer >= input_buffer_chars
      do net_read;
      input_buffer_pointer := input_buffer_pointer + 1;
      code := filter[ord(input_buffer[input_buffer_pointer])]
   until code <> 0;
   ch := chr(code+128)
end { get_from_frame } ;

{ Encode host SSMP primitive (except H-CHARACTER) and add to output frame }

procedure assemble(code_char: char; npars: pars0_range; pars: primpar);         { ASSEMBLE }

var
   pindex: pars_range;
   temp_str: string[8];
   chindex: 1..8;

begin { assemble }
   add_to_frame('=');
   add_to_frame(code_char);
   for pindex := 1 to npars do
   begin
      str(pars[pindex], temp_str);
      for chindex := 1 to length(temp_str)
      do add_to_frame(temp_str[chindex]);
      if pindex <> npars
      then add_to_frame(',')
   end;
   add_to_frame(';')
end { assemble } ;

{ Issue pending SSMP code as necessary }

procedure issue_pending(new_pend_state: pend_type);                             { ISSUE_PENDING }

var
   ppar: primpar;

begin { issue_pending }
   if new_pend_state <> pend_state then
   begin
      case pend_state of
         nopend: ;
         csrpend:
            begin
               ppar[1] := saved_row;  ppar[2] := saved_col;
               assemble('B', 2, ppar);
            end;
         fldpend:
            begin
               ppar[1] := curfield;
               assemble('L', 1, ppar);
            end;
         chinspend:
            begin
               while saved_ssmp_count > 255 do
               begin
                  ppar[1] := 255;
                  assemble('E', 1, ppar);
                  saved_ssmp_count := saved_ssmp_count - 255
               end;
               ppar[1] := saved_ssmp_count;
               assemble('E', 1, ppar)
            end;
         chdelpend:
            begin
               while saved_ssmp_count > 255 do
               begin
                  ppar[1] := 255;
                  assemble('F', 1, ppar);
                  saved_ssmp_count := saved_ssmp_count - 255
               end;
               ppar[1] := saved_ssmp_count;
               assemble('F', 1, ppar)
            end;
      end { case } ;
      pend_state := new_pend_state;
      saved_ssmp_count := 0
   end { new_pend_state <> pend_state }
end { issue_pending } ;

{ Generate H-CHARACTER to send text }

procedure hcharacter(ch: char);                                                 { HCHARACTER }

var
   code: smallint;
   ppar: primpar;

begin { hcharacter }
   issue_pending(nopend);
   { switch parity on if necessary }
   if ord(ch) < 200b
   then ch := chr(ord(ch) + 200b);
   code := encode[ord(ch)];
   if code = 0
   then add_to_frame(ch)
   else
   begin { it is coded }
      add_to_frame(':');
      add_to_frame(chr(code + 128))
   end { it is coded } ;
   if saved_col < ssmp_mode[tmaxcol]
   then saved_col := saved_col + 1
end { hcharacter } ;

{ Convert literal string into H-CHARACTER calls }

procedure ssmp_literal(the_string: tree_string);                                { SSMP_LITERAL }

var
   chindex: 1..maxint;

begin { ssmp_literal }
   for chindex := 1 to length(the_string)
   do hcharacter(the_string[chindex])
end { ssmp_literal } ;

{ Generate H-TOKEN to transfer token }

procedure htoken(reqcode: smallint);                                            { HTOKEN }

var
   ppar: primpar;

begin { htoken }
   issue_pending(nopend);
   ppar[1] := reqcode;  assemble('A', 1, ppar);  push_frame;
   if ssmp_mode[cursor] = 0 then
   begin { bound cursor }
      if (saved_row < box_top) or (saved_row > box_bottom) or
         (saved_col < box_left) or (saved_col > box_right)
      then
      begin
         saved_row := box_top;
         saved_col := box_left
      end
   end { bound cursor } ;
   ssmp_mode[tlevel] := 0;              { check this element on token return }
   token := withterm                    { control with remote terminal }
end { htoken } ;

{ Generate H-SETCURSOR to set the cursor position }

procedure hsetcursor(newrow: ssmp_row_range; newcol: ssmp_col_range);           { HSETCURSOR }

var
   ppar: primpar;

begin { hsetcursor }
   issue_pending(csrpend);
   saved_row := newrow;
   saved_col := newcol
end { hsetcursor } ;

{ Generate H-SETMODE to set a mode array element value }

procedure hsetmode(index: mode_range; ivalue: smallint);                        { HSETMODE }

var
   ppar: primpar;

begin { hsetmode }
   ppar[1] := index;
   ppar[2] := ivalue;
   assemble('C', 2, ppar);
   ssmp_mode[index] := ivalue
end { hsetmode } ;

{ Generate H-ERASETORIGHT to erase to end of row }

procedure herasetoright;                                                        { HERASETORIGHT }

var
   ppar: primpar;

begin { herasetoright }
   issue_pending(nopend);
   assemble('D', 0, ppar)
end { herasetoright } ;

{ Generate H-INSERTSPACE to insert SPACE characters }

procedure hinsertspace(nsp: smallint);                                          { HINSERTSPACE }

var
   ppar: primpar;

begin { hinsertspace }
   issue_pending(chinspend);
   saved_ssmp_count := saved_ssmp_count + nsp
end { hinsertspace } ;

{ Generate H-DELETECHAR to delete characters }

procedure hdeletechar(nch: smallint);                                           { HDELETECHAR }

var
   ppar: primpar;

begin { hdeletechar }
   issue_pending(chdelpend);
   saved_ssmp_count := saved_ssmp_count + nch
end { hdeletechar } ;

{ Generate H-SCROLLUP to scroll rows up }

procedure hscrollup(rowa, rowb: ssmp_row_range; nrows: smallint);               { HSCROLLUP }

var
   ppar: primpar;

begin { hscrollup }
   issue_pending(nopend);
   ppar[1] := rowa;  ppar[2] := rowb;  ppar[3] := nrows;
   assemble('H', 3, ppar)
end { hscrollup } ;

{ Generate H-SCROLLDOWN to scroll rows down }

procedure hscrolldown(rowa, rowb: ssmp_row_range; nrows: smallint);             { HSCROLLDOWN }

var
   ppar: primpar;

begin { hscrolldown }
   issue_pending(nopend);
   ppar[1] := rowa;  ppar[2] := rowb;  ppar[3] := nrows;
   assemble('I', 3, ppar);
end { hscrolldown } ;

{ Generate H-ERASEDISPLAY to erase screen image }

procedure herasedisplay;                                                        { HERASEDISPLAY }

var
   ppar: primpar;

begin { herasedisplay }
   if pend_state = fldpend
   then issue_pending(nopend)
   else pend_state := nopend;           { implied cursor movement to [0,0] }
   assemble('J', 0, ppar);
   saved_row := 0;  saved_col := 0
end { herasedisplay } ;

{ Generate H-ERASEFIELDS to erase all field defns }

procedure herasefields;                                                         { HERASEFIELDS }

var
   ppar: primpar;

begin { herasefields }
   if pend_state = fldpend
   then pend_state := nopend            { implicit selection of field zero }
   else issue_pending(nopend);
   assemble('K', 0, ppar);
   curfield := 0;  maxfield := 0;
   box_top := 0;  box_bottom := 0;
   box_left := 0;  box_right := 0
end { herasefields } ;

{ Generate H-SETFIELD to select new field definition }

procedure hsetfield(fidx: field_index);                                         { HSETFIELD }

var
   ppar: primpar;

begin { hsetfield }
   if pend_state = csrpend
   then pend_state := nopend;           { suppress any H-SETCURSOR }
   issue_pending(fldpend);
   if fidx > maxfield then maxfield := fidx;
   curfield := fidx;
   saved_row := box_top;  saved_col := box_left
end { hsetfield } ;

{ Generate H-SETUPDATE to set update limits for field }

procedure hsetupdate(rowa, rowb: ssmp_row_range; cola, colb: ssmp_col_range);   { HSETUPDATE }

var
   ppar: primpar;

begin { hsetupdate }
   if (box_top  = rowa) and (box_bottom = rowb) and
      (box_left = cola) and (box_right  = colb)
   then                                 { already set }
   else
   begin
      issue_pending(nopend);
      ppar[1] := rowa;  ppar[2] := rowb;
      ppar[3] := cola;  ppar[4] := colb;
      assemble('M', 4, ppar);
      box_top := rowa;
      box_bottom := rowb;
      box_left  := cola;
      box_right := colb
   end { need to issue }
end { hsetupdate } ;

{ Generate H-ERASETABS to erase all tabulation stops }

procedure herasetabs;                                                           { HERASETABS }

var
   ppar: primpar;

begin { herasetabs }
   assemble('N', 0, ppar)
end { herasetabs } ;

{ Generate H-SETTAB to set one tabulation stop }

procedure hsettab(tcol: ssmp_col_range);                                        { HSETTAB }

var
   ppar: primpar;

begin { hsettab }
   ppar[1] := tcol;
   assemble('O', 1, ppar)
end { hsettab } ;

{ Generate H-SESSION to enter or exit session }

procedure hsession(reqcode: smallint);                                          { HSESSION }

var
   ppar: primpar;

begin { hsession }
   issue_pending(nopend);
   ppar[1] := reqcode;
   assemble('P', 1, ppar);
   if reqcode = 0
   then init_shared                     { new session, init shared data }
   else push_frame                      { current session ends, flush packet }
end { hsession } ;

{ Generate H-REQTOKEN to request early return of token }

procedure hreqtoken;                                                            { HREQTOKEN }

var
   saved_suppress_output: boolean;
   ppar: primpar;

begin { hreqtoken }
   issue_pending(nopend);
   push_frame;
   saved_suppress_output := suppress_output;
   suppress_output := false;            { ensure request actually gets sent! }
   assemble('Q', 0, ppar);
   push_frame;
   suppress_output := saved_suppress_output;
   token_requested := true;
   token_here_after_request := false
   { will be set true when special token arrives }
end { hreqtoken } ;

{ Do the equivalent or CR and LF on an SSMP terminal }

procedure do_ssmp_cr;                                                           { DO_SSMP_CR }

begin { do_ssmp_cr }
   hsetcursor(saved_row, 0)
end { do_ssmp_cr } ;

procedure do_ssmp_lf;                                                           { DO_SSMP_LF }

begin { do_ssmp_lf }
   if saved_row < ssmp_mode[tmaxrow]
   then hsetcursor(saved_row+1, saved_col)
   else
   begin
      hscrollup(0, ssmp_mode[tmaxrow], 1);
      {*************************************************************}
      {                                                             }
      { Temporary fix because of a problem in Newcastle SSMP for PC }
      {                                                             }
      {*************************************************************}
      hsetcursor(saved_row, saved_col)
      {*** end of fix ***}
   end
end { do_ssmp_lf } ;

{ Make the real mode array conform to that supplied }

procedure put_mode(mcopy: mode_array);                                          { PUT_MODE }

var
   mindex: mode_range;

begin { put_mode }
   for mindex := notify to klast
   do if mcopy[mindex] <> ssmp_mode[mindex]
      then hsetmode(mindex, mcopy[mindex])
end { put_mode } ;

{ Get the next imput primitive from the input frame(s) }

procedure get_primitive(var code_char: char;                                    { GET_PRIMITIVE }
                        var npars: pars0_range;
                        var pars: primpar);

var
   alright, par_done: boolean;
   par: integer;
   ch: char;

begin { get_primitive }
   repeat
      alright := true;
      get_from_frame(ch);
      if ch = '=' then
      begin { not t-character }
         get_from_frame(code_char);
         npars := 0;
         par_done := false;
         get_from_frame(ch);
         repeat
            if ch in ['0'..'9'] then
            begin
               if npars <= max_pars
               then npars := npars + 1;
               par := 0;
               repeat
                  par := par * 10 + ord(ch) - ord('0');
                  get_from_frame(ch)
               until not (ch in ['0'..'9']);
               pars[npars] := par;
               if ch = ','
               then get_from_frame(ch)
            end
            else par_done := true
         until par_done;
         if ch <> ';'
         then alright := false
      end { starts with '=' }
      else if ch = ':' then
      begin { coded t-character }
         get_from_frame(ch);
         code_char := '*';
         npars := 1;
         pars[1] := decode[ord(ch)]
      end { coded t-character }
      else
      begin { uncoded t-character }
         code_char := '*';
         npars := 1;
         pars[1] := ord(ch) - 128
      end { uncoded t-character }
   until alright
end { get_primitive } ;

{ Clear the stacked SSMP codes (if any) }

procedure clear_ssmp_stack;                                                     { CLEAR_SSMP_STACK }

begin { clear_ssmp_stack }
   with stacked_ssmp_codes do           { no stacked input }
   begin
      code1 := 0;
      code2 := 0;
      row_diff := 0;
      col_diff := 0;
      tree_report := false;
      multiple_count := 0
   end
end { clear_ssmp_stack } ;

{ Negotiate screen management session with terminal }

function start_ssmp: boolean;                                                   { START_SSMP }

var
   success: boolean;

begin { start_ssmp }
   out_remainder := 0;
   output_buffer_pointer := 0;
   frame_number := 1;                   { first frame of output }
   pend_state := nopend;                { no pending operations }
   clear_ssmp_stack;                    { no stacked input }
   token_requested := false;
   token_here_after_request := false;
   writeln;                             { ensure LF before first frame }
   hsession(0);                         { initialises shared data structure }
   ssmp_literal('  Please press the RETURN key to continue.  ');
   htoken(1);
   net_read;                            { read first frame, then crude check }
                                        { for T-SETMODE }
   if input_buffer_chars >= 3
   then success := (input_buffer[1] = '=') and
                   (input_buffer[2] = 'C') and
                   (input_buffer[3] = '0')
   else success := false;
   if not success
   then writeln('Simple Screen Management Protocol ::: connection refused.');
   start_ssmp := success
end { start_ssmp } ;

{ Tidy up and close down screen management session }

procedure stop_ssmp;                                                            { STOP_SSMP }

begin { stop_ssmp }
   hsession(1);
   if term_is_direct then
   begin { terminal on direct line }
      ctrl_term(half_duplex, false)
   end { terminal on direct line }
end { stop_ssmp } ;

%page

{***********************}
{ SSMP APPLICATION CODE }
{***********************}

{ Host application, fetch and dispatch T-primitives }

procedure dispatch;                                                             { DISPATCH }

var
   primitive_char: char;
   npars: pars0_range;
   pars: primpar;
   request_code: smallint;
   image_subscript: integer;
   mindex: smallint;
   saved_box_top, saved_box_bottom, saved_box_left, saved_box_right: smallint;

   { Stack an SSMP input code as control char or ESC and printable char }

   procedure stack_code(the_code: smallint);

   begin { stack_code }
      the_code := the_code + 128;       { -> usual range }
      with stacked_ssmp_codes do
      begin
         if chr(the_code) in ascii_printable then
         begin { printable }
            code1 := esc_code;
            code2 := the_code
         end { printable }
         else
         begin { not printable }
            code1 := the_code;
            code2 := 0
         end { not printable }
      end { with }
   end { stack_code } ;

   procedure stack_multiple(the_code, counter: smallint);

   begin { stack_multiple }
      the_code := the_code + 128;       { -> usual range }
      with stacked_ssmp_codes do
      begin
         if chr(the_code) in ascii_printable then
         begin { printable }
            mult_code1 := esc_code;
            mult_code2 := the_code
         end { printable }
         else
         begin { not printable }
            mult_code1 := the_code;
            mult_code2 := 0
         end { not printable } ;
         multiple_count := counter
      end { with }
   end { stack_multiple } ;

   procedure stack_cursor_report(stack_row, stack_col: smallint);

   begin { stack_cursor_report }
      with stacked_ssmp_codes do
      begin
         row_diff := stack_row - saved_row;
         col_diff := stack_col - saved_col
      end { with }
   end { stack_cursor_report } ;

begin { dispatch }
   get_primitive(primitive_char, npars, pars);  { fetch next primitive }
   with stacked_ssmp_codes do
   case primitive_char of
      '*':       { T-CHARACTER(ch) }
         begin
            code1 := pars[1] + 128;
            if chr(code1) in ascii_printable
            then { normal printable char }
            else code1 := code1 - 256;  { special action for literal ctl ch }
            if saved_col < ssmp_mode[tmaxcol]
            then saved_col := saved_col + 1
         end;
      'A':       { T-TOKEN(reqcode) }
         begin
            request_code := pars[1];
            if request_code = 0
            then { skip the normal t-token(0) }
            else if request_code = 128 then
            begin { special token after H-REQTOKEN }
               token_requested := false;
               token_here_after_request := true
            end { special token after H-REQTOKEN }
            else if request_code = 129  { user session restart request }
            then stack_code(ssmp_mode[krestart])
            { if a vicious terminal sends an invalid code > 129 ignore it }
            else if request_code > 129
            then { ignore }
            else stack_code(request_code);
            token := withhost;
            { The following code sets the dsinvalid mode array element zero    }
            { whenever we receive t-token(0).  This is unnecessary except that }
            { once or twice (especially after a user break) the terminal has   }
            { a value of one for this element but we do not know about it.     }
            { Maybe the h-setmode command was corrupted in transit.            }
            if request_code = 0
            then hsetmode(dsinvalid, 0);
            if ssmp_mode[dsinvalid] <> 0 then
            begin { must reset modes }
               saved_box_top := box_top;
               saved_box_bottom := box_bottom;
               saved_box_left := box_left;
               saved_box_right := box_right;
               box_top := 0;
               box_bottom := 0;
               box_left := 0;
               box_right := 0;
               hsetmode(dsinvalid, 0);
               for mindex := notify to klast
               do hsetmode(mindex, ssmp_mode[mindex]);
               hsetupdate(saved_box_top, saved_box_bottom,
                          saved_box_left, saved_box_right)
            end { must reset modes }
         end;
      'B':       { T-SETCURSOR(newrow,newcol) }
         begin
            stack_cursor_report(pars[1], pars[2]);
            saved_row := pars[1];
            saved_col := pars[2]
         end;
      'C':       { T-SETMODE(index,ivalue) }
         begin
            ssmp_mode[pars[1]] := pars[2];
            { set mode_insert to match [icharmode] element }
            if pars[1] = icharmode
            then mode_insert := pars[2] = 1
         end;
      'D':       { T-ERASETORIGHT }
         stack_code(ssmp_mode[keraright ]);
      'E':       { T-INSERTSPACE(nsp) }
         stack_multiple(ssmp_mode[kinsspac  ], pars[1]);
      'F':       { T-DELETECHAR(nch) }
         stack_multiple(ssmp_mode[kdelchar  ], pars[1]);
      'G':       { T-ERASEPREV(nch) }
         begin
            stack_multiple(ssmp_mode[keraprev  ], pars[1]);
            saved_col := saved_col - pars[1]
         end;
      'H':       { T-INSERTLINE(nln) }
         stack_multiple(ssmp_mode[kinsline  ], pars[1]);
      'I':       { T-DELETELINE(nln) }
         stack_multiple(ssmp_mode[kdelline  ], pars[1]);
      'J': ;     { T-APPENDLINE }
      'K': ;     { T-SPLITLINE }
      'L':       { T-SETFIELD(fidx) }
         begin { treat as cursor position to top left }
            stack_cursor_report(box_top, box_left);
            saved_row := box_top;
            saved_col := box_left
         end { treat as cursor position to top left }
   end {case}
end { dispatch } ;

{ Decide whether the login connection is direct or network; we need to know }
{ this for SSMP use }

procedure decide_if_direct;                                                     { DECIDE_IF_DIRECT }

const
   xi$rlg = 40000b;
   xs$unk = 7;

type
   atype = packed array[1..128] of char;

var
   num, alen1, alen2, code, time: shortint;
   array1, array2: atype;

   procedure x$stat(key: shortint;
                    var num: shortint;
                    var array1: atype;
                    var alen1: shortint;
                    var array2: atype;
                    var alen2: shortint;
                    var code, time: shortint);
      extern;

begin { decide_if_direct }
   x$stat(xi$rlg, num, array1, alen1, array2, alen2, code, time);
   term_is_direct := code = xs$unk
end { decide_if_direct } ;

{ Start up SSMP and process initial dialogue if successful }

function init_ssmp: boolean;                                                    { INIT_SSMP }

begin { init_ssmp }
   screen_tabs_set := false;
   decide_if_direct;
   if term_is_direct then
   begin { terminal on direct line }
      ctrl_term(half_duplex, true);
      usual_ssmp_mode[intsignal] := 16;  { CTRL&P}
      reply_mode[intsignal] := 16
   end { terminal on direct line }
   else
   begin { terminal on network line }
      usual_ssmp_mode[intsignal] := 0;  { CTRL&P}
      reply_mode[intsignal] := 0
   end { terminal on network line } ;
   if start_ssmp then
   begin
      while token = withterm            { process reply to opening question }
      do dispatch;
      put_mode(usual_ssmp_mode);        { set usual mode values }
      htoken(1);
      while token = withterm            { process any replies }
      do dispatch;
      with vdu_table^ do                { copy returned information }
      begin
         cols_per_row  := ssmp_mode[tmaxcol] + 1;
         rows_per_page := ssmp_mode[tmaxrow] + 1;
         ch_ins_del_avail := ssmp_mode[reqshift] = 1
      end;
      if ssmp_mode[icharmode] <> ord(mode_insert)
      then hsetmode(icharmode, ord(mode_insert) );
      calculate_window_sizes;           { affected by new rows/cols }
      init_ssmp := true
   end { session started alright }
   else
   begin { connection refused }
      init_ssmp := false;
      if term_is_direct
      then ctrl_term(half_duplex, false)
   end { connection refused }
end { init_ssmp } ;

{ Set and remember special update limits }

procedure set_special_limits(rowa, rowb: ssmp_row_range;                        { SET_SPECIAL_LIMITS }
                             cola, colb: ssmp_col_range);

begin { set_special_limits }
   special_limits_set := true;
   special_top    := rowa;
   special_bottom := rowb;
   special_left   := cola;
   special_right  := colb;
   hsetupdate(rowa, rowb, cola, colb)
end { set_special_limits } ;

{ Set the update limits to include just the cursor position so that we can }
{ detect any command typed, including cursor movements }

procedure box_in_cursor;                                                        { BOX_IN_CURSOR }

begin { box_in_cursor }
   { ensure not in character insertion mode because SSMP terminal }
   { would try to insert the single character }
   if ssmp_mode[icharmode] <> 0
   then hsetmode(icharmode, 0);
   set_special_limits(saved_row, saved_row, saved_col, saved_col)
end { box_in_cursor } ;

{ Wait until the token returns from the terminal }

procedure wait_for_token(requested: boolean);                                   { WAIT_FOR_TOKEN }

var
   done: boolean;

begin { wait_for_token }
   token_here_after_request := false;
   done := false;
   repeat
      while token = withterm
      do dispatch;
      if requested
      then if token_here_after_request then
           begin { special token arrived }
              done := true;
              token_here_after_request := false
           end
           else                         { token was not the special one }
      else done := true                 { any will do }
   until done
end { wait_for_token } ;

{ Get the token by using H-REQTOKEN if necessary }

procedure get_token;                                                            { GET_TOKEN }

begin { get_token }
   if token = withterm then
   begin
      hreqtoken;
      wait_for_token(true)
   end
end { get_token } ;

{ Get a stacked SSMP input code if any present }

procedure get_stacked_code(var the_code: integer);                              { GET_STACKED_CODE }

begin { get_stacked_code }
   with stacked_ssmp_codes do
   begin
      the_code := 0;
      if code1 <> 0 then
      begin { stacked code found }
         the_code := code1;
         code1 := code2;
         code2 := 0
      end { stacked code found }
      else if multiple_count > 0 then
      begin { get one from repeated code }
         multiple_count := multiple_count - 1;
         the_code := mult_code1;
         code1 := mult_code2            { second part (if any) next time }
      end { get one from repeated code }
      { convert cursor position report into cursor movements }
      else if (row_diff <> 0) or (col_diff <> 0) then
      begin { stacked cursor movement }
         if row_diff < 0 then
         begin
            the_code := ssmp_mode[kcsrup] + 128;
            row_diff := row_diff + 1
         end
         else if row_diff > 0 then
         begin
            the_code := ssmp_mode[kcsrdown] + 128;
            row_diff := row_diff - 1
         end
         else if col_diff < 0 then
         begin
            the_code := ssmp_mode[kcsrleft] + 128;
            col_diff := col_diff + 1
         end
         else if col_diff > 0 then
         begin
            the_code := ssmp_mode[kcsrright] + 128;
            col_diff := col_diff - 1
         end
         else the_code := 0;            { cursor now in correct position }
         if chr(the_code) in ascii_printable then
         begin { escape code }
            code1 := the_code;
            the_code := esc_code
         end { escape code }
      end { cursor report }
      else if tree_report then
      begin
         if length(stacked_tree) > 0 then
         begin { not end of tree }
            the_code := ord(stacked_tree[1]);
            delete(stacked_tree, 1, 1)
         end { not end of tree }
         else
         begin { end of tree }
            if tree_cr_given
            then the_code := lf_code
            else the_code := can_code;
            tree_report := false
         end { end of tree }
      end { tree_report }
      else { no code found }
   end { with }
end { get_stacked_code } ;

{ Get a user reply from SSMP terminal.                                     }
{ Proceed by setting update limits round rest of current line and allowing }
{ the user to edit using <-, -> and DEL.                                   }

procedure get_ssmp_reply(message: long_string;                                  { GET_SSMP_REPLY }
                         highlight_it: boolean;
                         var answer: long_string;
                         var cr_given: boolean);

var
   image_row, image_col, rindex, next_code: integer;
   done: boolean;
   state: search_type;
   last_command: (start_up, del_code, left_code, right_code, char_code);
   image: record
      case boolean of
         false: (len: shortint; chars: packed array[line_chars_range] of char);
         true:  (strng: string[line_max_chars])
      end;

   procedure set_limits;

   var
      start_col: integer;

   begin { set_limits }
      if saved_row = image_row
      then start_col := image_col
      else start_col := 0;
      set_special_limits(saved_row, saved_row, start_col, ssmp_mode[tmaxcol])
   end { set_limits } ;

begin { get_ssmp_reply }
   { Preserve the current graphic rendition }
   reply_mode[selectgr] := ssmp_mode[selectgr];
   put_mode(reply_mode);                { standard set of operations available }
   done := false;
   repeat
      no_lf_screen_message(message, highlight_it);
      image_row := saved_row;
      image_col := saved_col;
      set_limits;
      last_command := start_up;
      for rindex := 1 to line_max_chars
      do image.chars[rindex] := chr(nul_code);  { fill with NULs }
      rindex := 1;
      state := searching;
      repeat
         get_stacked_code(next_code);
         if next_code = 0 then
         begin { no character obtained }
            if token = withterm
            then dispatch               { get next input code }
            else
            begin { token = withhost }
               if last_command in [right_code, char_code] then
               begin { moving forwards }
                  if saved_row = ssmp_mode[tmaxrow]
                  then rindex := rindex - 1
                  else
                  begin { not last row on screen }
                     hsetcursor(saved_row+1, 0);
                     set_limits;
                     if message_lines_to_clear < saved_row+1 then
                     begin { first time on this row }
                        message_lines_to_clear := saved_row+1;
                        herasetoright
                     end { first time on this row }
                  end { not last row on screen }
               end { moving forwards }
               else if last_command in [left_code, del_code] then
               begin { moving backwards }
                  if (saved_row = image_row) and (saved_col = image_col)
                  then rindex := rindex + 1  { at start of image }
                  else
                  begin
                     hsetcursor(saved_row-1, ssmp_mode[tmaxcol]);
                     set_limits;
                     if last_command = del_code
                     then hcharacter(' ')
                  end
               end { moving backwards } ;
               htoken(0)                { over to user }
            end { token = withhost }
         end { no character obtained }
         else
         begin { character obtained }
            { test for coded control character or ordinary printable char }
            if (next_code < 0) or
               ((next_code >= ord(first_print_char)) and
                (next_code <= ord( last_print_char)))
            then
            begin { printable character }
               { convert coded control char back to normal }
               if next_code < 0
               then next_code := next_code + 256;
               if (rindex >= 1) and (rindex <= line_max_chars)
               then image.chars[rindex] := chr(next_code);
               rindex := rindex + 1;
               last_command := char_code
            end { printable character }
            else
            begin { control code }
               if next_code = esc_code
               then get_stacked_code(next_code);
               if (next_code = cr_code) or  { CR or }
                  (next_code = can_code)  { CTRL&X }
               then state := found
               else if next_code = ssmp_mode[kcsrleft]+128 then  { <- }
               begin
                  rindex := rindex - 1;
                  last_command := left_code
               end
               else if next_code = ssmp_mode[kcsrright]+128 then  { -> }
               begin
                  rindex := rindex + 1;
                  last_command := right_code
               end
               else if next_code = ssmp_mode[keraprev]+128 then  { DEL }
               begin
                  rindex := rindex - 1;
                  last_command := del_code;
                  if (rindex >= 1) and (rindex <= line_max_chars)
                  then image.chars[rindex] := chr(nul_code)
               end
               else if next_code = ssmp_mode[krestart]+128 then
               begin
                  last_command := start_up;
                  state := not_found
               end
               else last_command := start_up
            end { not printable character }
         end { character obtained }
      until state <> searching;
      if state = found
      then done := true
      else vdu_control(clearscreen)     { user did a session restart }
   until done;
   cr_given := (next_code = cr_code);
   special_limits_set := false;
   put_mode(usual_ssmp_mode);
   clear_ssmp_stack;
   { do the equivalent of CR/LF }
   do_ssmp_cr; do_ssmp_lf;
   with image do
   begin
      len := line_max_chars;
      done := false;
      repeat                            { strip any trailing NULs }
         if len = 0
         then done := true
         else if chars[len] = chr(nul_code)
         then len := len - 1
         else done := true
      until done;
      for rindex := 1 to len            { convert internal NULs to spaces }
      do if chars[rindex] = chr(nul_code)
         then chars[rindex] := ' '
   end { with } ;
   answer := image.strng
end { get_ssmp_reply } ;

begin { pmed_ssmp }
end { pmed_ssmp } .
