%separate = true
%ecs = true
%ttybuffered = false
%epfargs = true
{ Sheffield University Editor }
{ Copyright (c) University of Sheffield 1990 }

{***********************}
{ VDU HANDLING ROUTINES }
{***********************}

subprogram pmed_vdu;

%listing off
%insert pmed_globals.ins.pas
%insert pmed_efs.ins.pas
%insert pmed_procs.ins.pas
%insert pmed_ssmp.ins.pas
%listing on

{ Print a single character on the terminal. We use tnoua rather than t1ou }
{ because t1ou converts LF into CR and LF }

procedure print_a_char(the_char: char);                                         { PRINT_A_CHAR }

type
   twoch = packed record
      ch1, ch2: char
   end;

var
   string: twoch;

   procedure tnoua(string: twoch; count: shortint);
      extern;

begin { print_a_char }
   if suppress_output
   then                                 { no output }
   else if term_is_ssmp
   then hcharacter(the_char)
   else
   begin
      string.ch1 := the_char;
      tnoua(string, 1)
   end
end { print_a_char } ;

{ Output a string of characters given the start subscript and length, }
{ inserting row and column values if requested }

procedure vdu_string(ch_point: vdu_char_pointer;                                { VDU_STRING }
                     string_row: row0_range;
                     string_col: col0_range);

var
   index: vdu_char_range;
   ch_code, last_code: integer;

   { output a cursor coordinate as a single ascii character }
   procedure ascii_plus(valu: integer);
   begin { ascii_plus }
      print_a_char(chr(valu + ch_code))
   end { ascii_plus } ;

   { similar for ADM12PLUS, which wraps around after column 80 }
   procedure adm12plus_row(rvalu, cvalu: integer);
   begin { adm12plus_row }
      if cvalu <= 80
      then ascii_plus(rvalu)
      else ascii_plus(rvalu+48)
   end { adm12plus_row } ;

   procedure adm12plus_col(rvalu, cvalu: integer);
   begin { adm12plus_col }
      if cvalu <= 80
      then ascii_plus(cvalu)
      else ascii_plus(cvalu-80)
   end { adm12plus_col } ;

   { output a cursor coordinate as its decimal value + offset }
   procedure dec_plus(valu: integer);
   begin { dec_plus }
      write(valu + ch_code:1)
   end { dec_plus } ;

   procedure sleep$(msecs: integer);
      extern;

begin { vdu_string }
   if suppress_output
   then                                 { soon to redraw screen - ignore }
   else with vdu_table^, ch_point do
   begin { output not suppressed }
      last_code := 0;
      for index := start_index to (start_index + string_length - 1) do
      begin
         ch_code := two_way.char_array[index];
         if last_code < 0 then
         begin
            case last_code of           { last_code < 0 - output col or row }
               -1: ascii_plus(string_row);
               -2: ascii_plus(string_col);
               -3: dec_plus(string_row);
               -4: dec_plus(string_col);
               -5: adm12plus_row(string_row, string_col);
               -6: adm12plus_col(string_row, string_col)
               otherwise abort('unrecognised action code in vducursor_string')
            end {case} ;
            ch_code := 0                { to clear "last_code < 0" switch }
         end
         else if ch_code < -100 then
         begin
            hook7(vdu_type, ch_code);   { user control hook }
            ch_code := 0
         end
         else if ch_code = 0            { code 0 => wait 0.1 secs }
         then sleep$(100)
         else if ch_code > 0 then       { code > 0 => output ascii char }
         print_a_char(chr(ch_code));
         last_code := ch_code
      end { for loop }
   end { output not suppressed }
end { vdu_string } ;

{ Control the vdu terminal - uses appropriate control codes from vdu record }

procedure vdu_control(control: controls);                                       { VDU_CONTROL }

   { select graphic rendition for SSMP terminal from info in table }

   procedure select_rendition(control: controls);

   var
      rendition: smallint;

   begin { select_rendition }
      with vdu_table^, control_array[control] do
      begin
         if string_length = 1           { single integer is rendition to use }
         then rendition := two_way.char_array[start_index]
         else rendition := defaultrend;
         hsetmode(selectgr, rendition)
      end { with }
   end { select_rendition } ;

begin { vdu_control }
   if term_is_ssmp then
   begin
      case control of
         backtab     : ;
         scrolloff   : ;
         charinsert  : hinsertspace(1);
         chardelete  : hdeletechar(1);
         clearatab   : ;
         cleartabs   : herasetabs;
         clearscreen : herasedisplay;
         scrollon    : ;
         ctrlcodeoff : ;
         ctrlcodeon  : ;
         cursorup    : if saved_row > 0
                       then hsetcursor(saved_row-1, saved_col);
         cursordown  : if saved_row < ssmp_mode[tmaxrow]
                       then hsetcursor(saved_row+1, saved_col);
         cursorleft  : if saved_col > 0
                       then hsetcursor(saved_row, saved_col-1);
         cursorright : if saved_col < ssmp_mode[tmaxcol]
                       then hsetcursor(saved_row, saved_col+1);
         foldoff     : select_rendition(foldoff);
         foldon      : select_rendition(foldon);
         horiztab    : ;
         arrowoff    : select_rendition(arrowoff);
         arrowon     : select_rendition(arrowon);
         lineerase   : herasetoright;
         lineinsert  : hscrolldown(saved_row, ssmp_mode[tmaxrow], 1);
         linedelete  :
            begin
               hscrollup  (saved_row, ssmp_mode[tmaxrow], 1);
               {*************************************************************}
               {                                                             }
               { Temporary fix because of a problem in Newcastle SSMP for PC }
               {                                                             }
               {*************************************************************}
               hsetcursor(saved_row, saved_col)
               {*** end of fix ***}
            end;
         pageerase   : hscrolldown(saved_row, ssmp_mode[tmaxrow],
                                   ssmp_mode[tmaxrow] - saved_row + 1);
         reverself   : ;
         settab      : hsettab(saved_col);
         messoff     : select_rendition(messoff);
         messon      : select_rendition(messon)
      end { case }
   end { ssmp terminal }
   else vdu_string(vdu_table^.control_array[control], 0, 0)
end { vdu_control } ;

{ Position vdu cursor to given place - uses pattern string in vdu record }

procedure vdu_position(vdu_row: row_range; vdu_col: col_range);                 { VDU_POSITION }

begin { vdu_position }
   if term_is_ssmp
   then hsetcursor(vdu_row-1, vdu_col-1)
   else vdu_string(vdu_table^.cursor_string, vdu_row, vdu_col)
end { vdu_position } ;

{ Clear the screen and position home (not all terminals do this automatically) }

procedure do_clear_screen;                                                      { DO_CLEAR_SCREEN }

begin { do_clear_screen }
   vdu_control(clearscreen);
   vdu_position(1, 1)
end { do_clear_screen } ;

{ Print a CR and LF, allowing for SSMP terminals }

procedure print_crlf;                                                           { PRINT_CRLF }

begin { print_crlf }
   if term_is_ssmp then
   begin
      do_ssmp_cr; do_ssmp_lf
   end
   else writeln
end { print_crlf } ;

{ Display a message (possibly highlighted) on the screen, splitting long }
{ messages over several screen lines, with no terminating CR/LF }

procedure no_lf_screen_message(the_message: long_string;                        { NO_LF_SCREEN_MESSAGE }
                               highlight_it: boolean);

var
   screen_row: row_range;
   max_length, split_pos, last_column_of_message: integer;
   temp_message: long_string;
   state: search_type;

   procedure expand_to_max_length(var the_message: long_string);

   begin { expand_to_max_length }
      last_column_of_message := length(the_message);
      if highlight_it and (not vdu_table^.mess_type1)
      then last_column_of_message := last_column_of_message + 1;
      while length(the_message) < max_length
      do insert(' ', the_message, length(the_message)+1)
   end { expand_to_max_length } ;

begin { no_lf_screen_message }
   if in_window_code then
   with vdu_table^ do
   begin
      if continue_screen_messages then
      begin { messages still on }
         repeat
            message_lines_to_clear := message_lines_to_clear + 1;
            if ((message_lines_to_clear mod rows_per_page) = 0) and
               (not in_help_system) then
            begin { pause the output }
               continue_screen_messages :=
                  question_and_answer('--more?--', false, true)
            end { pause the output } ;
            if continue_screen_messages then
            begin { messages still on }
               if message_lines_to_clear <= rows_per_page
               then screen_row := message_lines_to_clear
               else screen_row := rows_per_page;
               vdu_position(screen_row, 1);
               if lineerase_avail
               then vdu_control(lineerase);  { blank out line }
               if highlight_it then
               begin { highlight the message }
                  if mess_type1
                  then max_length := cols_per_row - 1
                  else
                  begin { place attribute off code }
                     { do not use last column in case of scroll }
                     vdu_position(screen_row, cols_per_row-1);
                     vdu_control(messoff);  { highlight off at end }
                     max_length := cols_per_row - 3
                  end { place attribute off code } ;
                  vdu_position(screen_row, 1);
                  vdu_control(messon)   { highlight on at start }
               end { highlighted message }
               else max_length := cols_per_row - 1;
               if length(the_message) > max_length then
               begin { more than one line }
                  split_pos := max_length + 1;
                  state := searching;
                  repeat
                     if the_message[split_pos] = ' '
                     then state := found
                     else if split_pos <= 30
                     then state := not_found
                     else split_pos := split_pos - 1
                  until state <> searching;
                  if state = not_found
                  then split_pos := max_length + 1;
                  temp_message := substr(the_message, 1, split_pos-1);
                  if state = found
                  then split_pos := split_pos + 1;  { skip the space }
                  the_message := substr(the_message, split_pos,
                                        length(the_message)-split_pos+1);
                  if not lineerase_avail
                  then expand_to_max_length(temp_message);
                  if term_is_ssmp
                  then ssmp_literal(temp_message)
                  else write(temp_message);
                  if not lineerase_avail
                  then vdu_position(screen_row, last_column_of_message+1);
                  if highlight_it and vdu_table^.mess_type1
                  then vdu_control(messoff);
                  print_crlf
               end { more than one line }
               else
               begin { fits on one line }
                  if not lineerase_avail
                  then expand_to_max_length(the_message);
                  if term_is_ssmp
                  then ssmp_literal(the_message)
                  else write(the_message);
                  if not lineerase_avail
                  then vdu_position(screen_row, last_column_of_message+1);
                  the_message := ''
               end { fits on one line }
            end { messages still on }
            else the_message  := ''
         until length(the_message) = 0
      end { messages still on }
   end { if in_window_code }
   else write(the_message)
end { no_lf_screen_message } ;

{ As for NO_LF_SCREEN_MESSAGE, but finish it off }

procedure screen_message(the_message: long_string; highlight_it: boolean);      { SCREEN_MESSAGE }

begin { screen_message }
   no_lf_screen_message(the_message, highlight_it);
   if in_window_code then
   begin
      if highlight_it and vdu_table^.mess_type1
      then vdu_control(messoff);
      if continue_screen_messages
      then print_crlf;
      if term_is_ssmp then              { ensure the SSMP user can see the }
      begin { message }
         issue_pending(nopend);         { issue any cursor pos etc. }
         push_frame
      end { message }
   end
   else writeln
end { screen_message } ;

{ Append a given message to the one being constructed }

procedure append_message(addition: long_string);                                { APPEND_MESSAGE }

begin { append_message }
   if length(message) + length(addition) <= line_max_chars
   then message := concat(message, addition)
end { append_message } ;

{ Append a given character to the message being constructed }

procedure append_char(addition: char);                                          { APPEND_CHAR }

begin { append_char }
   append_message(' ');
   message[length(message)] := addition
end { append_char } ;

{ Replace the position marker in a string by a new marker string }

procedure rep_marker(new_marker: long_string; var start_pos: integer);          { REP_MARKER }

begin { rep_marker }
   start_pos := pos('~', message);
   message := concat(
      substr(message, 1, start_pos-1),
      new_marker,
      substr(message, start_pos+1, length(message)-start_pos)
      )
end { rep_marker } ;

{ Insert a "char" at the marker in the given string, using octal if }
{ unprintable.                                                      }

procedure ins_ch(the_char: char);                                               { INS_CH }

var
   start_pos: integer;
   valu: integer;
   digit_no: 1..3;

begin { ins_ch }
   if the_char in ascii_printable then
   begin { printable }
      start_pos := pos('~', message);
      message[start_pos] := the_char
   end { printable }
   else
   begin { not printable }
      rep_marker('xxxx', start_pos);
      message[start_pos] := special[sescape];
      valu := ord(the_char);            { ascii code for char }
      for digit_no := 3 downto 1 do
      begin
         message[start_pos + digit_no] := chr( (valu mod 8) + ord('0') );
         valu := valu div 8
      end { for loop }
   end { not printable }
end { ins_ch } ;

{ Insert a "word" at the marker in the given string }

procedure ins_word(the_word: word; trim_it: boolean);                           { INS_WORD }

var
   start_pos, len: integer;
   windex: word_range;

begin { ins_word }
   len := word_max_chars;
   if trim_it then
   begin
      while (len > 1) and (the_word[len] = ' ')
      do len := len - 1
   end;
   rep_marker(substr('xxxxxxxx',1,len), start_pos);
   for windex := 1 to len
   do message[start_pos + windex - 1] := the_word[windex]
end { ins_word } ;

{ Insert an integer at the marker in the given string }

procedure ins_int(the_int, width: integer);                                     { INS_INT }

var
   int_string: string;
   sindex, start_pos: integer;

begin { ins_int }
   str(the_int, int_string);
   if width >= 0 then
   begin { precede with spaces if needed }
      for sindex := length(int_string)+1 to width
      do insert(' ', int_string, 1)
   end { precede with spaces if needed }
   else
   begin { follow ditto }
      for sindex := length(int_string)+1 to -width
      do int_string := concat(int_string,' ')
   end { follow ditto } ;
   rep_marker(int_string, start_pos)
end { ins_int } ;

{ Print a message after inserting a given string }

procedure str_screen_message(the_string: long_string;                           { STR_SCREEN_MESSAGE }
                             highlight_it: boolean;
                             the_str: long_string);

var
   start_pos: integer;

begin { str_screen_message }
   message := the_string;
   rep_marker(the_str, start_pos);
   screen_message(message, highlight_it)
end { str_screen_message } ;

{ Print a message after inserting a given character }

procedure ch_screen_message(the_string: long_string;                            { CH_SCREEN_MESSAGE }
                            highlight_it: boolean;
                            the_ch: char);

begin { ch_screen_message }
   message := the_string;
   ins_ch(the_ch);
   screen_message(message, highlight_it)
end { ch_screen_message } ;

{ Print a message after inserting a given word }

procedure word_screen_message(the_string: long_string;                          { WORD_SCREEN_MESSAGE }
                              highlight_it: boolean;
                              the_word: word;
                              trim_it: boolean);

begin { word_screen_message }
   message := the_string;
   ins_word(the_word, trim_it);
   screen_message(message, highlight_it)
end { word_screen_message } ;

{ Print a message after inserting a given integer }

procedure int_screen_message(the_string: long_string;                           { INT_SCREEN_MESSAGE }
                             highlight_it: boolean;
                             the_int, width: integer);

begin { int_screen_message }
   message := the_string;
   ins_int(the_int, width);
   screen_message(message, highlight_it)
end { int_screen_message } ;

{ Either read the table for a given vdu_type, or read all the tables and list }
{ their identification fields }

procedure read_vdu_table(just_list: boolean);                                   { READ_VDU_TABLE }

label 227;

var
   table_file: file of vdu_info;
   tables_read: integer;
   done: boolean;

   { test whether vdutables file was produced by correct version of VDUSETUP }

   function test_version: boolean;

   begin { test_version }
      if table_file^.stored_version = info_version
      then test_version := true
      else
      begin { incorrect version number }
         test_version := false;
         main_state := command_error;
         if (table_file^.stored_version > 0) and
            (table_file^.stored_version <= 20)
         then
         begin
            screen_message(concat('The vdu tables in file ',vdu_treename.strng,
               ' were produced by an old version of VDUSETUP and are not ',
               'compatible with this version of the editor.  You can bring ',
               'them up to date by running VDUSETUP in interactive mode.'),
                           true)
         end
         else str_screen_message('File ~ does not contain valid vdu tables.',
                                 true, vdu_treename.strng);
         screen_message('Please contact your System Administrator.', true)
      end { incorrect version number }
   end { test_version } ;

   { scan vdu table and set ssmp mode array to correspond }

   procedure set_mode_array;

   var
      mindex: mode_range;
      ch: char;
      esc_action: action_proc;
      esc_start, index: vdu_char_range;
      esc_number: vdu_char0_range;

      procedure tell_ssmp(chcode: smallint; action: action_proc);

      begin { tell_ssmp }
         case action of
            deleteonech: usual_ssmp_mode[kdelchar ] := chcode;
            deleteonel : usual_ssmp_mode[kdelline ] := chcode;
            godown     :
               begin
                  usual_ssmp_mode[kcsrdown ] := chcode;
                  reply_mode[kcsrdown ] := chcode
               end;
            goleft     :
               begin
                  usual_ssmp_mode[kcsrleft ] := chcode;
                  reply_mode[kcsrleft ] := chcode
               end;
            goright    :
               begin
                  usual_ssmp_mode[kcsrright] := chcode;
                  reply_mode[kcsrright] := chcode
               end;
            goup       :
               begin
                  usual_ssmp_mode[kcsrup   ] := chcode;
                  reply_mode[kcsrup   ] := chcode
               end;
            insertmode :
               if command_line_args.tidy_sw
               then { disable local insertmode switch }
               else usual_ssmp_mode[kinsmode ] := chcode;
            insertonech: usual_ssmp_mode[kinsspac ] := chcode;
            insertonel : usual_ssmp_mode[kinsline ] := chcode;
            tableft    : usual_ssmp_mode[kprevtab ] := chcode;
            tabright   : usual_ssmp_mode[knexttab ] := chcode;
            sinthost   : usual_ssmp_mode[kinthost ] := chcode;
            ssuspend   : usual_ssmp_mode[ksuspend ] := chcode;
            srestart   : usual_ssmp_mode[krestart ] := chcode;
            sleftupd   : usual_ssmp_mode[kleftupd ] := chcode;
            sfirstns   : usual_ssmp_mode[kfirstns ] := chcode;
            salastns   : usual_ssmp_mode[kalastns ] := chcode;
            seraprev   :
               begin
                  usual_ssmp_mode[keraprev ] := chcode;
                  reply_mode[keraprev ] := chcode
               end;
            seraline   : usual_ssmp_mode[keraline ] := chcode;
            seraright  : usual_ssmp_mode[keraright] := chcode;
            shomefld   : usual_ssmp_mode[khomefld ] := chcode;
            snewline   : usual_ssmp_mode[knewline ] := chcode
            otherwise                   { not an ssmp function }
         end { case }
      end { tell_ssmp } ;

   begin { set_mode_array }
      for mindex := kinthost to klast   { clear any existing entries }
      do usual_ssmp_mode[mindex] := 0;
      with vdu_table^ do
      begin
         for ch := chr(nul_code) to char.max do
         begin
            if not (ch in ascii_printable)
            then tell_ssmp(ord(ch)-128, action_array[ord(ch)].char_action)
         end { for loop } ;
         with action_array[esc_code] do
         begin
            esc_action := char_action;
            esc_start := char_pointer.start_index;
            esc_number := char_pointer.number_of_entries
         end { with } ;
         if esc_action in [sequence, posmacro, negmacro] then
         begin
            for index := esc_start downto esc_start-esc_number+1 do
            with two_way.entry_array[index]
            do tell_ssmp(ord(entry_char)-128, entry_rest.char_action)
         end;
         usual_ssmp_mode[ilinerow] := (rows_per_page*3) div 4
      end { with }
   end { set_mode_array } ;

   { Display a line of the list of terminal types }

   procedure display_line(test_type: integer; the_text: string);

   begin { display_line }
      if vdu_type = test_type
      then message := '->  '
      else message := '    ';
      append_message('~   ');
      ins_int(test_type, 3);
      append_message(the_text);
      screen_message(message, false)
   end { display_line } ;

   procedure onunit(not_used: integer);
   begin { onunit }
      if not just_list
      then int_screen_message('Vdu type ~ is not defined.', true, vdu_type, 1);
      main_state := command_error;
      goto 227
   end { onunit } ;

begin { read_vdu_table }
   decide_vdu_type;                     { in case not set by user }
   {---teststart---}
   if not mode_test then
   {---teststop---}
   on('ERROR_TRAP', onunit);
   openread(table_file,vdu_treename.strng);  { usually SYSOVL>VDUTABLES }
   tables_read := 0;
   if just_list then
   begin { just list tables in file }
      done := false;
      repeat
         get(table_file);
         if eof(table_file)
         then done := true
         else
         begin
            tables_read := tables_read + 1;
            if test_version
            then with table_file^ do
            begin
               if tables_read = 1 then
               begin
                  screen_message('Supported terminal types are: ', false);
                  display_line(0, 'Dumb terminal')
               end { first type } ;
               display_line(tables_read, vdu_identification)
            end
            else done := true           { wrong version number }
         end { not eof }
      until done;
      if tables_read = 0
      then screen_message('There are no tables in the file.', true)
   end { just list tables in file }
   else
   begin { load a particular table file }
      get(table_file);                  { just to check version }
      if (main_state = running) and test_version
      then
      begin
         seek(table_file, vdu_type);
         get(table_file);
         if vdu_table = nil
         then new(vdu_table);
         vdu_table^ := table_file^;
         with vdu_table^ do
         begin

            { vital question: Is it an SSMP terminal? }

            term_is_ssmp := (cols_per_row = 1) and (rows_per_page = 1);

            { are various hardware features available? }

            if term_is_ssmp then
            begin { SSMP terminal }
               { set usual rows/cols just for sensible split values - will be }
               { set correctly at SSMP startup }
               cols_per_row := 80;
               rows_per_page := 24;
               ch_ins_del_avail := true;  { but terminal may reject it later }
               line_ins_del_avail := true;
               lineerase_avail := true;
               scrolloff_avail := true;  { not needed in any case }
               horiztab_avail  := false;  { not available as output control }
               backtab_avail   := false;  { ditto }
               tabs_avail      := true;  { but can set screen tabs }
               reverself_avail := false;  { use lineinsert instead }
               ctrlcode_avail  := true;  { not needed }
               fold_avail      :=       { will usually be true }
                  control_array[foldon].string_length > 0;
               pageerase_avail := true;  { we can easily simulate it }
               set_mode_array
            end { SSMP terminal }
            else
            begin { not SSMP terminal }
               ch_ins_del_avail :=
                  (control_array[charinsert].string_length > 0) and
                  (control_array[chardelete].string_length > 0);
               line_ins_del_avail :=
                  (control_array[lineinsert].string_length > 0) and
                  (control_array[linedelete].string_length > 0);
               lineerase_avail := control_array[lineerase ].string_length > 0;
               scrolloff_avail := control_array[scrolloff ].string_length > 0;
               horiztab_avail  := control_array[horiztab  ].string_length > 0;
               backtab_avail   := control_array[backtab   ].string_length > 0;
               tabs_avail      := horiztab_avail or backtab_avail;
               reverself_avail := control_array[reverself ].string_length > 0;
               ctrlcode_avail  := control_array[ctrlcodeon].string_length > 0;
               fold_avail      := control_array[foldon    ].string_length > 0;
               pageerase_avail := control_array[pageerase ].string_length > 0
            end { not SSMP terminal }
         end { with } ;
         calculate_window_sizes
      end { table file alright }
   end { load a particular table file } ;
227:
end { read_vdu_table } ;

begin { pmed_vdu }
end { pmed_vdu } .
