%separate = true
%ecs = true
%ttybuffered = false
%epfargs = true
{ Sheffield University Editor }
{ Copyright (c) University of Sheffield 1990 }

{***************}
{ WINDOW EDITOR }
{***************}

subprogram pmed_wi;

%listing off
%insert pmed_globals.ins.pas
%insert pmed_efs.ins.pas
%insert pmed_procs.ins.pas
%insert pmed_vdu.ins.pas
%insert pmed_ssmp.ins.pas
%insert pmed_wprocs.ins.pas
%listing on

{ Entry point from line editor }

procedure window_code(number_of_args, arg1, arg2: integer;                      { WINDOW_CODE }
                      init_wi: boolean);

label 311;

type
   tidy_type = (tidy_ld, tidy_inspect, tidy_delete, tidy_finish, tidy_quit);

var
   windex: window0_range;
   rindex: row_range;
   main_command_worked, cominput_was_on: boolean;

   procedure read_command_line(var command_line: edline_type;
                               function read_a_char: char);
      extern;

   procedure obey_command_line(command_line: line_chars_type;
                               times: integer;
                               upto: line_chars_plus_range);
      extern;

%page
{*****************************}
{ INTERNAL UTILITY PROCEDURES }
{*****************************}

   { Move "wd_count" words from "from_string[fromsub]" to "to_string[tosub]". }
   { Preserves the words moved by copying either forwards or backwards        }

   procedure movewds(from_string: line_chars_type;
                     fromsub: integer;
                     var to_string: line_chars_type;
                     tosub: integer;
                     wd_count: integer);
      extern;                           { written in PMA }

{---teststart---}

{ Print all details of current window state.    }
{ Only called when using Source Level Debugger. }

procedure printall;                                                             { PRINTALL }

var
   cindex, clength: row_range;

   procedure writearow(the_row: integer);

   begin { writearow }
      with current_window do
      begin
         if the_row <= window_length then
         with window_status[the_row] do
         begin
            write(the_row:5);
            if file_line > 999999
            then write('    big')
            else write(file_line:7);
            write(offset:5);
            case row_status of
               row_normal: write(' norm');
               row_null  : write(' null')
            end;
            if row_drawn
            then write(' d')
            else write('  ')
         end
      end { with }
   end { writearow } ;

begin { printall }
   with current_buffer, current_window do
   begin
      clength := (window_length + 2) div 3;  { split into 3 columns }
      for cindex := 1 to clength do
      begin
         writearow(cindex); write(' |');
         writearow(cindex + clength); write(' |');
         writearow(cindex + clength + clength);
         writeln
      end;
      writeln('cursor_row = ',cursor_row:1,' cursor_col = ',cursor_col:1);
      writeln('need_to_redraw = ',need_to_redraw,
              '  suppress_output ',suppress_output);
      writeln('line_number = ',line_number:1,'  line_length = ',line_length:1,
              '  line_position = ',line_position:1);
      write  ('line_exists = ',line_exists,'  line_altered = ',line_altered,
              '  line_status = ');
      case line_status of
         normal: write('normal');
         top:    write('top   ');
         bottom: write('bottom')
      end;
      writeln;
      writeln('marked_line = ',marked_line:1,
              '  iw_marked_line = ',iw_marked_line:1,
              '  point_line_no = ',point_line_no:1,
              '  point_line_pos = ',point_line_pos:1)
   end
end { printall } ;

{---teststop---}

{ Window code version of Q & A routine - adjusts full duplex }

function wi_question_and_answer(message: long_string;                           { WI_QUESTION_AND_ANSWER }
                                highlight_it: boolean;
                                default_value: boolean): boolean;

var
   need_to_switch: boolean;

begin { wi_question_and_answer }
   { switch echoing back on for answer if necessary }
   need_to_switch := not echoing_on;
   if need_to_switch
   then ctrl_half_duplex(false);
   wi_question_and_answer :=
      question_and_answer(message, highlight_it, default_value);
   if need_to_switch
   then ctrl_half_duplex(true);
end { wi_question_and_answer } ;

{ Unrecognised sequence of characters }

procedure erroract_comm;                                                        { ERRORACT_COMM }

begin { erroract_comm }
   if (macro_head <> nil) and
      (macro_id_char = input_char)
   then display_error('You have not finished storing the macro yet.')
   else display_error('Not recognised.  Type {DEL}H if you need help.')
end { erroract_comm } ;

{ Deal with a user breakin }

{--------------------------------------------------------------------}
{ THIS CODE MUST NOT MAKE ANY DIRECT OR INDIRECT USE OF EFS ROUTINES }
{ WHICH AFFECT THE FILE, BECAUSE EFS IS PROBABLY INTERRUPTED AT AN   }
{ AWKWARD PLACE AND SUCH USE WILL CORRUPT IT.                        }
{--------------------------------------------------------------------}

{ Must only use primitives such as screen_message; display_message does }
{ screen_message and then replace_cursor, which affects EFS }

procedure deal_with_interrupt;                                                  { DEAL_WITH_INTERRUPT }

var
   lindex: integer;

   procedure quit_when_outputting;

   label 322;

   var
      result: boolean;

      procedure internal_quit_onunit(not_used: integer);

      begin { internal_quit_onunit }
         result := false;
         goto 322
      end { internal_quit_onunit } ;

   begin { quit_when_outputting }
      on('QUIT$', internal_quit_onunit);
      screen_message('Quit', true);
      result := wi_question_and_answer('Do you want to continue execution, but with screen updating suppressed? ',
                                    true, true);
   322:
      if result then
      begin { now suppress output }
         screen_message('Suppressing output while execution continues.', false);
         suppress_output := true        { suppresses further output }
      end { now suppress output }
      else
      begin { break in now }
         screen_message('Terminating execution.', false);
         macro_dropout := true;         { stop repeats etc. }
         main_state := quit_signal
      end { break in now }
   end { quit_when_outputting } ;

begin { deal_with_interrupt }
   lowest_empty_buffer := 1;
   if waiting_for_input then            { in get_next_char? }
   begin
      if command_line_args.no_line_sw
      then display_message('Use {DEL}Q if you want to quit.')
      else
      begin { exit allowed }
         main_state := quit_signal;
         goto 311                       { get out immediately }
      end { exit allowed }
   end { waiting for input }
   else quit_when_outputting
end { deal_with_interrupt } ;

procedure read_input_command(echo_on: boolean;                                  { READ_INPUT_COMMAND }
                             var desc: ptr_macro_desc;
                             var action: action_proc);

var
   default_repair, repair: repair_type;
   default_reply, reply: vdu_char_pointer;
   ignore_ch: char;
   the_code, input_code, input_code_level: integer;
   fiddle: record
      case boolean of
         false: (action_part: action_proc);
         true:  (ord_part: shortint)
      end;

   { get the information from a char entry, following sequences if necessary }

   procedure get_char_entry(the_entry: char_entry);

   var
      tindex: vdu_entry_range;
      count: integer;
      state: search_type;

   begin { get_char_entry }
      with vdu_table^, the_entry do
      begin
         action := char_action;
         if action in [sequence, posmacro, negmacro] then
         begin
            input_code_level := input_code_level + 1;
            input_char := get_next_char(input_code_level);
            if (input_char in ascii_printable) and
               (action in [posmacro, negmacro]) then
            with macro_pointers[input_char] do
            begin
               if action = posmacro then
               begin { posmacro }
                  desc := pointers[posm];
                  default_repair := posmacro_repair;
                  default_reply := posmacro_reply
               end { posmacro }
               else
               begin { negmacro }
                  desc := pointers[negm];
                  default_repair := negmacro_repair;
                  default_reply := negmacro_reply
               end { negmacro }
            end { with } ;
            if desc <> nil then
            begin { obey macro }
               repair := default_repair;
               need_to_output := true;
               reply := default_reply;
               desc^.loop_spec.end_condition := once_only  { obey once normally }
            end { desc <> nil }
            else
            with char_pointer do
            begin { not a macro }
               tindex := start_index;
               count := number_of_entries;
               state := searching;
               repeat
                  with two_way.entry_array[tindex] do
                  begin
                     if input_char = entry_char
                     then state := found
                     else if count > 0 then
                     begin
                        count := count - 1;
                        tindex := tindex- 1  { count backwards }
                     end
                     else state := not_found
                  end
               until state <> searching;
               if state = not_found then
               begin { character (sequence) not recognised }
                  action := erroract;
                  repair := default_repair;
                  need_to_output := false;
                  reply.start_index := 1;
                  reply.string_length := 0;
               end { character (sequence) not recognised }
               else get_char_entry(two_way.entry_array[tindex].entry_rest)
            end { DEL/SO but not macro }
         end { sequence/posmacro/negmacro }
         else
         begin { normal action }
            repair := char_repair;
            need_to_output := need_to_output or (char_ware = software);
            reply := char_reply
         end { normal action }
      end { with }
   end { get_char_entry } ;

begin { read_input_command }
   reading_action := true;              { routine may deliver an action }
   input_code_level := 1;
   { "need_to_output" may be set true in "get_ssmp_code", but cannot set it }
   { up yet because macro may still exist but be exhausted }
   need_to_output := false;
   input_code := get_next_code(input_code_level);
   need_to_output := need_to_output or
                     always_output or
                     (current_macro_desc <> nil) or
                     (not echo_on);
   desc := nil;
   if input_code < 0 then
   begin { literal control character }
      input_char := chr(input_code + 256);
      action := litchar
   end { literal control character }
   else if input_code < 256 then
   with current_window, vdu_table^ do
   begin { character read or got from buffer }
      input_char := chr(input_code);
      action := erroract;
      default_repair := ok;
      with default_reply do
      begin
         start_index := 1;
         string_length := 0
      end { with } ;
      { assume that any parity off characters received are special overseas }
      { alphabetics and store in file }
      if ord(input_char) < 128 then
      begin { parity off }
         action := storeonech;
         repair := ok;
         reply  := default_reply
      end { parity off }
      else get_char_entry(action_array[ord(input_char)]);
      if (repair = func) and func_expect_return
      then ignore_ch := get_next_char(1);  { wait for RETURN }
      if (not always_output) and (current_macro_desc = nil) and echo_on
      then
      begin { we need to repair screen }
         vdu_string(reply, 0, 0);       { send reply string if any }
         case repair of
            ok:       ;
            func:     repair_screen(func_repair, func_expect_return)
            otherwise repair_screen(repair, false)
         end
      end { we need to repair screen }
   end { character obtained }
   else
   begin { action code obtained directly }
      with fiddle do
      begin { fiddle to get symbolic type from ord }
         ord_part := input_code - 256;
         action := action_part
      end { fiddle to get symbolic type from ord } ;
      if action in [posmacro, negmacro] then
      begin { get macro id char }
         input_code_level := input_code_level + 1;
         input_char := get_next_char(input_code_level);
         with macro_pointers[input_char]
         do if action = posmacro
            then desc := pointers[posm]
            else desc := pointers[negm];
         if desc = nil then
         begin
            message := 'Macro "~" does not exist.';
            ins_ch(input_char);
            display_message(message);
            macro_dropout := true;
            action := nullact
         end
         else desc^.loop_spec.end_condition := once_only  { obey once normally }
      end { get macro id char }
      else if action = litchar then
      begin { literal char from ssmp }
         input_code_level := input_code_level + 1;
         input_char := get_next_char(input_code_level)
      end { literal char from ssmp }
   end { action obtained directly } ;
   reading_action := false;
   { store code read into macro if necessary }
   if macro_head = nil
   then                                 { not recording a macro }
   else if current_macro_desc = nil then
   begin { store action read into macro }
      if action = storeonech
      then the_code := ord(input_char)  { simple character - store its ord }
      else the_code := ord(action) + 256;  { action code }
      store_macro_code(the_code);
      if action in [posmacro, negmacro, litchar]
      then store_macro_code(ord(input_char))  { store char as well }
   end { store action read into macro }
   else { do not copy characters from macro }
end { read_input_command } ;

%page
{*******************}
{ ACTION PROCEDURES }
{*******************}

{ Break the current line into two lines at the cursor position }

procedure breakline_comm;                                                       { BREAKLINE_COMM }

var
   row_last: row_range;
   copy_line: line_chars_type;
   copy_length: line_chars0_range;
   old_below, new_below: integer;

begin { breakline_comm }
   if check_normal then
   with current_buffer, current_window do
   begin
      if lineerase_avail and window_at_right
      then vdu_control(lineerase);
      find_last(row_last);
      { make copy of part of line after line_position }
      expand_line;
      movewds(line_text, line_position,
              copy_line, 1,
              line_length + 2 - line_position);
      copy_length := line_length - line_position + 1;
      { truncate current line }
      line_text[line_position] := newline;
      line_length := line_position - 1;
      line_altered := true;
      { insert the copied part below the current }
      replace_line;
      line_text := copy_line;
      line_length := copy_length;
      insert_a_line;
      { adjust screen to match new file state }
      if cursor_row = window_length
      then replace_cursor               { rest is off screen - done }
      else
      begin
         old_below := row_last - cursor_row;
         new_below := ( (line_length - 1) div window_width) + 1;
         window_position(cursor_row+1, 1);
         with window_status[cursor_row + 1] do
         begin { set up first new line }
            file_line := line_number;
            offset := 0
         end { set up first new line } ;
         adjust_below(old_below, new_below, cursor_row+1, true, 1);
         draw_file_line(true)
      end;
      if not lineerase_avail or not window_at_right
      then draw_file_line(true)         { draw truncated line correctly now }
   end
end { breakline_comm } ;

{ Split the current line one or more times at the MODE_MARGIN column }

procedure split_at_margin;                                                      { SPLIT_AT_MARGIN }

var
   spaces, break_column, lowest_acceptable: integer;
   state: search_type;
   new_line_position, current_indent, lindex: line_chars_range;

begin { split_at_margin }
   strip_trailing_spaces;
   if token = withterm
   then { SSMP only process when we have token }
   else with current_buffer, current_window do
   begin { may act now }
      while line_position >= (margin_column + 2) do
      begin { keep within margin }
         break_column := margin_column + 1;
         lowest_acceptable := margin_column div 2;
         { move back to a non-space }
         state := searching;
         repeat
            if break_column < lowest_acceptable
            then state := not_found
            else if line_text[break_column] = ' '
            then break_column := break_column - 1
            else state := found
         until state <> searching;
         { move back to a space }
         state := searching;
         repeat
            if break_column < lowest_acceptable
            then state := not_found
            else if line_text[break_column] = ' '
            then state := found
            else break_column := break_column - 1
         until state <> searching;
         if state = found
         then break_column := break_column + 1  { char after last space }
         else break_column := margin_column + 1;
         if mode_indent then
         begin { line up on current indent }
            find_indent(current_indent, window_width);
            if current_indent > 1 then
            begin { it is indented }
               spaces := current_indent - 1;
               if line_length + spaces <= line_max_charsm1 then
               begin
                  for lindex := line_length+1 downto break_column
                  do line_text[lindex+spaces] := line_text[lindex];
                  for lindex := break_column to break_column+spaces-1
                  do line_text[lindex] := ' ';
                  line_length := line_length + spaces;
                  line_altered := true
               end { room for spaces }
               else current_indent := 1  { no room for indenting }
            end { line is indented }
         end { mode_indent }
         else current_indent := 1;
         new_line_position := line_position - break_column + current_indent;
         position_cursor_in_window(false, break_column, false);
         breakline_comm;
         form_draw_file_line(need_to_redraw, true);  { in case > 80 chars etc. }
         move_to_line(line_number + 1);
         position_cursor_in_window(false, new_line_position, false)
      end { while loop }
   end { may act now, with }
end { split_at_margin } ;

{ Carriage Return - go to next file line, one or more screen lines down }

procedure carret_comm;                                                          { CARRET_COMM }

var
   new_col: line_chars_range;
   saved_line_position: line_chars_range;

begin { carret_comm }
   with current_buffer, current_window do
   begin
      if mode_indent
      then find_indent(new_col, window_width);  { indent for old line }
      if need_to_output or term_is_ssmp then
      begin
         window_position(cursor_row, 1);
         if screen_is_split
         then do_cursor_move(cursordown)
         else print_newline
      end;
      cursor_col := 1;
      cursor_row := cursor_row + 1;
      cursor_moved(true, backtab);
      { skip until offset = 0 }
      while window_status[cursor_row].offset <> 0 do
      begin
         do_cursor_move(cursordown);
         cursor_row := cursor_row + 1;
         cursor_moved(true, cursordown)
      end;
      if mode_indent then
      begin
         strip_trailing_spaces;
         if (line_length <> 0) or       { if this line is non-blank or }
            (line_status <> normal)     { .bottom. then }
         then find_indent(new_col, window_width)  { indent up to first non-space }
         else;                          { else line up from previous line }
         if new_col > 1 then
         begin
            cursor_col := new_col;
            replace_cursor;
            cursor_moved(false, backtab)
         end { new_col > 1 }
      end { mode_indent } ;
      if mode_margin then
      begin
         saved_line_position := line_position;
         move_to_line(line_number - 1);
         if line_length < line_max_charsm1
         then line_position := line_length + 1
         else line_position := line_length;
         split_at_margin;
         move_to_line(line_number + 1);
         position_cursor_in_window(false, saved_line_position, false)
      end { mode_margin }
   end { with }
end { carret_comm } ;

{ Delete one character }

procedure deleteonech_comm(allow_redraw: boolean);                              { DELETEONECH_COMM }

begin { deleteonech_comm }
   with current_buffer, current_window do
   if check_normal then
   begin
      if need_to_output and ch_ins_del_avail then
      begin { output }
         vdu_control(chardelete);
         if window_at_right
         then { no further action necessary }
         else
         begin { need to char ins at right of window to realign }
            window_position(cursor_row, window_width);
            vdu_control(charinsert);
            replace_cursor
         end { realign windows }
      end { output } ;
      need_to_redraw := need_to_redraw or (line_length > window_width);
      if need_to_redraw and (token = withterm) then
      begin { SSMP - redraw later }
         if not token_requested
         then hreqtoken                 { Get token and redraw later }
      end { SSMP - redraw later } ;
      if line_position <= line_length then
      begin { cursor is within line }
         if not line_altered
         then push_line;                { save for OOPS command }
         movewds(line_text, line_position + 1,
                 line_text, line_position,
                 line_length - line_position + 1);
         line_length := line_length - 1;
         line_altered := true
      end { cursor is within line } ;
      if ch_ins_del_avail
      then { screen already ok }
      else if allow_redraw
      then form_draw_file_line(true, true)  { redraw whole line if necessary }
      else { action taken by caller }
   end { with, check_normal }
end { deleteonech_comm } ;

   procedure eraseline_comm; forward;

{ Delete one file line - may correspond to more than one screen line }

procedure deleteonel_comm(allow_redraw: boolean);                               { DELETEONEL_COMM }

var
   row_first, row_last, rindex, row_index: row_range;
   start_row, rows_gone: integer;
   saved_file_place: file_places;

begin { deleteonel_comm }
   if check_normal then
   with current_buffer, current_window, saved_file_place do
   begin
      push_line;                        { save for OOPS command }
      if (not window_mode) and (line_number = (iw_marked_line - 1)) then
      begin { IW - deleted last line input }
         cursor_to_line_start;          { move to start of line and do }
         cursor_col := 1;               { line erase instead }
         replace_cursor;
         line_position := 1;
         eraseline_comm
      end { IW - deleted last line input }
      else
      begin
         if need_to_output and line_ins_del_avail and window_full_width
         then do_linedelete(cursor_row, 1);
         if not find_first(line_number, row_first)
         then abort('deleteonel_comm: find_first failed');
         find_last(row_last);
         if cursor_row <> row_first then
         begin { move to first visible part }
            cursor_row := row_first;
            if token = withterm
            then suppress_output := true
            else replace_cursor
         end { move to first visible part } ;
         save_place(saved_file_place);
         { ensure cursor will end up on first row }
         while saved_position > window_width
         do saved_position := saved_position - window_width;
         rows_gone := row_last - row_first + 1;  { visible rows containing line }
         { delete folded parts (if any) }
         if rows_gone > 1 then
         begin
            if token = withterm
            then suppress_output := true
            else if line_ins_del_avail and window_full_width
            then do_linedelete(cursor_row, rows_gone - 1)
         end;
         for rindex := (row_last + 1) to window_length do
         begin { copy status up }
            window_status[rindex - rows_gone] := window_status[rindex];
            with window_status[rindex - rows_gone]
            do file_line := file_line - 1  { and reduce file_line }
         end { copy status up } ;
         delete_line;                   { delete line in file }
         load_a_line(line_number, false);  { ensure gap closed }
         row_index := window_length - rows_gone;
         move_down_screen(row_index);   { fill in window_status }
         if token = withterm then
         begin { redraw later }
            for rindex := window_length - rows_gone + 1 to window_length
            do window_status[rindex].row_drawn := false;
            fill_in_undrawn_rows := true;
            if not token_requested
            then hreqtoken
         end { redraw later }
         else if allow_redraw then
         begin
            if line_ins_del_avail and window_full_width
            then start_row := window_length - rows_gone + 1
            else start_row := cursor_row;
            draw_part_window(start_row, window_length, false)
         end;
         { may have deleted last drawn line }
         with window_status[cursor_row] do
         if (not row_drawn) and (token = withhost) then
         begin
            row_drawn := true;
            draw_window_line(cursor_row, false, true)
         end;
         restore_place(saved_file_place);
         if token = withhost
         then position_cursor_in_window(false, saved_position, false)
      end { not last line for IW }
   end { check_normal }
end { deleteonel_comm } ;

{ Do a QUIT operation }

procedure doquit_comm;                                                          { DOQUIT_COMM }

begin { doquit_comm }
   if always_output
   then ctrl_half_duplex(false);        { echoing on }
   clear_message;
   quit_proc(false);
   if always_output
   then ctrl_half_duplex(true);         { echoing off }
   replace_cursor;
   if main_state = quit_comm
   then exit_window := true
   else if main_state = command_error
   then main_state := running
end { doquit_comm } ;

   procedure wi_replace_proc;
      forward;
   procedure wi_spell_proc;
      forward;
   procedure wi_oops_proc;
      forward;

{ Obey a line of standard editor commands one or more times }

procedure obey_ed_command(var the_line: edline_type;                            { OBEY_ED_COMMAND }
                          var command_worked: boolean);

var
   row_first: row_range;
   saved_file_place: file_places;
   saved_window_number: window_range;
   old_buffer_number, new_buffer_number: buffer_range;
   old_vdu_type, new_vdu_type, new_window_number,
   saved_message_lines_to_clear: integer;
   windex: window_range;
   all_lines_redrawn, need_to_call: boolean;

begin { obey_ed_command }
   with current_buffer, current_window, the_line, vdu_table^ do
   begin
      old_buffer_number := current_buffer_number;
      save_place(saved_file_place);
      form_draw_file_line(need_to_redraw, true);  { ensure current line drawn }
      strip_trailing_spaces;            { to avoid strange effects for user }
      window_array[current_window_number] := current_window;
      if always_output
      then ctrl_half_duplex(false);     { echoing on }
      clear_message;
      vdu_position(1, 1);               { to show the user we are doing }
                                        { something, and also to position }
                                        { for first message line (if any) }
      if term_is_ssmp then
      begin { output to terminal }
         issue_pending(nopend);
         push_frame
      end { output to terminal } ;

      { preset various variables that are updated by command procedures }
      { so that we know which procedures have been used }
      must_redraw_screen := false;
      first_changed_line := maxint;
      last_changed_line := 0;
      split_command_used := false;
      replace_command_given := false;
      spell_command_given := false;
      wi_oops_count := 0;
      old_vdu_type := vdu_type;

      { preset per-buffer variables }
      if screen_is_split then
      begin { preset for each window visible }
         saved_window_number := current_window_number;
         for windex := 1 to window_count do
         begin
            switch_window(windex);
            must_redraw_window := false;
            some_lines_changed := false
         end { for loop } ;
         switch_window(saved_window_number)
      end { preset for each window visible }
      else
      begin { just current buffer }
         must_redraw_window := false;
         some_lines_changed := false
      end { just current buffer } ;

      { set this one last to avoid accidentally setting it true here }
      switch_buffer_used := false;

      { Now obey the command line }

      obey_command_line(ed_text, 1, text_length + 2);

      { Tidy up and allow for any changes made by the user }

      command_worked := main_state = running;
      if main_state <> running then
      begin { some problem }
         macro_dropout := true;         { stop repeats on any error }
         lowest_empty_buffer := 1;
         if main_state = quit_comm
         then exit_window := true       { exit if FILE or QUIT obeyed }
      end { some problem } ;
      edit_mode := true;                { disallow switching to input mode }
      new_buffer_number := current_buffer_number;
      current_window := window_array[current_window_number];

      { set correct echoing state according to terminal type and editing mode }
      if term_is_ssmp then
      begin { ssmp terminal }
         { allow for use of MODE command }
         if ssmp_mode[icharmode] <> ord(mode_insert)
         then hsetmode(icharmode, ord(mode_insert) );
         always_output := false         { never need to echo for SSMP }
      end { ssmp terminal }
      else always_output := mode_margin or mode_direct or
                            screen_is_split or mode_insert;
      if always_output
      then ctrl_half_duplex(true);      { echoing off }

      { Set message status and logout notification according to modes in }
      { case changed by user. }

      if mode_lon
      then lon$cn(1)                    { switch logout notification on  }
      else lon$cn(0);                   { switch logout notification off }
      set_defer(false, true);           { set new message status }

      { if lost marked line or switched buffers then switch to WI mode }
      if (iw_marked_line = 0) or (new_buffer_number <> old_buffer_number)
      then window_mode := true;

      { do not allow the cursor below the marked line in IW mode }
      if not window_mode and (line_number >= iw_marked_line)
      then move_to_line(iw_marked_line - 1);

      { do not leave on .null. line }
      if not line_exists
      then load_a_line(line_number, false);

      if message_lines_to_clear > 0 then
      begin { message was displayed }
         continue_screen_messages := true;
         if message_lines_to_clear < rows_per_page
         then screen_message('', false)  { add blank line to separate }
         else screen_message('Press RETURN to redraw the screen.', false)
      end { message was displayed } ;

      { if hit top or bottom or other error return to original file position }
      { in original buffer }
      if (main_state = top_or_bottom) or (main_state = command_error) then
      begin { reset position }
         main_state := running;
         if new_buffer_number <> old_buffer_number then
         begin { buffer changed }
            switch_buffer(old_buffer_number);
            new_buffer_number := old_buffer_number
         end { buffer changed } ;
         restore_place(saved_file_place)
      end { reset position } ;

      { If we were in IW mode and have switched buffers then clear IW flag }
      { in old buffer }
      if not window_mode and (new_buffer_number <> old_buffer_number) then
      begin { clear flag }
         switch_buffer(old_buffer_number);
         must_redraw_window := true;    { display switch to WI mode }
         iw_marked_line := 0;           { clear IW marked line in old buffer }
         switch_buffer(new_buffer_number)
      end { clear flag } ;

      { if vdu type changed then tidy up }
      if (vdu_type <> old_vdu_type) or (vdu_table = nil) then
      begin { vdu_type changed }
         new_vdu_type := vdu_type;
         vdu_type := old_vdu_type;
         read_vdu_table(false);
         vdu_string(exit_string, 0, 0);  { send exit string for old type }
         vdu_type := new_vdu_type;
         read_vdu_table(false);
         must_redraw_screen := true
      end { vdu_type changed } ;

      { Deal with user typing SPLIT command }

      if split_command_used then
      begin { user has reset screen split }
         screen_tabs_set := false;
         must_redraw_screen := true
      end { user has reset screen split } ;

      { Decide whether to switch windows }

      if switch_buffer_used or split_command_used then
      begin { buffer changed }
         new_window_number := 0;
         for windex := 1 to window_count
         do if window_array[windex].window_buffer = new_buffer_number
            then new_window_number := windex;
         if new_window_number > 0 then
         begin { new buffer is part of split screen }
            screen_is_split := true;
            if (current_window_number > 0) and (not must_redraw_screen) then
            begin { switching from one window to another }
               switch_window(new_window_number);
               saved_message_lines_to_clear := message_lines_to_clear;
               { we need to loop round all windows and redraw any that }
               { the user has changed }
               for windex := 1 to window_count do
               begin
                  switch_window(windex);
                  { if any lines changed we must redraw the whole window }
                  { because we only record one set of changed line numbers }
                  if some_lines_changed
                  then must_redraw_window := true;
                  if must_redraw_window
                  then need_to_call := true
                  else with window_status[cursor_row] do
                  begin { has file position moved? }
                     need_to_call := (file_line <> line_number) or
                                     ((cursor_col + offset) <> line_position)
                  end { has file position moved? } ;
                  if need_to_call then
                  begin { sort out this window }
                     { if there is a message on the screen cancel it because }
                     { message clearing in position_cursor_in_window will }
                     { cause havoc }
                     if message_lines_to_clear > 0 then
                     begin { blank out message area }
                        clear_message_area;
                        message_lines_to_clear := 0
                     end { blank out message area } ;
                     position_cursor_in_window(must_redraw_window,
                                               line_position, true);
                  end { sort out this window } ;
                  must_redraw_window := false
               end { for loop } ;
               switch_window(new_window_number);
               if (message_lines_to_clear = 0) and
                  (saved_message_lines_to_clear > 0) then
               begin { clear the message now }
                  message_lines_to_clear := saved_message_lines_to_clear;
                  clear_message
               end { clear the message now }
            end { switching from one window to another }
            else
            begin { screen becomes split }
               switch_window(new_window_number);
               screen_tabs_set := false;
               must_redraw_screen := true
            end { screen becomes split }
         end { new buffer is part of split screen }
         else
         begin { new buffer not in a window }
            screen_is_split := false;
            { must make sure that buffer number is correct in [0] before }
            { switching because switch_window calls switch_buffer }
            if current_window_number = 0 then
            begin { already in [0] }
               if window_buffer = new_buffer_number
               then { buffer was switched but back to original now }
               else
               begin { buffer changed }
                  window_buffer := new_buffer_number;
                  switch_buffer(window_buffer);
                  must_redraw_screen := true
               end { buffer changed }
            end { already in [0] }
            else
            begin { need to switch windows }
               window_array[0].window_buffer := new_buffer_number;
               switch_window(0);
               screen_tabs_set := false;
               must_redraw_screen := true
            end { need to switch windows }
         end { new buffer not in a window }
      end { buffer changed }
      else
      begin { buffer not changed }

         { redraw any lines changed by e.g. CHANGE command }

         if must_redraw_window or must_redraw_screen
         then { no point }
         else if some_lines_changed then
         begin { some range of lines changed }
            if message_lines_to_clear < rows_per_page then
            begin { at least one file line visible }
               save_place(saved_file_place);

               { omit any lines off top of window }
               with window_status[1]
               do if first_changed_line < file_line
                  then first_changed_line := file_line;

               { there is no need to redraw any lines obliterated by messages }
               if message_lines_to_clear >= window_top then
               begin { messages have covered some of window }
                  if message_lines_to_clear >= window_bottom
                  then first_changed_line := maxint  { all obliterated }
                  else with window_status[message_lines_to_clear+2-window_top] do
                  begin { just part covered }
                     if first_changed_line < file_line
                     then first_changed_line := file_line;
                  end { just part covered }
               end { messages have covered some of window } ;

               all_lines_redrawn := false;
               repeat
                  if first_changed_line > last_changed_line
                  then all_lines_redrawn := true
                  else if find_first(first_changed_line, row_first) then
                  begin
                     move_to_line(first_changed_line);
                     form_draw_file_line(true, false);
                     first_changed_line := first_changed_line + 1
                  end
                  else all_lines_redrawn := true
               until all_lines_redrawn;
               restore_place(saved_file_place)
            end { screen not full of messages }
         end { some range of lines changed }
      end { buffer not changed } ;

      if must_redraw_screen             { send terminal reset string }
      then vdu_string(reset_string, 0, 0);

      if main_state = quit_signal
      then { redraw later }
      else if must_redraw_screen then
      begin { redraw whole screen }
         setup_whole_screen(-(window_length div 2),
                             (window_length - 1) div 2);
         draw_whole_screen
      end { redraw whole screen }
      else position_cursor_in_window(must_redraw_window, line_position, true);

      if display_buffer_name then
      begin { show buffer name on screen }
         announce_buffer_name;
         replace_cursor
      end { show buffer name on screen } ;
      if not screen_tabs_set then
      begin { need to set tabs }
         set_screen_tabs;               { set new tabs if changed }
         replace_cursor
      end { need to set tabs } ;
      if replace_command_given
      then wi_replace_proc;
      if spell_command_given
      then wi_spell_proc;
      if wi_oops_count > 0
      then wi_oops_proc
   end { with }
end { obey_ed_command } ;

{ Obey a line of commands given as a string }

procedure obey_ed_str(the_commands: long_string;                                { OBEY_ED_STR }
                      var command_worked: boolean);

var
   temp_line: edline_type;
   chindex: line_chars_range;

begin { obey_ed_str }
   with temp_line do
   begin
      if special[ssemicolon] <> ';' then
      begin { allow for non-standard semi symbol }
         for chindex := 1 to length(the_commands)
         do if the_commands[chindex] = ';'
            then the_commands[chindex] := special[ssemicolon]
      end { allow for non-standard semi symbol } ;
      for chindex := 1 to length(the_commands)
      do ed_text[chindex] := the_commands[chindex];
      ed_text[length(the_commands)+1] := newline;
      text_length := length(the_commands);
      being_obeyed := false
   end { with } ;
   obey_ed_command(temp_line, command_worked)
end { obey_ed_str } ;

{ Obey a line of commands after substituting an integer value }

procedure obey_ed_int(the_commands: long_string; the_int: integer);             { OBEY_ED_INT }

var
   command_worked: boolean;

begin { obey_ed_int }
   message := the_commands;
   ins_int(the_int, 1);
   obey_ed_str(message, command_worked)
end { obey_ed_int } ;

{ Read and obey a line of ordinary editor commands }

procedure edcom_comm;                                                           { EDCOM_COMM }

var
   edcom_col: 1..maxint;
   the_command: long_string;
   command_worked, cr_given: boolean;

   function extract_next: char;

   begin { extract_next }
      if edcom_col > length(the_command)
      then extract_next := newline
      else
      begin
         extract_next := the_command[edcom_col];
         edcom_col := edcom_col + 1
      end
   end { extract_next } ;

begin { edcom_comm }
   clear_message;
   get_long_string('Give command: ', the_command, cr_given,
                   false, true, true, false);
   clear_message;
   if cr_given and (length(the_command) > 0) then
   begin
      edcom_col := 1;
      read_command_line(ed_com_line, extract_next);
      ed_com_line_valid := ed_command;
      obey_ed_command(ed_com_line, command_worked)
   end
   else display_message('Function abandoned.')
end { edcom_comm } ;

{ End of macro definition; either by user typing end-of-macro code or at exit }
{ from screen editor }

procedure emacro_comm(by_user: boolean);                                        { EMACRO_COMM }

begin { emacro_comm }
   if current_macro_desc <> nil
   then                                 { do nothing if obeyed from macro }
   else if macro_head = nil
   then display_error('Not storing a macro.')
   else
   begin
      new(macro_pointers[macro_id_char].pointers[macro_id_type]);
      with macro_pointers[macro_id_char].pointers[macro_id_type]^ do
      begin
         code_start := macro_head;
         macro_head := nil;
         total_count := store_macro_count;
         if by_user
         then total_count := total_count - 1;  { exclude emacro code }
         macro_being_obeyed := false
      end
   end
end { emacro_comm } ;

{ Erase from cursor to end of line }

procedure eraseline_comm;                                                       { ERASELINE_COMM }

var
   saved_line_length: line_chars0_range;

begin { eraseline_comm }
   if check_normal then
   with current_buffer, current_window do
   begin
      push_line;                        { save for OOPS command }
      if line_position > line_length
      then { no action needed }
      else
      begin { within line }
         if need_to_output and lineerase_avail and window_at_right
         then vdu_control(lineerase);
         saved_line_length := line_length;
         line_length := line_position - 1;
         line_text[line_position] := newline;
         line_altered := true;
         if (saved_line_length <= window_width) and
            lineerase_avail and window_at_right
         then { display is already OK }
         else
         begin { need to update screen }
            if token = withterm then
            begin { SSMP - cannot act now }
               need_to_redraw := true;
               if not token_requested
               then hreqtoken           { Get token and redraw later }
            end { SSMP - cannot act now }
            else form_draw_file_line(true, true)  { show rest of long line }
         end { need to update screen }
      end { within line }
   end { with }
end { eraseline_comm } ;

{ Exit from window editor unless -Noexit or -Screen was given }

procedure exitwi_comm;                                                          { EXITWI_COMM }

begin { exitwi_comm }
   if command_line_args.no_line_sw
   then display_message('Not allowed.  Type {DEL}H if you need help.')
   else exit_window := true
end { exitwi_comm } ;

{ Do a FILE or SAVE operation }

procedure dofile_or_dosave_proc(fors: boolean);                                 { DOFILE_OR_DOSAVE_PROC }

var
   new_treename: tree_type;
   cr_given, result: boolean;

begin { dofile_or_dosave_proc }
   with current_buffer, main_file do
   begin
      message := 'Give filename: ';
      if treename.len > 0
      then insert(concat(' (default ',treename.strng,')'), message, 14)
   end;
   clear_message;
   get_treename(message, new_treename, cr_given, true, false, false, true);
   if macro_dropout
   then { error in treename }
   else
   begin { treename alright }
      clear_message;
      if cr_given then
      begin { go ahead and save it }
         if always_output
         then ctrl_half_duplex(false);  { echoing on }
         result := save_proc(new_treename, true);
         if always_output
         then ctrl_half_duplex(true);   { echoing off }
         if fors then
         begin
            if result then
            begin { FILE - quit after successful save }
               if always_output
               then ctrl_half_duplex(false);  { echoing on }
               quit_proc(false);
               if always_output
               then ctrl_half_duplex(true);  { echoing off }
               if main_state = quit_comm
               then exit_window := true
            end { FILE - quit after successful save }
         end;
         if main_state = command_error
         then main_state := running;
      end { cr_given }
      else display_message('Function abandoned.')
   end { treename alright } ;
   replace_cursor                       { in case of any error messages }
end { dofile_or_dosave_proc } ;

{ Cursor down or linefeed }

procedure godown_comm(meta_call: boolean;                                       { GODOWN_COMM }
                      arrow_type: controls);

var
   saved_macro_dropout, was_normal_line, result: boolean;

begin { godown_comm }
   with current_buffer, current_window, vdu_table^ do
   begin
      saved_macro_dropout := macro_dropout;
      was_normal_line := line_status = normal;
      if need_to_output then
      begin { echo it }
         if screen_is_split or (arrow_type = cursordown)
         then do_cursor_move(cursordown)
         else print_newline
      end { echo it } ;
      cursor_row := cursor_row + 1;
      cursor_moved(true, arrow_type);
      if meta_call then
      begin { go to bottom of file }
         macro_dropout := saved_macro_dropout;  { avoid stopping if on line 1 }
         if window_mode then
         begin
            if main_file.file_is_open and not suppress_output
            then display_message('Going to bottom of file.');
            move_to_line(large_int)     { go to .bottom. }
         end
         else move_to_line(iw_marked_line -1);  { IW - just go to before mark }
         if main_state = quit_signal
         then                           { redraw later }
         else position_cursor_in_window(false, 1, false)
      end { go to bottom of file }
      else if not was_normal_line
      then result := check_normal       { object to cursor down at .bottom. }
   end { with }
end { godown_comm } ;

{ Cursor home }

procedure gohome_comm;                                                          { GOHOME_COMM }

begin { gohome_comm }
   with current_window, vdu_table^.home_location do
   begin
      cursor_row := home_row;
      cursor_col := home_col;
      if cursor_row > window_length
      then cursor_row := window_length;
      if cursor_col > window_width
      then cursor_col := window_width;
      if need_to_output
      then window_position(cursor_row, cursor_col)
   end;
   cursor_moved(true, backtab)
end { gohome_comm } ;

{ Cursor left }

procedure goleft_comm(meta_call: boolean);                                      { GOLEFT_COMM }

var
   saved_file_place: file_places;
   saved_macro_dropout: boolean;

begin { goleft_comm }
   if check_normal then
   with current_buffer, current_window do
   begin
      save_place(saved_file_place);
      saved_macro_dropout := macro_dropout;
      if need_to_output
      then do_cursor_move(cursorleft);
      cursor_col := cursor_col - 1;
      cursor_moved(false, cursorleft);
      if meta_call or                   { DEL cursorleft or }
         (saved_file_place.saved_position = 1)  { left at position 1 of file line }
      then
      begin
         restore_place(saved_file_place);  { in case cursor left went onto
                                              previous line }
         if meta_call
         then macro_dropout := saved_macro_dropout  { in case hit .top. }
         else macro_dropout := true;    { cursor left at column 1 }
         position_cursor_in_window(false, 1, false)  { move to column 1 }
      end
   end
end { goleft_comm } ;

{ Cursor right }

procedure goright_comm(meta_call: boolean);                                     { GORIGHT_COMM }

var
   new_position: line_chars_range;

begin { goright_comm }
   if check_normal then
   with current_buffer, current_window do
   begin
      if need_to_output
      then do_cursor_move(cursorright);
      cursor_col := cursor_col + 1;
      cursor_moved(false, cursorright);
      if (line_position >= line_max_charsm1) and (token = withhost)
      then position_cursor_in_window(false, line_position, false);
      if meta_call then
      begin { move cursor to last char of line + 1 }
         strip_trailing_spaces;
         if line_length < line_max_charsm1
         then new_position := line_length + 1
         else new_position := line_length;  { on last char for full line }
         position_cursor_in_window(false, new_position, false)
      end { move cursor to last char of line + 1 }
   end { check_normal, with }
end { goright_comm } ;

{ Cursor up or reverse linefeed }

procedure goup_comm(meta_call: boolean;                                         { GOUP_COMM }
                    arrow_type: controls);

var
   saved_macro_dropout: boolean;

begin { goup_comm }
   with current_window do
   begin
      saved_macro_dropout := macro_dropout;
      if need_to_output then
      begin { echo it }
         if screen_is_split or (arrow_type = cursorup)
         then do_cursor_move(cursorup)
         else vdu_control(reverself)
      end { echo it } ;
      cursor_row := cursor_row - 1;
      cursor_moved(true, arrow_type);
      if meta_call then
      begin { go to top of file }
         move_to_line(1);
         position_cursor_in_window(false, 1, false);
         do_cursor_move(cursorup);      { ensure .top. displayed }
         cursor_row := cursor_row - 1;
         cursor_moved(true, cursorup);
         macro_dropout := saved_macro_dropout  { avoid stopping accidentally }
      end { meta_call }
   end { with }
end { goup_comm } ;

{ Put some help text onto the screen }

procedure helpact_comm;                                                         { HELPACT_COMM }

begin { helpact_comm }
   if always_output
   then ctrl_half_duplex(false);        { echoing on }
   do_clear_screen;
   help_system('SCREEN');
   draw_whole_screen;
   if always_output
   then ctrl_half_duplex(true)          { echoing off }
end { helpact_comm } ;

{ Insert one character }

procedure insertonech_comm(allow_redraw: boolean);                              { INSERTONECH_COMM }

var
   try_position: integer;
   saved_line_position: line_chars_range;

   procedure delete_at_right_edge;

   begin { delete_at_right_edge }
      with current_window do
      begin
         window_position(cursor_row, window_width);
         vdu_control(chardelete);
         replace_cursor
      end { with }
   end { delete_at_right_edge } ;

begin { insertonech_comm }
   if check_normal then
   with current_buffer, current_window, vdu_table^ do
   if line_position <= line_length then  { no action if after line_length }
   begin
      if need_to_output and ch_ins_del_avail then
      begin
         if not window_at_right
         then delete_at_right_edge
         else if data_protect then
         begin
            { calculate line position for RH edge of window }
            try_position := window_width + window_status[cursor_row].offset;
            if try_position <= line_max_charsm1 then
            begin
               saved_line_position := line_position;
               line_position := try_position;
               expand_line;
               if line_text[try_position] <> ' '
               then delete_at_right_edge;
               line_position := saved_line_position
            end
         end { data_protect } ;
         vdu_control(charinsert)
      end { output code } ;
      if line_length < line_max_charsm1 then
      begin { there is room }
         if not line_altered
         then push_line;                { save for OOPS command }
         movewds(line_text, line_position,
                 line_text, line_position + 1,
                 line_length + 2 - line_position);
         if (line_length = window_width)and  { allow for just spaces at end }
            (line_text[line_length + 1] = ' ')  { of line }
         then line_text[line_length + 1] := newline
         else line_length := line_length + 1;
         line_text[line_position] := ' ';
         line_altered := true;
         need_to_redraw := need_to_redraw or (line_length > window_width)
      end { there is room in line }
      else
      begin { no room in line }
         strip_trailing_spaces;
         if line_length < line_max_charsm1
         then insertonech_comm(allow_redraw)
         else
         begin { still no space }
            if not suppress_output then
            begin { message and suppress }
               display_error('File line is full.');
               suppress_output := true;
               need_to_redraw := true
            end { message and suppress }
         end { still no space }
      end { no room in line } ;
      if need_to_redraw and (token = withterm) then
      begin { SSMP - redraw later }
         if not token_requested
         then hreqtoken                 { Get token and redraw later }
      end { SSMP - redraw later } ;
      if ch_ins_del_avail
      then { screen already ok }
      else if allow_redraw
      then form_draw_file_line(true, true)  { redraw whole line if necessary }
      else { action taken by caller }
   end { check_normal, with, within line }
end { insertonech_comm } ;

   procedure obey_input_command(desc: ptr_macro_desc;
                                action: action_proc);
      forward;

{ Toggle insert mode.  If entering insert mode (and not SSMP) then echoing }
{ needs to be switched off. }

procedure insertmode_comm;                                                      { INSERTMODE_COMM }

begin { insertmode_comm }
   mode_insert := not mode_insert;
   if term_is_ssmp then
   begin { SSMP: just inform terminal }
      hsetmode(icharmode, ord(mode_insert) )
   end { SSMP }
   else with current_buffer, current_window do
   begin
      if current_macro_desc = nil then
      begin { not obeying a macro }
         if mode_insert
         then display_message('Insert mode.')
         else display_message('Overtype mode.')
      end { not obeying a macro } ;
      { Set echoing off when in insert mode }
      always_output := mode_margin or mode_direct or
                       screen_is_split or mode_insert;
      ctrl_half_duplex(always_output)
   end { not SSMP, with }
end { insertmode_comm } ;

{ Insert one (blank) line on the screen and in the file.                    }
{ It is inserted before the first part of the file_line, which may          }
{ involve scrolling the screen to find it. The "new_text" argument controls }
{ whether the new line is blank, duplicate or from stack                    }

procedure insertonel_comm(new_text: insertonel_type; allow_redraw: boolean);    { INSERTONEL_COMM }

var
   rindex: row_range;
   ok_to_continue, needop: boolean;
   newcol: line_chars_range;
   copy_line: line_chars_type;
   copy_length: line_chars0_range;
   saved_line_position: line_chars_range;

begin { insertonel_comm }
   if new_text = do_duplicate
   then ok_to_continue := check_normal  { cannot duplicate .bottom. }
   else ok_to_continue := true;
   if ok_to_continue then
   with current_buffer, current_window, vdu_table^ do
   begin
      needop := need_to_output;
      { if user has inserted inside fold and already done on screen }
      { then undo lineinsert }
      if (window_status[cursor_row].offset >0)
         and (not needop) then
      begin
         if token = withterm then
         begin { do it later }
            if not token_requested
            then hreqtoken;
            suppress_output := true
         end { do it later }
         else
         begin { do it now }
            vdu_control(linedelete);    { cancel it }
            needop := true;             { now need to reoutput }
            draw_window_line(window_length, true, true)  { replace bottom row }
         end { do it now }
      end;
      if new_text = do_duplicate then
      begin { make a copy of the current line }
         copy_line := line_text;
         copy_length := line_length;
         saved_line_position := line_position
      end { make a copy of the current line } ;
      if new_text = do_oops
      then saved_line_position := line_position;
      if token = withhost then
      begin
         cursor_to_line_start;          { cursor -> start of file line }
         if needop and line_ins_del_avail and window_full_width
         then do_lineinsert(cursor_row, 1);
      end;
      { save line_position now to avoid cursor ending up on folded part of }
      { blank line and confusing the user }
      if new_text = do_blank
      then saved_line_position := line_position;

      { [cursor_row] is now ok - increment line numbers for rest of screen and }
      { move down }

      for rindex := window_length downto (cursor_row + 1) do
      begin
         window_status[rindex] := window_status[rindex - 1];
         with window_status[rindex]
         do file_line := file_line + 1
      end;

      move_to_line(line_number - 1);    { insert a blank file line before }
      if mode_indent                    { current line }
      then find_indent(newcol, window_width);
      insert_blank_line;
      if new_text = do_blank then
      begin { do_blank }
         if token = withterm
         then line_position := saved_line_position
         else if mode_indent then
         begin
            line_position := newcol;
            if line_ins_del_avail and window_full_width
            then position_cursor_in_window(false, line_position, false)
         end
         else position_cursor_in_window(false, saved_line_position, false);
         if (not line_ins_del_avail or not window_full_width) and
            allow_redraw and (token = withhost) then
         begin
            draw_part_window(cursor_row, window_length, false);
            replace_cursor;
            if mode_indent
            then position_cursor_in_window(false, newcol, false)
         end { draw part window } ;
         { SSMP: if only part window in update limits we will have lost a line }
         if (token = withterm) and (last_update_row < window_length) then
         begin { replace lines later }
            window_status[last_update_row + 1].row_drawn := false;
            fill_in_undrawn_rows := true;
            if not token_requested
            then hreqtoken
         end { replace lines later }
      end { do_blank }
      else
      begin { duplicate or oops }
         if new_text = do_duplicate then
         begin { duplicate - use copy }
            line_text := copy_line;
            line_length := copy_length;
            line_altered := true
         end { duplicate - use copy }
         else pop_line;                 { OOPS - get line from stack }
         if line_ins_del_avail and window_full_width
         then form_draw_file_line(true, true)  { display new line }
         else
         begin
            rindex := cursor_row;
            move_down_screen(rindex);
            draw_part_window(cursor_row, window_length, false);
            replace_cursor
         end;
         position_cursor_in_window(false, saved_line_position, false)
      end { duplicate or oops }
   end { with }
end { insertonel_comm } ;

{ Join the next file line onto the current at the cursor position }

procedure joinline_comm;                                                        { JOINLINE_COMM }

var
   row_last: row_range;
   copy_line: line_chars_type;
   saved_line_position: line_chars_range;
   copy_length: line_chars0_range;
   old_below, new_below: integer;

begin { joinline_comm }
   if check_normal then
   with current_buffer, current_window do
   begin
      push_line;                        { save for OOPS command }
      saved_line_position := line_position;
      move_to_line(line_number + 1);
      if (not window_mode) and (line_number = iw_marked_line)
      then display_error('You are on the last IW line, and cannot join to the next.')
      else if line_status = bottom
      then display_error('Cannot join to .bottom.')
      else
      begin
         if window_status[window_length].file_line < line_number
         then row_last := window_length  { line below is not in window }
         else find_last(row_last);
         { copy and delete the line below }
         copy_line := line_text;
         copy_length := line_length;
         if copy_length + saved_line_position > line_max_chars
         then display_error('New line would be too long.')
         else
         begin
            delete_line;
            { append to original line }
            move_to_line(line_number - 1);
            line_position := saved_line_position;
            expand_line;
            movewds(copy_line, 1,
                    line_text, saved_line_position,
                    copy_length);
            line_length := saved_line_position + copy_length - 1;
            line_text[line_length + 1] := newline;
            line_altered := true;
            { adjust screen status to match new file state }
            if cursor_row <> window_length then
            begin { code only needed if not last row }
               new_below := (line_length - window_status[cursor_row].offset - 1)
                           div window_width;
               old_below := row_last - cursor_row;
               window_position(cursor_row+1, 1);
               if new_below > 0 then
               with window_status[cursor_row + 1] do
               begin { set up first row status }
                  file_line := line_number;
                  offset := window_status[cursor_row].offset + window_width
               end { set up first row status } ;
               adjust_below(old_below, new_below, cursor_row+1, true, -1);
            end { not last screen row } ;
            draw_file_line(true)
         end { result not too long }
      end { next line not .bottom. }
   end { if check_normal }
end { joinline_comm } ;

{ Load a file above the current line }

procedure lfile_comm;                                                           { LFILE_COMM }

var
   new_line_number, old_line_number, overlap: integer;
   row_first, row_index: row_range;
   done, redraw: boolean;

begin { lfile_comm }
   if read_load_treename then
   with current_buffer, current_window, vdu_table^ do
   begin
      cursor_to_line_start;
      cursor_col := 1;
      if (current_macro_desc = nil) and (not suppress_output)
      then display_message('Loading... ')
      else replace_cursor;
      old_line_number := line_number;   { line number before loading }
      move_to_line(line_number - 1);
      done := load_a_file;
      if line_status = top
      then load_a_line(1, false)
      else move_to_line(line_number + 1);
      if not done
      then replace_cursor               { failed }
      else
      begin
         new_line_number := line_number;  { line number after loading }
         with window_status[cursor_row] do
         begin
            file_line := old_line_number;
            offset := 0;
            row_status := row_normal;
            row_drawn := true
         end;
         row_index := cursor_row;
         move_down_screen(row_index);
         move_to_line(new_line_number);  { back to correct place in file }
         overlap := window_length div 4;
         if find_first(new_line_number, row_first)
         then redraw := false           { new place still on screen }
         else if new_line_number >
                 window_status[window_length].file_line + overlap
         then redraw := true            { too far off screen }
         else
         begin { only a little way off screen }
            redraw := false;
            row_first := window_length + 1
         end { only a little way off screen } ;
         if full_screen and line_ins_del_avail and window_full_width and
            not redraw then
         begin { do by line ins/del }
            clear_message;              { clear & replace cursor }
            if (row_first - cursor_row) <= max_ins_del then
            begin { do by lineinsert }
               do_lineinsert(cursor_row, row_first - cursor_row);
               for row_index := cursor_row to (row_first - 1) do
               begin
                  draw_window_line(row_index, false, true)
               end
            end { do by lineinsert }
            else
            begin { do by redraw }
               window_position(cursor_row, 1);
               draw_part_window(cursor_row, window_length, false)
            end { do by redraw } ;
            replace_cursor;
            move_to_line(new_line_number);
            position_cursor_in_window(false, 1, false)
         end { do by line ins/del }
         else form_draw_window(false);
         { if buffer used it may be visible on split screen so redraw }
         with load_file
         do if load_use = use_buffer
            then redraw_buffer_if_visible(load_buffer_number, true)
      end { file alright }
   end { with }
end { lfile_comm } ;

{ Mark the current line on the screen for use in unload functions etc. }

procedure markit_comm(meta_switch: boolean);                                    { MARKIT_COMM }

begin { markit_comm }
   with current_buffer, current_window do
   if meta_switch then
   begin { switch on mark }
      if line_status = bottom
      then display_error('Cannot mark .bottom.')
      else
      begin
         unmark;
         marked_line := line_number;
         cursor_to_line_start;          { in case on fold }
         mark_the_line(cursor_row)
      end
   end { switch on mark }
   else unmark                          { switch off mark }
end { markit_comm } ;

{ Cursor up or down half a screen }

procedure move_many_comm(direction: controls);                                  { MOVE_MANY_COMM }

var
   distance, lines: integer;
   done: boolean;

begin { move_many_comm }
   with current_buffer, current_window do
   begin
      distance := window_length div 2;  { half number of lines in window }
      lines := 1;
      done := false;
      repeat
         if direction = cursordown
         then if line_status = bottom   { avoid 'not allowed at bottom' msg }
              then macro_dropout := true
              else godown_comm(false, direction)
         else goup_comm(false, direction);
         if lines < distance
         then lines := lines + 1
         else done := true
      until done or macro_dropout
   end { with } ;
   adjust_outside_margins               { do not leave too near top or bottom }
end { move_many_comm } ;

{ Move down one windowful }

procedure downscreen_comm;                                                      { DOWNSCREEN_COMM }

var
   target_line: integer;

begin { downscreen_comm }
   with current_buffer, current_window do
   begin
      target_line := window_status[window_length].file_line;
      if (not window_mode) and (target_line >= iw_marked_line)
      then target_line := iw_marked_line - 1;
      move_to_line(target_line);
      form_draw_args(-screen_margin, window_length - screen_margin - 1, false)
   end
end { downscreen_comm } ;

{ Move up one screenful }

procedure upscreen_comm;                                                        { UPSCREEN_COMM }

begin { upscreen_comm }
   with current_buffer, current_window do
   begin
      move_to_line(window_status[1].file_line);
      if line_status = top
      then move_to_line(1);
      form_draw_args(screen_margin - window_length + 1, screen_margin, false)
   end { with }
end { upscreen_comm } ;

{ Insert some blank lines or delete some blank lines }

procedure openit_comm(meta_switch: boolean);                                    { OPENIT_COMM }

var
   lindex, lines_to_delete: integer;
   saved_file_place: file_places;
   row_index: row_range;

begin { openit_comm }
   with current_buffer, current_window do
   if meta_switch then
   begin { insert 12 blank lines }
      for lindex := 1 to 12
      do insertonel_comm(do_blank, false);
      if not line_ins_del_avail or not window_full_width then
      begin
         draw_part_window(cursor_row, window_length, false);
         replace_cursor
      end
   end { insert 12 blank lines }
   else with saved_file_place do
   begin { delete any blank lines }
      strip_trailing_spaces;
      if (line_length = 0) and (line_status = normal) then
      begin
         save_place(saved_file_place);
         lines_to_delete := 0;
         { count the blank lines }
         repeat
            lines_to_delete := lines_to_delete + 1;
            load_a_line(line_number+1, false);
            strip_trailing_spaces
         until
            (line_length <> 0) or       { non-blank line or }
            (line_status <> normal) or  { .bottom. or }
            macro_dropout or            { breakin or }
            (main_state <> running) or  { some file system error or }
            ((not window_mode) and (line_number = iw_marked_line - 1));  { IW at marked line }
         restore_place(saved_file_place);
         if (lines_to_delete <= max_ins_del) and
            line_ins_del_avail and window_full_width and
            lineerase_avail and window_at_right
         then
         begin
            for lindex := 1 to lines_to_delete
            do deleteonel_comm(true)
         end
         else
         begin
            quick_delete(saved_line, saved_line+lines_to_delete-1);
            row_index := cursor_row;
            move_down_screen(row_index);  { set up screen status correctly }
            draw_part_window(cursor_row, window_length, false);
            replace_cursor
         end
      end
      else if current_macro_desc = nil
      then display_error('Line is not blank.')
   end { delete any blank lines }
end { openit_comm } ;

{ Put the character marker at the cursor position, or remove the character }
{ marker }

procedure pointch_comm(meta_switch: boolean);                                   { POINTCH_COMM }

var
   seton: boolean;
   saved_file_place: file_places;
   old_point_line_pos: line_chars0_range;
   row_first: row_range;

begin { pointch_comm }
   if check_normal then
   with current_buffer, current_window, saved_file_place do
   begin

      { first remove the current pointer if necessary }

      save_place(saved_file_place);
      if (point_line_pos <> 0) then
      begin { there was already a pointer }
         old_point_line_pos := point_line_pos;
         point_line_pos := 0;
         if find_first(point_line_no, row_first) then
         begin { old pointer still on screen }
            move_to_line(point_line_no);
            position_cursor_in_window(false, old_point_line_pos, false);
            if old_point_line_pos > line_length
            then put_char_on_screen(' ')
            else put_char_on_screen(line_text[old_point_line_pos]);
            restore_place(saved_file_place);
            position_cursor_in_window(false, saved_position, true)
         end { old pointer still on screen } ;
         point_line_no := 0
      end { there was already a pointer } ;

      { now switch on at cursor position }

      if meta_switch then
      begin
         seton := window_status[cursor_row].offset = 0;
         set_half(seton);               { set 'half intensity' status to
                                          opposite of usual }
         print_a_char('^');
         vdu_control(cursorleft);
         set_half(false);               { back to full intensity }
         point_line_pos := saved_position;
         point_line_no := line_number
      end
   end
end { pointch_comm } ;

{ Redraw current file line }

procedure redrawl_comm;                                                         { REDRAWL_COMM }

begin { redrawl_comm }
   clear_message;
   form_draw_file_line(true, true)
end { redrawl_comm } ;

{ Redraw the whole screen }

procedure redraws_comm;                                                         { REDRAWS_COMM }

begin { redraws_comm }
   form_draw_file_line(need_to_redraw, true);  { reformat current line if necessary }
   screen_tabs_set := false;            { set tabs when user requests redraw }
   draw_whole_screen
end { redraws_comm } ;

{ Repeat the following command a number of times }

procedure repeatit_comm;                                                        { REPEATIT_COMM }

var
   ch: char;
   answer_text: tree_type;
   cr_given, prompt, obey: boolean;
   desc: ptr_macro_desc;
   action: action_proc;
   loop_info: looprec;

begin { repeatit_comm }
   clear_message;
   if mode_quiet
   then get_treename('Give count or terminator: ',
                     answer_text, cr_given, false, true, true, false)
   else
   begin
      if current_macro_desc = nil
      then screen_message('Give count or terminator', true);
      get_treename(
         '(M for marker, E ch for equal character, N ch for not equal): ',
         answer_text, cr_given, false, true, true, false)
   end;
   obey := false;
   if cr_given and (answer_text.len > 0) then
   with current_buffer, loop_info, answer_text do
   begin { analyse the answer }
      ch := chars[1];
      if ch in ['M','m'] then           { m => use mark }
      begin
         if marked_line = 0
         then display_error('No line is marked.')
         else
         begin
            if marked_line < line_number
            then  end_condition := stop_before_mark
            else  end_condition := stop_after_mark;
            loop_counter := 1;
            obey := true;
            clear_message
         end { a line is marked }
      end { repeat to mark }
      else if ch in ['E','e','N','n'] then
      begin { test for "equals character" }
         if ch in ['E','e']
         then end_condition := stop_eq_char
         else end_condition := stop_ne_char;
         prompt := false;
         if len = 3 then
         begin { 3 chars  typed }
            if chars[2] = ' '
            then loop_test_char := chars[3]  { test char }
            else prompt := true
         end { 3 chars  typed }
         else if len = 2 then
         begin { 2 chars typed }
            if chars[2] = ' '
            then prompt := true
            else loop_test_char := chars[2]
         end { 2 chars typed }
         else prompt := true;
         clear_message;
         if prompt then
         begin { prompt for test character }
            message := 'Give character to test for ';
            if end_condition = stop_ne_char
            then append_message('not ');
            append_message( 'equal: ');
            get_treename(message, answer_text, cr_given, false, true, true,
                         false);
            if cr_given then
            begin { char given }
               if length(answer_text.strng) > 0
               then loop_test_char := answer_text.strng[1]
               else loop_test_char := ' ';
               obey := true;
               clear_message
            end { char given }
            else display_message('Function abandoned.')
         end { prompt for test character }
         else obey := true;
         if always_output and (current_macro_desc = nil)
         then ctrl_half_duplex(true)    { echoing off }
      end { E or N }
      else if ch = '*' then
      begin { repeat forever }
         loop_counter := large_int;
         end_condition := ntimes;
         clear_message;
         obey := true
      end { repeat forever }
      else
      begin { convert repetition counter }
         if tree_to_dec(answer_text, loop_counter) then
         begin
            end_condition := ntimes;
            clear_message;
            obey := loop_counter > 0
         end
         else display_error('Count error.')
      end { number }
   end { analyse the answer }
   else display_message('Function abandoned.');
   if obey then
   begin { perform at least once }
      if (current_macro_desc = nil) and
         not mode_quiet
      then with loop_info do
      begin
         message := 'Next command will be obeyed ';
         case end_condition of
            once_only:
               append_message('once only.');
            ntimes:
               begin
                  if loop_counter >= large_int
                  then append_message('forever.')
                  else
                  begin
                     append_message('~ times.');
                     ins_int(loop_counter, 1)
                  end
               end;
            stop_after_mark, stop_before_mark:
               append_message('to the marked line.');
            stop_eq_char:
               begin
                  append_message('to "~" character.');
                  ins_ch(loop_test_char)
               end;
            stop_ne_char:
               begin
                  append_message('to not "~" character.');
                  ins_ch(loop_test_char)
               end
         end { case } ;
         display_message(message)
      end { display confirmation message } ;
      if term_is_ssmp and (current_macro_desc = nil) then
      begin
         box_in_cursor;
         read_input_command(true, desc, action);
         wait_for_token(false);
         special_limits_set := false;
         repair_screen(rest1, false);   { replace the ch possibly overwritten }
         need_to_output := true         { output all for ssmp }
      end
      else read_input_command(true, desc, action);
      if desc <> nil then
      begin { macro - copy loop info into desc }
         desc^.loop_spec := loop_info;
         obey_input_command(desc, action)
      end { macro - copy loop info into desc }
      else
      begin { command - obey here }
         repeat
            obey_input_command(desc, action);
            need_to_output := true;     { from now on they are repeats }
         until test_for_end(loop_info)
      end { command - obey here }
   end { perform at least once }
end { repeatit_comm } ;

{ Erase previous character (SSMP only) }

procedure seraprev_comm;                                                        { SERAPREV_COMM }

begin { seraprev_comm }
   with current_buffer, current_window do
   if check_normal and (line_position > 1) and term_is_ssmp then
   begin
      if cursor_col = 1
      then wait_for_token(false);
      cursor_col := cursor_col - 1;
      cursor_moved(false, cursorleft);
      if line_position <= line_length then
      begin { cursor is within line }
         if not line_altered
         then push_line;                { save for OOPS command }
         line_text[line_position] := ' ';
         line_altered := true;
         if token = withhost            { i.e. at screen boundary }
         then form_draw_file_line(true, true)
      end { cursor is within line }
   end
end { seraprev_comm } ;

{ Start of macro definition }

procedure smacro_comm(meta_switch: boolean);                                    { SMACRO_COMM }

var
   done, cr_given: boolean;
   ch: char;
   answer: tree_type;

begin { smacro_comm }
   if always_output
   then ctrl_half_duplex(false);        { echoing on }
   clear_message;
   message := 'Give macro identifying character or ^ch for opposite macro: ';
   done := false;
   repeat
      get_treename(message, answer, cr_given, false, true, true, false);
      if answer.len <= 0
      then cr_given := false;
      if cr_given then
      begin { char given }
         macro_id_char := answer.chars[1];
         if (macro_id_char = '^') and (answer.len > 1) then
         begin { may be neg macro }
            macro_id_char := answer.chars[2];
            macro_id_type := negm
         end { may be neg macro }
         else
         begin { single char typed }
            macro_id_type := posm
         end { single char typed } ;
         if (macro_id_char in ascii_printable) and
            not (macro_id_char in ['(',')'])
         then done := true
         else message := 'Not allowed - please type another: '
      end { char given }
      else done := true
   until done;
   clear_message;
   if cr_given then
   begin { macro char given }
      { delete old macro if necessary }
      with macro_pointers[macro_id_char] do
      if (not meta_switch) and          { if deleteing macro and }
         (pointers[macro_id_type] = nil)  { it does not exist }
      then display_error('No macro defined.')
      else
      begin
         free_macro(macro_pointers[macro_id_char].pointers[macro_id_type]);
         free_mbufs(macro_head)         { clear out old if any }
      end;
      if meta_switch then
      begin
         new_macro_buffer(macro_head);
         store_macro_count := 0
      end
   end { macro char given }
   else display_message('Function abandoned.');
   if always_output
   then ctrl_half_duplex(true)          { echoing off }
end { smacro_comm } ;

{ Insert or delete spaces on the current line }

procedure spaces_comm(meta_switch: boolean);                                    { SPACES_COMM }

const
   spaces_to_insert = 20;

var
   posn, lindex: line_chars_range;
   new_spaces: 0..spaces_to_insert;

begin { spaces_comm }
   if check_normal then                 { not .bottom. }
   with current_buffer, current_window, vdu_table^ do
   begin
      strip_trailing_spaces;
      if (line_position <= line_length) then  { only act if inside line_length }
      if meta_switch then
      begin { insert spaces }
         new_spaces := spaces_to_insert;
         if (line_length + new_spaces) > line_max_charsm1
         then display_error('File line is full.')
         else
         begin { line not full }
            if not line_altered
            then push_line;             { save for OOPS command }
            movewds(line_text, line_position,
                    line_text, line_position + new_spaces,
                    line_length + 2 - line_position);
            for lindex := line_position to (line_position + new_spaces - 1)
            do line_text[lindex] := ' ';
            line_length := line_length + new_spaces;
            line_altered := true;
            if need_to_redraw or        { already need to redraw }
               (line_length > window_width) or  { long line }
               (not ch_ins_del_avail) or  { no charinsert }
               (control_array[charinsert].string_length > 2)  { long ins seq }
            then form_draw_file_line(true, true)
            else
            begin { do by charinsert }
               if data_protect or screen_is_split then
               begin
                  window_position(cursor_row, window_width-new_spaces+1);
                  for lindex := 1 to new_spaces
                  do vdu_control(chardelete);
                  replace_cursor
               end;
               for lindex := 1 to new_spaces
               do vdu_control(charinsert)
            end { do by charinsert }
         end { line not full }
      end { insert spaces }
      else
      begin { delete spaces up to non-space }
         posn := line_position;
         if line_text[posn] = ' ' then
         begin
            repeat posn := posn + 1     { find next non-space char }
            until line_text[posn] <> ' ';
            delete_some_chs(posn)       { delete the spaces }
         end
         else if current_macro_desc = nil
         then display_error('Character is not blank.')
      end { delete spaces }
   end { check_normal, with }
end { spaces_comm } ;

{ Store one character }

procedure storeonech_comm;                                                      { STOREONECH_COMM }

var
   make_space: boolean;

begin { storeonech_comm }
   if check_normal then
   with current_buffer, current_window, vdu_table^ do
   begin
      if not line_altered
      then push_line;                   { save for OOPS command }
      { disallow storing two special chars that Primos uses }
      if (input_char = chr(dc1_code)) or (input_char = newline)
      then input_char := '?';
      if mode_insert
      then if term_is_ssmp
           then make_space := current_macro_desc <> nil
           else make_space := true
      else make_space := false;
      if make_space then
      begin { make space for character }
         insertonech_comm(true)         { make space in the line }
      end { make space for character } ;
      if need_to_output then
      begin { echo it }
         { suppress output at bottom RH corner of screen }
         if screen_is_split and
            ((cursor_col + window_left - 1) = cols_per_row ) and
            ((cursor_row + window_top  - 1) = rows_per_page)
         then need_to_redraw := true
         else print_a_char(input_char)
      end { echo it } ;
      expand_line;
      line_text[line_position] := input_char;  { store the ch }
      line_altered := true;
      need_to_redraw := need_to_redraw or  { redraw later if overlaying }
         ((input_char <> ' ') and       { non-blank character onto }
         (window_status[cursor_row].offset > 0) and  { folded part because it }
         fold_avail);                   { appears in wrong intensity }
      cursor_col := cursor_col + 1;
      cursor_moved(false, backtab);
      if (line_position >= line_max_charsm1) and (token = withhost)
      then position_cursor_in_window(false, line_position, false);
      if mode_margin
      then split_at_margin
   end { with }
end { storeonech_comm } ;

{ Tab left }

procedure tableft_comm;                                                         { TABLEFT_COMM }

var
   state: search_type;
   new_col: line_chars_range;
   tabno: tabs_range;

begin { tableft_comm }
   with current_buffer, current_window do
   begin { perform a back tab }
      if need_to_output
      then vdu_control(backtab);
      if line_position = 1
      then macro_dropout := true;       { tabbing left off screen }
      if number_of_tabs < 1
      then state := not_found           { no tabs - go to col 1 }
      else
      begin
         tabno := number_of_tabs;
         state := searching;
         repeat
            if htabs[tabno] < line_position
            then state := found
            else if tabno <= 1
            then state := not_found
            else tabno := tabno - 1
         until state <> searching
      end;
      if state = found
      then new_col := htabs[tabno]
      else new_col := 1;
      if (state = found) and backtab_avail and
         (line_position <= window_width) and (not screen_is_split)
      then
      begin { done by hardware }
         line_position := new_col;
         cursor_col := new_col
      end { done by hardware }
      else position_cursor_in_window(false, new_col, false)  { must simulate }
   end { perform a back tab }
end { tableft_comm } ;

{ Tab right }

procedure tabright_comm;                                                        { TABRIGHT_COMM }

var
   state: search_type;
   new_col: line_chars_range;
   tabno: tabs_range;

begin { tabright_comm }
   with current_buffer, current_window do
   begin
      if need_to_output
      then vdu_control(horiztab);
      tabno := 1;                       { find first screen tab after cursor }
      state := searching;
      repeat
         if tabno > number_of_tabs
         then state := not_found
         else if htabs[tabno] > line_position
         then state := found
         else if tabno >= tabs_max
         then state := not_found
         else tabno := tabno + 1
      until state <> searching;
      if state = found
      then new_col := htabs[tabno]
      else
      begin
         new_col := line_position;
         macro_dropout := true          { off end of tabs }
      end;
      if (state = found) and horiztab_avail and
         (new_col <= window_width) and (not screen_is_split)
      then
      begin { done by hardware }
         line_position := new_col;
         cursor_col := new_col
      end { done by hardware }
      else position_cursor_in_window(false, new_col, false)  { must simulate }
   end
end { tabright_comm } ;

{ Set or clear a tabstop at the current cursor column }

procedure tabstop_comm(meta_switch: boolean);                                   { TABSTOP_COMM }

var
   tabno, tab_index: tabs_range;
   tabpos: integer;

   procedure findtab;                   { find first tab >= col }

   var
      state: search_type;

   begin { findtab }
      tabno := 1;
      state := searching;
      repeat
         if tabno > number_of_tabs
         then state := not_found
         else if htabs[tabno] >= current_buffer.line_position
         then state := found
         else if tabno < tabs_max
         then tabno := tabno + 1
         else state := not_found
      until state <> searching;
      if state = found
      then tabpos := htabs[tabno]
      else tabpos := large_int
   end { findtab } ;

begin { tabstop_comm }
   with current_buffer, current_window, vdu_table^ do
   begin
      screen_tabs_set := false;         { if anything goes wrong, at least }
                                        { correct tabs will be set next time }
      findtab;
      if meta_switch then
      begin { switch on a tab }
         if tabpos = line_position
         then                           { already a tab }
         else if number_of_tabs < tabs_max then
         begin
            for tab_index := number_of_tabs downto tabno
            do htabs[tab_index + 1] := htabs[tab_index];
            number_of_tabs := number_of_tabs + 1;
            htabs[tabno] := line_position;
            replace_cursor;             { in case user typed ahead }
            if line_position <= window_width
            then vdu_control(settab)
         end
         else display_error('Too many tabs.')
      end { switch on }
      else
      begin { switch off a tab }
         if tabpos = line_position then
         begin
            for tab_index := (tabno + 1) to number_of_tabs
            do htabs[tab_index - 1] := htabs[tab_index];
            number_of_tabs := number_of_tabs - 1;
            replace_cursor;             { in case user typed ahead }
            if line_position <= window_width then
            begin { clear the tab }
               if control_array[clearatab].string_length > 0
               then vdu_control(clearatab)
               else
               begin
                  set_screen_tabs;
                  replace_cursor
               end
            end { clear the tab }
         end
         else display_error('Not a tab.')
      end { switch off }
   end { with }
end { tabstop_comm } ;

{ Unload and optionally delete the lines between the cursor and the marked }
{ line inclusive }

procedure unlfile_comm(delete: boolean);                                        { UNLFILE_COMM }

var
   start_line, stop_line, lineno, new_line_number, overlap: integer;
   row_first, row_index: row_range;
   screen_ok, redraw: boolean;

begin { unlfile_comm }
   if read_load_treename then
   with current_buffer, current_window, load_file do
   begin
      if line_status = bottom then
      begin { move up a line }
         do_cursor_move(cursorup);
         cursor_row := cursor_row - 1;
         cursor_moved(true, cursorup)
      end { move up a line } ;
      if always_output
      then ctrl_half_duplex(false);     { echoing on }
      original_pathname := false;
      rename_done := false;
      open_file(k$writ, load_file, 0, true, true);
      if always_output
      then ctrl_half_duplex(true);      { echoing off }
      if not file_is_open then
      begin
         replace_cursor;
         macro_dropout := true
      end
      else
      begin { file opened ok }
         if (current_macro_desc = nil) and (not suppress_output)
         then display_message('Unloading... ');
         if marked_line = 0             { default is one line }
         then stop_line := line_number
         else stop_line := marked_line;
         unmark;                        { do not need mark any more }
         if stop_line < line_number then
         begin { swap so that line_number <= stop_line}
            lineno := line_number;
            move_to_line(stop_line);
            stop_line := lineno
         end { swap so that line_number <= stop_line} ;
         start_line := line_number;
         screen_ok := true;             { assume still ok so far }
         if quick_unload(start_line, stop_line) then
         begin
            new_line_number := line_number + 1;
            clear_message;
            move_to_line(new_line_number);
            if delete then
            begin
               { if only a few lines and first line on screen ... }
               if ((stop_line - start_line) < max_ins_del) and
                  (start_line >= window_status[1].file_line) and
                  find_first(start_line,row_first) and
                  line_ins_del_avail and window_full_width
               then
               begin { delete lines individually on screen }
                  cursor_row := row_first;
                  cursor_col := 1;
                  replace_cursor;
                  for lineno := start_line to stop_line
                  do deleteonel_comm(true)
               end { delete in situ }
               else
               begin { just redraw screen }
                  quick_delete(start_line, stop_line);
                  screen_ok := false    { must redraw screen }
               end { just redraw screen }
            end { delete after unload }
         end { quick_unload worked }
         else { quick_unload failed } ;
         close_file(load_file);

         { we are now at the line after the area unloaded }

         if not window_mode and (line_number >= iw_marked_line) then
         begin
            move_to_line(line_number - 1);
            insert_blank_line;          { reinstate blank line }
            screen_ok := false          { force redraw }
         end;
         overlap := window_length div 4;
         if find_first(new_line_number, row_first)
         then redraw := false           { new place still on screen }
         else if new_line_number >
                 window_status[window_length].file_line + overlap
         then redraw := true            { too far off screen }
         else
         begin { only a little way off screen }
            redraw := false;
            row_first := window_length + 1
         end { only a little way off screen } ;
         if not screen_ok or redraw
         then form_draw_window(false)
         else
         begin { screen still ok }
            if delete
            then clear_message          { cursor placed ok already }
            else
            begin { move cursor to correct place }
               replace_cursor;
               move_to_line(new_line_number);
               position_cursor_in_window(false, 1, false)
            end { move cursor }
         end { screen still ok } ;
         { if buffer used it may be visible on split screen so redraw }
         if load_use = use_buffer
         then redraw_buffer_if_visible(load_buffer_number, true)
      end { file opened ok }
   end { treename alright, with }
end { unlfile_comm } ;

{ Convert cursor position to end of line to either upper or lower case }

procedure up_and_lowcase_proc(up: boolean);                                     { UP_AND_LOWCASE_PROC }

var
   lindex: line_chars_range;
   temp, lpos, saved_point: line_chars0_range;
   from_set: set_of_char;
   diff: integer;
   ch: char;

begin { up_and_lowcase_proc }
   if check_normal then
   with current_buffer do
   begin
      if up then
      begin
         from_set := lower_alphabet;
         diff := ord('A') - ord('a')
      end
      else
      begin
         from_set := upper_alphabet;
         diff := ord('a') - ord('A')
      end;
      if (point_line_pos = 0) or (point_line_no <> line_number)
      then                              { alter to end of line }
         if line_position <= line_length
         then saved_point := line_length
         else saved_point := line_position
      else
      begin { alter to pointer }
         saved_point := point_line_pos;
         point_line_no := 0;
         point_line_pos := 0            { clear the pointer }
      end { alter to pointer } ;
      lpos := line_position;
      if lpos > saved_point then
      begin
         temp := lpos;
         lpos := saved_point;
         saved_point := temp
      end;
      push_line;                        { save for OOPS command }
      for lindex := lpos to saved_point do
      begin
         ch := line_text[lindex];
         if ch in from_set
         then line_text[lindex] := chr(ord(ch) + diff)
      end { for loop } ;
      line_altered := true;
      form_draw_file_line(true, true)
   end { normal line, with }
end { up_and_lowcase_proc } ;

{ Switch between WI and IW modes }

procedure wiswitch_comm;                                                        { WISWITCH_COMM }

var
   row_index, row_last: row_range;
   mline: integer;
   saved_line_position: line_chars_range;

begin { wiswitch_comm }
   with current_buffer, current_window do
   begin
      window_mode := not window_mode;   { invert mode }
      if window_mode then
      begin { now in WI mode }
         mline := iw_marked_line;
         iw_marked_line := 0;
         if (mline - 1) < (window_status[window_length].file_line) then
         begin { there are some null lines on screen }
            move_to_line(mline - 1);    { last valid line }
            find_last(row_last);
            row_index := row_last;
            move_down_screen(row_index);
            draw_part_window(row_last+1, window_length, true)
         end { null lines on screen }
      end { now in WI mode }
      else
      begin { now in IW mode }
         cursor_to_line_start;
         iw_marked_line := line_number;
         insertonel_comm(do_blank, false);  { insert the first blank line }
         saved_line_position := line_position;
         row_index := cursor_row;
         move_down_screen(row_index);
         if pageerase_avail and window_at_bottom then
         begin { use pageerase }
            if cursor_row < window_length then
            begin
               window_position(cursor_row, 1);
               vdu_control(pageerase)   { blank out rest of page }
            end
         end { use pageerase }
         else if lineerase_avail and window_at_right then
         begin { use lineerase }
            for row_index := cursor_row to window_length do
            begin { blank out null lines }
               window_position(row_index, 1);
               vdu_control(lineerase)
            end { blank out null lines }
         end { use lineerase }
         else draw_part_window(cursor_row, window_length, false);
         position_cursor_in_window(false, saved_line_position, false)
      end { now IW } ;
      if not mode_quiet then
      begin
         if window_mode
         then display_message('Now in WI mode')
         else display_message('Now in IW mode')
      end { not mode_quite } ;
      replace_cursor
   end { with }
end { wiswitch_comm } ;

{ Re-execute the last editor command stored in the command buffer }

procedure xcomm_comm(once: boolean);                                            { XCOMM_COMM }

var
   command_worked: boolean;

begin { xcomm_comm }
   if ed_com_line_valid = nothing_defined
   then display_error('No editor command in buffer.')
   else obey_ed_command(ed_com_line, command_worked)
end { xcomm_comm } ;

{ Delete character at cursor location, or characters from cursor to
  "point_line_pos" inclusive }

procedure zap_comm;                                                             { ZAP_COMM }

var
   temp, chars_gone, saved_point: line_chars0_range;

begin { zap_comm }
   if check_normal then
   with current_buffer do
   begin
      if (point_line_pos = 0) or (point_line_no <> line_number)
      then saved_point := line_position  { no pointer - delete one ch }
      else
      begin
         saved_point := point_line_pos;
         point_line_no := 0;
         point_line_pos := 0            { clear the pointer }
      end;
      if line_position > saved_point then
      begin { ensure saved_point >= line_position }
         temp := line_position;
         line_position := saved_point;
         saved_point := temp;
         position_cursor_in_window(false, line_position, false)
      end { ensure saved_point >= line_position } ;
      expand_line;
      if saved_point > line_length      { move saved_point within line_length }
      then saved_point := line_length;
      delete_some_chs(saved_point + 1)  { delete the selected characters }
   end
end { zap_comm } ;

{ Perform text formatting as for TEXT command, but keeping screen up-to-date
  as we go }

procedure formtext_comm;                                                        { FORMTEXT_COMM }

var
   lines_to_delete, relative_pos, old_mark, old_left_margin, index: integer;
   last_was_marked: boolean;
   from_line, to_line: line_chars_type;

   procedure move_relative(relline: integer);

   var
      target_line: integer;

   begin { move_relative }
      with current_buffer, current_window do
      begin
         target_line := line_number - relative_pos + relline;
         while window_status[cursor_row].file_line > target_line do
         begin
            do_cursor_move(cursorup);
            cursor_row := cursor_row - 1;
            cursor_moved(true, cursorup)
         end;
         while window_status[cursor_row].file_line < target_line do
         begin
            do_cursor_move(cursordown);
            cursor_row := cursor_row + 1;
            cursor_moved(true, cursordown)
         end;
         relative_pos := relline
      end
   end { move_relative } ;

   function get_from_line(var from_line: line_chars_type;
                          var from_length: line_chars0_range): boolean;

   begin { get_from_line }
      if last_was_marked or (main_state <> running)
      then get_from_line := false
      else with current_buffer do
      begin
         move_relative(lines_to_delete);
         if line_status = normal then
         begin
            strip_trailing_spaces;
            from_line := line_text;
            from_length := line_length;
            get_from_line := true;
            last_was_marked := line_number = marked_line;
            lines_to_delete := lines_to_delete + 1;
            if lines_to_delete > 2 then
            begin { do not get too far ahead }
               move_relative(0);
               deleteonel_comm(true);
               position_cursor_in_window(false, 1, false);
               lines_to_delete := lines_to_delete - 1
            end { do not get too far ahead }
         end { not EOF }
         else get_from_line := false
      end { some more lines available }
   end { get_from_line } ;

   procedure write_to_line(to_line: line_chars_type;
                           to_length: line_chars0_range);

   begin { write_to_line }
      with current_buffer do
      begin
         move_relative(0);
         if lines_to_delete > 0
         then lines_to_delete := lines_to_delete - 1  { reuse existing line }
         else insertonel_comm(do_blank, true);  { insert new line }
         position_cursor_in_window(false, 1, false);
         line_text := to_line;
         line_length := to_length;
         line_altered := true;
         form_draw_file_line(true, true);
         relative_pos := -1
      end { with }
   end { write_to_line } ;

begin { formtext_comm }
   with current_buffer, current_window do
   begin
      old_left_margin := point_line_pos;
      point_line_no := 0;
      point_line_pos := 0;              { clear the pointer after use }
      if marked_line = 0
      then marked_line := line_number
      else if marked_line < line_number then
      begin
         old_mark := marked_line;
         unmark;
         marked_line := line_number;
         cursor_to_line_start;
         mark_the_line(cursor_row);
         move_to_line(old_mark);
         position_cursor_in_window(false, 1, false)
      end;
      last_was_marked := false;
      lines_to_delete := 0;             { lines whose text has been processed }
                                        { but not yet deleted }
      relative_pos := 0;                { position relative to first of above }
      text_format(old_left_margin, get_from_line, write_to_line);
      if main_state = running
      then unmark;
      move_relative(0);
      for index := 1 to lines_to_delete  { get rid of last few old lines }
      do deleteonel_comm(true);
      { leave cursor at column 1 }
      position_cursor_in_window(false, 1, true)
   end { with }
end { formtext_comm } ;

{ Switch the cursor to the next window cyclically }

procedure nextwind_comm(direction: direction_type);                             { NEXTWIND_COMM }

var
   new_window_number: integer;

begin { nextwind_comm }
   if screen_is_split then
   begin { screen is split }
      clear_message;                    { clear any previously announced name }
      if direction = go_forwards
      then if current_window_number = window_count
           then new_window_number := 1
           else new_window_number := current_window_number + 1  { forwards }
      else if current_window_number = 1
           then new_window_number := window_count
           else new_window_number := current_window_number - 1;  { backwards }
      switch_window(new_window_number);
      if mode_disp
      then announce_buffer_name;
      replace_cursor
   end { screen is split }
   else display_error('The screen is not split.')
end { nextwind_comm } ;

{ Test the pattern of upper and lower case in a string of characters }
{ ignoring any non-alphabetic characters                             }
{ Answer is:                                                         }
{   mixed_case:   AbcDefG                                            }
{   all_upper:    ABCDEFG                                            }
{   all_lower:    abcdefg                                            }
{   first_upper:  Abcdefg or just A                                  }

function test_case_pattern(function next_char: char): case_patterns;            { TEST_CASE_PATTERN }

var
   first_char: char;
   rest_state, result: case_patterns;

   function test_rest: case_patterns;

   var
      done: boolean;
      second_char, ch: char;
      test_set: set_of_char;

   begin { test_rest }
      repeat
         done := true;
         second_char := next_char;
         if second_char = newline
         then test_rest := all_lower    { so that "A" => first_upper }
         else if second_char in lower_alphabet then
         begin
            test_set := upper_alphabet + [newline];
            repeat ch := next_char
            until ch in test_set;
            if ch = newline
            then test_rest := all_lower
            else test_rest := mixed_case
         end
         else if second_char in upper_alphabet then
         begin
            test_set := lower_alphabet + [newline];
            repeat ch := next_char
            until ch in test_set;
            if ch = newline
            then test_rest := all_upper
            else test_rest := mixed_case
         end
         else done := false             { ignore non-alpha }
      until done
   end { test_rest } ;

begin { test_case_pattern }
   first_char := next_char;
   if first_char in lower_alphabet then
   begin
      rest_state := test_rest;
      if rest_state in [all_upper, first_upper]
      then result := mixed_case
      else result := rest_state
   end
   else if first_char in upper_alphabet then
   begin
      rest_state := test_rest;
      if rest_state = all_lower
      then result := first_upper
      else result := rest_state
   end
   else result := test_rest;
   test_case_pattern := result
end { test_case_pattern } ;

{ Obey one or more OOPS commands given via DEL C }

procedure wi_oops_proc;                                                         { WI_OOPS_PROC }

var
   oops_index: wi_oops_range;

begin { wi_oops_proc }
   if wi_oops_count > wi_oops_top
   then wi_oops_count := wi_oops_top;
   if wi_oops_count = 0
   then display_error('No saved lines available.')
   else
   begin
      for oops_index := 1 to wi_oops_count
      do insertonel_comm(do_oops, true)
   end;
   { ensure that the current line is updated to file; otherwise any changes }
   { are not pushed }
   replace_line
end { wi_oops_proc } ;

{ Get a single character from the user as answerto a question. }
{ Echoing must be off if not SSMP. }

procedure get_answer_char(var answer_char: char);                               { GET_ANSWER_CHAR }

var
   chcode: integer;

begin { get_answer_char }
   if term_is_ssmp then
   begin { SSMP terminal }
      box_in_cursor;
      chcode := get_ssmp_code(1);
      answer_char := chr(chcode);
      wait_for_token(false);
      special_limits_set := false;
      repair_screen(rest1, false)
   end { SSMP terminal }
   else
   begin { not SSMP terminal }
      if eoln(input) then readln;       { ignore CR }
      read(answer_char)
   end { not SSMP terminal } ;
   if answer_char in lower_alphabet
   then answer_char := chr(ord('A') - ord('a') + ord(answer_char))
end { get_answer_char } ;

{ Obey the REPLACE command given via DEL C }

procedure wi_replace_proc;                                                      { WI_REPLACE_PROC }

label 321;

var
   locate_command, first_command: string;
   do_prompt, terminate, options_ok, answer_ok, file_line_is_full,
   user_typing, cr_given, command_worked: boolean;
   replace_case, replace_ident, replace_mark, adjust_case: boolean;
   options_tree: tree_type;
   saved_marked_line, rep_sub_2: integer;
   sindex: tree_range;
   chindex: integer;
   ch: char;
   code: shortint;

   function analyse_options: boolean;

   var
      opindex: 1..8;
      ch: char;

   begin { analyse_options }
      analyse_options := true;
      replace_case := false;
      replace_ident := false;
      for opindex := 1 to 8 do
      begin
         ch := replace_options[opindex];
         if ch = ' '
         then                           { ignore spaces }
         else if (ch = 'c') or (ch = 'C')
         then replace_case := true
         else if (ch = 'i') or (ch = 'I')
         then replace_ident := true
         else analyse_options := false
      end { for loop }
   end { analyse_options } ;

   procedure store(the_char: char);

   begin { store }
      insert(the_char, locate_command, length(locate_command)+1)
   end { store } ;

   procedure perform_replace;

   var
      old_length, new_line_length, chindex, rep_sub: integer;
      ch: char;
      do_in_situ: boolean;
      rep_string: tree_type;

      function next_char: char;
      begin { next_char }
         with current_buffer do
         begin
            if rep_sub > last_match_position
            then next_char := newline
            else
            begin { in range }
               next_char := line_text[rep_sub];
               rep_sub := rep_sub + 1
            end { in range }
         end { with }
      end { next_char } ;

      procedure upcase(start, finish: integer);

      var
         index: integer;
         ch: char;

      begin { upcase }
         with rep_string do
         begin
            for index := start to finish do
            begin
               ch := chars[index];
               if ch in lower_alphabet
               then chars[index] := chr(ord('A') - ord('a') + ord(ch))
            end { for loop }
         end { with }
      end { upcase } ;

      procedure lowcase(start, finish: integer);

      var
         index: integer;
         ch: char;

      begin { lowcase }
         with rep_string do
         begin
            for index := start to finish do
            begin
               ch := chars[index];
               if ch in upper_alphabet
               then chars[index] := chr(ord('a') - ord('A') + ord(ch))
            end { for loop }
         end { with }
      end { lowcase } ;

   begin { perform_replace }
      rep_string := replace_ptr_2^;
      with current_buffer, current_window, rep_string do
      begin
         old_length := last_match_position - line_position + 1;
         new_line_length := line_length - old_length + len;
         if new_line_length > line_max_charsm1 then
         begin { line overflow }
            terminate := true;
            file_line_is_full := true
         end { line overflow }
         else
         begin { line length OK }
            { ensure the line is long enough - the string matched could have }
            { had trailing spaces }
            line_position := line_position + len;
            expand_line;
            line_position := line_position - len;
            if adjust_case then
            begin { adjust case of replacement string }
               rep_sub := line_position;
               case test_case_pattern(next_char) of
                  mixed_case: ;
                  all_upper: upcase(1, len);
                  all_lower: lowcase(1, len);
                  first_upper:
                     begin
                        upcase(1, 1);
                        lowcase(2, len)
                     end
               end { case }
            end { adjust case of replacement string } ;
            { Perform the change in situ on the screen if...         }
            {    still prompting for each change                     }
            {    no char ins/del necessary or char ins/del available }
            {    not too many char ins/del                           }
            do_in_situ :=
               do_prompt and
               ((old_length = len) or
                (ch_ins_del_avail and window_at_right)) and
               (abs(len - old_length) <= 20);
            if do_in_situ then
            begin
               for chindex := old_length+1 to len
               do insertonech_comm(true);
               for chindex := len+1 to old_length
               do deleteonech_comm(true);
               for chindex := 1 to len do
               begin
                  ch := chars[chindex];
                  line_text[line_position] := ch;
                  line_altered := true;
                  set_half(window_status[cursor_row].offset > 0);
                  put_char_on_screen(ch);
                  cursor_col := cursor_col + 1;
                  cursor_moved(false, backtab)
               end;
               set_half(false)
            end { do in situ }
            else
            begin { not in situ }
               if new_line_length <= 0
               then line_length := 0
               else
               begin
                  movewds(line_text, line_position+old_length,
                          line_text, line_position+len,
                          line_length-line_position+1-old_length);
                  line_length := new_line_length;
                  for chindex := 1 to len
                  do line_text[line_position+chindex-1] := chars[chindex]
               end;
               line_altered := true;
               form_draw_file_line(true, true);
               line_position := line_position + len;
               position_cursor_in_window(false, line_position, false)
            end { not in situ }
         end { line length OK }
      end { with }
   end { perform_replace } ;

   procedure replace_quit_onunit(not_used: integer);

   begin { replace_quit_onunit }
      if user_typing then
      begin
         answer_ok := true;
         terminate := true;
         goto 321
      end
      else
      begin
         macro_dropout := true;
         main_state := quit_signal
      end
   end { replace_quit_onunit } ;

   function next_char_2: char;

   begin { next_char_2 }
      with replace_ptr_2^ do
      begin
         if rep_sub_2 > len
         then next_char_2 := newline
         else
         begin { in range }
            next_char_2 := chars[rep_sub_2];
            rep_sub_2 := rep_sub_2 + 1
         end { in range }
      end { with }
   end { next_char_2 } ;

begin { wi_replace_proc }
   with current_buffer, current_window do
   begin
      if replace_ptr_1 = nil then
      begin { need to prompt for strings }
         new(replace_ptr_1);
         new(replace_ptr_2);
         replace_ptr_1^.len := 0;
         with replace_ptr_1^ do
         begin
            while len <= 0 do
            begin
               get_treename('Give search string     : ', replace_ptr_1^,
                            cr_given, false, true, true, false);
               if cr_given
               then if len = 0
                    then display_message('String must be at least 1 character.')
                    else
               else len := 1            { just to exit }
            end
         end;
         if cr_given
         then get_treename('Give replacement string: ', replace_ptr_2^,
                           cr_given, false, true, true, false);
         if cr_given then
         begin { user pressed RETURN at end }
            repeat
               get_treename(
                  'Give options if any (Case independent, Identifier): ',
                  options_tree, cr_given, false, true, true, false);
               if cr_given then
               begin
                  replace_options := '        ';
                  for chindex := 1 to options_tree.len
                  do if chindex <= 8
                     then replace_options[chindex] := options_tree.chars[chindex];
                  options_ok := analyse_options;
                  if not options_ok
                  then display_message
                     ('Options must contain only the letters C and I.')
               end
               else options_ok := true  { just to exit }
            until options_ok;
            clear_message
         end { user pressed RETURN at end }
      end { need to prompt for strings }
      else
      begin { strings given }
         options_ok := analyse_options;
         cr_given := true
      end { strings given } ;

      if not cr_given
      then display_message('Function abandoned.')
      else if not options_ok
      then display_error('Error in format of REPLACE command.')
      else
      begin { go ahead and replace }
         if replace_case then
         begin { 'C' option given }
            rep_sub_2 := 1;
            case test_case_pattern(next_char_2) of
               all_upper, all_lower: adjust_case := true
               otherwise adjust_case := false
            end
         end { 'C' option given }
         else adjust_case := false;
         if marked_line = 0 then
         begin { no line is marked }
            replace_mark := false;
            first_command := 'top'      { allows for buffer chaining }
         end { no line is marked }
         else
         begin { line is marked }
            replace_mark := true;
            if marked_line < line_number then
            begin { swap places }
               saved_marked_line := marked_line;
               unmark;
               marked_line := line_number;
               str(saved_marked_line, first_command);
               insert('point ',first_command, 1)
            end { swap places }
            else first_command := '';
            line_position := 1          { <- col 1 }
         end { line is marked } ;

         { Construct the appropriate Locate-type command }

         with replace_ptr_1^ do
         begin
            locate_command := '';
            if replace_mark
            then store('M');
            if replace_ident
            then store('I');
            if replace_case
            then store('C');
            store('S');                 { start at current }
            store('L');
            store(' ');
            for sindex := 1 to len do
            begin
               ch := chars[sindex];
               if (ch in string_terminators) or
                  (ch = special[sescape])
               then store(special[sescape]);
               store(ch)
            end { for loop }
         end { with } ;

         { Loop obeying Locate command and doing replacement }

         need_to_output := true;        { used when doing ch ins/del }
         user_typing := false;
         on('QUIT$', replace_quit_onunit);
         ctrl_half_duplex(true);
         terminate := false;
         do_prompt := true;
         file_line_is_full := false;
         repeat
            if length(first_command) > 0
            then obey_ed_str(concat(first_command,';',locate_command),
                             command_worked)
            else obey_ed_str(locate_command, command_worked);
            first_command := '';
            if not command_worked then
            begin { not found }
               terminate := true;
               macro_dropout := false   { ok to hit eof }
            end { not found }
            else if replace_mark and (line_number > marked_line)
            then terminate := true;
            if terminate
            then { all done }
            else if do_prompt then
            begin { ask whether to replace }
               if term_is_ssmp or (tty$in >= 0)  { skip if typed ahead }
               then display_next_message('Replace?');
               user_typing := true;
               repeat
                  answer_ok := true;
                  get_answer_char(ch);
                  case ch of
                     'Y': perform_replace;
                     'N':
                        begin { move on one character position }
                           do_cursor_move(cursorright);
                           cursor_col := cursor_col + 1;
                           cursor_moved(false, cursorright)
                        end { move on one character position } ;
                     'Q': terminate := true;
                     'C':
                        begin { continue without display }
                           perform_replace;
                           do_prompt := false;
                           { use screen_message to avoid replace_cursor }
                           screen_message('Performing rest of replaces.', true);
                           { suppress any further screen redraws }
                           suppress_output := true
                        end { continue without display }
                     otherwise
                        begin
                           if main_state = running then
                           begin
                              display_message('Please press Y, N, Q or C.');
                              answer_ok := false
                           end
                           else terminate := true
                        end
                  end { case }
               until answer_ok;
            321:
               user_typing := false;
               if do_prompt             { do not clear message after 'C' }
               then clear_message
            end { prompt for replace }
            else perform_replace        { replace without prompting }
         until terminate;
         suppress_output := false;      { may have been suppressed by 'C' }
         if not always_output
         then ctrl_half_duplex(false);  { echoing back to usual state }
         if do_prompt and not term_is_ssmp then
         begin { clear any type ahead }
            tty$rs([bit2],code);        { clear terminal input }
            if code <> 0
            then error_print(k$irtn, code, 'Clearing input', 'TTY$RS')
         end { clear any type ahead } ;
         if do_prompt and (main_state <> quit_signal) and
            (not macro_dropout) then
         begin { normal end }
            unmark;
            display_next_message('End of REPLACE.')
         end { normal end }
         else
         begin { no display or interrupted }
            marked_line := 0;
            suppress_output := true;
            position_cursor_in_window(true, line_position, true);
            suppress_output := false;
            draw_whole_screen;
            if file_line_is_full
            then display_next_error('File line is full.');
            if main_state = quit_signal then
            begin
               main_state := running;
               display_next_message('REPLACE terminated.')
            end
            else display_next_message('End of REPLACE.')
         end { no display or interrupted }
      end { go ahead and replace } ;
      dispose(replace_ptr_1);
      replace_ptr_1 := nil;
      dispose(replace_ptr_2);
      replace_ptr_2 := nil
   end { with }
end { wi_replace_proc } ;

{ Obey a SPELL command given via DEL C }

procedure wi_spell_proc;                                                        { WI_SPELL_PROC }

   procedure show_line;

   begin { show_line }
      with current_buffer
      do position_cursor_in_window(false, line_position, true)
   end { show_line } ;

   procedure ask_user(msg: string; var answer_char: char);

   begin { ask_user }
      ctrl_half_duplex(true);           { echoing off }
      display_next_message(msg);
      get_answer_char(answer_char);
      clear_message;
      ctrl_half_duplex(false)           { echoing on }
   end { ask_user } ;

   procedure show_changed_line;

   begin { show_changed_line }
      show_line;                        { ensure cursor in right place }
      clear_message;
      form_draw_file_line(true, true)
   end { show_changed_line } ;

begin { wi_spell_proc }
   ctrl_half_duplex(false);             { echoing on for answers }
   spell_proc(show_line, ask_user, show_changed_line);
   show_line;
   if always_output
   then ctrl_half_duplex(true)          { echoing back to usual state }
end { wi_spell_proc } ;

{ Obey the window code for ED -TIDY. }

procedure wi_tidy_proc(tidy_arg: tidy_type);                                    { WI_TIDY_PROC }

   { attach to a pathname that may be longer than the Primos maximum }

   procedure attach_to_long_path(the_pathname: long_string);

   var
      code: shortint;
      temp_pathname, first_part: long_string;
      last_gt, first_gt, pindex: integer;

   begin { attach_to_long_path }
      if length(the_pathname) <= 120 then
      begin
         at$(k$seth, the_pathname, code);
         if code <> 0
         then error_print(k$irtn, code, the_pathname, 'AT$')
      end
      else
      begin { long pathname }
         temp_pathname := the_pathname;
         last_gt := 0;
         for pindex := 1 to 120
         do if temp_pathname[pindex] = '>'
            then last_gt := pindex;
         if last_gt = 0
         then display_next_error('No ''>'' in first 128 chars')
         else
         begin { last '>' found }
            temp_pathname := substr(the_pathname, 1, last_gt-1);
            at$(k$seth, temp_pathname, code);
            if code <> 0
            then error_print(k$irtn, code, temp_pathname, 'AT$');
            temp_pathname := substr(the_pathname, last_gt+1, length(the_pathname) - last_gt);
            while length(temp_pathname) > 0 do
            begin
               first_gt := pos('>', temp_pathname);
               if first_gt = 0 then
               begin { no '>' left }
                  first_part := temp_pathname;
                  temp_pathname := ''
               end { no '>' left }
               else
               begin { split at first '>' }
                  first_part := substr(temp_pathname, 1, first_gt-1);
                  delete(temp_pathname, 1, first_gt)
               end { split at first '>' } ;
               at$rel(k$seth, first_part, code);
               if code <> 0
               then error_print(k$irtn, code, first_part, 'AT$REL')
            end { while loop }
         end { last '>' found }
      end { long pathname }
   end { attach_to_long_path } ;

   { Get a filename from the current line of the current buffer }

   procedure get_filename_from_line(var the_filename: string);

   var
      done: boolean;
      col: line_chars_range;

   begin { get_filename_from_line }
      the_filename := '';
      with current_buffer do
      begin
         if line_length < 3
         then { line too short }
         else if line_text[2] <> ' '
         then { format wrong }
         else
         begin { loop to extract filename }
            done := false;
            col := 3;
            repeat
               if (line_text[col] = ' ') or
                  (line_text[col] = newline) or
                  (length(the_filename) >= 80)
               then done := true
               else
               begin { add next char }
                  insert(line_text[col], the_filename, length(the_filename)+1);
                  col := col + 1
               end { add next char }
            until done
         end { loop to extract filename }
      end { with }
   end { get_filename_from_line } ;

   { Test whether entry exists and what type it is }

   procedure test_entry(the_pathname: long_string;
                        var file_exists: boolean;
                        var file_type: shortint;
                        var is_binary: boolean);

   var
      buffer: packed array[1..2048] of char;
      unit, rnw, code: shortint;
      chars_read, count, chindex, position: integer;
      ch: char;
      ignore_next: boolean;

   begin { test_entry }
      call_srsfx$(k$read+k$getu, the_pathname, unit, file_type, code);
      file_exists := code = 0;
      if (code = 0) and (file_type in [sam_type, dam_type, cam_type]) then
      begin { test whether binary }
         position := 0;
         prwf$$(k$read, unit, addressof(buffer), 1024, position, rnw, code);
         if code = e$eof
         then code := 0;
         if code = 0
         then chars_read := rnw*2
         else chars_read := 0;
         if chars_read > 100 then
         begin { test file }
            { count characters with parity off }
            count := 0;
            ignore_next := false;
            for chindex := 1 to chars_read do
            begin
               if ignore_next
               then ignore_next := false
               else
               begin { test this character }
                  ch := buffer[chindex];
                  if ord(ch) < 200b
                  then count := count + 1;
                  if (ch = chr(dc1_code)) or  { space suppression }
                     ( (ch = newline) and odd(chindex) )  { filler after LF }
                  then ignore_next := true
               end { test this character }
            end { for loop } ;
            { more than 5% parity off? }
            is_binary := (count / chars_read) > 0.05
         end { test file }
         else is_binary := false
      end { test whether binary }
      else error_print(k$irtn, code, the_pathname, 'SRSFX$');
      close_unit(unit, code)
   end { test_entry } ;

   { Construct a CPL file to delete the marked files.  ld_cpl_file is set }
   { to the filename, or '' if not to be run. }

   procedure write_cpl_file(the_treename: tree_type);

   var
      infile, cpl_file: text;
      count, space_pos: integer;
      line, filename: string;
      is_mfd, obey_it: boolean;

   begin { write_cpl_file }
      ld_cpl_filename := substr(ld_pathname, 1, original_length);
      ld_cpl_filename := concat(ld_cpl_filename,'>T$PMED_DELETE_FILES.CPL');
      reset(infile, the_treename.strng);
      rewrite(cpl_file, ld_cpl_filename);
      writeln(cpl_file,'&severity &error &ignore');
      if length(ld_pathname) <= 200
      then writeln(cpl_file,'type Cleaning up directory ',ld_pathname);
      { this is an MFD if pathname ends with '>MFD' and there are no other '>' }
      is_mfd := (pos('>MFD', ld_pathname) = length(ld_pathname) - 3) and
                (pos('>'   , ld_pathname) = length(ld_pathname) - 3);
      if is_mfd
      then writeln(cpl_file,'avail')
      else writeln(cpl_file,'list_quota -br');
      writeln(cpl_file,'type Deleting entries');
      clear_message;
      count := 0;
      while not eof(infile) do
      begin
         readln(infile, line);
         if length(line) < 3
         then { too short }
         else if not (line[1] in ['d','D'])
         then { not marked }
         else if line[2] <> ' '
         then { invalid format }
         else
         begin { extract filename }
            delete(line, 1, 2);
            space_pos := pos(' ', line);
            if space_pos = 0
            then { invalid format }
            else
            begin { space found }
               filename := substr(line, 1, space_pos-1);
               { include pathname only if in sub-directory }
               writeln(cpl_file,'delete ''',filename,'''');
               if count = 0 then
               begin { give first message }
                  message := 'About to delete following entries';
                  if length(ld_pathname) <= 1000
                  then message := concat(message,' in ',ld_pathname);
                  screen_message(message, false)
               end { give first message } ;
               screen_message(filename, false);
               count := count + 1
            end { space found }
         end { extract filename }
      end { while loop } ;
      if is_mfd
      then writeln(cpl_file,'avail')
      else writeln(cpl_file,'list_quota -br');
      close(cpl_file);
      if count = 0
      then obey_it := false             { no files marked for deletion }
      else obey_it := wi_question_and_answer('Alright to carry on? ',
                                             false, true);
      if not obey_it then
      begin { delete it }
         delete_primos_file(ld_cpl_filename);
         ld_cpl_filename := ''
      end { delete it }
   end { write_cpl_file } ;

   { Return to the current ld file buffer }

   procedure tidy_ld_proc;

   begin { tidy_ld_proc }
      obey_ed_int('buffer ~; next', ld_buffer)
   end { tidy_ld_proc } ;

   { Inspect a file or go down into a sub-directory }

   procedure tidy_inspect_proc;

   var
      filename: string;
      ld_filename, new_path: long_string;
      command_worked, new_path_exists, type_alright, is_binary: boolean;
      new_path_type, code: shortint;

   begin { tidy_inspect_proc }
      { if we are already in buffer 1 inspecting a file then move on to }
      { the next }
      if current_buffer_number = 1
      then obey_ed_int('buffer ~; next', ld_buffer)
      else obey_ed_int('buffer ~', ld_buffer);
      get_filename_from_line(filename);
      if (length(filename) = 0) or      { no filename }
         (filename[1] = '<')            { user is on pathname line }
      then display_next_error('Failed to get entry name from line.')
      else if length(ld_pathname) + 1 + length(filename) <= line_max_chars then
      begin { path length alright }
         new_path := concat(ld_pathname,'>',filename);
         test_entry(filename, new_path_exists, new_path_type, is_binary);
         if new_path_exists then
         begin { entry exists }
            if new_path_type in [sam_type, dam_type, cam_type] then
            begin { inspect file }
               if is_binary then
               begin { ask about apparent binary file }
                  display_next_message(concat(filename,
                                              ' appears to be a binary file.'));
                  if wi_question_and_answer('Do you still want to inspect it? ',
                                            true, true)
                  then is_binary := false
               end { ask about apparent binary file } ;
               if is_binary
               then obey_ed_str('next', command_worked)
               else
               begin { go ahead and inspect }
                  obey_ed_str('overlay *', command_worked);
                  obey_ed_str(concat('buffer 1;ed ',filename), command_worked)
               end { go ahead and inspect }
            end { inspect file }
            else if new_path_type = 4 then
            begin { go down into directory }
               obey_ed_str('overlay *', command_worked);
               if ld_buffer < buffer_max then
               begin { new buffer number alright }
                  ld_buffer := ld_buffer + 1;
                  ld_pathname := new_path;
                  at$rel(k$seth, filename, code);
                  if code <> 0
                  then error_print(k$irtn, code, filename, 'AT$REL');
                  write_ld_file(ld_filename);
                  obey_ed_int(concat('buffer ~; ed ',ld_filename,'; point 9'),
                              ld_buffer)
               end { new buffer number alright }
               else display_next_error('Too many levels.')
            end { go down into directory }
            else display_next_error('Wrong type of object.')
         end { entry exists }
         else { error printed in test_entry }
      end { path length alright }
      else display_next_error('Pathname too long')
   end { tidy_inspect_proc } ;

   { Mark current file/dir for deletion later }

   procedure tidy_delete_proc;

   var
      command_worked: boolean;

   begin { tidy_delete_proc }
      { do not put command after 'overlay' because mode_semi may be false }
      obey_ed_int('buffer ~; overlay d', ld_buffer);
      obey_ed_str('next', command_worked)
   end { tidy_delete_proc } ;

   { Exit from this directory level, optionally deleting any marked files }

   procedure tidy_finish_proc(do_deletes: boolean);

   var
      chindex, last_gt: integer;
      temp_treename: tree_type;
      command_worked: boolean;

   begin { tidy_finish_proc }
      if current_buffer_number <> ld_buffer
      then display_next_error('Type {DEL}L to return to file list.')
      else
      begin { in correct buffer }
         obey_ed_str('save', command_worked);
         { delete buffer 1 if it exists to close the file }
         obey_ed_str('delbuf 1', command_worked);
         if buffer_pointers[1].bptr <> nil
         then { delbuf failed - let user try again }
         else
         begin { buffer 1 clear }
            temp_treename := current_buffer.main_file.treename;
            obey_ed_str('ed', command_worked);  { just to clear screen }
            if do_deletes
            then write_cpl_file(temp_treename)
            else ld_cpl_filename := '';
            delete_primos_file(temp_treename.strng);
            if (ld_buffer > 2) and (length(ld_cpl_filename) > 0) then
            begin { perform the deletes }
               obey_ed_str(concat('! resume ',ld_cpl_filename), command_worked);
               delete_primos_file(ld_cpl_filename)
            end { perform the deletes } ;
            ld_buffer := ld_buffer - 1;
            if ld_buffer > 1 then
            begin { switch back to previous ld file buffer }
               { remove last entry from pathname }
               last_gt := 0;
               for chindex := 1 to length(ld_pathname)
               do if ld_pathname[chindex] = '>'
                  then last_gt := chindex;
               if last_gt > 0
               then delete(ld_pathname, last_gt, length(ld_pathname) - last_gt + 1);
               attach_to_long_path(ld_pathname);
               obey_ed_int('buffer ~; next', ld_buffer);
               obey_ed_int('delbuf ~', ld_buffer+1)
            end { switch back to previous ld file buffer }
            else obey_ed_str('qf', command_worked)  { CPL file will be obeyed after exit }
         end { buffer 1 clear }
      end { in correct buffer }
   end { tidy_finish_proc } ;

begin { wi_tidy_proc }
   case tidy_arg of
      tidy_ld:      tidy_ld_proc;
      tidy_inspect: tidy_inspect_proc;
      tidy_delete:  tidy_delete_proc;
      tidy_finish:  tidy_finish_proc(true);
      tidy_quit:    tidy_finish_proc(false)
   end { case } ;
   replace_cursor                       { in case error_print called }
end { wi_tidy_proc } ;

{ Perform initialisation actions on entering the window editor }

procedure entry_actions;                                                        { ENTRY_ACTIONS }

var
   oops_index: wi_oops_range;

begin { entry_actions }
   ctrl_term(rekord, false);            { note terminal state now }
   switch_como(false);                  { como off }
   set_defer(true, true);
   if not mode_lon
   then lon$cn(0);                      { switch logout notification off }
   if vdu_table = nil
   then read_vdu_table(false);
   if main_state = running then
   begin { tables alright }
      if term_is_ssmp then
      begin { SSMP }
         if init_ssmp                   { set up mode array etc. }
         then { worked alright }
         else main_state := command_error
      end { SSMP }
      else
      with vdu_table^ do
      begin { not SSMP }
         token := withhost;             { just to simplify some checks }
         { If DC1 or DC3 is used as a command then disable XON operation }
         { to allow them to be used as normal characters }
         if (action_array[dc1_code].char_action = erroract) and
            (action_array[dc3_code].char_action = erroract)
         then { DC1 and DC3 are not commands }
         else ctrl_term(xonrec, false);
         vdu_string(reset_string, 0, 0);  { reset the terminal to a suitable }
                                        { initial state }
         hook6(vdu_type)
      end { not SSMP }
   end { tables alright } ;
   if main_state = running then
   begin { alright so far }
      in_window_code := true;
      hook3(vdu_type);
      get_erase_and_kill;
      interrupt_deferred := false;
      special_limits_set := false;
      getting_single_char := false;
      character_omitted := false;
      reading_action := false;
      need_to_redraw := false;
      fill_in_undrawn_rows := false;
      first_update_row := 1;
      last_update_row := 1;
      suppress_output := false;
      message_lines_to_clear := 0;
      continue_screen_messages := true;
      screen_tabs_set := false;         { set tabs on entry }
      waiting_for_input := false;
      macro_dropout := false;
      current_macro_desc := nil;
      echoing_on := true;
      half_state := false;              { not in half intensity }
      for oops_index := 1 to wi_oops_max  { clear out stack }
      do wi_oops_stack[oops_index] := nil;
      wi_oops_top := 0
   end { main_state = running }
   else switch_como(true)               { como back on again }
end { entry_actions } ;

{ Perform tidying-up actions on leaving the window editor }

procedure exit_actions;                                                         { EXIT_ACTIONS }

var
   oops_index: wi_oops_range;

begin { exit_actions }
   with vdu_table^ do
   begin
      if term_is_ssmp
      then stop_ssmp
      else
      begin
         ctrl_term(reinstate, false);   { restore terminal state }
         if mess_type1                  { tidy up terminal }
         then vdu_control(messoff);
         if arrow_type1
         then vdu_control(arrowoff);
         vdu_control(foldoff);
         vdu_control(cleartabs);
         vdu_string(exit_string, 0, 0)  { send exit string }
      end { not ssmp } ;
      for oops_index := 1 to wi_oops_max
      do if wi_oops_stack[oops_index] <> nil
         then dispose(wi_oops_stack[oops_index]);
      switch_como(true);
      set_defer(false, false);
      if macro_head <> nil
      then emacro_comm(false);          { terminate the current macro }
      in_window_code := false;
      hook4(vdu_type);
      get_erase_and_kill;
      lon$cn(1)                         { switch logout notification on }
   end { with }
end { exit_actions } ;

{ On-unit for QUIT$ }

procedure wi_quit_onunit(not_used: integer);                                    { WI_QUIT_ONUNIT }

var
   code: shortint;

begin { wi_quit_onunit }
   suppress_output := false;            { just to ensure screen controls work }
   if term_is_ssmp then
   begin
      if token = withterm
      then get_token
      else interrupt_deferred := true   { deal with it later }
   end
   else
   begin
      tty$rs([bit1,bit2], code);        { clear terminal buffers }
      vdu_string(vdu_table^.reset_string, 0, 0);  { reset in case the user has }
                                        { pressed BREAK because he is in }
                                        { trouble with the terminal }
      hook6(vdu_type)
   end;
   lowest_empty_buffer := 1;
   if interrupt_deferred
   then main_state := quit_signal       { stop any efs action }
   else deal_with_interrupt
end { wi_quit_onunit } ;

{ Onunit for LOGOUT$ and ERROR_TRAP - tidy up and exit WI, then continue }
{ to signal }

procedure wi_general_onunit(not_used: integer);                                 { WI_GENERAL_ONUNIT }

var
   code: shortint;

begin { wi_general_onunit }
   { position to bottom of screen }
   vdu_position(vdu_table^.rows_per_page, 1);
   writeln;
   exit_actions;                        { perform exitwi actions }
   cnsig$(code)
end { wi_general_onunit } ;

%page
{************************************}
{ MAIN LOOP TO OBEY INPUT CHARACTERS }
{************************************}

{ Obey a macro or command read from the terminal or a macro buffer }

procedure obey_input_command;                                                   { OBEY_INPUT_COMMAND }
                            { desc: ptr_macro_desc; action: action_proc); }

begin { obey_input_command }
   if desc <> nil then
   with desc^ do
   begin { macro takes precedence over action }
      if macro_being_obeyed
      then display_error('Macro is already being obeyed.')
      else
      begin { not already in use }
         macro_being_obeyed := true;
         current_buf := code_start;
         codes_used := 0;
         previous := current_macro_desc;
         current_macro_desc := desc
      end { not already in use }
   end { obey macro }
   else
   begin { obey command }
      case action of
         breakline:   breakline_comm;
         carret:      carret_comm;
         deleteonech: deleteonech_comm(true);
         deleteonel:  deleteonel_comm(true);
         dofile:      if command_line_args.tidy_sw
                      then wi_tidy_proc(tidy_finish)
                      else dofile_or_dosave_proc(true);
         doquit:      if command_line_args.tidy_sw
                      then wi_tidy_proc(tidy_quit)
                      else doquit_comm;
         dosave:      dofile_or_dosave_proc(false);
         downmany:    move_many_comm(cursordown);
         downscreen:  downscreen_comm;
         dunlfile:    if command_line_args.tidy_sw
                      then wi_tidy_proc(tidy_delete)
                      else unlfile_comm(true);
         duplicate:   insertonel_comm(do_duplicate, true);
         edcom:       edcom_comm;
         emacro:      emacro_comm(true);
         eraseline:   eraseline_comm;
         erroract:    erroract_comm;
         exitwi:      exitwi_comm;
         formtext:    formtext_comm;
         godown:      godown_comm(false, cursordown);
         godownx:     godown_comm(true , cursordown);
         gohome:      gohome_comm;
         goleft:      goleft_comm(false);
         goleftx:     goleft_comm(true);
         goright:     goright_comm(false);
         gorightx:    goright_comm(true);
         goup:        goup_comm(false, cursorup);
         goupx:       goup_comm(true , cursorup);
         helpact:     helpact_comm;
         insertmode:  if command_line_args.tidy_sw
                      then wi_tidy_proc(tidy_inspect)
                      else insertmode_comm;
         insertonech: insertonech_comm(true);
         insertonel:  insertonel_comm(do_blank, true);
         joinline:    joinline_comm;
         lfchar:      godown_comm(false, backtab);
         lfcharx:     godown_comm(true , backtab);
         lfile:       if command_line_args.tidy_sw
                      then wi_tidy_proc(tidy_ld)
                      else lfile_comm;
         lowcase:     up_and_lowcase_proc(false);
         markon:      markit_comm(true);
         markoff:     markit_comm(false);
         nullact:     ;
         openit:      openit_comm(true);
         closeit:     openit_comm(false);
         pointchon:   pointch_comm(true);
         pointchoff:  pointch_comm(false);
         redrawl:     redrawl_comm;
         redraws:     redraws_comm;
         repeatit:    repeatit_comm;
         revlfchar:   goup_comm(false, backtab);
         revlfcharx:  goup_comm(true , backtab);
         smacro:      smacro_comm(true);
         dmacro:      smacro_comm(false);
         ispaces:     spaces_comm(true);
         dspaces:     spaces_comm(false);
         storeonech:  storeonech_comm;
         tableft:     tableft_comm;
         tabright:    tabright_comm;
         tabstopon:   tabstop_comm(true);
         tabstopoff:  tabstop_comm(false);
         unlfile:     unlfile_comm(false);
         upcase:      up_and_lowcase_proc(true);
         upmany:      move_many_comm(cursorup);
         upscreen:    upscreen_comm;
         wiswitch:    wiswitch_comm;
         xcomm:       xcomm_comm(true);
         zap:         zap_comm;
         litchar:     storeonech_comm;
         sinthost:    ;
         ssuspend:    ;
         srestart:    redraws_comm;
         sleftupd:    ;
         sfirstns:    ;
         salastns:    ;
         seraprev:    seraprev_comm;
         seraline:    ;
         seraright:   eraseline_comm;
         shomefld:    ;
         snewline:    ;
         nextwind:    nextwind_comm(go_forwards);
         prevwind:    nextwind_comm(go_backwards)
         { otherwise can happen when running old editor with new vdu tables }
         otherwise    erroract_comm
      end { case }
   end { obey command } ;
   if interrupt_deferred then
   begin
      deal_with_interrupt;
      interrupt_deferred := false
   end
end { obey_input_command } ;

%page
{************************************}
{ INITIAL COMMAND ANALYSIS AND SETUP }
{************************************}

begin { window_code }
   entry_actions;
   if main_state = running then         { check OK so far }
   with current_buffer, current_window, vdu_table^ do
   begin
      { if screen not split then use special [0] window }
      { start in window 0 unless the user's current buffer is one of the }
      { ones in the split }
      current_window_number := 0;
      for windex := 1 to window_count
      do if window_array[windex].window_buffer = current_buffer_number
         then current_window_number := windex;
      screen_is_split := current_window_number > 0;
      if not screen_is_split
      then with window_array[0] do
      begin { not in split }
         window_buffer := current_buffer_number;
         current_window_number := 0
      end { not in split } ;
      { set window mode to init_wi in first window, and WI mode in others }
      for windex := 0 to window_count
      do with window_array[windex]
         do if windex = current_window_number
            then window_mode := init_wi
            else window_mode := true;
      current_window := window_array[current_window_number];
      switch_buffer(window_buffer);
      entry_line_position := current_buffer.line_position;
      if term_is_ssmp
      then always_output := false       { never need to echo for SSMP }
      else
      begin { not ssmp }
         always_output := mode_margin or mode_direct or
                          screen_is_split or mode_insert;
         if always_output
         then ctrl_half_duplex(true)    { we will do our own echoing }
      end { not smmp } ;
      { process and check arguments }
      if window_mode then
      begin { WINDOW command }
         if number_of_args = 0 then
         begin { no arguments given - set default }
            arg1 := -(window_length div 2);  { to fill window }
            arg2 :=  (window_length - 1) div 2
         end { no arguments given - set default }
         else if (arg2 = 0) and (arg1 > 0) then
         begin { WI 5 => WI 0 5 }
            arg2 := arg1;
            arg1 := 0
         end { WI 5 => WI 0 5 } ;
         { adjust arguments to fit in window if necessary }
         if arg1 > 0                    { arg1 must be <= 0 }
         then arg1 := 0;
         if arg2 < 0                    { arg2 must be >= 0 }
         then arg2 := 0;
         if (arg2 - arg1) > (window_length - 1) then
         begin { more than a windowful }
            arg2 := arg1 + window_length- 1;  { adjust arg2 back }
            if arg2 < 0 then
            begin { arg1 must have been large and -ve }
               arg2 := 0;
               arg1 := 1 - window_length
            end { arg1 must have been large and -ve }
         end { more than a windowful } ;
         full_screen := (arg2 - arg1) = (window_length - 1)
      end { WI }
      else
      begin { IW command }
         if number_of_args = 0          { IW => IW -5 }
         then arg1 := -5
         else if arg1 > 0               { IW 3 => IW -3 }
         then arg1 := -arg1;
         if arg1 < (1 - window_length)  { more than a windowful? }
         then arg1 := 1 - window_length;
         full_screen := true;
         { insert first blank line }
         if line_exists and (line_status = normal)
         then                           { current line ok }
         else load_a_line(line_number - 1, false);  { close the gap }
         if mode_indent
         then find_indent(entry_line_position, window_width)  { start at current indent col }
         else entry_line_position := 1;
         insert_blank_line;
         iw_marked_line := line_number + 1  { set flag on next line }
      end { IW command } ;
      on('QUIT$', wi_quit_onunit);
      on('LOGOUT$', wi_general_onunit);
      on('ERROR_TRAP', wi_general_onunit);
      setup_whole_screen(arg1, arg2);
      draw_whole_screen;
      position_cursor_in_window(false, entry_line_position, false);
      if mode_disp then                 { MODE DISP: display filename on entry }
      begin
         announce_buffer_name;
         replace_cursor
      end;
      exit_window := false;
      repeat
         main_state := running;         { always clear this }
         cominput_was_on := cominput_on;
         if cominput_on
         then read_cominput_line(ed_com_line);
         if cominput_on then
         begin { obey command from cominput }
            obey_ed_command(ed_com_line, main_command_worked)
         end { obey command from cominput }
         else
         begin { read and obey }
            if cominput_was_on
            then replace_cursor;        { because of 'eof' message }
            read_input_command(true, macro_ptr, action_code);
            obey_input_command(macro_ptr, action_code)
         end { read and obey } ;
         if main_state = quit_signal then
         begin { user quit }
            form_draw_window(false);
            int_screen_message('Quit at line ~', true, line_number, 1);
            if cominput_on
            then stop_cominput;
            replace_cursor
         end { user quit }
      until exit_window;
   311:
      strip_trailing_spaces;            { do not leave as spurious length }
      iw_marked_line := 0;
      set_half(false);
      if screen_is_split
      then rindex := rows_per_page
      else
      begin { not split }
         rindex := rows_per_page;       { find last non-blank line }
         while ( (window_status[rindex].row_status = row_null) or
               (not window_status[rindex].row_drawn)) and
               (rindex > 1)
         do rindex := rindex - 1
      end { not split } ;
      if (rindex >= (rows_per_page - 2)) and lineerase_avail then
      begin
         for rindex := rows_per_page - 2 to rows_per_page do
         begin { clear last few lines of screen }
            vdu_position(rindex, 1);
            vdu_control(lineerase)
         end { clear last few lines of screen } ;
         vdu_position(rows_per_page - 3,1)  { leave cursor near screen bottom }
      end
      else vdu_position(rindex, 1);     { leave on last non-blank line }
      exit_actions;
      writeln;
      if main_state <> quit_comm
      then writeln('Edit')
   end { with } ;
   command_line_args.window_sw := false
end { window_code } ;

begin { pmed_wi }
end { pmed_wi } .
