%separate = true
%ecs = true
%ttybuffered = false
%epfargs = true
{ Sheffield University Editor }
{ Copyright (c) University of Sheffield 1990 }

{**************************************}
{ PROCEDURES USED BY THE WINDOW EDITOR }
{**************************************}

subprogram pmed_wprocs;

%listing off
%insert pmed_globals.ins.pas
%insert pmed_efs.ins.pas
%insert pmed_procs.ins.pas
%insert pmed_ssmp.ins.pas
%insert pmed_vdu.ins.pas
%listing on

   { Move "wd_count" words from "from_string[fromsub]" to "to_string[tosub]". }
   { Preserves the words moved by copying either forwards or backwards        }

   procedure movewds(from_string: line_chars_type;
                     fromsub: integer;
                     var to_string: line_chars_type;
                     tosub: integer;
                     wd_count: integer);
      extern;                           { written in PMA }

   procedure obey_command_line(command_line: line_chars_type;
                               times: integer;
                               upto: line_chars_plus_range);
      extern;

{ Set or reset the user's message status; we set it to -defer during }
{ window editing so that the screen is not messed up by messages, but }
{ this can be overridden by MODE MSG }

procedure set_defer(rekord, switch_on: boolean);                                { SET_DEFER }

type
   timdat_type = record
      not_used: array[1..11] of shortint;
      userno: shortint
   end;

var
   timdat_info: timdat_type;
   code: shortint;

   procedure timdat(var timdat_info: timdat_type;
                    length: shortint);
    extern;

   procedure msg$st(key: shortint;
                    userno: shortint;
                    notused1: shortint;
                    notused2: shortint;
                    notused3: shortint;
                    notused4: shortint;
                    var msgst: shortint);
      extern;

   procedure mgset$(msgst: shortint; var code: shortint); extern;

begin { set_defer }
   if rekord then
   begin { record current message status }
      timdat(timdat_info, 12);          { get userno }
      msg$st(1, timdat_info.userno, 0, 0, 0, 0, msgst)  { get old status }
   end { record current message status } ;
   if switch_on then
   begin { switch to requested state }
      if mode_msg
      then mgset$(k$acpt, code)         { set -accept }
      else mgset$(k$defr, code)         { set -defer }
   end { switch to requested state }
   else mgset$(msgst, code)             { restore old status }
end { set_defer } ;

{ Position vdu cursor to given place relative to current window }

procedure window_position(window_row: row_range; window_col: col_range);        { WINDOW_POSITION }

begin { window_position }
   with current_window
   do vdu_position(window_row + window_top  - 1,
                   window_col + window_left - 1)
end { window_position } ;

{ After outputting a character at last column of screen we position to (1,1) }
{ and send LF to foil wraparound action on terminals which ignore next LF }

procedure foil_wraparound;                                                      { FOIL_WRAPAROUND }

begin { foil_wraparound }
   if term_is_ssmp
   then { no problem with SSMP terminals }
   else with current_window, vdu_table^ do
   begin
      if window_right = cols_per_row then
      begin { really need to do it }
         vdu_position(1, 1);
         print_a_char(newline)
      end { really need to do it }
   end { with }
end { foil_wraparound } ;

{ Perform a given cursor movement unless doing so would violate window edges }

procedure do_cursor_move(the_control: controls);                                { DO_CURSOR_MOVE }

begin { do_cursor_move }
   if screen_is_split then
   with current_window do
   begin
      case the_control of
         cursorup:
            if cursor_row > 1
            then vdu_control(cursorup);
         cursordown:
            if cursor_row < window_length
            then vdu_control(cursordown);
         cursorleft:
            if cursor_col > 1
            then vdu_control(cursorleft);
         cursorright:
            if cursor_col < window_width
            then vdu_control(cursorright)
      end { case }
   end { screen_is_split, with }
   else vdu_control(the_control)
end { do_cursor_move } ;

{ Do a lineinsert operation, allowing for SSMP terminals and split screen }

procedure do_lineinsert(the_row: row_range; the_number: integer);               { DO_LINEINSERT }

var
   rindex: integer;

begin { do_lineinsert }
   if the_number = 0
   then { no need to do anything }
   else with current_window, vdu_table^ do
   begin
      if term_is_ssmp then
      begin { use SSMP multiple scroll }
         { remember then hscrolldown uses absolute screen coordinates }
         { counting from zero }
         hscrolldown(the_row + window_top - 2, window_bottom - 1, the_number)
      end { ssmp }
      else
      begin
         for rindex := 1 to the_number do
         begin
            if data_protect or screen_is_split then
            begin { first delete line at end }
               window_position(window_length, 1);
               vdu_control(linedelete);
               window_position(the_row, 1)
            end { first delete line at end } ;
            vdu_control(lineinsert)
         end { for loop }
      end { not ssmp terminal }
   end { with }
end { do_lineinsert } ;

{ Do a linedelete operation, allowing for SSMP terminals and split screen }

procedure do_linedelete(the_row: row_range; the_number: integer);               { DO_LINEDELETE }

var
   rindex: integer;

begin { do_linedelete }
   if the_number = 0
   then { no need to do anything }
   else with current_window do
   begin
      if term_is_ssmp then
      begin { use SSMP multiple scroll }
         { remember then hscrollup uses absolute screen coordinates }
         { counting from zero }
         hscrollup(the_row + window_top - 2, window_bottom - 1, the_number)
      end { ssmp }
      else
      begin
         for rindex := 1 to the_number do
         begin
            vdu_control(linedelete);
            if screen_is_split then
            begin { do compensating insert at end }
               window_position(window_length, 1);
               vdu_control(lineinsert);
               window_position(the_row, 1)
            end { do compensating insert at end }
         end { for loop }
      end { not ssmp terminal }
   end { with }
end { do_linedelete } ;

{ Move to a given line in the editor workfile unless already there }

procedure move_to_line(ln: integer);                                            { MOVE_TO_LINE }

var
   cline: integer;

begin { move_to_line }
   with current_buffer do
   if line_status = top
   then cline := 0                      { .top. is really line 0 }
   else cline := line_number;
   if ln <> cline then
   begin
      load_a_line(ln, false);
      strip_trailing_spaces             { to include char pointer if nec }
   end
end { move_to_line } ;

{ Add the current line to top of the OOPS stack, losing oldest if necessary }

procedure push_line;                                                            { PUSH_LINE }

var
   oops_index: wi_oops_range;
   temp_ptr: ptr_edline;

begin { push_line }
   if wi_oops_top < wi_oops_max then
   begin { add to top of stack }
      wi_oops_top := wi_oops_top + 1;
      if wi_oops_stack[wi_oops_top] = nil
      then new(wi_oops_stack[wi_oops_top])
   end { add to top of stack }
   else
   begin { remove bottom and shuffle down one }
      temp_ptr := wi_oops_stack[1];
      for oops_index := 2 to wi_oops_max
      do wi_oops_stack[oops_index-1] := wi_oops_stack[oops_index];
      wi_oops_stack[wi_oops_max] := temp_ptr
   end { remove bottom and shuffle down one } ;
   { copy current line to top of stack }
   with wi_oops_stack[wi_oops_top]^, current_buffer do
   begin
      ed_text := line_text;
      text_length := line_length;
      being_obeyed := false
   end { with }
end { push_line } ;

{ Copy the top of the OOPS stack to current line and remove entry }

procedure pop_line;                                                             { POP_LINE }

begin { pop_line }
   with current_buffer do
   begin
      if wi_oops_top = 0 then
      begin { stack empty }
         line_length := 0   { null line }
      end { stack empty }
      else with wi_oops_stack[wi_oops_top]^ do
      begin { stack not empty }
         line_text := ed_text;
         line_length := text_length;
         wi_oops_top := wi_oops_top - 1
      end { stack not empty } ;
      line_altered := true
   end { with }
end { pop_line } ;

{ Print a string on the terminal }

procedure print_a_string(the_string: string);                                   { PRINT_A_STRING }

var
   chindex: 1..maxint;

begin { print_a_string }
   for chindex := 1 to length(the_string)
   do print_a_char(the_string[chindex])
end { print_a_string } ;

{ Switch command output off or on }

procedure switch_como(swon: boolean);                                           { SWITCH_COMO }

const
   como_unit = -4;

var
   code: shortint;
   como_path: tree_type;

begin { switch_como }
   if swon then
   begin { switch on at exit }
      if como_switched_off then
      begin { switch como back on again }
         como$$(20b, 0, 0, 0, code);
         if code <> 0
         then error_print(k$irtn, code, 'Switching on', 'COMO$$')
      end { switch como back on again }
   end { switch on }
   else
   begin { switch off at entry }
      como_switched_off := false;       { usual default }
      if mode_como
      then                              { leave como alone }
      else
      with como_path do
      begin
         gpath$(k$unit, como_unit, chars, tree_max_chars, len, code);
         if code = 0 then
         begin
            como_switched_off := true;
            como$$(10b, 0, 0, 0, code);
            if code <> 0
            then error_print(k$irtn, code, 'Switching off', 'COMO$$')
         end
         else { GPATH$ failed: probably COMO file not open }
      end { with }
   end { switch off }
end { switch_como } ;

{ Set vdu into or out of half intensity, and remember }

procedure set_half(new_state: boolean);                                         { SET_HALF }

begin { set_half }
   if new_state <> half_state then
   begin { need to change state }
      half_state := new_state;
      case half_state of
         true:  vdu_control(foldon);
         false: vdu_control(foldoff)
      end { case }
   end { need to change state }
end { set_half } ;

{ Replace screen cursor at recorded cursor position and move to corresponding }
{ line and position in file }

procedure replace_cursor;                                                       { REPLACE_CURSOR }

var
   try_line_position: integer;

begin { replace_cursor }
   with current_buffer, current_window, window_status[cursor_row] do
   begin
      window_position(cursor_row, cursor_col);
      move_to_line(file_line);          { ensure on correct file line }
      try_line_position := cursor_col + offset;
      if try_line_position > line_max_charsm1
      then line_position := line_max_charsm1
      else line_position := try_line_position  { set correct line_position }
   end { with }
end { replace_cursor } ;

{ Force the cursor down one row, leaving column unchanged }

procedure print_newline;                                                        { PRINT_NEWLINE }

begin { print_newline }
   with current_window do
   begin
      if screen_is_split and (cursor_row = window_length) then
      begin { at bottom of window }
         if line_ins_del_avail and window_full_width then
         begin { use line ins/del }
            window_position(1, 1);
            vdu_control(linedelete);
            window_position(window_length, 1);
            vdu_control(lineinsert);
            replace_cursor
         end { use line ins/del }
         else suppress_output := true   { redraw whole window }
      end { at bottom of window }
      else if term_is_ssmp
      then do_ssmp_lf
      else print_a_char(newline)
   end { with }
end { print_newline } ;

{ Put the arrow marker on a screen row }

procedure mark_the_line(draw_row: row_range);                                   { MARK_THE_LINE }

begin { mark_the_line }
   if not suppress_output then
   with vdu_table^ do
   begin
      if arrow_type1 then
      begin { type 1 }
         window_position(draw_row, 1);
         print_a_char(' ');
         vdu_control(arrowon);
         print_a_string('--->');
         vdu_control(arrowoff);
         print_a_char(' ')
      end { type 1 }
      else
      begin { type 2 }
         window_position(draw_row, 6);  { switch inverse off first to avoid }
         vdu_control(arrowoff);         { the whole screen going white }
         window_position(draw_row, 1);
         vdu_control(arrowon);
         print_a_string('--->')
      end { type 2 } ;
      replace_cursor
   end
end { mark_the_line } ;

{ Set screen tabulation stops to columns in tabs array }

procedure set_screen_tabs;                                                      { SET_SCREEN_TABS }

var
   cindex: tabs_range;
   tabpos: line_chars_range;

begin { set_screen_tabs }
   if screen_tabs_set
   then { already set - no need to continue }
   else if suppress_output
   then { control chars would have no effect }
   else with current_window do
   begin
      screen_tabs_set := true;
      if tabs_avail then
      begin
         vdu_control(cleartabs);
         { do not set tabs for vertical split }
         if window_full_width then
         begin
            for cindex := 1 to number_of_tabs do
            begin
               tabpos := htabs[cindex];
               if tabpos <= window_width then  { if on screen }
               begin { set physical tab }
                  vdu_position(1, htabs[cindex]);
                  vdu_control(settab)
               end { in range }
            end { for loop }
         end { window_full_width }
      end { tabs_avail }
   end { not already set and output not suppressed, with }
end { set_screen_tabs } ;

{ Find the first row on the window containing part or all of the given }
{ line number, returning "true" if found }

function find_first(theline: integer; var row_first: row_range): boolean;       { FIND_FIRST }

var
   state: search_type;

begin { find_first }
   with current_window do
   if (theline < window_status[1].file_line) or
      (theline > window_status[window_length].file_line)
   then find_first := false             { way off screen }
   else
   begin
      row_first := 1;
      state := searching;
      repeat
         if window_status[row_first].file_line = theline
         then state := found
         else if row_first < window_length
         then row_first := row_first + 1
         else state := not_found
      until state <> searching;
      find_first := state = found
   end
end { find_first } ;

{ Find the last row on the screen containing part or all of the current
  file line }

procedure find_last(var row_last: row_range);                                   { FIND_LAST }

var
   ln: integer;

begin { find_last }
   with current_buffer, current_window do
   begin
      if line_status = top              { treat .top. as line 0 }
      then ln := 0
      else ln := line_number;
      row_last := window_length;
      while window_status[row_last].file_line <> ln
      do row_last := row_last - 1
   end { with }
end { find_last } ;

{ Move down screen, setting up status as we go, starting from a given row }
{ which is already set up }

procedure move_down_screen(var row_index: row_range);                           { MOVE_DOWN_SCREEN }

var
   rindex: row_range;
   state: search_type;
   ln, off: integer;

begin { move_down_screen }
   with current_buffer, current_window do
   begin
      with window_status[row_index] do
      begin
         ln := file_line;               { line number in editor workfile }
         off := offset                  { start offset }
      end;
      state := searching;
      repeat
         move_to_line(ln);
         if row_index >= window_length
         then state := found            { hit bottom of screen }
         else
         begin
            off := off + window_width;  { increase offset by one window width }
            if off >= line_length then
            begin { on to next line }
               off := 0;
               ln := ln + 1
            end { on to next line } ;
            row_index := row_index + 1;
            with window_status[row_index] do
            if (line_status = bottom) or  { line below .bottom. }
               ((not window_mode) and   { or IW and >= mark }
                (ln >= iw_marked_line))
            then
            begin { null line }
               file_line := large_int;
               offset := 0;
               row_status := row_null;
               row_drawn := false
            end { null line }
            else
            begin { normal line }
               file_line := ln;
               offset := off;
               row_status := row_normal;
               row_drawn := window_status[row_index - 1].row_drawn  { copy from above }
            end { normal line }
         end { still on screen }
      until state <> searching
   end { with }
end { move_down_screen } ;

{ Move up the screen, filling in window_status as we go }

procedure move_up_screen(var row_index: row_range;                              { MOVE_UP_SCREEN }
                         allow_borrow: boolean);

var
   rindex: row_range;
   state: search_type;
   ln, off: integer;

begin { move_up_screen }
   with current_buffer, current_window do
   begin
      with window_status[row_index] do  { get initial values from line }
      begin
         ln := file_line;
         off := offset
      end;
      state := searching;
      repeat
         if line_status = top
         then state := not_found
         else
         begin
            { If we have hit the top of the screen, we may be able to borrow }
            { a null line from the bottom and carry on }
            if row_index = 1            { at top of screen }
            then  if allow_borrow and   { if allowed }
                     window_mode and    { only do it for window }
                     (window_status[window_length].row_status = row_null) then
                  begin { borrow the line }
                     for rindex := window_length downto 2
                     do window_status[rindex] := window_status[rindex-1];
                     row_index := row_index + 1
                  end { borrow the line }
                  else state := found;  { really hit top of screen }
            if state = searching then
            begin { now have empty row to process }
               row_index := row_index - 1;
               with window_status[row_index] do
               begin
                  if off > 0            { folded line? }
                  then off := off - window_width
                  else
                  begin
                     ln := ln - 1;      { move to previous line }
                     move_to_line(ln);
                     off := ((line_length - 1) div window_width) * window_width
                  end;
                  file_line := ln;
                  offset := off;
                  row_status := row_normal;
                  row_drawn := window_status[row_index+1].row_drawn  { copy from below }
               end
            end { now have empty row to process }
         end { not top }
      until state <> searching
   end { with }
end { move_up_screen } ;

{ Set up the window status array for a part of the file pivoting around }
{ the current line in the file. }

procedure setup_window_status(arg1, arg2: integer);                             { SETUP_WINDOW_STATUS }

var
   saved_file_place: file_places;
   row_index, rindex, pivot_row: row_range;
   pivot_line: integer;
   start_offset: 0..maxint;
   drawall, row_found: boolean;

begin { setup_window_status }
   with current_buffer, current_window do
   begin
      { If screen is split and WI mode then force arg values to fill window }
      if screen_is_split and window_mode then
      begin { force full window }
         arg1 := -(window_length div 2);
         arg2 := (window_length - 1) div 2
      end { force full window } ;
      pivot_row := 1 - arg1;
      if window_mode then
      begin { WI mode }
         if (not line_exists)           { do not start on .top. }
         then load_a_line(line_number, false);  { or a null line }
         iw_marked_line := 0
      end { WI mode } ;
      save_place(saved_file_place);
      if line_status = top
      then pivot_line := 0              { treat .top. as line zero }
      else pivot_line := line_number;

      { set up the first pivotal row }

      with window_status[pivot_row] do
      begin
         file_line := pivot_line;
         offset := (line_position - 1) div window_width * window_width;
         start_offset := offset;
         row_status := row_normal;
         row_drawn := true
      end { with } ;

      { Start at pivot and move down the screen }

      row_index := pivot_row;
      move_down_screen(row_index);

      { Start at pivot and move up the screen }

      row_index := pivot_row;
      move_up_screen(row_index, true);

      { We may not have reached top of screen because of hitting top of file }

      if row_index > 1 then
      begin

         { Copy row statuses up to top of screen }

         for rindex := 1 to (window_length - row_index + 1)
         do  window_status[rindex] := window_status[rindex-1+row_index];
         { pivot_row is now further up the screen }
         pivot_row := pivot_row - row_index + 1;

         { Fill in rest of index by moving down }

         row_index := pivot_row;
         move_down_screen(row_index)
      end { row_index > 1 } ;
      cursor_row := 1;
      row_found := false;
      repeat
         with window_status[cursor_row] do
         begin
            if (file_line = pivot_line) and (offset = start_offset)
            then row_found := true
            else cursor_row := cursor_row + 1
         end
      until row_found;

      { indicate the subset of lines to be drawn initially }

      drawall :=                        { true if all window to be drawn }
         (arg2 - arg1 + 1) = window_length;
      for rindex := 1 to window_length
      do window_status[rindex].row_drawn :=
            drawall or
          ( (rindex >= (cursor_row + arg1)) and
            (rindex <= (cursor_row + arg2)) );
      restore_place(saved_file_place);
      cursor_col := (line_position - 1) mod window_width + 1
   end { with }
end { setup_window_status } ;

{ Put a character on the screen, including all processing for special cases }

procedure put_char_on_screen(ch: char);                                         { PUT_CHAR_ON_SCREEN }

begin { put_char_on_screen }
   if term_is_ssmp
   then print_a_char(ch)                { SSMP can handle any character }
   else if ord(ch) in [000b..037b,200b..237b] then
   begin { control char }
      if ctrlcode_avail then
      begin { vdu can cope }
         vdu_control(ctrlcodeon);
         print_a_char(ch);
         vdu_control(ctrlcodeoff)
      end { vdu can cope }
      else print_a_char('?')
   end { control char }
   else if ch > last_print_char
   then print_a_char('?')
   else print_a_char(ch)
end { put_char_on_screen } ;

{ Switch to a given window number }

procedure switch_window(new_window_number: window0_range);                      { SWITCH_WINDOW }

begin { switch_window }
   if new_window_number = current_window_number
   then { nothing to do }
   else
   begin { different window }
      window_array[current_window_number] := current_window;
      current_window_number := new_window_number;
      current_window := window_array[current_window_number];
      switch_buffer(current_window.window_buffer)
   end { different window }
end { switch_window } ;

{ Draw bars to delineate split screen; only called when screen is split }

procedure draw_window_frame(first_row, last_row: row_range;                     { DRAW_WINDOW_FRAME }
                            do_all_characters: boolean);

var
   rindex: row_range;
   cindex: col_range;
   windex: window_range;
   last_used: integer;

begin { draw_window_frame }
   { calculate last line of screen actually used for windows }
   last_used := windows_down*rows_down - 1;
   if last_row > last_used
   then last_row := last_used;
   for rindex := first_row to last_row do
   begin
      if (rindex mod rows_down) = 0 then
      begin { horizontal bar }
         vdu_position(rindex, 1);
         for windex := 1 to windows_across do
         begin
            for cindex := 1 to cols_across - 1
            do print_a_char('-');
            if windex <> windows_across
            then print_a_char('+')
         end { for loop }
      end { horizontal bar }
      else if do_all_characters then
      begin { vertical bar with spaces }
         vdu_position(rindex, 1);
         for windex := 1 to windows_across do
         begin
            for cindex := 1 to cols_across - 1
            do print_a_char(' ');
            if windex <> windows_across
            then print_a_char('|')
         end { for loop }
      end { vertical bar with spaces }
      else
      begin { fill in columns }
         for windex := 1 to windows_across-1 do
         begin
            vdu_position(rindex, windex*cols_across);
            print_a_char('|')
         end { for loop }
      end { fill in columns }
   end { for loop } ;
   if term_is_ssmp
   then { no problem with SSMP terminals }
   else
   begin
      vdu_position(1, 1);
      print_a_char(newline)
   end { not ssmp }
end { draw_window_frame } ;

{ Set up status for all windows on screen }

procedure setup_whole_screen(arg1, arg2: integer);                              { SETUP_WHOLE_SCREEN }

var
   saved_window, windex: window_range;

begin { setup_whole_screen }
   if current_window_number > 0 then
   begin { screen is split }
      saved_window := current_window_number;
      for windex := 1 to window_count do
      begin
         switch_window(windex);
         setup_window_status(arg1, arg2)
      end { for loop } ;
      switch_window(saved_window)
   end { screen is split }
   else setup_window_status(arg1, arg2)
end { setup_whole_screen } ;

{ Draw one line of the window, consisting of part or all of a file line. }
{ Make use of cursor positioning to speed up output.                     }
{ If lineerase required first and not available, then simulate it by     }
{ outputting spaces.                                                     }

procedure draw_window_line(draw_row: row_range;                                 { DRAW_WINDOW_LINE }
                           replace_after: boolean;
                           already_clear: boolean);

var
   spaces: line_chars0_range;
   start_off, stop_pos, write_col, lindex: integer;
   code: shortint;
   test_line_pos: line_chars0_range;
   need_scrolloff, first_pos_done, last_col_full,
   need_long_way: boolean;
   line_ch: char;

   { Position to start of draw_row if needed }

   procedure position_if_needed;
   begin { position_if_needed }
      if not first_pos_done then
      begin
         window_position(draw_row, 1);
         first_pos_done := true
      end
   end { position_if_needed } ;

   { Write out a character, making use of cursor addressing if worthwhile }

   procedure write_a_ch(ch: char);

   var
      spindex: integer;

   begin { write_a_ch }
      if (ch = ' ') and (not need_long_way)
      then spaces := spaces + 1         { just stack up spaces }
      else
      begin
         if spaces > 0 then
         begin { output the stacked spaces first }
            if (spaces >= 8) or         { big jump or }
               (not first_pos_done)     { first visible part of line }
            then
            begin { it is worth positioning }
               window_position(draw_row, write_col + spaces);
               first_pos_done := true
            end { it is worth positioning }
            else
            begin { just output the spaces }
               position_if_needed;
               for spindex := 1 to spaces
               do print_a_char(' ')
            end { just output the spaces } ;
            write_col := write_col + spaces;
            spaces := 0
         end { output spaces }
         else position_if_needed;
         put_char_on_screen(ch);
         write_col := write_col + 1
      end
   end { write_a_ch } ;

   { Write out .top. or .bottom. }

   procedure write_a_word(theword: word);

   var
      windex: word_range;
      lindex, last_col: integer;

   begin { write_a_word }
      position_if_needed;
      for windex := 1 to word_max_chars
      do print_a_char(theword[windex]);
      if need_long_way then
      with current_window, vdu_table^ do
      begin
         { avoid placing character at bottom right of screen }
         if (window_right = cols_per_row) and (window_bottom = rows_per_page)
         then last_col := window_width - 1
         else last_col := window_width;
         for lindex := word_max_chars + 1 to last_col
         do print_a_char(' ');
         if last_col = window_width
         then foil_wraparound
      end { need_long_way, with }
   end { write_a_word } ;

begin { draw_window_line }
   if suppress_output
   then { user has typed ahead - suppress redundant output until cleared }
   else with current_buffer, current_window, window_status[draw_row],
             vdu_table^ do
   if (row_drawn and (row_status <> row_null)) or (not already_clear) then
   begin
      need_long_way := (not already_clear) and  { awkward combination }
         (not (lineerase_avail and window_at_right));
      move_to_line(file_line);
      if (not already_clear) or         { must draw to over-write }
         (line_length > 0) or           { actually draw if not blank }
         (line_status <> normal)        { at .top. or .bottom. }
      then
      begin { only draw visible lines }
         if point_line_no = file_line   { set "test_line_pos" for printing }
         then test_line_pos := point_line_pos  { column marker }
         else test_line_pos := 0;
         first_pos_done := false;       { not positioned yet }
         if (not already_clear) and window_at_right then
         begin { need to clear the row }
            position_if_needed;
            vdu_control(lineerase)
         end { need to clear the row } ;
         set_half(offset > 0);
         if row_drawn and (row_status <> row_null) then
         begin { really draw the line }
            case line_status of
               top:
                  write_a_word('.top.   ');
               bottom:
                  write_a_word('.bottom.');
               normal:
                  begin
                     start_off := offset;
                     stop_pos := start_off + window_width;
                     if (stop_pos > line_length) and (not need_long_way)
                     then stop_pos := line_length;
                     last_col_full :=
                        need_long_way or
                        ((stop_pos - offset = window_width) and
                        (line_text[stop_pos] <> ' '));
                     need_scrolloff :=
                        (draw_row = window_length) and
                        last_col_full and
                        (window_right = cols_per_row) and
                        (window_bottom = rows_per_page);
                     if need_scrolloff then
                     begin
                        if scrolloff_avail
                        then vdu_control(scrolloff)  { suppress scrolling }
                        else if ch_ins_del_avail and window_at_right
                        then stop_pos := stop_pos - 2  { treat last 2 specially }
                        else if line_ins_del_avail and window_full_width and
                                lineerase_avail and window_at_right then
                        begin { draw using lineinsert }
                           { draw it by first drawing it one line above, }
                           { moving it down and then redrawing the destroyed }
                           { line }
                           draw_row := draw_row - 1;
                           window_position(draw_row, 1);
                           first_pos_done := true;
                           vdu_control(lineerase)
                        end { draw using lineinsert }
                        else
                        begin { just ignore last position }
                           stop_pos := stop_pos - 1;
                           { attempt to put it back later }
                           character_omitted := true
                        end { just ignore last position }
                     end;
                     spaces := 0;
                     write_col := 1;
                     for lindex := (start_off + 1) to stop_pos do
                     begin
                        if lindex <= line_length  { may be off end (no lineerase) }
                        then line_ch := line_text[lindex]
                        else line_ch := ' ';
                        if lindex = test_line_pos then
                        begin
                           set_half(offset = 0);  { set opposite intensity }
                           write_a_ch('^');  { write the character marker }
                           set_half(offset > 0)  { switch back }
                        end
                        else write_a_ch(line_ch)
                     end;
                     if need_scrolloff then
                     begin
                        if scrolloff_avail
                        then vdu_control(scrollon)
                        else if ch_ins_del_avail and window_at_right then
                        begin { fill using charinsert }
                           stop_pos := stop_pos + 2;  { restore value }
                           if stop_pos <= line_length
                           then write_a_ch(line_text[stop_pos])  { last char }
                           else write_a_ch(' ');
                           vdu_control(cursorleft);  { move it to last place }
                           vdu_control(charinsert);
                           if (stop_pos-1) <= line_length
                           then write_a_ch(line_text[stop_pos-1])  { next to last }
                           else write_a_ch(' ')
                        end { fill using charinsert }
                        else if line_ins_del_avail and window_full_width and
                                lineerase_avail and window_at_right then
                        begin { move redrawn row down and draw again }
                           window_position(draw_row, 1);
                           do_lineinsert(draw_row, 1);
                           draw_window_line(draw_row, false, false);
                           draw_row := draw_row + 1
                        end { move redrawn row down and draw again }
                        else            { do nothing }
                     end;
                     if last_col_full
                     then foil_wraparound
                  end { normal line }
            end {case}
         end { really draw the line }
         else write_a_word('        ');  { just blank out }
         if (not always_output) and     { test whether user has typed ahead }
            echoing_on and              { and possibly corrupted the screen }
            (not term_is_ssmp) and
            (tty$in < 0)
         then
         begin { user has typed ahead }
            if not term_is_ssmp then
            begin
               tty$rs([bit1], code);    { clear any waiting output }
               if code <> 0
               then error_print(k$irtn, code, 'Clearing output', 'TTY$RS')
            end;
            print_a_char(chr(nul_code));  { send NUL in case last char actually }
                                        { output was an ESC }
            screen_tabs_set := false;   { may not have finished setting them }
            { force message to bottom of screen }
            message_lines_to_clear := 1000;
            screen_message('', false);
            screen_message('', false);
            screen_message('Processing character(s) typed ahead.', false);
            screen_message('', false);
            screen_message('', false);
            if term_is_ssmp             { ensure frame is empty }
            then push_frame;
            suppress_output := true     { set it now because after it is set }
                                        { all controls are ignored }
         end { user has typed ahead } ;
         if replace_after               { if requested and line actually }
         then replace_cursor;           { visible then replace cursor }
         set_half(false)                { leave screen in full intensity }
      end { only draw visible lines } ;
      if (file_line = marked_line) and (marked_line <> 0) and (offset = 0)
      then mark_the_line(draw_row)      { add the arrow marker }
   end { output not suppressed, with, actually output the line }
end { draw_window_line } ;

   procedure clear_message; forward;

{ Draw a windowful of lines from the file; the window may be already clear, }
{ which simplifies output }

procedure draw_window(already_clear: boolean);                                  { DRAW_WINDOW }

var
   rindex: row_range;

begin { draw_window }
   if screen_is_split
   then clear_message                   { may affect other windows }
   else message_lines_to_clear := 0;    { message has gone }
   continue_screen_messages := true;
   character_omitted := false;
   if suppress_output
   then                                 { type-ahead still being processed }
   else with current_window do
   begin { output not suppressed }
      set_screen_tabs;
      vdu_control(foldoff);             { can get left on because of quits }
      half_state := false;
      if not screen_is_split then
      begin { screen not split }
         vdu_control(clearscreen);      { no need to position to (1,1) }
         already_clear := true
      end { screen not split } ;
      for rindex := 1 to window_length
      do draw_window_line(rindex, false, already_clear)
   end { output not suppressed } ;
   replace_cursor                       { note - may change file line }
end { draw_window } ;

{ Draw part of the current window }

procedure draw_part_window(row1, row2: row_range; already_clear: boolean);      { DRAW_PART_WINDOW }

var
   rindex: row_range;
   lindex: integer;

begin { draw_part_window }
   with current_buffer, current_window do
   begin
      { do not labour redrawing nearly all screen if no lineerase }
      if (lineerase_avail and window_at_right) or
         (row2-row1 <= 11) or
         already_clear or
         screen_is_split
      then
      begin
         for rindex := row1 to row2
         do draw_window_line(rindex, false, already_clear)
      end { redraw without clearing screen first }
      else draw_window(false)
   end { with }
end { draw_part_window } ;

{ Draw the whole screen by clearing and then drawing each window }

procedure draw_whole_screen;                                                    { DRAW_WHOLE_SCREEN }

var
   saved_window: window0_range;
   windex: window_range;

begin { draw_whole_screen }
   message_lines_to_clear := 0;         { avoid double redraws }
   continue_screen_messages := true;
   if screen_is_split then
   begin { screen is split }
      vdu_control(clearscreen);
      draw_window_frame(1, vdu_table^.rows_per_page, false);
      saved_window := current_window_number;
      for windex := 1 to window_count do
      begin
         switch_window(windex);
         draw_window(true)
      end;
      switch_window(saved_window);
      replace_cursor
   end { screen is split }
   else draw_window(false)
end { draw_whole_screen } ;

{ Draw some lines of the screen by clearing them and then calling }
{ draw_part_window for any affected windows }

procedure draw_part_screen(row1, row2: row_range);                              { DRAW_PART_SCREEN }

var
   rindex, first_affected, last_affected: row_range;
   windex: window_range;
   saved_window, wtop, wbottom: integer;

begin { draw_part_screen }
   if not screen_is_split
   then draw_part_window(row1, row2, false)  { not split screen }
   else
   begin { split screen }
      for rindex := row1 to row2 do
      begin
         vdu_position(rindex, 1);
         vdu_control(lineerase)
      end;
      draw_window_frame(row1, row2, not lineerase_avail);
      saved_window := current_window_number;
      for windex := 1 to window_count do
      begin
         if windex = current_window_number then
         with current_window do
         begin { current window }
            wtop := window_top;
            wbottom := window_bottom
         end { current window }
         else with window_array[windex] do
         begin { different window }
            wtop := window_top;
            wbottom := window_bottom
         end { different window } ;
         { calculate screen rows affected that lie within this window }
         first_affected := row1;
         if first_affected < wtop
         then first_affected := wtop;
         last_affected := row2;
         if last_affected > wbottom
         then last_affected := wbottom;
         if last_affected >= first_affected then
         with current_window do
         begin
            switch_window(windex);
            draw_part_window(first_affected - window_top + 1,
                              last_affected - window_top + 1, true);
            replace_cursor              { sets correct file line }
         end { need to redraw }
      end { for loop } ;
      switch_window(saved_window)
   end { split screen }
end { draw_part_screen } ;

{ Reformat and redraw the whole window, using the arguments supplied }

procedure form_draw_args(arg1, arg2: integer; already_clear: boolean);          { FORM_DRAW_ARGS }

begin { form_draw_args }
   need_to_redraw := false;             { line has been reformatted }
   full_screen := true;                 { screen is now fully drawn }
   strip_trailing_spaces;
   setup_window_status(arg1, arg2);
   draw_window(already_clear)
end { form_draw_args } ;

{ Reformat and redraw the whole window.  Use default arguments to centre the }
{ cursor in the window. }

procedure form_draw_window(already_clear: boolean);                             { FORM_DRAW_WINDOW }

begin { form_draw_window }
   with current_window
   do form_draw_args( -(window_length div 2), (window_length - 1) div 2 ,
                      already_clear)
end { form_draw_window } ;

{ Draw the whole of the current line on the screen }

procedure draw_file_line(repcur: boolean);                                      { DRAW_FILE_LINE }

var
   rindex: row_range;

begin { draw_file_line }
   with current_buffer, current_window do
   begin
      for rindex := 1 to window_length do
      with window_status[rindex] do
      begin
         if file_line = line_number then
         begin
            row_drawn := true;
            draw_window_line(rindex, false, false)
         end
      end;
      if repcur
      then replace_cursor
   end { with }
end { draw_file_line } ;

{ Redraw the given buffer if it is on screen in a window; positions to .top. }
{ before redrawing because position may have been lost by clear_message }

procedure redraw_buffer_if_visible(the_buffer: buffer_range;                    { REDRAW_BUFFER_IF_VISIBLE }
                                   go_to_top: boolean);

var
   the_window, saved_window_number: window0_range;
   windex: window_range;

begin { redraw_buffer_if_visible }
   if screen_is_split then
   begin
      the_window := 0;
      for windex := 1 to window_count do
      begin
         if windex = current_window_number
         then if current_window.window_buffer = the_buffer
              then the_window := windex
              else
         else if window_array[windex].window_buffer = the_buffer
         then the_window := windex
      end { for loop } ;
      if the_window > 0 then
      begin { buffer is on screen }
         saved_window_number := current_window_number;
         switch_window(the_window);
         if go_to_top
         then load_a_line(0, false);
         form_draw_window(false);
         switch_window(saved_window_number);
         replace_cursor
      end { buffer is on screen }
   end { split screen }
   else { not possible if screen not split }
end { redraw_buffer_if_visible } ;

{ Unmark the marked line (if any) and update the screen to correspond }

procedure unmark;                                                               { UNMARK }

var
   row_first: row_range;

begin { unmark }
   with current_buffer, current_window do
   if marked_line <> 0 then
   begin
      move_to_line(marked_line);
      marked_line := 0;
      if find_first(line_number, row_first) then
      with window_status[row_first] do
      begin { part at least of line is on screen }
         if offset = 0 then
         begin { first row of marked line visible }
            { cannot just remove the marker because the user may have done }
            { some character inserts etc. }
            row_drawn := true;
            draw_window_line(row_first, false, false)
         end { first row of marked line visible }
      end { part at least of line is on screen } ;
      replace_cursor
   end
end { unmark } ;

{ Append spaces to the current line until it covers the cursor position }

procedure expand_line;                                                          { EXPAND_LINE }

var
   lindex: line_chars_range;
   new_length: integer;

begin { expand_line }
   with current_buffer, current_window do
   begin
      if line_position <= line_length
      then                              { already long enough }
      else if line_position >= line_max_chars
      then abort('expand_line: past max line length')
      else
      begin { expand to an integral number of window lines }
         new_length := ((line_position-1) div window_width + 1) * window_width;
         if new_length >= line_max_chars
         then new_length := line_max_charsm1;
         for lindex := (line_length + 1) to new_length do
            line_text[lindex] := ' ';
         line_text[new_length + 1] := newline;
         line_length := new_length
      end { expand to an integral number of window lines }
   end { with }
end { expand_line } ;

{ Update the part of the screen below the cursor to allow for some screen rows }
{ being deleted. }

procedure update_for_delete(old_below, new_below: integer;                      { UPDATE_FOR_DELETE }
                            delete_row: row_range;
                            needop: boolean;
                            lineinc: integer);

var
   counter, stop_row: integer;
   saved_file_place: file_places;
   row_index: row_range;
   start_row, rows_deleted: integer;
   ln, off: integer;

begin { update_for_delete }
   with current_buffer, current_window do
   begin
      rows_deleted := old_below - new_below;
      counter := rows_deleted;
      if not needop
      then counter := counter - 1;      { one already done }
      if line_ins_del_avail and window_full_width then
      begin { delete the screen rows }
         do_linedelete(delete_row, counter)
      end { delete the screen rows } ;
      { move status lines and adjust file_line entry }
      for row_index := (delete_row + new_below) to
                       (window_length - rows_deleted) do
      begin
         window_status[row_index] := window_status[row_index + rows_deleted];
         with window_status[row_index] do
            file_line := file_line + lineinc
      end;
      { set window_status for file line }
      stop_row := delete_row + new_below - 1;
      if stop_row > window_length
      then stop_row := window_length;
      with window_status[delete_row] do  { [delete_row] already set by caller }
      begin
         ln := file_line;
         off := offset
      end;
      for row_index := delete_row to stop_row do
      with window_status[row_index] do
      begin
         file_line := line_number;
         offset := off;
         off := off + window_width;
         row_status := row_normal;
         row_drawn := true
      end;
      { fill in lines at bottom of screen and redraw }
      save_place(saved_file_place);
      row_index := window_length - rows_deleted;
      move_down_screen(row_index);
      if line_ins_del_avail and window_full_width
      then start_row := window_length - rows_deleted + 1
      else start_row := delete_row;
      draw_part_window(start_row, window_length, false);
      restore_place(saved_file_place)
   end
end { update_for_delete } ;

{ Update the part of the screen below the cursor to allow for some screen rows }
{ being inserted. }

procedure update_for_insert(old_below, new_below: integer;                      { UPDATE_FOR_INSERT }
                            insert_row: row_range;
                            needop: boolean;
                            lineinc: integer);

var
   counter, stop_row, rows_inserted: integer;
   saved_file_place: file_places;
   row_index: row_range;
   ln, off: integer;

begin { update_for_insert }
   with current_buffer, current_window, vdu_table^ do
   begin
      rows_inserted := new_below - old_below;
      counter := rows_inserted;
      if not needop
      then counter := counter - 1;      { one already done }
      { update window_status to match }
      for row_index := window_length downto (insert_row + new_below) do
      begin
         window_status[row_index] := window_status[row_index - rows_inserted];
         with window_status[row_index]
         do file_line := file_line + lineinc
      end;
      { set window_status for rows inserted }
      stop_row := insert_row + new_below - 1;
      if stop_row > window_length
      then stop_row := window_length;
      with window_status[insert_row] do  { [insert_row] already set by caller }
      begin
         ln := file_line;
         off := offset
      end;
      for row_index := insert_row to stop_row do
      with window_status[row_index] do
      begin
         file_line := line_number;
         offset := off;
         off := off + window_width;
         row_status := row_normal;
         row_drawn := true
      end;
      if line_ins_del_avail and window_full_width then
      begin { use line ins/del }
         do_lineinsert(insert_row, counter)
      end { use line ins/del }
      else
      begin { need to redraw }
         save_place(saved_file_place);
         draw_part_window(insert_row, window_length, false);
         restore_place(saved_file_place)
      end { need to redraw }
   end { with }
end { update_for_insert } ;

{ Adjust the screen status below the cursor to allow for screen rows inserted
  or deleted }

procedure adjust_below(old_below, new_below: integer;                           { ADJUST_BELOW }
                       the_row: row_range;
                       needop: boolean;
                       lineinc: integer);

begin { adjust_below }
   if new_below >= old_below
   then update_for_insert(old_below, new_below, the_row, needop, lineinc)
   else update_for_delete(old_below, new_below, the_row, needop, lineinc)
end { adjust_below } ;

{ Set the cursor position to correspond to the given file line and position.
  There are one or two rare circumstances in which the appropriate spot for
  the cursor is no longer on the screen; we could scroll the screen, but since
  that proves very complicated (we may already be scrolling the screen!) we
  just put it back where it was. }

procedure set_cursor_pos(lineno: integer;                                       { SET_CURSOR_POS }
                         position: line_chars_range);

var
   row_first, row_last: row_range;

begin { set_cursor_pos }
   with current_buffer, current_window do
   begin
      if find_first(lineno, row_first) then
      begin { part or whole line on screen }
         move_to_line(lineno);
         find_last(row_last);
         if position > (window_status[row_last].offset + window_width)
         then                           { appropriate fold not visible }
         else if position <= (window_status[row_first].offset)
         then                           { ditto off top }
         else
         begin { at last - correct place is visible }
            cursor_row := row_last;
            while position <= window_status[cursor_row].offset
            do cursor_row := cursor_row - 1;
            cursor_col := position - window_status[cursor_row].offset
         end { at last - correct place is visible }
      end { part or whole line on screen } ;
      line_position := window_status[cursor_row].offset + cursor_col
   end { with }
end { set_cursor_pos } ;

{ Reformat if necessary and redraw the current file line on the screen. If }
{ force_draw is false the line is only redrawn if reformatting is necessary. }

procedure form_draw_file_line(force_draw, repcur: boolean);                     { FORM_DRAW_FILE_LINE }

var
   row_first, row_last, row_index, rindex: row_range;
   saved_line_position: line_chars_range;
   line_to_find, rows_for_line, correct_last, old_below, new_below,
   last_offset: integer;
   cursor_line: integer;
   reformat: boolean;

begin { form_draw_file_line }
   if current_buffer.line_status = top
   then line_to_find := 0
   else line_to_find := current_buffer.line_number;
   with current_buffer, current_window do
   if find_first(line_to_find, row_first) then
   begin { part or all on screen }
      with window_status[cursor_row] do
      begin
         cursor_line := file_line;      { current file line }
         if (cursor_col + offset) <= line_max_charsm1  { and position }
         then line_position := cursor_col + offset
         else line_position := line_max_charsm1  { at end of max line }
      end;
      saved_line_position := line_position;
      strip_trailing_spaces;
      if line_number = cursor_line      { cursor is still on this line }
      then expand_line;                 { ensure cursor position will still }
                                        { be visible }
      find_last(row_last);
      rows_for_line := (line_length + window_width - 1) div window_width;
      if rows_for_line = 0
      then rows_for_line := 1;
      if window_status[row_first].offset > 0 then
      begin { line is partly off top of screen }
         last_offset := (rows_for_line - 1) * window_width;
         { Allow for line extending off bottom of screen }
         with window_status[window_length]
         do if (line_number = file_line) and (last_offset > offset)
            then last_offset := offset;
         if force_draw or
            need_to_redraw or
            (window_status[row_last].offset <> last_offset)  { reformat needed }
         then
         begin
            row_index := row_last + 1;
            { Allow for line off bottom of screen }
            if row_index > window_length
            then row_index := window_length;
            move_up_screen(row_index, false);
            for rindex := 1 to (row_index - 1) do
            begin
               with window_status[rindex] do
               begin
                  file_line := 0;
                  offset := 0;
                  row_status := row_null;
                  row_drawn := false
               end
            end;
            draw_part_window(row_index, row_last, false);
            set_cursor_pos(cursor_line, saved_line_position);
            if repcur
            then replace_cursor
         end { redraw line }
      end { partly off top of screen }
      else
      begin { first part of line is on screen }
         correct_last := row_first + rows_for_line - 1;
         if correct_last > window_length
         then correct_last := window_length;
         reformat := row_last <> correct_last;
         if reformat then
         begin
            old_below := row_last - row_first + 1;
            new_below := correct_last - row_first + 1;
            window_position(row_first, 1);
            adjust_below(old_below, new_below, row_first, true, 0)
         end;
         if force_draw or reformat then
         begin
            draw_file_line(false);
            set_cursor_pos(cursor_line, saved_line_position);  { replace cursor }
            if repcur
            then replace_cursor         { at correct "file" position }
         end
      end { first part of line is on screen }
   end { part or all on screen } ;
   need_to_redraw := false
end { form_draw_file_line } ;

{ Insert a new blank line into the file below the current line }

procedure insert_blank_line;                                                    { INSERT_BLANK_LINE }

begin { insert_blank_line }
   with current_buffer do
   begin
      replace_line;
      line_length := 0;
      line_text[1] := newline;
      insert_a_line
   end
end { insert_blank_line } ;

{ Clear the message area without reinstating file lines.  Use when calls to }
{ such as replace_cursor would destroy file positions.                      }

procedure clear_message_area;                                                   { CLEAR_MESSAGE_AREA }

var
   lindex: integer;

begin { clear_message_area }
   if message_lines_to_clear > 0 then
   with vdu_table^ do
   begin
      if message_lines_to_clear > rows_per_page
      then message_lines_to_clear := rows_per_page;
      for lindex := 1 to message_lines_to_clear do
      begin
         vdu_position(lindex, 1);
         vdu_control(lineerase)
      end { for loop } ;
      if term_is_ssmp then
      begin
         issue_pending(nopend);
         push_frame
      end { SSMP terminal }
   end { with }
end { clear_message_area } ;

{ Clear the message (if any) on the top few lines of the screen }

procedure clear_message;                                                        { CLEAR_MESSAGE }

begin { clear_message }
   if message_lines_to_clear > 0 then
   with vdu_table^ do
   begin
      if message_lines_to_clear > rows_per_page
      then message_lines_to_clear := rows_per_page;
      draw_part_screen(1, message_lines_to_clear);
      replace_cursor;
      message_lines_to_clear := 0;
      continue_screen_messages := true;
      if term_is_ssmp then
      begin { SSMP terminal }
         issue_pending(nopend);
         push_frame
      end { SSMP terminal }
   end { with }
end { clear_message } ;

{ Alter screen status to allow for screen scrolling one line up }

procedure scrolled_one_line;                                                    { SCROLLED_ONE_LINE }

var
   row_index: row_range;

begin { scrolled_one_line }
   with current_buffer, current_window do
   begin
      for row_index := 2 to window_length
      do window_status[row_index-1] := window_status[row_index];
      with window_status[window_length] do
      begin
         row_drawn := false;
         if line_length > (offset + window_width)
         then offset := offset + window_width  { already long enough for }
                                        { another fold }
         else
         begin
            file_line := file_line + 1;  { go on to next file line }
            offset := 0
         end
      end { for loop }
   end { with }
end { scrolled_one_line } ;

{ Take a crudely updated cursor position and allow for screen wrap-around etc. }
{ 'arrow_type' is set to the control value corresponding to a cursor arrow if  }
{ appropriate, or else any other control value (e.g. backtab)                  }

procedure cursor_moved(new_row: boolean;                                        { CURSOR_MOVED }
                       arrow_type: controls);

var
   cursor_ok, fold, different_row, do_complete_redraw: boolean;
   screen_has_scrolled, half_setting: boolean;
   row_index: row_range;
   index, newcol: integer;
   saved_line_position: line_chars_range;
   char_to_output, missing_char: char;

begin { cursor_moved }
   different_row := new_row;
   do_complete_redraw := false;
   fold := false;
   screen_has_scrolled := false;
   with current_buffer, current_window, vdu_table^ do
   begin

      { first ensure that column number is within bounds }

      if cursor_col < 1 then
      begin { off screen to left }
         if term_is_ssmp                { should have caused token return }
         then wait_for_token(false);
         if not left_pos_ok then
         begin { move cursor to expected place }
            window_position(cursor_row, window_width);
            do_cursor_move(cursorup)
         end { move cursor to expected place } ;
         cursor_col := window_width;    { wrap around to right }
         cursor_row := cursor_row - 1;
         different_row := true
      end { cursor_col < 1 }
      else if cursor_col > window_width then
      begin { off screen to right }
         if term_is_ssmp
         then wait_for_token(false);
         if screen_is_split
         then cursor_ok := false
         else if arrow_type = cursorright
         then cursor_ok := right_pos_ok  { was a cursor movement arrow }
         else cursor_ok := right_prt_ok;  { was a printing character }
         if cursor_ok then
         begin
            { Some terminals make the cursor "pause" at or after column 80 }
            { when the user has just filled column 80 - we output another  }
            { character and then backspace so that we know where we are }
            if cursor_row >= window_length
            then char_to_output := ' '
            else
            begin { not off bottom of screen }
               with window_status[cursor_row+1] do
               begin
                  set_half(offset > 0);
                  move_to_line(file_line);
                  if offset + 1 <= line_length
                  then char_to_output := line_text[offset+1]
                  else char_to_output := ' '
               end { with } ;
               move_to_line(window_status[cursor_row].file_line)
            end { not off bottom of screen } ;
            put_char_on_screen(char_to_output);
            set_half(false);
            foil_wraparound;
            if cursor_row < window_length
            then window_position(cursor_row+1, 1)
            else window_position(cursor_row, 1)  { allow for scroll }
         end { cursor_ok }
         else if cursor_row < window_length
         then window_position(cursor_row+1, 1)
         else if screen_is_split
         then { cursor down code will scroll window }
         else
         begin { use LF to scroll }
            window_position(cursor_row, 1);
            print_newline
         end { use LF to scroll } ;
         cursor_col := 1;               { wrap around to left }
         cursor_row := cursor_row + 1;
         different_row := true;
         fold := true                   { => folded long line }
      end { cursor_col > window_width } ;

      { Now check that the row is within bounds }

      if cursor_row < 1 then
      begin { off top of screen }
         if term_is_ssmp
         then wait_for_token(false);
         if ((arrow_type = cursorup) and not up_pos_ok) or
            screen_is_split then
         begin { scroll a line upwards }
            if reverself_avail and (not screen_is_split) then
            begin { use reverse linefeed if poss }
               window_position(1, 1);
               vdu_control(reverself)
            end { use reverse linefeed if poss }
            else if line_ins_del_avail and window_full_width then
            begin { use line ins/del }
               window_position(1, 1);
               do_lineinsert(1, 1);
               window_position(1, cursor_col)
            end { use line ins/del }
            else do_complete_redraw := true
         end { scroll a line upwards } ;
         cursor_row := 1;
         for row_index := window_length downto 2 do
            window_status[row_index] := window_status[row_index - 1];
         if window_status[2].offset > 0 then
         begin { earlier part of folded line }
            { [1] is already a copy of [2] }
            with window_status[1] do
            begin
               offset := offset - window_width;
               row_drawn := false       { ensure drawn later }
            end
         end { earlier part of folded line }
         else with window_status[1] do
         begin
            file_line := file_line - 1;
            { move temporarily to the new file line just to find how long it }
            { is, but we must move back again in case of redraw later }
            move_to_line(file_line);
            offset := ((line_length-1) div window_width) * window_width;
            move_to_line(window_status[2].file_line);
            row_status := row_normal;
            row_drawn := false          { ensure drawn later }
         end;
         replace_cursor;                { back to correct column }
         { any message has moved down one line }
         if message_lines_to_clear > 0
         then message_lines_to_clear := message_lines_to_clear + 1
      end { cursor_row < 1 }
      else if cursor_row > window_length then
      begin { off bottom of screen }
         if term_is_ssmp
         then wait_for_token(false);
         if ((arrow_type = cursordown) and not down_pos_ok) or
            screen_is_split then
         begin { scroll a line downwards }
            if not screen_is_split then
            begin
               { reposition cursor in case vdu placed it somewhere else }
               window_position(window_length, cursor_col);
               print_newline
            end
            else if line_ins_del_avail and window_full_width then
            begin { use line ins/del }
               window_position(1, 1);
               do_linedelete(1, 1);
               window_position(window_length, cursor_col)
            end { use line ins/del }
            else do_complete_redraw := true
         end { scroll a line downwards } ;
         cursor_row := window_length;
         scrolled_one_line;
         if message_lines_to_clear > 0  { one message line has scrolled off }
         then message_lines_to_clear := message_lines_to_clear - 1;
         screen_has_scrolled := true
      end { cursor_row > window_length } ;

      { if row has changed deal with inserting, redrawing etc. }

      if different_row then
      with window_status[cursor_row] do
      begin
         if fold and (line_status = normal) then
         begin
            if offset = 0 then
            begin { add extra part to folded line }
               for row_index := window_length downto cursor_row
               do window_status[row_index] := window_status[row_index-1];
               offset := offset + window_width;
               if row_drawn then
               begin
                  if line_ins_del_avail and window_full_width then
                  begin
                     do_lineinsert(cursor_row, 1);
                     window_position(cursor_row, cursor_col)
                  end
                  else
                  begin
                     draw_part_window(cursor_row, window_length, false);
                     replace_cursor
                  end
               end;
               row_drawn := true;
               if need_to_redraw then
               begin
                  draw_file_line(true);
                  need_to_redraw := false
               end
            end { add extra part to folded line }
            else                        { line already folded }
            begin
               form_draw_file_line(need_to_redraw, true);
               with window_status[cursor_row] do
               if not row_drawn then
               begin
                  row_drawn := true;
                  draw_window_line(cursor_row, true, true)
               end
            end
         end { fold }
         else
         begin { on to ordinary line }
            if (not window_mode ) and
             ( (row_status = row_null) or  { hit a null line }
               (file_line = iw_marked_line))  { or the marked line }
            then
            begin { IW - need to insert line into file }
               insert_blank_line;
               file_line := line_number;
               offset := 0;
               row_status := row_normal;
               move_to_line(line_number - 1)  { in case of redraw below }
            end { IW - need to insert line into file } ;

            { must redraw after new line inserted for IW or window_status is }
            { not up-to-date and we get things twisted }

            if token = withterm then
            begin { SSMP - cannot act now }
               if need_to_redraw and (not token_requested)
               then hreqtoken           { Get token and redraw later }
            end { SSMP - cannot act now }
            else form_draw_file_line(need_to_redraw, true);

            { new "with" because form_draw_file_line may change "cursor_row" }
            with window_status[cursor_row] do
            begin
               move_to_line(file_line);
               if line_status = bottom then
               begin
                  if cursor_row > 1 then
                  begin { look at previous row }
                     move_to_line(window_status[cursor_row-1].file_line);
                     case line_status of
                        top, normal:    { first occurrence of .bottom. }
                           begin
                              move_to_line(window_status[cursor_row].file_line);
                              if not row_drawn then
                              begin
                                 row_drawn := true;
                                 draw_window_line(cursor_row, true, false)
                              end
                           end;
                        bottom:         { below .bottom. line on screen }
                           begin
                              cursor_row := cursor_row - 1;
                              replace_cursor;
                              macro_dropout := true  { stop any repeats }
                           end
                     end {case}
                  end { look at previous row }
                  else                  { avoid moving off screen! }
               end { .bottom. }
               else
               begin { not .bottom. }
                  if not row_drawn then
                  begin
                     row_drawn := true;
                     if not do_complete_redraw
                     then draw_window_line(cursor_row, true, true)
                  end;
                  if line_status = top then  { do not leave cursor on .top. }
                  begin
                     cursor_row := cursor_row + 1;  { do not perform "cursordown" because }
                     replace_cursor;    { it may be at bottom of screen by now }
                     macro_dropout := true  { stop any repeats }
                  end
               end { not .bottom. }
            end { with }
         end { not fold } ;
         if token = withhost
         then clear_message;            { clear any displayed message when }
                                        { cursor changes rows }
         if (line_status <> normal) then
         begin
            cursor_col := 1;
            if token = withhost
            then window_position(cursor_row, cursor_col)
         end
      end { if different_row } ;

      { set file line position }

      with window_status[cursor_row] do
      begin
         if (cursor_col + offset) <= line_max_charsm1
         then line_position := cursor_col + offset
         else
         begin { off end of editor max line length }
            if token = withhost then
            begin
               newcol := line_max_charsm1 - offset;
               for index := (cursor_col + offset) downto line_max_chars
               do vdu_control(cursorleft);  { backspace to valid position }
               if newcol > 0
               then cursor_col := newcol  { on same screen row }
               else
               begin
                  cursor_row := cursor_row - 1;
                  cursor_col := newcol + window_width;
                  cursor_moved(true, cursorleft)
               end
            end { token = withhost } ;
            line_position := line_max_charsm1;  { corrected line_position }
            need_to_redraw := true;
            macro_dropout := true       { stop any repeats }
         end { off end of editor max line length }
      end { with } ;
      if do_complete_redraw then
      begin { dumb terminal action }
         saved_line_position := line_position;
         form_draw_window(false);
         set_cursor_pos(line_number, saved_line_position);
         replace_cursor
      end { do_complete_redraw } ;
      if character_omitted and screen_has_scrolled then
      begin { dumb terminal was unable to write }
                                        { last row/col of screen }
         with window_status[window_length-1] do
         begin
            move_to_line(file_line);
            half_setting := offset > 0;
            if offset + window_width <= line_length
            then missing_char := line_text[offset + window_width]
            else missing_char := ' '
         end;
         if missing_char = ' '
         then                           { false alarm }
         else
         begin
            window_position(window_length-1, window_width);
            set_half(half_setting);
            put_char_on_screen(missing_char);
            set_half(false);
            foil_wraparound
         end;
         replace_cursor;
         with window_status[window_length]
         do if offset + window_width > line_length
            then character_omitted := false
            else character_omitted := line_text[offset+window_width] <> ' '
      end { character_omitted and screen_has_scrolled }
   end { with }
end { cursor_moved } ;

{ Move the cursor to the screen row containing the first part of the current
  file line.  It may be necessary to roll the screen to reach it. }

procedure cursor_to_line_start;                                                 { CURSOR_TO_LINE_START }

begin { cursor_to_line_start }
   with current_window do
   begin
      while window_status[cursor_row].offset > 0 do
      begin
         do_cursor_move(cursorup);
         cursor_row := cursor_row - 1;
         cursor_moved(true, cursorup)
      end { while loop }
   end { with }
end { cursor_to_line_start } ;

{ Move the cursor to the screen row containing the last part of the current
  file line.  It may be necessary to scroll the screen to reach it. }

procedure cursor_to_line_end;                                                   { CURSOR_TO_LINE_END }

var
   test_length: integer;

begin { cursor_to_line_end }
   with current_buffer, current_window do
   begin
      { first move to left margin to avoid passing through a zone off end of }
      { maximum length line }
      cursor_col := 1;
      window_position(cursor_row, cursor_col);
      cursor_moved(false, backtab);
      strip_trailing_spaces;
      if line_length = 0
      then test_length := 1             { treat 0 as special case for testing }
      else test_length := line_length;
      { move back to within line_length }
      while window_status[cursor_row].offset >= test_length do
      begin
         do_cursor_move(cursorup);
         cursor_row := cursor_row - 1;
         cursor_moved(true, cursorup)
      end;
      { move forward to last fold }
      form_draw_file_line(need_to_redraw, true);  { redraw now if necessary }
      while (window_status[cursor_row].offset + window_width) < test_length do
      begin
         do_cursor_move(cursordown);
         cursor_row := cursor_row + 1;
         cursor_moved(true, cursordown)
      end { while loop }
   end { with }
end { cursor_to_line_end } ;

{ Adjust cursor to be outside "margin" rows from screen top or bottom }
{ if possible }

procedure adjust_outside_margins;                                               { ADJUST_OUTSIDE_MARGINS }

var
   saved_col: col0p1_range;
   row_count: integer;
   done: boolean;

begin { adjust_outside_margins }
   if message_lines_to_clear > 0
   then { do not adjust if message on screen to avoid clearing it }
   else if screen_is_split
   then { not worth the effort }
   else with current_buffer, current_window do
   begin
      if cursor_row > window_length - screen_margin then
      begin { near bottom of screen }
         saved_col := cursor_col;
         row_count := screen_margin;
         done := false;
         repeat
            if (line_status <> normal) or
               (row_count <= 0) or
               ((not window_mode) and (line_number >= iw_marked_line - 1))
               { avoids generating extra blank lines }
            then done := true
            else
            begin
               row_count := row_count - 1;
               do_cursor_move(cursordown);
               cursor_row := cursor_row + 1;
               cursor_moved(true, cursordown)
            end
         until done;
         for row_count := 1 to screen_margin - row_count do
         begin
            do_cursor_move(cursorup);
            cursor_row := cursor_row - 1;
            cursor_moved(true, cursorup)
         end;
         if cursor_col <> saved_col then
         begin { column changed because of hitting .bottom. }
            cursor_col := saved_col;
            replace_cursor
         end { column changed because of hitting .bottom. }
      end { near bottom of screen }
      else if cursor_row <= screen_margin then
      begin { near top of screen }
         row_count := screen_margin;
         done := false;
         repeat
            if (line_number <= 1) or
               (row_count <= 0)
            then done := true
            else
            begin
               row_count := row_count - 1;
               do_cursor_move(cursorup);
               cursor_row := cursor_row - 1;
               cursor_moved(true, cursorup)
            end
         until done;
         for row_count := 1 to screen_margin - row_count do
         begin
            do_cursor_move(cursordown);
            cursor_row := cursor_row + 1;
            cursor_moved(true, cursordown)
         end
      end { near top of screen }
      else                              { within central portion - leave }
   end { do adjust, with }
end { adjust_outside_margins } ;

{ Test whether the file line that we have moved to still corresponds to a line }
{ of the window - redraw the window if not, or if "must_redraw" is true.       }

procedure position_cursor_in_window(must_redraw: boolean;                       { POSITION_CURSOR_IN_WINDOW }
                                    target_position: line_chars_range;
                                    do_margin: boolean);

var
   redraw: boolean;
   lposition: line_chars_range;
   overlap, new_line_number, chindex, rows_for_line, visible_length,
   cursor_file_line: integer;
   direction: (down, up, stay);

begin { position_cursor_in_window }
   with current_buffer, current_window do
   begin
      if line_status = top
      then load_a_line(1, false);       { move from .top. to line 1 }
      if not window_mode and            { IW - do not allow to go below mark }
         (line_number >= iw_marked_line)
      then move_to_line(iw_marked_line-1);

      { move to new line by scrolling screen if within "overlap" lines of }
      { being on the screen }

      if screen_is_split then
      begin { split screen - do not scroll }
         redraw := must_redraw;
         with window_status[1] do
         begin
            if line_number < file_line
            then redraw := true
            else if (line_number = file_line) and
                    (target_position <= offset)
            then redraw := true
         end { with } ;
         with window_status[window_length] do
         begin
            if line_number > file_line
            then redraw := true
            else if (line_number = file_line) and
                    (target_position > offset + window_width)
            then redraw := true
         end { with }
      end { split screen }
      else
      begin { not split }
         overlap := window_length div 2;  { half a screen }
         redraw := must_redraw or
            (line_number < (window_status[1].file_line - overlap)) or
            (line_number > (window_status[window_length].file_line + overlap))
      end { not split } ;
      if redraw then
      begin
         line_position := target_position;
         form_draw_window(false)
      end;
      strip_trailing_spaces;
      rows_for_line := (line_length + window_width - 1) div window_width;
      if rows_for_line = 0
      then rows_for_line := 1;
      visible_length := window_width * rows_for_line;  { columns visible }
      { the cursor may have been off the end of the non-blank portion of }
      { the line }
      with window_status[cursor_row] do
      begin
         cursor_file_line := file_line;
         if (cursor_file_line = line_number) and  { still on same line }
            (visible_length < offset + window_width)  { need to expand? }
         then visible_length := offset + window_width
      end;
      if visible_length > line_max_charsm1
      then visible_length := line_max_charsm1;
      if target_position > visible_length  { adjust line position within bounds }
      then lposition := visible_length
      else lposition := target_position;
      cursor_col := 1;                  { move vertically at left margin to }
                                        { avoid moving through forbidden zones }
                                        { after long lines }
      new_line_number := line_number;
      if need_to_redraw and (new_line_number = cursor_file_line)
      then form_draw_file_line(true, false);
      if redraw
      then { cursor already in place }
      else replace_cursor;

      { move onto correct file line }

      if new_line_number > line_number
      then direction := down
      else if new_line_number < line_number
      then direction := up
      else if lposition > window_status[cursor_row].offset + window_width
      then direction := down
      else if lposition <= window_status[cursor_row].offset
      then direction := up
      else direction := stay;           { already on correct row }
      if direction = down then
      begin { move down screen }
         while (new_line_number > line_number) or
               (lposition > window_status[cursor_row].offset + window_width)
         do
         begin
            do_cursor_move(cursordown);
            cursor_row := cursor_row + 1;
            cursor_moved(true, cursordown)
         end { while loop }
      end { direction = down }
      else if direction = up then
      begin { move up screen }
         while (new_line_number < line_number) or
               (lposition <= window_status[cursor_row].offset)
         do
         begin
            do_cursor_move(cursorup);
            cursor_row := cursor_row - 1;
            cursor_moved(true, cursorup)
         end { while loop }
      end { direction = up } ;
      if do_margin
      then adjust_outside_margins;

      { move to correct column }

      cursor_col := lposition mod window_width;
      if cursor_col = 0
      then cursor_col := window_width;
      if cursor_col <> 1
      then replace_cursor;

      { move to correct cursor position if after line_length }

      for chindex := line_position - 1 downto target_position do
      begin
         do_cursor_move(cursorleft);
         cursor_col := cursor_col - 1;
         cursor_moved(false, cursorleft)
      end { for loop } ;
      for chindex := line_position + 1 to target_position do
      begin
         do_cursor_move(cursorright);
         cursor_col := cursor_col + 1;
         cursor_moved(false, cursorright)
      end { for loop }
   end { with }
end { position_cursor_in_window } ;

{ Delete a sequence of characters in the current line from line_position up to }
{ but not including posn. }

procedure delete_some_chs(posn: line_chars_range);                              { DELETE_SOME_CHS }

var
   chars_gone: line_chars0_range;
   lindex: line_chars_range;
   redraw: boolean;

begin { delete_some_chs }
   with current_buffer, current_window do
   begin
      push_line;                        { save for OOPS command }
      chars_gone := posn - line_position;
      redraw := line_length > window_width;  { redraw whole line if folded }
      movewds(line_text, posn,
              line_text, line_position,
              line_length + 2 - posn);
      line_length := line_length - chars_gone;
      line_altered := true;
      if redraw or
         need_to_redraw or
         (chars_gone > 20) or
         (not ch_ins_del_avail) or
         (not window_at_right)
      then form_draw_file_line(true, true)  { redraw whole line }
      else
         for lindex := 1 to chars_gone  { output some char deletes }
         do vdu_control(chardelete)
   end
end { delete_some_chs } ;

{ Ensure that the cursor is not on row one of the screen, to avoid a message }
{ getting in the way }

procedure ensure_not_row_one;                                                   { ENSURE_NOT_ROW_ONE }

var
   saved_macro_dropout: boolean;

begin { ensure_not_row_one }
   with current_buffer, current_window do
   begin
      if (cursor_row + window_top - 1) = 1 then
      begin { scroll screen to move cursor off top row }
         saved_macro_dropout := macro_dropout;
         do_cursor_move(cursorup);
         cursor_row := cursor_row - 1;
         cursor_moved(true, cursorup);
         if cursor_row = 1 then
         begin { normal case - screen scrolled }
            do_cursor_move(cursordown);
            cursor_row := cursor_row + 1;
            cursor_moved(true, cursordown)
         end { normal case - screen scrolled }
         else                           { bounced back off .top. }
            macro_dropout := saved_macro_dropout  { set in cursor_moved }
      end { scroll screen to move cursor off top row }
   end { with }
end { ensure_not_row_one } ;

{ Print an ordinary message on next line of screen }

procedure display_next_message(message: long_string);                           { DISPLAY_NEXT_MESSAGE }

begin { display_next_message }
   screen_message(message, true);
   replace_cursor
end { display_next_message } ;

{ Print an ordinary message on top line of screen }

procedure display_message(message: long_string);                                { DISPLAY_MESSAGE }

begin { display_message }
   clear_message;
   ensure_not_row_one;
   display_next_message(message)
end { display_message } ;

{ Print an error message }

procedure display_error(message: long_string);                                  { DISPLAY_ERROR }

begin { display_error }
   display_message(message);
   macro_dropout := true
end { display_error } ;

{ Print an error message }

procedure display_next_error(message: long_string);                             { DISPLAY_NEXT_ERROR }

begin { display_next_error }
   display_next_message(message);
   macro_dropout := true
end { display_next_error } ;

{ Check that the line we are about to alter is not top or bottom of file }

function check_normal: boolean;                                                 { CHECK_NORMAL }

begin { check_normal }
   with current_buffer, current_window do
   begin
      if line_status = normal
      then check_normal := true
      else
      begin { do not let user change .bottom. }
         check_normal := false;
         macro_dropout := true;
         if token = withterm then
         begin { SSMP - cannot act now }
            if not token_requested
            then hreqtoken;             { Get token and redraw later }
            suppress_output := true
         end { token = withterm }
         else
         begin { may act now }
            draw_window_line(cursor_row, false, false);
            display_message('Not allowed at bottom of file.')
         end { may act now }
      end { not already normal line }
   end { with }
end { check_normal } ;

{ Test for end of repetition (of macro or command) }

function test_for_end(var loop_spec: looprec): boolean;                         { TEST_FOR_END }

var
   ended: boolean;

begin { test_for_end }
   with current_buffer, loop_spec do
   begin
      if main_state = command_error then  { can happen if e.g. error on paging }
      begin
         main_state := running;
         macro_dropout := true
      end;
      ended := macro_dropout;           { always end if any error has occurred }
      if not ended then
      case end_condition of             { test for end of repetition }
         once_only:
            ended := true;
         ntimes:
            begin
               loop_counter := loop_counter - 1;
               ended := loop_counter <= 0
            end;
         stop_after_mark:
            begin
               ended := line_number >= marked_line;
               if ended then unmark
            end;
         stop_before_mark:
            begin
               ended := line_number <= marked_line;
               if ended then unmark
            end;
         stop_eq_char:
            begin
               expand_line;
               ended := line_text[line_position] = loop_test_char
            end;
         stop_ne_char:
            begin
               expand_line;
               ended := line_text[line_position] <> loop_test_char
            end
      end;
      if ended then
      begin
         if (not mode_quiet) and (current_macro_desc = nil)
         then display_message('End of repetition.')
      end;
      test_for_end := ended
   end
end { test_for_end } ;

{ Get the next character from the macro buffer }

function get_macro_code(var the_code: integer): boolean;                        { GET_MACRO_CODE }

begin { get_macro_code }
   if current_macro_desc = nil
   then get_macro_code := false         { no macro executing }
   else with current_macro_desc^, current_buf^ do
   begin { macro is being executed }
      if (not macro_dropout) and (codes_used < code_count) then
      begin { get char from current buffer }
         codes_used := codes_used + 1;
         the_code := macro_text[codes_used];
         get_macro_code := true
      end { get char from current buffer }
      else if next <> nil then
      begin { try next buffer }
         current_buf := next;
         codes_used := 0;
         get_macro_code := get_macro_code(the_code)
      end { try next buffer }
      else
      begin { end of macro - test for repeats }
         if test_for_end(loop_spec) then
         begin { no repeats - continue calling macro }
            macro_being_obeyed := false;  { (if any) }
            current_macro_desc := previous;
            get_macro_code := get_macro_code(the_code)
         end { no repeats - continue calling macro }
         else
         begin { repeat - back to beginning }
            current_buf := code_start;
            codes_used := 0;
            get_macro_code := get_macro_code(the_code)
         end { repeat - back to beginning }
      end { end of macro - test for repeats }
   end { macro is being executed }
end { get_macro_code } ;

{ Get an input action code from SSMP input }

function get_ssmp_code(input_code_level: integer): integer;                     { GET_SSMP_CODE }

var
   the_code: integer;
   rindex: row_range;

begin { get_ssmp_code }
   repeat
      get_stacked_code(the_code);
      if the_code = 0 then
      begin { no code available }
         if token = withterm
         then dispatch                  { get next input code }
         else
         begin { ask terminal for more }
            if need_to_redraw
            then form_draw_file_line(need_to_redraw, true);
            if suppress_output then
            begin { refresh whole screen }
               suppress_output := false;
               form_draw_window(false);
               fill_in_undrawn_rows := false
            end { refresh whole screen }
            else if fill_in_undrawn_rows then
            with current_window do
            begin
               for rindex := 1 to window_length do
               with window_status[rindex] do
               begin
                  if not row_drawn then
                  begin { draw the row }
                     row_drawn := true;
                     draw_window_line(rindex, false, true)
                  end { draw the row }
               end { for loop } ;
               replace_cursor;
               fill_in_undrawn_rows := false
            end { fill in } ;
            if input_code_level > 1 then
            begin { multi-character code not fitting in with SSMP standards }
               { ensure not in character insertion mode because SSMP terminal }
               { would try to insert the single character }
               if ssmp_mode[icharmode] <> 0
               then hsetmode(icharmode, 0);
               hsetupdate(saved_row, saved_row, saved_col, saved_col);
               getting_single_char := true
            end { multi-character code }
            else if not special_limits_set then
            with current_window do
            begin { determine correct update limits }
               if window_mode then
               begin { WI mode }
                  first_update_row := window_length;
                  last_update_row := 1;
                  for rindex := 1 to window_length do
                  begin
                     { line numbers below .bottom. were originally set to }
                     { large_int but may have been reduced by line delete }
                     if window_status[rindex].row_drawn and
                       (window_status[rindex].file_line < fairly_large_int)
                     then
                     begin
                        if rindex < first_update_row
                        then first_update_row := rindex;
                        if rindex > last_update_row
                        then last_update_row := rindex
                     end
                  end { for loop }
               end { WI mode }
               else
               begin { IW mode }
                  first_update_row := 1;
                  last_update_row := window_length
               end { IW mode } ;
               { exclude .top. if visible }
               if (first_update_row = 1) and
                  window_status[1].row_drawn and
                  (window_status[1].file_line = 0)
               then first_update_row := 2;
               { include .bottom. only if cursor is on it }
               if cursor_row > last_update_row
               then last_update_row := cursor_row;
               { set these limits, converting window-relative row, col to }
               { SSMP absolute co-ords (start at zero) }
               hsetupdate(first_update_row + window_top - 2,
                          last_update_row  + window_top - 2,
                          window_left - 1,
                          window_right - 1);
               getting_single_char := false
            end { determine correct update limits }
            else if getting_single_char then
            begin
               { special limits were set }
               hsetupdate(special_top,  special_bottom,
                          special_left, special_right);
               getting_single_char := false
            end;
            htoken(0)                   { over to user }
         end { token = withhost }
      end { no code available }
   until the_code <> 0;
   get_ssmp_code := the_code;
   { If we were getting a single char (e.g. DEL X) then the token should }
   { have been returned }
   if getting_single_char
   then wait_for_token(false);
   { If we now have the token the item received violated box limits etc. }
   { so we will need to output }
   if token = withhost then
   begin
      need_to_output := true;
      { restore insert mode if changed (e.g. when getting DEL code) }
      if ssmp_mode[icharmode] <> ord(mode_insert)
      then hsetmode(icharmode, ord(mode_insert) )
   end { we have token }
end { get_ssmp_code } ;

{ Get the next code either from the terminal or from the macro buffer. }
{ Macro text is stored in a chain of macro buffers. }

function get_next_code(input_code_level: integer): integer;                     { GET_NEXT_CODE }

var
   the_code: integer;
   the_char: char;

begin { get_next_code }
   if get_macro_code(the_code) then
   begin { got one from current macro }
      if reading_action
      then                              { may deliver any code }
      else the_code := the_code mod 255  { normally deliver 0..255 }
   end { got one from current macro }
   else
   begin { not got from macro }
      if suppress_output then
      begin { output was suppressed }
         if (not term_is_ssmp) and (tty$in >= 0) then
         begin { type-ahead now cleared }
            suppress_output := false;
            form_draw_window(false)
         end { type-ahead now cleared }
      end { output was suppressed } ;
      macro_dropout := false;           { back to reading from keyboard }
      waiting_for_input := true;        { used in the QUIT$ on-unit }
      if term_is_ssmp
      then the_code := get_ssmp_code(input_code_level)
      else
      begin { not ssmp }
         c1in(the_char);
         the_code := ord(the_char)
      end { not ssmp } ;
      waiting_for_input := false;
      if macro_head = nil
      then { not storing a macro }
      else if reading_action
      then { action code may be stored later }
      else store_macro_code(the_code)   { store code in macro being recorded }
   end { not got from macro } ;
   get_next_code := the_code
end { get_next_code } ;

{ Get next input character, from terminal or macro }

function get_next_char(input_code_level: integer): char;                        { GET_NEXT_CHAR }

begin { get_next_char }
   get_next_char := chr(get_next_code(input_code_level))
end { get_next_char } ;

{ Read a long string, processing erase and kill characters }

procedure get_long_string(message: long_string;                                 { GET_LONG_STRING }
                          var the_string: long_string;
                          var cr_given: boolean;
                          upcaseit,
                          allow_spaces,
                          copy_quotes,
                          do_specials: boolean);

var
   ch: char;
   done: boolean;
   first_space: line_chars0_range;
   tindex: line_chars_range;
   temp_string: record
      case boolean of
         false: (len: shortint; chars: packed array[line_chars_range] of char);
         true:  (strng: string[line_max_chars])
      end;

begin { get_long_string }
   with temp_string do
   begin
      if current_macro_desc = nil then
      begin { set up the screen }
         ensure_not_row_one;
         if term_is_ssmp then
         with stacked_ssmp_codes do
         begin
            get_ssmp_reply(message, true, stacked_tree, tree_cr_given);
            tree_report := true
         end { SSMP }
         else
         begin
            no_lf_screen_message(message, true);
            if always_output
            then ctrl_half_duplex(false)  { echoing on }
         end { not SSMP }
      end { set up the screen } ;
      len := 0;
      done := false;                    { read and process erase and kill }
      repeat
         ch := get_next_char(1);
         if ch = newline then
         begin { newline read }
            done := true;               { finish at end of line }
            cr_given := true
         end { newline read }
         else if ch = chr(can_code) then
         begin { CAN  read }
            done := true;               { CAN cancels operation }
            cr_given := false
         end { CAN  read }
         else if do_specials and (ch = special[skill])
         then len := 0                  { reset to start }
         else if do_specials and (ch = special[serase])
         then if len > 0                { erase one character }
              then len := len - 1
              else
         else if not copy_quotes and (ch = '''')
         then                           { ignore quotes }
         else if len < line_max_chars then
         begin
            len := len + 1;
            if upcaseit and (ch in ['a'..'z'])
            then ch := chr(ord('A') - ord('a') + ord(ch));  { upcase it }
            chars[len] := ch
         end
         else                           { no space for character }
      until done;
      if allow_spaces
      then                              { no need to check }
      else
      begin { check no spaces in the_string }
         first_space := 0;
         for tindex := len downto 1
         do if chars[tindex] = ' '
            then first_space := tindex;
         if first_space <> 0
         then display_next_error('No spaces allowed in string.')
      end { check no spaces in the_string } ;
      the_string := strng;
      if always_output and (current_macro_desc = nil)
      then ctrl_half_duplex(true)       { echoing off }
   end { with } ;
   if (current_macro_desc = nil) and vdu_table^.mess_type1
   then vdu_control(messoff)
end { get_long_string } ;

{ Read a tree name, processing erase and kill characters }

procedure get_treename(message: long_string;                                    { GET_TREENAME }
                       var the_tree: tree_type;
                       var cr_given: boolean;
                       upcaseit, allow_spaces, copy_quotes,
                       do_abbrevs: boolean);

var
   long_temp: long_string;

begin { get_treename }
   get_long_string(message, long_temp, cr_given,
                   upcaseit, allow_spaces, copy_quotes, true);
   with the_tree do
   begin
      if length(long_temp) > tree_max_chars
      then strng := substr(long_temp, 1, tree_max_chars)
      else strng := long_temp;
      if do_abbrevs
      then expand_abbrevs(the_tree)
   end { with }
end { get_treename } ;

{ Prompt in message area and read a treename for load/unload }

function read_load_treename: boolean;                                           { READ_LOAD_TREENAME }

var
   cr_given: boolean;

begin { read_load_treename }
   with load_file do
   begin
      file_type := sam_type;
      clear_message;
      get_treename('Give filename or buffer number: ',
                   treename, cr_given, true, false, false, true);
      if macro_dropout
      then read_load_treename := false  { invalid treename }
      else if cr_given and (treename.len > 0)
      then read_load_treename := true
      else
      begin { CAN or no name given - abandon }
         if current_macro_desc = nil
         then display_message('Function abandoned.');
         read_load_treename := false
      end { cancelled }
   end { with }
end { read_load_treename } ;

{ Repair the screen after reading the command }

procedure repair_screen(repair: repair_type; return_received: boolean);         { REPAIR_SCREEN }

var
   chars, count, lpos: integer;
   saved_line_position: line_chars_range;
   test_row: row_range;
   row_changed, lf_prob: boolean;

begin { repair_screen }
   case repair of
      ok:    chars := 0;
      rest1: chars := 1;
      rest2: chars := 2;
      rest3: chars := 3;
      rest4: chars := 4;
      func:  abort('repair type "func" in repair_screen')
   end;
   if chars > 0 then
   with current_buffer, current_window, vdu_table^ do
   begin
      { avoid a repair that exactly fills the screen line because of }
      { terminals that "pause" at last column }
      if cursor_col + chars - 1 = window_width
      then chars := chars + 1;
      if return_received
      then test_row := window_length - 1
      else test_row := window_length;
      if (cursor_row >= test_row) and
         ((cursor_col + chars) >= window_width) and
         (not term_is_ssmp)
      then suppress_output := true      { take easy way out for awkward cases }
      else
      begin
         if return_received and (cursor_row = window_length) then
         begin { allow for screen scrolling because of function key }
            scrolled_one_line;
            cursor_row := cursor_row - 1
         end { allow for screen scrolling because of function key } ;
         window_position(cursor_row, cursor_col);
         if line_status <> normal
         then draw_window_line(cursor_row, true, false)  { draw .top. etc in full }
         else if (line_number = marked_line) and (line_position <= 6)
         then mark_the_line(cursor_row)  { replace the arrow marker }
         else
         begin
            saved_line_position := line_position;
            lf_prob := false;
            for count := 1 to chars do
            with window_status[cursor_row] do
            begin
               move_to_line(file_line);
               lpos := cursor_col + offset;
               if (lpos = point_line_pos) and
                  (file_line = point_line_no)
               then
               begin { write the character marker }
                  set_half(offset = 0);  { set opposite intensity }
                  print_a_char('^')
               end { write the character marker }
               else
               begin { write the correct character }
                  set_half(offset > 0);
                  if lpos > line_length
                  then put_char_on_screen(' ')
                  else put_char_on_screen(line_text[lpos])
               end { write the correct character } ;
               if cursor_col < window_width
               then cursor_col := cursor_col + 1
               else
               begin
                  cursor_col := 1;
                  cursor_row := cursor_row + 1;
                  { assumes printable ch moves to next row after right margin }
                  lf_prob := true
               end
            end { move forwards } ;
            set_half(false);
            row_changed := false;
            for count := 1 to chars do
            begin
               vdu_control(cursorleft);
               if cursor_col > 1
               then cursor_col := cursor_col - 1
               else
               begin
                  cursor_col := window_width;
                  cursor_row := cursor_row - 1;
                  row_changed := true
               end
            end { move backwards } ;
            if row_changed and (not left_pos_ok or not window_full_width)
            then window_position(cursor_row, cursor_col);
            move_to_line(window_status[cursor_row].file_line);
            line_position := saved_line_position;
            if lf_prob then
            begin { went over column 80 - next newline may be ignored }
               foil_wraparound;
               replace_cursor
            end { went over col 80 }
         end { replace individual chars } ;
         if term_is_ssmp then
         begin
            issue_pending(nopend);
            push_frame
         end { show the user }
      end { replace chars }
   end { chars > 0 }
end { repair_screen } ;

begin { pmed_wprocs }
end { pmed_wprocs } .
