program termtest (input, output);
 
 
{ ******************************************************************* }
{ *                                                                 * }
{ *                         T E R M T E S T                         * }
{ *                                                                 * }
{ ******************************************************************* }
{ *                                                                 * }
{ *       A program to exercise a remote terminal emulation         * }
{ *      operating via the Simple Screen Management Protocol.       * }
{ *                                                                 * }
{ *                                                                 * }
{ *                           Alan Hunter                           * }
{ *    Newcastle University Computing Laboratory -- November 1985   * }
{ *                                                                 * }
{ ******************************************************************* }
 
 
const
 
{ ******************************************************************* }
{ *                                                                 * }
{ *            International Standard 646 character set             * }
{ *                                                                 * }
{ ******************************************************************* }
 
    blank = 32;
 
    { numerics }
    dig0 = 48;   dig1 = 49;   dig2 = 50;   dig3 = 51;   dig4 = 52;
    dig5 = 53;   dig6 = 54;   dig7 = 55;   dig8 = 56;   dig9 = 57;
 
    { lower case alphabetics }
    leta = 97;   letb = 98;   letc = 99;   letd = 100;  lete = 101;
    letf = 102;  letg = 103;  leth = 104;  leti = 105;  letj = 106;
    letk = 107;  letl = 108;  letm = 109;  letn = 110;  leto = 111;
    letp = 112;  letq = 113;  letr = 114;  lets = 115;  lett = 116;
    letu = 117;  letv = 118;  letw = 119;  letx = 120;  lety = 121;
    letz = 122;
 
    { upper case alphabetics }
    capa = 65;   capb = 66;   capc = 67;   capd = 68;   cape = 69;
    capf = 70;   capg = 71;   caph = 72;   capi = 73;   capj = 74;
    capk = 75;   capl = 76;   capm = 77;   capn = 78;   capo = 79;
    capp = 80;   capq = 81;   capr = 82;   caps = 83;   capt = 84;
    capu = 85;   capv = 86;   capw = 87;   capx = 88;   capy = 89;
    capz = 90;
 
    { other graphic symbols }
    exclam = 33;     dquote = 34;     sharp = 35;      dollar = 36;
    percent = 37;    amper = 38;      squote = 39;     lparen = 40;
    rparen = 41;     star = 42;       plus = 43;       comma = 44;
    minus = 45;      period = 46;     slash = 47;      colon = 58;
    semicol = 59;    less = 60;       equals = 61;     greater = 62;
    question = 63;   atsign = 64;     lbrack = 91;     backslash = 92;
    rbrack = 93;     cflex = 94;      uscore = 95;     grave = 96;
    lbrace = 123;    bar = 124;       rbrace = 125;    tilde = 126;
 
    { control characters }
    nul = 0;
    ctla = 1;    ctlb = 2;    ctlc = 3;    ctld = 4;    ctle = 5;
    ctlf = 6;    ctlg = 7;    ctlh = 8;    ctli = 9;    ctlj = 10;
    ctlk = 11;   ctll = 12;   ctlm = 13;   ctln = 14;   ctlo = 15;
    ctlp = 16;   ctlq = 17;   ctlr = 18;   ctls = 19;   ctlt = 20;
    ctlu = 21;   ctlv = 22;   ctlw = 23;   ctlx = 24;   ctly = 25;
    ctlz = 26;
    esc = 27;    fs = 28;     gs = 29;     rs = 30;     us = 31;
    del = 127;
 
{ ******************************************************************* }
{ *                                                                 * }
{ *                      Shared data structure                      * }
{ *                                                                 * }
{ ******************************************************************* }
 
    zmaxrow = 59;        { allows a maximum of 60 rows }
    zmaxcol = 131;       { allows a maximum of 132 columns }
    isospace  = blank;   { to fit with The Book }
    zfieldlimit = 239;   { allows a maximum of 240 fields }
    tlevel    = 0;     tmaxrow   = 1;     tmaxcol   = 2;
    dsinvalid = 3;     notify    = 4;     selectgr  = 5;
    reqshift  = 6;     cursor    = 7;     icharmode = 8;
    ilinerow  = 9;     intsignal = 10;    kinthost  = 11;
    ksuspend  = 12;    krestart  = 13;    kenter    = 14;
    kcsrup    = 15;    kcsrdown  = 16;    kcsrleft  = 17;
    kcsrright = 18;    knexttab  = 19;    kprevtab  = 20;
    kleftupd  = 21;    kfirstns  = 22;    kalastns  = 23;
    kinsmode  = 24;    keraright = 25;    kinsspac  = 26;
    kdelchar  = 27;    keraprev  = 28;    kinsline  = 29;
    kdelline  = 30;    keraline  = 31;    kappline  = 32;
    ksplline  = 33;    knextfld  = 34;    kprevfld  = 35;
    khomefld  = 36;    knewline  = 37;
 
{ ******************************************************************* }
{ *                                                                 * }
{ *                Network encoding and transmission                * }
{ *                                                                 * }
{ ******************************************************************* }
 
    defrend    = 0;   { selectgr : default rendition }
    bold       = 1;   { selectgr : bold, or increased intensity }
    faint      = 2;   { selectgr : faint, or reduced intensity }
    italicised = 3;   { selectgr : italicised, or slanted }
    underlined = 4;   { selectgr : underlined }
    slowblink  = 5;   { selectgr : slowly blinking (less than 150/min) }
    rapidblink = 6;   { selectgr : rapidly blinking (150/min or more) }
    negative   = 7;   { selectgr : negative image (reverse video) }
    concealed  = 8;   { selectgr : concealed (displayed as SPACEs) }
    crossedout = 9;   { selectgr : crossed out (marked for deletion) }
 
    maxmode = knewline;  { maximum allocated mode index }
    framesize = 60;      { characters per SSMP frame output }
    maxframenbr = 2;     { frames per packet }
 
{ ******************************************************************* }
{ *                                                                 * }
{ *                           Application                           * }
{ *                                                                 * }
{ ******************************************************************* }
 
    { ## none ## }
 
 
type 
 
{ ******************************************************************* }
{ *                                                                 * }
{ *                      Shared data structure                      * }
{ *                                                                 * }
{ ******************************************************************* }
 
    toktype = (withhost, withterm); 
    rowtype = 0..zmaxrow;               { in use, 0..maxrow }
    coltype = 0..zmaxcol;               { in use, 0..maxcol }
    isochar = 0..127;                   { iso 646 7-bit character set }
    savedfield =                        { each stored field definition }
        record
            fldtop, fldbottom : rowtype;    { stored update limit rows }
            fldleft, fldright : coltype     { stored update limit cols }
        end;
    fieldindex = 0..zfieldlimit;        { in use, 0..((maxrow+1)*4-1) }
    modeindex  = 0..63;                 { 37 is highest defined index }
    smallint   = 0..255;                { parameter range }
    modearray  = array [modeindex]
                        of smallint;    { terminal emulation status }
    tabstop    = (notab, tabset);       { tabulation stop }
 
{ ******************************************************************* }
{ *                                                                 * }
{ *                Network encoding and transmission                * }
{ *                                                                 * }
{ ******************************************************************* }
 
    table   = packed array [smallint] of smallint;
    textlit = packed array [1..24] of char;
    charstr = packed array [1..16] of isochar;
    charlen = 0..16;
    netbuff = packed array [1..256] of isochar;
    primpar = array [1..4] of smallint;
    rowtext = array [coltype] of isochar;
    ncftype = (nocf, csrpend, fldpend);
 
{ ******************************************************************* }
{ *                                                                 * }
{ *                           Application                           * }
{ *                                                                 * }
{ ******************************************************************* }
 
    astype = (
                entry,          { negotiation reply; first menu entry }
                menu,           { main menu selection }
                charset,        { graphic rendition selection }
                fhost,          { first part of host primitive test }
                thost,          { exit from host primitive test }
                ooc,            { exit from ooc message test }
                rogue,          { exit from rogue application test }
                soak,           { exit from soak test }
                dismode,        { let the user select an element }
                modmode,        { accept modified value }
                boxes);         { token return from edit screen }
 
 
var 
 
{ ******************************************************************* }
{ *                                                                 * }
{ *                      Shared data structure                      * }
{ *                                                                 * }
{ ******************************************************************* }
 
    token : toktype;                    { access control token }
    row   : rowtype;                    { cursor row }
    col   : coltype;                    { cursor column }
    image : array [rowtype,coltype]
                        of isochar;     { screen image }
    curfield, maxfield : fieldindex;
    field : array [fieldindex]
                        of savedfield;  { stored field definitions }
    boxtop, boxbottom : rowtype;        { update limits - rows }
    boxleft, boxright : coltype;        { update limits - columns }
    mode  : modearray;                  { terminal emulation status }
    tabs  : array [coltype] of tabstop; { tabulation stops array }
 
{ ******************************************************************* }
{ *                                                                 * }
{ *                Network encoding and transmission                * }
{ *                                                                 * }
{ ******************************************************************* }
 
    masterlevel : smallint;       { session level after negotiation }
    firstnr : boolean;            { true before first netread }
    ssmpstate : isochar;          { parser state }
    tparidx, tsum : integer;      { used in parameter decoding }
    gotdigit : boolean;           { used in parameter decoding }
    tprid : isochar;              { returned prim id (star=character) }
    tpars : primpar;              { returned primitive parameters }
    cfstate : ncftype;            { setcursor/setfield optimisation }
    littoiso, isotolit : table;   { native to/from iso conversions }
    filter : table;               { filter input and strip parity }
    encode, decode : table;       { iso to/from ssmp conversions }
    inbuff, outbuff : netbuff;    { input and output frame buffers }
    inlen,                        { characters in input buffer }
    inptr, outptr,                { buffer pointers }
    framenbr,                     { index of output frame in packet }
    outremdr : smallint;          { characters available in frame }
 
{ ******************************************************************* }
{ *                                                                 * }
{ *                           Application                           * }
{ *                                                                 * }
{ ******************************************************************* }
 
    screenmode : boolean;         { true if frames may be fetched }
    tokencode : smallint;         { set to 1 for resticted transfer }
    domore : boolean;             { true if menu change before H-TOKEN }
    newcode : smallint;           { new menu screen code }
    appstate : astype;            { application processing state }
    menumode,                     { mode array for menu-driven screens }
    mdismode,                     { mode array for "mode array" screen }
    mmodmode,                     { ... same after element selected }
    editmode : modearray;         { mode array for editing screens }
    lastmode : smallint;          { last selected mode array index }
    lastrend : smallint;          { last rendition for char display }
 
 
{ ******************************************************************* }
{ *                                                                 * }
{ *               T h e   S S M P   M o d u l e                     * }
{ *               =============================                     * }
{ *                                                                 * }
{ *    All SSMP service procedures are defined below.  If this      * }
{ *    section is split into a separate module, the procedures and  * }
{ *    functions listed in the following explanatory text will      * }
{ *    need to be exported.   Variables defined as "Shared Data     * }
{ *    Structure" or "Network Encoding and Transmission" are        * }
{ *    required by these SSMP routines only.                        * }
{ *                                                                 * }
{ ******************************************************************* }
 
 
{ ####################### General Information #########################
 
 The following procedures service requests to generate the equivalent
 host primitives.  With the exception of "reqtoken", they should only
 be called whilst the token is with the host application.
 
 Note that optimisation of H-SETCURSOR and H-SETFIELD is performed
 automatically.  Only essential instances will be generated.
 
    character -- generate H-CHARACTER to send text
    sendtoken -- generate H-TOKEN to transfer token
    setcursor -- generate H-SETCURSOR to set the cursor position
    setmode -- generate H-SETMODE to set a mode array element value
    erasetoright -- generate H-ERASETORIGHT to erase to end of row
    insertspace -- generate H-INSERTSPACE to insert SPACE characters
    deletechar -- generate H-DELETECHAR to delete characters
    soundalarm -- generate H-SOUNDALARM to sound the audible alarm
    scrollup -- generate H-SCROLLUP to scroll rows up
    scrolldown -- generate H-SCROLLDOWN to scroll rows down
    erasedisplay -- generate H-ERASEDISPLAY to erase screen image
    erasefields -- generate H-ERASEFIELDS to erase all field defns
    setfield -- generate H-SETFIELD to select new field definition
    setupdate -- generate H-SETUPDATE to set update limits for field
    erasetabs -- generate H-ERASETABS to erase all tabulation stops
    settab -- generate H-SETTAB to set one tabulation stop
    session -- generate H-SESSION to enter or exit session
    reqtoken -- generate H-REQTOKEN to request early return of token
 
 The following procedures allow interrogation of the shared data
 structure variables.  Use of functions and procedures for this purpose
 hides the data from the application so that all assignments are under
 the control of procedures in the SSMP module and can be vetted.
 
    maxrow -- returns current session maximum row value
    maxcol -- returns current session maximum column value
    fieldlimit -- returns current session maximum field index
    getloc -- returns current cursor position row and column
    getfield -- returns current field index
    getmode -- returns a copy of the mode array
    getrow -- returns the text of the requested row
 
 The next three procedures are to allow convenient replacement of
 shared data structure values.
 
    put9ode -- makes the real mode array conform to that supplied
    putrow -- replaces the text for the specified row
    literal -- convert literal string into 1 to 24 character calls
 
 Note that "putmode" will issue only those H-SETMODEs required to
 correct the mode array.  A convenient way of quickly switching
 terminal operating modes is to keep a suitable copy of the mode array
 for each set required and using "putmode" to establish that state from
 any other.
 
 The second procedure, "putrow", will optimise generation of the
 required host primitives.  It will use H-SETCURSOR to skip ten or more
 consecutive blanks where this is possible (that is, either default
 rendition set or mode[reqshift]=1).
 
 If the application requires to refresh the terminal copy of the shared
 data structure, the following procedure will do this.  Note that since
 the graphic rendition of each character is not held in the data
 structure, any highlighting will be lost and must be separately
 re-established if required.
 
    refresh -- refreshes the whole shared data structure
 
 The following procedures wrap up the session entry negotiation and the
 recommended generation of primitives at session close.
 
    startssmp -- negotiate screen management session with terminal
    stopssmp -- tidy up and close down screen management session
 
 The key to use of the whole SSMP module is the last primitive:
 
    getprimitive -- process input until a primitive has been decoded
 
 This function should be repeatedly called to fetch each primitive from
 the terminal.  Input frames will be fetched automatically as required.
 Each call returns exactly one terminal-generated primitive decoded
 into the primitive identifier and parameters.  The primitive may be
 processed as appropriate.  Only T-TOKEN must be processed, and only in
 response to this primitive may host primitives other than H-REQTOKEN
 be generated.
 
 When each primitive is presented by "getprimitive" the host copy of
 the shared data structure will already have been updated.  The parser
 calls the relative T-primitive semantic routine to perform this action
 before returning control.  Calls to "getloc", "getfield", "getmode"
 and "getrow" at this point will therefore return correct values.
 
 The semantic routine called on recognition of T-TOKEN checks
 mode[TLEVEL] and mode[DSINVALID] for values consistent with the
 current SSMP session.  If they are not consistent, the values are
 corrected and "refresh" is called to correct the terminal.  On being
 presented with T-TOKEN, the application is freed from the necessity of
 checking session integrity and may act on the request code as
 appropriate.
 
 The master session level will be taken from the first T-SETMODE
 setting it.  Note that it is the responsibility of the host
 application calling "getprimitive" to check the screen dimensions
 offered.  This should be done at the first T-TOKEN; if the values are
 unsuitable a suitable text message should be output and "stopssmp"
 called.
 
###################### End of General Information ################### }
 
 
    { initliterals -- set up tables to convert native to/from ISO 646 }
    procedure initliterals;
    var
        j : smallint;
        lca, uca : char;
    begin
        lca := 'a';  uca := 'A';  { to please some compilers! }
        if lca > uca then
        begin  { assume ISO 646 }
            for j := 0 to 127 do
            begin  { strip parity only }
                littoiso[j] := j;  littoiso[j+128] := j;
                isotolit[j] := j;  isotolit[j+128] := j
            end
        end
        else
        begin  { assume EBCDIC }
            for j := 0 to 255 do
                littoiso[j] := grave;      { check character }
            for j := 0 to 9 do
                littoiso[j+240] := j + dig0;    { dig0..dig9 }
            for j := 0 to 8 do
            begin
                littoiso[j+129] := j + leta;    { leta..leti }
                littoiso[j+193] := j + capa;    { capa..capi }
                littoiso[j+145] := j + letj;    { letj..letr }
                littoiso[j+209] := j + capj     { capj..capr }
            end;
            for j := 0 to 7 do
            begin
                littoiso[j+162] := j + lets;    { lets..letz }
                littoiso[j+226] := j + caps     { caps..capz }
            end;
            littoiso[64]  := blank;
            littoiso[90]  := exclam;     littoiso[127] := dquote;
            littoiso[123] := sharp;      littoiso[91]  := dollar;
            littoiso[108] := percent;    littoiso[80]  := amper;
            littoiso[125] := squote;     littoiso[77]  := lparen;
            littoiso[93]  := rparen;     littoiso[92]  := star;
            littoiso[78]  := plus;       littoiso[107] := comma;
            littoiso[96]  := minus;      littoiso[75]  := period;
            littoiso[97]  := slash;      littoiso[122] := colon;
            littoiso[94]  := semicol;    littoiso[76]  := less;
            littoiso[126] := equals;     littoiso[110] := greater;
            littoiso[111] := question;   littoiso[124] := atsign;
            littoiso[173] := lbrack;     littoiso[224] := backslash;
            littoiso[189] := rbrack;     littoiso[95]  := cflex;
            littoiso[109] := uscore;     littoiso[121] := grave;
            littoiso[192] := lbrace;     littoiso[79]  := bar;
            littoiso[208] := rbrace;     littoiso[161] := tilde;
 
            { reverse table }
            for j := 0 to 255 do isotolit[j] := 121;
            for j := 0 to 255 do isotolit[littoiso[j]] := j;
            isotolit[96] := 121;
        end
    end;  {initliterals}
 
 
    { initcoding -- initialise tables for SSMP primitive encoding }
    procedure initcoding;
    var
        j : smallint;
    begin
        { i/o system and session initialisation }
        masterlevel := 0;
        outremdr := 0;  outptr := 0;  framenbr := 1;  cfstate := nocf;
        ssmpstate := nul;
 
        { input filter table to pass valid characters and strip parity }
        for j := 0 to 127 do filter[j] := nul;
        filter[blank] := blank;
        for j := lparen to semicol do filter[j] := j;
        filter[equals] := equals;
        for j := capa to capz do
        begin
            filter[j] := j;  filter[j+32] := j+32
        end;
        for j := 0 to 127 do filter[j+128] := filter[j];
 
        { decoding table for incoming T-CHARACTER primitives }
        for j := 0 to 127 do
        begin
           decode[j] := 32;
           encode[j] := 32
        end;
 
        decode[48] := 0;
        for j := 0 to 25 do decode[65+j]  := j + 1;
        for j := 0 to 4  do decode[49+j]  := j + 27;
        for j := 0 to 6  do decode[97+j]  := j + 33;
        for j := 0 to 5  do decode[104+j] := j + 58;
        decode[110] := 64;
        for j := 0 to 4  do decode[111+j] := j + 91;
        decode[116] := 96;
        for j := 0 to 4  do decode[117+j] := j + 123;
 
        { encoding table for generating H-CHARACTER primitives }
        for j := 0 to 127 do encode[decode[j]] := j;
        encode[32] := 32;
    end;  {initcoding}
 
 
    { initshared -- initialise our copy of the shared data structure }
    procedure initshared; 
    var 
        fidx : fieldindex; 
        trow : rowtype;
        tcol : coltype; 
        midx : modeindex; 
    begin 
        token := withhost; 
        row := 0;  col := 0; 
 
        { stored field definitions } 
        for fidx := 0 to zfieldlimit do
        with field[fidx] do 
        begin 
            fldtop := 0;   fldbottom := 0; 
            fldleft := 0;  fldright := 0; 
        end; 
        curfield := 0;  maxfield := 0; 
 
        { update limits } 
        boxtop := 0;   boxbottom := 0; 
        boxleft := 0;  boxright := 0; 
 
        { mode array } 
        for midx := 0 to 63 do mode[midx] := 0;
 
        { tabulation stops array} 
        for tcol := 0 to zmaxcol do tabs[tcol] := notab;
 
        { not true ssmp, but assign all elements of the screen array
          to avoid upsetting Pascal run time checking routines }
        for trow := 0 to zmaxrow do
            for tcol := 0 to zmaxcol do image[trow,tcol] := grave
    end;  {initshared} 
 
 
    { The following five procedures (netopen, netclose, netread,
      netwrite, netpush) will almost certainly need to be rewritten
      to be efficient when this program is tailored for a specific
      operating system.  }
 
 
    { netopen -- prepare network link for SSMP frame input/output }
    procedure netopen;
    begin
        firstnr := true;
    end;  {netopen}
 
 
    { netclose -- release any resources acquired by netopen }
    procedure netclose;
    begin
    end;  {netclose}
 
 
    { netread -- get next frame from network }
    procedure netread;
    var
        j : integer;
        c : char;
    begin
        if firstnr then firstnr := false
                   else readln;
        j := 1;
        repeat
            read(c);  inbuff[j] := littoiso[ord(c)];  j := j + 1;
        until eoln;
        inbuff[j] := ctlm;  inbuff[j+1] := ctlj;
        inlen := j + 1;  inptr := 0
    end;  {netread}
 
 
    { netwrite -- output assembled frames to the network }
    procedure netwrite;
    var
        j : smallint;
        c : isochar;
    begin
        for j := 1 to outptr do
        begin
            c := outbuff[j];
            if c >= 32 then write(chr(isotolit[c]))
            else
            begin
                if c = 13 then writeln
                else
                if c = 11 then write(chr(11))
            end
        end
    end;  {netwrite}
 
 
    { netpush -- push last frame to network, if necessary }
    procedure netpush;
    begin
    end;  {netpush}
 
 
    { endframe -- end current output frame }
    procedure endframe;
    begin
        if outptr > 0 then
        begin
            if outbuff[outptr] = isospace then
            begin  { foil trailing space trim }
                outbuff[outptr+1] := dollar;
                outptr := outptr + 1
            end;
            outbuff[outptr+1] := ctlm;  outbuff[outptr+2] := ctlj;
            outptr := outptr + 2;
            if framenbr < maxframenbr
            then framenbr := framenbr + 1
            else
            begin
                netwrite;
                framenbr := 1;  outptr := 0
            end
        end;
        outremdr := 0
    end;  {endframe}
 
 
    { pushframe -- push frame(s) to network }
    procedure pushframe;
    begin
        endframe;
        if outptr > 0 then
        begin
            netwrite;
            framenbr := 1;  outptr := 0
        end;
        netpush
    end;  {pushframe}
 
 
    { assemble -- encode host SSMP primitive and add to output frame }
    procedure assemble (primid : isochar;
                        npars : smallint;
                        pars : primpar);
    var
        pstr, istr : charstr;
        plen, ilen, chindex : charlen;
        pindex, jindex, itemp : smallint;
    begin
        if primid = star then
        begin  { H-CHARACTER }
            itemp := encode[pars[1]];  plen := 1;
            if itemp = blank then  pstr[1] := pars[1]
            else
            begin  { escaped set }
                pstr[1] := colon;  pstr[2] := itemp;  plen := 2
            end
        end
        else
        begin  { not H-CHARACTER }
            pstr[1] := equals;  pstr[2] := primid;  plen := 2;
            if npars = 0 then plen := 3
            else
            for pindex := 1 to npars do
            begin
                itemp := pars[pindex];  ilen := 0;
                while itemp>0 do
                begin
                    ilen := ilen + 1;
                    istr[16-ilen] := (itemp mod 10) + dig0;
                    itemp := itemp div 10
                end;
                if ilen=0 then
                begin
                    ilen := 1;  istr[15] := dig0;
                end;
                for jindex := ilen downto 1 do
                begin
                    plen := plen + 1;  pstr[plen] := istr[16-jindex]
                end;
                plen := plen + 1;  pstr[plen] := comma;
            end;
            pstr[plen] := semicol;
        end;
 
        { add encoded primitive to output frame }
        for chindex := 1 to plen do
        begin
            if outremdr = 0 then
            begin
                endframe;
                outptr := outptr + 1;  outbuff[outptr] := ctlk;
                outremdr := framesize
            end;
            outptr := outptr + 1;  outbuff[outptr] := pstr[chindex];
            outremdr := outremdr - 1
        end
    end;  {assemble}
 
 
    { cftest -- issues pending H-SETCURSOR or H-SETFIELD as necessary }
    procedure cftest (newcfstate : ncftype);
    var
        ppar : primpar;
    begin
        if newcfstate <> cfstate then
        begin
            case cfstate of
 
            nocf:
                ;
 
            csrpend:
                begin
                    ppar[1] := row;  ppar[2] := col;
                    assemble(capb, 2, ppar);
                end;
 
            fldpend:
                begin
                    ppar[1] := curfield;  assemble(capl, 1, ppar);
                end;
 
            end;  {case}
            cfstate := newcfstate
        end
    end;  {cftest}
 
 
    { maxrow -- returns current session maximum row value }
    function maxrow : rowtype;
    begin
        maxrow := mode[tmaxrow]
    end;  {maxrow}
 
 
    { maxcol -- returns current session maximum column value }
    function maxcol : coltype;
    begin
        maxcol := mode[tmaxcol]
    end;  {maxcol}
 
 
    { fieldlimit -- returns current session maximum field index }
    function fieldlimit : fieldindex;
    begin
        fieldlimit := ((maxrow + 1) * 4) - 1
    end;  {fieldlimit}
 
 
    { eraserow -- erase part of one row to SPACE characters }
    procedure eraserow (rowa : rowtype; 
                        cola, 
                        colb : coltype); 
    var 
        ctemp : coltype; 
    begin 
        for ctemp := cola to colb 
            do image[rowa,ctemp] := isospace 
    end;  {eraserow} 
 
 
    { erasebox -- erase a rectangle of positions to SPACE characters }
    procedure erasebox (rowa, rowb : rowtype; 
                        cola, colb : coltype); 
    var 
        rtemp : rowtype; 
    begin 
        for rtemp := rowa to rowb 
            do eraserow(rtemp, cola, colb) 
    end;  {erasebox} 
 
 
    { rowcopy -- copy part of row to same column range of another row }
    procedure rowcopy (rowa, rowb : rowtype; 
                       cola, colb : coltype); 
    var 
        ctemp : coltype; 
    begin 
        for ctemp := cola to colb 
            do image[rowb,ctemp] := image[rowa,ctemp] 
    end;  {rowcopy} 
 
 
    { rightshift -- shift row characters to right within column range }
    procedure rightshift (rowa  : rowtype; 
                          cola, 
                          colb  : coltype; 
                          ncols : smallint); 
    var 
        ctemp : coltype; 
    begin 
        if (colb - cola) < ncols 
            then eraserow(rowa, cola, colb) 
        else 
        begin 
            for ctemp := colb - ncols downto cola 
                do image[rowa,ctemp+ncols] := image[rowa,ctemp]; 
            eraserow(rowa, cola, cola+ncols-1) 
        end 
    end;  {rightshift} 
 
 
    { leftshift -- shift row characters to left within column range }
    procedure leftshift (rowa  : rowtype; 
                         cola, 
                         colb  : coltype; 
                         ncols : smallint); 
    var 
        ctemp : coltype; 
    begin 
        if (colb - cola) < ncols 
            then eraserow(rowa, cola, colb) 
        else 
        begin 
            for ctemp := cola + ncols to colb 
                do image[rowa,ctemp-ncols] := image[rowa,ctemp]; 
            eraserow(rowa, colb-ncols+1, colb) 
        end 
    end;  {leftshift} 
 
 
    { character -- generate H-CHARACTER to send text }
    procedure character (ch : isochar); 
    var
        ppar : primpar;
    begin 
        cftest(nocf);   { flush any H-SETFIELD or H-SETCURSOR }
        ppar[1] := ch;  assemble(star, 1, ppar);
        image[row,col] := ch; 
        if col < maxcol then col := col + 1
    end;  {character} 
 
 
    { literal -- convert literal string into 1 to 24 character calls }
    procedure literal (txt : textlit);
    const
        eolit = '&';
    var
        copying : boolean;
        cidx, tidx : smallint;
    begin
        tidx := 1;  copying := txt[1] <> eolit;
 
        while copying do
        begin
            cidx := ord(txt[tidx]);  character(littoiso[cidx]);
            copying := tidx < 24;
            if copying then
            begin
                tidx := tidx + 1;  copying := txt[tidx] <> eolit
            end;
        end
    end;  {literal}
 
 
    { sendtoken -- generate H-TOKEN to transfer token }
    procedure sendtoken (reqcode : smallint);
    var
        ppar : primpar;
    begin
        cftest(nocf);   { flush any H-SETFIELD or H-SETCURSOR }
        ppar[1] := reqcode;  assemble(capa, 1, ppar);  pushframe;
        if mode[cursor] = 0 then 
        begin  { bound cursor } 
            if (row < boxtop) or (row > boxbottom) 
                or (col < boxleft) or (col > boxright) 
                    then 
                    begin 
                        row := boxtop;  col := boxleft 
                    end 
        end; 
        mode[tlevel] := 0;   { check this element on token return }
        token := withterm    { control with remote terminal emulation }
    end;  {sendtoken}
 
 
    { setcursor -- generate H-SETCURSOR to set the cursor position }
    procedure setcursor (newrow : rowtype; 
                          newcol : coltype); 
    var
        ppar : primpar;
    begin
        cftest(csrpend);   { flush any H-SETFIELD, defer H-SETCURSOR }
        row := newrow;  col := newcol 
    end;  {setcursor} 
 
 
    { setmode -- generate H-SETMODE to set a mode array element value }
    procedure setmode (index : modeindex; 
                        ivalue : smallint); 
    var
        ppar : primpar;
    begin
        ppar[1] := index;  ppar[2] := ivalue;  assemble(capc, 2, ppar);
        mode[index] := ivalue 
    end;  {setmode} 
 
 
    { erasetoright -- generate H-ERASETORIGHT to erase to end of row }
    procedure erasetoright; 
    var
        ppar : primpar;
    begin
        cftest(nocf);   { flush any H-SETFIELD or H-SETCURSOR }
        assemble(capd, 0, ppar);
        eraserow(row, col, maxcol) 
    end;  {erasetoright} 
 
 
    { insertspace -- generate H-INSERTSPACE to insert SPACE characters }
    procedure insertspace (nsp : smallint); 
    var
        ppar : primpar;
    begin
        cftest(nocf);   { flush any H-SETFIELD or H-SETCURSOR }
        ppar[1] := nsp;  assemble(cape, 1, ppar);
        if mode[reqshift] = 1 
            then rightshift(row, col, maxcol, nsp) 
    end;  {insertspace} 
 
 
    { deletechar -- generate H-DELETECHAR to delete characters }
    procedure deletechar (nch : smallint); 
    var
        ppar : primpar;
    begin
        cftest(nocf);   { flush any H-SETFIELD or H-SETCURSOR }
        ppar[1] := nch;  assemble(capf, 1, ppar);
        if mode[reqshift] = 1 
            then leftshift(row, col, maxcol, nch) 
    end;  {deletechar} 
 
 
    { soundalarm -- generate H-SOUNDALARM to sound the audible alarm }
    procedure soundalarm;
    var
        ppar : primpar;
    begin
        assemble(capg, 0, ppar)
    end;  {soundalarm}
 
 
    { scrollup -- generate H-SCROLLUP to scroll rows up }
    procedure scrollup (rowa, 
                         rowb  : rowtype; 
                         nrows : smallint); 
    var 
        rtemp : rowtype; 
        ppar : primpar;
    begin
        ppar[1] := rowa;  ppar[2] := rowb;  ppar[3] := nrows;
        assemble(caph, 3, ppar);
        if (rowb-rowa) < nrows 
            then erasebox(rowa, rowb, 0, maxcol) 
        else 
        begin 
            for rtemp := rowa+nrows to rowb 
                do rowcopy(rtemp, rtemp-nrows, 0, maxcol); 
            erasebox(rowb-nrows+1, rowb, 0, maxcol) 
        end 
    end;  {scrollup} 
 
 
    { scrolldown -- generate H-SCROLLDOWN to scroll rows down }
    procedure scrolldown (rowa, 
                           rowb  : rowtype; 
                           nrows : smallint); 
    var 
        rtemp : rowtype; 
        ppar : primpar;
    begin
        ppar[1] := rowa;  ppar[2] := rowb;  ppar[3] := nrows;
        assemble(capi, 3, ppar);
        if (rowb-rowa) < nrows 
            then erasebox(rowa, rowb, 0, maxcol) 
        else 
        begin 
            for rtemp := rowb-nrows downto rowa 
                do rowcopy(rtemp, rtemp+nrows, 0, maxcol); 
            erasebox(rowa, rowa+nrows-1, 0, maxcol) 
        end 
    end;  {scrolldown} 
 
 
    { erasedisplay -- generate H-ERASEDISPLAY to erase screen image }
    procedure erasedisplay; 
    var
        ppar : primpar;
    begin
        if cfstate = csrpend
            then cfstate := nocf    { implied cursor movement to [0,0] }
            else cftest(nocf);      { flush any H-SETFIELD }
        assemble(capj, 0, ppar);
        erasebox(0, maxrow, 0, maxcol); 
        row := 0;  col := 0 
    end;  {erasedisplay} 
 
 
    { erasefields -- generate H-ERASEFIELDS to erase all field defns }
    procedure erasefields; 
    var 
        ftemp : fieldindex; 
        ppar : primpar;
    begin
        if cfstate = fldpend
            then cfstate := nocf    { implicit selection of field zero }
            else cftest(nocf);      { flush any H-SETCURSOR }
        assemble(capk, 0, ppar);
        curfield := 0;  maxfield := 0; 
        for ftemp := 0 to fieldlimit do 
            with field[ftemp] do 
            begin 
                fldtop := 0;  fldbottom := 0; 
                fldleft := 0;  fldright := 0 
            end; 
        boxtop := 0;  boxbottom := 0; 
        boxleft := 0;  boxright := 0 
    end;  {erasefields} 
 
 
    { setfield -- generate H-SETFIELD to select new field definition }
    procedure setfield (fidx : fieldindex); 
    var
        ppar : primpar;
    begin
        cfstate := fldpend;   { suppress any H-SETCURSOR }
        if fidx > maxfield then maxfield := fidx; 
        curfield := fidx; 
        with field[curfield] do 
        begin 
            boxtop := fldtop;  boxbottom := fldbottom; 
            boxleft := fldleft;  boxright := fldright 
        end; 
        row := boxtop;  col := boxleft 
    end;  {setfield} 
 
 
    { setupdate -- generate H-SETUPDATE to set update limits for field }
    procedure setupdate (rowa, 
                          rowb : rowtype; 
                          cola, 
                          colb : coltype); 
    var
        ppar : primpar;
    begin
        cftest(nocf);   { flush any H-SETFIELD or H-SETCURSOR }
        ppar[1] := rowa;  ppar[2] := rowb;
        ppar[3] := cola;  ppar[4] := colb;  assemble(capm, 4, ppar);
        with field[curfield] do 
        begin 
            boxtop := rowa;     fldtop := boxtop; 
            boxbottom := rowb;  fldbottom := boxbottom; 
            boxleft  := cola;   fldleft := boxleft; 
            boxright := colb;   fldright := boxright 
        end 
    end;  {setupdate} 
 
 
    { erasetabs -- generate H-ERASETABS to erase all tabulation stops }
    procedure erasetabs; 
    var 
        ctemp : coltype; 
        ppar : primpar;
    begin
        assemble(capn, 0, ppar);
        for ctemp := 0 to maxcol 
            do tabs[ctemp] := notab 
    end;  {erasetabs} 
 
 
    { settab -- generate H-SETTAB to set one tabulation stop }
    procedure settab (tcol : coltype); 
    var
        ppar : primpar;
    begin
        ppar[1] := tcol;  assemble(capo, 1, ppar);
        tabs[tcol] := tabset 
    end;  {settab} 
 
 
    { session -- generate H-SESSION to enter or exit session }
    procedure session (reqcode : smallint);
    var
        ppar : primpar;
    begin
        cftest(nocf);   { flush any [H-SETFIELD or] H-SETCURSOR }
        ppar[1] := reqcode;  assemble(capp, 1, ppar);
        if reqcode = 0
            then initshared  { new session, init shared data structure }
            else pushframe   { current session ends, flush to network }
    end;  {session}
 
 
    { reqtoken -- generate H-REQTOKEN to request early return of token }
    procedure reqtoken; 
    var
        ppar : primpar;
    begin
        cftest(nocf);   { flush any H-SETFIELD or H-SETCURSOR }
        pushframe;  assemble(capq, 0, ppar);  pushframe
    end;  {reqtoken} 
 
 
    { getloc -- returns current cursor position row and column }
    procedure getloc (var currow : rowtype;
                      var curcol : coltype);
    begin
        currow := row;  curcol := col
    end;  {getloc}
 
 
    { getfield -- returns current field index }
    procedure getfield (var curfdx : fieldindex);
    begin
        curfdx := curfield
    end;  {getfield}
 
 
    { getmode -- returns a copy of the mode array }
    procedure getmode (var mcopy : modearray);
    begin
        mcopy := mode
    end;  {getmode}
 
 
    { putmode -- makes the real mode array conform to that supplied }
    procedure putmode (mcopy : modearray);
    var
        midx : modeindex;
    begin
        for midx := notify to knewline do
            if mcopy[midx] <> mode[midx] then setmode(midx, mcopy[midx])
    end;  {putmode}
 
 
    { getrow -- returns the text of the requested row }
    procedure getrow (r : rowtype;
                      var line : rowtext);
    var
        tcol, mcol : coltype;
    begin
        mcol := maxcol;
        for tcol := 0 to mcol do line[tcol] := image[r,tcol];
        if mcol < zmaxcol then
            for tcol := mcol+1 to zmaxcol do line[tcol] := blank
    end;  {getrow}
 
 
    { putrow -- replaces the text for the specified row }
    procedure putrow (r : rowtype;
                      var line : rowtext);
    var
        tcol, mcol : coltype;
        done : boolean;
        ch : isochar;
        bcount, bidx : smallint;
    begin
        done := false;  mcol := maxcol;  setcursor(r, 0);
 
        if mode[selectgr] <> defrend then
        begin
            if mode[reqshift] = 1
            then insertspace(mcol+1)   { clear to selected rendition }
            else
            begin  { must output each character }
                for tcol := 0 to mcol do character(line[tcol]);
                done := true
            end
        end
        else erasetoright;   { clear to default rendition spaces }
 
        { if the text has not yet been output, use the following
          algorithm which will skip large areas of white space }
        tcol := 0;  bcount := 0;
        while not done do
        begin
            ch := line[tcol];
            if ch = blank then bcount := bcount + 1
            else
            begin
                if bcount > 0 then
                begin
                    if bcount >= 10 then setcursor(r, tcol)
                    else
                    for bidx := 1 to bcount do character(blank);
                    bcount := 0
                end;
                character(ch)
            end;
            if tcol = mcol then done := true
                           else tcol := tcol + 1
        end
    end;  {putrow}
 
 
    { refresh -- refreshes the whole shared data structure }
    procedure refresh;
    var
        svrow, trow : rowtype;
        svcol, tcol : coltype;
        fidx, fcur, fmax : fieldindex;
        fdefs : array [fieldindex] of savedfield;
        midx : modeindex;
        mval : smallint;
        tdefs : array [coltype] of tabstop;
        text : rowtext;
    begin
        svrow := row;  svcol := col;  { note cursor position }
 
        { stored field definitions }
        fcur := curfield;  fmax := maxfield;
        for fidx := 0 to fmax do fdefs[fidx] := field[fidx];
        erasefields;
        for fidx := 0 to fmax do
        with fdefs[fidx] do
        begin
            setfield(fidx);
            setupdate(fldtop, fldbottom, fldleft, fldright)
        end;
        if fcur <> curfield then setfield(fcur);
 
        { mode array }
        if masterlevel = 0 then masterlevel := 1;  { safety check }
        mode[tlevel] := masterlevel;  mode[dsinvalid] := 0;
        mode[selectgr] := defrend;
        for midx := 0 to maxmode do
        begin
            mval := mode[midx];  setmode(midx, mval)
        end;
 
        { tabulation stops array }
        for tcol := 0 to maxcol do tdefs[tcol] := tabs[tcol];
        erasetabs;
        for tcol := 0 to maxcol do
            if tdefs[tcol] = tabset then settab(tcol);
 
        { screen image and cursor position }
        for trow := 0 to maxrow do
        begin
            getrow(trow, text);  putrow(trow, text)
        end;
        setcursor(svrow, svcol)
 
    end;  {refresh}
 
 
    { tcharacter -- interpret T-CHARACTER character replaced report }
    procedure tcharacter (ch : isochar); 
    begin 
        image[row,col] := ch; 
        if col < boxright then col := col + 1 
    end;  {tcharacter} 
 
 
    { ttoken -- interpret T-TOKEN return of token }
    procedure ttoken (reqcode : smallint); 
    begin 
        token := withhost; 
        if (mode[dsinvalid] <> 0)
            or (mode[tlevel] <> masterlevel) then refresh;
        { application will interpret the request code parameter }
    end;  {ttoken} 
 
 
    { tsetcursor -- interpret T-SETCURSOR cursor positioned report }
    procedure tsetcursor (newrow : rowtype; 
                          newcol : coltype); 
    begin 
        row := newrow;  col := newcol 
    end;  {tsetcursor} 
 
 
    { tsetmode -- interpret T-SETMODE mode array correction report }
    procedure tsetmode (index : modeindex; 
                        ivalue : smallint); 
    var
        newval : smallint;
    begin 
        newval := ivalue;
        if index <= dsinvalid then
        begin
            case index of
 
            tlevel:
                if newval = 0 then mode[dsinvalid] := 1
                else
                if masterlevel = 0 then masterlevel := newval;
 
            tmaxrow:
                if newval > zmaxrow then newval := zmaxrow;
 
            tmaxcol:
                if newval > zmaxcol then newval := zmaxcol;
 
            dsinvalid:
                newval := 1;
 
            end  {case}
        end
        else
        if (index = icharmode) and (newval > 1) then newval := 0
        else ivalue := 0;
        mode[index] := newval;
    end;  {tsetmode} 
 
 
    { terasetoright -- interpret T-ERASETORIGHT right erasure report }
    procedure terasetoright; 
    begin 
        if (mode[keraright] > 0) or (mode[keraline] > 0) 
            then eraserow(row, col, boxright) 
    end;  {terasetoright} 
 
 
    { tinsertspace -- interpret T-INSERTSPACE SPACEs inserted report }
    procedure tinsertspace (nsp : smallint); 
    begin 
        if (mode[kinsspac] > 0) 
            or (mode[icharmode] = 1) then 
                rightshift(row, col, boxright, nsp) 
    end;  {tinsertspace} 
 
 
    { tdeletechar -- interpret T-DELETECHAR characters deleted report }
    procedure tdeletechar (nch : smallint); 
    begin 
        if mode[kdelchar] > 0 
            then leftshift(row, col, boxright, nch) 
    end;  {tdeletechar} 
 
 
    { teraseprev -- interpret T-ERASEPREV previous char erasure report }
    procedure teraseprev (nch : smallint); 
    var 
        temp : smallint; 
    begin 
        if mode[keraprev] > 0 then 
        begin 
            { Note:  nch <= col - boxleft } 
            for temp := 1 to nch do 
            begin 
                col := col - 1;  image[row,col] := isospace 
            end 
        end 
    end;  {teraseprev} 
 
 
    { tinsertline -- interpret T-INSERTLINE line(s) inserted report }
    procedure tinsertline (nln : smallint); 
    var 
        rtemp : rowtype; 
    begin 
        if mode[kinsline] > 0 then 
        begin 
            if (boxbottom-row) < nln 
            then 
                erasebox(row, boxbottom, boxleft, boxright) 
            else 
            begin 
                for rtemp := boxbottom-nln downto row 
                    do rowcopy(rtemp, 
                                  rtemp+nln, boxleft, boxright); 
                erasebox(row, row+nln-1, boxleft, boxright) 
            end 
        end 
    end;  {tinsertline} 
 
 
    { tdeleteline -- interpret T-DELETELINE line(s) deleted report }
    procedure tdeleteline (nln : smallint); 
    var 
        rtemp : rowtype; 
    begin 
        if mode[kdelline] > 0 then 
        begin 
            if (boxbottom-row) < nln 
            then 
                erasebox(row, boxbottom, boxleft, boxright) 
            else 
            begin 
                for rtemp := row+nln to boxbottom 
                    do rowcopy 
                        (rtemp, rtemp-nln, boxleft, boxright); 
                erasebox(boxbottom-nln+1, boxbottom, 
                                             boxleft, boxright) 
            end 
        end 
    end;  {tdeleteline} 
 
 
    { tappendline -- interpret T-APPENDLINE line appended report }
    procedure tappendline; 
    var 
        rtemp : rowtype; 
    begin 
        if mode[kappline] > 0 then 
        begin 
            col := boxleft; 
            if (row >= mode[ilinerow]) or (row = boxbottom) then 
            begin 
                if boxtop < boxbottom then 
                    for rtemp := boxtop+1 to row 
                        do rowcopy(rtemp, rtemp-1, boxleft, boxright)
            end 
            else 
            begin 
                row := row + 1; 
                for rtemp := boxbottom-1 downto row 
                   do rowcopy(rtemp, rtemp+1, boxleft, boxright) 
            end; 
            eraserow(row, boxleft, boxright) 
        end 
    end;  {tappendline} 
 
 
    { tsplitline -- interpret T-SPLITLINE line split report }
    procedure tsplitline; 
    var 
        rtemp : rowtype; 
        cinit : coltype; 
    begin 
        if mode[ksplline] > 0 then 
        begin 
            cinit := col; 
            col := boxleft; 
            if (row >= mode[ilinerow]) or (row = boxbottom) then 
            begin 
                if boxtop < boxbottom then 
                    for rtemp := boxtop+1 to row 
                        do rowcopy(rtemp, rtemp-1, boxleft, boxright)
            end 
            else 
            begin 
                for rtemp := boxbottom-1 downto row 
                  do rowcopy(rtemp, rtemp+1, boxleft, boxright); 
                row := row + 1 
            end; 
            if row > boxtop then 
                eraserow(row-1, cinit, boxright); 
            leftshift(row, boxleft, boxright, cinit-boxleft) 
        end 
    end;  {tsplitline} 
 
 
    { tsetfield -- interpret T-SETFIELD field selection report }
    procedure tsetfield (fidx : fieldindex); 
    begin 
        { Note: fidx <= maxfield } 
        curfield := fidx;
        with field[curfield] do
        begin
            boxtop := fldtop;  boxbottom := fldbottom;
            boxleft := fldleft;  boxright := fldright
        end;
        row := boxtop;  col := boxleft
    end;  {tsetfield} 
 
 
    { parsenum -- parse a stream of characters into n numbers }
    function parsenum (expected : integer;
                       c : isochar;
                       var fin : boolean) : isochar;
    var
        nst : isochar;
    begin
        fin := false;
        nst := ssmpstate;       { state change on completion or error }
 
        if (c >= dig0) and (c <= dig9) then
        begin
            tsum := (10 * tsum) + (c - dig0);  gotdigit := true;
            if tsum > 255 then nst := question  { parameter too large }
        end else
        if ((c = comma) or (c = semicol))       { parameter terminator }
            and gotdigit then                   { at least one digit }
        begin
            tparidx := tparidx + 1;  tpars[tparidx] := tsum;
            tsum := 0;  gotdigit := false;
            if tparidx = expected then
            begin  { got enough pars, check termination }
                if c = semicol then
                begin
                    fin := true;  nst := nul    { finished }
                end
                else nst := question            { too many parameters }
            end
            else
            if c <> comma then nst := nul       { early termination }
        end else nst := question;               { unexpected character }
        parsenum := nst
    end;  {parsenum}
 
 
    { parsesemi -- check termination for primitives with no parameters }
    function parsesemi (c : isochar;
                        var ok : boolean) : isochar;
    var
        flag : boolean;
    begin
        flag := c = semicol;
        if flag then parsesemi := nul           { finished }
                else parsesemi := question;     { unexpected character }
        ok := flag
    end;  {parsesemi}
 
 
    { parsechar -- finite state machine emulation for ssmp }
    function parsechar (ch : isochar) : boolean;
    var
        nextst, tch : isochar;
        complete : boolean;
    begin
        nextst := question;  complete := true;
        case ssmpstate of
 
        nul:       { ground state, process after case to catch errors }
            ;
 
        colon:     { an encoded character follows }
            begin
                tch := decode[ch];
                if tch <> blank then
                begin
                    tprid := star;  tpars[1] := tch;  tcharacter(tch);
                    nextst := nul
                end
            end;
 
        equals:    { start of a primitive other than T-CHARACTER }
            if (ch >= capa) and (ch <= capl) then
            begin
                nextst := ch;  tprid := ch;  complete := false;
                tparidx := 0;  tsum := 0;  gotdigit := false
                                          { initialise number parser }
            end;
 
        capa:      { parsing T-TOKEN(reqcode) }
            begin
                nextst := parsenum(1, ch, complete);
                if complete then
                begin
                    ttoken(tpars[1]);
                    inlen := 0            { discard remainder of frame }
                end
            end;
 
        capb:      { parsing T-SETCURSOR(newrow, newcol) }
            begin
                nextst := parsenum(2, ch, complete);
                if complete then
                begin
                    complete :=
                        (tpars[1] <= maxrow) and (tpars[2] <= maxcol);
                    if complete then tsetcursor(tpars[1], tpars[2])
                end
            end;
 
        capc:      { parsing T-SETMODE(index, ivalue) }
            begin
                nextst := parsenum(2, ch, complete);
                if complete then
                begin
                    complete := tpars[1] <= 63;
                    if complete then tsetmode(tpars[1], tpars[2])
                end
            end;
 
        capd:      { parsing T-ERASETORIGHT }
            begin
                nextst := parsesemi(ch, complete);
                if complete then terasetoright
            end;
 
        cape:      { parsing T-INSERTSPACE(nch) }
            begin
                nextst := parsenum(1, ch, complete);
                if complete then tinsertspace(tpars[1])
            end;
 
        capf:      { parsing T-DELETECHAR(nch) }
            begin
                nextst := parsenum(1, ch, complete);
                if complete then tdeletechar(tpars[1])
            end;
 
        capg:      { parsing T-ERASEPREV(nch) }
            begin
                nextst := parsenum(1, ch, complete);
                if complete then teraseprev(tpars[1])
            end;
 
        caph:      { parsing T-INSERTLINE(nln) }
            begin
                nextst := parsenum(1, ch, complete);
                if complete then tinsertline(tpars[1])
            end;
 
        capi:      { parsing T-DELETELINE(nln) }
            begin
                nextst := parsenum(1, ch, complete);
                if complete then tdeleteline(tpars[1])
            end;
 
        capj:      { parsing T-APPENDLINE }
            begin
                nextst := parsesemi(ch, complete);
                if complete then tappendline
            end;
 
        capk:      { parsing T-SPLITLINE }
            begin
                nextst := parsesemi(ch, complete);
                if complete then tsplitline
            end;
 
        capl:      { parsing T-SETFIELD(fidx) }
            begin
                nextst := parsenum(1, ch, complete);
                if complete then
                begin
                    complete := tpars[1] <= maxfield;
                    if complete then tsetfield(tpars[1])
                end
            end
 
        end;  {case}
 
        if nextst = question then
        begin      { process in ground state }
            nextst := nul;  complete := false;
            if (ch = equals) or (ch = colon) then nextst := ch
            else
            if ch <> semicol then
            begin  { T-CHARACTER }
                tprid := star;   tpars[1] := ch;  tcharacter(ch);
                complete := true
            end
        end;
 
        ssmpstate := nextst;  parsechar := complete
    end;  {parsechar}
 
 
    { getprimitive -- process input until a primitive has been decoded }
    procedure getprimitive (var pcode : isochar;
                            var ppars : primpar);
    var
        nextch : isochar;
    begin
        repeat
            repeat
                while inptr >= inlen do netread;
                inptr := inptr + 1;
                nextch := filter[inbuff[inptr]]
            until nextch <> nul;
        until parsechar(nextch);
        pcode := tprid;  ppars := tpars
    end;  {getprimitive}
 
 
    { startssmp -- negotiate screen management session with terminal }
    procedure startssmp (var success : boolean);
    begin
        initliterals;  initcoding;
        netopen;
 
        session(0);                { initialises shared data structure }
        literal('  Please press the RETUR');
        literal('N key to continue.  &   ');
        sendtoken(1);
        netread;    { read first frame, then crude check for T-SETMODE }
        success := (inbuff[1] = equals) and (inbuff[2] = capc)
            and (inbuff[3] = dig0);
        if not success then
        begin
           writeln('Simple Screen Management Protocol :::',
                   ' connection refused.');
           netclose
        end
    end;  {startssmp }
 
 
    { stopssmp -- tidy up and close down screen management session }
    procedure stopssmp;
    begin
        scrollup(0, maxrow, 1);  setcursor(maxrow, 0);  session(1);
        netclose
    end;  {stopssmp}
 
 
{ ******************************************************************* }
{ *                                                                 * }
{ *               T h e   A p p l i c a t i o n                     * }
{ *               =============================                     * }
{ *                                                                 * }
{ *    All SSMP service procedures have been defined above.         * }
{ *    All references to the SSMP shared data structure by the      * }
{ *    application are by procedure or function calls, never by     * }
{ *    direct reference to the variables.                           * }
{ *                                                                 * }
{ ******************************************************************* }
 
 
    { menureply -- returns code in range 0..12 after menu reply }
    function menureply (req : smallint) : smallint;
    var
        creq, rc : smallint;
        r : rowtype;
        c : coltype;
        curline : rowtext;
    begin
        creq := req;  rc := 12;
        if creq = 0 then
        begin  { single character box exit }
            getloc(r, c);                              { where? }
            getrow(r, curline);  creq := curline[c];   { get reply }
            setcursor(r, c);  character(blank)         { clear it }
        end;
        if (creq >= dig0) and (creq <= dig9) then rc := creq - dig0
        else
        if (creq = letm) or (creq = capm) then rc := 10
        else
        if (creq = lets) or (creq = caps) then rc := 11;
        menureply := rc
    end;  {menureply}
 
 
    { genescmsg -- display the general screen exit message }
    procedure genescmsg;
    begin
        literal('To exit:    Press ESC th');
        literal('en a menu selection (0..');
        literal('9, S) or M for the menu.');
    end;  {genescmsg}
 
 
    { putbox -- draw a box to show the limits of the screen offered }
    procedure putbox;
    var
        r : rowtype;
        c : coltype;
    begin
        erasedisplay;  character(plus);
        for c := 1 to maxcol-1 do character(minus);
        character(plus);
        c := maxcol;
        for r := 1 to maxrow-1 do
        begin
            setcursor(r, 0);  character(bar);
            setcursor(r, c);  character(bar)
        end;
        setcursor(maxrow, 0);  character(plus);
        for c := 1 to maxcol-1 do character(minus);
        character(plus);
    end;  {putbox}
 
 
    { putmenu -- put up main menu prompts }
    procedure putmenu;
    var
        j : smallint;
    begin
        appstate := menu;  putmode(menumode);
        putbox;  setcursor(0, (maxcol-47) div 2);
        setmode(selectgr, negative);
        literal(' S S M P   T e r m i n a');
        literal(' l   E x e r c i s e r &');
        setmode(selectgr, defrend);
        for  j := 0 to 9 do
        begin
            setcursor(j+4, 8);
            character(j+dig0);  character(colon);
            character(blank);  character(blank);
            case j of
 
            0:  begin
                    literal('ISO 646 characters and g');
                    literal('raphic rendition display');
                end;
 
            1:  begin
                    literal('Exercise host-generated ');
                    literal('primitives&             ');
                end;
 
            2:  begin
                    literal('Out-of-context message d');
                    literal('isplay&                 ')
                end;
 
            3:  begin
                    literal('Illegally formed primiti');
                    literal('ves:  test resilience&  ');
                end;
 
            4:  begin
                    literal('Soak test:  about 18000 ');
                    literal('characters down the line');
                end;
 
            5:  begin
                    literal('Display and modification');
                    literal(' of the mode array&     ');
                end;
 
            6:  begin
                    literal('Test image with a SINGLE');
                    literal(' data entry field&      ');
                end;
 
            7:  begin
                    literal('As previous, but with te');
                    literal('st pattern in place&    ');
                end;
 
            8:  begin
                    literal('Test image with MULTIPLE');
                    literal(' data entry fields&     ');
                end;
 
            9:  begin
                    literal('As previous, but with te');
                    literal('st pattern in place&    ');
                end;
 
            end  {case}
        end;
 
        setcursor(15, 8);  literal('S:  Stop&               ');
        setcursor(18, 8);
        literal('Key one digit or the let');
        literal('ter "S": &              ');
        erasefields;  setupdate(18, 18, 41, 41)
    end;  {putmenu}
 
 
    { firstmenu -- first entry to menu, set up a few things }
    procedure firstmenu;
    var
        j : smallint;
    begin
        setmode(reqshift, 1);         { try for host shift primitives }
 
        { set up session control functions as recommended }
        setmode(kinthost, 33);        { Interrupt Host:   ESC then "!" }
        setmode(ksuspend, 62);        { Suspend Session:  ESC then ">" }
        setmode(krestart, 63);        { Restart Session:  ESC then "?" }
 
        getmode(menumode);            { save the menu mode array }
 
        { set up initial mode array for the array display screen }
        mdismode := menumode;
        mdismode[kenter]    :=  13;   { Enter:            CTRL/M  (CR) }
        mdismode[knextfld]  :=  15;   { Next Field:       CTRL/O }
        mdismode[kprevfld]  :=  16;   { Previous Field:   CTRL/P }
 
        { set a variant used when entering a new element value }
        mmodmode := menumode;
        mmodmode[kenter]    :=  13;   { Enter:            CTRL/M  (CR) }
        mmodmode[kcsrleft]  :=   8;   { Cursor Left:      CTRL/H  (BS) }
        mmodmode[kcsrright] :=  12;   { Cursor Right:     CTRL/L }
        mmodmode[kleftupd]  :=  24;   { Line Start:       CTRL/X }
        mmodmode[keraprev]  := 127;   { Erase Previous:   DEL }
 
        { set up initial mode array for the four editing screens }
        editmode := menumode;
        editmode[ilinerow]  :=  16;   { Preferred insertion row }
        editmode[kenter]    :=  14;   { Enter:            CTRL/N }
        editmode[kcsrup]    :=  11;   { Cursor Up:        CTRL/J  (LF) }
        editmode[kcsrdown]  :=  10;   { Cursor Down:      CTRL/K }
        editmode[kcsrleft]  :=   8;   { Cursor Left:      CTRL/H  (BS) }
        editmode[kcsrright] :=  12;   { Cursor Right:     CTRL/L }
        editmode[keraprev]  := 127;   { Erase Previous:   DEL }
        editmode[knexttab]  :=   9;   { Next Tab:         CTRL/I  (HT) }
        editmode[kprevtab]  :=   6;   { Previous Tab:     CTRL/F }
        editmode[kleftupd]  :=  24;   { Line Start:       CTRL/X }
        editmode[kfirstns]  :=  26;   { Text Start:       CTRL/Z }
        editmode[kalastns]  :=   3;   { Text End:         CTRL/C }
        editmode[kinsmode]  :=  23;   { Insertion Toggle: CTRL/W }
        editmode[keraright] :=  25;   { Erase to Right:   CTRL/Y }
        editmode[kinsspac]  :=  19;   { Insert Space:     CTRL/S }
        editmode[kdelchar]  :=   5;   { Delete Character: CTRL/E }
        editmode[kinsline]  :=  17;   { Insert Line:      CTRL/Q }
        editmode[kdelline]  :=  21;   { Delete Line:      CTRL/U }
        editmode[keraline]  :=  92;   { Erase Line:       ESC then "\" }
        editmode[khomefld]  :=   7;   { Home Field:       CTRL/G }
        editmode[knextfld]  :=  15;   { Next Field:       CTRL/O }
        editmode[kprevfld]  :=  16;   { Previous Field:   CTRL/P }
        editmode[knewline]  :=  13;   { New Line:         CTRL/M  (CR) }
 
        { establish tab stops }
        for j := 1 to (maxcol div 8) do settab(j*8);
 
        { put up the menu }
        putmenu
    end;  {firstmenu}
 
 
    { puthexdigit -- send one hexadecimal digit to the screen }
    procedure puthexdigit (digit : smallint);
    begin
        if digit <= 9 then character(dig0+digit)
                      else character(capa-10+digit)
    end;  {puthexdigit}
 
 
    { putrend -- put rendition name on character set display }
    procedure putrend (sgrcode : smallint);
    begin
        setcursor(sgrcode+2, 46);
        character(blank);  character(sgrcode+dig0);
        character(colon);  character(blank);  character(blank);
        case sgrcode of
 
        0:  literal('Default rendition&      ');
        1:  literal('Bold&                   ');
        2:  literal('Faint&                  ');
        3:  literal('Italicised&             ');
        4:  literal('Underlined&             ');
        5:  literal('Slowly blinking&        ');
        6:  literal('Rapidly blinking&       ');
        7:  literal('Negative image&         ');
        8:  literal('Concealed&              ');
        9:  literal('Crossed out&            ');
 
        end;
        character(blank)
    end;  {putrend}
 
 
    { markrend -- mark rendition as ok or not available }
    procedure markrend (sgrcode : smallint;
                        ok : boolean);
    begin
        setcursor(sgrcode+2, 71);
        if ok then literal(' OK&                    ')
              else literal('N/A&                    ')
    end;  {markrend}
 
 
    { putaxes -- put out labelled axes for character set display }
    procedure putaxes;
    var
        r : rowtype;
        c : coltype;
        j : smallint;
    begin
        erasedisplay;
        setcursor(0, 8);  character(bar);
        for j := 0 to 7 do
        begin
            character(blank);   character(dig0+j);
            character(letx);    character(blank)
        end;
        setcursor(1, 5);
        for j := 1 to 37 do character(minus);
        setcursor(1, 8);  character(plus);
        for j := 0 to 15 do
        begin
            setcursor(j+2, 5);  character(letx);  puthexdigit(j);
            character(blank);  character(bar)
        end;
 
        lastrend := 0;
        setmode(selectgr, negative);  putrend(0);
        setmode(selectgr, defrend);
        for j := 1 to 9 do putrend(j);
        setcursor(13, 46);  literal(' M:  Menu&              ');
        setcursor(14, 46);  literal(' S:  Stop&              ');
 
        setcursor(17, 46);  literal(' Next?&                 ');
        erasefields;  setupdate(17, 17, 55, 55);
 
        if maxrow >=21 then
        begin
            setcursor(21, (maxcol-60) div 2);
            setmode(selectgr, negative);
            literal(' ISO 646 character set w');
            literal('ith SELECTGR values (fro');
            literal('m ISO 6429) &           ');
            setmode(selectgr, defrend)
        end
    end;  {putaxes}
 
 
    { putcharset -- put up the ISO 646 character set grid }
    procedure putcharset;
    var
        i, j : smallint;
    begin
        for j := 0 to 15 do
        for i := 0 to 7 do
        begin
            if i = 0 then setcursor(j+2, (i*4)+10)
                     else character(blank);
            character(blank);  character((i*16)+j);  character(blank)
        end;
        tokencode := 1
    end;  {putcharset}
 
 
    { firstcharset -- entry to character set display }
    procedure firstcharset;
    begin
        appstate := charset;  putmode(menumode);
        putaxes;  putcharset
    end;  {firstcharset}
 
 
    { incharset -- process reply from character set display }
    procedure incharset (reply : smallint);
    var
        newcode : smallint;
        mcopy : modearray;
        ok : boolean;
    begin
        if tokencode = 1 then
        begin  { return from resticted transfer }
            getmode(mcopy);
            ok := mcopy[selectgr] = lastrend; { did the last one work? }
            setmode(selectgr, defrend);       { reset it }
            markrend(lastrend, ok);           { mark "ok" or "n/a" }
            tokencode := 0                    { let the user at it }
        end
        else
        begin  { user reply }
            newcode := menureply(reply);
            if newcode <= 9 then
            begin
                putrend(lastrend);            { reset last choice }
                setmode(selectgr, negative);  { change rendition }
                putrend(newcode);             { emphasise new choice }
                setmode(selectgr, newcode);   { requested rendition }
                putcharset;                   { output characters }
                lastrend := newcode
            end
            else
            begin
                case newcode of
 
                10: putmenu;                  { "M"|"m" : Menu }
                11: screenmode := false;      { "S"|"s" : Stop }
                12: soundalarm                { anything else }
 
                end  {case}
            end
        end
    end;  {incharset}
 
 
    { dmsprompt -- prompt for 0..9, M or S reply }
    procedure dmsprompt;
    var
        mrow : rowtype;
        j : smallint;
        ln : rowtext;
    begin
        mrow := maxrow;
        getrow(mrow-2, ln);
        if ln[0] <> bar then
        begin
            setcursor(mrow-3, 0);  erasetoright;
            setcursor(mrow-2, 0);  erasetoright;
            setcursor(mrow-1, 0);  erasetoright
        end;
        setcursor(mrow-2, 2);  genescmsg;
        erasefields
    end;  {dmsprompt}
 
 
    { dmsprocess -- process for 0..9, M or S reply }
    procedure dmsprocess (reply : smallint);
    var
        mqcode : smallint;
    begin
        newcode := menureply(reply);  domore := true
    end;  {dmsprocess}
 
 
    { putlogo -- put out "SSMP" in block capitals }
    procedure putlogo;
    type
        oneline = packed array [1..12] of char;
        blkletter = array [1..12] of oneline;
    var
        blks, blkm, blkp : blkletter;
        chline : oneline;
        x, y : 1..12;
        z : 0..3;
        srow : rowtype;
        scol, tcol : coltype;
        ch : char;
        rch : isochar;
        textline : rowtext;
    begin
        blks[1]  := ' SSSSSSSSSS ';  blkm[1]  := 'MM        MM';
        blks[2]  := 'SSSSSSSSSSSS';  blkm[2]  := 'MMM      MMM';
        blks[3]  := 'SS        SS';  blkm[3]  := 'MMMM    MMMM';
        blks[4]  := 'SS          ';  blkm[4]  := 'MM MM  MM MM';
        blks[5]  := 'SSS         ';  blkm[5]  := 'MM  MMMM  MM';
        blks[6]  := ' SSSSSSSSS  ';  blkm[6]  := 'MM   MM   MM';
        blks[7]  := '  SSSSSSSSS ';  blkm[7]  := 'MM        MM';
        blks[8]  := '         SSS';  blkm[8]  := 'MM        MM';
        blks[9]  := '          SS';  blkm[9]  := 'MM        MM';
        blks[10] := 'SS        SS';  blkm[10] := 'MM        MM';
        blks[11] := 'SSSSSSSSSSSS';  blkm[11] := 'MM        MM';
        blks[12] := ' SSSSSSSSSS ';  blkm[12] := 'MM        MM';
 
        blkp[1]  := 'PPPPPPPPPPP ';
        blkp[2]  := 'PPPPPPPPPPPP';
        blkp[3]  := 'PP        PP';
        blkp[4]  := 'PP        PP';
        blkp[5]  := 'PP        PP';
        blkp[6]  := 'PPPPPPPPPPPP';
        blkp[7]  := 'PPPPPPPPPPP ';
        blkp[8]  := 'PP          ';
        blkp[9]  := 'PP          ';
        blkp[10] := 'PP          ';
        blkp[11] := 'PP          ';
        blkp[12] := 'PP          ';
 
        srow := (maxrow - 12) div 2;  scol := (maxcol-70) div 2;
        for y := 12 downto 1 do
        begin
            for tcol := 0 to maxcol do textline[tcol] := blank;
            for z := 0 to 3 do
            begin
                case z of
                0:  chline := blks[y];
                1:  chline := blks[y];
                2:  chline := blkm[y];
                3:  chline := blkp[y];
                end;  {case}
                for x := 1 to 12 do
                begin
                    ch := chline[x];
                    if ch <> ' ' then
                    begin
                        if ch = 'S' then rch := caps
                        else if ch = 'M' then rch := capm
                        else rch := capp;
                        textline[scol+x+(12-y)+(z*14)] := rch
                    end
                end
            end;
            putrow(srow+y, textline)
        end
    end;  {putlogo}
 
 
    { hosttitle -- put out title for host test screen }
    procedure hosttitle;
    begin
        setcursor(0, (maxcol-32) div 2);
        literal(' Test Host-generated Pri');
        literal('mitives &               ')
    end;  {hosttitle}
 
 
    { hostbars -- put out horizontal bars for host test screen }
    procedure hostbars;
    var
        srow : rowtype;
        width : smallint;
    begin
        srow := (maxrow -12) div 2;  srow := srow + 1;
        width := maxcol + 1;
        setcursor(srow-3, 0);   character(plus);
        setcursor(srow-3, 0);   insertspace(width);
        setcursor(srow+12, 0);  insertspace(width);
    end;  {hostbars}
 
 
    { hostrend -- refresh parts of image not in default rendition }
    procedure hostrend;
    begin
        setmode(selectgr, negative);
        hosttitle;  hostbars;
        setmode(selectgr, defrend);
    end;  {hostrend}
 
 
    { hosttest -- test host primitives }
    procedure hosttest;
    var
        r, srow, mrow : rowtype;
        mcopy : modearray;
        shiftok, nrendok : boolean;
    begin
        if tokencode <> 1 then
        begin  { first entry -- check out the terminal }
            appstate := fhost;  putmode(menumode);
            erasedisplay;  setmode(selectgr, negative);
            hosttitle;
            tokencode := 1;
        end
        else
        begin  { return from restricted transfer }
            appstate := thost;
            getmode(mcopy);
            shiftok := mcopy[reqshift] = 1;
            nrendok := mcopy[selectgr] = negative;
 
            setmode(selectgr, defrend);
            putlogo;
            srow := (maxrow -12) div 2;  srow := srow + 1;
            mrow := srow + 6;
            for r := srow to (srow + 11) do
            begin
                setcursor(r, 0);  character(greater);  character(colon);
                if r <> mrow then
                begin
                    if r < mrow then deletechar(mrow-r)
                                else insertspace(r-mrow)
                end;
            end;
            scrollup(1, mrow, 4);  scrollup(mrow-3, maxrow, 4);
            scrolldown(1, maxrow, 3);
 
            setmode(selectgr, negative);
            hostbars;
            setmode(selectgr, defrend);
 
            setcursor(srow-4, 2);
            literal('The block letters "SSMP"');
            literal(' below should &         ');
            if shiftok then
                literal('no longer &             ');
            literal('be slanted.&            ');
 
            if shiftok and nrendok then
            begin
                setcursor(srow+13, 2);
                literal('The row above should be ');
                literal('the second of two full-w');
                literal('idth negative image bars');
                character(period);
            end;
 
            tokencode := 0;  dmsprompt
        end
    end;  {hosttest}
 
 
    { ooctest -- test out-of-context message handling }
    procedure ooctest;
    var
        j : smallint;
    begin
        appstate := ooc;  putmode(menumode);  putbox;
        setcursor(0, (maxcol-29) div 2);
        setmode(selectgr, negative);
        literal(' Out-of-context message ');
        literal('test &                  ');
        setcursor(maxrow, (maxcol-12) div 2);
        literal(' bottom row &           ');
        setmode(selectgr, defrend);
 
        for j := 1 to 8 do
        begin
            setcursor(j+2, 10);
            literal('Out-of-context message t');
            literal('est.  This is line # &  ');
            character(j+dig0);  character(period);
            if j = 4 then
            begin
                literal('  ---> Now!&            ');
                writeln(' ');
                writeln('** Mailer:  Message from Santa@North-Pole');
                writeln('** Operator:  Coffee m/c down indefinitely');
                writeln('** Operator:  Widget maintenance 18:00-18:30')
            end
        end;
        dmsprompt    { if it gets past here, recovery worked }
    end;  {ooctest}
 
 
    { rendinfo -- explain why non-default renditions have vanished }
    procedure rendinfo;
    begin
        setcursor(13, 2);
        literal('With the single exceptio');
        literal('n of this paragraph, all');
        literal(' text now on the screen&');
        setcursor(14, 2);
        literal('was placed there by the ');
        literal('program''s SSMP-level err');
        literal('or recovery routines.&  ');
        setcursor(15, 2);
        literal('The rendition of each ch');
        literal('aracter is not held in t');
        literal('he SSMP shared data&    ');
        setcursor(16, 2);
        literal('structure so non-default');
        literal(' renditions in the image');
        literal(' have been lost.&       ');
        setcursor(17, 2);
        literal('A host application can t');
        literal('rap and process such an ');
        literal('event; you can see this&');
        setcursor(18, 2);
        literal('in action by invoking Se');
        literal('ssion Restart (ESC "?") ');
        literal('from menu selection #1.&');
    end;  {rendinfo}
 
 
    { inject -- injects characters into the frames output }
    procedure inject (ichar : isochar);
    begin
        { WARNING -- this facility should not normally be used
          by a host application.  It is provided here to test the
          resilience of a terminal emulation and accesses the SSMP
          module output buffer directly.  DON'T DO IT!  }
 
        if outremdr = 0 then
        begin
            endframe;
            outptr := outptr + 1;  outbuff[outptr] := ctlk;
            outremdr := framesize
        end;
        outptr := outptr + 1;  outbuff[outptr] := ichar;
        outremdr := outremdr - 1
    end;  {inject}
 
 
    { injnumber -- calls inject to feed in a parameter number }
    procedure injnumber (nbr : smallint);
    var
        n : smallint;
    begin  { not exactly beautiful Pascal, but only used by "rogueapp" }
        n := nbr;
        if n >= 100 then
        begin
            inject(dig0 + (n div 100));  n := n mod 100
        end;
        inject(dig0 + (n div 10));  n := n mod 10;
        inject(dig0 + n)
    end;  {injnumber}
 
 
    { rogueapp -- test recovery from badly formed host primitives }
    procedure rogueapp;
    var
        r, crow : rowtype;
        c : coltype;
        j : smallint;
        midx : modeindex;
        ch : isochar;
    begin
        { WARNING!  This procedure references the encode and decode
          tables directly to test error recovery.  DON'T DO IT! }
 
        appstate := rogue;  putmode(menumode);  erasedisplay;
        setcursor(0, (maxcol-64) div 2);  setmode(selectgr, negative);
        literal(' Test Resilience to ille');
        literal('gally formed host-genera');
        literal('ted primitives &        ');
        setmode(selectgr, defrend);
 
        setcursor(2, 2);
        literal('For each test, the illeg');
        literal('al sequence is bracketed');
        literal(' by "abcd" and "wxyz".& ');
 
        setcursor(3, 2);
        literal('Correspondence between "');
        literal('Expected" and "Result" f');
        literal('ields need not be exact.');
 
        setmode(selectgr, negative);
        setcursor(5, 8);   literal(' Description of test &  ');
        setcursor(5, 41);  literal(' Expected &             ');
        setcursor(5, 60);  literal(' Result &               ');
        setmode(selectgr, defrend);
 
        { identify and perform each test }
        for j := 0 to 9 do
        begin
            crow := j + 7;
            setcursor(crow, 4);  character(j+dig0);
            character(colon);  character(blank);  character(blank);
 
            case j of
 
            0:  literal('Unallocated mode indices');
            1:  literal('Invalid simple encodings');
            2:  literal('Invalid ":" encodings&  ');
            3:  literal('Too few parameters&     ');
            4:  literal('Too many parameters&    ');
            5:  literal('Cursor row out of range&');
 
            6:  begin
                    literal('Cursor column out of ran');
                    literal('ge&                     ');
                end;
 
            7:  literal('Field index out of range');
            8:  literal('Parameter too large&    ');
            9:  begin
                    literal('Invalid primitive identi');
                    literal('fier&                   ');
                end;
 
            end;  {case}
 
            { first, what you should see }
            setcursor(crow, 40);
            case j of
 
            0:  literal(' abcdwxyz &             ');
            1:  literal(' abcdwxyz &             ');
            2:  literal(' abcd(zwxyz &           ');
            3:  literal(' abcdwxyz &             ');
            4:  literal(' abcd,50wxyz &          ');
            5:  literal(' abcdwxyz &             ');
            6:  literal(' abcdwxyz &             ');
            7:  literal(' abcdwxyz &             ');
            8:  literal(' abcd6wxyz &            ');
            9:  literal(' abcdR0wxyz &           ');
 
            end;  {case}
 
            { now the actual test sequence }
            setcursor(crow, 60);
            literal('abcd&                   ');
            case j of
 
            0:  begin  { Unallocated mode indices }
                    for midx := 38 to 63 do setmode(midx, 1)
                end;
 
            1:  begin  { Invalid simple encodings }
                    for ch := blank to tilde do
                        if (encode[ch] <> blank) and (ch <> colon)
                            and (ch <> equals)
                                then inject(ch)
                end;
 
            2:  begin  { Invalid ":" encodings }
                    for ch := blank to tilde do
                        if decode[ch] = blank then
                        begin
                            inject(colon);  inject(ch)
                        end
                end;
 
            3:  begin  { Too few parameters }
                    inject(equals);  inject(capb);
                    injnumber(20);  inject(semicol);
                end;
 
            4:  begin  { Too many parameters }
                    inject(equals);  inject(capb);
                    injnumber(20);  inject(comma);
                    injnumber(60);  inject(comma);
                    injnumber(50);  inject(semicol);
                end;
 
            5:  begin  { Cursor row out of range }
                    inject(equals);  inject(capb);
                    injnumber(maxrow+1);  inject(comma);
                    injnumber(maxcol div 2);  inject(semicol)
                end;
 
            6:  begin  { Cursor column out of range }
                    inject(equals);  inject(capb);
                    injnumber(maxrow div 2);  inject(comma);
                    injnumber(maxcol+1);  inject(semicol)
                end;
 
            7:  begin  { Field index out of range }
                    inject(equals);  inject(capl);
                    injnumber((maxrow+1)*4);  inject(semicol)
                end;
 
            8:  begin  { Parameter too large }
                    inject(equals);  inject(cape);
                    inject(dig2);  inject(dig5);  inject(dig6);
                    inject(semicol)
                end;
 
            9:  begin  { Invalid primitive id }
                    inject(equals);  inject(capr);
                    inject(dig0);  inject(semicol)
                end;
 
            end;  {case}
            literal('wxyz&                   ');
        end;
 
        { a last note of caution }
        setcursor(18, 2);
        literal('Warning:  The illegal pr');
        literal('imitives were not genera');
        literal('ted via the SSMP module&');
        setcursor(19, 2);
        literal('and Session Restart (ESC');
        literal(' "?") will not preserve ');
        literal('the "Result" images.&   ');
 
        dmsprompt
    end;  {rogueapp}
 
 
    { soaktest -- write the screen image the hard way }
    procedure soaktest;
    var
        r : rowtype;
        c : coltype;
    begin
        appstate := soak;  putmode(menumode);  erasedisplay;
        setcursor(0, (maxcol-11) div 2);
        setmode(selectgr, negative);
        literal(' Soak Test &            ');
        setmode(selectgr, defrend);
 
        for r := maxrow downto 1 do
        for c := maxcol downto 0 do
        begin
            setcursor(r, c);  character(((r+c) mod 94) + 33)
        end;
        dmsprompt
    end;  {soaktest}
 
 
    { putkey -- display the keystroke equivalent of a command code }
    procedure putkey (ccode : smallint;
                      key : boolean);
    var
        done : boolean;
        temp, cct : smallint;
        idx : 0..3;
        s : array [0..3] of isochar;
    begin
        character(sharp);
        puthexdigit(ccode div 16);  puthexdigit(ccode mod 16);
        character(blank);  character(blank);
 
        done := key;
        if key then
        begin
            if (ccode >= blank) and (ccode <= tilde) then
            begin
                literal('ESC "&                  ');
                character(ccode);  character(dquote);
            end
            else if (ccode >= ctla) and (ccode <= ctlz) then
            begin
                literal('CTRL/&                  ');
                character(capa-ctla+ccode);  character(blank);
            end
            else if ccode = del then literal('DEL    &                ')
            else done := false
        end;
 
        if not done then
        begin  { output as decimal }
            character(lbrack);
            for idx := 1 to 3 do s[idx] := blank;
            temp := ccode;  idx := 3;
            repeat
                s[idx] := dig0 + (temp mod 10);
                temp := temp div 10;  idx := idx - 1
            until temp = 0;
            cct := 0;
            for idx := 1 to 3 do
            if s[idx] <> blank then
            begin
                character(s[idx]);  cct := cct + 1
            end;
            character(rbrack);
            for cct := cct to 4 do character(blank)
        end;
    end;  {putkey}
 
 
    { putitem -- display the requested mode array element with value }
    procedure putitem(mindex : smallint;
                      mvalue : smallint);
    var
        srow : rowtype;
        scol : coltype;
    begin
        srow := 2 + (mindex mod 14);
        scol := 1 + ((mindex div 14) * 25);
        setcursor(srow, scol);  character(blank);
 
        case mindex of
        0:  literal('TLEVEL    &             ');
        1:  literal('TMAXROW   &             ');
        2:  literal('TMAXCOL   &             ');
        3:  literal('DSINVALID &             ');
        4:  literal('NOTIFY    &             ');
        5:  literal('SELECTGR  &             ');
        6:  literal('REQSHIFT  &             ');
        7:  literal('CURSOR    &             ');
        8:  literal('ICHARMODE &             ');
        9:  literal('ILINEROW  &             ');
        10: literal('INTSIGNAL &             ');
        11: literal('KINTHOST  &             ');
        12: literal('KSUSPEND  &             ');
        13: literal('KRESTART  &             ');
        14: literal('KENTER    &             ');
        15: literal('KCSRUP    &             ');
        16: literal('KCSRDOWN  &             ');
        17: literal('KCSRLEFT  &             ');
        18: literal('KCSRRIGHT &             ');
        19: literal('KNEXTTAB  &             ');
        20: literal('KPREVTAB  &             ');
        21: literal('KLEFTUPD  &             ');
        22: literal('KFIRSTNS  &             ');
        23: literal('KALASTNS  &             ');
        24: literal('KINSMODE  &             ');
        25: literal('KERARIGHT &             ');
        26: literal('KINSSPAC  &             ');
        27: literal('KDELCHAR  &             ');
        28: literal('KERAPREV  &             ');
        29: literal('KINSLINE  &             ');
        30: literal('KDELLINE  &             ');
        31: literal('KERALINE  &             ');
        32: literal('KAPPLINE  &             ');
        33: literal('KSPLLINE  &             ');
        34: literal('KNEXTFLD  &             ');
        35: literal('KPREVFLD  &             ');
        36: literal('KHOMEFLD  &             ');
        37: literal('KNEWLINE  &             ');
        end;  {case}
 
        if mindex >=  11 then putkey(mvalue, true)
                         else putkey(mvalue, false);
        setcursor(srow, scol+10);
    end;  {putitem}
 
 
    { mdisstate -- setup state and messages for mode element display }
    procedure mdisstate;
    begin
        appstate := dismode;
        putmode(mdismode);
        setcursor(18, 0);  erasetoright;
        genescmsg;
        setcursor(19, 0);  erasetoright;
        literal('To modify:  Select -- CT');
        literal('RL/O (next) or CTRL/P (p');
        literal('revious) -- then RETURN.');
        setcursor(21, 0);  erasetoright;
        literal('In the test screens, ESC');
        literal(' then [QWERTYUIOP] sets ');
        literal('rendition [1234567890].&');
        setmode(selectgr, concealed);
    end;  {mdisstate}
 
 
    { mmodstate -- setup state and messages for new value entry }
    procedure mmodstate;
    begin
        appstate := modmode;  putmode(mmodmode);
        setcursor(21, 0);  erasetoright;
 
        setcursor(18, 0);  erasetoright;
        literal('Enter a replacement valu');
        literal('e and press RETURN.&    ');
 
        setcursor(19, 0);  erasetoright;
        literal('Replies are &           ');
        setmode(selectgr, negative);
        literal('<char>&                 ');
        setmode(selectgr, defrend);
        character(comma);  character(blank);
        setmode(selectgr, negative);
        character(cflex);                           { EBCDIC!! }
        literal('<letter>&               ');
        setmode(selectgr, defrend);
        literal(' or &                   ');
        setmode(selectgr, negative);
        literal('#<hexnbr>&              ');
        setmode(selectgr, defrend);   character(period);
 
        setcursor(19, 48);
        setmode(selectgr, bold);
        character(greater);  character(colon);
        setmode(selectgr, defrend);
        setfield(0);  setupdate(19, 19, 50, 61);
    end;  {mmodstate}
 
 
    { firstmode -- first entry to mode display }
    procedure firstdismode;
    var
        m : smallint;
        r : rowtype;
        c : coltype;
        f : fieldindex;
    begin
        appstate := dismode;  putmode(mdismode);
        erasedisplay;  erasefields;
        setcursor(0, (maxcol-62) div 2);
        setmode(selectgr, negative);
        literal(' Mode array values for t');
        literal('he function test screens');
        literal(' (Menu: 6..9) &         ');
        setmode(selectgr, defrend);
        for m := 0 to 37 do
        begin
            putitem(m, editmode[m]);
            if m > 3 then
            begin
                getloc(r, c);  getfield(f);
                setfield(f+1);  setupdate(r, r, c, c)
            end
        end;
        mdisstate;
        setfield(0);  setcursor(0, 0)
    end;  {firstdismode}
 
 
    { modeselect -- mode display element selection or exit }
    procedure modeselect (reply : smallint);
    var
        f : fieldindex;
    begin
        if reply = ctlm then
        begin
            getfield(f);
            if f <> 0 then
            begin
                lastmode := f + 3;
                setmode(selectgr, negative);
                putitem(lastmode, editmode[lastmode]);
                mmodstate   { includes rendition reset }
            end
        end
        else
        begin
            newcode := menureply(reply);  domore := true
        end
    end;  {modeselect}
 
 
    { decodehex -- decode one hex digit }
    function decodehex (ch : isochar;
                        var nbr : smallint) : boolean;
    var
        temp : smallint;
    begin
        temp := 0;  decodehex := true;
        if (ch >= dig0) and (ch <= dig9) then temp := ch - dig0
        else
        if (ch >= leta) and (ch <= letf) then temp := ch - leta + 10
        else
        if (ch >= capa) and (ch <= capf) then temp := ch - capa + 10
        else decodehex := false;
        nbr := temp
    end;  {decodehex}
 
 
    { modemodify -- element selected; this reply signals the new value }
    procedure modemodify (reply : smallint);
    var
        inline : rowtext;
        cbuff : array [1..12] of isochar;
        newval, temp : smallint;
        c : coltype;
        cidx : 1..13;
        ok : boolean;
    begin
        { fetch reply text from input area }
        getrow(19, inline);  cidx := 1;
        for c := 50 to 61 do
        if inline[c] <> blank then
        begin
            cbuff[cidx] := inline[c];  cidx := cidx + 1
        end;
        for cidx := cidx to 12 do cbuff[cidx] := blank;
 
        { attempt to decode reply }
        if cbuff[1] <> blank then
        begin
            ok := true;
            if cbuff[2] = blank then newval := cbuff[1]
            else if cbuff[1] = cflex then
            begin
                if (cbuff[2] >= leta) and (cbuff[2] <= letz)
                    then newval := (cbuff[2] - leta) + 1
                else if (cbuff[2] >= capa) and (cbuff[2] <= capz)
                    then newval := (cbuff[2] - capa) + 1
                else ok := false;
                if cbuff[3] <> blank then ok := false
            end
            else if cbuff[1] = sharp then
            begin
                if decodehex(cbuff[2], newval) then
                begin
                    if cbuff[3] <> blank then
                    begin
                        if decodehex(cbuff[3], temp)
                            then newval := newval*16 + temp
                            else ok := false
                    end
                end
                else ok := false;
                if cbuff[4] <> blank then ok := false
            end
            else ok := false
        end
        else ok := false;
 
        { check the value against legal values }
        if ok then
        begin
            if lastmode >= kinthost
                then ok := (newval <> esc) and (newval <= del)
            else if lastmode = selectgr
                then ok := newval <= crossedout
            else if (lastmode >= reqshift) and (lastmode <= icharmode)
                then ok := newval <= 1
            else if lastmode = ilinerow
                then ok := newval <= editmode[tmaxrow]
            else if lastmode = intsignal
                then ok := newval < blank
        end;
 
        { reset mode element value if still ok }
        if ok then editmode[lastmode] := newval
              else soundalarm;
 
        { rewrite mode element display }
        putitem(lastmode, editmode[lastmode]);
        setupdate(0, 0, 0, 0);   { remove entry box }
        mdisstate;
        setfield(lastmode-3)     { restore cursor to last choice }
    end;  {modemodify}
 
 
    { boxcommon -- common entry for all editing tests }
    procedure boxcommon (pattern : boolean);
    var
        r, mrow : rowtype;
        c : coltype;
    begin
        appstate := boxes;  putmode(editmode);
        setmode(selectgr, defrend);  erasedisplay;
        setmode(selectgr, negative);
        literal(' Functions Test &       ');
        setmode(selectgr, defrend);
        setcursor(0, 22);
        literal('Last T-TOKEN request cod');
        literal('e:&                     ');
        if pattern then
        begin
            setmode(selectgr, defrend);
            mrow := maxrow;  if mrow > 21 then mrow := 21;
            for r := 2 to mrow do
            begin
                setcursor(r, 0);
                for c := 0 to maxcol do
                    character(((r+c) mod 94) + 33)
            end
        end
    end;  {boxcommon}
 
 
    { addfield -- adds a new field definition }
    procedure addfield (rowa, rowb : rowtype;
                        cola, colb : coltype;
                        first : boolean);
    var
        r : rowtype;
        c : coltype;
        f : fieldindex;
    begin
        { set up field }
        if first then erasefields else            { first }
        begin
            getfield(f);  setfield(f+1)       { subsequent }
        end;
        setupdate(rowa, rowb, cola, colb);
 
        { draw a box around fields surrounded by protected areas }
        if (rowa <> 0) and (rowb <> maxrow)
            and (cola <> 0) and (colb <> maxcol) then
        begin
            setmode(selectgr, negative);
            setcursor(rowa-1, cola-1);  character(plus);
            for c := cola to colb do character(minus);
            character(plus);
            for r := rowa to rowb do
            begin
                setcursor(r, cola-1);  character(bar);
                setcursor(r, colb+1);  character(bar)
            end;
            setcursor(rowb+1, cola-1);  character(plus);
            for c := cola to colb do character(minus);
            character(plus);
        end;
    end;  {addfield}
 
 
    { putescmsg -- tell the user how to escape }
    procedure putescmsg;
    begin
        setmode(selectgr, defrend);
        if maxrow >= 23 then setcursor(23, 0)
                        else setcursor(1, 0);
        genescmsg
    end;  {putescmsg}
 
 
    { firstsingle -- output image for single box editing test }
    procedure firstsingle (pattern : boolean);
    begin
        appstate := boxes;  boxcommon(pattern);
        addfield(4, 19, 5, maxcol-4, true);
        putescmsg;
        setmode(selectgr, editmode[selectgr]);  setcursor(4, 5)
    end;  {firstsingle}
 
 
    { firstmultiple -- output image for multiple boxes editing test }
    procedure firstmultiple (pattern : boolean);
    var
        c : coltype;
    begin
        appstate := boxes;  boxcommon(pattern);
 
        { five rectangular fields }
        addfield(3, 3, 3, 34, true);
        addfield(5, 7, 3, 34, false);
        addfield(9, 13, 3, 34, false);
        addfield(3, 5, 39, 70, false);
        addfield(7, 13, 39, 70, false);
 
        { construct a multiple-line field from three one-liners }
        addfield(16, 16, 10, maxcol, false);
        addfield(17, 17, 0, maxcol, false);
        addfield(18, 18, 0, maxcol-15, false);
 
        { draw round that last one (too much for addfield) }
        setcursor(15, 9);  character(plus);
        for c := 10 to maxcol do character(minus);
        setcursor(16, 0);  for c := 0 to 8 do character(minus);
        character(plus);
        setcursor(18, maxcol-14);  character(plus);
        for c := maxcol-13 to maxcol do character(minus);
        setcursor(19, 0);  for c := 0 to maxcol-15 do character(minus);
        character(plus);
 
        { messages, correct the rendition, select the first field }
        if editmode[cursor] <> 1 then
        begin
            setcursor(1, 0);
            literal(' For a better demonstrat');
            literal('ion with this screen set');
            literal(' Mode[CURSOR]=1 (ESC "5"');
            literal('). &                    ');
        end;
        putescmsg;
        setmode(selectgr, editmode[selectgr]);  setfield(0)
    end;  {firstmultiple}
 
 
    { getpfnbr -- get program function key number from request code }
    function getpfnbr (reply : smallint) : smallint;
    var
        initial, final : smallint;
    begin
        initial := reply;  final := 0;
        if (reply>=capa) and (reply<=capz) then initial := initial + 32;
 
        if (initial>=dig0) and (initial<=dig9) then
        begin
            if initial=dig0 then final := 10
                            else final := initial - dig0
        end
        else
        if (initial>=leta) and (initial<=letz) then
        begin
            case initial of
            letq:   final := 11;
            letw:   final := 12;
            lete:   final := 13;
            letr:   final := 14;
            lett:   final := 15;
            lety:   final := 16;
            letu:   final := 17;
            leti:   final := 18;
            leto:   final := 19;
            letp:   final := 20;
            leta:   final := 21;
            lets:   final := 22;
            letd:   final := 23;
            letf:   final := 24;
            letg:   ;
            leth:   ;
            letj:   ;
            letk:   ;
            letl:   ;
            letz:   ;
            letx:   ;
            letc:   ;
            letv:   ;
            letb:   ;
            letn:   ;
            letm:   ;
            end;
        end;
        getpfnbr := final
    end;  {getpfnbr}
 
 
    { checktoken -- token returned from edit screen }
    procedure checktoken (reply : smallint);
    var
        r : rowtype;
        c : coltype;
        pfkey : smallint;
    begin
        getloc(r, c);
        getmode(editmode);   { some elements may have changed }
        setcursor(0, 50); setmode(selectgr, defrend);
        putkey(reply, true);
 
        { check for menu request }
        if reply = 0 then newcode := 12
                     else newcode := menureply(reply);
        if newcode <= 11 then domore := true
        else
        begin  { continue }
            pfkey := getpfnbr(reply);
            if (pfkey>=11) and (pfkey<=20) then
            begin  { change of rendition }
                if pfkey = 20 then editmode[selectgr] := defrend
                              else editmode[selectgr] := pfkey - 10
            end;
            setmode(selectgr, editmode[selectgr]);  setcursor(r, c)
        end
    end;  {checktoken}
 
 
    { action -- application main menu reply processing }
    procedure action (reply : smallint);
    begin
        newcode := menureply(reply);
        if newcode = 10 then soundalarm       { already there }
                        else domore := true   { menu selection }
    end;  {action}
 
 
    { service -- application T-TOKEN processing }
    procedure service (rc : smallint);
    begin
        domore := false;
        case appstate of
 
        entry:
            if (maxrow < 23) or (maxcol < 79) then
            begin
                setcursor(maxrow, 0);
                literal('Needs 24 rows by 80 colu');
                literal('mns minimum screen size.');
                screenmode := false
            end
            else firstmenu;             { first entry to menu }
 
        menu:       action(rc);         { main menu selection }
        charset:    incharset(rc);      { graphic rendition selection }
        fhost:      hosttest;           { start of host primitive test }
 
        thost:      begin               { exit host primitive test }
                        if rc=129           { check for krestart }
                            then hostrend;  { do it properly here! }
                        dmsprocess(rc)
                    end;
 
        ooc:        begin               { exit ooc message test }
                        if rc=129           { check for krestart }
                            then rendinfo;  { explain why rends vanish }
                        dmsprocess(rc)
                    end;
 
        rogue:      dmsprocess(rc);     { exit rogue application test }
        soak:       dmsprocess(rc);     { exit soak test }
        dismode:    modeselect(rc);     { let user select an element }
        modmode:    modemodify(rc);     { accept modified value }
        boxes:      checktoken(rc);     { return from edit screen }
        end;  {case}
 
        { possible menu selection exit }
        if domore then
        begin
            case newcode of
 
            0:  firstcharset;           { character set display }
            1:  hosttest;               { host primitives test }
            2:  ooctest;                { out-of-context message test }
            3:  rogueapp;               { rogue application test }
            4:  soaktest;               { soak test }
            5:  firstdismode;           { mode array display }
            6:  firstsingle(false);     { single box }
            7:  firstsingle(true);      { single box with pattern }
            8:  firstmultiple(false);   { multiple box }
            9:  firstmultiple(true);    { multiple box with pattern }
            10: putmenu;                { "M"|"m" : Menu }
            11: screenmode := false;    { "S"|"s" : Stop }
            12: soundalarm;             { anything else }
 
            end  {case}
        end;
 
        { candle still burning?  return the token! }
        if screenmode then sendtoken(tokencode);
    end;  {service}
 
 
    { dispatch -- host application, fetch and dispatch T-primitives }
    procedure dispatch;
    var
        pc : isochar;
        pars : primpar;
    begin
        getprimitive(pc, pars);   { fetch next primitive }
        case pc of
 
        star:   ;   { T-CHARACTER(ch) }
        capa:       { T-TOKEN(reqcode) }
                service(pars[1]);         { *must* act on request code }
        capb:   ;   { T-SETCURSOR(newrow,newcol) }
        capc:   ;   { T-SETMODE(index,ivalue) }
        capd:   ;   { T-ERASETORIGHT }
        cape:   ;   { T-INSERTSPACE(nsp) }
        capf:   ;   { T-DELETECHAR(nch) }
        capg:   ;   { T-ERASEPREV(nch) }
        caph:   ;   { T-INSERTLINE(nln) }
        capi:   ;   { T-DELETELINE(nln) }
        capj:   ;   { T-APPENDLINE }
        capk:   ;   { T-SPLITLINE }
        capl:   ;   { T-SETFIELD(fidx) }
 
        end   {case}
    end;  {dispatch}
 
 
    { startapp -- application initialisation }
    procedure startapp;
    begin
        appstate := entry;  tokencode := 0
    end;  {startapp}
 
 
    { stopapp -- application closedown }
    procedure stopapp;
    begin
    end;  {stopapp}
 
 
begin  { termtest main routine }
 
    startssmp(screenmode);         { ssmp session entry negotiation }
    if screenmode then             { if successful, subsequent session }
    begin
        startapp;                      { application initialisation }
        while screenmode do dispatch;  { fetch and process primitives }
        stopapp;                       { application closedown }
        stopssmp                       { ssmp session exit }
    end;
 
end.   { main program }
