%separate = true
%ecs = true
%ttybuffered = true
{ VDUSETUP for Sheffield University Editor }
{ Copyright (c) University of Sheffield 1990 }

{************************************************}
{ MAIN PROGRAM AND INTERACTIVE CODE FOR VDUSETUP }
{************************************************}

program main(input, output);  { VDUSETUP }

%insert pmed_globals.ins.pas
%insert vdusetup_globals.ins.pas

{---extstart---}
{ Declarations for procedures in procs file }

   procedure to_up_case(var the_str: string);
      extern;
   procedure to_low_case(var the_str: string12);
      extern;
   procedure init_exp_record;
      extern;
   procedure empty_exp_record;
      extern;
   function compress_vdu_info: boolean;
      extern;
   procedure expand_vdu_info;
      extern;
   procedure dispose_key_sequence(var the_sequence: sequences);
      extern;
   procedure write_char_name(ch_code: integer; var chars_written: integer);
      extern;
   procedure write_key_sequence(to_report: boolean;
                                the_sequence: sequences;
                                var chars_written: integer);
      extern;
   procedure store_key_code(key_code: integer; var sequence: sequences);
      extern;
   function get_code_from_sequence(sequence: sequences; seq_no: integer): integer;
      extern;
   procedure append_sequence(from_seq: sequences; var to_seq: sequences);
      extern;
   procedure copy_sequence(from_seq: sequences; var to_seq: sequences);
      extern;
   function read_one_source(write_output_files: boolean): boolean;
      extern;
   function store_input_sequence(the_sequence: sequences;
                                 the_action: action_proc;
                                 the_repair: repair_type;
                                 the_ware: ware_type;
                                 the_reply: sequences): boolean;
      extern;
   procedure find_input_sequence(the_action: action_proc;
                                 var input_sequence: sequences);
      extern;
   function sequence_defined(the_sequence: sequences): boolean;
      extern;
   procedure delete_input_sequence(the_sequence: sequences;
                                   delete_inferiors: boolean);
      extern;
   procedure dl_in_set(the_set: set_of_action_proc);
      extern;
   procedure count_func(var count: integer; delete_them: boolean);
      extern;
   function check_action(word_str: string12; var action: action_proc): boolean;
      extern;
   procedure write_source_file;
      extern;
   procedure do_batch;
      extern;
   procedure list_action_names;
      extern;
{---extstop---}

procedure do_interactive;                                                       { DO_INTERACTIVE }

const
   menu_max = 7;
   ex_code_delay = 3.0;                 { seconds delay before executing code }
   read_delay = 5.0;                    { seconds delay for reading }

type
   cont_type = (rekord, half_duplex, xonrec, reinstate);
   keyword_type = string[22];
   ptr_message_line = ^message_line;
   message_line = record
      next_line: ptr_message_line;
      line_type: (normal_line, no_nl, ask_for_cont);
      line_text: string
   end;

   ptr_key2 = ^key2_type;
   key2_type = record
      next_key: ptr_key2;
      line_head: ptr_message_line;
      key_entry: keyword_type
   end;

   ptr_key1 = ^key1_type;
   key1_type = record
      next_key: ptr_key1;
      key2_head: ptr_key2;
      key_entry: keyword_type
   end;

   procs = (
      down_menu,
      up_menu,
      character_delete,
      character_insert,
      clear_screen,
      command_sequences,
      ctrl_ch,
      cursor_addressing,
      cursor_down,
      cursor_left,
      cursor_right,
      cursor_up,
      diag_print,
      exit,
      function_keys,
      general_input_keys,
      home_key,
      line_delete,
      line_erase,
      line_insert,
      long_highlight,
      make_changes,
      message_highlight,
      new_file,
      no_testing,
      old_file,
      page_erase,
      pointer_highlight,
      redisplay_menu,
      reverse_line_feed,
      save_binary,
      save_source,
      scroll_suppress,
      tabulation,
      term_reset,
      title);

   menu_range = 1..menu_max;
   ptr_menu = ^ menus;
   menu_line = record
      line_key: char;
      line_done: three_state;
      line_text: string;
      proc: procs;
      next_menu: ptr_menu
   end;
   menus = record
      menu_length: menu_range;
      menu_entries: array[menu_range] of menu_line
   end;


var
   message_head: ptr_key1;              { message index on heap }
   top_menu,                            { level 1 }
   make_changes_menu,                   { level 2 }
   essential_codes_menu,                { level 3 }
   output_codes_menu,
   input_codes_menu,
   cursor_movement_menu,                { level 4 }
   screen_modifying_menu,
   optional_extras_menu,
   highlights_menu: ptr_menu;
   testing_code, file_modified, must_set_x_codes: boolean;
   all_done: three_state;
   go_to_top: boolean;
   duplex_word: shortint;               { current duplex state of terminal }
                                        { saved between switching off echo }
                                        { and switching it on again }
   perform_tests: boolean;              { do tests if true }

{ get a yes or no answer to a question }

function yes_or_no(default_answer: three_state): boolean;                       { YES_OR_NO }

var
   answer: alfa;
   ch: char;
   ans: three_state;

begin { yes_or_no }
   ans := is_unknown;
   repeat
      while (input^ = ' ') and (not eoln(input))
      do get(input);
      readln(answer);
      ch := answer[1];
      if ch in ['Y','y']
      then ans := is_true
      else if ch in ['N','n','Q','q']
      then ans := is_false
      else if ch = ' '
      then ans := default_answer;
      if ans = is_unknown
      then write('Please answer Yes or No: ')
   until ans <> is_unknown;
   writeln;                             { confirmation }
   yes_or_no := ans = is_true
end { yes_or_no } ;

{ Find the start of a chain of message lines identified by two keys. }
{ Optionally create if not found }

procedure find_message(key1, key2: keyword_type;                                { FIND_MESSAGE }
                       create: boolean;
                       var key2_ptr: ptr_key2);

var
   state: search_type;
   key1_ptr: ptr_key1;

begin { find_message }
   key1_ptr := message_head;
   state := searching;
   repeat
      if key1_ptr = nil
      then state := not_found
      else if key1_ptr^.key_entry = key1
      then state := found
      else key1_ptr := key1_ptr^.next_key
   until state <> searching;
   if (state = not_found) and create then
   begin
      new(key1_ptr);
      with key1_ptr^ do
      begin
         key_entry := key1;
         next_key := message_head;
         key2_head := nil
      end;
      message_head := key1_ptr;
      state := found
   end;
   if state = not_found
   then key2_ptr := nil
   else
   begin
      key2_ptr := key1_ptr^.key2_head;
      state := searching;
      repeat
         if key2_ptr = nil
         then state := not_found
         else if key2_ptr^.key_entry = key2
         then state := found
         else key2_ptr := key2_ptr^.next_key
      until state <> searching;
      if (state = not_found) and create then
      begin
         new(key2_ptr);
         with key2_ptr^ do
         begin
            key_entry := key2;
            next_key := key1_ptr^.key2_head;
            line_head := nil
         end;
         key1_ptr^.key2_head := key2_ptr;
         state := found
      end;
      if state = not_found
      then key2_ptr := nil
   end
end { find_message } ;

{ Read the file of messages and store on the heap for rapid access }

procedure read_vdu_messages;                                                    { READ_VDU_MESSAGES }

var
   vdu_messages: text;
   all_done, done: boolean;
   line, rest, key1, key2: string;
   posin: integer;
   head, tail: ptr_message_line;
   key2_ptr: ptr_key2;
   line_ptr: ptr_message_line;

begin { read_vdu_messages }
   write('Reading VDU_MESSAGES ...');
   reset(vdu_messages,'VDU_MESSAGES');
   message_head := nil;
   new(head);                           { always have one dummy to ease adding }
                                        { in lines of text }
   readln(vdu_messages, line);
   all_done := false;
   repeat
      rest := substr(line, 4, length(line)-3);
      posin := pos(' ', rest);
      if posin = 0
      then posin := length(rest) + 1;
      key1 := substr(rest, 1, posin-1);
      if posin < length(rest)
      then key2 := substr(rest, posin+1, length(rest)-posin)
      else key2 := '';
      find_message(key1, key2, true, key2_ptr);
      tail := head;                     { start with one dummy message }
      head^.next_line := nil;
      done := false;
      repeat
         if eof(vdu_messages) then
         begin
            done := true;
            all_done := true
         end
         else
         begin
            readln(vdu_messages, line);
            if length(line) >= 2
            then if substr(line, 1, 2) = '$$'
                 then done := true;
            if not done then
            begin
               new(line_ptr);
               tail^.next_line := line_ptr;
               tail := line_ptr;
               with line_ptr^ do
               begin
                  next_line := nil;
                  if length(line) >= 2
                  then
                  begin
                     if line[length(line)] = '~' then
                     begin
                        line_type := no_nl;
                        line := substr(line, 1, length(line)-1)
                     end
                     else if line[1] = '~' then
                     begin
                        line_type := ask_for_cont;
                        line := substr(line, 2, length(line)-1)
                     end
                     else line_type := normal_line
                  end
                  else line_type := normal_line;
                  line_text := line
               end;
            end { process a line } ;
            key2_ptr^.line_head := head^.next_line
         end { not eof }
      until done;
   until all_done;
   dispose(head);                       { get rid of dummy }
   writeln
end { read_vdu_messages } ;

{ Display a message identified by two keys }

procedure display_message(key1, key2: string);                                  { DISPLAY_MESSAGE }

var
   key2_ptr: ptr_key2;
   line_ptr: ptr_message_line;
   done: boolean;

begin { display_message }
   find_message(key1, key2, false, key2_ptr);
   if key2_ptr = nil
   then writeln('Message ',key1,' ',key2,' not found.')
   else
   begin
      line_ptr := key2_ptr^.line_head;
      done := false;
      while (line_ptr <> nil) and (not done) do
      with line_ptr^ do
      begin
         write(line_text);
         case line_type of
            normal_line: writeln;
            no_nl: ;
            ask_for_cont: if not yes_or_no(is_true)
                          then done := true
         end { case } ;
         line_ptr := next_line
      end { while loop }
   end
end { display_message } ;

{ Wait a given number of seconds }

procedure delay(secs: real);                                                    { DELAY }

   procedure sleep$(millisecs: integer);
      extern;

begin { delay }
   sleep$(round(secs*1000))
end { delay } ;

{ List the titles of the vdu tables in a file }

procedure list_tables(print_title: string;                                      { LIST_TABLES }
                      var table_count: integer;
                      test_id: string;
                      var type_found: integer);

begin { list_tables }
   seek(binary_file, 1);
   get(binary_file);
   table_count := 0;
   type_found := 0;
   writeln;
   while not eof(binary_file) do
   begin
      table_count := table_count + 1;
      with binary_file^ do
      begin
         if table_count = 1 then
         begin
            writeln(print_title);
            writeln
         end;
         if vdu_identification = test_id then
         begin
            if type_found = 0
            then type_found := table_count
            else type_found := -1;      { duplicate }
            write('->  ')
         end
         else write('    ');
         writeln(table_count:3,'   ',vdu_identification)
      end;
      get(binary_file)
   end;
   if type_found < 0
   then type_found := 0;
   writeln
end { list_tables } ;

{ Convert a file of version 1 vdu_info records to version 2 }

procedure convert_1_to_2(vers1_fn, vers2_fn: string);                           { CONVERT_1_TO_2 }

type
   unchanged_type_a = array[1..2561] of shortint;
   unchanged_type_b = array[1..922]  of shortint;

   vdu_info1 = record                   { version 1 - bare bones }
      stored_version1: integer;
      vdu_id_length1: integer;
      unchangeda1: unchanged_type_a;
      unchangedb1: unchanged_type_b
   end;

   vdu_info2 = record                   { version 2 - bare bones }
      stored_version2: integer;
      vdu_id_length2: shortint;
      unchangeda2: unchanged_type_a;
      arrow_type1: boolean;
      unchangedb2: unchanged_type_b
   end;

var
   vers1_file: file of vdu_info1;
   vers2_file: file of vdu_info2;

begin { convert_1_to_2 }
   reset(vers1_file, vers1_fn);
   open(vers2_file, vers2_fn);          { ensure DAM file }
   {---teststart---}
{  writeln('Sizes of vdu_info records:');  {----}
{  writeln('Version 1: ',sizeof(vers1_file^):1);  {----}
{  writeln('Version 2: ',sizeof(vers2_file^):1);  {----}
   {---teststop---}
   while not eof(vers1_file) do
   with vers1_file^, vers2_file^ do
   begin
      stored_version2 := 2;
      vdu_id_length2 := vdu_id_length1;
      unchangeda2 := unchangeda1;
      arrow_type1 := false;
      unchangedb2 := unchangedb1;
      put(vers2_file);
      get(vers1_file)
   end;
   display_message('convert_1_to_2','after')
end { convert_1_to_2 } ;

procedure set_control_name(the_control: controls;                               { SET_CONTROL_NAME }
                           var control_name: string12);

var
   char_name: char12;
   len: 0..12;
   chindex: 1..12;
   ch: char;

begin { set_control_name }
   control_name := '            ';
   char_name := vdu_control_names[the_control];
   len := 0;
   for chindex := 1 to 12 do
   begin
      ch := char_name[chindex];
      control_name[chindex] := ch;
      if ch <> ' '
      then len := chindex
   end;
   control_name := substr(control_name, 1, len)
end { set_control_name } ;

{ Set or reset terminal characteristics }

procedure control_term(cont: cont_type; switch_on: boolean);                    { CONTROL_TERM }

var
   short_or_bits: record
      case boolean of
         true:  (int: shortint);
         false: (bits: packed array[1..16] of boolean)
      end;
   rubbish: shortint;

   function duplx$(newmode: shortint): shortint; extern;

begin { control_term }
   with short_or_bits do
   case cont of
      rekord: duplex_word := duplx$(-1);  { record old state }
      reinstate: rubbish := duplx$(duplex_word);  { reinstate }
      half_duplex:
         begin
            int := duplx$(-1);          { get current state }
            bits[1] := switch_on;       { set half duplex }
            bits[2] := switch_on;       { and no lf echo }
            rubbish := duplx$(int)
         end;
      xonrec:
         begin
            int := duplx$(-1);          { get current state }
            bits[3] := switch_on;       { allow XON/XOFF }
            rubbish := duplx$(int)
         end
   end { case}
end { control_term } ;

{ Look at a file and check for binary version number }

procedure check_binary(filename: string;                                        { CHECK_BINARY }
                       var exists, is_binary: boolean;
                       var binary_version: integer);

label 421;

var
   word_file: file of integer;

   procedure onunit(not_used: integer);
   begin { onunit }
      goto 421
   end { onunit } ;

begin { check_binary }
   exists := false;
   on('IO_ERROR', onunit);
   reset(word_file, filename);
   exists := true;
   if eof(word_file)
   then is_binary := false
   else
   begin
      binary_version := word_file^;
      is_binary := binary_version >= 0  { not characters }
   end;
421:
end { check_binary } ;

{ Send a full sequence to the screen }

procedure send_full_sequence(the_sequence: sequences;                           { SEND_FULL_SEQUENCE }
                             row: row0_range;
                             col: col0_range);

var
   index: integer;
   ch_code, last_code: integer;

   { output a cursor coordinate as a single ascii character }
   procedure ascii_plus(valu: integer);
   begin { ascii_plus }
      write(chr(valu + ch_code))
   end { ascii_plus } ;

   { similar for ADM12PLUS, which wraps around after column 80 }
   procedure adm12plus_row(rvalu, cvalu: integer);
   begin { adm12plus_row }
      if cvalu <= 80
      then ascii_plus(rvalu)
      else ascii_plus(rvalu+48)
   end { adm12plus_row } ;

   procedure adm12plus_col(rvalu, cvalu: integer);
   begin { adm12plus_col }
      if cvalu <= 80
      then ascii_plus(cvalu)
      else ascii_plus(cvalu-80)
   end { adm12plus_col } ;

   { output a cursor coordinate as its decimal value + offset }
   procedure dec_plus(valu: integer);
   begin { dec_plus }
      write(valu + ch_code:1)
   end { dec_plus } ;

   procedure sleep$(msecs: integer);
      extern;

begin { send_full_sequence }
   with the_sequence do
   begin
      last_code := 0;
      for index := 1 to seq_len do
      begin
         ch_code := get_code_from_sequence(the_sequence, index);
         if last_code < 0 then
         begin
            case last_code of           { last_code < 0 - output col or row }
               -1: ascii_plus(row);
               -2: ascii_plus(col);
               -3: dec_plus(row);
               -4: dec_plus(col);
               -5: adm12plus_row(row, col);
               -6: adm12plus_col(row, col)
               otherwise halt('unrecognised action code in vducursor_string')
            end {case} ;
            ch_code := 0                { to clear 'last_code < 0' switch }
         end
         else if ch_code < -100
         then ch_code := 0              { user control hook }
         else if ch_code = 0            { code 0 => wait 0.1 secs }
         then sleep$(100)
         else if ch_code > 0            { code > 0 => output ascii char }
         then write(chr(ch_code));
         last_code := ch_code
      end { for loop }
   end { with }
end { send_full_sequence } ;

{ Send a sequence of characters to the screen to try it out }

procedure send_key_sequence(the_sequence: sequences);                           { SEND_KEY_SEQUENCE }

begin { send_key_sequence }
   send_full_sequence(the_sequence, 0, 0)
end { send_key_sequence } ;

{ Position the cursor to a given row and column }

procedure vdu_position(row, col: integer);                                      { VDU_POSITION }

begin { vdu_position }
   send_full_sequence(expanded.exp_cursor_string, row, col)
end { vdu_position } ;

{ Send a terminal control using the expanded information }

procedure vdu_control(the_control: controls);                                   { VDU_CONTROL }

begin { vdu_control }
   send_full_sequence(expanded.exp_control_array[the_control], 0, 0)
end { vdu_control } ;

{ Clear the screen by using user-defined code, or output newlines if }
{ not defined yet }

procedure do_clear_screen;                                                      { DO_CLEAR_SCREEN }

begin { do_clear_screen }
   with expanded.exp_control_array[clearscreen] do
   begin
      if perform_tests and (seq_len > 0) then
      begin
         vdu_control(clearscreen);
         vdu_position(1, 1)             { cannot assume that clear screen will }
                                        { leave cursor at 'home' position }
      end
      else writeln
   end
end { do_clear_screen } ;

{ Check whether vdu description is modified }

function alright_to_replace: boolean;                                           { ALRIGHT_TO_REPLACE }

var
   result: boolean;

begin { alright_to_replace }
   if file_modified then
   begin
      display_message('alright_to_replace','');
      if yes_or_no(is_unknown) then
      begin
         result := true;
         file_modified := false
      end
      else result := false
   end
   else result := true;
   if result and perform_tests
   then send_key_sequence(expanded.exp_exit_string);
   alright_to_replace := result
end { alright_to_replace } ;

{ Draw a box round the screen }

procedure draw_screen_box(do_brh: boolean);                                     { DRAW_SCREEN_BOX }

var
   row, col: integer;

begin { draw_screen_box }
   with vdu_table_rec, expanded do
   begin
      vdu_position(1, 1);
      write('+');
      row := 1;
      for col := 2 to cols_per_row-1 do
      begin
         vdu_position(row, col);
         write('-')
      end;
      vdu_position(1, cols_per_row);
      write('+');
      col := cols_per_row;
      for row := 2 to rows_per_page-1 do
      begin
         vdu_position(row, col);
         write('|')
      end;
      if do_brh then
      begin { do bottom rh corner }
         if exp_control_array[scrolloff].seq_len > 0 then
         begin { use scroll off/on }
            vdu_control(scrolloff);
            vdu_position(rows_per_page, cols_per_row);
            write('+');
            vdu_control(scrollon)
         end { use scroll off/on }
         else if exp_control_array[charinsert].seq_len > 0 then
         begin { use character insert }
            vdu_position(rows_per_page, cols_per_row-1);
            write('+');
            vdu_control(cursorleft);
            vdu_control(charinsert)
         end { use character insert }
         else { miss it out }
      end { do bottom rh corner } ;
      row := rows_per_page;
      for col := cols_per_row-1 downto 2 do
      begin
         vdu_position(row, col);
         write('-')
      end;
      vdu_position(rows_per_page, 1);
      write('+');
      col := 1;
      for row := rows_per_page-1 downto 2 do
      begin
         vdu_position(row, col);
         write('|')
      end
   end
end { draw_screen_box } ;

procedure create_menus;                                                         { CREATE_MENUS }

begin { create_menus }
   new(top_menu);
   new(make_changes_menu);
   new(essential_codes_menu);
   new(output_codes_menu);
   new(input_codes_menu);
   new(cursor_movement_menu);
   new(screen_modifying_menu);
   new(optional_extras_menu);
   new(highlights_menu);

   with top_menu^ do
   begin
      menu_length := 7;
      with menu_entries[1] do
      begin
         line_key := '1';
         line_text := 'Start a new terminal description';
         proc := new_file; next_menu := nil
      end;
      with menu_entries[2] do
      begin
         line_key := '2';
         line_text := 'Load an existing terminal description';
         proc := old_file; next_menu := nil
      end;
      with menu_entries[3] do
      begin
         line_key := '3';
         line_text := 'Make changes to terminal description';
         proc := make_changes; next_menu := nil
      end;
      with menu_entries[4] do
      begin
         line_key := '4';
         line_text := 'Make changes to terminal description without testing';
         proc := no_testing; next_menu := nil
      end;
      with menu_entries[5] do
      begin
         line_key := '5';
         line_text := 'Save terminal description in binary form for editor';
         proc := save_binary; next_menu := nil
      end;
      with menu_entries[6] do
      begin
         line_key := '6';
         line_text := 'Save terminal description in source form';
         proc := save_source; next_menu := nil
      end;
      with menu_entries[7] do
      begin
         line_key := 'Q';
         line_text := 'Quit';
         proc := up_menu; next_menu := nil
      end
   end;

   { Note that 'display_menu' assumes that entry 1 is the essentials menu }

   with make_changes_menu^ do
   begin
      menu_length := 3;
      with menu_entries[1] do
      begin
         line_key := '1';
         line_text := 'Essential terminal information';
         proc := down_menu; next_menu := essential_codes_menu
      end;
      with menu_entries[2] do
      begin
         line_key := '2';
         line_text := 'Output codes';
         proc := down_menu; next_menu := output_codes_menu
      end;
      with menu_entries[3] do
      begin
         line_key := '3';
         line_text := 'Input codes';
         proc := down_menu; next_menu := input_codes_menu
      end
   end;
   with essential_codes_menu^ do
   begin
      menu_length := 6;
      with menu_entries[1] do
      begin
         line_key := '1';
         line_text := 'Title for terminal';
         proc := title; next_menu := nil
      end;
      with menu_entries[2] do
      begin
         line_key := '2';
         line_text := 'Clear the screen';
         proc := clear_screen; next_menu := nil
      end;
      with menu_entries[3] do
      begin
         line_key := '3';
         line_text := 'Reset terminal sequence';
         proc := term_reset; next_menu := nil
      end;
      with menu_entries[4] do
      begin
         line_key := '4';
         line_text := 'Exit from screen editor sequence';
         proc := exit; next_menu := nil
      end;
      with menu_entries[5] do
      begin
         line_key := '5';
         line_text := 'Cursor addressing';
         proc := cursor_addressing; next_menu := nil
      end;
      with menu_entries[6] do
      begin
         line_key := '6';
         line_text := 'Cursor movement';
         proc := down_menu; next_menu := cursor_movement_menu
      end
   end;
   with output_codes_menu^ do
   begin
      menu_length := 4;
      with menu_entries[1] do
      begin
         line_key := '1';
         line_text := 'Screen modifying codes';
         proc := down_menu; next_menu := screen_modifying_menu
      end;
      with menu_entries[2] do
      begin
         line_key := '2';
         line_text := 'Optional extras';
         proc := down_menu; next_menu := optional_extras_menu
      end;
      with menu_entries[3] do
      begin
         line_key := '3';
         line_text := 'Highlighting codes';
         proc := down_menu; next_menu := highlights_menu
      end;
      with menu_entries[4] do
      begin
         line_key := '4';
         line_text := 'The ''bottom right-hand corner'' problem';
         proc := scroll_suppress; next_menu := nil
      end
   end;
   with input_codes_menu^ do
   begin
      menu_length := 4;
      with menu_entries[1] do
      begin
         line_key := '1';
         line_text := 'Editor command sequences';
         proc := command_sequences; next_menu := nil
      end;
      with menu_entries[2] do
      begin
         line_key := '2';
         line_text := 'General terminal keys';
         proc := general_input_keys; next_menu := nil
      end;
      with menu_entries[3] do
      begin
         line_key := '3';
         line_text := 'Function keys';
         proc := function_keys; next_menu := nil
      end;
      with menu_entries[4] do
      begin
         line_key := '4';
         line_text := 'The HOME key';
         proc := home_key; next_menu := nil
      end
   end;
   with cursor_movement_menu^ do
   begin
      menu_length := 4;
      with menu_entries[1] do
      begin
         line_key := '1';
         line_text := 'Cursor right';
         proc := cursor_right; next_menu := nil
      end;
      with menu_entries[2] do
      begin
         line_key := '2';
         line_text := 'Cursor left';
         proc := cursor_left; next_menu := nil
      end;
      with menu_entries[3] do
      begin
         line_key := '3';
         line_text := 'Cursor up';
         proc := cursor_up; next_menu := nil
      end;
      with menu_entries[4] do
      begin
         line_key := '4';
         line_text := 'Cursor down';
         proc := cursor_down; next_menu := nil
      end
   end;
   with screen_modifying_menu^ do
   begin
      menu_length := 5;
      with menu_entries[1] do
      begin
         line_key := '1';
         line_text := 'Line erase';
         proc := line_erase; next_menu := nil
      end;
      with menu_entries[2] do
      begin
         line_key := '2';
         line_text := 'Line insert';
         proc := line_insert; next_menu := nil
      end;
      with menu_entries[3] do
      begin
         line_key := '3';
         line_text := 'Line delete';
         proc := line_delete; next_menu := nil
      end;
      with menu_entries[4] do
      begin
         line_key := '4';
         line_text := 'Character insert';
         proc := character_insert; next_menu := nil
      end;
      with menu_entries[5] do
      begin
         line_key := '5';
         line_text := 'Character delete';
         proc := character_delete; next_menu := nil
      end
   end;
   with optional_extras_menu^ do
   begin
      menu_length := 4;
      with menu_entries[1] do
      begin
         line_key := '1';
         line_text := 'Page erase';
         proc := page_erase; next_menu := nil
      end;
      with menu_entries[2] do
      begin
         line_key := '2';
         line_text := 'Tabulation';
         proc := tabulation; next_menu := nil
      end;
      with menu_entries[3] do
      begin
         line_key := '3';
         line_text := 'Displaying control characters';
         proc := ctrl_ch; next_menu := nil
      end;
      with menu_entries[4] do
      begin
         line_key := '4';
         line_text := 'Reverse line feed';
         proc := reverse_line_feed; next_menu := nil
      end
   end;
   with highlights_menu^ do
   begin
      menu_length := 3;
      with menu_entries[1] do
      begin
         line_key := '1';
         line_text := 'Editor messages';
         proc := message_highlight; next_menu := nil
      end;
      with menu_entries[2] do
      begin
         line_key := '2';
         line_text := 'Line marker';
         proc := pointer_highlight; next_menu := nil
      end;
      with menu_entries[3] do
      begin
         line_key := '3';
         line_text := 'Folded parts of long lines';
         proc := long_highlight; next_menu := nil
      end
   end;
end { create_menus } ;

{ Ask the user for a single character }

procedure get_single_char(var ch_code: integer);                                { GET_SINGLE_CHAR }

var
   char_ok, done: boolean;
   answer: string[8];
   ch: char;
   aindex, chars_written: integer;

   procedure t1in(var ch: char); extern;

begin { get_single_char }
   char_ok := false;
   repeat
      display_message('get_single_char','typecode');
      done := false;
      repeat
         readln(answer);
         if length(answer) > 3
         then writeln('Too many digits.')
         else if length(answer) = 0
         then writeln('Too few digits.')
         else
         begin { number of digits alright }
            ch_code := 0;
            for aindex := 1 to length(answer) do
            begin
               ch := answer[aindex];
               if ch in ['0'..'7']
               then ch_code := ch_code*8 + ord(ch) - ord('0')
               else writeln('Non-octal character.')
            end { for loop } ;
            if (ch_code >= 0) and (ch_code <= 255)
            then done := true
            else writeln('Ascii code out of range.')
         end { number of digits alright } ;
         if not done
         then write('Please try again: ')
      until done;
      writeln;
      write('The character is ASCII ');
      if ch_code = ord(' ')
      then writeln('space')
      else
      begin { not space }
         write('''');
         write_char_name(ch_code, chars_written);
         writeln('''')
      end { not space } ;
      writeln;
      write('Is the character correct? ');
      char_ok := yes_or_no(is_unknown)
   until char_ok
end { get_single_char } ;

procedure get_key_sequence(var the_sequence: sequences;                         { GET_KEY_SEQUENCE }
                           allow_null: boolean);

var
   sequence_ok, done, last_was_slash: boolean;
   ch: char;
   chars_written, ch_code: integer;

   { Get a single character, even if it is CTRL-P! }

   procedure in_single(var ch: char);

   label 431;

      procedure t1in(var ch: char); extern;

      procedure quit_onunit(not_used: integer);
      begin { quit_onunit }
         ch := chr(dle_code);           { CTRL-P is DLE }
         goto 431
      end { quit_onunit } ;

   begin { in_single }
      on('QUIT$', quit_onunit);
      t1in(ch);
   431:
      { cannot allow parity-off characters in input sequences because array }
      { is only 200b .. 377b }
      if ord(ch) < 200b
      then ch := chr(ord(ch) + 200b)
   end { in_single } ;

begin { get_key_sequence }
   file_modified := true;               { getting new key sequence }
   sequence_ok := false;
   with the_sequence do
   repeat
      display_message('get_key_sequence','type');
      if allow_null
      then writeln('Just give the // if you want to specify a null sequence.');
      writeln;
      control_term(half_duplex, true);  { echoing off }
      control_term(xonrec, false);      { XON recognition off }
      done := false;
      last_was_slash := false;
      the_sequence.seq_len := 0;
      repeat
         in_single(ch);
         ch_code := ord(ch);
         write_char_name(ch_code, chars_written);
         write(' ');
         if ch = '/'
         then if last_was_slash
              then done := true
              else last_was_slash := true
         else
         begin { not '/' }
            if last_was_slash then
            begin { store delayed '/' }
               store_key_code(ord('/'), the_sequence);
               last_was_slash := false
            end { store delayed '/' } ;
            store_key_code(ch_code, the_sequence)
         end { not '/' }
      until done;
      writeln;
      control_term(reinstate, false);   { restore terminal state }
      writeln;
      if (seq_len > 0) or allow_null then
      begin
         writeln('The sequence received was: ');
         write_key_sequence(false, the_sequence, chars_written);
         writeln;
         writeln;
         sequence_ok := true
      end
      else display_message('get_key_sequence','zero')
   until sequence_ok
end { get_key_sequence } ;

{ Test whether a key sequence contains any DLE (CTRL-P) characters }

function test_for_dle(the_sequence: sequences): boolean;                        { TEST_FOR_DLE }

var
   sindex: integer;

begin { test_for_dle }
   test_for_dle := false;
   with the_sequence do
   begin
      for sindex := 1 to seq_len
      do if get_code_from_sequence(the_sequence, sindex) = dle_code
         then test_for_dle := true
   end
end { test_for_dle } ;

{ Get the type of an action: software or hardware }

procedure get_ware(the_action: action_proc; var the_ware: ware_type);           { GET_WARE }

begin { get_ware }
   if action_proc_info[the_action].always_soft
   then the_ware := software            { this one cannot be hardware }
   else
   begin
      display_message('get_ware','ques');
      if yes_or_no(is_unknown)
      then the_ware := hardware
      else the_ware := software
   end
end { get_ware } ;

{ Ask the user for a reply sequence }

procedure get_reply(the_sequence: sequences;                                    { GET_REPLY }
                    var the_reply: sequences;
                    explain: boolean);

var
   done: boolean;

begin { get_reply }
   if explain then
   begin
      write('Do you want an explanation of reply sequences? ');
      if yes_or_no(is_true)
      then display_message('get_reply','explain');
   end;
   display_message('get_reply','intro');
   if yes_or_no(is_false) then
   begin
      done := false;
      repeat
         display_message('get_reply','type1');
         get_key_sequence(the_reply, false);
         if perform_tests then
         begin
            display_message('get_reply','test');
            write('Sending input sequence: ');
            send_key_sequence(the_sequence);
            delay(ex_code_delay);
            write('Sending reply sequence: ');
            send_key_sequence(the_reply);
            writeln;
            write('Did it work? ');
            done := yes_or_no(is_unknown)
         end
         else done := true
      until done
   end
   else the_reply := init_sequence
end { get_reply } ;

{ Test the effect of a character sequence on the screen }

procedure get_repair(the_sequence: sequences;                                   { GET_REPAIR }
                     add_one: boolean;
                     var the_repair: repair_type;
                     the_reply: sequences);

var
   chars_gone, max_gone, lindex: integer;
   done: boolean;

begin { get_repair }
   if perform_tests then
   begin
      display_message('get_repair','intro');
      for lindex := 1 to 10
      do vdu_control(cursorleft);
      delay(ex_code_delay);
      send_key_sequence(the_sequence);
      send_key_sequence(the_reply);
      display_message('get_repair','howmany1')
   end
   else display_message('get_repair','howmany2');
   repeat
      done := true;
      readln(chars_gone);
      if add_one and (chars_gone >= 0)
      then chars_gone := chars_gone + 1;
      case chars_gone of
         0: the_repair := ok;
         1: the_repair := rest1;
         2: the_repair := rest2;
         3: the_repair := rest3;
         4: the_repair := rest4
         otherwise
            begin
               if add_one
               then max_gone := 3
               else max_gone := 4;
               write('Must be in the range 0 to ',max_gone:1,': ');
               done := false
            end
      end { case }
   until done
end { get_repair } ;

{ Get the name of an action_proc from the user }

procedure get_action(var the_action: action_proc);                              { GET_ACTION }

var
   done: boolean;
   the_name: string12;

begin { get_action }
   done := false;
   repeat
      write('Give an action name or press RETURN for a list: ');
      readln(the_name);
      if length(the_name) = 0
      then list_action_names
      else if check_action(the_name, the_action) then
      begin
         writeln;
         to_low_case(the_name);
         display_message('get_action',the_name);
         writeln;
         write('Is this the correct action? ');
         done := yes_or_no(is_unknown)
      end
      else writeln('Not recognised.')
   until done
end { get_action } ;

{ Get a simple string of characters }

procedure get_simple_string(var line_done: three_state;                         { GET_SIMPLE_STRING }
                            var the_string: sequences;
                            name_for_user: string12;
                            send_it: boolean);

var
   get_sequence, done: boolean;
   chars_written: integer;

begin { get_simple_string }
   if line_done = is_unknown then
   begin
      if the_string.seq_len > 0
      then line_done := is_true
      else line_done := is_false
   end
   else
   begin
      line_done := is_true;
      if the_string.seq_len > 0 then
      begin
         writeln('Current terminal ',name_for_user,' string is');
         write_key_sequence(false, the_string, chars_written);
         writeln;
         writeln;
         write('Do you want to redefine it? ');
         get_sequence := yes_or_no(is_unknown)
      end
      else get_sequence := true;
      if get_sequence then
      repeat
         display_message(name_for_user,'intro');
         get_key_sequence(the_string, true);
         if send_it and perform_tests and (the_string.seq_len > 0) then
         begin
            write('Sending sequence to terminal...');
            delay(ex_code_delay);
            send_key_sequence(the_string)
         end;
         writeln;
         write('Is the sequence alright? ');
         done := yes_or_no(is_unknown)
      until done
   end { not just testing }
end { get_simple_string } ;

{ Get and test a single vdu control code such as char insert }

procedure single_code(var line_done: three_state;                               { SINGLE_CODE }
                      the_control: controls;
                      the_action: action_proc;
                      must_exist: boolean;
                      the_ware: ware_type;
                      name_for_user: string;
                      procedure do_test_proc);

var
   read_seq, get_seq, ask_test, do_test, exit_from_proc, result: boolean;
   control_name: string12;
   chars_written: integer;

begin { single_code }
   with expanded do
   if line_done = is_unknown then
   begin { just checking }
      if exp_control_array[the_control].seq_len > 0
      then line_done := is_true
      else line_done := is_false
   end { just checking }
   else
   begin
      line_done := is_false;
      set_control_name(the_control, control_name);
      delete_input_sequence(exp_control_array[the_control], false);
      if exp_control_array[the_control].seq_len > 0 then
      begin
         writeln('Current setting is:');
         write('   ',vdu_control_names[the_control],':  ');
         write_key_sequence(false, exp_control_array[the_control], chars_written);
         writeln;
         write('Do you want to redefine it? ');
         read_seq := yes_or_no(is_unknown)
      end
      else read_seq := true;
      ask_test := not read_seq;
      exit_from_proc := false;
      repeat
         if read_seq then
         begin
            display_message(control_name,'intro');
            if must_exist
            then get_seq := true
            else
            begin
               write('Can your terminal do ''',name_for_user,'''? ');
               get_seq := yes_or_no(is_unknown)
            end;
            if get_seq then
            begin
               display_message(control_name,'type');
               get_key_sequence(exp_control_array[the_control], false)
            end
            else
            begin
               exp_control_array[the_control].seq_len := 0;
               exit_from_proc := true;
               line_done := is_true
            end
         end { read sequence }
         else read_seq := true;         { read it next time if any }
         if not perform_tests then
         begin
            exit_from_proc := true;
            line_done := is_true
         end
         else if not exit_from_proc then
         begin
            if ask_test then
            begin
               write('Do you want to test it? ');
               do_test := yes_or_no(is_unknown);
               ask_test := false
            end
            else do_test := true;
            if do_test then
            begin
               if the_control <> clearscreen
               then do_clear_screen;
               display_message(control_name,'test');
               writeln;
               do_test_proc;
               writeln;
               writeln;
               write('Did it work? ');
               if yes_or_no(is_unknown) then
               begin
                  exit_from_proc := true;
                  line_done := is_true
               end
            end { do test }
            else
            begin
               exit_from_proc := true;
               line_done := is_true
            end;
            if not exit_from_proc then
            begin
               exp_control_array[the_control].seq_len := 0;
               do_clear_screen;
               write('Try again? ');
               exit_from_proc := not yes_or_no(is_unknown)
            end
         end { terminal can do it }
      until exit_from_proc;
      if exp_control_array[the_control].seq_len > 0 then
      begin
         delete_input_sequence(exp_control_array[the_control], true);
         result := store_input_sequence(exp_control_array[the_control],
                   the_action, ok, the_ware, init_sequence)
      end
   end { not just testing }
end { single_code } ;

{ Get and test a cursor movement arrow code }

procedure cursor_arrow_code(var line_done: three_state;                         { CURSOR_ARROW_CODE }
                            the_control: controls;
                            the_action: action_proc;
                            name_for_user: string;
                            procedure do_test_proc;
                            procedure do_margin_proc;
                            var margin_bool: boolean);

var
   read_seq, ask_test, do_test, exit_from_proc, result, done,
   tested : boolean;
   control_name: string12;
   chars_written: integer;
   input_sequence: sequences;
   the_ware: ware_type;

begin { cursor_arrow_code }
   with expanded do
   if line_done = is_unknown then
   begin { just checking }
      if exp_control_array[the_control].seq_len > 0
      then line_done := is_true
      else line_done := is_false
   end { just checking }
   else
   begin
      input_sequence := init_sequence;
      line_done := is_false;
      set_control_name(the_control, control_name);
      if exp_control_array[the_control].seq_len > 0 then
      begin
         writeln('Current settings for ',name_for_user,' are:');
         write('   Output sequence:  ');
         write_key_sequence(false, exp_control_array[the_control], chars_written);
         writeln;
         writeln('   Margin alright = ',margin_bool);
         writeln;
         write('Do you want to redefine them? ');
         read_seq := yes_or_no(is_unknown)
      end
      else read_seq := true;
      ask_test := not read_seq;
      exit_from_proc := false;
      repeat
         if read_seq then
         begin
            must_set_x_codes := true;
            delete_input_sequence(exp_control_array[the_control], true);
            display_message(control_name,'intro');
            display_message(control_name,'type');
            repeat
               get_key_sequence(input_sequence, the_control=cursordown);
               if input_sequence.seq_len = 0 then
               begin { cursor down is lf }
                  display_message(control_name,'lf');
                  get_key_sequence(input_sequence, false);
                  exp_control_array[the_control].seq_len := 0;
                  store_key_code(lf_code, exp_control_array[the_control]);
                  the_ware := software
               end { cursor down is lf }
               else
               begin { sequence received }
                  copy_sequence(input_sequence, exp_control_array[the_control]);
                  if test_for_dle(input_sequence) then
                  begin
                     done := false;
                     repeat
                        display_message('cursor_arrow_code','dle');
                        get_key_sequence(input_sequence, false);
                        done := not test_for_dle(input_sequence)
                     until done;
                     the_ware := software
                  end { sequence contains dle }
                  else the_ware := hardware
               end { sequence received }
            until input_sequence.seq_len > 0;
            delete_input_sequence(input_sequence, true);
            result := store_input_sequence(input_sequence,
                      the_action, ok, the_ware, init_sequence);
            delay(ex_code_delay)
         end { read sequence }
         else read_seq := true;         { read it next time if any }
         if not perform_tests then
         begin
            exit_from_proc := true;
            line_done := is_true
         end
         else if not exit_from_proc then
         begin
            if ask_test then
            begin
               write('Do you want to test them? ');
               do_test := yes_or_no(is_unknown);
               ask_test := false
            end
            else do_test := true;
            if do_test then
            begin
               repeat
                  tested := true;
                  do_clear_screen;
                  display_message(control_name,'test');
                  writeln;
                  do_test_proc;
                  writeln;
                  writeln;
                  write('Did it work? ');
                  if yes_or_no(is_unknown) then
                  begin
                     exit_from_proc := true;
                     line_done := is_true;
                     do_clear_screen;
                     display_message(control_name,'margin');
                     writeln;
                     do_margin_proc;
                     display_message(control_name,'ques');
                     margin_bool := yes_or_no(is_unknown)
                  end
                  else
                  begin { did not work }
                     if the_control = cursordown then
                     begin { cursor down }
                        display_message('cursordown','lfmod');
                        if yes_or_no(is_unknown) then
                        begin { cursor down is lf }
                           display_message('cursordown','lfout');
                           exp_control_array[the_control].seq_len := 0;
                           store_key_code(lf_code,
                                          exp_control_array[the_control]);
                           tested := false  { try again }
                        end { cursor down is lf }
                     end { cursor down }
                  end { did not work }
               until tested
            end { do test }
            else
            begin
               exit_from_proc := true;
               line_done := is_true
            end;
            if not exit_from_proc then
            begin
               exp_control_array[the_control].seq_len := 0;
               do_clear_screen;
               write('Try again? ');
               exit_from_proc := not yes_or_no(is_unknown)
            end
         end { testing }
      until exit_from_proc;
      dispose_key_sequence(input_sequence)
   end { not just testing }
end { cursor_arrow_code } ;

{ Do a test for a highlighting code }

procedure test_highlight(on_control, off_control: controls);                    { TEST_HIGHLIGHT }

begin { test_highlight }
   writeln;
   write  ('Not highlighted');
   vdu_control(on_control);
   write  ('Highlighted');
   vdu_control(off_control);
   writeln('Not highlighted');
   writeln
end { test_highlight } ;

{ Get and test a pair of control codes such as inverse video on/off }

procedure onoff_code(var line_done: three_state;                                { ONOFF_CODE }
                     on_control, off_control: controls;
                     null_allowed: boolean;
                     name_for_user: string;
                     procedure do_test_proc(switchon, switchoff: controls);
                     gap_allowed: boolean;
                     var there_is_a_gap: boolean);

var
   read_seq, ask_test, do_test, exit_from_proc: boolean;
   on_control_name, off_control_name: string12;
   chars_written: integer;

begin { onoff_code }
   with expanded do
   if line_done = is_unknown then
   begin { just checking }
      if (exp_control_array[ on_control].seq_len > 0) or
         (exp_control_array[off_control].seq_len > 0)
      then line_done := is_true
      else line_done := is_false
   end { just checking }
   else
   begin
      line_done := is_false;
      set_control_name(on_control, on_control_name);
      set_control_name(off_control, off_control_name);
      if (exp_control_array[on_control].seq_len > 0) or
         (exp_control_array[off_control].seq_len > 0)
      then
      begin
         writeln('Current settings are:');
         write('   ',vdu_control_names[on_control],':  ');
         write_key_sequence(false, exp_control_array[on_control], chars_written);
         writeln;
         write('   ',vdu_control_names[off_control],':  ');
         write_key_sequence(false, exp_control_array[off_control], chars_written);
         writeln;
         if gap_allowed
         then if there_is_a_gap
              then writeln('(Highlight occupies a space on screen.)')
              else writeln('(Highlight does not occupy a space on screen.)');
         write('Do you want to redefine them? ');
         read_seq := yes_or_no(is_unknown)
      end
      else read_seq := true;
      ask_test := not read_seq;
      exit_from_proc := false;
      repeat
         if read_seq then
         begin
            display_message(on_control_name,'intro');
            write('Can your terminal do ',name_for_user,'? ');
            if yes_or_no(is_unknown) then
            begin
               display_message(on_control_name,'type1');
               get_key_sequence(exp_control_array[on_control], false);
               display_message(on_control_name,'type2');
               get_key_sequence(exp_control_array[off_control], null_allowed)
            end
            else
            begin { not implemented }
               exp_control_array[on_control].seq_len := 0;
               exp_control_array[off_control].seq_len := 0;
               exit_from_proc := true;
               line_done := is_true
            end { not implemented }
         end { read sequence }
         else read_seq := true;         { read it next time if any }
         if not perform_tests then
         begin
            exit_from_proc := true;
            line_done := is_true
         end
         else if not exit_from_proc then
         begin
            if ask_test then
            begin
               writeln;
               write('Do you want to test them? ');
               do_test := yes_or_no(is_unknown);
               ask_test := false
            end
            else do_test := true;
            if do_test then
            begin
               do_clear_screen;
               display_message(on_control_name,'test');
               writeln;
               do_test_proc(on_control, off_control);
               write('Did it work? ');
               if yes_or_no(is_unknown) then
               begin
                  display_message(on_control_name, 'is_gap');
                  there_is_a_gap := yes_or_no(is_unknown);
                  if (not gap_allowed) and there_is_a_gap
                  then display_message(on_control_name,'gap')
                  else
                  begin
                     exit_from_proc := true;
                     line_done := is_true
                  end
               end { it worked }
               else do_clear_screen
            end { do test }
            else
            begin
               exit_from_proc := true;
               line_done := is_true
            end;
            if not exit_from_proc then
            begin
               exp_control_array[on_control].seq_len := 0;
               exp_control_array[off_control].seq_len := 0;
               write('Try again? ');
               exit_from_proc := not yes_or_no(is_unknown)
            end
         end { terminal can do it }
      until exit_from_proc
   end { not just testing }
end { onoff_code } ;

{ Set the extended cursor arrow input sequences }

procedure set_x_codes;                                                          { SET_X_CODES }

var
   the_repair: repair_type;
   input_sequence, x_input_sequence: sequences;
   result: boolean;

   procedure set_x(the_action, x_action: action_proc);
   begin { set_x }
      with expanded do
      begin
         find_input_sequence(the_action, input_sequence);
         if input_sequence.seq_len > 0 then
         begin { it is defined }
            copy_sequence(exp_poscommand, x_input_sequence);
            append_sequence(input_sequence, x_input_sequence);
            if not store_input_sequence(x_input_sequence, x_action,
                                        the_repair, hardware,
                                        exp_poscommand_reply)
            then writeln('Failed to store input sequence.')
         end { it is defined }
      end
   end { set_x } ;

begin { set_x_codes }
   with vdu_table_rec, expanded do
   begin
      if (cols_per_row = 1) and
         (rows_per_page = 1)
      then must_set_x_codes := false    { SSMP terminal - do not tinker }
      else
      begin
         input_sequence := init_sequence;
         x_input_sequence := init_sequence;
         case exp_poscommand_repair of
            rest2: the_repair := rest1;
            rest3: the_repair := rest2;
            rest4: the_repair := rest3
            otherwise the_repair := ok
         end;
         if must_set_x_codes then
         begin
            must_set_x_codes := false;  { stored now }
            dl_in_set([goupx, godownx, goleftx, gorightx,
                       lfcharx, revlfcharx, carret]);
            { delete BS if defined }
            store_key_code(bs_code, input_sequence);
            delete_input_sequence(input_sequence, true);
            if exp_poscommand.seq_len > 0 then
            begin
               set_x(goup     , goupx     );
               set_x(godown   , godownx   );
               set_x(goleft   , goleftx   );
               set_x(goright  , gorightx  );
               set_x(lfchar   , lfcharx   );
               set_x(revlfchar, revlfcharx)
            end;
            { add BS and DEL BS }
            input_sequence.seq_len := 0;
            store_key_code(bs_code, input_sequence);
            result := store_input_sequence(input_sequence, goleft, ok, hardware,
                                           init_sequence);
            copy_sequence(exp_poscommand, x_input_sequence);
            append_sequence(input_sequence, x_input_sequence);
            result := store_input_sequence(x_input_sequence, goleftx,
                                           the_repair, hardware,
                                           exp_poscommand_reply)

         end { set x codes } ;
         input_sequence.seq_len := 0;
         store_key_code(212b, input_sequence);  { linefeed }
         delete_input_sequence(input_sequence, true);
         result := store_input_sequence(input_sequence, carret, ok, hardware,
                                        init_sequence);
         dispose_key_sequence(input_sequence);
         dispose_key_sequence(x_input_sequence)
      end { not SSMP terminal }
   end { with }
end { set_x_codes } ;

procedure character_delete_proc(var line_done: three_state);                    { CHARACTER_DELETE_PROC }

   procedure do_test;
   var
      vindex: integer;
   begin { do_test }
      write('abcdef');
      for vindex := 1 to 3
      do vdu_control(cursorleft);
      delay(ex_code_delay);
      vdu_control(chardelete)
   end { do_test } ;

begin { character_delete_proc }
   single_code(line_done, chardelete, deleteonech, false, hardware,
               'character delete', do_test)
end { character_delete_proc } ;

procedure character_insert_proc(var line_done: three_state);                    { CHARACTER_INSERT_PROC }

var
   index: integer;
   old_line_done: three_state;

   procedure do_test;
   var
      vindex: integer;
   begin { do_test }
      write('abcdef');
      for vindex := 1 to 3
      do vdu_control(cursorleft);
      delay(ex_code_delay);
      vdu_control(charinsert)
   end { do_test } ;

begin { character_insert_proc }
   with vdu_table_rec, expanded do
   begin
      old_line_done := line_done;
      single_code(line_done, charinsert, insertonech, false, hardware,
                  'character insert', do_test);
      if old_line_done = is_unknown
      then                              { just checking }
      else
      begin
         if exp_control_array[charinsert].seq_len > 0 then
         begin { test for data protect }
            do_clear_screen;
            display_message('charinsert','protect');
            for index := 1 to cols_per_row-2
            do write(' ');
            write('X');
            vdu_control(cursorleft);
            vdu_control(charinsert);
            delay(ex_code_delay);
            vdu_control(charinsert);
            display_message('charinsert','ques');
            exp_char_protect := not yes_or_no(is_unknown)
         end { test for data protect }
         else exp_char_protect := false
      end
   end
end { character_insert_proc } ;

procedure clear_screen_proc(var line_done: three_state);                        { CLEAR_SCREEN_PROC }

   procedure do_test;
   begin { do_test }
      readln;
      vdu_control(clearscreen)
   end { do_test } ;

begin { clear_screen_proc }
   single_code(line_done, clearscreen, redraws, true, software,
               'clear screen', do_test)
end { clear_screen_proc } ;

procedure command_sequences_proc(var line_done: three_state);                   { COMMAND_SEQUENCES_PROC }

var
   chars_written, line_number: integer;
   infile: text;
   state: (reading_pos, reading_neg, all_read);
   line, action_name: string;
   command_ch: char;
   the_repair: repair_type;
   the_action: action_proc;
   the_sequence, the_reply: sequences;
   read_seq, result: boolean;

   procedure add_repair(repair1, repair2: repair_type; var result: repair_type);

   var
      sum: integer;

   begin { add_repair }
      sum := ord(repair1) + ord(repair2) - 2;
      case sum of
         0: result := ok;
         1: result := rest1;
         2: result := rest2;
         3: result := rest3
         otherwise result := rest4
      end { case }
   end { add_repair } ;

begin { command_sequences_proc }
   with vdu_table_rec, expanded do
   if line_done = is_unknown then
   begin { just checking }
      if (exp_poscommand.seq_len > 0) and
         (exp_negcommand.seq_len > 0) and
         (exp_posmacro  .seq_len > 0) and
         (exp_negmacro  .seq_len > 0)
      then line_done := is_true
      else line_done := is_false
   end { just checking }
   else
   begin { not just checking }
      the_sequence := init_sequence;
      the_reply := init_sequence;
      line_done := is_true;
      if (exp_poscommand.seq_len > 0) and
         (exp_negcommand.seq_len > 0) and
         (exp_posmacro  .seq_len > 0) and
         (exp_negmacro  .seq_len > 0)
      then
      begin
         writeln('Current editor command strings are:');
         write('''Normal''   commands are introduced by: ');
         write_key_sequence(false, exp_poscommand, chars_written);
         writeln;
         write('''Opposite'' commands are introduced by: ');
         write_key_sequence(false, exp_negcommand, chars_written);
         writeln;
         write('''Normal''   macros   are introduced by: ');
         write_key_sequence(false, exp_posmacro  , chars_written);
         writeln;
         write('''Opposite'' macros   are introduced by: ');
         write_key_sequence(false, exp_negmacro  , chars_written);
         writeln;
         writeln;
         write('Do you want to redefine them? ');
         read_seq := yes_or_no(is_unknown)
      end
      else read_seq := true;
      if read_seq then
      begin
         must_set_x_codes := true;
         delete_input_sequence(exp_poscommand, true);
         delete_input_sequence(exp_negcommand, true);
         delete_input_sequence(exp_posmacro  , false);
         exp_posmacro_defined := false;
         delete_input_sequence(exp_negmacro  , false);
         exp_negmacro_defined := false;
         display_message('command_sequences_proc','intro');
         get_key_sequence(exp_poscommand, false);
         get_reply(exp_poscommand, exp_poscommand_reply, true);
         get_repair(exp_poscommand, true, exp_poscommand_repair,
                                          exp_poscommand_reply);
         display_message('command_sequences_proc','type2');
         get_key_sequence(exp_negcommand, false);
         get_reply(exp_negcommand, exp_negcommand_reply, true);
         get_repair(exp_negcommand, true, exp_negcommand_repair,
                                          exp_negcommand_reply);
         display_message('command_sequences_proc','type3');
         if yes_or_no(is_unknown) then
         begin { use the same for macros }
            copy_sequence(exp_poscommand, exp_posmacro);
            posmacro_repair := exp_poscommand_repair;
            copy_sequence(exp_poscommand_reply, exp_posmacro_reply);
            copy_sequence(exp_negcommand, exp_negmacro);
            negmacro_repair := exp_negcommand_repair;
            copy_sequence(exp_negcommand_reply, exp_negmacro_reply)
         end { use the same for macros }
         else
         begin
            display_message('command_sequences_proc','type4');
            get_key_sequence(exp_posmacro, false);
            get_reply(exp_posmacro, exp_posmacro_reply, true);
            get_repair(exp_posmacro, true, posmacro_repair,
                                       exp_posmacro_reply);
            display_message('command_sequences_proc','type5');
            get_key_sequence(exp_negmacro, false);
            get_reply(exp_negmacro, exp_negmacro_reply, true);
            get_repair(exp_negmacro, true, negmacro_repair,
                                       exp_negmacro_reply)
         end;
         delete_input_sequence(exp_posmacro  , true);
         delete_input_sequence(exp_negmacro  , true);
         delete_input_sequence(exp_poscommand, true);
         delete_input_sequence(exp_negcommand, true);
         display_message('command_sequences_proc','process');
         reset(infile, 'VDU_PATTERN');
         line_number := 0;
         state := reading_pos;
         copy_sequence(exp_poscommand, the_sequence);
         the_repair := exp_poscommand_repair;
         copy_sequence(exp_poscommand_reply, the_reply);
         repeat
            readln(infile, line);
            line_number := line_number + 1;
            if line = '//' then
            begin
               state := succ(state);
               if state = reading_neg then
               begin
                  copy_sequence(exp_negcommand, the_sequence);
                  the_repair := exp_negcommand_repair;
                  copy_sequence(exp_negcommand_reply, the_reply)
               end
            end
            else
            begin
               command_ch := line[1];
               action_name := substr(line, 3, length(line)-2);
               if length(action_name) > 12
               then action_name := substr(action_name, 1, 12);
               if check_action(action_name, the_action) then
               begin { name recognised }
                  store_key_code(ord(command_ch), the_sequence);
                  if store_input_sequence(the_sequence, the_action,
                                          the_repair, software, the_reply)
                  then                  { stored correctly }
                  else writeln('Failed to store input sequence');
                  with the_sequence     { remove command_ch }
                  do seq_len := seq_len - 1
               end { name recognised }
               else writeln('Action name ''',action_name,
                            ''' not recognised on line ',line_number:1,
                            ' of file VDU_PATTERN.')
            end
         until state = all_read;
         result := store_input_sequence(exp_posmacro, posmacro,
                   posmacro_repair, software, exp_posmacro_reply);
         result := store_input_sequence(exp_negmacro, negmacro,
                   negmacro_repair, software, exp_negmacro_reply);

         { Store DEL DEL etc as nulls }

         copy_sequence(exp_poscommand, the_sequence);
         append_sequence(exp_poscommand, the_sequence);
         if not sequence_defined(the_sequence) then
         begin
            add_repair(exp_poscommand_repair, exp_poscommand_repair,
                       the_repair);
            result := store_input_sequence(the_sequence, nullact, the_repair,
                      software, exp_poscommand_reply)
         end;
         copy_sequence(exp_negcommand, the_sequence);
         append_sequence(exp_negcommand, the_sequence);
         if not sequence_defined(the_sequence) then
         begin
            add_repair(exp_negcommand_repair, exp_negcommand_repair,
                       the_repair);
            result := store_input_sequence(the_sequence, nullact, the_repair,
                      software, exp_negcommand_reply)
         end;
         add_repair(exp_poscommand_repair, exp_negcommand_repair,
                    the_repair);
         copy_sequence(exp_poscommand_reply, the_reply);
         append_sequence(exp_negcommand_reply, the_reply);
         copy_sequence(exp_poscommand, the_sequence);
         append_sequence(exp_negcommand, the_sequence);
         if not sequence_defined(the_sequence)
         then result := store_input_sequence(the_sequence, nullact, the_repair,
                        software, the_reply);
         copy_sequence(exp_negcommand, the_sequence);
         append_sequence(exp_poscommand, the_sequence);
         if not sequence_defined(the_sequence)
         then result := store_input_sequence(the_sequence, nullact, the_repair,
                        software, the_reply)
      end { read sequence } ;
      dispose_key_sequence(the_sequence);
      dispose_key_sequence(the_reply);
      delay(ex_code_delay)
   end { not just testing }
end { command_sequences_proc } ;

procedure ctrl_ch_proc(var line_done: three_state);                             { CTRL_CH_PROC }

var
   is_gap: boolean;

   procedure test_ctrl_ch(on_control, off_control: controls);

   var
      ch_code: 200b..237b;

   begin { test_ctrl_ch }
      writeln;
      writeln('         1         2         3');
      writeln('12345678901234567890123456789012');
      for ch_code := 200b to 237b do
      begin
         vdu_control(on_control);
         write(chr(ch_code));
         vdu_control(off_control)
      end { for loop } ;
      writeln;
      writeln
   end { test_ctrl_ch } ;

begin { ctrl_ch_proc }
   is_gap := false;
   onoff_code(line_done, ctrlcodeon, ctrlcodeoff, true,
              'displaying control characters',
              test_ctrl_ch, false, is_gap)
end { ctrl_ch_proc } ;

procedure cursor_addressing_proc(var line_done: three_state);                   { CURSOR_ADDRESSING_PROC }

var
   read_seq, ask_test, do_test, exit_from_proc: boolean;
   chars_written: integer;
   cursor_intro, cursor_sep, cursor_term: sequences;
   row_before_col, offsets_ok: boolean;
   encoding: (ascii_code_value, decimal_value, adm12plus_value);
   row_offset, col_offset, row, col: integer;
   code_type, top_number, left_number: integer;

   procedure store_row_or_col(do_row: boolean);

   var
      addr_code: integer;

   begin { store_row_or_col }
      with expanded do
      begin
         if do_row then
         begin
            case encoding of
               ascii_code_value: addr_code := -1;
               decimal_value:    addr_code := -3;
               adm12plus_value:  addr_code := -5
            end { case } ;
            store_key_code(addr_code, exp_cursor_string);
            store_key_code(row_offset, exp_cursor_string)
         end
         else
         begin
            case encoding of
               ascii_code_value: addr_code := -2;
               decimal_value:    addr_code := -4;
               adm12plus_value:  addr_code := -6
            end { case } ;
            store_key_code(addr_code, exp_cursor_string);
            store_key_code(col_offset, exp_cursor_string)
         end
      end
   end { store_row_or_col } ;

begin { cursor_addressing_proc }
   with expanded do
   if line_done = is_unknown then
   begin { just checking }
      if exp_cursor_string.seq_len > 0
      then line_done := is_true
      else line_done := is_false
   end { just checking }
   else
   begin
      line_done := is_false;
      cursor_intro := init_sequence;
      cursor_sep := init_sequence;
      cursor_term := init_sequence;
      if exp_cursor_string.seq_len > 0 then
      begin
         writeln('Current cursor positioning information: ');
         writeln('Screen is ',vdu_table_rec.rows_per_page:1,' rows by ',
                 vdu_table_rec.cols_per_row :1,' columns.');
         write('Cursor string is   ');
         write_key_sequence(false, exp_cursor_string, chars_written);
         writeln;
         writeln;
         write('Do you want to redefine it? ');
         read_seq := yes_or_no(is_unknown)
      end
      else read_seq := true;
      ask_test := not read_seq;
      exit_from_proc := false;
      repeat
         if read_seq then
         begin
            display_message('cursor_addr','intro');

            { Read rows and columns }

            with vdu_table_rec do
            begin
               repeat
                  write('Give number of rows on screen: ');
                  readln(row);
                  if (row < 1) or (row > rows_max) then
                  begin
                     writeln('Must be in the range 1 to ',rows_max:1);
                     row := 0
                  end
               until row <> 0;
               rows_per_page := row;
               repeat
                  write('Give number of columns on screen: ');
                  readln(col);
                  if (col < 10) or (col > cols_max) then
                  begin
                     writeln('Must be in the range 10 to ',cols_max:1);
                     col := 0
                  end
               until col <> 0;
               cols_per_row := col
            end;

            { Read cursor addressing information }

            display_message('cursor_addr','desc');
            get_key_sequence(cursor_intro, false);
            display_message('cursor_addr','type1');
            if yes_or_no(is_unknown) then
            begin
               display_message('cursor_addr','type2');
               get_key_sequence(cursor_sep, false)
            end
            else cursor_sep.seq_len := 0;
            display_message('cursor_addr','type3');
            if yes_or_no(is_unknown) then
            begin
               display_message('cursor_addr','type4');
               get_key_sequence(cursor_term, false)
            end
            else cursor_term.seq_len := 0;
            display_message('cursor_addr','type5');
            row_before_col := yes_or_no(is_unknown);
            repeat
               display_message('cursor_addr','type6');
               repeat
                  readln(code_type);
                  if code_type = 1
                  then encoding := ascii_code_value
                  else if code_type = 2
                  then encoding := decimal_value
                  else
                  begin
                     writeln('Please answer 1 or 2: ');
                     code_type := 0
                  end
               until code_type <> 0;
               if encoding = ascii_code_value then
               begin
                  display_message('cursor_addr','type7');
                  get_single_char(row_offset);
                  row_offset := row_offset - 1;
                  display_message('cursor_addr','type8');
                  get_single_char(col_offset);
                  col_offset := col_offset - 1;
                  if (vdu_table_rec.rows_per_page + row_offset > 255) or
                     (vdu_table_rec.cols_per_row  + col_offset > 255)
                  then
                  begin
                     display_message('cursor_addr','toobig');
                     if yes_or_no(is_unknown) then
                     begin
                        encoding := adm12plus_value;
                        offsets_ok := true
                     end
                     else
                     begin
                        display_message('cursor_addr','notsingle');
                        offsets_ok := false
                     end
                  end
                  else offsets_ok := true
               end { single char }
               else
               begin { decimal addrs }
                  display_message('cursor_addr','type9');
                  readln(top_number);
                  row_offset := top_number - 1;
                  display_message('cursor_addr','type10');
                  readln(left_number);
                  col_offset := left_number - 1;
                  offsets_ok := true
               end { decimal addrs }
            until offsets_ok;

            { Copy information gathered into exp_cursor_string }

            copy_sequence(cursor_intro, exp_cursor_string);
            store_row_or_col(row_before_col);
            append_sequence(cursor_sep, exp_cursor_string);
            store_row_or_col(not row_before_col);
            append_sequence(cursor_term, exp_cursor_string)
         end { read sequence }
         else read_seq := true;         { read it next time if any }
         if not perform_tests then
         begin
            exit_from_proc := true;
            line_done := is_true
         end
         else if not exit_from_proc then
         begin
            if ask_test then
            begin
               writeln;
               write('Do you want to test cursor addressing? ');
               do_test := yes_or_no(is_unknown);
               ask_test := false
            end
            else do_test := true;
            if do_test then
            begin
               do_clear_screen;
               display_message('cursor_addr','test1');
               col := 15;
               for row := 4 to 9 do
               begin
                  vdu_position(row, col);
                  write('*');
                  col := col - 1
               end;
               for row := 8 downto 4 do
               begin
                  vdu_position(row, col);
                  write('*');
                  col := col - 1
               end;
               for row := 1 to 10
               do writeln;
               write('Did it work? ');
               if yes_or_no(is_unknown) then
               begin
                  do_clear_screen;
                  display_message('cursor_addr','test2');
                  delay(ex_code_delay);
                  draw_screen_box(false);
                  vdu_position(11, 3);
                  write('Was the box drawn correctly? ');
                  if yes_or_no(is_unknown) then
                  begin
                     exit_from_proc := true;
                     line_done := is_true
                  end
               end { first test worked }
               else display_message('cursor_addr','fail1')
            end { do test }
            else
            begin
               exit_from_proc := true;
               line_done := is_true
            end;
            if not exit_from_proc then
            begin
               exp_cursor_string.seq_len := 0;
               write('Try again? ');
               if yes_or_no(is_unknown)
               then do_clear_screen
               else exit_from_proc := true
            end
         end { terminal can do it }
      until exit_from_proc;
      dispose_key_sequence(cursor_intro);
      dispose_key_sequence(cursor_sep);
      dispose_key_sequence(cursor_term)
   end { not just testing }
end { cursor_addressing_proc } ;

procedure cursor_down_proc(var line_done: three_state);                         { CURSOR_DOWN_PROC }

   procedure do_test;
   begin { do_test }
      write('abc');
      delay(ex_code_delay);
      vdu_control(cursordown);
      write('def')
   end { do_test } ;

   procedure do_margin;
   begin { do_margin }
      vdu_position(vdu_table_rec.rows_per_page, 1);
      do_test;
      vdu_position(13, 1)
   end { do_margin } ;

begin { cursor_down_proc }
   cursor_arrow_code(line_done, cursordown, godown, 'cursor down',
                     do_test, do_margin, vdu_table_rec.down_pos_ok)
end { cursor_down_proc } ;

procedure cursor_left_proc(var line_done: three_state);                         { CURSOR_LEFT_PROC }

   procedure do_test;
   begin { do_test }
      write('abc');
      delay(ex_code_delay);
      vdu_control(cursorleft);
      writeln('def')
   end { do_test } ;

   procedure do_margin;
   var
      index: integer;
   begin { do_margin }
      write('----------');
      delay(ex_code_delay);
      for index := 1 to vdu_table_rec.cols_per_row
      do vdu_control(cursorleft);
      writeln('X');
   end { do_margin } ;

begin { cursor_left_proc }
   cursor_arrow_code(line_done, cursorleft, goleft, 'cursor left',
                     do_test, do_margin, vdu_table_rec.left_pos_ok)
end { cursor_left_proc } ;

procedure cursor_right_proc(var line_done: three_state);                        { CURSOR_RIGHT_PROC }

   procedure do_test;
   begin { do_test }
      write('abc');
      delay(ex_code_delay);
      vdu_control(cursorright);
      writeln('def')
   end { do_test } ;

   procedure do_margin;
   var
      index: integer;
   begin { do_margin }
      write('----------');
      delay(ex_code_delay);
      for index := 1 to vdu_table_rec.cols_per_row
      do vdu_control(cursorright);
      writeln('X')
   end { do_margin } ;

begin { cursor_right_proc }
   cursor_arrow_code(line_done, cursorright, goright, 'cursor right',
                     do_test, do_margin, vdu_table_rec.right_pos_ok)
end { cursor_right_proc } ;

procedure cursor_up_proc(var line_done: three_state);                           { CURSOR_UP_PROC }

   procedure do_test;
   begin { do_test }
      write('abc');
      delay(ex_code_delay);
      vdu_control(cursorup);
      write('def')
   end { do_test } ;

   procedure do_margin;
   begin { do_margin }
      vdu_position(1, 1);
      do_test;
      vdu_position(13, 1)
   end { do_margin } ;

begin { cursor_up_proc }
   cursor_arrow_code(line_done, cursorup, goup, 'cursor up',
                     do_test, do_margin, vdu_table_rec.up_pos_ok)
end { cursor_up_proc } ;

{ Diagnostic print of input sequences }

procedure diag_print_proc;                                                      { DIAG_PRINT_PROC }

var
   lines_printed: integer;
   printing: boolean;

   procedure pause;
   begin { pause }
      lines_printed := lines_printed + 1;
      if lines_printed >= 23 then
      begin
         lines_printed := 0;
         write('--more?--');
         printing := yes_or_no(is_true)
      end
   end { pause } ;

   procedure print_table(var ptr: ptr_char_table; indent: integer);

   var
      ch: char;
      chars_written: integer;
      unset_ptr: ptr_char_table;

   begin { print_table }
      if ptr = nil
      then { not defined - ignore }
      else
      with expanded do
      begin { table is defined }
         if printing then
         begin
            writeln(' ':indent,'--------------------');
            pause;
            for ch := chr(nul_code) to char.max do
            with ptr^[ch], full_pointer do
            begin
               if full_action <> emptyact  { omit empty entries }
               then
               begin
                  if printing then
                  begin
                     write(' ':indent,char_names[ch],' ',
                           action_proc_info[full_action].act_name,
                           '(',start_index:1,',',number_of_entries:1,')  ',
                           repair_names[full_repair]:7,
                           ware_names[full_ware]);
                     write_key_sequence(true, full_reply, chars_written);
                     write(' ');
                     if table_ptr = nil then
                     begin
                        writeln('NIL');
                        pause
                     end
                     else if table_ptr = unset_ptr then
                     begin
                        writeln('UNSET');
                        pause
                     end
                     else
                     begin
                        writeln('PTR');
                        pause;
                        print_table(table_ptr, indent+3)
                     end
                  end
               end
            end { for loop } ;
            if printing then
            begin
               writeln(' ':indent,'--------------------');
               pause
            end
         end
      end { table exists }
   end { print_table } ;

begin { diag_print_proc }
   lines_printed := 0;
   printing := true;
   print_table(expanded.exp_input_sequences_head, 0);
   if printing then
   begin
      write('Press RETURN: ');
      readln
   end
end { diag_print_proc } ;

procedure exit_proc(var line_done: three_state);                                { EXIT_PROC }

begin { exit_proc }
   get_simple_string(line_done, expanded.exp_exit_string, 'exit', false)
end { exit_proc } ;

procedure function_keys_proc(var line_done: three_state);                       { FUNCTION_KEYS_PROC }

var
   done, first_time, result: boolean;
   the_sequence, the_reply: sequences;
   the_action: action_proc;
   the_ware: ware_type;
   num_func_keys: integer;

begin { function_keys_proc }
   if line_done = is_unknown then
   begin
      count_func(num_func_keys, false);
      if num_func_keys > 0
      then line_done := is_true
      else line_done := is_false
   end
   else with vdu_table_rec do
   begin
      count_func(num_func_keys, false);
      if num_func_keys > 0 then
      begin
         writeln('There are ',num_func_keys:1,' function keys defined.');
         writeln
      end;
      display_message('function_keys','ques');
      if yes_or_no(is_unknown) then
      begin
         if num_func_keys > 0 then
         begin
            display_message('function_keys','delete');
            if yes_or_no(is_unknown)
            then count_func(num_func_keys, true)  { delete existing ones }
         end;
         display_message('function_keys','intro');
         the_sequence := init_sequence;
         the_reply := init_sequence;
         first_time := true;
         line_done := is_false;
         repeat
            get_key_sequence(the_sequence, false);
            with the_sequence do
            begin
               func_expect_return :=
                  get_code_from_sequence(the_sequence, seq_len) = lf_code;
               if func_expect_return
               then seq_len := seq_len - 1  { trim off the newline }
            end;
            done := false;
            repeat
               get_action(the_action);
               if the_action in [posmacro, negmacro] then
               begin
                  writeln('The macro actions cannot be defined here.');
                  writeln('They are set up with the editor commands.')
               end
               else done := true
            until done;
            if first_time then
            begin
               first_time := false;
               the_ware := software;
               get_reply(the_sequence, the_reply, true);
               get_repair(the_sequence, false, func_repair, the_reply)
            end;
            delete_input_sequence(the_sequence, true);
            result := store_input_sequence(the_sequence, the_action, func,
                                           the_ware, the_reply);
            write('Any more function keys? ');
            if not yes_or_no(is_unknown)
            then line_done := is_true
            else do_clear_screen
         until line_done = is_true;
         dispose_key_sequence(the_sequence);
         dispose_key_sequence(the_reply)
      end
      else line_done := is_true;        { no function keys }
   end
end { function_keys_proc } ;

procedure general_input_keys_proc(var line_done: three_state);                  { GENERAL_INPUT_KEYS_PROC }

var
   done, result, first_time: boolean;
   the_sequence, the_reply: sequences;
   the_action: action_proc;
   the_repair: repair_type;
   the_ware: ware_type;

begin { general_input_keys_proc }
   if line_done = is_unknown
   then line_done := is_false           { no obvious way of checking }
   else
   begin
      display_message('general_input','ques');
      if yes_or_no(is_unknown) then
      begin
         display_message('general_input','intro');
         the_sequence := init_sequence;
         the_reply := init_sequence;
         first_time := true;
         line_done := is_false;
         repeat
            get_key_sequence(the_sequence, false);
            done := false;
            repeat
               get_action(the_action);
               if the_action in [posmacro, negmacro] then
               begin
                  writeln('The macro actions cannot be defined here.');
                  writeln('They are set up with the editor commands.')
               end
               else done := true
            until done;
            get_ware(the_action, the_ware);
            get_reply(the_sequence, the_reply, first_time);
            get_repair(the_sequence, false, the_repair, the_reply);
            delete_input_sequence(the_sequence, true);
            result := store_input_sequence(the_sequence, the_action, the_repair,
                                           the_ware, the_reply);
            write('Any more input sequences? ');
            if not yes_or_no(is_unknown)
            then line_done := is_true
            else
            begin
               first_time := false;
               do_clear_screen
            end
         until line_done = is_true;
         dispose_key_sequence(the_sequence);
         dispose_key_sequence(the_reply)
      end
      else line_done := is_true         { no keys defined }
   end
end { general_input_keys_proc } ;

procedure home_key_proc(var line_done: three_state);                            { HOME_KEY_PROC }

var
   chars_written, temp: integer;
   home_sequence: sequences;
   read_seq, exit_from_proc, ask_test, do_test: boolean;
   done: boolean;

begin { home_key_proc }
   with vdu_table_rec, home_location, expanded do
   begin
      home_sequence := init_sequence;
      find_input_sequence(gohome, home_sequence);
      if line_done = is_unknown then
      begin { just checking }
         if home_sequence.seq_len > 0
         then line_done := is_true
         else line_done := is_false
      end { just checking }
      else
      begin
         if home_sequence.seq_len > 0 then
         begin
            write('Home key is: ');
            write_key_sequence(false, home_sequence, chars_written);
            writeln;
            writeln('Home location is row ',home_row:1,'  column ',home_col:1,'.');
            writeln;
            write('Do you want to redefine them? ');
            read_seq := yes_or_no(is_unknown)
         end
         else read_seq := true;
         ask_test := not read_seq;
         exit_from_proc := false;
         repeat
            if read_seq then
            begin
               dl_in_set([gohome]);
               display_message('home_key_proc','intro');
               if yes_or_no(is_unknown) then
               begin { There is a HOME key }
                  display_message('home_key_proc','type1');
                  get_key_sequence(home_sequence, false);
                  display_message('home_key_proc','type2');
                  done := false;
                  repeat
                     write('row: ');
                     readln(temp);
                     if (temp >= 1) and (temp <= rows_per_page) then
                     begin
                        home_row := temp;
                        done := true
                     end
                     else writeln('Row must be in range 1 to ',
                                  rows_per_page:1)
                  until done;
                  done := false;
                  repeat
                     write('column: ');
                     readln(temp);
                     if (temp >= 1) and (home_col <= cols_per_row) then
                     begin
                        home_col := temp;
                        done := true
                     end
                     else writeln('Column must be in range 1 to ',
                                  cols_per_row:1)
                  until done;
                  delete_input_sequence(home_sequence, true);
                  if store_input_sequence(home_sequence, gohome, ok,
                                          software, init_sequence)
                  then                  { stored successfully }
                  else writeln('Failed to store input sequence')
               end { HOME key exists }
               else
               begin
                  home_row := 1;        { just to be tidy }
                  home_col := 1;
                  exit_from_proc := true;
                  line_done := is_true
               end { no HOME key }
            end { read sequence }
            else read_seq := true;      { read it next time if any }
            if not perform_tests then
            begin
               exit_from_proc := true;
               line_done := is_true
            end
            else if not exit_from_proc then
            begin
               if ask_test then
               begin
                  write('Do you want to test them? ');
                  do_test := yes_or_no(is_unknown);
                  ask_test := false
               end
               else do_test := true;
               if do_test then
               begin
                  do_clear_screen;
                  display_message('home_key_proc','test');
                  vdu_position(13, 1);
                  delay(ex_code_delay);
                  vdu_position(home_row, home_col);
                  write('H');
                  vdu_position(13, 1);
                  delay(ex_code_delay);
                  send_key_sequence(home_sequence);
                  write('X');
                  vdu_position(13,1);
                  write('Did it work? ');
                  if yes_or_no(is_unknown) then
                  begin
                     exit_from_proc := true;
                     line_done := is_true
                  end
               end { do test }
               else
               begin
                  exit_from_proc := true;
                  line_done := is_true
               end
            end { may test } ;
            if not exit_from_proc then
            begin
               delete_input_sequence(home_sequence, false);
               write('Try again? ');
               exit_from_proc := not yes_or_no(is_unknown)
            end
         until exit_from_proc
      end { not just testing } ;
      dispose_key_sequence(home_sequence)
   end { with }
end { home_key_proc } ;

procedure line_delete_proc(var line_done: three_state);                         { LINE_DELETE_PROC }

   procedure do_test;
   begin { do_test }
      writeln('line 1');
      writeln('line 2');
      write  ('line 3');
      vdu_control(cursorup);
      delay(ex_code_delay);
      vdu_control(linedelete);
      writeln
   end { do_test } ;

begin { line_delete_proc }
   single_code(line_done, linedelete, deleteonel, false, hardware,
               'line delete', do_test)
end { line_delete_proc } ;

procedure line_erase_proc(var line_done: three_state);                          { LINE_ERASE_PROC }

   procedure do_test;
   var
      vindex: integer;
   begin { do_test }
      writeln('12345678901234567890');
      write  ('xxxxxxxxxxxxxxxxxxxx');
      vdu_control(cursorup);
      for vindex := 1 to 10
      do vdu_control(cursorleft);
      delay(ex_code_delay);
      vdu_control(lineerase);
      writeln
   end { do_test } ;

begin { line_erase_proc }
   single_code(line_done, lineerase, eraseline, false, hardware,
               'line erase', do_test);
end { line_erase_proc } ;

procedure line_insert_proc(var line_done: three_state);                         { LINE_INSERT_PROC }

var
   old_line_done: three_state;

   procedure do_test;
   begin { do_test }
      writeln('line 1');
      write  ('line 2');
      delay(ex_code_delay);
      vdu_control(lineinsert);
      writeln
   end { do_test } ;

begin { line_insert_proc }
   with vdu_table_rec, expanded do
   begin
      old_line_done := line_done;
      single_code(line_done, lineinsert, insertonel, false, hardware,
                  'line insert', do_test);
      if old_line_done = is_unknown
      then                              { just checking }
      else
      begin
         if exp_control_array[lineinsert].seq_len > 0 then
         begin { test for data protect }
            do_clear_screen;
            display_message('lineinsert','protect');
            vdu_position(rows_per_page, 1);
            write('XXXXXXXXXX');
            vdu_position(13,1);
            delay(ex_code_delay);
            vdu_control(lineinsert);
            display_message('lineinsert','ques');
            exp_line_protect := not yes_or_no(is_unknown)
         end { test for data protect }
         else exp_char_protect := false
      end
   end { with }
end { line_insert_proc } ;

procedure long_highlight_proc(var line_done: three_state);                      { LONG_HIGHLIGHT_PROC }

var
   is_gap: boolean;

begin { long_highlight_proc }
   is_gap := false;
   onoff_code(line_done, foldon, foldoff, false,
              'highlighting for long lines',
              test_highlight, false, is_gap);
end { long_highlight_proc } ;

procedure message_highlight_proc(var line_done: three_state);                   { MESSAGE_HIGHLIGHT_PROC }

var
   is_gap: boolean;

begin { message_highlight_proc }
   with expanded do
   begin
      is_gap := exp_mess_type = 2;
      onoff_code(line_done, messon, messoff, false,
                 'highlighting for messages',
                 test_highlight, true, is_gap);
      if is_gap
      then exp_mess_type := 2
      else exp_mess_type := 1
   end
end { message_highlight_proc } ;

procedure page_erase_proc(var line_done: three_state);                          { PAGE_ERASE_PROC }

   procedure do_test;
   var
      vindex: integer;
   begin { do_test }
      writeln('12345678901234567890');
      write  ('xxxxxxxxxxxxxxxxxxxx');
      vdu_control(cursorup);
      for vindex := 1 to 10
      do vdu_control(cursorleft);
      delay(ex_code_delay);
      vdu_control(pageerase);
      writeln
   end { do_test } ;

begin { page_erase_proc }
   single_code(line_done, pageerase, redraws, false, software,
               'page erase', do_test)
end { page_erase_proc } ;

procedure pointer_highlight_proc(var line_done: three_state);                   { POINTER_HIGHLIGHT_PROC }

var
   is_gap: boolean;

begin { pointer_highlight_proc }
   with expanded do
   begin
      is_gap := exp_arrow_type = 2;
      onoff_code(line_done, arrowon, arrowoff, false,
                 'highlighting for line pointer',
                 test_highlight, true, is_gap);
      if is_gap
      then exp_arrow_type := 2
      else exp_arrow_type := 1
   end
end { pointer_highlight_proc } ;

procedure reverse_line_feed_proc(var line_done: three_state);                   { REVERSE_LINE_FEED_PROC }

   procedure do_test;
   begin { do_test }
      vdu_position(1, 1);
      write('abc');
      delay(ex_code_delay);
      vdu_control(reverself);
      write('def');
      vdu_position(13,1)
   end { do_test } ;

begin { reverse_line_feed_proc }
   single_code(line_done, reverself, revlfchar, false, hardware,
               'reverse line feed', do_test)
end { reverse_line_feed_proc } ;

{ Write the binary record out to disk }

procedure save_binary_proc(var line_done: three_state);                         { SAVE_BINARY_PROC }

var
   exit_from_proc, exists, is_binary: boolean;
   binary_version, selected, type_found, table_count: integer;

begin { save_binary_proc }
   line_done := is_false;               { do not set for top level }
   set_x_codes;                         { DEL -> etc. }
   display_message('save_binary_proc','compress');
   if compress_vdu_info then
   begin
      exit_from_proc := false;
      repeat
         display_message('save_binary_proc','intro');
         readln(binary_fn);
         writeln;
         selected := 0;
         check_binary(binary_fn, exists, is_binary, binary_version);
         if exists then
         begin
            if is_binary then
            begin
               if binary_version = info_version then
               begin
                  open(binary_file, binary_fn);
                  list_tables('The file already contains the following terminal descriptions:',
                              table_count,
                              vdu_table_rec.vdu_identification,
                              type_found);
                  selected := 0;
                  if type_found > 0 then
                  begin
                     write('Replace terminal type ',type_found:1,'? ');
                     if yes_or_no(is_unknown)
                     then selected := type_found
                  end;
                  if selected = 0 then
                  begin
                     write('Add to end of file? ');
                     if yes_or_no(is_unknown)
                     then selected := table_count + 1
                     else if table_count = 1
                     then selected := 1
                     else
                     begin
                        write('Which terminal number do you want to replace? ');
                        repeat
                           readln(selected);
                           if (selected < 1) or (selected > table_count) then
                           begin
                              write('Must be 1 to ',table_count:1,': ');
                              selected := 0
                           end
                        until selected <> 0
                     end
                  end
               end { correct version }
               else display_message('save_binary_proc','version')
            end
            else writeln('Not a binary file.')
         end { file exists }
         else
         begin
            writeln('Binary file ',binary_fn,' does not exist.');
            write('Create new binary file? ');
            if yes_or_no(is_unknown) then
            begin
               open(binary_file, binary_fn);
               selected := 1
            end
         end { file does not exist } ;
         if selected > 0 then
         begin
            seek(binary_file, selected);
            binary_file^ := vdu_table_rec;
            put(binary_file);
            list_tables('The file now contains the following terminal descriptions:',
                        table_count,
                        vdu_table_rec.vdu_identification,
                        type_found);
            write('Press RETURN: ');
            readln;
            file_modified := false;     { binary saved }
            exit_from_proc := true
         end;
         close(binary_file);
         if not exit_from_proc then
         begin
            write('Try again? ');
            exit_from_proc := not yes_or_no(is_unknown)
         end
      until exit_from_proc
   end { compressed successfully }
   else
   begin
      display_message('save_binary_proc','failed');
      write('Press RETURN: ');
      readln
   end
end { save_binary_proc } ;

procedure save_source_proc(var line_done: three_state);                         { SAVE_SOURCE_PROC }

var
   exists, is_binary, alright: boolean;
   binary_version: integer;

begin { save_source_proc }
   line_done := is_false;               { do not set for top level }
   display_message('save_source_proc','intro');
   repeat
      readln(source_fn);
      writeln;
      to_up_case(source_fn);
      check_binary(source_fn, exists, is_binary, binary_version);
      if exists then
      begin
         writeln('File ',source_fn,' already exists.');
         write('Alright to replace? ');
         alright := yes_or_no(is_unknown);
         if not alright
         then write('Please give another filename: ')
      end
      else alright := true
   until alright;
   set_x_codes;                         { DEL -> etc. }
   rewrite(report_file, source_fn);
   write_source_file;
   rewrite(report_file,'@tty');
   file_modified := false               { source saved }
end { save_source_proc } ;

procedure scroll_suppress_proc(var line_done: three_state);                     { SCROLL_SUPPRESS_PROC }

var
   read_seq, ask_test, do_test, exit_from_proc: boolean;
   chars_written, type_read: integer;

begin { scroll_suppress_proc }
   with expanded do
   begin
      if line_done = is_unknown then
      begin { just checking }
         if (exp_control_array[charinsert].seq_len > 0) or
            (exp_control_array[scrolloff].seq_len > 0)
         then line_done := is_true
         else line_done := is_false
      end { just checking }
      else
      begin
         line_done := is_false;
         if (exp_control_array[charinsert].seq_len > 0) or
            (exp_control_array[scrolloff].seq_len > 0)
         then
         begin
            writeln('Characters are placed in the bottom right-hand corner using ');
            if exp_control_array[scrolloff].seq_len > 0 then
            begin
               write('Scroll suppress: ');
               write_key_sequence(false, exp_control_array[scrolloff],
                                  chars_written);
               writeln;
               write('Scroll enable:   ');
               write_key_sequence(false, exp_control_array[scrollon ],
                                  chars_written);
               writeln
            end
            else writeln('your terminal''s ''character insert'' function.');
            writeln;
            write('Do you want to redefine this? ');
            read_seq := yes_or_no(is_unknown)
         end
         else read_seq := true;
         ask_test := not read_seq;
         exit_from_proc := false;
         repeat
            if read_seq then
            begin
               display_message('scroll_suppress_proc','intro');
               if yes_or_no(is_unknown) then
               begin { terminal can do it unaided }
                  { set both to NUL to convince the editor its alright }
                  exp_control_array[scrolloff].seq_len := 0;
                  store_key_code(200b, exp_control_array[scrolloff]);
                  exp_control_array[scrollon ].seq_len := 0;
                  store_key_code(200b, exp_control_array[scrollon ])
               end { scroll suppress not needed }
               else
               begin
                  display_message('scroll_suppress_proc','type1');
                  repeat
                     readln(type_read);
                     if (type_read < 1) or (type_read > 3) then
                     begin
                        write('Must be 1 to 3: ');
                        type_read := 0
                     end
                  until type_read <> 0;
                  if type_read = 1 then
                  begin
                     display_message('scroll_suppress_proc','type2');
                     get_key_sequence(exp_control_array[scrolloff], false);
                     display_message('scroll_suppress_proc','type3');
                     get_key_sequence(exp_control_array[scrollon ], false)
                  end
                  else
                  begin { no scroll suppress feature }
                     exp_control_array[scrolloff].seq_len := 0;
                     exp_control_array[scrollon ].seq_len := 0;
                     if exp_control_array[charinsert].seq_len > 0
                     then               { use charinsert to convince user }
                     else
                     begin { do not test otherwise }
                        if type_read = 2 then
                        begin
                           writeln;
                           writeln('Warning - character insert code has not been defined yet.');
                           writeln;
                           write('Press RETURN: ');
                           readln
                        end;
                        line_done := is_true;
                        exit_from_proc := true
                     end { do not test otherwise }
                  end { no scroll suppress }
               end { scroll suppress needed }
            end { read sequence }
            else read_seq := true;      { read it next time }
            if not perform_tests then
            begin
               exit_from_proc := true;
               line_done := is_true
            end
            else if not exit_from_proc then
            begin
               if ask_test then
               begin
                  write('Do you want to test it? ');
                  do_test := yes_or_no(is_unknown);
                  ask_test := false
               end
               else do_test := true;
               if do_test then
               begin
                  do_clear_screen;
                  display_message('scroll_suppress_proc','test');
                  delay(ex_code_delay);
                  draw_screen_box(true);
                  vdu_position(14, 3);
                  write('Is the whole box visible? ');
                  if yes_or_no(is_unknown) then
                  begin
                     line_done := is_true;
                     exit_from_proc := true
                  end
               end { do test }
            end { not testing } ;
            if not exit_from_proc then
            begin
               exp_control_array[scrolloff].seq_len := 0;
               exp_control_array[scrollon ].seq_len := 0;
               write('Try again? ');
               if yes_or_no(is_unknown)
               then do_clear_screen
               else exit_from_proc := true
            end
         until exit_from_proc;
      end { just checking }
   end { with }
end { scroll_suppress_proc } ;

procedure tabulation_proc(var line_done: three_state);                          { TABULATION_PROC }

var
   read_seq, ask_test, do_test, exit_from_proc: boolean;
   chars_written: integer;
   tab_ware, bt_ware: ware_type;
   tested_ok, result1, result2: boolean;

   procedure set_all_zero;

   begin { set_all_zero }
      with expanded do
      begin
         exp_control_array[horiztab ].seq_len := 0;
         exp_control_array[backtab  ].seq_len := 0;
         exp_control_array[cleartabs].seq_len := 0;
         exp_control_array[settab   ].seq_len := 0;
         exp_control_array[clearatab].seq_len := 0;
      end
   end { set_all_zero } ;

begin { tabulation_proc }
   with expanded do
   if line_done = is_unknown then
   begin { just checking }
      if exp_control_array[horiztab].seq_len > 0
      then line_done := is_true
      else line_done := is_false
   end { just checking }
   else
   begin
      line_done := is_false;
      delete_input_sequence(exp_control_array[horiztab], false);
      delete_input_sequence(exp_control_array[backtab], false);
      if exp_control_array[horiztab].seq_len > 0 then
      begin
         writeln('Current tabulation strings are: ');
         write('Horizontal tab is  ');
         write_key_sequence(false, exp_control_array[horiztab ], chars_written);
         writeln;
         write('Backwards tab is   ');
         write_key_sequence(false, exp_control_array[backtab  ], chars_written);
         writeln;
         write('Clear all tabs is  ');
         write_key_sequence(false, exp_control_array[cleartabs], chars_written);
         writeln;
         write('Set a tab is       ');
         write_key_sequence(false, exp_control_array[settab   ], chars_written);
         writeln;
         write('Clear a tab is     ');
         write_key_sequence(false, exp_control_array[clearatab], chars_written);
         writeln;
         writeln;
         write('Do you want to redefine them? ');
         read_seq := yes_or_no(is_unknown)
      end
      else read_seq := true;
      ask_test := not read_seq;
      exit_from_proc := false;
      repeat
         if read_seq then
         begin
            file_modified := true;
            set_all_zero;
            display_message('tabulation','intro');
            if yes_or_no(is_unknown)
            then tab_ware := hardware
            else tab_ware := software;
            if tab_ware = hardware then
            begin
               display_message('tabulation','type1');
               get_key_sequence(exp_control_array[horiztab], false);
               display_message('tabulation','type2');
               if yes_or_no(is_unknown)
               then bt_ware := hardware
               else bt_ware := software;
               if bt_ware = hardware then
               begin { hardware back tab avail }
                  display_message('tabulation','type3');
                  get_key_sequence(exp_control_array[backtab], false)
               end { hardware back tab avail }
               else
               begin { no hardware back tab }
                  exp_control_array[backtab].seq_len := 0;
                  display_message('tabulation','type4');
                  if yes_or_no(is_unknown) then
                  begin { get substitute back tab key }
                     display_message('tabulation','type5');
                     get_key_sequence(exp_control_array[backtab], false)
                  end { get substitute back tab key }
               end { no hardware for back tab } ;
               display_message('tabulation','type6');
               get_key_sequence(exp_control_array[cleartabs], false);
               display_message('tabulation','type7');
               get_key_sequence(exp_control_array[settab], false);
               display_message('tabulation','type8');
               if yes_or_no(is_unknown) then
               begin
                  display_message('tabulation','type9');
                  get_key_sequence(exp_control_array[clearatab], false)
               end
            end { hardware tabs }
            else
            begin { not hardware tabs }
               bt_ware := software;     { assume no back tab either }
               display_message('tabulation','type10');
               if yes_or_no(is_unknown) then
               begin
                  display_message('tabulation','type11');
                  get_key_sequence(exp_control_array[horiztab], false)
               end;
               display_message('tabulation','type4');
               if yes_or_no(is_unknown) then
               begin
                  display_message('tabulation','type3');
                  get_key_sequence(exp_control_array[backtab], false)
               end
            end { not hardware tabs }
         end { read sequence }
         else
         begin
            read_seq := true;           { read it next time if any }
            tab_ware := software;
            bt_ware := software;
            if exp_control_array[horiztab].seq_len > 0 then
            begin
               tab_ware := hardware;
               if exp_control_array[backtab].seq_len > 0
               then bt_ware := hardware
            end
         end;
         if not perform_tests then
         begin
            exit_from_proc := true;
            line_done := is_true
         end
         else if not exit_from_proc then
         begin
            if tab_ware = software
            then do_test := false
            else if ask_test then
            begin
               writeln;
               write('Do you want to test tabulation? ');
               do_test := yes_or_no(is_unknown);
               ask_test := false
            end
            else do_test := true;
            if do_test then
            begin
               do_clear_screen;
               vdu_control(cleartabs);
               { test setting a tab }
               display_message('tabulation','test1');
               write('  ');
               vdu_control(settab);
               write('   ');
               vdu_control(settab);
               writeln('     ');
               writeln('1234567890');
               delay(ex_code_delay);
               vdu_control(horiztab);
               writeln('X');
               writeln;
               write('Did it work? ');
               tested_ok := yes_or_no(is_unknown);
               if tested_ok and (exp_control_array[clearatab].seq_len > 0) then
               begin { test clear a tab if avail }
                  display_message('tabulation','test2');
                  write('  ');
                  vdu_control(clearatab);
                  writeln;
                  writeln('1234567890');
                  delay(ex_code_delay);
                  vdu_control(horiztab);
                  writeln('X');
                  writeln;
                  write('Did it work? ');
                  tested_ok := yes_or_no(is_unknown)
               end { test clear a tab if avail } ;
               if tested_ok then
               begin { test clear all tabs }
                  display_message('tabulation','test3');
                  vdu_control(cleartabs);
                  write('        ');
                  vdu_control(settab);
                  writeln;
                  writeln('1234567890');
                  delay(ex_code_delay);
                  vdu_control(horiztab);
                  writeln('X');
                  writeln;
                  write('Did it work? ');
                  tested_ok := yes_or_no(is_unknown)
               end { test clear all tabs } ;
               if tested_ok and (bt_ware = hardware) then
               begin { test hardware back tab }
                  display_message('tabulation','test4');
                  vdu_control(cleartabs);
                  write('  ');
                  vdu_control(settab);
                  writeln;
                  writeln('1234567890');
                  write('          ');
                  delay(ex_code_delay);
                  vdu_control(backtab);
                  writeln('X');
                  writeln;
                  write('Did it work? ');
                  tested_ok := yes_or_no(is_unknown)
               end { test hardware back tab } ;
               if tested_ok then
               begin
                  exit_from_proc := true;
                  line_done := is_true
               end
            end { do test }
            else
            begin
               exit_from_proc := true;
               line_done := is_true
            end;
            if not exit_from_proc then
            begin
               set_all_zero;
               write('Try again? ');
               if yes_or_no(is_unknown)
               then do_clear_screen
               else exit_from_proc := true
            end
         end { terminal can do it }
      until exit_from_proc;

      { Delete the input sequences again because they may have changed from }
      { software to hardware, in which case they were not deleted at the top }
      delete_input_sequence(exp_control_array[horiztab], true);
      delete_input_sequence(exp_control_array[backtab], true);

      if exp_control_array[horiztab].seq_len > 0
      then result1 := store_input_sequence(exp_control_array[horiztab], tabright,
                                           ok, tab_ware, init_sequence)
      else result1 := true;
      if exp_control_array[backtab].seq_len > 0
      then result2 := store_input_sequence(exp_control_array[backtab], tableft,
                                           ok, bt_ware, init_sequence)
      else result2 := true;
      if result1 and result2
      then                              { both stored successfully }
      else
      begin
         write('Press RETURN: ');
         readln
      end;
      if tab_ware = software
      then exp_control_array[horiztab].seq_len := 0;
      if bt_ware = software
      then exp_control_array[backtab].seq_len := 0
   end { not just testing }
end { tabulation_proc } ;

procedure term_reset_proc(var line_done: three_state);                          { TERM_RESET_PROC }

begin { term_reset_proc }
   get_simple_string(line_done, expanded.exp_reset_string, 'reset', true)
end { term_reset_proc } ;

procedure title_proc(var line_done: three_state);                               { TITLE_PROC }

var
   change_title: boolean;

begin { title_proc }
   with vdu_table_rec do
   if line_done = is_unknown then
   begin { just checking }
      if vdu_identification = 'Not set'
      then line_done := is_false
      else line_done := is_true
   end { just checking }
   else
   begin
      if vdu_identification = 'Not set'
      then change_title := true
      else
      begin
         writeln('Current terminal identification is:');
         writeln;
         writeln(vdu_identification);
         writeln;
         write('Do you want to change it? ');
         change_title := yes_or_no(is_unknown)
      end;
      if change_title then
      begin
         display_message('title_proc','intro');
         repeat
            readln(vdu_identification);
            if length(vdu_identification) = 0
            then display_message('title_proc','error')
         until length(vdu_identification) > 0;
         file_modified := true          { title changed }
      end;
      line_done := is_true
   end
end { title_proc } ;

procedure obey_a_proc(proc: procs; var line_done: three_state);                 { OBEY_A_PROC }

begin { obey_a_proc }
   case proc of
      character_delete:   character_delete_proc(line_done);
      character_insert:   character_insert_proc(line_done);
      clear_screen:       clear_screen_proc(line_done);
      command_sequences:  command_sequences_proc(line_done);
      ctrl_ch:            ctrl_ch_proc(line_done);
      cursor_addressing:  cursor_addressing_proc(line_done);
      cursor_down:        cursor_down_proc(line_done);
      cursor_left:        cursor_left_proc(line_done);
      cursor_right:       cursor_right_proc(line_done);
      cursor_up:          cursor_up_proc(line_done);
      diag_print:         diag_print_proc;
      exit:               exit_proc(line_done);
      function_keys:      function_keys_proc(line_done);
      general_input_keys: general_input_keys_proc(line_done);
      home_key:           home_key_proc(line_done);
      line_delete:        line_delete_proc(line_done);
      line_erase:         line_erase_proc(line_done);
      line_insert:        line_insert_proc(line_done);
      long_highlight:     long_highlight_proc(line_done);
      message_highlight:  message_highlight_proc(line_done);
      page_erase:         page_erase_proc(line_done);
      pointer_highlight:  pointer_highlight_proc(line_done);
      reverse_line_feed:  reverse_line_feed_proc(line_done);
      save_binary:        save_binary_proc(line_done);
      save_source:        save_source_proc(line_done);
      scroll_suppress:    scroll_suppress_proc(line_done);
      tabulation:         tabulation_proc(line_done);
      term_reset:         term_reset_proc(line_done);
      title:              title_proc(line_done);
   end { case }
end { obey_a_proc } ;

procedure init_menus;                                                           { INIT_MENUS }

var
   mindex: menu_range;
   all_done: three_state;

   procedure set_line_done(var the_menu: menus; var all_done: three_state);

   var
      mindex: menu_range;

   begin { set_line_done }
      with the_menu do
      begin
         all_done := is_true;
         for mindex := 1 to menu_length do
         with menu_entries[mindex] do
         begin
            line_done := is_unknown;
            case proc of
               up_menu:   line_done := is_false;
               down_menu: set_line_done(next_menu^, line_done);
               otherwise  obey_a_proc(proc, line_done)
            end { case } ;
            if line_done = is_false
            then all_done := is_false
         end { for loop }
      end { with }
   end { set_line_done } ;

begin { init_menus }
   must_set_x_codes := false;           { starting new desc }
   with top_menu^ do
   begin
      for mindex := 1 to menu_length do
      with menu_entries[mindex] do
      begin
         line_done := is_false
      end
   end;
   set_line_done(make_changes_menu^, all_done);
end { init_menus } ;

procedure display_menu(level: integer;                                          { DISPLAY_MENU }
                       menu_title: string;
                       var the_menu: menus;
                       var all_done: three_state;
                       var go_to_top: boolean);

label 422, 423;

var
   getting_response, level_done, allow_it, select_done, go_up: boolean;
   state: search_type;
   lindex: menu_range;
   the_proc: procs;
   answer: alfa;
   key_ch: char;
   chindex: integer;

   procedure make_changes_proc(var line_done: three_state; do_tests: boolean);

   begin { make_changes_proc }
      line_done := is_false;
      perform_tests := do_tests;
      if perform_tests
      then send_key_sequence(expanded.exp_reset_string);
      display_menu(level+1, 'Make changes to terminal description',
                   make_changes_menu^, all_done, go_to_top)
   end { make_changes_proc } ;

   { Load an old file specified by the user }

   procedure load_old_file;

   var
      the_fn: string;
      exit_from_proc, exists, is_binary: boolean;
      binary_version, table_count, type_found, selected: integer;

   begin { load_old_file }
      empty_exp_record;
      exit_from_proc := false;
      display_message('load_old_file','intro');
      repeat
         repeat
            write('Give filename for input: ');
            readln(the_fn);
            writeln;
            check_binary(the_fn, exists, is_binary, binary_version);
            if not exists
            then writeln('File not found.')
         until exists;
         if is_binary then
         begin
            binary_fn := the_fn;
            if binary_version = 1 then
            begin
               display_message('load_old_file','version1');
               if yes_or_no(is_unknown) then
               begin { convert from version 1 to 2 }
                  write('Give filename for version 2 file: ');
                  repeat
                     readln(the_fn);
                     check_binary(the_fn, exists, is_binary,
                                  binary_version);
                     if exists then
                     begin
                        writeln('File ',the_fn,' already exists.');
                        write('Please give another name: ')
                     end
                  until not exists;
                  convert_1_to_2(binary_fn, the_fn);
                  binary_fn := the_fn;
                  binary_version := 2
               end { convert from version 1 to 2 }
               else                     { user does not want to convert }
            end;
            if binary_version = info_version then
            begin
               open(binary_file, binary_fn);
               list_tables('Terminal descriptions in file', table_count,
                           '', type_found);
               if table_count = 0
               then writeln('The file is empty.')
               else if table_count = 1 then
               begin
                  selected := 1;
                  exit_from_proc := true
               end
               else
               begin
                  writeln;
                  write('Which terminal number? ');
                  repeat
                     readln(selected);
                     if (selected < 1) or (selected > table_count) then
                     begin
                        write('Must be 1 to ',table_count:1,': ');
                        selected := 0
                     end
                  until selected <> 0;
                  exit_from_proc := true
               end { choose which table } ;
               if exit_from_proc then
               begin
                  seek(binary_file, selected);
                  get(binary_file);
                  vdu_table_rec := binary_file^;
                  expand_vdu_info
               end;
               close(binary_file)
            end { correct version number }
            else writeln('File contains old version of vdu tables.')
         end { binary file }
         else
         begin
            source_fn := the_fn;
            reset(source_file, source_fn);
            vdu_type_count := 0;
            exit_from_proc := read_one_source(false);
            close(source_file)
         end { source file } ;
         if exit_from_proc then
         begin
            writeln;
            writeln('Terminal description loaded is:');
            writeln(vdu_table_rec.vdu_identification);
            writeln;
            write('Correct one? ');
            exit_from_proc := yes_or_no(is_unknown)
         end
      until exit_from_proc
   end { load_old_file } ;

   procedure new_file_proc(var line_done: three_state);

   begin { new_file_proc }
      line_done := is_false;            { do not set for top level }
      if alright_to_replace then
      begin
         empty_exp_record;
         init_menus;                    { init before load so all 'false' }
         display_message('new_file_proc','ques');
         if yes_or_no(is_unknown)
         then load_old_file;
         make_changes_proc(line_done, true)
      end
   end { new_file_proc } ;

   procedure old_file_proc(var line_done: three_state);

   begin { old_file_proc }
      line_done := is_false;            { do not set for top level }
      if alright_to_replace then
      begin
         perform_tests := false;
         load_old_file;
         init_menus                     { set appropriate false/true }
      end { alright to replace }
   end { old_file_proc } ;

   procedure quit_onunit(not_used: integer);

   begin { quit_onunit }
      writeln;
      writeln('QUIT');
      if getting_response
      then goto 423                     { leave this level }
      else goto 422                     { redisplay menu }
   end { quit_onunit } ;

   procedure error_onunit(not_used: integer);

   begin { error_onunit }
      goto 422
   end { error_onunit } ;

begin { display_menu }
   with the_menu, vdu_table_rec, expanded do
   begin
      getting_response := true;         { safety measure }
      if not testing_code then
      begin
         on('QUIT$', quit_onunit);
         on('ERROR_TRAP', error_onunit)
      end;
   422:
      level_done := false;
      repeat

         { Display menu and wait for valid response }

         getting_response := true;
         do_clear_screen;
         if level = 1 then
         begin
            if vdu_identification = 'Not set'
            then                        { do not bother to print }
            else
            begin
               write(vdu_identification);
               if file_modified
               then write('      (modified)');
               writeln;
               for chindex := 1 to length(vdu_identification)
               do write('-')
            end
         end
         else
         begin
            writeln('level: ',level:1,'         ',menu_title);
            write  ('-------','-'    ,'         ');
            for chindex := 1 to length(menu_title)
            do write('-')
         end;
         writeln;
         writeln;
         writeln('   key           action');
         writeln;
         for lindex := 1 to menu_length do
         with menu_entries[lindex] do
         begin
            write('    ',line_key,'    ');
            if (level > 1) and (line_done = is_true)
            then write('done')
            else write('    ');
            writeln('    ',line_text);
            writeln
         end;
         if level > 1 then
         begin
            writeln('    U            Up to previous menu');
            writeln
         end;
         if level > 2 then
         begin
            writeln('    T            Back to Top-level menu');
            writeln
         end;
         go_up := false;
         go_to_top := false;
         select_done := false;
         repeat
            write('Give key: ');
            readln(answer);
            key_ch := answer[1];
            answer[1] := ' ';
            if answer = '        ' then
            begin { single character given }
               if key_ch in ['a'..'z']
               then key_ch := chr(ord('A') - ord('a') + ord(key_ch));
               lindex := 1;
               state := searching;
               repeat
                  if key_ch = menu_entries[lindex].line_key
                  then state := found
                  else if lindex < menu_length
                  then lindex := lindex + 1
                  else state := not_found
               until state <> searching;
               if state = found then
               begin
                  the_proc := menu_entries[lindex].proc;
                  if the_proc = down_menu then
                  begin
                     allow_it := true;
                     if (level = 2) and (lindex > 1) and
                        (menu_entries[1].line_done = is_false)
                     then
                     begin
                        display_message('display_menu','essential');
                        allow_it := false
                     end
                     else if (level = 3) and
                            ((exp_cursor_string.seq_len <= 0) or
                             (exp_control_array[clearscreen].seq_len <= 0))
                     then
                     begin
                        display_message('display_menu','need');
                        allow_it := false
                     end;
                     if allow_it
                     then select_done := true
                  end { down_menu }
                  else select_done := true
               end
               else
               begin { test for standard responses }
                  if level > 1 then
                  begin
                     state := found;
                     if key_ch = 'U'
                     then go_up := true
                     else if key_ch = 'T'
                     then go_to_top := true
                     else state := not_found
                  end;
                  if key_ch = 'P' then
                  begin
                     the_proc := diag_print;
                     state := found
                  end
                  else if key_ch = 'M' then
                  begin
                     the_proc := redisplay_menu;
                     state := found
                  end;
                  if state = found
                  then select_done := true
                  else writeln('Key not recognised.  You can type M to redraw the menu.')
               end { test for standard responses }
            end { single character given }
            else writeln('Please give a single character.')
         until select_done;

         { Perform action specified by user }

         getting_response := false;
         if go_up or go_to_top
         then level_done := true
         else with menu_entries[lindex] do
         begin
            case the_proc of
               up_menu:
                  level_done := true;
               down_menu:
                  begin
                     display_menu(level+1, line_text, next_menu^,
                                  line_done, go_to_top);
                     if (level > 1) and go_to_top
                     then level_done := true
                  end;
               new_file:
                  new_file_proc(line_done);
               old_file:
                  old_file_proc(line_done);
               make_changes:
                  make_changes_proc(line_done, true);
               no_testing:
                  make_changes_proc(line_done, false);
               diag_print:
                  obey_a_proc(the_proc, line_done);
               redisplay_menu: ;
               otherwise
                  begin
                     do_clear_screen;
                     writeln(line_text);
                     for chindex := 1 to length(line_text)
                     do write('-');
                     writeln;
                     writeln;
                     obey_a_proc(the_proc, line_done)
                  end
            end { case }
         end
      until level_done;
   423:
      all_done := menu_entries[1].line_done;
      for lindex := 2 to menu_length
      do if menu_entries[lindex].line_done = is_false
         then all_done := is_false
   end { with }
end { display_menu } ;

begin { do_interactive }
   interactive := true;
   perform_tests := false;
   testing_code := false;
   {---teststart---}
{  write('Testing code? ');             {----}
{  testing_code := yes_or_no(is_unknown);    {----}
   {---teststop---}
   read_vdu_messages;
   control_term(rekord, false);         { note terminal state now }
   rewrite(report_file,'@tty');         { to avoid passing files all over }
   create_menus;
   init_menus;
   write('Do you want to read the introductory information? ');
   if yes_or_no(is_true)
   then display_message('do_interactive','intro');
   file_modified := false;
   repeat
      display_menu(1, 'Top level menu', top_menu^, all_done, go_to_top);
   until alright_to_replace;
   control_term(reinstate, false);      { restore terminal state }
end { do_interactive } ;

begin { main }
   vdu_punctuation := [' ','{',chr(ff_code)];
   ascii_printable := [first_print_char .. last_print_char];
   writeln('[VDUSETUP for Sheffield Editor version ',version_string,']');
   writeln(copyright_string);
   entry_length := sizeof(vdu_table_rec.two_way.entry_array[1]);
   {---teststart---}
{  with vdu_table_rec do  {----}
{  begin  {----}
{     writeln('Record sizes:');         {----}
{     writeln('   vdu_char_pointer: ',sizeof(control_array[backtab]));  {----}
{     writeln('   char_entry:       ',sizeof(action_array[ord(' ')]));  {----}
{     writeln('   char_plus_entry:  ',sizeof(two_way.entry_array[1]));  {----}
{     writeln('   char_array:       ',sizeof(two_way.char_array));  {----}
{     writeln('   entry_array:      ',sizeof(two_way.entry_array));  {----}
{     writeln('   vdu_info:         ',sizeof(vdu_table_rec))  {----}
{  end;  {----}
   {---teststop---}
   init_exp_record;
   repeat
      write('Interactive or Batch? ');
      readln(answer)
   until answer[1] in ['i','I','b','B'];
   if answer[1] in ['i','I']
   then do_interactive
   else do_batch
end { main } .
