%separate = true
%ecs = true
%ttybuffered = true
{ VDUSETUP for Sheffield University Editor }
{ Copyright (c) University of Sheffield 1990 }

{***********************************************}
{ COMMON PROCEDURES AND BATCH CODE FOR VDUSETUP }
{***********************************************}

subprogram vdusetup_procs;

%insert pmed_globals.ins.pas
%insert vdusetup_globals.ins.pas

{ Dispose of a used sequence chain }

procedure dispose_key_sequence(var the_sequence: sequences);                    { DISPOSE_KEY_SEQUENCE }

var
   ptr, next_ptr: ptr_seq_block;

begin { dispose_key_sequence }
   with the_sequence do
   begin
      ptr := first_seq_block;
      first_seq_block := nil;
      while ptr <> nil do
      begin
         next_ptr := ptr^.next_seq_block;
         dispose(ptr);
         ptr := next_ptr
      end;
      seq_len := 0
   end
end { dispose_key_sequence } ;

{ Initialise all the expanded and compressed vdu information }

procedure init_exp_record;                                                      { INIT_EXP_RECORD }

var
   control: controls;
   cindex: vdu_char_range;
   ch: char;
   empty_char_entry: char_entry;

begin { init_exp_record }
   { Construct initial records }
   with init_sequence do
   begin
      seq_len := 0;
      first_seq_block := nil
   end;
   with init_full_entry do
   begin
      full_action := emptyact;
      with full_pointer do
      begin
         start_index := 1;
         number_of_entries := 0
      end;
      full_repair := ok;
      full_ware := software;
      full_reply := init_sequence;
      table_ptr := nil
   end;
   with init_table_rec do
   begin
      stored_version := info_version;
      { Set all blank first just to initialise memory }
      vdu_identification := '                                                                                ';
      vdu_identification := 'Not set';
      with empty_char_pointer do
      begin
         start_index := 1;
         string_length := 0
      end;
      for control := controls.min to controls.max
      do control_array[control] := empty_char_pointer;
      reset_string := empty_char_pointer;
      exit_string := empty_char_pointer;
      cursor_string := empty_char_pointer;
      for cindex := 1 to vdu_chars_max  { clear char area }
      do vdu_table_rec.two_way.char_array[cindex] := 0;
      mess_type1 := false;
      arrow_type1 := false;
      func_expect_return := false;
      func_repair := ok;
      cols_per_row := 1;
      rows_per_page := 1;
      with home_location do
      begin
         home_row := 1;
         home_col := 1
      end;
      right_pos_ok := false;
      right_prt_ok := true;
      left_pos_ok := false;
      up_pos_ok := false;
      down_pos_ok := false;
      data_protect := false;
      posmacro_repair := ok;
      negmacro_repair := ok;
      posmacro_reply := empty_char_pointer;
      negmacro_reply := empty_char_pointer;
      with empty_char_entry do
      begin
         char_action := erroract;
         char_repair := ok;
         char_ware   := software;
         char_reply  := empty_char_pointer
      end;
      for ch := chr(nul_code) to char.max
      do action_array[ord(ch)] := empty_char_entry
   end;

   { Initialise expanded vdu information }

   with expanded do
   begin
      for control := controls.min to controls.max
      do exp_control_array[control] := init_sequence;
      exp_reset_string := init_sequence;
      exp_exit_string := init_sequence;
      exp_cursor_string := init_sequence;
      exp_mess_type := 0;
      exp_arrow_type := 0;
      exp_char_protect := false;
      exp_line_protect := false;
      exp_posmacro := init_sequence;
      exp_negmacro := init_sequence;
      exp_posmacro_reply := init_sequence;
      exp_negmacro_reply := init_sequence;
      exp_posmacro_defined := false;
      exp_negmacro_defined := false;
      exp_input_sequences_head := nil;
      exp_poscommand := init_sequence;
      exp_negcommand := init_sequence;
      exp_poscommand_repair := ok;
      exp_negcommand_repair := ok;
      exp_poscommand_reply := init_sequence;
      exp_negcommand_reply := init_sequence
   end;
   vdu_table_rec := init_table_rec
end { init_exp_record } ;

{ Dispose an input table and all its inferiors }

procedure empty_input_table(var ptr: ptr_char_table);                           { EMPTY_INPUT_TABLE }

var
   ch: char;

begin { empty_input_table }
   if ptr <> nil then
   begin
      for ch := chr(nul_code) to char.max do
      with ptr^[ch] do
      begin
         if full_action in [sequence, posmacro, negmacro]
         then empty_input_table(table_ptr)
         else dispose_key_sequence(full_reply)
      end;
      dispose(ptr);
      ptr := nil
   end
end { empty_input_table } ;

{ Empty all the expanded and compressed vdu information }

procedure empty_exp_record;                                                     { EMPTY_EXP_RECORD }

var
   control: controls;

begin { empty_exp_record }
   with expanded do
   begin
      for control := controls.min to controls.max
      do exp_control_array[control].seq_len := 0;
      exp_reset_string.seq_len   := 0;
      exp_exit_string.seq_len    := 0;
      exp_cursor_string.seq_len  := 0;
      exp_posmacro.seq_len := 0;
      exp_negmacro.seq_len := 0;
      exp_posmacro_reply.seq_len := 0;
      exp_negmacro_reply.seq_len := 0;
      exp_posmacro_defined := false;
      exp_negmacro_defined := false;
      empty_input_table(exp_input_sequences_head);
      exp_poscommand.seq_len := 0;
      exp_negcommand.seq_len := 0;
      exp_poscommand_repair := ok;
      exp_negcommand_repair := ok;
      exp_poscommand_reply.seq_len := 0;
      exp_negcommand_reply.seq_len := 0
   end;
   vdu_table_rec := init_table_rec
end { empty_exp_record } ;

{ Convert string to upper case }

procedure to_up_case(var the_str: string);                                      { TO_UP_CASE }

var
   index: 1..80;
   ch: char;

begin { to_up_case }
   for index := 1 to length(the_str) do
   begin
      ch := the_str[index];
      if ch in ['a'..'z']
      then the_str[index] := chr(ord('A') - ord('a') + ord(ch));
   end { for loop }
end { to_up_case } ;

{ Convert string to lower case }

procedure to_low_case(var the_str: string12);                                   { TO_LOW_CASE }

var
   index: 1..12;
   ch: char;

begin { to_low_case }
   for index := 1 to length(the_str) do
   begin
      ch := the_str[index];
      if ch in ['A'..'Z']
      then the_str[index] := chr(ord('a') - ord('A') + ord(ch));
   end { for loop }
end { to_low_case } ;

{ Write out the list of possible action names }

procedure list_action_names;                                                    { LIST_ACTION_NAMES }

var
   count: integer;
   act: action_proc;

begin { list_action_names }
   writeln('Possible action proc names are:');
   count := 0;
   for act := action_proc.min to action_proc.max do
   begin
      if act in [sequence, emptyact]
      then { do not print special internal ones }
      else
      begin { not special internal one }
         if count <= 0 then
         begin
            count := 6;
            writeln;
         end;
         count := count - 1;
         write(action_proc_info[act].act_name)
      end { not special internal one }
   end;
   writeln;
   writeln
end { list_action_names } ;

{ Interpret a word as the name of an action_proc }

function check_action(word_str: string12; var action: action_proc): boolean;    { CHECK_ACTION }

var
   act: action_proc;
   state: search_type;
   word: char12;
   fiddle: record
      case boolean of
         false: (strng: string12);
         true:  (len: shortint; chars: char12)
      end;

begin { check_action }
   fiddle.chars := '            ';      { space fill }
   fiddle.strng := word_str;
   to_low_case(fiddle.strng);
   word := fiddle.chars;
   action := action_proc.min;
   state := searching;
   repeat
      if action_proc_info[action].act_name = word
      then state := found
      else if action = action_proc.max
      then state := not_found
      else action := succ(action)
   until state <> searching;
   if (state = found) and (action in [sequence, emptyact])
   then state := not_found;             { disallow internal ones as user input }

   check_action := (state = found)
end { check_action } ;

{ Print the list of valid character names }

procedure list_char_names;                                                      { LIST_CHAR_NAMES }

var
   count: integer;
   ch: char;

begin { list_char_names }
   writeln('Possible character names are:');
   count := 0;
   for ch := chr(nul_code) to char.max do
   begin
      if count <= 0 then
      begin { end the line }
         count := 16;
         writeln;
      end { end the line } ;
      count := count - 1;
      write(char_names[ch],' ')
   end { for loop } ;
   writeln;
   writeln
end { list_char_names } ;

{ Store a key code in a character sequence }

procedure store_key_code(key_code: integer; var sequence: sequences);           { STORE_KEY_CODE }

   procedure store_code(len: integer; var ptr: ptr_seq_block);
   begin { store_code }
      if len >= seq_max
      then store_code(len-seq_max, ptr^.next_seq_block)
      else
      begin
         if (len = 0) and (ptr = nil) then
         begin
            new(ptr);
            ptr^.next_seq_block := nil
         end;
         with ptr^
         do seq_chs[len+1] := key_code
      end
   end { store_code } ;

begin { store_key_code }
   with sequence do
   begin
      store_code(seq_len, first_seq_block);
      seq_len := seq_len + 1
   end
end { store_key_code } ;

{ Get a character code from a sequence }

function get_code_from_sequence(sequence: sequences; seq_no: integer): integer;  { GET_CODE_FROM_SEQUENCE }

var
   sub: integer;
   ptr: ptr_seq_block;

begin { get_code_from_sequence }
   with sequence do
   begin
      if seq_no > seq_len
      then halt('get_code_from_sequence: out of range')
      else
      begin
         sub := seq_no;
         ptr := first_seq_block;
         while sub > seq_max do
         begin
            sub := sub - seq_max;
            ptr := ptr^.next_seq_block
         end;
         get_code_from_sequence := ptr^.seq_chs[sub]
      end
   end
end { get_code_from_sequence } ;

{ Append one sequence to another }

procedure append_sequence(from_seq: sequences; var to_seq: sequences);          { APPEND_SEQUENCE }

var
   chindex, ch_code: integer;

begin { append_sequence }
   for chindex := 1 to from_seq.seq_len do
   begin
      ch_code := get_code_from_sequence(from_seq, chindex);
      store_key_code(ch_code, to_seq)
   end
end { append_sequence } ;

{ Copy one sequence to another }

procedure copy_sequence(from_seq: sequences; var to_seq: sequences);            { COPY_SEQUENCE }

begin { copy_sequence }
   to_seq.seq_len := 0;
   append_sequence(from_seq, to_seq)
end { copy_sequence } ;

{ Write a char in visible form }

procedure write_char_name(ch_code: integer; var chars_written: integer);        { WRITE_CHAR_NAME }

var
   chname: char3;

begin { write_char_name }
   chname := char_names[chr(ch_code)];
   if chname[3] <> ' '
   then chars_written := 3
   else if chname[2] <> ' '
   then chars_written := 2
   else chars_written := 1;
   write(report_file, chname:chars_written)
end { write_char_name } ;

{ Write out a sequence of characters in source form }

procedure write_key_sequence(to_report: boolean;                                { WRITE_KEY_SEQUENCE }
                             the_sequence: sequences;
                             var chars_written: integer);

var
   sindex: integer;
   ch_code, the_code, len: integer;
   ch: char;

   { Write an integer as three octal digits }

   procedure write_octal(int: integer);

   var
      valu, power, digit: integer;
      index: 1..3;

   begin { write_octal }
      valu := int;
      power := 64;
      for index := 1 to 3 do
      begin
         digit := valu div power;
         valu := valu mod power;
         power := power div 8;
         write(report_file, digit:1)
      end;
      write(report_file,'b');
      chars_written := chars_written + 4
   end { write_octal } ;

begin { write_key_sequence }
   with the_sequence do
   begin
      chars_written := 0;
      if seq_len = 0 then
      begin
         if not to_report then
         begin
            write('null sequence');
            chars_written := 13
         end
      end
      else
      begin
         for sindex := 1 to seq_len do
         begin
            ch_code := get_code_from_sequence(the_sequence, sindex);
            if (ch_code >= 128) and (ch_code <= 255) then
            begin
               ch := chr(ch_code);
               { Print comment symbols as octal because of special meaning }
               if to_report and (ch in ['{','}'])
               then write_octal(ch_code)
               else
               begin
                  write_char_name(ch_code, len);
                  chars_written := chars_written + len
               end
            end { chcode 128..255 }
            else if ch_code <= 9 then
            begin
               the_code := ch_code;
               if the_code < 0 then
               begin
                  write(report_file,'-');
                  the_code := - the_code
               end
               else write(report_file,'+');
               write(report_file, the_code:1);
               chars_written := chars_written + 2;
               if the_code > 9
               then chars_written := chars_written + 1;
               if the_code > 99
               then chars_written := chars_written + 1
            end { ch_code <= 9 }
            else write_octal(ch_code);
            write(report_file,' ');
            chars_written := chars_written + 1;
            if chars_written > 70 then
            begin
               writeln(report_file);
               chars_written := 0
            end
         end { for loop }
      end { not null }
   end { with }
end { write_key_sequence } ;

{ Store a complete input sequence description in the expanded tables }

function store_input_sequence(the_sequence: sequences;                          { STORE_INPUT_SEQUENCE }
                              the_action: action_proc;
                              the_repair: repair_type;
                              the_ware: ware_type;
                              the_reply: sequences): boolean;

var
   {---teststart---}
{  chars_written: integer;              {----}
   {---teststop---}
   store_worked: boolean;

   { store in table }

   procedure store_sequence(var ptr: ptr_char_table; level: integer);

   var
      ch_code: integer;

      { Create a new input character table }

      procedure create_new_table;

      var
         ch: char;

      begin { create_new_table }
         new(ptr);
         for ch := chr(nul_code) to char.max
         do ptr^[ch] := init_full_entry
      end { create_new_table } ;

   begin { store_sequence }
      ch_code := get_code_from_sequence(the_sequence, level);
      if ptr = nil
      then create_new_table;            { set up new table }
      with vdu_table_rec, expanded, ptr^[chr(ch_code)] do
      begin
         if the_sequence.seq_len = level then
         begin { this is the end of the sequence }
            if full_action = emptyact
            then full_action := the_action
            else if full_action = sequence
                 then if the_action in [posmacro, negmacro]
                      then full_action := the_action
                      else
                      begin
                         writeln;
                         writeln('This input sequence is the same as the beginning of an earlier one.');
                         store_worked := false
                      end
                 else
                 begin
                    writeln;
                    writeln('You are attempting to redefine a previous sequence.');
                    store_worked := false
                 end;
            if store_worked then
            begin
               if full_action = posmacro then
               begin
                  posmacro_repair := the_repair;
                  copy_sequence(the_reply, exp_posmacro_reply)
               end
               else if full_action = negmacro then
               begin
                  negmacro_repair := the_repair;
                  copy_sequence(the_reply, exp_negmacro_reply)
               end
               else
               begin
                  full_repair := the_repair;
                  full_ware := the_ware;
                  copy_sequence(the_reply, full_reply)
               end
            end
         end { this is the end of the sequence }
         else if full_action in [sequence, posmacro, negmacro, emptyact] then
         begin
            if full_action = emptyact
            then full_action := sequence;
            store_sequence(table_ptr, level+1)
         end
         else
         begin
            writeln;
            writeln('The first part of this sequence has already been defined.');
            store_worked := false
         end
      end
   end { store_sequence } ;

   procedure sleep$(msecs: integer); extern;

begin { store_input_sequence }
   with expanded do
   begin
      store_worked := true;
      {---teststart---}
{     if interactive then               {----}
{     begin                             {----}
{        write('storing: ');            {----}
{        write_key_sequence(false, the_sequence, chars_written);  {----}
{        write(' ',action_proc_info[the_action].act_name,  {----}
{              ' ',repair_names[the_repair],  {----}
{              ' ',ware_names[the_ware],  {----}
{              ' ');                    {----}
{        write_key_sequence(true, the_reply, chars_written);  {----}
{        writeln(' //');                {----}
{     end;                              {----}
      {---teststop---}
      if the_sequence.seq_len = 0 then
      begin
         writeln;
         writeln('The input sequence is null.');
         store_worked := false
      end
      else if the_action = posmacro then
      begin
         if exp_posmacro_defined then
         begin
            writeln;
            writeln('Action "posmacro" has already been defined and may not be redefined.');
            store_worked := false
         end
         else exp_posmacro_defined := true
      end
      else if the_action = negmacro then
      begin
         if exp_negmacro_defined then
         begin
            writeln;
            writeln('Action "negmacro" has already been defined and may not be redefined.');
            store_worked := false
         end
         else exp_negmacro_defined := true
      end;
      if store_worked
      then store_sequence(exp_input_sequences_head, 1);
      store_input_sequence := store_worked
   end { with }
end { store_input_sequence } ;

{ Find a given action in the input sequence tables }
{ Returns null string if not found }

procedure find_input_sequence(the_action: action_proc;                          { FIND_INPUT_SEQUENCE }
                              var input_sequence: sequences);

var
   result: search_type;

   function find_sequence(var ptr: ptr_char_table): search_type;

   var
      ch: char;
      state: search_type;

   begin { find_sequence }
      if ptr = nil
      then state := not_found           { not defined - ignore }
      else
      begin
         ch := chr(nul_code);
         state := searching;
         repeat
            store_key_code(ord(ch), input_sequence);
            with ptr^[ch] do
            begin
               if full_action = the_action
               then state := found
               else if full_action in [sequence, posmacro, negmacro] then
               begin
                  if find_sequence(table_ptr) = found
                  then state := found   { found lower down }
               end;
               if state = searching then
               begin
                  if ch < char.max
                  then ch := succ(ch)
                  else state := not_found
               end
            end;
            if state <> found
            then with input_sequence
                 do seq_len := seq_len - 1
         until state <> searching
      end;
      find_sequence := state
   end { find_sequence } ;

begin { find_input_sequence }
   input_sequence.seq_len := 0;
   result := find_sequence(expanded.exp_input_sequences_head)
end { find_input_sequence } ;

{ Find whether a given input sequence is defined }

function sequence_defined(the_sequence: sequences): boolean;                    { SEQUENCE_DEFINED }

   procedure scan_table(var ptr: ptr_char_table; level: integer);

   var
      ch_code: integer;

   begin { scan_table }
      ch_code := get_code_from_sequence(the_sequence, level);
      if ptr = nil
      then                              { not defined }
      else with expanded, ptr^[chr(ch_code)] do
      begin
         if the_sequence.seq_len = level then
         begin { end of sequence }
            if full_action in [emptyact, erroract]
            then                        { not defined }
            else sequence_defined := true  { same or longer sequence defined }
         end { end of sequence }
         else if full_action in [sequence, posmacro, negmacro]
         then scan_table(table_ptr, level+1)
         else if full_action in [emptyact, erroract]
         then                           { alright to extend }
         else sequence_defined := true  { shorter sequence defined }
      end
   end { scan_table } ;

begin { sequence_defined }
   sequence_defined := false;
   if the_sequence.seq_len > 0
   then scan_table(expanded.exp_input_sequences_head, 1)
   else                                 { null sequence }
end { sequence_defined } ;

{ Delete an input sequence from the expanded tables }

procedure delete_input_sequence(the_sequence: sequences;                        { DELETE_INPUT_SEQUENCE }
                                delete_inferiors: boolean);

   {---teststart---}
{var                                    {----}
{  chars_written: integer;              {----}
   {---teststop---}

   procedure delete_sequence(var ptr: ptr_char_table; level: integer);

   var
      ch_code: integer;

   begin { delete_sequence }
      ch_code := get_code_from_sequence(the_sequence, level);
      if ptr = nil
      then                              { not defined - ignore }
      else with expanded, ptr^[chr(ch_code)] do
      begin
         if the_sequence.seq_len = level then
         begin { end of sequence }
            if full_action = posmacro then
            begin
               full_action := sequence;
               exp_posmacro_defined := false
            end
            else if full_action = negmacro then
            begin
               full_action := sequence;
               exp_negmacro_defined := false
            end
            else if full_action = sequence
            then                        { first part of longer seq }
            else full_action := emptyact;
            if full_action = sequence then
            begin { delete inferiors if asked }
               if delete_inferiors
               then empty_input_table(table_ptr);
               if table_ptr = nil
               then full_action := emptyact
            end { delete inferiors if asked }
         end { end of sequence }
         else if full_action in [sequence, posmacro, negmacro]
         then delete_sequence(table_ptr, level+1)
         else                           { not found }
      end
   end { delete_sequence } ;

begin { delete_input_sequence }
   if the_sequence.seq_len > 0 then
   begin
      {---teststart---}
{     write('delete_input_sequence: ');  {----}
{     write_key_sequence(false, the_sequence, chars_written);  {----}
{     if delete_inferiors then write('  and inferiors');  {----}
{     writeln;                          {----}
      {---teststop---}
      delete_sequence(expanded.exp_input_sequences_head, 1)
   end
   else { just ignore null sequence }
end { delete_input_sequence } ;

{ Delete all input sequences with actions in a given set }

procedure dl_in_set(the_set: set_of_action_proc);                               { DL_IN_SET }

   procedure delete_set(var ptr: ptr_char_table);

   var
      ch: char;

   begin { delete_set }
      if ptr = nil
      then                              { not defined - ignore }
      else
      with expanded do
      begin
         for ch := chr(nul_code) to char.max do
         with ptr^[ch] do
         begin
            if full_action in the_set then
            begin { found one }
               if full_action = posmacro then
               begin
                  full_action := sequence;
                  exp_posmacro_defined := false
               end
               else if full_action = negmacro then
               begin
                  full_action := sequence;
                  exp_negmacro_defined := false
               end
               else if full_action = sequence
               then                     { first part of longer seq }
               else full_action := emptyact;
               if (full_action = sequence) and (table_ptr = nil)
               then full_action := emptyact
            end { found one } ;
            if full_action in [sequence, posmacro, negmacro]
            then delete_set(table_ptr)
         end { for loop }
      end { table exists }
   end { delete_set } ;

begin { dl_in_set }
   delete_set(expanded.exp_input_sequences_head)
end { dl_in_set } ;

{ Count function key entries and delete if requested }

procedure count_func(var count: integer; delete_them: boolean);                 { COUNT_FUNC }

   procedure count_table(var ptr: ptr_char_table);

   var
      ch: char;

   begin { count_table }
      if ptr = nil
      then                              { not defined - ignore }
      else
      with expanded do
      begin
         for ch := chr(nul_code) to char.max do
         with ptr^[ch] do
         begin
            if full_repair = func then
            begin { found one }
               count := count + 1;
               if delete_them
               then ptr^[ch] := init_full_entry
            end { found one }
            else if full_action in [sequence, posmacro, negmacro]
            then count_table(table_ptr)
         end { for loop }
      end { table exists }
   end { count_table } ;

begin { count_func }
   count := 0;
   count_table(expanded.exp_input_sequences_head)
end { count_func } ;

{ Write out a source representation of the expanded vdu information }

procedure write_source_file;                                                    { WRITE_SOURCE_FILE }

const
   slash_position = 28;
   com_position   = 41;
   input_position = 13;

var
   control: controls;
   input_sequence: sequences;

   procedure write_a_sequence(the_sequence: sequences; the_id: char12);

   var
      chars_written, spaces_needed: integer;

   begin { write_a_sequence }
      write_key_sequence(true, the_sequence, chars_written);
      spaces_needed := slash_position - chars_written - 1;
      if spaces_needed < 2
      then spaces_needed := 2;
      write(report_file,' ':spaces_needed);
      chars_written := chars_written + spaces_needed;
      write(report_file,'//');
      chars_written := chars_written + 2;
      spaces_needed := com_position - chars_written - 1;
      if spaces_needed < 2
      then spaces_needed := 2;
      write(report_file,' ':spaces_needed);
      writeln(report_file,'{ ',the_id,' }')
   end { write_a_sequence } ;

   procedure write_input_table(ptr: ptr_char_table);

   var
      ch: char;

      procedure print_the_sequence(the_entry: full_entry);

      var
         chars_written, spaces_needed: integer;
         repair_name, ware_name: char12;
         reply: sequences;

      begin { print_the_sequence }
         with vdu_table_rec, expanded, the_entry, input_sequence do
         begin
            if (seq_len = 1) and (full_action = storeonech)
            then  { do not write out default characters }
            else
            begin
               write_key_sequence(true, input_sequence, chars_written);
               spaces_needed := input_position - chars_written - 1;
               if spaces_needed < 2
               then spaces_needed := 2;
               write(report_file,' ':spaces_needed);
               write(report_file,action_proc_info[full_action].act_name);
               if full_action = posmacro then
               begin
                  repair_name := repair_names[posmacro_repair];
                  ware_name := ware_names[software];
                  reply := exp_posmacro_reply
               end
               else if full_action = negmacro
               then
               begin
                  repair_name := repair_names[negmacro_repair];
                  ware_name := ware_names[software];
                  reply := exp_negmacro_reply
               end
               else
               begin
                  repair_name := repair_names[full_repair];
                  ware_name := ware_names[full_ware];
                  reply := full_reply
               end;
               write(report_file,repair_name:8,ware_name);
               write_key_sequence(true, reply, chars_written);
               writeln(report_file,'//')
            end
         end
      end { print_the_sequence } ;

   begin { write_input_table }
      for ch := chr(nul_code) to char.max do
      with ptr^[ch] do
      begin
         store_key_code(ord(ch), input_sequence);
         case full_action of
            sequence: write_input_table(table_ptr);
            posmacro, negmacro:
               begin
                  print_the_sequence(ptr^[ch]);
                  if table_ptr <> nil
                  then write_input_table(table_ptr)
               end;
            emptyact, erroract:         { do nothing }
            otherwise print_the_sequence(ptr^[ch])
         end;
         with input_sequence
         do seq_len := seq_len - 1
      end
   end { write_input_table } ;

begin { write_source_file }
   with vdu_table_rec, expanded do
   begin
      writeln(report_file);
      writeln(report_file,vdu_identification);
      writeln(report_file);
      for control := controls.min to controls.max
      do write_a_sequence(exp_control_array[control],
                          vdu_control_names[control]);
      write_a_sequence(exp_reset_string ,'reset       ');
      write_a_sequence(exp_exit_string  ,'exit        ');
      write_a_sequence(exp_cursor_string,'cursor      ');
      writeln(report_file);
      writeln(report_file,exp_mess_type:1-input_position,
                          '{ mess_type          }');
      writeln(report_file,exp_arrow_type:1-input_position,
                          '{ arrow_type         }');
      writeln(report_file, func_expect_return:1-input_position,
                          '{ func_expect_return }');
      writeln(report_file,repair_names[func_repair],' ':input_position-13,
                          '{ func_repair        }');
      writeln(report_file,cols_per_row:-5,' ':input_position-6,
                          '{ cols_per_row       }');
      writeln(report_file,rows_per_page:-5,' ':input_position-6,
                          '{ rows_per_page      }');
      with home_location
      do writeln(report_file,home_row:-4,home_col:-4,' ':input_position-9,
                          '{ home location      }');
      writeln(report_file, right_pos_ok:1-input_position,
                          '{ right_pos_ok       }');
      writeln(report_file, right_prt_ok:1-input_position,
                          '{ right_prt_ok       }');
      writeln(report_file, left_pos_ok :1-input_position,
                          '{ left_pos_ok        }');
      writeln(report_file, up_pos_ok   :1-input_position,
                          '{ up_pos_ok          }');
      writeln(report_file, down_pos_ok :1-input_position,
                          '{ down_pos_ok        }');
      writeln(report_file, data_protect:1-input_position,
                          '{ data_protect       }');
      writeln(report_file);
      writeln(report_file,'{ Input sequences }');
      writeln(report_file);
      input_sequence := init_sequence;
      if exp_input_sequences_head <> nil
      then write_input_table(exp_input_sequences_head);
      dispose_key_sequence(input_sequence);
      writeln(report_file,'//');
      writeln(report_file)
   end { with }
end { write_source_file } ;

{ Check whether the input sequence being stored looks like a command from the }
{ standard set }

procedure check_for_commands(input_sequence: sequences;                         { CHECK_FOR_COMMANDS }
                             the_action: action_proc;
                             the_repair: repair_type;
                             the_reply: sequences);

   {---teststart---}
{var                        {----}
{  chars_written: integer;  {----}
   {---teststop---}

begin { check_for_commands }
   with expanded do
   begin
      if (the_action = ispaces) and (the_repair <> func) and
         (input_sequence.seq_len > 1)
      then
      begin
         copy_sequence(input_sequence, exp_poscommand);
         with exp_poscommand            { trim off the last character }
         do seq_len := seq_len - 1;
         {---teststart---}
{        write('"normal  " command is ');  {----}
{        write_key_sequence(false, exp_poscommand, chars_written);  {----}
{        writeln;  {----}
         {---teststop---}
         exp_poscommand_repair := the_repair;
         copy_sequence(the_reply, exp_poscommand_reply);
         {---teststart---}
{        writeln('repair = ',repair_names[exp_poscommand_repair]);  {----}
{        write('reply is ');            {----}
{        write_key_sequence(false, exp_poscommand_reply, chars_written);  {----}
{        writeln  {----}
         {---teststop---}
      end;
      if (the_action = dspaces) and (the_repair <> func) and
         (input_sequence.seq_len > 1)
      then
      begin
         copy_sequence(input_sequence, exp_negcommand);
         with exp_negcommand
         do seq_len := seq_len - 1;
         {---teststart---}
{        write('"opposite" command is ');  {----}
{        write_key_sequence(false, exp_negcommand, chars_written);  {----}
{        writeln;  {----}
         {---teststop---}
         exp_negcommand_repair := the_repair;
         {---teststart---}
{        writeln('repair = ',repair_names[exp_negcommand_repair]);  {----}
         {---teststop---}
         copy_sequence(the_reply, exp_negcommand_reply);
         {---teststart---}
{        write('reply is ');            {----}
{        write_key_sequence(false, exp_negcommand_reply, chars_written);  {----}
{        writeln  {----}
         {---teststop---}
      end;
      if (the_action = posmacro) then
      begin
         copy_sequence(input_sequence, exp_posmacro);
         {---teststart---}
{        write('"normal  " macro   is ');  {----}
{        write_key_sequence(false, exp_posmacro, chars_written);  {----}
{        writeln  {----}
         {---teststop---}
      end;
      if (the_action = negmacro) then
      begin
         copy_sequence(input_sequence, exp_negmacro);
         {---teststart---}
{        write('"opposite" macro   is ');  {----}
{        write_key_sequence(false, exp_negmacro, chars_written);  {----}
{        writeln  {----}
         {---teststop---}
      end
   end { with }
end { check_for_commands } ;

{ Expand vdu_table_rec into expanded strings and input tables }

procedure expand_vdu_info;                                                      { EXPAND_VDU_INFO }

var
   control:controls;
   ch: char;
   input_sequence: sequences;

   procedure expand_sequence(comp_seq: vdu_char_pointer;
                             var exp_seq: sequences);

   var
      chindex: vdu_char_range;

   begin { expand_sequence }
      exp_seq.seq_len := 0;             { ? }
      with comp_seq, vdu_table_rec do
      begin
         for chindex := start_index to start_index+string_length-1
         do store_key_code(two_way.char_array[chindex], exp_seq)
      end
   end { expand_sequence } ;

   procedure copy_table(var tab_ptr: ptr_char_table; ch_ptr: vdu_entry_pointer);

   var
      ch: char;
      eindex: vdu_entry_range;

   begin { copy_table }
      new(tab_ptr);
      for ch := chr(nul_code) to char.max
      do tab_ptr^[ch] := init_full_entry;
      with vdu_table_rec, ch_ptr do
      begin
         for eindex := start_index downto start_index-number_of_entries+1 do
         with two_way.entry_array[eindex], tab_ptr^[entry_char], entry_rest do
         begin
            store_key_code(ord(entry_char), input_sequence);
            full_action := char_action;
            if full_action in [posmacro, negmacro, sequence]
            then copy_table(table_ptr, char_pointer)
            else
            begin
               full_repair := char_repair;
               full_ware := char_ware;
               expand_sequence(char_reply, full_reply);
               table_ptr := nil
            end;
            check_for_commands(input_sequence, full_action, full_repair,
                               full_reply);
            with input_sequence
            do seq_len := seq_len - 1
         end { for eindex }
      end { with }
   end { copy_table } ;

begin { expand_vdu_info }
   with vdu_table_rec, expanded do
   begin
      for control := controls.min to controls.max
      do expand_sequence(control_array[control],
                         exp_control_array[control]);
      expand_sequence(reset_string, exp_reset_string);
      expand_sequence(exit_string, exp_exit_string);
      expand_sequence(cursor_string, exp_cursor_string);
      expand_sequence(posmacro_reply, exp_posmacro_reply);
      expand_sequence(negmacro_reply, exp_negmacro_reply);

      if mess_type1
      then exp_mess_type := 1
      else exp_mess_type := 2;
      if arrow_type1
      then exp_arrow_type := 1
      else exp_arrow_type := 2;
      exp_char_protect := data_protect;
      exp_line_protect := data_protect;

      new(exp_input_sequences_head);
      input_sequence := init_sequence;
      for ch := chr(nul_code) to char.max do
      with exp_input_sequences_head^[ch], action_array[ord(ch)] do
      begin
         exp_input_sequences_head^[ch] := init_full_entry;
         input_sequence.seq_len := 0;
         store_key_code(ord(ch), input_sequence);
         if char_action in [erroract, storeonech]
         then { leave it empty }
         else
         begin { not erroract or storeonech }
            full_action := char_action;
            if full_action in [posmacro, negmacro, sequence]
            then copy_table(table_ptr, char_pointer)
            else
            begin
               full_repair := char_repair;
               full_ware := char_ware;
               full_reply := init_sequence;
               expand_sequence(char_reply, full_reply);
               table_ptr := nil
            end;
            check_for_commands(input_sequence, full_action, full_repair,
                               full_reply)
         end { not erroract or storeonech }
      end { for loop, with } ;
      dispose_key_sequence(input_sequence)
   end { with vdu_table_rec }
end { expand_vdu_info } ;

{ Compress all the vdu information into the binary record }

function compress_vdu_info: boolean;                                            { COMPRESS_VDU_INFO }

label 521;

var
   current_char_start, current_entry_start, spaces_left: integer;
   control: controls;
   vdu_pt: vdu_entry_pointer;           { not set because top level stored }
                                        { in array not pointer }

   { Store a sequence of characters in the binary file }

   procedure store_sequence_in_binary(the_sequence: sequences;
                                      var the_pointer: vdu_char_pointer);

   var
      sindex, ch_code: integer;

   begin { store_sequence_in_binary }
      with vdu_table_rec, the_sequence, the_pointer do
      begin
         start_index := current_char_start;  { initialise the pointer }
         string_length := 0;
         if spaces_left < 1 then
         begin
            writeln;
            writeln('The area which contains character codes is full.');
            writeln('The length of the area is ',vdu_chars_max:1);
            goto 521
         end
         else
         begin
            spaces_left := spaces_left - seq_len;
            for sindex := 1 to the_sequence.seq_len do
            begin
               ch_code := get_code_from_sequence(the_sequence, sindex);
               two_way.char_array[current_char_start] := ch_code;
               current_char_start := current_char_start + 1;
               string_length := string_length + 1
            end { for }
         end
      end { with }
   end { store_sequence_in_binary } ;

   { Store an ASCII table of input characters in the vdu_info record }

   procedure store_table(var ptr: ptr_char_table;
                         level: integer;
                         var vdu_pt: vdu_entry_pointer);

   var
      ch: char;

      { Store a input character entry in the vdu tables }

      procedure store_char_plus_entry(var pointer: vdu_entry_pointer;
                                      the_ch: char;
                                      the_rec: full_entry);

      begin { store_char_plus_entry }
         with pointer, vdu_table_rec do
         begin
            if spaces_left < entry_length then
            begin
               writeln;
               writeln('The area which contains character codes is full.');
               writeln('The length of the area is ',vdu_chars_max:1);
               goto 521
            end
            else
            begin
               spaces_left := spaces_left - entry_length;
               with two_way.entry_array[current_entry_start], entry_rest,
                    the_rec do
               begin
                  entry_char := the_ch;
                  char_action := full_action;
                  if char_action in [sequence, posmacro, negmacro]
                  then char_pointer := full_pointer
                  else
                  begin
                     char_repair := full_repair;
                     char_ware   := full_ware;
                     char_reply  := empty_char_pointer;
                     store_sequence_in_binary(full_reply, char_reply)
                  end
               end;
               current_entry_start := current_entry_start - 1;
               number_of_entries := number_of_entries + 1
            end
         end
      end { store_char_plus_entry } ;

   begin { store_table }
      for ch := chr(nul_code) to char.max do
      with ptr^[ch] do
      begin
         if (full_action in [sequence, posmacro, negmacro]) and
            (table_ptr <> nil)
         then store_table(table_ptr, level+1, full_pointer)
      end;
      if level > 1 then
      with vdu_pt do
      begin { not top level - store entry list }
         start_index := current_entry_start;
         number_of_entries := 0;
         for ch := chr(nul_code) to char.max do
         with ptr^[ch] do
         begin
            if full_action <> emptyact
            then store_char_plus_entry(vdu_pt, ch, ptr^[ch])
         end
      end { not top level - store entry list }
      else
      begin { top level - store in action_array }
         for ch := chr(nul_code) to char.max do
         with vdu_table_rec, ptr^[ch], action_array[ord(ch)] do
         begin
            if full_action = emptyact then
            begin { fill in default }
               if ch in ascii_printable
               then full_action := storeonech
               else full_action := erroract;
               full_repair := ok;
               full_ware := hardware
            end { fill in default } ;
            char_action := full_action;
            if full_action in [sequence, posmacro, negmacro]
            then char_pointer := full_pointer
            else
            begin
               char_repair := full_repair;
               char_ware   := full_ware;
               char_reply  := empty_char_pointer;
               store_sequence_in_binary(full_reply, char_reply)
            end
         end { for loop }
      end { top level table }
   end { store_table } ;

begin { compress_vdu_info }
   compress_vdu_info := false;
   current_char_start := 1;
   current_entry_start := vdu_entry_max;
   spaces_left := vdu_chars_max;
   with vdu_table_rec, expanded do
   begin
      for control := controls.min to controls.max
      do store_sequence_in_binary(exp_control_array[control],
                                  control_array[control]);
      store_sequence_in_binary(exp_reset_string, reset_string);
      store_sequence_in_binary(exp_exit_string, exit_string);
      store_sequence_in_binary(exp_cursor_string, cursor_string);
      store_sequence_in_binary(exp_posmacro_reply, posmacro_reply);
      store_sequence_in_binary(exp_negmacro_reply, negmacro_reply);
      mess_type1 := exp_mess_type = 1;
      arrow_type1 := exp_arrow_type = 1;
      data_protect := exp_char_protect or exp_line_protect;
      if exp_input_sequences_head <> nil
      then store_table(exp_input_sequences_head, 1, vdu_pt)
   end;
   compress_vdu_info := true;
   writeln(vdu_chars_max-spaces_left:1,' code spaces used out of ',
           vdu_chars_max:1,'.');
521:
   { exit for errors, leaving function result false }
end { compress_vdu_info } ;

{ Read a single source file and process it into an output file }

function read_one_source(write_output_files: boolean): boolean;                 { READ_ONE_SOURCE }

label 522;

{ Print an error message and abort }

procedure error_exit;                                                           { ERROR_EXIT }

begin { error_exit }
   writeln('Execution aborted at line ',line_no:1,' column ',in_pos:1,
           ' of the input file ',source_fn);
   if string_name = '            '
   then writeln(extra_information)
   else writeln('while reading output string "',string_name,'".');
   goto 522
end { error_exit } ;

{ Read a line of text from the input file }

procedure read_in_line;                                                         { READ_IN_LINE }

var
   done: boolean;

begin { read_in_line }
   if eof(source_file) then
   begin
      writeln('End of file found while reading input file ',source_fn);
      writeln('Check for missing // terminator(s)');
      goto 522
   end
   else
   begin
      readln(source_file, in_line: in_length);
      done := false;
      repeat
         if in_length > 0
         then if in_line[in_length] = ' '
              then in_length := in_length - 1
              else done := true
         else done := true
      until done;
      line_no := line_no + 1;
      in_pos := 1
   end
end { read_in_line } ;

{ Skip any blank lines }

procedure skip_blank_lines;                                                     { SKIP_BLANK_LINES }

var
   done: boolean;

begin { skip_blank_lines }
   done := false;
   repeat
      if in_length > 0
      then done := true
      else if eof(source_file)
      then done := true
      else read_in_line
   until done
end { skip_blank_lines } ;

{ Skip spaces and vdu_punctuation }

procedure skip_spaces;                                                          { SKIP_SPACES }

var
   ch: char;
   done: boolean;

   procedure get_next_char;
   begin { get_next_char }
      if in_pos > in_length
      then
         repeat read_in_line
         until in_pos <= in_length
      else in_pos := in_pos + 1
   end { get_next_char } ;

   procedure skip_a_comment;
   begin { skip_a_comment }
      repeat in_pos := in_pos + 1
      until (in_line[in_pos] = '}') or  { end of comment or }
            (in_pos > in_length);       { end of line }
      get_next_char
   end { skip_a_comment } ;

begin { skip_spaces }
   if in_pos > in_length
   then get_next_char;
   done := false;
   repeat
      ch := in_line[in_pos];
      if ch in vdu_punctuation
      then if ch = '{'
      then skip_a_comment
      else get_next_char                { skip spaces and form feeds }
      else done := true
   until done
end { skip_spaces } ;

{ Get the next word from the input file }

procedure get_next_word(var word: string12);                                    { GET_NEXT_WORD }

var
   word_length: 0..12;
   done: boolean;

begin { get_next_word }
   skip_spaces;
   word_length := 0;
   word := '            ';
   done := false;
   repeat
      if word_length = 12 then
      begin
         writeln;
         writeln('Word beginning "', word, '" is too long.');
         error_exit
      end
      else
      begin
         word_length := word_length + 1;
         word[word_length] := in_line[in_pos];
         in_pos := in_pos + 1;
         if in_pos > in_length
         then done := true              { end of line }
         else if in_line[in_pos] in vdu_punctuation
         then done := true
      end
   until done;
   word := substr(word, 1, word_length)
end { get_next_word } ;

{ Convert a word to integer, allowing for 'B' suffix if present }

function convert(word: string12): integer;                                      { CONVERT }

var
   len, start_sub, windex: integer;
   factor, valu: integer;
   ch: char;
   positive: boolean;

begin { convert }
   len := length(word);
   if word[len] in ['b', 'B'] then
   begin { B suffix => octal }
      factor := 8;
      len := len - 1
   end { B suffix => octal }
   else factor := 10;                   { normally decimal }
   ch := word[1];
   if ch in ['-', '+'] then
   begin
      positive := ch = '+';
      start_sub := 2
   end
   else
   begin
      positive := true;
      start_sub := 1
   end;
   valu := 0;
   for windex := start_sub to len do
   begin
      ch := word[windex];
      if ch in ['0'..'9']
      then valu := valu*factor + ord(ch) - ord('0')
      else
      begin { non-numeric }
         writeln;
         writeln('"', ch, '" found when digit expected.');
         error_exit
      end { non-numeric }
   end { for loop } ;
   if positive
   then convert := valu
   else convert := -valu
end { convert } ;

{ Read a word and try to recognise a character symbol }

procedure read_char_symbol(var sym_text: string12;                              { READ_CHAR_SYMBOL }
                           var ch_code: integer;
                           allow_neg: boolean;
                           var recognised: boolean);

var
   ch: char;

   procedure numeric_code;

   begin { numeric_code }
      ch_code := convert(sym_text);
      if (ch_code <= 377b) and (ch_code >= 0)
      then { normal range 0..377b }
      else if ch_code < -100
      then { -101 etc. for user hook }
      else if allow_neg and (ch_code < 0)
      then { -1 etc. allowed for position }
      else
      begin { out of range }
         writeln;
         writeln('Character code ',ch_code:1,' is out of range.');
         if ch_code < 0 then
         begin { value < 0 }
            writeln('-1, -2 etc. are only allowed in the cursor addressing string.');
            writeln('Check that all preceding strings have their // terminator.')
         end { value < 0 } ;
         error_exit
      end { out of range }
   end { numeric_code } ;

   procedure ascii_char_name;

   var
      long_word: string;
      index: 1..3;
      ch: char;
      ch3: char3;
      state: search_type;

   begin { ascii_char_name }
      if length(sym_text) > 3
      then recognised := false
      else
      begin { <= 3 chars }
         long_word := concat(sym_text,'   ');
         for index := 1 to 3 do
         begin
            ch := long_word[index];
            if ch in ['a'..'z']
            then ch := chr(ord('A') - ord('a') + ord(ch));
            ch3[index] := ch
         end;
         ch_code := 200b;
         state := searching;
         repeat
            if ch3 = char_names[chr(ch_code)]
            then state := found
            else if ch_code < 237b
            then ch_code := ch_code + 1
            else state := not_found
         until state <> searching;
         if ch3 = char_names[' '] then
         begin
            ch_code := ord(' ');
            state := found
         end
         else if ch3 = char_names[chr(377b)] then
         begin
            ch_code := 377b;
            state := found
         end
         else if ch3 = 'PSE' then
         begin
            ch_code := 0;
            state := found
         end;
         if state = not_found
         then recognised := false
      end { <= 3 chars }
   end { ascii_char_name } ;

begin { read_char_symbol }
   recognised := true;
   get_next_word(sym_text);
   ch := sym_text[1];
   if length(sym_text) = 1
   then ch_code := ord(sym_text[1])
   else if ch in ['-','+','0'..'9']
   then numeric_code
   else ascii_char_name
end { read_char_symbol } ;

{ Read a word and interpret it a an integer }

procedure read_integer(var the_int: integer);                                   { READ_INTEGER }

var
   word: string12;

begin { read_integer }
   get_next_word(word);
   the_int := convert(word)
end { read_integer } ;

{ Read a word and interpret as a Boolean }

procedure read_boolean(var the_bool: boolean);                                  { READ_BOOLEAN }

var
   word: string12;

begin { read_boolean }
   get_next_word(word);
   to_low_case(word);
   if word = 'true'
   then the_bool := true
   else if word = 'false'
   then the_bool := false
   else
   begin
      writeln;
      writeln('"',word,'" found when "true" or "false" expected.');
      error_exit
   end
end { read_boolean } ;

{ Read a word and interpret it as the name of an ware_type }

procedure read_ware(var ware: ware_type);                                       { READ_WARE }

var
   state: search_type;
   word: char12;
   fiddle: record
      case boolean of
         false: (strng: string12);
         true:  (len: shortint; chars: char12)
      end;

begin { read_ware }
   get_next_word(fiddle.strng);
   to_low_case(fiddle.strng);
   word := fiddle.chars;
   ware := ware_type.min;
   state := searching;
   repeat
      if ware_names[ware] = word
      then state := found
      else if ware = ware_type.max
      then state := not_found
      else ware := succ(ware)
   until state <> searching;

   if state <> found
   then
   begin
      writeln;
      writeln('"',word,'" found when "hardware" or "software" expected.');
      error_exit
   end
end { read_ware } ;

{ Read a word and interpret it as the name of a repair_type }

procedure read_repair(var repair: repair_type);                                 { READ_REPAIR }

var
   rep: repair_type;
   state: search_type;
   word: char12;
   fiddle: record
      case boolean of
         false: (strng: string12);
         true:  (len: shortint; chars: char12)
      end;

begin { read_repair }
   get_next_word(fiddle.strng);
   to_low_case(fiddle.strng);
   word := fiddle.chars;
   repair := repair_type.min;
   state := searching;
   repeat
      if repair_names[repair] = word
      then state := found
      else if repair = repair_type.max
      then state := not_found
      else repair := succ(repair)
   until state <> searching;

   if state <> found then
   begin { not found }
      writeln;
      writeln('"',word,'" found when "repair" name expected.');
      writeln('Possible names are:');
      for rep := repair_type.min to repair_type.max
      do write(repair_names[rep]);
      writeln;
      writeln;
      error_exit
   end { not found }
end { read_repair } ;

{ Read a string of control characters }

procedure read_control_string(var the_sequence: sequences;                      { READ_CONTROL_STRING }
                              allow_neg: boolean);

var
   done, recognised: boolean;
   ch_code: integer;
   sym_text: string12;

begin { read_control_string }
   with vdu_table_rec do
   begin
      the_sequence.seq_len := 0;
      done := false;
      repeat
         read_char_symbol(sym_text, ch_code, allow_neg, recognised);
         if recognised
         then store_key_code(ch_code, the_sequence)
         else if sym_text = '//'
         then done := true
         else
         begin
            writeln;
            writeln('Characters "',sym_text,
                    '" found when a character code was expected.');
            writeln('Possible character names are:');
            list_char_names;
            error_exit
         end
      until done
   end
end { read_control_string } ;

{ Read the VDU identification string }

procedure read_vdu_ident;                                                       { READ_VDU_IDENT }

var
   tindex: tree_range;

begin { read_vdu_ident }
   with vdu_table_rec do
   begin
      extra_information := 'while reading VDU identification line.';
      vdu_identification := '                                                                                ';
      for tindex := 1 to 80
      do vdu_identification[tindex] := in_line[tindex];
      if in_length > 80
      then in_length := 80;
      vdu_identification := substr(vdu_identification, 1, in_length);
      writeln('Terminal type', vdu_type_count:3, ':  ',
              vdu_identification);
      read_in_line
   end
end { read_vdu_ident } ;

{ Read the sequences of control characters }

procedure read_control_chars;                                                   { READ_CONTROL_CHARS }

var
   control: controls;

begin { read_control_chars }
   with vdu_table_rec, expanded do
   begin
      for control := controls.min to controls.max do
      begin
         string_name := vdu_control_names[control];
         read_control_string(exp_control_array[control], false)
      end;
      string_name := '            ';
      extra_information := 'while reading "vdu reset string".';
      read_control_string(exp_reset_string , false);
      extra_information := 'while reading "vdu exit string".';
      read_control_string(exp_exit_string  ,false);
      extra_information := 'while reading "cursor addressing string".';
      read_control_string(exp_cursor_string,true )
   end
end { read_control_chars } ;

{ Read various odds and ends }

procedure read_bits;                                                            { READ_BITS }

   function read_named_integer(name: char12;
                               min_val, max_val: integer): integer;

   var
      the_int: integer;

   begin { read_named_integer }
      read_integer(the_int);
      if (the_int < min_val) or (the_int > max_val) then
      begin
         writeln;
         writeln(the_int,' is out of range.');
         writeln('It must be in the range ',min_val:1,' to ',max_val:1);
         error_exit
      end;
      read_named_integer := the_int
   end { read_named_integer } ;

begin { read_bits }
   with vdu_table_rec, expanded do
   begin
      extra_information := 'while reading "messtype".';
      exp_mess_type := read_named_integer('messtype    ', 1, 2);
      extra_information := 'while reading "arrowtype".';
      exp_arrow_type := read_named_integer('arrowtype   ', 1, 2);
      extra_information := 'while reading "funcexpectreturn".';
      read_boolean(func_expect_return);
      writeln(report_file);
      extra_information := 'while reading "funcrepair".';
      read_repair(func_repair);
      extra_information := 'while reading "colsperrow".';
      cols_per_row  := read_named_integer('colsperrow  ', 1, cols_max);
      extra_information := 'while reading "rowsperpage".';
      rows_per_page := read_named_integer('rowsperpage ', 1, rows_max);
      with home_location do
      begin
         extra_information := 'while reading "home.row".';
         home_row := read_named_integer('home.row    ', 1, rows_per_page);
         extra_information := 'while reading "home.col".';
         home_col := read_named_integer('home.col    ', 1, cols_per_row )
      end;
      extra_information := 'while reading "rightposok".';
      read_boolean(right_pos_ok);
      extra_information := 'while reading "rightprtok".';
      read_boolean(right_prt_ok);
      extra_information := 'while reading "leftposok".';
      read_boolean(left_pos_ok);
      extra_information := 'while reading "upposok".';
      read_boolean(up_pos_ok);
      extra_information := 'while reading "downposok".';
      read_boolean(down_pos_ok);
      extra_information := 'while reading "dataprotect".';
      read_boolean(data_protect);
      exp_char_protect := data_protect;
      exp_line_protect := data_protect
   end
end { read_bits } ;

{ Read the keyboard input sequences }

procedure read_input_sequences;                                                 { READ_INPUT_SEQUENCES }

var
   action_read, slash_read, recognised, input_sequences_done: boolean;
   input_sequence, reply: sequences;
   action: action_proc;
   repair: repair_type;
   ware: ware_type;
   ch_code: integer;
   sym_text: string12;

begin { read_input_sequences }
   input_sequences_done := false;
   extra_information := 'while reading input sequences.';
   input_sequence := init_sequence;
   reply := init_sequence;
   repeat
      action_read := false;
      slash_read := false;
      input_sequence.seq_len := 0;
      repeat
         read_char_symbol(sym_text, ch_code, false, recognised);
         if recognised then
         begin { it is a character }
            { cannot allow parity-off characters in input sequences because }
            { array is only 200b .. 377b }
            if ch_code < 200b
            then ch_code := ch_code + 200b;
            store_key_code(ch_code, input_sequence)
         end { it is a character }
         else if check_action(sym_text, action)
         then action_read := true
         else if sym_text = '//'
         then slash_read := true
         else
         begin { input symbol not recognised }
            writeln;
            writeln('Symbol "',sym_text,
                    '" found where a character or an action proc expected.');
            list_char_names;
            list_action_names;
            error_exit
         end { input symbol not recognised }
      until action_read or slash_read;
      if action_read then
      begin
         read_repair(repair);
         read_ware(ware);
         read_control_string(reply, false);
         if store_input_sequence(input_sequence, action, repair, ware, reply)
         then check_for_commands(input_sequence, action, repair, reply)
         else error_exit
      end
      else if input_sequence.seq_len = 0
      then input_sequences_done := true  { end of list }
      else
      begin
         writeln('// found when action name expected.');
         error_exit
      end
   until input_sequences_done;
   dispose_key_sequence(input_sequence)
end { read_input_sequences } ;

begin { read_one_source }
   line_no := 0;
   read_one_source := false;
   read_in_line;                        { read first line of file }
   skip_blank_lines;                    { skip to vdu ident }
   repeat
      vdu_type_count := vdu_type_count + 1;
      string_name := '            ';
      vdu_table_rec := init_table_rec;  { start with a clean sheet }
      empty_exp_record;                 { clear all vdu information }
      read_vdu_ident;                   { read various parts of table }
      read_control_chars;
      read_bits;
      read_input_sequences;
      if not eof(source_file)           { skip rest of input line }
      then read_in_line;
      if write_output_files then
      begin
         if not compress_vdu_info       { copy all into binary rec }
         then error_exit;
         write(binary_file, vdu_table_rec);  { write tables file }
         page(report_file);
         write_source_file
      end;
      skip_blank_lines
   until eof(source_file);
   read_one_source := true;
522:
end { read_one_source } ;

{ Read a batch of source files and process to output files }

procedure do_batch;                                                             { DO_BATCH }

label 222;

var
   done, translated_ok: boolean;

   procedure onunit(not_used: integer);

   begin { onunit }
      writeln;
      writeln('Please try again.');
      goto 222
   end { onunit } ;

begin { do_batch }
   interactive := false;
   write('Give output file name for binary vdu tables: ');
   readln(binary_fn);
   to_up_case(binary_fn);
   open(binary_file, binary_fn);        { ensure DAM file }
   rewrite(binary_file, binary_fn);     { ensure truncated }
   write('Give output report file name: ');
   readln(report_fn);
   to_up_case(report_fn);
   rewrite(report_file, report_fn);
   vdu_type_count := 0;
   done := false;
   writeln('Give input filename(s), one per line; just press RETURN to finish.');
   translated_ok := true;
   repeat
222:  write('Filename: ');
      readln(source_fn);
      to_up_case(source_fn);
      if length(source_fn) <= 0
      then done := true
      else
      begin
         on('ERROR_TRAP', onunit);      { trap any bad input filenames }
         reset(source_file, source_fn);
         revert('ERROR_TRAP');
         translated_ok := read_one_source(true);
         if not translated_ok
         then done := true
      end
   until done;
   if translated_ok then
   begin
      writeln('Execution finished with no errors.');
      writeln(vdu_type_count:1, ' vdu types processed.')
   end
end { do_batch } ;

begin { vdusetup_procs }
end { vdusetup_procs } .
